:- module(pac, [
		befun/0, efun/0,
		bekind/1, bekind/2, ekind/0,
		betrs/1, etrs/0,
		nopac/1,
		eval/2, eval/3,
		expand_pac/2,
		expand_arg/5,
		expand_arg_assert/2, expand_arg_compile/2,
		expand_basic_phrase/4,
		expand_basic_phrase/5,
		expand_core/5,
		expand_exp/7, expand_exp/8,
		expand_fun/2, expand_goal/5,
		expand_kind_rule/5, expand_phrase/5,
		compile_kind_rule/6,
		expand_query/3, flip_list/3, goal_to_pred/5,
		partial_args_match/2,
		phrase_to_pred/5, regex/3
	       ]).

:- use_module(pac(reduce)).
:- use_module(pac(basic)).
:- use_module(pac('expand-etc')).
:- use_module(pac('anti-subst')).
:- use_module(pac('odict-expand')).
:- use_module(pac(op)).
:- use_module(pac('expand-word')).
:- use_module(library(gensym)).

:- meta_predicate user:maplist(:).
user:maplist(_).

maplist_assert([]).
maplist_assert([X|Y]) :- assert_one(X),
						maplist_assert(Y).
%
assert_one(A:(B,C))	:-!, assert_one(A:B),
		 assert_one(A:C).
assert_one(A)		:- assert(A).

%
assert_in_user([]).
assert_in_user([X|Y]) :- assert_in_user_one(X),
	assert_in_user(Y).
assert_in_user_one(M:(B,C))	:- !,
	assert_in_user_one(M:B),
	assert_in_user_one(M:C).
assert_in_user_one((B,C))	:- !,
	assert_in_user_one(B),
	assert_in_user_one(C).
assert_in_user_one(M:(H:-B)) :-!, assert(user:H :- M:B).
assert_in_user_one((H:-B)) :-!, assert(user:H :- B).
assert_in_user_one(M:H:-B) :-!, assert(user:H :- M:B).
assert_in_user_one(H:-B) :-!, assert(user:H :- B).
assert_in_user_one(_:H) :- !, assert(user:H).
assert_in_user_one(H) :- assert(user:H).

% 2011 April, June, Kuniaki Mukai
% 2012 March, April
% 2014 June, September
% 2015 February

% [2015/03/01]  Argument order of expand_exp was changed by swap_args_of.

:- use_module(pac('odict-attr')).
:- use_module(pac('odict-expand')).

:-	meta_predicate(bekind(:)).
:-	meta_predicate(bekind(:,?)).
:-	meta_predicate(let(?,:,?)).
:-	meta_predicate(let_exp(?,:,?)).

%   compiles clauses in X.
pac_compile_aux_clauses(X) :- compile_aux_clauses(X).

		/*************************
		*         regex          *
		*************************/

%%  parse_string(+W:regex, +S:tring, -S0:string).
%
%   True if W is expand to qeury for parsing string S (minus S0).

parse_string(W, X, Y)				   :-
	once(phrase_to_pred(W, user, [U,V] :- G, P, [])),
	expand_dict_goal(G, G0),
	maplist(expand_dict_clause, P, Aux),
	maplist_assert(Aux),
	string_codes(X, U),
	call(G0),
	string_codes(Y, V).

%  parse_string(+W:regex, +S:codes, -S0:codes)  is nondet.
%
%   True if W is expand to qeury for parsing codes S (minus S0).

% ?- parse_codes(w(*(char(lower)) + *(char(digit))), `abc123`, Y).
% ?- parse_codes((w(+char(lower), A), w(+char(digit), B)), `abc123`, Y).
parse_codes(W, X, Y)				   :-
	once(phrase_to_pred(W, user, [U,V] :- G, P, [])),
	expand_dict_goal(G, G0),
	maplist(expand_dict_clause, P, Aux),
	maplist_assert(Aux),
	U=X,
	V=Y,
	call(G0).

%%  regex(+W:regex, +S:string, -S0:string) is nondet
%
%   True if W is expande to qeury for parsing string S (minus S0), that is,
%   a  version of phrase/3 for string.
%
%	?- regex(w(".*"), "abc", R).
%	?- regex(w("."), "abc", R).
%	?- regex(wl(".*"), "abc", R).
%	?- regex(wl(".*", A), "abc", R).

regex(X, Y, Z) :- once(expand_phrase(X, [], G, P, [])),
	maplist(expand_dict_clause, P, Aux),
	maplist_assert(Aux),
	string_codes(Y, Y0),
	phrase(G, Y0, Z0),
	string_codes(Z, Z0).

% ?-  pac:expand_query(user, maplist(pred([a]), [A]), G), call(G).
% ?-  pac:expand_query(pac, call(pred([a]), X), G), call(G).

		/***************************************************
		*     managing  begin-block of KIND definitions    *
		***************************************************/

save_back_quotes_flag :- current_prolog_flag(back_quotes, X),
	nb_setval(back_quotes_flag, X),
	set_prolog_flag(back_quotes, symbol_char).

restore_back_quotes_flag :- nb_getval(back_quotes_flag, X),
	set_prolog_flag(back_quotes, X).

%%	bekind(+X) is det.
%	Enable a kind block X.
bekind(X) :- bekind(X, []).

%% 	bekind(X:epac, O:options)  is det
%	Enable a kind block X with options O.
bekind(M:X, Opts)	:-
	nb_setval(pac_block, kind),
	nb_setval(pac_defsort_name, M:X),
	nb_setval(begin_options, Opts),
	nb_setval(rules, []),
	save_back_quotes_flag.

%%	ekind is det.
%	Disable kind block.

ekind	:-
	end_kind_basic,
	nb_setval(pac_defsort_name, []),
	nb_setval(rules, []),
	nb_setval(pac_block, []),
	restore_back_quotes_flag.
%
end_kind_basic :-  nb_getval(begin_options, Opts),
	(	memberchk(stop, Opts)
	->  EndOpts = [X= quote(X)]
	;	EndOpts =  []
	),
	end_kind_basic(EndOpts).

%
end_kind_basic(EndOpts) :-
	nb_current(pac_defsort_name, F), !,
	nb_getval(rules, Rs),
	once(kind_term(F, F1)),
	append(EndOpts, Rs, Ts),
	reverse(Ts, R0s),
	nb_getval(begin_options, Opts),
	(	memberchk(nonvar, Opts)
	->	R1s = [( A = [] :- var(A), !, fail)|R0s]
	;	R1s = R0s
	),
	maplist(expand_kind_rule(F1, Opts), R1s, R2s),
	ignore(pac_compile_aux_clauses(R2s)).

%
make_clause(sgn(F, N), (A = '`'(A)) :- true) :-
	functor(A, F, N).
make_clause(F/N, C)					:- make_clause(sgn(F, N), C).
make_clause(A, (X = '`'(X))			:- B):-  complete_args(A, [X], B).

%%	betrs(+X) is det.
%  Enable trs block X.
%  betrs : BEgin of Term Rewriting System

 :- meta_predicate betrs(:).
 :- meta_predicate betrs(:,?).

betrs(X)	   :- betrs(X, []).
%
betrs(X, Opts) :-
	nb_setval(pac_block, trs),
	nb_setval(pac_defsort_name, X),
	nb_setval(begin_options, Opts),
	nb_setval(rules, []),
	save_back_quotes_flag.

%%	etrs is det.
%  Disable trs block.
%  etrs : End of Term Rewriting System

etrs :-
	end_trs_basic,
	nb_setval(pac_defsort_name, []),
	nb_setval(rules, []),
	nb_setval(pac_block, []),
	restore_back_quotes_flag.

%
end_trs_basic :-
	nb_current(pac_defsort_name, F), !,
	nb_getval(rules, Rs),
	reverse(Rs, R0s),
	pac_etc:etc_trs(F, R0s, [], _, R1s, []),
	ignore(pac_compile_aux_clauses(R1s)).
%
end_trs_basic(_).

%%	befun  is det
%	Enable a function declaration mode.
befun :- nb_setval(pac_defsort_name, fun),
	nb_setval(pac_block, fun),
	nb_setval(begin_options, []),
	nb_setval(rules, []),
	save_back_quotes_flag.

% Toy version for arithmetic-like functional programming.
% :- befun.
%		A1 = B1.
%		....
%		An = Bn.
% :- efun.

efun :- nb_current(pac_defsort_name, fun),
	nb_getval(rules, Rs),
	reverse(Rs, R0s),
	maplist(expand_fun, R0s, R1s),
	ignore(pac_compile_aux_clauses(R1s)),
	nb_setval(pac_defsort_name, []),
	nb_setval(pac_block, []),
	nb_setval(rules, []),
	restore_back_quotes_flag.

% ?- pac:expand_fun(f(a,1)= g(a,a), Code).
%@ Code =  (f(a, 1, _23460)	:- (a(_23594), a(_23628)), g(_23594, _23628, _23460)).

% ?- pac:expand_fun(xadd(X+Y) = plus(xadd(X), xadd(Y)), R).
%@ R =  (xadd(X+Y, _24940) :- (xadd(X, _25096), xadd(Y, _25222)), plus(_25096, _25222, _24940)).

expand_fun(( H = B ), ( H0 :- G )):-
	complete_args(H, [V], H0),
	once(expand_fun(B, G, V)).

%
expand_fun({G},  (call(G) -> V = 1; V = 0), V).
expand_fun(if(E, E1, E2),
	   (F, (V == 1 -> F1, W = V1; F2, W = V2)),
	   W)			:-
				expand_fun(E, F, V),
				expand_fun(E1, F1, V1),
				expand_fun(E2, F2, V2).
expand_fun(is(X), V is X, V).
expand_fun(B, G, V)	:- once(args_inside_out(B, B0)),
	once(expand_exp(B0, =, V, [], G,  P, [])),
	( P \== [] ->  throw(error(expand_fun(B))); true ).


% test for eval/2.
% ?- qcompile(util(misc)).

% ?- eval(X, Y).
% ?- eval(:append@[1,2]@[3,4], Y).
% ?- eval(:append @ :append([1],[2]) @ :append([3],[4]), Y).
% ?- eval(:(=)@ hello, V).
% ?- eval((=)@ hello, V).
% ?- eval(:((=)@ hello), V).
% ?- eval((:(=)@ quote(hello)), V).
% ?- eval((:(=)@ (hello)), V).
% ?- eval(:((:(=))) @ quote(hello), V).  % error.
% ?- eval(=, V).
% ?- eval(a, V).
% ?- eval(append, V).
% ?- eval(append@a, V).
% ?- eval(append@a, V).
% ?- eval(append@a@b, V).
% ?- eval((append@a@b), V).
% ?- eval(pred([X,Y,Z])@a@b, V).
% ?- eval((pred([X,Y,Z]))@a@b, V).
%@ V = pac:'pac#12'(a, b).
% ?- eval((pred([X,Y,Z])), V).
% ?- eval(:pred([X,Y,Z])@a@b, V).
% ?- eval(plus@1@2, V).
% ?- eval((:plus)@1@2, V).
% ?- eval((plus)@1@2, V).
% ?- eval(:(#plus)@1@2, V).
% ?- eval(a@ b@ c, V).
% ?- eval(a@ quote(b)@ quote(c), V).
% ?- eval(a@ pred([X])@ pred([Y]), V).
% ?- eval(a@ (user: #pred([X]))@ (user: #pred([Y])), V).
% ?- eval(a@ ( pred([X]))@ (pred([Y])), V).
% ?- eval(a@ (user: pred([X]))@ (user: pred([Y])), V).
% ?- eval(a@ (pred([X]))@ (pred([Y])), V).
% ?- eval(a@ b @ c, V).
% ?- eval(a@ (=) @ (=), V).
% ?- eval(pred([X]), Y).
% ?- eval(:(((=))) @ hello, V).
% ?- eval(:(((=))) @ quote(hello), V).
% ?- eval(:(((=)))@ hello, V).
% ?- eval(:(pred([X,Y,Z] :-	append(Y,X,Z)))@[1,2,3]@[a,b,c], V).
% ?- eval(:((pred([X,Y,Z] :- append(Y,X,Z))))@[1,2,3]@[a,b,c], V).
% ?- eval(:maplist @ ([X]\ (:append(X) @ :append([a],[b]))) @ [[1,2],[3,4]], R).
% ?- eval(:(maplist) @ (fun([X]>>(:append(X) @ :append([a],[b])))), V).
% ?- eval(misc:set([1,2]+[3,4]), X).
% ?- eval((misc:set):: ((+) @ [1,2] @ [3,4]), X).
% ?- nopac(eval( ((misc:set) @ ((+) @ [1,2] @ [3,4])), X)).
% ?- eval( ([X]>>X)@1, Y).
% ?- eval( :([X]>>X)@1, Y).
% ?- eval((fun([X]>>X))@1, Y).
% ?- eval(:((fun([X]>>X)))@1, Y).
% ?- eval((fun([X]>>X))@1, Y).
% ?- eval((:fun([X]>>X))@1, Y).
% ?- eval(:maplist@(pred([X,a(X)]))@[1,2], Y).
% ?- eval(fun([X,F]>> (:F)@X)@1@(=), V).
% ?- eval(:fun([X,F]>> (:F)@X)@1@(=), V).
% ?- eval(:maplist @ pred([X,F,Y]:- call(F, X, Y)) @ [1,2] @ [=, succ], V).
% ?- eval(:maplist @ fun([X,F]>> (:F)@X) @ [1,2] @ [=, succ], V).
% ?- eval(:maplist @ fun([X,F]>> (:F)@X) @ [1,2] @ [=, succ], V).
% ?- eval(:maplist@ (fun([X]>> f(X)))@[1,2], Y).
% ?- eval(:maplist@ (fun([X]>> :succ(X)))@[1,2], Y).
% ?- eval(:maplist@ #(pred([X,X])) @[1,2], Y).
% ?- eval( fun([X]>>X ) @2, Y).
% ?- eval( :fun([X]>>X ) @2, Y).
% ?- eval( :((fun([X]->X))) @2, Y).
% ?- eval( :fun([A,B]>> (:fun([X,Y] >> (misc:set(X+Y))) @ A @ B))@[1,2]@[3,4], Y).

%%	args_inside_out(+T:term, -E:<pac-expression>) is det.
%   Convert T a functional expression to an pac expression
%   such that it is accepted for eval/2 test.

% ?- args_inside_out(1, X).
% ?- args_inside_out(A, X).
% ?- args_inside_out([1,2], X).
% ?- args_inside_out(quote(a), X).
% ?- args_inside_out(user:f(a, b), X).
% ?- args_inside_out(user:f(1, b), X).

% ?- args_inside_out(append([a, b], [c, d]), Exp), eval(Exp, V, [run]).
% ?- args_inside_out(append(append([a, a], [b, b]), append([c, c], [d, d])), Exp), eval(Exp, V, [run]).
%@ Exp = @(@(:(append), @(@(:(append), [a, a]), [b, b])), @(@(:(append), [c, c]), [d, d])),
%@ V = [a, a, b, b, c, c, d, d] .

args_inside_out(B, B) :- (var(B); number(B); is_list(B)).
args_inside_out(quote(B), quote(B)).
args_inside_out(@@(A), @@(A)).
args_inside_out('`'(A),'`'(A)).
args_inside_out(::(S, B), ::(S, B)).
args_inside_out(@(A, B), @(A, B)).
args_inside_out(B, G) :-
	(	B = M:B0 ->  F0 = M:F
	;	B0 = B,
		F0 = (:F)
	),
	B0 =.. [F|As],
	maplist(args_inside_out, As, Cs),
	args_inside_out(F0, Cs, G).

%
args_inside_out(F, [], F).
args_inside_out(F, [A|B], G) :- args_inside_out(F@A, B, G).

		/**********************************************
		*     expand conditional equations  (KIND)    *
		**********************************************/

% ?- kind_term(m:f//2, T).
% ?- kind_term(m:f, T).
% ?- kind_term(f//2, T).
kind_term(M:(F//N), M:FN) :- !, functor(FN, F, N).
kind_term(F//N, FN) :- !, functor(FN, F, N).
kind_term(X, X).
% ?- eliminate_with(m:f(A), with(B, b), R).
% ?- eliminate_with(f(A), with(B, b), R).
% ?- eliminate_with(m:f(A), with([B], b), R).

eliminate_with(_, R, R)	:- var(R), !.

eliminate_with(KTerm, with(A, B), B):-!,
	(	is_list(A) -> As = A
	;	As = [A]
	),
	(	KTerm = _:F -> true
	;	F = KTerm
	),
	F=..[_|Fs],
	append(As, _, Fs).  % passing args from left.
eliminate_with(_, X, X).

% ?- pac: expand_kind_rule(s(U), [], (A*A)= with(M, A+M), Y).
% ?- pac: expand_kind_rule(m:s(A), [], a = b, Y).
% ?- pac: expand_kind_rule(s(A), [], a = b, Y).
% ?- pac: expand_kind_rule(s(A), [], a = b, Y).
% ?- pac: expand_kind_rule(s(A), [], (a = b):-true, Y).
% ?- pac: expand_kind_rule(s(A), [], (a = b):-c, Y).
% ?- pac: expand_kind_rule(set(A), [], user, (1, 2) = 3, Y).
% ?- pac: expand_kind_rule(set(A), [], user, &(1, 2) = 3+4, Y).


% ?- expand_kind_rule(tm:set,[flip([2,1])],(_13664=[]:-tm:var(_13664),!), Y).
% ?- compile_kind_rule(tm:set,[flip([2,1])],(_13664=[]:-tm:var(_13664),!), Y, P, []).

expand_kind_rule(S, _, sgn(Sgn, A, F), Y) :-!,
	sgn_to_kind(Sgn, A, F, L = R),
	expand_kind_rule(S, L, R, true, Y).
expand_kind_rule(S, Options, X, Y) :-
	(	 X = (Head	:- Body) -> true
	;	 Head = X,
		 Body = true
	),
	(	Head = (L = R)
	->  expand_kind_rule(S, L, R, Body, Y0),
		(	S= _:S0 -> true
		;	S0 = S
		),
		functor(S0, F, _),
		(	memberchk(flip(Flip), Options)
		->	flip_clause(F, Flip, Y0, Y)
		;	Y = Y0
		)
	;	expand_only_goal(X, [], Y)
	).


%%  expand_kind_rule(+K:kind,+L:term,+R:exp,-G:goal, -C:clause)
%	is semidet
%
%   True if a given kind (a named group of conditional equations) rule
%
%			(L = R			:- G)
%
%   is converted to an equivelnt clause C.
%
%   Simplest usage:
% ?- pac:expand_kind_rule(word_am, X + Y, :am_concat@X@Y, true,  H :- G).
% ?- expand_kind_rule(word_am, X + Y, xargs([A, B, C] :- am_concat(A, B, C))@X@Y, true, H:-G).
% ?- expand_kind_rule(w, a, :(xargs([A, B, C] :- f@X@Y)), true, H:-G).
% ?- expand_kind_rule(w, a, (:xargs([A,V] :- f(A,V)))@X, true, H:-G).
% ?- expand_kind_rule(w, a, (xargs([A,V] :- f(A,V)))@X, true, H:-G).
% ?- expand_kind_rule(w, a, b, true,  H	:- G).
% ?- expand_kind_rule(w, a, b@c, true,  H :- G).
% ?- expand_kind_rule(w, a, b@c, true, H :- G).
% ?- expand_kind_rule(w, a, b, true, H	:- G).
% ?- expand_kind_rule(cond(A), X, with(S, S::X), true, R).
% ?- expand_kind_rule(cond(A), X;Y,  (;)  @X @ '`'(Y), true, R).
%
expand_kind_rule(F, _, _, _, _)	:- var(F), throw('variable kind 3').
expand_kind_rule(F, L, R, B, Cl):- nonvar(L), L= with(As, L0), !,
	expand_kind_rule(F, L0, with(As, R), B, Cl).
expand_kind_rule(F, L, R, Body, H:-G):-
	(	Body = true -> B = (!)
	;	B = Body
	),
	once(expand_goal(B, [], B0, P, P0)),
	slim_exp_goal(B0, B1),
	copy_term(F, F0),
	once(complete_args(F0, [L, V], H)),
	once(eliminate_with(F0, R, R0)),
	once(expand_exp(R0, F0, V, [], R1, P0, [])),
	slim_exp_goal((B1, R1), G),
	ignore(pac_compile_aux_clauses(P)).

% ?- pac:expand_exp(1, f(A), V, [], E, P, []).
% ?- pac:expand_exp(M, f(A), V, [], E, P, []).
% ?- pac:expand_kind_rule(a(M), b, with(A, c(A)), d, Cl).
% ?- nopac(pac:expand_kind_rule(a, b, @(pred([A,B]),b),  d, Cl)).

		/****************************************
		*     compile a conditional equation    *
		****************************************/

% ?- compile_kind_rule(s, [], a = b, Y, P, Q).
% ?- compile_kind_rule(s, [], a = b:-true, Y, P, Q).
% ?- compile_kind_rule(s, [flip([2,1])], a = b:-true, Y, P, Q).

compile_kind_rule(S, _, sgn(Sgn, A, F), Y, P, Q) :-!,
	sgn_to_kind(Sgn, A, F, L = R),
	compile_kind_rule(S, L, R, true, Y, P, Q).
compile_kind_rule(S, Options, X, Y, P, Q) :-
%	snap(compile_kind_rule(S, Options, X)),
	(	X = (Head :- Body) -> true
	;	Head = X,
		Body = true
	),
	(	Head = (L = R)
	->	once(compile_kind_rule(S, L, R, Body, Y0, P, Q)),
		functor(S, F, _),
		(	memberchk(flip(Flip), Options)
		-> 	flip_clause(F, Flip, Y0, Y)
		; 	Y = Y0
		)
	;	expand_only_goal(X, [], Y)
	).

% ?- compile_kind_rule(s, a, b, true, Y, P, []).
%
compile_kind_rule(F, _, _, _, _, _, _)  :- var(F), !, throw('variable kind 3').
compile_kind_rule(F, L, R, B, Cl, P, Q) :- nonvar(L), L=with(As, L0), !,
	compile_kind_rule(F, L0, with(As, R), B, Cl, P, Q).
compile_kind_rule(F, L, R, Body, (H :- G), P, Q):-
	(	Body = true -> B = (!)
	;	B = Body
	),
	once(expand_goal(B, [], B0, P, P0)),
	slim_exp_goal(B0, B1),
	copy_term(F, F0),
	once(complete_args(F0, [L, V], H)),
	once(eliminate_with(F0, R, R0)),
	once(expand_exp(R0, F0, V, [], R1, P0, Q)),
	slim_exp_goal((B1, R1), G).

% ?- trace.
% ?- sgn_to_kind([], 0, [], E).
% ?- sgn_to_kind(f, 3, g, E).
%
sgn_to_kind(S, 0, F, S = quote(F)):-!.
sgn_to_kind(S, N, F, L = R):- length(As, N),
	L =.. [S|As],
	mk_right(:(F), As, R).
%
mk_right(H, [],  H).
mk_right(R, [X|Xs], R0):- mk_right(R@X, Xs, R0).

		/*******************************************
		*     expansion of pac macros    (top)     *
		*******************************************/

%%  expand_pac(+X:clause, -Y:clause) is det

%    X is expanded to Y.
%
expand_pac(X, _)  :- none_pac_term(X), !, fail.
expand_pac(X, []) :- nb_current(pac_block, Block),
	Block \== [],
	!,
	( Block == kind
	->	nb_current(pac_defsort_name, F),
		strip_module(F, M, _),
		once(expand_only_goal(X, M, X0))
	;	X0 = X
	),
	nb_getval(rules, Rs),
	nb_setval(rules, [X0|Rs]).
expand_pac(X, Y)  :- once(expand_clause(X, Y)).

%
none_pac_term(X) :- var(X).
none_pac_term(begin_of_file).
none_pac_term(:-(_)).
none_pac_term(end_of_file).

%
expand_only_goal(H :- B, M, H:-B0):- !,
	once(expand_goal(B, M, B0, P, [])),
	ignore(pac_compile_aux_clauses(P)).
expand_only_goal(H, _,  H).



%%  expand_query(?M:module, +Q:goal, -Q0:goal) is det
%
%   Given top level query Q is expanded to Q0 at module M.

% ?- expand_query(user, maplist(pred([a]), [X,Y]), G), call(G).
% ?- expand_query(user, maplist(pred([{a:1}]), [X,Y]), G), call(G).
% ?- expand_query(user, X={a:1}, G), call(G).
% ?- expand_query(pac, maplist(pred([a]), [X,Y]), G), call(G).
% ?- expand_query(pac, dict(phrase(append@[a,b], [c,d], [])), G).

% Open dict aware query expansion.
% ?- expand_query(user, dict(phrase(append@[a,b], [c,d], X)), R).

% ?- phrase(append([a,b]), [c,d], X).
% ?-let(F, fun([X] >> (append@[a,b]@X))), phrase(F, [c,d], Y).
% ?-let(F, fun([X] >> (:append@[a,b]@X))), phrase(F, [c,d], Y).
% ?-let(F, fun([X] >> (append@[a,b]@X))), phrase(F, [c,d], Y).
% ?-let(F, fun([X] >> (append@[a,b]@X))), eval(F@[c,d], U).
% ?-let(F, fun([X] >> (append@[a,b]@X))), eval(:(F@[c,d]), U).
% ?-eval(:(:(fun([X] >> (append@[a,b]@X)))@[c,d]), U).
% ?-let(F, fun([X] >> (append@[a,b]@X))), eval((:F)@[c,d], U).

expand_query(_, [], []).
expand_query(M, M:[X|Y], M:[X|Y]).
expand_query(M, X, Y) :-
	once(expand_goal(X, M, Y, Z, [])),
	assert_in_user(Z).

		/************************
		*     displaying pac    *
		************************/

%% show(+X:term) is det
%
% ?- show(pred([X, X])).
% ?- show(phrase(w("."))).
% ?- show(phrase(wl("[a-zA-Z]+", A, []))).
% ?- show(phrase(wl("[^a-zA-Z]+", A, []))).
% ?- show(phrase(wl("[^b-wB-W]+", A, []))).

% ?- time(pac:(call(pred([1]), A))).
% ?- pac:time(call(pred([1]), A)).
% ?- pac:show(nopac(call(pred([X]), 1))).

show(X) :- let(F, X, Y), show(F, Y).

show(F, Y)			 :-
	predsort(compare_clause, Y, Y0),
	maplist(copy_term, Y0, Y1),
	numbervars(F, 0, _, [singletons(true)]),
	write(F),
	once(show_where_clause(Y1)).
%
show_where_clause([]).
show_where_clause(Y) :- write(', where\n'),
	maplist(numbervars_write(0), Y).

% ?- numbervars(f(X,Y), 0, _), term_string(f(X,Y), R, [numbervars(true)]).
%%  show(E:exp) is det
%
%   Display expanded form of an expression E.

% ?- show_exp(a).
% ?- show_exp(pred([1])).
% ?- show_exp(X\X).
% ?- show(X\X).
% ?- show_exp(#(X\X)@ #(hello)).
% ?- eval(#(X\X) @ #(hello), V).
% ?- eval((X\X) @ #(hello), V).
% ?- eval((X\X) @ quote(hello), V).
% ?- eval((X\X) @ hello, V).
show_exp(E)	:-
	expand_exp(E, =, V, [], G, P, []),
	maplist(copy_term, P, P0),
	numbervars((V,G), 0, _, [singletons(true)]),
	show_val_with_goal(V, G),
	show_where_clause(P0).
%
show_val_with_goal(V,true) :- write(V).
show_val_with_goal(V, G)   :- write(V),
	write("  with "),
	write(G).

% ?- show_phrase(w(".***********")).
% ?- show_phrase(wl("[^b-wB-W]+", A, [])).
% ?- show_phrase(wl("[^\s\n\t\r]++++++++", A, [])).
% ?- show_phrase(wl("[^\s\n\t\r]++++++++++++++++++", A, [])).
% ?- show_phrase(wl("[^\s\n\t\r]++++++++++++++++++++++++", A, [])).
% ?- show_phrase(wl("[^\s\n\t\r]+++++++++++++++++++++++++++++++++++", A, [])).
% ?- show_phrase(wl("[^\s\n\t\r]++++++++++++++", A, [])).
% ?- show_phrase(wl("[^\s\n\t\r]++++++++++++", A, [])).
% ?- show_am("[^\s\n\t\r]++++++++").
% ?- show_am("[^\s\n\t\r]******************************+").
% ?- show_phrase(wl("[^\s\n\t\r]", A, [])).
% ?- show_phrase(wl("[^\s\n\t\r]", A, [])).
% ?- show_phrase(wl("[\s\n\t\r]+", A, [])).
% ?- show_phrase(wl("[\s\n\t\r]+++", A, [])).
% ?- show_phrase(wl("[\s\n\t\r]+++++++++", A, [])).
% ?- show_phrase(wl("[a-zA-Z]++++++++++************", A, [])).
% ?- show_phrase(wl("[^a-zA-Z]++++++++++************++", A, [])).
% ?- show_phrase(wl("[^a-zA-Z]++++++++++************+++++++", A, [])).
% ?- show_phrase(wl("[^a-zA-Z]" ^ (>=(5)), A, [])).
% ?- show_phrase(wl("[^a-zA-Z]" ^ (3-5), A, [])).
% ?- show_am("[^a-zA-Z]" ^ (3-5)).
% ?- show_am("a" ^ (3-5)).
% ?- show_am("ab" ^ (3-5)).
% ?- listing(pac_word:word_am).
% ?- listing(set).

%% show(+X:phrase) is det
%
%   True if the  expanded phrase of pac X is displayed.

show_phrase(P) :- show(phrase(P)).

% ?- module(pac).
% ?- phrase_to_pred(apply(f), user, X, P, []).
% ?- phrase_to_pred(w(*(.)), user, X, P, []).
% ?- phrase_to_pred((a,b), user, X, P, []).
phrase_to_pred(X, M, [U,V] :- G, P, Q):-
	once(expand_phrase(X, M, G0, P, P0)),
	dcg_translate_rule('DUMMY' --> G0, 'DUMMY'(U, V) :- G1),
	once(expand_goal(G1, M, G, P0, Q)).

% ?- pac:expand_pac_pipe((a,b), [], user, R, P, []).
% ?- pac:expand_pac_pipe((a(A),b(B)), [A,B], user, R, P, []).
expand_pac_pipe(Pipe,  Globals, M, Ref, P, Q) :-
	phrase_to_pred(Pipe, M, Pac, P, P0),
	expand_core(pred(Globals, Pac), M, Ref, P0, Q).

% ?- goal_to_pred(call([V,V], [X,_]), [], U, P, []).
% ?- goal_to_pred(call(pred([V,V]), [X,_]), [], U, P, []).
goal_to_pred(G, M, FVs :- Body, P, Q):- once(expand_arg(G, M, Body, P, Q)),
	term_variables(Body, FVs).

numbervars_write(N, C) :- numbervars(C, N, _, [singletons(true)]), writeln(C).
%
numbervars_writeq(N, C)	:- numbervars(C, N, _, [singletons(true)]), writeq(C), nl.
%
compare_clause(C, X, Y)	  :- strip_module(X, _, H0:- _),
	strip_module(Y, _, K0 :- _),
	strip_module(H0, _, H),
    strip_module(K0, _, K),
	functor(H, FH, _),
	functor(K, FK, _),
	compare(C0, FH, FK),
	change_order(C0, C).

%
change_order(<, >).
change_order(>, <).
change_order(=, <).

		/*******************
		*     using pac    *
		*******************/

%%  let(-F:var, +X:goal)  is det
%
%   True if a F is  unified with expanded goal X.

:- meta_predicate user:let(?, :).

user:let(F, X) :- let(F, X, Y),
	 maplist(expand_dict_clause, Y, Aux),
	 maplist_assert(Aux).
user:show(X):- show(X).
user:show_phrase(X):- show_phrase(X).
user:show_exp(X):- show_exp(X).


%%  let(-F:var, +X:goal, -Y:clauses)  is det
%
%   True if a F is  unified with expanded goal X, and Y with
%   generated helper clauses.

%
let(Y, M:X, Z) :- 	once(expand_arg(X, M, Y, Z, [])).

% ?- pac:expand_core(pred(F, ([X] :- f(X)) & ([A]:- g(A))), [], G, P, []).
%%  let_exp(-X:var, +E:exp)  is det
%
%   True if X is  unified with expanded exp E.

let_exp(X, E) :- let_exp(E, X, Y),
	maplist(expand_dict_clause, Y, Aux),
	maplist_assert(Aux).

%%  let_exp(-X:var, +E:exp, Y:clauses)  is det
%
%   True if X is  unified with expanded exp E, and Y with
%   generated helper clauses.

let_exp(E, M:X, Y) :- expand_exp(E, =, M, X, Y, []).

		/**************************************
		*       Recognize block structures    *
		*     for possible use in the future  *
		**************************************/
%
begin_end(kind, bekind(_), ekind).
begin_end(kind, bekind(_, _), ekind).
begin_end(trs, betrs(_), etrs).
begin_end(fun, befun(_), efun).

% ?- module(pac).
% ?- structured_pac([ :- betrs(a),b,c,:-etrs], Y).
% ?- structured_pac([ :- bekind(a),b,c,:-ekind], Y).

%
structured_pac(X, Y) :- structured_pac(Y, X, []).

structured_pac([], [], []).
structured_pac([A|As], X, Y)			:-
	structured_pac_one(A, X, X0),
	structured_pac(As, X0, Y).
%
structured_pac_one(block(Kind, B, A), [	:- B|X], Y):-
	begin_end(Kind, B, Endname), !,
	structured_pac_block(Endname, A, X, Y).
structured_pac_one(A, [A|X], X).

%
structured_pac_block(End, [], [			  :- End|X], X):-!.
structured_pac_block(End, [X|R],[X|Y], Z) :-
	structured_pac_block(End, R, Y, Z).

		/**************************************************
		*     Eliminte pacs from a clause recursively.    *
		**************************************************/

%%  expand_clause(+X:goal, -Y:goal) is det
%
%   True if Y is unified with an expanded  goal of  X.

% ?- expand_clause(a :- call(pred([X]:-write(X)), hello), L).

expand_clause(X, Y) :- 	prolog_load_context(module, M),
	expand_clause(X, M, [Y|R], []),
	!,
 	ignore(pac_compile_aux_clauses(R)).
%
attach_prefix_to_clause(M, A :- B, (M:A) :- (M:B)):-!.
attach_prefix_to_clause(M, A, M:A).
%
expand_sgn(A, S/N-F, [T, R]:-(EvalArgs, apply(F0, YsR))) :- !,
			expand_arg_assert(F, F0),
			length(Xs, N),
			T =..[S|Xs],
			expand_sgn_calls(Xs, A, Ys, EvalArgs),
			append(Ys, [R], YsR).
expand_sgn(A, S-F, R)		 :- expand_sgn(A, S/0-F, R).

%
expand_sgn_calls([], _, [], true).
expand_sgn_calls([X|Xs], A, [Y|Ys], (call(A, X, Y), Rest)) :-
	expand_sgn_calls(Xs, A, Ys, Rest).

%%  expand_clause(+X:clause, -Y:module, -Z:clauses) is det.
%   True if Y is unified with an expanded  goal of  X,
%   and Z with generated helper clauses.

% ?- module(pac).
% ?- expand_clause(a		--> (X\X), R).
% ?- expand_clause(a		--> (X\X), R).
% ?- expand_clause((a		--> b,c), R).
% ?- expand_clause(a		--> w(".*"), R).
% ?- expand_clause((a		--> w(".*"), (X\[X,X])), R).
% ?- expand_clause(a		--> [a,b], R).
% ?- expand_clause(a		--> w(*char(kanji)), R).
% ?- expand_clause(a({b:1}) --> [], R).
% ?- expand_clause(a({b:1}) --> c({d:2}), R).

% a tiny helper for updating module prefix
update_prefix(M, N, N0)	:-
	(  M == []
	->	N0 = N
	;	N0 = M
	).
%
sgn_subst(A+B, A0+B0, F) :- sgn_subst(A,A0,F), sgn_subst(B, B0,F).
sgn_subst(A*B, A0*B0, F) :- sgn_subst(A,A0,F), sgn_subst(B, B0,F).
sgn_subst(\(A,B), \(A0,B0), F) :- sgn_subst(A, A0, F), sgn_subst(B, B0, F).
sgn_subst(A, A, _)		 :- is_list(A), !.
sgn_subst(sgn(A), A, _)	 :- !.
sgn_subst(A, A0, F)		 :- call(F, A, A0).

% ?- zip_algebra(([a-1]+[b-2])\ [b-2], R).
zip_algebra(A+B, C)	  :- zip_algebra(A, A0),
					  zip_algebra(B, B0),
					  append(A0, B0, C).
zip_algebra(\(A, B), C) :- zip_algebra(A, A0),
					  zip_algebra(B, B0),
					  subtract(A0, B0, C).
zip_algebra(A * B, C) :- zip_algebra(A, A0),
					  zip_algebra(B, B0),
					  intersection(A0, B0, C).
zip_algebra(X, X).

% ?- pac:phrase_ref(A, (a,b(A),c), P).
% ?- pac:phrase_ref_show(A, (a,b(A),c), P).
% ?- pac:phrase_list_ref(A, [a,b(A),c], P).
% ?- pac:phrase_list_ref_show(A, [a,b(A),c], P).

% Expand an phrase and return the ref handle

% ?-  trace, pac:phrase_ref([A,B], (append(A), append(B)), G).
phrase_ref(X, Y, X0) :- translate_phrase_to_clause(X, Y, X0, Cs),
	maplist(expand_dict_clause, Cs, Aux),
	maplist_assert(Aux).

% ?-  pac:phrase_ref_show([A,B], (append(A), append(B)), G).
phrase_ref_show(X, Y, X0) :- translate_phrase_to_clause(X, Y, X0, Cs),
			    show(X0, Cs).

%%	translate_phrase_to_clause(+Globals, +Phrase:phrase, -PhraseRef, -Cs) is det.
%
%       Translate a DCG phrase Phrase with pac preds and regex to
%	an equivalent list Cs of Prolog clauses and a reference PhraseRef which
%	has the global variables appearing in Globals.

% ?- pac:translate_phrase_to_clause([A,B], (append(A), append(B)), G, Cs),
% maplist(assert, Cs), A=[a,b], B=[c,d], call(G, [1,2], V).
% ?- pac:translate_phrase_to_clause([A,B], (append(A), append(B)), G, Cs).

translate_phrase_to_clause(Globals, Phrase, PhraseRef, Cs) :-
	once(expand_phrase(Phrase, [], Phrase0, Aux, [])),
	new_pac_name(Pname),
	canonical_global(Globals, Canonical_Globals),
	make_ref(Canonical_Globals, Pname, PhraseRef),
	dcg_translate_rule(PhraseRef --> Phrase0, C),
	expand_clause(C, C0s),
	append([C0s], Aux, Cs).

%
make_ref([], R, R) :- !.
make_ref(X, R, R0) :- complete_args(R, X, R0).

% Expand an phrase of the list form and return the ref handle
phrase_list_ref(X, Y, X0)	   :- period_to_comma(Y, Y0),
	phrase_ref(X, Y0, X0).
%
phrase_list_ref_show(X, Y, X0) :- period_to_comma(Y, Y0),
	phrase_ref_show(X, Y0, X0).

% ?- module(pac).
% ?- period_to_comma([a,b,c], R).
period_to_comma([X], X).
period_to_comma([X, Y|Z], (X,R)) :- period_to_comma([Y|Z], R).

% for not expansion as builtin dicts. (To be revised. )
is_role(role(X, Y), X, Y).
is_role(P, X, Y) :- P=..[(.), X, Y].

%%  expand_goal(+G:goal, +M:atom, -G0:goal, -P:list, -Q:list) is det
%
%   True if G is converted to G0  and the generated helper predicates in
%   P (minus Q)  so that G is equivalent to G0 with P (minus Q).
%
%   Simples usage:
%	?- pac:expand_goal((call(pred([a]), X), write(X)),  user, G,  L, []).
%	G = (user:'pred#2'(X), user:write(X)),
%	L = [ (user:'pred#2'(a)	:- true)]

% ?- expand_goal((call([X]\[X,X],1, V), Y=V), user, G,  L, []).

% expand_goal(X, M, MX, P, P)			   :- var(X), attach_prefix(M, X, MX).
expand_goal(X, _, X, P, P)			   :- var(X),!.
expand_goal(M:X, N, Y, P, Q)		   :- update_prefix(M, N, N0),
	expand_goal(X, N0, Y, P, Q).
expand_goal((X,Y), M, (X0, Y0), P, Q)  :-
	expand_goal(X, M, X0, P, P0),
	expand_goal(Y, M, Y0, P0, Q).
expand_goal((X;Y), M, (X0;Y0), P, Q)   :-
	expand_goal(X, M, X0, P, P0),
	expand_goal(Y, M, Y0, P0, Q).
expand_goal((X->Y), M, (X0->Y0), P, Q) :-
	expand_goal(X, M, X0, P, P0),
	expand_goal(Y, M, Y0, P0, Q).
expand_goal(if(A, B, C), M, Y, P, Q)   :- once(expand_goal_if_cond(A, A0)),
	expand_goal((A0->B;C), M, Y, P, Q).
expand_goal(if(A, B), M, Y, P, Q)	   :- expand_goal(if(A,B,true), M, Y, P, Q).
expand_goal(??(A), _, true, P, P)	   :- once(A).
expand_goal(\+(X), M, \+(Y), P, Q)	   :- expand_goal(X, M, Y, P, Q).
expand_goal(nopac(X), _M, X, P, P).
expand_goal(X@Y, M, G, P, Q)		   :- expand_atmark_goal(X@Y, M, G, P, Q).
expand_goal(X, M, MX, P, P)			   :- is_list(X), attach_prefix(M, X, MX).
	% For not loading files.
expand_goal(X, M, G, P, Q) :- once(expand_basic(X, M, G, P, Q)).

%
expand_atmark_goal(X@Y, Mod, G, P, Q):-
	expand_exp(X@Y, [], =, G_atomic, Mod, G_aux, P, Q),
	slim_goal((G_aux, G_atomic), G).

% For var X, if(X, A, B)  => (X==true -> A; B)
% otherwise,  if(X, A, B) => (X -> A; B).
% tiny optimization to prevent from (call(<module-ref>:X) -> A; B).
expand_goal_if_cond(X,  X == true) :- var(X).
expand_goal_if_cond(X,  X).

%
expand_phrase_list([], _, [], P, P).
expand_phrase_list([A|Xs], M, [A|Ys], P, Q)			   :- var(A),
	expand_phrase_list(Xs, M, Ys, P, Q).
expand_phrase_list([phrase(A, B)|Xs], M, [Y|Ys], P, Q) :-
	once(expand_core(phrase(A, B), M, Y, P, R)),
	expand_phrase_list(Xs, M, Ys, R, Q).
expand_phrase_list([phrase(A)|Xs], M, [Y|Ys], P, Q)	   :-
	once(expand_core(phrase([], A), M, Y, P, R)),
	expand_phrase_list(Xs, M, Ys, R, Q).
expand_phrase_list([A|Xs], M, [X0|Ys], P, Q)		   :-
	once(expand_phrase(A, M, X0, P, R)),
	expand_phrase_list(Xs, M, Ys, R, Q).

%%  expand_core(+Pac:pac, +M:atom, -G:pred, -P:list, -Q:list) is det
%
%   True if Pac is converted to a fresh predicate G in P (minus Q)
%   at a module M.
%
%   Simplest usage:
%	?-  expand_core(pred([X, Y]	:- X\==Y), user, G, P, []).
%	G = user:'pred#1',
%	P = [ (user:'pred#1'(X, Y)	:- user: (X\==Y))]

% ?- expand_core(X\(X\X), user, R, P, []).
% ?- expand_core(X\ #(X\X), user, R, P, []).
% ?- expand_core(X\X, user, R, P, []).
% ?- let(H, rec(F, [], ([[],X,X] :- true)&([[X|Y],Z,[X|U]]:- call(F, Y, Z, U)))), eval(H@[1,2]@[3,4], R).
% ?- new_pac_name([A], V).
% ?- call(pipe((=, =, =)), hello, X).
% ?- maplist(pipe((=, =, =)), [a,b,c], X).
% ?- maplist(pipe((pipe((=, =)), =)), [hello], X).
% ?- maplist(pipe([A, B], (append([A]), append([B]))), [[1],[2],[3]], R),  A = hello, B=world.

% [2015/04/02] added mutual recursion with globals.

expand_core(pred(Global, Cs), M, G, P, Q) :-
	once(new_pac_name(Global, FC)),
	attach_prefix(M, FC, G),
	once(expand_pac_clauses(Cs, M, G, P, Q)).
expand_core(pred(Cs), M, G, P, Q)		  :-
	expand_core(pred([], Cs), M, G, P, Q).
expand_core(global(C, Cs), M, G, P, Q)	  :-
	expand_core(pred(C, Cs), M, G, P, Q).
expand_core(mrec(C, Defs), M, G, P, Q)	  :- Defs = [A = _|_],
	canonical_global(C, C0),
	maplist(recursive_pac_name(C0), Defs),
	attach_prefix(M, A, G),
	expand_mrec_system(Defs, M, P, Q).
expand_core(mrec(Defs), M, G, P, Q)		  :-
	expand_core(mrec([], Defs), M, G, P, Q).
expand_core(rec(F, C, Cs), M, G, P, Q)	  :-
	expand_core(mrec(C, [ F=pred(Cs) ]), M, G, P, Q).
expand_core(rec(F, Cs), M, G, P, Q)		  :-
	expand_core(mrec([ F = pred(Cs) ]), M, G, P, Q).
expand_core(E, M, G, P, Q)	 :- normal_fun(E, =, E0), !,
	expand_fun_to_pred(E0, M, G, P, Q).
expand_core(pipe(E), M, G, P, Q)		  :- expand_pac_pipe(E, [], M, G, P, Q).
expand_core(pipe(Vs, E),  M, G, P, Q)	  :- expand_pac_pipe(E, Vs, M, G, P, Q).
%
expand_pac_rule_body(_, [], [], P, P).
expand_pac_rule_body(M, [H :- B|Ns],[H:-B0|N0s], P, Q):-
 	once(expand_goal(B, M, B0, P, P0)),
	expand_pac_rule_body(M, Ns, N0s, P0, Q).
%
recursive_pac_name([],  F = _) :- new_pac_name(F).
recursive_pac_name(C,   F = _) :- new_pac_name(F0),
	complete_args(F0, [C], F).

%
expand_mrec_system([A = Pred|R], M, P, Q) :-
	attach_prefix(M, A, PacRef),
	expand_pred_to_ref(Pred, M, PacRef, P, P0),
	expand_mrec_system(R, M, P0, Q).
expand_mrec_system([], _, P, P).

%
expand_pred_to_ref(pred(Cs), M, PacRef, P, Q) :-
	expand_pac_clauses(Cs, M, PacRef, P, Q).
expand_pred_to_ref(Cs, M, PacRef, P, Q)		  :-
	expand_pac_clauses(Cs, M, PacRef, P, Q).

%
expand_pac_clauses(Cs, M, PacRef, P, Q)	:-
	flat(&, Cs, Ds),
	maplist(canonical_pred, Ds, Ns),
	expand_pac_rule_body(M, Ns, N0s, P, P0),
	maplist(slim_clause(PacRef), N0s, Es),
	append(Es, Q, P0).

%
funs_to_preds(A&B, S, M, A0&B0, P, Q)	:-
	funs_to_preds(A, S, M, A0, P, P0),
	funs_to_preds(B, S, M, B0, P0, Q).
funs_to_preds(A, S, M, A0, P, Q) :- once(fun_to_pred(A, S, M, A0, P, Q)).

%
fun_to_pred(A >> R, S, M, G, P, Q)	:- \+ is_list(A),
	fun_to_pred([A] >> R, S, M, G, P, Q).
fun_to_pred(L >> R, S, M, L0 :- R0, P, Q):- append(L, [V], L0),
	once(expand_exp(R, [], S, V, M, R0, P, Q)).

%
slim_clause(MFC, Xs	:- B, H:-B0):- once(complete_args(MFC, Xs, H)),
	reduce:slim_goal(B, B0).

%
canonical_pred(X, Y) :-
	(	is_list(X)	-> Y = (X:-true)
	;	nonvar(X), X = (H:- _), is_list(H) -> Y = X
	;	throw('invalid pred clause'(X))
	).

% ?- pac:new_pac_name(N).
gensym_pac_name(N) :- nb_current(pac_name_prefix, N),
		 N \== [],
		 !.
gensym_pac_name('pac#').
% gensym_pac_name('pred#').


new_pac_name(N)	:- var(N), !,
		gensym_pac_name(G),
		gensym(G, N).
new_pac_name(_).

% ?- pac:new_pac_name(A, R).
% ?- pac:new_pac_name(A*B, R).
% ?- A=a, pac:new_pac_name(A+A, R).
new_pac_name(Global, PacRef) :- new_pac_name(Global, _, PacRef).

%
new_pac_name(_, _, PacRef)		   :- nonvar(PacRef), !.
new_pac_name(Global, Name, PacRef) :- term_variables(Global, Vs),
    once(canonical_global(Vs, Vs0)),
	new_pac_name(Name),
   	PacRef =.. [Name|Vs0].
%
canonical_global(X, [X]):- var(X), !.
canonical_global([X|Y], [X|Y]).
canonical_global([], []).
canonical_global(X, [X]).

% expand_arglist/6.
expand_arglist([], _, [], true, P, P).
expand_arglist([A|As], M, [B|Bs],(H,H0), P, Q) :- expandable_meta_arg(A),
	once(expand_meta_arg(A, M, B, H, P, P0)),
	expand_arglist(As, M, Bs, H0, P0, Q).
expand_arglist([A|As], M, [A|Bs], H, P, Q)	   :-
	expand_arglist(As, M, Bs, H, P, Q).

% expand_arglist/5
expand_arglist([], _, [], P, P).
expand_arglist([A|As], M, [B|Bs], P, Q)	:- expandable_meta_arg(A),
	once(expand_arg(A, M, B, P, P0)),
	expand_arglist(As, M, Bs, P0, Q).
expand_arglist([A|As], M, [A|Bs], P, Q)	:-
	expand_arglist(As, M, Bs, P, Q).

%  Variables and applications are not expandable.
expandable_meta_arg(A) :- nonvar(A), \+ number(A).


% ?- call([X]\ (:X@ 1), =, V).
% ?- call([X]\ (X@ 1), =, V).
% ?- call(fun([X]-> (:X@ 1)), =, V).
% ?- call(fun([X]-> (X@ 1)), =, V).

% ?-  pac:meta_call(f([1,?,?]), f([a,b,c]), A, B, Y, U, V).
meta_call(I, X, A, B, Y, SF, SFV) :- I =..[F,I0],
	X=..[_, X0],
	Y=..[F, Y0],
	once(collect_calls(I0, X0, Y0, A, B, SF, SFV)).

% ?- pac:collect_calls([1, ?, ?], [a, b, c], A, B, C, D, E).
collect_calls(_, [], [], [], [], [], []).
collect_calls([:|X], [A|Y], [B|Z], [A|U], [B|V],  P, Q)	:-
	collect_calls(X, Y, Z, U, V, P, Q).
collect_calls([^|X], [A|Y], [B|Z], [A|U], [B|V],  P, Q)	:-
	collect_calls(X, Y, Z, U, V, P, Q).
collect_calls([//|X], [A|Y], [B|Z], U, V,  [A|P],[B|Q])	:-
	collect_calls(X, Y, Z, U, V, P, Q).
collect_calls([N|X], [A|Y], [B|Z], [A|U], [B|V], P, Q)	:- integer(N),
	collect_calls(X, Y, Z, U, V, P, Q).
collect_calls([_|X], [A|Y], [A|Z], U, V, P, Q)			:-
	collect_calls(X, Y, Z, U, V, P, Q).

		/***********************
		*     expand phrase    *
		***********************/

%%  expand_phrase(+G:phrase, +M:atom, -G0:phrase, -P:list, -Q:list) is det
%
%   True if G is converted to G0 and helper predicates in P (minus Q)
%   at a module M.
%
%   Simplest usage:
%	?- expand_phrase((pred([X, f(X)]), pred(U, g(U))), user, G,  L, []).
%	G = (user:'pred#3', user:'pred#4'(U)),
%	L = [ (user:'pred#3'(X, f(X)) :- true), (user:'pred#4'(U, g(U)):-true)] .

% ?- module(pac).
% ?- expand_phrase((call([X]\[X,X],1, V), Y=V), user, G,  L, []).
% ?- expand_phrase(w(".*"), user, G,  L, []).
% ?- expand_phrase(wl(".*"), user, G,  L, []).
% ?- expand_phrase(maplist(phrase(wl(".*"))), user, G,  L, []).
% ?- phrase(append@[a], [b,c], R).
% ?- phrase(pred([X,Y] :- append([a], X, Y)), [b,c], R).
% ?- phrase(fun([X] -> :append([a, b], X)), [c,d], R).
% ?- phrase(pred([X, Y] :- append([a,b])@X@Y), [c, d], R).
% ?- let(F, pred([X,Y,Z] :- append(X, Y, Z))), phrase(F@[a], [b,c], V).
% ?- let(F, (pred([X,Y,Z] :- append(X, Y, Z)))).
% ?- let(F, pred([X,Y,Z] :- append(X, Y, Z))), phrase(F@[a], [b,c], V).

expand_phrase(X, _, X, P, P)			 :- var(X), !.
expand_phrase(M:X, N, Y, P, Q)			 :- update_prefix(M, N, N0),
	expand_phrase(X, N0, Y, P, Q).
expand_phrase((X,Y), M, (X0, Y0), P, Q)	 :-
	expand_phrase(X, M, X0, P, P0),
	expand_phrase(Y, M, Y0, P0, Q).
expand_phrase((X|Y), M, G, P, Q)		:- expand_phrase((X;Y), M, G, P, Q).
expand_phrase((X;Y), M, (X0;Y0), P, Q)	 :-
	expand_phrase(X, M, X0, P, P0),
	expand_phrase(Y, M, Y0, P0, Q).
expand_phrase((X->Y), M, (X0->Y0), P, Q) :-
	expand_phrase(X, M, X0, P, P0),
	expand_phrase(Y, M, Y0, P0, Q).
expand_phrase(\+(X), M, \+(Y), P, Q)	 :- expand_phrase(X, M, Y, P, Q).
expand_phrase({E}, M, {E0}, P, Q)		 :- once(expand_goal(E, M, E0, P, Q)).
expand_phrase(X@Y, M, ({H}, G), P, Q)	 :-
	expand_meta_arg(X@Y, M, G, H, P, Q).
expand_phrase(X, _, X, P, P)			 :- is_list(X).  % for not loading files.
expand_phrase(X, _, Y, P, P)			 :- string(X), string_codes(X, Y).
expand_phrase(bind(X), M, bind(Y), P, Q) :- % for bind_bind
	expand_phrase(X, M, Y, P, Q).
expand_phrase(shift(X), _, {shift(Y)}, P, Q) :-
	expand_arg(X, [], Y, P, Q).
expand_phrase(sed(X), M, Y, P, Q)		 :- must_be(nonvar, X),
	expand_sed(X, [F, W, A]), !,
	expand_recognize_act(F, W, A, M, Y, P, Q).
expand_phrase(X, M, G, P, Q)			 :- once(expand_basic_phrase(X, M, G, P, Q)).

% ?- expand_basic_phrase(maplist(maplist(phrase(wl(".")))), user, G, P, []).

% Let ^ and >> be associative with each other.
expand_sed((F^W)>>A,	[F, W, A]).
expand_sed(F^(W >> A),	[F, W, A]).
expand_sed(W  >> A,		[[], W, A]).
expand_sed(s/Regex/S,	[[], w(Regex),S]).
expand_sed(sl/Regex/S,	[[], wl(Regex),	S]).
expand_sed(a/Regex/B,	[[], w(Regex, X), (X + B)]).
expand_sed(al/Regex/B,	[[], wl(Regex, X), (X + B)]).
expand_sed(b/Regex/B,	[[], w(Regex, X), (B + X)]).
expand_sed(bl/Regex/B,	[[], wl(Regex, X),	(B + X)]).
expand_sed(d/Regex, X)	:- expand_sed(s/Regex/ "", X).
expand_sed(dl/Regex, X)	:- expand_sed(sl/Regex/ "", X).
expand_sed(w/Regex/Before/After,  [[], w(Regex, X),  (Before + X + After)]).
expand_sed(wl/Regex/Before/After, [[], wl(Regex, X), (Before + X + After)]).

% ?- expand_phrase(bind(a), [], G, P, []).
warning_w(E) :- write( 'non ground compound regex found: '),
	writeln(E).

% ?- eval(X@Y, V, [goal(G)]), X=(=), Y=1, call(G).
% ?- eval(X@Y@Z, V, [goal(G)]), X=(:append), Y=[a,b], Z=[c, d], call(G).
% ?- expand_phrase(w("a+b"), user, G, P, []).
% ?- expand_phrase(w("a"), user, G, P, []).
% ?- expand_phrase(w("aa"), user, G, P, []).
% ?- expand_w("a", user, _G5350, _G5351, []).

expand_basic_phrase(X, Y, Z, U) :-
	once(expand_basic_phrase(X, Y, Z, U, [])).

%
expand_basic_phrase(w(E), _, w(E), P, P)	 :- var(E).
expand_basic_phrase(w(E), M, Y, P, Q)		 :-
	expand_w(E, M, Y, P, Q).
expand_basic_phrase(w(E, A), M, Y, P, Q)	 :-
	expand_w(E, A, [], M, Y, P, Q).
expand_basic_phrase(w(E, A, B), M, Y, P, Q)	 :-
	expand_w(E, A, B, M, Y, P, Q).
expand_basic_phrase(wl(E), _, wl(E), P, P)	 :- var(E).
expand_basic_phrase(wl(E), M, Y, P, Q)		 :-
	expand_wl(E, M, Y, P, Q).
expand_basic_phrase(wl(E, A), M, Y, P, Q)	 :-
	expand_wl(E, A, [], M, Y, P, Q).
expand_basic_phrase(wl(E, A, B), M, Y, P, Q) :-
	expand_wl(E, A, B, M, Y, P, Q).
% '::' is obsolete  [2015/03/04]
expand_basic_phrase(S::E, M, F, P, R)		 :-
	once(expand_exp(E, S, V, M, G, P, Q)),
	term_variables(S-E, Vs),
	once(expand_core(pred(Vs, [V] :- G), M, F, Q, R)).
expand_basic_phrase(X, M, G, P, Q) :- once(expand_basic(X, M, G, P, Q)).

% ?- module(pac).
% ?- expand_basic(maplist(phrase(wl("."))), user, G, P, []).
% ?- expand_basic(maplist(=(a), []), user, G, P, []).
% ?- expand_basic_phrase(maplist(phrase(wl("."))), user, G, P, []).
% ?- predicate_property(profile(a=b),  meta_predicate(P)).
% ?- predicate_property(call(a=b, b), meta_predicate(P)).
% ?- predicate_property(call(a=b, b,c), meta_predicate(P)).
% ?- pac:expand_basic(maplist(phrase(=), [[a]], R), user, G, P, []).
% ?- pac:expand_basic(maplist(phrase(wl(".")), [[a]], R), user, G, P, []).
% ?- pac:expand_basic(maplist(phrase(wl(".")), [[a]], R), user, G, P, []).
% ?- predicate_property(xxx:maplist(phrase(=), X, Y), meta_predicate(U)).

% ?- expand_basic(call(X, Y), user, G, P, []).
% ?- expand_basic(setup_call_cleanup(X, Y, Z), user, G, P, []).
% ?- expand_basic(setup_call_cleanup(pred([]):-true, Y, Z), user, G, P, []).

expand_basic(E, M, G, P, Q)	:- 	expand_core(E, M, G, P, Q)
	;	expand_etc(E, M, G, P, Q)
	;	expand_atomic_goal(E, M, E0, H, P, P0),
		expand_meta_args(E0, M, G0, P0, Q),
		slim_goal((H, G0), G).

% ?- meta_property(user:call(A@B, C), U).
% ?- expand_meta_args(user:call(X@Y), user, U, P, []).
expand_meta_args(X, _, X, P, P)   :-  var(X), !.
expand_meta_args(M:X, _, Y, P, Q)   :-!, expand_meta_args(X, M, Y, P, Q).
expand_meta_args((X,Y), M, (X0, Y0), P, Q)  :- !,
	expand_meta_args(X, M, X0, P, P0),
	expand_meta_args(Y, M, Y0, P0, Q).
expand_meta_args((X;Y), M, (X0;Y0), P, Q)   :-!,
	expand_meta_args(X, M, X0, P, P0),
	expand_meta_args(Y, M, Y0, P0, Q).
expand_meta_args(E, M, G, P, Q):-
	meta_property(M:E, U), !,
	U=..[_|As],
	E=..[E0|Bs],
	expand_meta_arg_list(As, Bs, M, Cs, Aux, P, Q),
	G0=..[E0|Cs],
	slim_goal((Aux, M:G0), G).
expand_meta_args(E, M, Em, P, P):-attach_prefix(M, E, Em).
%
indicator_for_meta(:):-!.
indicator_for_meta(A):- integer(A).
%
expand_meta_arg_list([], _, _, [], true, P, P):-!.
expand_meta_arg_list([A|As], [B|Bs], M, [C|Cs], (Pre, Pre0), P, Q):-
	indicator_for_meta(A),
	expand_meta_arg(B, M, C, Pre, P, P0), !,
	expand_meta_arg_list(As, Bs, M, Cs, Pre0, P0, Q).
expand_meta_arg_list([_|As], [B|Bs], M, [B|Cs], Pre, P, Q):-
	expand_meta_arg_list(As, Bs, M, Cs, Pre, P, Q).
%
 :- meta_predicate expand_arg_assert(:, ?).
 :- meta_predicate expand_arg_compile(:, ?).
expand_arg_assert(X, Y)				  :- expand_arg(X, [], Y, P, []),
 	ignore(maplist(assert, P)).
expand_arg_compile(X, Y)			  :- expand_core(X, [], Y, P, []),
 	ignore(pac_compile_aux_clauses(P)).
%
expand_meta_arg(X, _, X, true, P, P) :- var(X), !.
expand_meta_arg(M:X, _, M:X, true, P, P) :- var(X), !.  %  ad hoc ?
expand_meta_arg(X@Y, M, U, H, P, Q)	  :-!,
	expand_exp(X@Y, [], =, U, M, H, P, Q).
expand_meta_arg(M:X, N, Y, H, P, Q)	  :- update_prefix(M, N, N0),
	expand_meta_arg(X, N0, Y, H, P, Q).
expand_meta_arg(X, _, Y, true, P, P)  :- (string(X); is_list(X)), !, Y = X.
expand_meta_arg((X,Y), M, (X0, Y0), (G0, H0), P, Q)  :-
	expand_meta_arg(X, M, X0, G0, P, P0),
	expand_meta_arg(Y, M, Y0, H0, P0, Q).
expand_meta_arg(X, M, Z, true, P, Q)  :- once(expand_basic(X, M, Z, P, Q)).

% ?- disable_pac_query.
% ?- eval(:(fun([X]>>X)) @2, R).
% ?- eval(:pred([X,X]) @2, R).
% ?- pac:expand_exp(fun([X]>>X), [], =, U, user, H, P, []).
expand_arg(X, _, X, P, P)	:-   var(X), !.
expand_arg(X, M, G, P, Q)	:-
	(	expand_meta_arg(X, M, G, H, P, Q),
		H = true
	->	true
	;	throw('NON ATOMIC meta argument found.')
	).
%
canonical_exp(X) :- ( number(X); string(X); is_list(X) ).

%%	eval(+E:exp, -V, +O:options) is det.
%	Convert E to a goal G (see below) such that  execution of G
%	makes V be unified with the value.
%
%   options in O:
%	*	goal(G) :  G is a goal whose execution gives the value V of
%				   the expression E.
%	*	pac(X-Y):  generated helper predicates in the form of d-list X-Y.
%	*	inside_out(true/false) : if true given, subterms of E
%			are evaluated recursively unless it is a expression with
%			a kind prefix like <kind>::<expression>.
%	*	run(true/false): 'true' is default, and then run the goal G.

% ?- eval( 1, V).
% ?- eval( =(append(=([a,b]), =([1,2]))), V, [goal(G)]).
% ?- eval( =(append(=([a,b]), =([1,2]))), V).
% ?- eval(append@[a,b]@[1,2], V).
% ?- eval(append@(append@[a]@[b])@[1,2], V).
% ?- eval(append(append([a,b], [c,d]), append([a,b], [c,d])), V).
% ?- eval(append@(append([a,b])@(append([c,d])@ append([a,b], [c,d])))@[x,y,z], V).
% ?- eval(append@(append([a,b])@(append([c,d])@ append([a,b], [c,d])))@[x,y,z], V).
% ?- eval(pred([X,X])@ #a, V).
% ?- eval(quote(f(a))@ #b, V).
% ?- eval(pred([X,Y])@ #b, V).
% ?- E = (=(append([a,b], [c,d]))), eval(E, V, [inside_out(true)]).
% ?- eval(xargs([X,Y]	:- Y is X)@1, R, [goal(G)]).
% ?- eval(xargs([X,Y]	:- Y is X)@1, R).

% ?- eval(xargs([X,Y,Z]	:- append(Y, X, Z))@[a]@[b], R, [run(false)]).
% ?- eval(1, V).
% ?- [util(misc), util(math)].
% ?- eval(misc:set(pow([1,2,3])), X).
% ?- eval(misc:set::pow([1,2,3]), X).
% ?- eval(([X,Y] \ (append(Y,X))) @[1,2,3]@[a,b,c], R).
% ?- eval(fun([X,Y] >> (append(Y,X))) @[1,2,3]@[a,b,c], R).
% ?- eval(fun([X, Y] >> append(X, Y))@[1,2,3]@[a,b,c], R).
% ?- eval(append([1,2], [3,4]), V).
% ?- eval(append@[1,2]@[3,4], V).

% ?- enable_pac_query.

% ?- maplist((pred([a])), [X,Y]).
% ?- eval(append([a,b],[c,d]), V).
% ?- E=1, eval(E, V).
% ?- E =append([a,b],[c,d]), eval(:E, V).
% ?- E =append([a,b],[c,d]), eval(E, V).

% ?- E =append([a,b],[c,d]), expand_exp(E,[val(V), goal(G)]), call(G).
% ?- E =append([a,b],[c,d]),eval(E, V).
% ?- E =append([a,b],[c,d]),eval(E, V, [goal(G)]).

% test eval / fun / (\) construction.
%
% ?- [util(misc)], [pac(op)], module(pac).
% ?- eval(quote(X\X), V).
% ?- eval(fun([F,X,Y]>> F@X@Y)@ quote(append) @ [1,2,3]@[a, b, c], V).
% ?- eval((fun([S, X] >> (S@X))@ quote(misc:set) @ quote(pow([1,2,3]))), Exp),
%   call(Exp, U).
% ?- eval((fun([S, X] >> (S@X))@ quote(misc:set) @ quote(pow([1,2,3]))), Exp),
% listing(Exp).
% ?- eval((#fun([S, X] >> (S@X))@ quote(misc:set) @ quote(pow([1,2,3]))), Exp).

% ?- eval(:fun([S, X] >> (S@X))@ quote(misc:set) @ quote(pow([1,2,3])), Exp), call(Exp, V).
%@ Exp = misc:set(pow([1, 2, 3])),
%@ V = [[3], [3, 1], [3, 2, 1], [3, 2], [2], [2, 1], [1], []].

% ?- eval(:(:fun([S, X] >> (S@X))@ quote(misc:set) @ quote(pow([1,2,3]))), V).
% ?- eval( fun([X,Y]>> append(Y,X))@[1,2,3]@[a,b,c], V).
% ?- eval( fun([X]>> quote(X)) @ 1, V).
% ?- eval( fun([X] >> #(X\ quote(X))) @ 1, F), call(F, hello, R).
% ?- eval( fun([X] >> #(X\ quote(X))) @ 1, F).
% ?- eval( (X\ #(X\quote(X))) @ 1, F), call(F, hello, V).
% ?- eval( fun([X,Y]>> append(Y,X))@[1,2,3]@[a,b,c], V).
% ?- call(fun([X]>>(set::(X+X))), [1,2,3], Z).
% ?- call([X]\(list::(X+X)), [1,2,3], Z).
% ?- eval( (X\ (append@X@X)) @[1,2,3], Z).
% ?- eval( pred([X, V]	 :- append(X, X, V)) @[1,2,3], V).
% ?- eval( :maplist @ pred([X, X]) @ [1,2,3], R).

% ?- [pac(op)], module(pac).
% ?- nopac(expand_exp( fun([X] -> X), [], =, V, user, G, P, [])).
% ?- trace, nopac(expand_exp( fun([X] -> (set::X)), [], =, V, user, G, P, [])).
% ?- trace, nopac(expand_exp( fun([X] -> (set::X)), [], call, V, user, G, P, [])).
% ?- nopac(eval( fun([X] -> #(fun([X]-> quote(X)))) @ 1, F)), call(F, hello, V).
% ?- nopac(eval( fun([X] -> #(fun([X]-> quote(X)))) @ 1, F)), call(F, hello, V).
% ?- nopac(eval( fun([X] -> #(fun([X]-> X))) @ 1, F)), call(F, hello, V).
% ?- nopac(eval( fun([X] -> #(fun([X], [Y]-> X@Y))) @  quote(integer), F)).
% ?- nopac(expand_exp( fun([F, X] -> (F@X)), [],  =, V, user, G, P, [])).
% ?- nopac(expand_exp( fun([F, X] -> (F@X)), [],  set, V, user, G, P, [])).

% ?- nopac(expand_exp( (#(X\X)@ #(Y\Y))@ (Z\Z), [], set, V, user, G, P, [])).
% ?- nopac(eval( ((X\X)@ (Y\Y))@ (Z\Z), V)).
% ?- eval( ((X\X)@ (Y\Y))@ (Z\Z), V).
%@ V = pac:'pac#3'(pac:'pac#1', pac:'pac#2').

% ?- eval( :((X\X)@ (Y\Y))@ (Z\Z), V).
% ?- eval( :((X\X)@ (Y\Y)), V).
%@ V = pac:'pac#7'.
%@ V = pac:'pac#7'.
% ?- eval( (X\X)@ (Y\Y)@(Z\Z), V).
%@ V = pac:'pac#11'(pac:'pac#9', pac:'pac#10').
% ?- eval( (X\X)@ (=) @ (=), V).

% ?- show_exp((X\X)@ (Y\Y)@ (Z\Z)).
%@ pac#14(pac#12,pac#13), where
%@ pac#12(A,A):-true
%@ pac#13(A,A):-true
%@ pac#14(A,A):-true
%@ true .


%%	expand_exp(+E:exp, +O:options) is det.
%	Convert E to a goal G (see below) such that  execution of G
%	makes V be unified with the value.
%
%   options in O:
%	*	goal(G) :  G is a goal whose execution gives the value V of
%				   the expression E.
%	*	pac(X-Y):  generated helper predicates in the form of d-list X-Y.
%	*	inside_out(true/false) : if true given, subterms of E
%			are evaluated recursively unless it is a expression with
%			a kind prefix like <kind>::<expression>.
%	*	run(true/false): 'true' is default, and then run the goal G.
%
%	*	val(V) : V is to be unified with the value of E.canonical_exp(

% ?- [util(misc)].

% ?- eval(:append([1],[2]), V).
% ?- eval(:append@[1]@[2], V).
% ?- eval(:append@ (:append@[1]@[2])@ (:append@[3]@[4]), V).
% ?- eval((X\X), V), call(V, hello, U).


% ?- eval((pred([X, X])), V), call(V, hello, U).
% ?- eval((pred([X, X])), V), call(V, 1, U).
% ?- eval(fun(X\X)@ hello, V).
% ?- eval((X\X)@hello, V), call(V, X).
% ?- eval(#(fun(X>>X))@hello, V), call(V, X).
% ?- eval(#(fun(X>>X))@hello, V), call(V, X).
% ?- eval((fun(X>>X))@ hello, V).
% ?- eval(:(fun(X>>X))@ quote(hello), V).
% ?- eval(:(fun(X>>X))@ hello, V).
%@ V = hello.

% ?- eval(:(:fun(X>>X)@append)@[1,2]@[3,4], V).
% ?- eval(#(fun([F, X]>> (F@X)))@ (=) @hello, V), call(V, R).
% ?- eval((fun([F, X]>> (F@X))), V), call(V, =, hello, R), listing(V).
% ?- eval(:fun([F, X]>> (F@X)) @ (=) @hello, V).
% ?- nopac(pac:eval(misc:set([1,2]), X)).
% ?- expand_exp(#(f)@ ((:b)@b1)@ (:c) @ (:z), [], =, V, m, G, P, []).
% ?- expand_exp(#(f)@(:b@b1)@c@z, [], =, V, m, G, P, []).

% ?- disable_pac_query.
% ?- enable_pac_query.
% ?- [util(misc), util(math)].
% ?- eval(misc:set(pow([1,2])), V).
% ?- eval((misc:set)::pow([1,2]), V).
%@ V = [[2], [2, 1], [1], []].
% ?- expand_exp(S::pow([1,2]), [], =, V, user, G, P, []), S=(misc:set),
%	call(G).
% ?- expand_basic(call(fun([X] >> X), A, B), user, G, P, []).
% ?- numlist(1, 3, A), expand_basic(maplist(fun([X] >> X), A, B), user, G, P, []),
%	 maplist(assert, P), call(G).
% ?- expand_basic(maplist(fun([F, X] >> F@X), A, B, C), user, G, P, []).
% ?- numlist(1, 3, A), maplist(fun([X] >> X), A, B).
% ?- numlist(1, 3, A), maplist(fun([X] >> :succ@X), A, B).
% ?- enable_pac_query.
% ?- disable_pac_query.
% ?- maplist(fun([X, F] >> (:F)@X), [1, 2], [=, succ], C).
% ?- maplist(fun([X, F] >> (:F)@X), [1, 2], [=, succ], C).
% ?- maplist(fun([X, F] >> F@X), [1, 2], [=, succ], C).
% ?- F=plus, X = 1,  maplist(F@X, [1, 2], C).
% ?- nopac(eval(fun([X, F] >> :(F)@X) @ 1 @ succ, C)), call(C, V).
% ?- eval(:maplist @ fun([X]>> X) @ [1,2], X).
% ?- nopac(eval(f @ fun([X] >> X) @ [1,2], X)).
% ?- eval(xargs([X, Y] >> (Y is X))@3, V).
% ?- eval(xargs([X,Y] >> (:append(X,Y)))@[1,2]@[3,4], V).
% ?- eval(xargs([Exp, V]:- V is Exp)@ quote(1+2), U, [goal(G)]), call(G).
% ?- eval(xargs([Exp, V]:- V is Exp)@ 1, U, [goal(G)]), call(G).
% ?- nopac(eval(::(misc:set, @(pow, [1,2])), E)).
% ?- nopac(eval(maplist @ fun([X] -> X) @ (:numlist(1,10)), Y)).
% ?- maplist(fun([]>> quote(a)), [X,Y]).
eval(E, V):- context_module(M),
	expand_exp(E, =, V, M, G, P, []), !,
	maplist_assert(P),
    call(G).

% ?- pac:eval(1, V, [goal(G), run]).
% ?- pac:eval(S::pow([1,2]), V, [goal(G)]),
%    S = misc:set, call(G).

% ?- pac:eval((M:K)::Pow@([1,2]), V, [goal(G)]),
%	M = util, Pow = pow, K = set, call(G).
%@ M = util,
%@ K = set,
%@ Pow = pow,
%@ V = [[2], [2, 1], [1], []],
%@ G = '$kind'(util:set, pow, [[1, 2]], [[2], [2, 1], [1], []], pac).

% ?-  eval(:fun([F,X,Y]-> (:F)@X@Y) @ append @ [1,2,3] @ [a, b, c], V,
%	[goal(G)]).
% ?- listing(pac:'pac#5').

eval(E, V, Opts):-
	(	memberchk(kind(K), Opts) -> true
	;	K = (=)
	),
	(	memberchk(module(M), Opts) -> true
	;	context_module(M)
	),
	(	memberchk(aux(P), Opts) -> true
	;	true
	),
	(	memberchk(goal(G), Opts) -> true
	;	true
	),
	expand_exp(E, K, V, M, G, P, []), !,
	maplist_assert(P),
	(	( memberchk(run, Opts); memberchk(run(true), Opts))
	->  call(G)
	;   true
	).

%
assert_dlist(P)		:- var(P), !.
assert_dlist([]).
assert_dlist([C|P])	:- assert_one(C),
	assert_dlist(P).

%
expand_dict_clause_olist(P0, P)			 :- var(P0), !, P = P0.
expand_dict_clause_olist([], [])		 :- !.
expand_dict_clause_olist([C|Cs], [D|Ds]) :-
	expand_dict_clause(C, D),
	expand_dict_clause_olist(Cs, Ds).

%
unify_options([], _).
unify_options([A|R], Options) :- memberchk(A, Options), !,
       unify_options(R, Options).
unify_options([_|R], Options) :- unify_options(R, Options).

		/**************************************************
		*     expanding limited functional expressions    *
		**************************************************/

%%  expand_exp(+E:exp,+S:kind,-U:term,+M:module,-G:goal, -P:list,-Q:list)
%	is semidet

%
%   True if G is a goal such that execution of G
%   unifies a term U with the value of expression E,
%	where  H is the helper predidates for G generated as
%   a difference list  P with a tail Q, i.e, append(H, Q, P) is true.
%	Typically atomic subgoals of G are of the form M:S(A, B).
%	expand_exp/7 is used when compiling an equations in a kind,
%	a set of equations, to a clause of the predicate M:S/2.
%
%   Simplest usage:
%	?- expand_exp(f(a), k, V, m, G, P, []).
%	G = m:k(f(a), V),
%	P = [].
%
%	?- expand_exp(f(a), call, V, m, G, P, []).
%	G = m:f(a, V),
%	P = [].
%
%	?- expand_exp(=(hello), call, V, user, G, P, []).
%	G = user: (hello=V),
%	P = [].
%
%   Features:
%  1. PAC/Lambda/Kind  are macros expanded on loading.
%  2. The sorted expression `(S :: E) = V' is eqivalent to `S(E, V)'.
%  3. Backquote '`'  is for  quoting term.
%  4. no special runtime predicate is necessary.
%  5. Arguments (in @ ) is evaluated before call.
%  6. let/2 for dynamic compiling pacs.

% ?- expand_exp(=(1), [val(V), goal(G)]), call(G).
% ?- expand_exp(=(1), [val(V), run(true)]).
% ?- expand_exp(:append([a,b],[c,d]),  [val(V), goal(G), run(true)]).
% ?- expand_exp(f(a)@b, =, V, m, G, P, []).
% ?- expand_exp(append(a)@b@c, =, V, m, G, P, []).
% ?- expand_exp((misc:set):: (pow@A), =, E, [], G,  P, []), A = [1,2], call(G).
% ?- expand_exp(misc:set(pow(A)), =, E, [], G,  P, []), A = [1,2], call(G).
% ?- expand_exp(misc:set::pow@A, =, E, [], G,  P, []), A = [1,2], call(G).
% ?- expand_exp(misc:set(pow@A), =, E, [], G,  P, []), write(G), A = [1,2], call(G).   %% error.
% ?- expand_exp(pow@[1,2], set(A), E, [], X,  P, []).
% ?- expand_exp((X\X), =, V, user, G, P, []).
% ?- expand_exp(X\(X\X), =, V, user, G, P, []).
% ?- expand_exp(pow@[1,2], set, E, [], X,  P, []).
% ?- expand_exp('`'(pow)@[1,2], set, E, [], X,  P, []).
% ?- expand_exp(pow@[1,2], set, E, [], X,  P, []), call(X).
% ?- expand_exp(pow(pow([1,2])), set, E, [], X,  P, []), call(X).
% ?- expand_exp(pow@(pow([1,2])), set, E, [], X,  P, []), call(X).
% ?- expand_exp(pow@(pow@[1,2]), set, E, [], X,  P, []), call(X).
% ?- expand_exp('`'(pow)@(pow@[1,2]), set, E, [], X,  P, []), call(X).
% ?- expand_exp(:ff(a), word_am, V,  user, G, P, []).
% ?- expand_exp((math:powerset)@A, set, V, misc, G, P, []).
% ?- expand_exp(1, [], set, V, user, G, P, []).
% ?- expand_exp(X, [], set, V, user, G, P, []).
% ?- expand_exp(X, [], call, V, user, G, P, []).
% ?- expand_exp(pow, [a], set, V, user, G, P, []).
% ?- expand_exp(append, [[a],[b]], call, V, user, G, P, []), call(G).
% ?- expand_exp(:append, [a,b], set, V, user, G, P, []).
% ?- expand_exp(append, [a,b], set, V, user, G, P, []).
% ?- expand_exp(m:append, [a,b], set, V, user, G, P, []).
% ?- expand_exp(quote(append), [a,b], call, V, user, G, P, []).
% ?- expand_exp('`'(append(x)), [a,b], call, V, user, G, P, []).
% ?- expand_exp(pred([X,X,X]), [a,b], call, V, user, G, P, []).
% ?- expand_exp(pred([X,Y,X+Y]), [a,b], call, V, user, G, P, []).
% ?- expand_exp(a@b, [], set, V, user, G, P, []).
% ?- expand_exp(a@b, [1, 2], set, V, user, G, P, []).
% ?- expand_exp(append@[a,b]@[c,d], call, V, user, G, P, []), call(G).
% ?- expand_exp(F@X@Y, call, V, user, G, P, []).
% ?- expand_exp(F@X@Y, set, V, user, G, P, []).
% ?- expand_exp(F@ X @ Y, call, V, user, G, P, []).
% ?- expand_exp(F@ (S @ X)@ (T@ Y), call, V, user, G, P, []).
% ?- expand_exp(F@ (S @ X)@ (T@ Y), set, V, user, G, P, []).
% ?- expand_exp(F@ (S @ X)@ (T@ Y), Unknown, V, user, G, P, []).
% ?- nopac( expand_exp((pred([X,Y,Z]))@a@b, [], set, V, user, G, P, [])).
% ?- expand_exp(([X]\X)@2, [], set, V, user, G, P, []).
% ?- expand_exp('`'(and)@(imply@X@Y)@(imply@Y@X), [], macro, V, user, G, P, []).
% ?- expand_exp(:union, [a, b], set, V, user, G, P, []).
% ?- expand_exp((:union)@a, [], set, V, user, G, P, []).
% ?- expand_exp(a, call, V, user, G, P, []).
% ?- expand_exp(A, call, V, user, G, P, []).
% ?- expand_exp(A@a, call, V, user, G, P, []).
% ?- expand_exp(A@a, set, V, user, G, P, []).
% ?- expand_exp(A@a, S, V, user, G, P, []).
% ?- expand_exp({a:1}, [], =, V, [], G, P, []).
% ?- expand_exp({b:{a:1}}, [], =, V, [], G, P, []).
% ?- expand_exp(:F, [], call, V, [], G, P, []).

% expand_exp/7
expand_exp(E, S, U, M, G, P, Q):-
	once(expand_exp(E, [], S, U, M, G, P, Q)).

% expand_exp/8
expand_exp(E, L, S, V, M, G, P, P):- var(E), !,
	expand_exp_var(L, E, S, V, M, G).
expand_exp(E, _, _, E, _, true, P, P):- canonical_exp(E), !.
expand_exp(E, L, _, G, M, true, P, Q):- expand_core(E, M, E0, P, Q), !,
	complete_args(E0, L, G).
expand_exp(::(S, E), L, _, V, M, G,  P, Q):-!,
	expand_exp(E, L, S, V, M, G, P, Q).
expand_exp(::, _, S, S, _,true,  P, P):- !.
expand_exp(@(E, E0), L, S, V, M, H, P, Q):- !,
	eliminate_atmark(E, [E0], [F|L0]),
	expand_exp_list(L0, S, L1, M, G, P, P0),
	append(L1, L, L2),
	expand_exp(F, L2, S, V, M, G1, P0, Q),
	slim_exp_goal((G, G1), H).
expand_exp(@(E), L, S, V, M, H, P, Q):- !,
	expand_exp(E, [], S, E0, M, H0, P, P0),
	expand_exp(E0, L, S, V, M, H1, P0, Q),
	slim_exp_goal((H0, H1), H).
expand_exp(@@(E), _, S, E0, M, H, P, P):- !,
	copy_term(E, E0),
	term_variables(E, Vs),
	term_variables(E0, V0s),
	maplist(make_kind_call(S, M), Vs, V0s, Calls),
	list_to_comma(Calls, H0),
	slim_exp_goal(H0, H).
expand_exp($(E), L, S, V, M, H, P, Q):- !,
	call(E, E0),
	expand_exp(E0, L, S, V, M, H, P, Q).
expand_exp(quote(E), L, _S, E0, _, true, P, P):- !,
	( var(E) -> E0 = E; complete_args(E, L, E0) ).
expand_exp('`'(E), L, S, V, M, G, P, Q):- !,
	expand_exp(quote(E), L, S, V, M, G, P, Q).
expand_exp(:(E), L, _, V, M, G, P, Q):-!,
 	expand_exp(E, [], =, E0, M, G0, P, Q),
	append(L, [V], L0),
	complete_args(M:E0, L0, G1),
	slim_exp_goal((G0, G1), G).
expand_exp(M:E, L, _, M:E0, _, G, P, Q):- !,
	expand_exp(E, L, =, E0, M, G, P, Q).
expand_exp(xargs(Fun), L, S, V, M, G, P, Q):- !,
	expand_xargs(Fun, L, S, V, M, G, P, Q).
expand_exp(flip(C), L, S, V, M, G, P, Q):- !,
	expand_exp(xargs(C), L, S, V, M, G, P, Q).
expand_exp(flip(F, C), L, S, V, M, G, P, Q):- !,
	flip_list(F, A, B),
	expand_exp(xargs(A:- apply(C, B)),
		   L, S, V, M, G, P, Q).
expand_exp({X}, _L, _S, V, M, G, P, Q):-!,
	(	chk_odict
	-> 	anti_subst({X}, V, Aux),
		anti_subst:expand_aux(Aux, M, G, P, Q)
	;	V = {X},  G=true, Q = P
	).
expand_exp(X, _, S, V, M, H, P, Q):- is_role(X, A, B), !,
	expand_dict_access(A, S, V0, M, H0, P, P0),
	expand_dict_access(V0, B, S, V, M, H1, P0, Q),
	slim_exp_goal((H0, H1), H).
expand_exp(E, L, S, V, M, G, P, P):- S == call, !,
	attach_prefix(M, E, E0),
	append(L, [V], L0),
	complete_args(E0, L0, G).
expand_exp(E, L, S, V, _, true, P, P):-  S == (=), !,
	complete_args(E, L, V).
expand_exp(E, L, S, V, M, G, P, P):-
	attach_prefix(M, S, S0),
	complete_args(E, L, E0),
	complete_args(S0, [E0, V], G).

%
expand_xargs(Args:-Goal, L, _, V, _, Goal, P, P):-!,  % last arg is outpu.
	append(L, _, Args),
	append(_, [V], Args).
expand_xargs(Fun, L, _, V, M, G, P, Q):-
	normal_fun(Fun, =, fun(_, _, Args>>Exp)),
	append(L, L0, Args),
	expand_exp(Exp, L0, =, V, M, G, P, Q).

%
normal_fun(fun(X\E), S,  fun(S, [], X>>E)).
normal_fun(fun(X->E), S, fun(S, [], X>>E)).
normal_fun(fun(X>>E), S, fun(S, [], X>>E)).
normal_fun(fun(S, A, Funs), _, fun(S, A, Funs)).
normal_fun(fun(A, Funs), S, fun(S, A, Funs)).
normal_fun(X\E, S,  fun(S, [], X>>E)).
normal_fun(X>>E, S, fun(S, [], X>>E)).

%
expand_fun_to_pred(fun(S, A, Funs), M, G, P, Q) :-
	funs_to_preds(Funs, S, M, Clauses, P, P0),
	expand_core(pred(A, Clauses), M, G, P0, Q).

% expand_dict_access
% /7.
% ?- pac:expand_dict_access(role(X, a), =, V, user, H, P, []).
% ?- pac:expand_dict_access(role(role(X,a), b), =, V, user, H, P, []).
% ?- pac:expand_dict_access(role(X, a(k)), =, V, user, H, P, []).
% ?- pac:expand_dict_access(role(X, role(Y, a)), =, V, user, H, P, []).
expand_dict_access(X, S, V, M, H, P, Q):- nonvar(X),  is_role(X, A, B), !,
  	expand_dict_access(A, S, V0, M, H0, P, P0),
	expand_dict_access(V0, B, S, V, M, H1, P0, Q),
	slim_exp_goal((H0, H1), H).
expand_dict_access(X, S, D, M, H, P, Q):- expand_exp(X, S, D, M, H, P, Q).
% /8.
expand_dict_access(X, R, S, V, M, H, P, Q):- nonvar(R), is_role(R, _, _), !,
	expand_dict_access(R, S, V0, M, H0, P, P0),
	expand_dict_access(X, V0, S, V, M, H1, P0, Q),
	slim_exp_goal((H0, H1), H).
expand_dict_access(X, K, _, V, _M, H, P, Q):-
	expand_exp(K, =, K0, [], H0, P, Q),
	slim_goal((H0, role(K0, X, V)), H).

%
expand_role_list(Y, S, Path, M, H, P, Q):- nonvar(Y), is_role(Y, Y1, Y2), !,
	expand_role_list(Y1, S, Path1, M, H1, P, P0),
	expand_role_list(Y2, S, Path2, M, H2, P0, Q),
	slim_exp_goal((H1, H2, append(Path1, Path2, Path)), H).
expand_role_list(Y, S, [V], M, H, P, Q):-
	expand_exp(Y, S, V, M, H, P, Q).

%
expand_exp_list([], _, [], _ , true, P, P).
expand_exp_list([E|L], S, [E0|L0], M, (G, G0), P, Q):-
	expand_exp_arg(E, S, E0, M, G, P, R),
	expand_exp_list(L, S, L0,  M, G0, R, Q).

% expand_exp_var/6
expand_exp_var([], E, S, V, M, G):- !, expand_exp_var(E, S, V, M, G).
expand_exp_var(L, E, S, V, M, '$kind'(S, E, L, V, M)).

% expand_exp_var/5
expand_exp_var(E, S, V, M, '$kind'(S, E, [], V, M)):- (var(S); var(M)), !.
expand_exp_var(E, =, E, _, true):-!.
expand_exp_var(E, call, V, M, call(M:E, V)):-!.
expand_exp_var(E, S, V, M, call(M:S, E, V)).

% ?- qcompile(util(misc)).
% ?- misc:set(pow([1,2]), X).
% ?- eval(E:: (Pow@A), X, [goal(G)]), Pow = pow, E=misc:set, A=[a,b], call(G).

% Run time for eval.
user:'$kind'(S, E, _, _, _):- (var(S);var(E)), !,
	throw(uninstantiated_term_or_kind).
user:'$kind'(_, :(E), L, V, M):-!,
	complete_args(E, L, E0),
	call(M:E0, V).
user:'$kind'(_, M:E, L, V, _):-!,
	complete_args(E, L, E0),
	call(M:E0, V).
user:'$kind'(call, E, L, V, M):-!,
	complete_args(E, L, E0),
	call(M:E0, V).
user:'$kind'(=, E, L, V, _):-!, complete_args(E, L, V).
user:'$kind'(S, E, L, V, M):-
	complete_args(E, L, E0),
	call(M:S, E0, V).

%
expand_exp_arg(E, S, U, M, G, P, Q) :-
	once(expand_exp_basic_arg(E, S, U, M, G, P, Q)).
expand_exp_arg(E, S, U, M, G, P, Q) :-
	once(expand_exp(E, [], S, U, M, G, P, Q)).

%
expand_exp_basic_arg(E, S, U, M, G, P, P):- var(E), !,
	(	( S = call; S = (=) )
	->  U = E,
		G = true
	;	attach_prefix(M, S, S0),
	 	complete_args(S0, [E,U], G)
	).
expand_exp_basic_arg(E, _, E, _, true, P, P):-  canonical_exp(E), !.
expand_exp_basic_arg(E, _, U, M, true, P, Q):-
	once(	expand_core(E, M, U, P, Q)
	    ;	expand_etc(E, M, U, P, Q)
	    ).
%
eliminate_atmark(E, L,  [E|L]):- var(E).
eliminate_atmark(@(E,E0), L,  L0):- eliminate_atmark(E, [E0|L], L0).
eliminate_atmark(E, L,  [E|L]).

% some tiny runtime.
list_to_comma([X,Y|Z], (X, U)):- list_to_comma([Y|Z], U).
list_to_comma([X], X).
list_to_comma([], true).

%
nopac(X):- call(X).		% to prevent from pac-expansion

%
flat(X, Y, Z):- once(flat(X, Y, Z, [])).
%
flat(_, A, [A|P], P):- var(A).
flat(;, A;B, P, Q):- flat(;, A, P, P0),
	flat(;,  B, P0, Q).
flat(&, A&B, P, Q):- flat(&, A, P, P0),
	flat(&,  B, P0, Q).
flat(_, A, [A|P], P).

%
attach_prefix_head(M, X:-Y, MX:-Y):-
	attach_prefix(M, X, MX).

% ?- pac_to_pred([X], U).
pac_to_pred(X, _):- var(X), throw(pac_syntax_error(X)).
pac_to_pred(H:-G, [P, F, G]):-  pac_to_pred_head(H, [P,F]).
pac_to_pred(H, [P, F,true]):- pac_to_pred_head(H, [P,F]).

%
pac_to_pred_head(P, _):- var(P), throw(pac_syntax_error(P)).
pac_to_pred_head(P, [P, []]):- is_list(P).
pac_to_pred_head(P, _) :- throw(pac_syntax_error(P)).

% ?- pac:meta_property(maplist(0), I).
% ?- pac:meta_property(xxx:maplist(=(0), _), I).
meta_property([]:E, I):-!, meta_property(E, I).
meta_property(E, I)	:- predicate_property(E, meta_predicate(I)).

%
slim_exp_goal(G, G0):- once(slim_exp_aux(G, G1)),
	reduce:slim_goal(G1, G0).
%
slim_exp_aux(M:A, M:B):- slim_exp_aux(A, B).
slim_exp_aux((X=Y), true):- X==Y.
slim_exp_aux((H,G), (H0, G0)):- slim_exp_aux(H, H0),
	slim_exp_aux(G, G0).
slim_exp_aux(X, G):-  X =.. [xargs, Pred |Args],
	pass_xargs(Pred, Args, G).
slim_exp_aux(H, H).

% ?- pass_xargs([X,Y]:- a(X,Y), [1], G).
% ?- pass_xargs([X,Y]:- pred(X, [U]:-a(X,Y)), [1], G).

pass_xargs(Ps :- Body, Args, G):-
	append(Args, _, Ps),
	slim_exp_goal(Body,  G).
pass_xargs(Ps, Args, true):- is_list(Ps),
	append(Args, _, Ps).
pass_xargs(F, [A, B], G):- complete_args(F, [B, A], G).

%
make_kind_call(S, M, V, U, call(M:S, V, U)).
%
make_kind_list(_, _, [], [], _, true, P, P).
make_kind_list(F, S, [V|Vs], [U|Us], M, (G, U=U0, Gs), P, Q):-
	once(expand_exp(F@V, S, U0, M, G, P, P0)),
	make_kind_list(F, S, Vs, Us, M, Gs, P0, Q).

% ?- flip_call([1,2], p, Q).
% ?- flip_call([2,1], p, Q).
% ?- flip_call([2,1], p(A), Q).
flip_call(P, F, pred(Vs, A :- apply(F, B))):- term_variables(F, Vs),
	flip_list(P, A, B).

% ?- module(mpac).
% ?- flip_list([1,2,3], [a,b,c], Y).
% ?- flip_list([3,1,2], [a,b,c], Y).
% ?- flip_list([1], [a], Y).
% ?- flip_list([1,2], [a,b,c], Y).
% ?- flip_list([3,1,2], X, [a,b,c]).

flip_list([], _, []):-!.
flip_list([I|Is], Y, [X|Xs]):-
	nth1(I, Y, X),
	flip_list(Is, Y, Xs).
%
partial_args_match([],_).
partial_args_match(_,[]).
partial_args_match([A|As],[A|Bs]):- partial_args_match(As, Bs).

% Tiny helpers
zip([A|B], [C|D], [A-C|R]):- zip(B, D, R).
zip([], [], []).

% ?- flip_clause(f, [], f, X).
% ?- flip_clause(f, [1,2], f(a,b),  X).
% ?- flip_clause(f, [2,1], f(a,b),  X).
% ?- flip_clause(f, [1,3,2], f(a,b,c),  X).
% ?- flip_clause(f, [1,3,2],(f(a,b,c):- true, g(a), f(x, y, z)), X).

flip_clause(F, P, T, T0):-
 	walk_term(T, T0, flip_functor_args(F, P)).
%
flip_functor_args(F, P, T, T0):-
	T =.. [F|As],
	length(P, N),
	length(As, N),
	flip_list(P, As, Bs),
	T0 =.. [F|Bs].
flip_functor_args(_, _, T, T).

% ?- walk_term((a,b), X, =).
% ?- walk_term(a:-b, X, =).
% ?- walk_goal(a:-b, X, =).
walk_term(A, A, _) :- var(A).
walk_term(M:A, M:B, F):- walk_term(A, B, F).
walk_term(H:-B, H0:-B0, F):- walk_term(H, H0, F), once(walk_goal(B, B0, F)).
walk_term(A, A0, F):- call(F, A, A0).

% ?- walk_goal((a,b), X, =).
walk_goal(A,A,_) :- var(A).
walk_goal((A0,B0), (A,B), F) :-  walk_goal(A0, A, F), walk_goal(B0, B, F).
walk_goal((A0;B0), (A,B), F) :-  walk_goal(A0, A, F), walk_goal(B0, B, F).
walk_goal((A0->B0), (A->B), F) :-  walk_goal(A0, A, F), walk_goal(B0, B, F).
walk_goal(\+(A0), \+(A), F) :-  walk_goal(A0, A, F).
walk_goal(M:A, M:A0,  F):-  walk_goal(A, A0, F).
walk_goal(A, A0, F):- call(F, A, A0), !.
