:- module(pac_word,[let_w/2, let_wd/2, let_wl/2, let_wld/2,
	adjacent_interval/3, am_finals/2,
	expand_recognize_act/7,
	expand_w/5, expand_w/7, expand_wl/5, expand_wl/7,
	regex_am/2, regex_coalgebra/2, regex_compare/3,
	word/3, x_char_boole_form/2
				   ]).

:- use_module(pac('parse-regex')).
:- use_module(pac('interval-boole')).
:- use_module(pac('anti-subst')).
:- use_module(pac(reduce)).
:- use_module(pac('expand-pac')).
:- use_module(pac(op)).

%%  word_am(X:regex, Y:automata) is det
%
%   True if X is converted to Y.

%
% ?- coalgebra:show_am("a").
% ?- coalgebra:show_am(".*").
% ?- coalgebra:show_am(@($)).
% ?- coalgebra:show_am(a).
% ?- coalgebra:show_am(a^(3-5)).
% ?- coalgebra:show_am(a^(>=(5))).
% ?- regex_coalgebra(a^(>=(1)), X).
% ?- regex_am(a^(>=(1)), X).

word_am(char(X),A1)	:-!, word_am(char_boole_form(X),A2),
	word_am(i_boole(A2),A3),
	am_char(A3,A1).
word_am('I'(X),A1)	:-!, am_char_simple(X,A1).
word_am([],A1)		:-!,am_unit(A1).
word_am(coa(A,B),coa(A,B)):-!.
word_am(X+Y,A1)		:-!, word_am(X,A2),
	word_am(Y,A3),
	am_concat(A2,A3,A1).
word_am((X|Y),A1)	:-!, word_am(X,A2),
	word_am(Y,A3),
	am_cup(A2,A3,A1).
word_am((X\Y),A1)	:-!, word_am(X,A2),
	word_am(Y,A3),
	am_minus(A2,A3,A1).
word_am(X&Y,A1)		:-!, word_am(X,A2),
	word_am(Y,A3),
	am_cap(A2,A3,A1).
word_am(+X,A1)		:-!, word_am(X,A),
	word_am(A+ *(A),A1).
word_am(*(X),A1)		:-!, word_am(X,A2), am_star(A2,A1).
word_am(\+ _,A1)	:-!, word_am_aux(A1).
word_am(\ _,A1)		:-!, word_am_aux(A1).
word_am(?(X),A1):-!, word_am(([]|X),A1).
word_am(E^ >=(N),A1):-!, word_am(E,E0),
	word_am(E0^N,A2),
	word_am(*E0,A3),
	word_am(A2+A3,A1).
word_am(E^(I-J),A1)	:-K is J-I, !,
	word_am(E,E0),
	word_am(E0^I,A2),
	word_am(E0^ =<(K),A3),
	word_am(A2+A3,A1).
word_am(E^ =<(N),A1):-!, word_am(E,A2),
	am_power_lower(N,A2,A1).
word_am(E^N,A1)		:-!, word_am(E,A2), am_power(N,A2,A1).
word_am(rev(E),A1)	:-!, word_am(E,A2),
	am_reverse(A2,A1).
word_am(special(X,Is),A1):-!, am_special(X,Is,A1).

%
word_am_aux(*('.'), _).

			/****************************
			*     Special characters    *
			****************************/

%
regex_macro($,	special($)).
regex_macro(comment,	"/\\*.*\\*/").
regex_macro(paragraph,	".*$$$*").

%
final_special_pred(at_end_of_tape).

%
special_char_class($, '\n').

% ?- x_char_boole_form($, X), iboole:i_boole(X, I).
x_char_boole_form(X, I):- special_char_class(X, X0),
	once(char_boole_form(X0, I)).

% So far, only "at_end_of_tape" is an example of special tokens.
% ?- am_special($, [1,2,3], E).

am_special($, Interval_list,  coa(E, 1)):-
	sort(Interval_list, Interval_list0),
	E0 = [1- [ special(at_end_of_tape),
		   Interval_list0 - 2	],
		   2 - [[]]],
	split_goto(E0, E).

%
compile_special_char($, Ints, N, I, J, M,
		     [R|P], P) :-
	(	N ==0,
		Head0 =.. [I, [V |X], Y] ,
		CallJ =.. [J, X, Y]
	;       N ==2,
		Head0 =.. [I, A, B, [V |X], Y],
		CallJ =.. [J, A, B, X, Y]
	),
	( Ints = (U-U) -> (U = V, V_test_code=true)
	;   once(generate_code_test(Ints, V, V_test_code))),
	slim_goal((V_test_code, CallJ), Body0),
	prefix_convention(M, Head0:-Body0, R).
%
compile_special_pred(at_end_of_tape, [I, _, N, M], Head:-true):-
	(	N ==0, 	Head0 =..[I, [],[]]
	;       N ==2,  Head0 =..[I, A, A, [],[]]
	),
	prefix_convention(M, Head0, Head).


% ?-  phrase(w(char(alnum) + *(char(alpha))), `a1bc`, S).
% ?-  phrase(w(ab+cd), `abcde`, S).
% ?-  phrase(w("abcd"), `abcde`, S).
% ?-  phrase(w(i(2)+i(3)), [2,3], S).
% ?-  regex_coalgebra("23", S).
% ?-  regex_coalgebra(i(2)+i(3), S).

% ?-  phrase(w(i(2)+i(3)), [2,3], S).
% ?-  phrase(w(i(2)^(=<(3))), [2,2], S).
% ?-  phrase(w(i(2)^(>=(3))), [2,2,2,2], S).

% ?- let_w(X, ".*"), phrase(X, `abc`, R).
% ?- let_wd(X, ".*"), call(X, A, [], `abc`, R).
% ?- let_wl(X, ".*"), phrase(X, `abc`, R).
% ?- let_wld(X, ".*"), call(X, A, [], `abc`, R).

let_w(F, X) :- once(let_expand_regex(w, 0, X, user, F)).
let_wd(F, X):- once(let_expand_regex(w, 2, X, user, F)).
let_wl(F, X):- once(let_expand_regex(wl, 0, X, user, F)).
let_wld(F, X):- once(let_expand_regex(wl, 2, X, user, F)).

%
let_expand_regex(W, N, Regex, M_prefix, Pred):-
	regex_coalgebra_code(Regex, coa(C, I, _)),
	(	W == wl -> null_last_coa(C, C0); C0=C),
	expand_coa(N, C0, I, M_prefix, Pred, P, []),
	maplist(assert, P).

% ?- let_sed(X,  w("a") >> =([])).
% ?- regex_word("a", _4774, _4776).
% ?- let_sed(X,  w("a") >> =([])), call(X, `abac`, S), basic:smash(S, S0).
% ?- let_sed(X,  w("a") >> =([])), call(X, `abac`, S), basic:smash(S).
% ?- let_sed(X,  w("a") >> =([])), call(X, `abac`, S), basic:smash(S).
% ?- let_sed(X,  w("a") >> =([])), call(X, `abac`, S), basic:smash(S).
% ?- let_sed(X,  (w("a", A), w("b", B)) >> append(B, A)), call(X, `abab`, S), basic:smash(S).
% ?- let_sed(X,  s/"a"/"x"), call(X, `abac`, S), basic:smash(S).
% ?- let_sed(X,  b/"a"/"x"), call(X, `abac`, S), basic:smash(S).
% ?- let_sed(X,  a/"a"/"x"), call(X, `abac`, S), basic:smash(S).
% ?- let_sed(X,  d/"a"), call(X, `abac`, S), basic:smash(S).
% ?- let_sed(X,  w/"a"/"xxx"/"yyy"), call(X, `abac`, S), basic:smash(S).
% ?- let_sed(F,  wl("a+", A) >> pred(A, ([B]:- length(A, L), length(B,L), maplist(=(0'b), B)))), call(F, `aaa`, R), basic:smash(R).
% ?- let_sed(F,  (w(".*", A), wl("b+")) >> pred(A, [A])),  call(F, `aaabbccc`, R), basic:smash(R).
% ?- basic:herbrand(_, `(w(".*", A), wl("b+")) >> pred(A, [A])`, H), let_sed(F, H),  call(F, `aaabbccc`, R), basic:smash(R).

user:let_sed(X, Y) :- prolog_load_context(module, M),
	once(let_sed(X, M, Y)).

%
let_sed(X, M, S) :- pac:expand_sed(S, [F, W, A]),
	once(expand_recognize_act(F, W, A, M, X, P, [])),
	maplist(assert, P).

% ?- regex_coalgebra("[^\\./]*", X).
% ?- regex_coalgebra(( "\"([^\"\\\\]|(\\\\.))*\"" | "'([^'\\\\]|(\\\\.))*'" | "[^ \t\"']+"), X).
% ?- regex_coalgebra(*(a|b)+ [a]+(*([]))+ (*(*([]|[]))), N).
% ?- regex_coalgebra(*(a|b)+ "a"+(*([]))+ (*(*([]|[]))), N).
% ?- regex_coalgebra(*(.), X).
% ?- regex_coalgebra(".*", X).
% ?- regex_coalgebra(ab+cd, X).
% ?- regex_coalgebra(ab+cd, X, [code]).

% ?- regex_coalgebra_code("a*", X).
% ?- regex_coalgebra_code(ab+cd, X).
% ?- regex_coalgebra_char(".*", X).
% ?- regex_coalgebra_char("a*", X).
% ?- regex_coalgebra_char("[adlz]", X).
% ?- regex_coalgebra_char(ab+cd, X).
% ?- regex_coalgebra_char(a, X).
% ?- regex_coalgebra_char(i([-1, -1]), X).
% ?- regex_coalgebra_char("a", X).

regex_coalgebra(X, coa(E0, I, Fs), Options):- once(regex_am(X,  coa(E, I))),
	am_finals(coa(E, _), Fs),
	(  memberchk(char, Options)
	->	map_coa_char_int(E, E0)
	;	E0 = E
	).

%
regex_coalgebra(X, C):- regex_coalgebra(X, C, [code]).
%
regex_coalgebra_code(X, C):- regex_coalgebra(X, C, [code]).
%
regex_coalgebra_char(X, C):- regex_coalgebra(X, C, [char]).

		/***************************************
		*     State Transition by Automaton    *
		***************************************/

% ?- regex_coalgebra_char("a*", X).
% ?- regex_coalgebra_char("a", X).
% ?- map_coa_char_int([1-[[97-97]-2], 2-[[]]], _508).
% ?- regex_coalgebra_char("a*", X), state_move([a], 1, S, X).
% ?- (regex_coalgebra_char("a*", X), state_move([a,a,a], 1, S, X)).
% ?- (regex_coalgebra("a*", X, [char]), state_move([a,a], 1, S, X)).
% ?- regex_coalgebra("\001\*", X, [code]).
% ?- regex_coalgebra(i([-1, -1]), X).

state_move([], S, S, _).
state_move([X|R], S, S0, Coa):-
	once(state_move_one(X, S, S1, Coa)),
	state_move(R, S1, S0, Coa).
%
state_move_one(X, S, S1, coa(E,_,_)):- memberchk(S-Dots, E),
	member(Intervals-S1, Dots),
	member(A-B, Intervals),
	greater_or_eq(X, A),
	smaller_or_eq(X, B).
%
greater_or_eq(_, inf).
greater_or_eq(X, I):- X @>= I.
%
smaller_or_eq(_, sup).
smaller_or_eq(X, I):- X @=< I.

			/***************************
			*   map character classes  *
			***************************/
% ?-  phrase(w(char(alnum) + *(char(alpha))), `a1bc`, S).
% ?-  phrase(w(ab+cd), `abcde`, S).
% ?-  phrase(w("abcd"), `abcde`, S).

% ?-  map_coa_char_int(R, [1-[[a-a]-2], 2-[[]]]).
% ?-  map_coa_char_int(R, [1-[[a-a]-2], 2-[[]]]), map_coa_char_int(R, S).

map_coa_char_int(X, Y):-
	maplist(coa_char_int_aux, X, Y).
%
coa_char_int_aux(I-R, I-R0):-  maplist(map_goto_char_int, R, R0).

%
map_goto_char_int([], []).
map_goto_char_int(Is-S, Js-S):- maplist(goto_char_int_aux, Is, Js).
%
goto_char_int_aux(A-B, A0-B0):- char_int(A, A0), char_int(B, B0).

%
char_int(inf, inf).
char_int(sup, sup).
char_int(X, Y):-
	(	(	is_of_type(negative_integer, X)
		;	is_of_type(negative_integer, Y)
		)
	->	X = Y
	;	char_code(Y, X)
	).

			/*********************
			*     goto states    *
			*********************/

%  ?-  am_char_dict([2-[[inf-sup]-3]], C, [], P).
%@ C = [2-[_G2137-3]],
%@ P = [[inf-sup]-_G2137].

am_char_dict([], [], P, P).
am_char_dict([[]|Z], [[]|Z0], P, Q):- am_char_dict(Z, Z0, P, Q).
am_char_dict([X-Y|Z], [X-Y0|Z0], P, Q):-
	once(goto_dict(Y, Y0, P, R)),
	am_char_dict(Z, Z0, R, Q).

%
goto_dict([], [], P, P).
goto_dict([[]|X],[[]|Y], P, Q):- goto_dict(X, Y, P, Q).
goto_dict([I-S|X],[V-S|Y], P, Q):- add_char_dict(I, P, P0, V),
	goto_dict(X, Y, P0, Q).

%
add_char_dict(I, P, P, V):- memberchk(I-V, P).
add_char_dict(I, P, [I-V|P], V).

%
expand_char_class(U, V):- maplist(expand_char_class_aux, U, V).

%
expand_char_class_aux(X-Y, X-Y0):-
	maplist(expand_char_class_aux_aux, Y, Y0).
%
expand_char_class_aux_aux([], []):- !.
expand_char_class_aux_aux(V-S, Gotos) :-
	maplist(expand_char_class_aux_aux_aux(S), V, Gotos).
%
expand_char_class_aux_aux_aux(S, A, A-S).

% ?- split_goto([1-[[], [1,2,3]-s, [2,3]-t], 1-[[1,2,3]-s, [2,3]-t]], X).
split_goto(X, Y):-	maplist(split_goto_aux, X, Y).

%
split_goto_aux(I-S, I-S0):- distribute_state(S, S0, []).

%
distribute_state(X, Y):- distribute_state(X, Y, []).

%
distribute_state([X-S|Y], Z, U):- distribute_state(X, S, Z, V),
	distribute_state(Y, V, U).
distribute_state([A|X], [A|Y], Z):- distribute_state(X, Y, Z).
distribute_state([], X, X).

%
distribute_state([A|R], S, [A-S|P], Q):- distribute_state(R, S, P, Q).
distribute_state([], _, Z, Z).

			/*****************************
			*     automaton synthesis    *
			*****************************/

%
merge([X,X|Y], U):- merge([X|Y], U).
merge([X|R], [X|U]):- merge(R, U).
merge([], []).

% ?- regex_am((.), R).
% ?- regex_am(($), R).
% ?- regex_am("abc", R).

regex_am(X, Y):- regex_word(X, X0, Basic_interval_dict),
 	word_am(X0, Y0),
 	am_connected_region(Y0, Y0_reachable),
 	am_remove_empty_states(Y0_reachable, Y0_slim),
	once(am_remove_useless_states(Y0_slim, Y0_useful)),
 	once(am_char_back(Y0_useful, Basic_interval_dict, Y1)),
 	am_normal(Y1, Y).

%
word_normal_am --> word_am,
	am_connected_region,
	am_remove_empty_states,
	am_remove_useless_states,
	am_normal.
%
regex_word(X, Y):- regex_word(X, Y, _).

%%  regex_word(+E:regex, -W: word, -D:intval_dict)  is det
%
%   True if E is conveted to W with D.
%
%   Simplest usage:
%
%	?- pac_word:regex_word(("."), Y, Z).
%	Y = 'I'([1]),
%	Z = [inf-sup-1] .
%	?- regex_word(char(alpha), Y, Z).
%	Y = 'I'([1, 2]),
%	Z = [65-90-1, 97-122-2] .
%	?- regex_word(special($), Y, Z).
%	Y = special($, [x($, 1)]),
%	Z = [x($, 10-10)-x(1)] .

x_snd(x(_, I), I):-!.
x_snd(I, I).

% ?- regex_word((special($) + a), Y, Z).
regex_word(X, Y, Basic_interval_dict):-
	once(parse_interval(X, Y, P, [])),
	sort(P, SortedP),
	once(merge(SortedP, PX)),
	zip_hyphen(A, Target_interval_list, PX),
	maplist(x_snd, A,  A0),
	zip_hyphen(A0, Refinement, Pzip),
	refine_char_dict(Pzip),
	basic_interval_index(Refinement,
	     Partition_in_index,
	     Basic_interval_dict0),
	zip_hyphen(A, Partition_in_index, Char_dict),
	mark_basic_interval_dict(Basic_interval_dict0,
		 Char_dict,
		 Basic_interval_dict),
	mark_interval_block_list(Char_dict,
		 Partition_in_index,
		 Target_interval_list).

% [2014/09/02]
refine_char_dict(Tzip):-  zip_hyphen(Left, Right, Tzip),
	iboole:m_partition(Left, Right).

%%  mark_basic_interval_dict(+X:zip, +P:zip, -Y:zip) is det
%
%   True if Y is unified with a basic interval dict built from X and Y.
%
%   sample use:
%	?-  mark_basic_interval_dict([(a-b)-1, (c-d)-2, (e-f)-3], [x($, _)-[1],
%		x(#, _)-[2], [_]-[3,4]],	Y).
%	Y = [x($, a-b)-x(1), x(#, c-d)-x(2), e-f-3] .

mark_basic_interval_dict(X, P, Y):-
	maplist(mark_basic_interval_dict_aux(P), X, Y).

%
mark_basic_interval_dict_aux(P, A-N, x(S, A)- x(N)):-
				member(x(S,_)-L, P),
				memberchk(N, L).
mark_basic_interval_dict_aux(_, U, U).

%
mark_interval_block_list(Assoc, IB, IBX):-
	maplist(mark_interval_block(Assoc), IB, IBX).

%
mark_interval_block(Assoc, L, LX):-
	maplist(mark_interval(Assoc), L, LX).

%
mark_interval(Assoc, I, x(S,I)):- member(x(S, _)-W, Assoc),
	memberchk(I, W).
mark_interval(_, I, I).

%
am_char_back(coa(X, I), M, coa(Y, I)):-	maplist(am_char_back_aux(M), X, Y).

%
am_char_back_aux(M, I-A, I-B):- maplist(am_char_back_aux_aux(M), A, B).

%
am_char_back_aux_aux(M, x(_X, K)-J, A-J) :-  memberchk(A-x(K), M).
am_char_back_aux_aux(M, C-J, D-J) :-  memberchk(D-C, M).
am_char_back_aux_aux(_, U, U).

%
basic_interval_index(Ks, Js, Basic_interval_dict):-
	maplist(zip_hyphen, Ks, Js,  U),
	flatten(U, V),
	sort(V, Basic_interval_dict0),
	once(merge(Basic_interval_dict0, Basic_interval_dict)),
	length(Basic_interval_dict, N),
	(	N==0
	->	true
	; 	numlist(1, N, NZip),
		zip_hyphen(_, NZip, Basic_interval_dict)
	).


			/************************
			*     parse interval    *
			************************/

%%  parse_interval(+X:word, -Y:word, -I:intervals, -I0:intervals) is det
%
%   True if Y is unified with X such that  all intervals in X are
%   extracted into I (minus I0), so that Y and X are equivalent
%   after refinement of the intervals I (minus I0).
%
%   Simple usage:
%	?- parse_interval("[a-z]", R, X, []).
%	R = 'I'(_G5557),
%	X = [[97-122]-_G5557] .

% ?-  phrase(w(i([-1, -1])), [-1], X).
% ?-  expand_phrase(w(i([-1, -1])), [], G, L, []).

% ?- regex_coalgebra("abc*", X, [code]).
% ?- (regex_coalgebra("(\000\|\001\)*", X, [code]), state_move([0,1,0,0], 1, S, X)).
% ?- regex_coalgebra(i(-1), X,  [code]).
% ?- regex_coalgebra(i([-1, -2]), X,  [code]).
% ?- (regex_coalgebra(*(i([-5 - -1])), X, [code]), state_move([-1, -2, -3], 1, S, X)).
% ?- (regex_coalgebra(i(-1) + *(i(0)) + i(-2), X, [code]), state_move([-1, -2], 1, S, X)).
% ?- (regex_coalgebra(i(-1) + *(i(0)) + i(-2), X, [code]), state_move([-1, 0,0, 0, 0, -2], 1, S, X)).
% ?- show_phrase(w("a*$")).
% ?- coalgebra:show_am("abc*").
%@ true.


parse_interval(X, Y) --> i_parse(X, Y).

%
i_parse((.), 'I'(U))	--> [[inf-sup]-U].
i_parse(char(X), 'I'(U))	--> [I-U],
	{ once(char_boole_form(X, X0)),  i_boole(X0, I) }.
i_parse(i(X), 'I'(U)) --> [I - U],
	{ once(i_normal_class_form(X, Y)),  i_boole(Y, I) }.
i_parse('C'(X), U)	--> i_parse_codes(X, U).
i_parse([], [])		--> [].
i_parse(special(X),  special(X, U)) --> [x(X, I)-U],
 	{ x_char_boole_form(X, X0),  i_boole(X0, I) }.
i_parse(out(X), U)	-->  i_parse(char(out(X)), U).
i_parse(dot(X), U)	-->  i_parse(char(X), U).
i_parse(X+Y, U+V)	-->  i_parse(X, U), i_parse(Y, V).
i_parse(X|Y, U|V)	-->  i_parse(X, U), i_parse(Y, V).
i_parse(X\Y, U\V)	-->  i_parse(X, U), i_parse(Y, V).
i_parse(X&Y, U&V)	-->  i_parse(X, U), i_parse(Y, V).
i_parse(+(X), +(U))	-->  i_parse(X, U).
i_parse(*(X), *(U))	-->  i_parse(X, U).
i_parse(\+(X), U)	-->  i_parse( *(.) \ X, U).
i_parse(\(X), U)	-->  i_parse( *(.) \ X, U).
i_parse(rev(X), rev(U))	-->  i_parse(X, U).
i_parse(?(X), ?(U))	-->  i_parse(X, U).
i_parse(E^N, E0^N)	-->  i_parse(E, E0).
i_parse(regexlist(L), regexlist(L0))   -->  i_parse_list(L, L0).
i_parse(@(X), U)	--> {regex_macro(X, Y)}, i_parse(Y, U).
i_parse("", [])		--> [].
i_parse(X, U)		--> {string(X)},
 	{parse_regex(X, Y)},
	i_parse(Y, U).
i_parse(X, U)		--> {atom(X)},  {atom_codes(X, Y)}, i_parse_codes(Y, U).
i_parse(X, U)		--> {number(X)}, {number_codes(X, Y)}, i_parse_codes(Y, U).
i_parse(X, _)		--> {throw(i_parse:'unknown hybrid regex found'(X))}.

%
i_parse_list([E|Es], [F|Fs]) --> i_parse(E, F),
	i_parse_list(Es, Fs).
i_parse_list([], [])	-->  [].

%
i_parse_codes([X], 'I'(U))  --> [[X-X] - U].
i_parse_codes([X, Y|Z], 'I'(U) + R) --> [[X-X]- U],
	i_parse_codes([Y|Z], R).
i_parse_codes([], [])  --> [].

% Note.
% ?- eval(:(=) @ (:(=) @ (:(=) @ 1)), V).

% ?- am_linear([a,b], X).
% ?- am_linear([], X).

am_linear(Cs, coa(E, 1)):- once(am_linear(Cs, 1, E, [])).

am_linear([], N, [N-[[]]|L], L).
am_linear([A|R], N, [N-[A-N0]|L], L0):- N0 is N+1,
 	am_linear(R, N0, L, L0).

%%  char_boole_form(+F:char_exp, -F0:char_exp) is det
%
%   True if F is converted to F0 such that all boolean expressions on intevals
%   are evaluated.
%
%   A simple usage:
%	?- char_boole_form(a|b, X).
%	X = ([97-97];[] | [98-98];[]).

char_boole_form((.), [inf-sup]).
char_boole_form([], []).
char_boole_form(X-Y, [A0-A1]):-
	(atom(X), atom_length(X, 1); integer(X)),
	(atom(Y), atom_length(Y, 1); integer(Y)),
	( atom(X) -> char_code(X, X0); X0=X ),
	( atom(Y) -> char_code(Y, Y0); Y0=Y ),
	msort([X0,Y0], [A0,A1]).
char_boole_form(X, Y):- atom(X), reserved_char_class(Assoc),
	memberchk(X-C, Assoc),
	maplist(char_boole_form_aux, C, C0),
	char_boole_form(C0, Y).
char_boole_form(X, Y):- atom(X), atom_codes(X, Y0),
	char_boole_form(Y0, Y).
char_boole_form(X, [X-X]):- integer(X).
char_boole_form([X|Y], X0;Y0):- char_boole_form(X, X0),
	char_boole_form(Y, Y0).
char_boole_form(dot(X), Y):- char_boole_form(X, Y).
char_boole_form(out(X), out(Y)):- char_boole_form(X, Y).
char_boole_form(^(X), Y):- char_boole_form(out(X), Y).
char_boole_form(X^N, Y^N):- char_boole_form(X, Y).
char_boole_form(\+(X), Y):- char_boole_form(out(X), Y).
char_boole_form(\(X), Y):- char_boole_form(out(X), Y).
char_boole_form(X|Y, X0|Y0):- char_boole_form(X, X0),
	char_boole_form(Y, Y0).
char_boole_form(X;Y, Z):- char_boole_form(X|Y, Z).
char_boole_form(\(X,Y), \(X0,Y0)):- char_boole_form(X, X0),
	char_boole_form(Y, Y0).
char_boole_form(X&Y, X0&Y0):- char_boole_form(X, X0),
	char_boole_form(Y, Y0).

%
char_boole_form_aux(A-B, A0-B0):-  char_code(A, A0), char_code(B, B0).
char_boole_form_aux(A, A0):- char_code(A, A0).


% ?- i_normal_class_form([-1, -2],  X).
%@ X = ([-1- -1];[-2- -2];[]) .

i_normal_class_form([], []).
i_normal_class_form(X, [X-X]):- integer(X).
i_normal_class_form([X-Y|Z], ([X-Y]; U)):- i_normal_class_form(Z, U).
i_normal_class_form([X|Z], ([X-X]; U)):- i_normal_class_form(Z, U).

%%% Tiny helpers
zip_hyphen([], [], []).
zip_hyphen([A|B], [C|D], [A-C|R]):- zip_hyphen(B, D, R).

%
zip_comma([], [], []).
zip_comma([A|B], [C|D], [(A,C)|R]):- zip_comma(B, D, R).

%
byte_in_bits(N, Bits):-  poly:bits(N, [], Bits0),
	length(Bits, 8),
	append(P, Bits0, Bits),
	maplist(=(0), P).


			/******************************************
			*     automaton sysnthesis basic steps    *
			******************************************/


%%  am_determine(+C:coa, -D:coa) is det
%
%   True if D is unified with a deterministic version of C.

am_determine(coa(E, I), D):- am_determine(E, I, E0, I0),
	am_fresh(coa(E0, I0), D0),
	am_clean(D0, D).

am_determine(E, I, D, [I]):-  once(am_determine([[I]], [], E, D, [])).

am_determine([], _, _, P, P).
am_determine([S|R], H, E, P, Q):-  memberchk(S, H),
	am_determine(R, H, E, P, Q).
am_determine([S|R], H, E, [S-C|P], Q):-
	expand_power_state(S, E, C, R0, R),
	am_determine(R0, [S|H], E, P, Q).

%%  am_slim(+C:coa, -D:coa) is det
%
%   True if D is unified with a slimmed version of C.

am_slim(coa(E, I), coa(D, I)):- once(am_slim([I], [], E, D, [])).

am_slim([], _, _, P, P).
am_slim([S|R], H, E, P, Q):-  memberchk(S, H),
	am_slim(R, H, E, P, Q).
am_slim([I|A], H, E, [I-U|P], Q):- once(select(I-U, E, E0)),
	once(am_slim(U, A, [I|H], E0, P, Q)).

%
am_slim([], A, H, E, P, Q):- am_slim(A, H, E, P, Q).
am_slim([_-G|R], A, H, E, P, Q):- am_slim(R, [G|A], H, E, P, Q).
am_slim([_|R], A, H, E, P, Q):-  am_slim(R, A, H, E, P, Q).

% ?- expand_power_state([x,y], [x-[a-2, b-3], y-[a-2, b-4, c-5]], R, A, []).
%@ R = [a-[2], b-[3, 4], c-[5]],
%@ A = [[2], [3, 4], [5]] .
% ?- expand_power_state([x,y], [x-[[], a-2, b-3], y-[a-2, b-4, c-5]], R, A, []).
%@ R = [[], a-[2], b-[3, 4], c-[5]],
%@ A = [[2], [3, 4], [5]] .
% ?- trace, expand_power_state([x,y], [x-[special(s), a-2, b-3], y-[a-2, b-4, c-5]], R, A, []).

expand_power_state(S, E, Coa, P, Q):-
	foldl(expand_power_state_aux(E), S, [], Pairs0),
	sort(Pairs0, Pairs),
	once(merge_pairs(Pairs, Coa, [], P, Q)).
%
expand_power_state_aux(E, K, L, M):-  memberchk(K-B, E),
    foldl(cons, B, L, M).

cons(X, Y, [X|Y]).

% ?-is_deterministic(coa([1-[a-1, b-1]], _)).
% ?-is_deterministic(coa([1-[a-1, a-1]], _)).
is_deterministic(coa(E, _)):- is_deterministic(E).

is_deterministic([_-A|R]):- \+ consecutive_same_key(A),
		is_deterministic(R).
is_deterministic([]).

%
consecutive_same_key([A-_, A-_|_]).
consecutive_same_key([_, A|B]):- consecutive_same_key([A|B]).


			/*************************************
			*     automata state minimization    *
			*************************************/

%%  am_minimum(+C:coa, -D:coa) is det
%
%   True if D is unified with a state minimized version of C.

% ?-  regex_coalgebra(".*****",  X).
%@ X = coa([1-[[], dot([inf-sup])-1]], 1, [1]) .
% ?- am_minimum([2-[[], (inf-sup)-3], 3-[[], (inf-sup)-3]], 2, R).
%@ R = coa([1-[[], inf-sup-1]], 1) .

% ?- am_minimum(coa([1-[a-2, b-1], 2-[a-1, b-2], 3-[a-3, b-1]],  1), D).
%@ D = coa([1-[a-1, b-1]], 1) .

am_minimum(coa(E,I), C):- once(am_minimum(E, I, C)).

am_minimum(E, I, coa(E3, J)):-  minimum_am_qmap(E, Qmap),
	(	Qmap == []
	->	E3 = E,
		J  = I
	; 	member(X-I0,  Qmap),
		memberchk(I, X),
		quotient_coa(E, Qmap, E1),
		am_clean(coa(E1, I0), coa(E2, J)),
		sort(E2, E3)
	).

make_singleton(I, [I]).
%
minimum_am_qmap(Coa, Qmap):-
	once(am_conflict_pairs(Coa, Conflicts)),
	sort(Conflicts, CSorted),
	once(pairs_to_assoc(CSorted, Inicon, [])),
	all_states(Coa, All),
	assoc_product(All, All, Prod, []),
	once(assoc_subtract(Prod, Inicon, Rel)),
	maplist(make_singleton, All, Singletons),
	once(coa_union_find(Rel, Singletons, Clusters)),
	(	Clusters == []
	->	Qmap = []
	;	length(Clusters, N),
		numlist(1, N, S),
		zip_hyphen(Clusters, S, Qmap)
	).

%
% quotient_coa(Eqs, Qmap, Eqs0) :-
% 	maplist(pred(Qmap, [S-L, S0-L0]:-
% 		(	member(Cluster-S0, Qmap),
% 			memberchk(S, Cluster),
% 			maplist(pred(Qmap,
% 					([A-G, A-G0]:-
% 						(member(GCluster-G0, Qmap),
% 						 memberchk(G, GCluster)))
% 					&  [X, X] ),
% 				    L, L1 ),
% 			predsort(compare_right, L1, L0)
% 		)), Eqs, Eqs0).

%
quotient_coa(Eqs, Qmap, Eqs0) :- maplist(quotient_coa_aux(Qmap), Eqs, Eqs0).

%
quotient_coa_aux(Qmap, S-L, S0-L0) :-
			member(Cluster-S0, Qmap),
			memberchk(S, Cluster),
			maplist(quotient_coa_aux_aux(Qmap), L, L1),
			predsort(compare_right, L1, L0).

%
quotient_coa_aux_aux(Qmap, A-G, A-G0) :-
	member(GCluster-G0, Qmap),
	memberchk(G, GCluster).
quotient_coa_aux_aux(_, X, X).

%
am_normal(coa([], _), C):- am_empty(C).
am_normal(coa(A,I), coa(B, I)):-  sort(A, C), am_normal(C, B, []).

%
am_normal([], P, P).
am_normal([I-A|R], [I-B|P], Q):-
	am_normal(I, R, R0, As, []),
	append([A|As], H),
	sort(H, G),
	am_normal_body(G, B),
	am_normal(R0,  P,  Q).

am_normal_body([], []).
am_normal_body([[]|G], [[]|B]) :- am_normal_body(G, B).
am_normal_body([special(X)|G], [special(X)|B]) :- am_normal_body(G, B).
am_normal_body(G, B) :-  keysort_right(G, G0),
	join_char(G0, B0),
	maplist(am_normal_body_aux, B0, B).

%
am_normal_body_aux(I-X, J-X):- x_normal(I, J).

%
am_normal(I, [I-A|R], R0, [A|P], Q):- am_normal(I, R, R0, P, Q).
am_normal(_, R, R, P, P).

%
compare_right(C, X-Z, Y-Z):- compare(C, X, Y).
compare_right(C, _-X, _-Y):- compare(C, X, Y).
compare_right(=, X, X).
compare_right(<, _, _-_).
compare_right(>, _-_, _).
compare_right(>, _, _).

% ?- x_compare(C, x(2), 1-2).
x_compare(C,  x(X), x(Y)):- compare(C, X, Y).
x_compare(<,  x(_), _-_).
x_compare(>,  _-_, x(_)).
x_compare(C,  A, B):- iboole:i_compare(C, A, B).

%
x_normal(X, Y):- x_separate(X, X0, R),
	iboole:i_normal(R, R0),
	append(X0, R0, Y).

x_separate([], [], []).
x_separate([A-B|R], U, [A-B|V]):-  x_separate(R, U, V).
x_separate([X|R], [X|U], V):- x_separate(R, U, V).


% ?- merge_pairs([a-2, a-3, a-2,  b-1, b-3, c-2], R, [], S, []).
%@ R = [a-[2, 3], b-[1, 3], c-[2]],
%@ S = [[2, 3], [1, 3], [2]].

merge_pairs([I-X|Cs], [I-Pow|D0], D, [Pow|A0], A):-
	power_state(I, Cs, Cs0, Pow0, []),
	sort([X|Pow0], Pow),
	merge_pairs(Cs0, D0, D, A0, A).
merge_pairs([X|Cs], [X|D0], D, A0, A):-
	merge_pairs(Cs, D0, D, A0, A).
merge_pairs([], D, D, A, A).

%
power_state(I, [I-G|R], R0, [G|P], Q):-
	power_state(I, R, R0, P, Q).
power_state(_, A, A, B, B).

% ?- join_char_interval([], []).
% ?- join_char_interval([(3-4)-t], R).
%@ R = [[3-4]-t] .
% ?- coa:join_char_interval([(inf-4)-t, (5-sup)-t], R).
%@ R = [[inf-4, 5-sup]-t].
% ?- join_char_interval([(10-sup)-t, (20-20)-s, (inf-4)-t], R).
%@ R = [[20-20]-s, [10-sup, inf-4]-t].

keysort_right(X,Y):- predsort(keysort_right_aux, X, Y).
%
keysort_right_aux(=, [], []).
keysort_right_aux(<, [], _).
keysort_right_aux(>, _, []).
keysort_right_aux(C, _-U, _-V):- U @> V, C=(>); C=(<).

%
join_char_interval(A, B):- keysort_right(A, A0), join_char(A0, B).

%
join_char([], []).
join_char([[]|R], [[]|S]):- join_char(R, S).
join_char([X-A|R], [[X|Xs]-A|S]):-
	once(join_char(A, R, Xs, R0)),
	join_char(R0, S).

%
join_char(_, [], [], []).
join_char(A, [X-A|R], [X|Xs], S):-
	join_char(A, R, Xs, S).
join_char(_, R, [], R).

% ?- merge_target([a-1], R, []).
%@ R = [a-[1]].
% ?- merge_target([a-1, a-2], R, []).
%@ R = [a-[1, 2]] .
% ?- merge_target([a-1, a-2, b-3], R, []).
%@ R = [a-[1, 2], b-[3]] .
% ?- merge_target([[], a-1, a-2, b-3], R, []).
%@ R = [[], a-[1, 2], b-[3]] .
% ?- merge_target([[], a-1,  b-3, a-2], R).
%@ R = [[], a-[1, 2], b-[3]] .

merge_target(X, Y):- sort(X, X0), once(merge_target(X0, Y, [])).

merge_target([], P, P).
merge_target([[]|R], [[]|P], Q):- merge_target(R, P, Q).
merge_target([I-X|R], [I-Xs|P], Q):-
	merge_target(I, R, R0, [X], Xs),
	merge_target(R0, P, Q).

%
merge_target(I, [I-X|R], R0, P, Q):- contract_insert(X, P, P0),
	merge_target(I, R, R0, P0, Q).
merge_target(_, R, R, P, P).


			/********************
			*     union-find    *
			********************/
% ?- coa_union_find([a-b,x-y, x-x, y-z, b-c], [], R).
% ?- coa_union_find([a-b,x-y, x-x, y-z, b-c], [], R).

coa_union_find([], X, X).
coa_union_find([X-Ys|R],C,D):-coa_union_find(Ys, X, C,C1),
	coa_union_find(R,C1,D).

coa_union_find([], _, P, P).
coa_union_find([Y|Ys], X, P, Q):- once(coa_union_find_one(X, Y, P, P0)),
	coa_union_find(Ys, X, P0, Q).

coa_union_find_one(X,Y,Z,U):-find_cluster(X,Z,C,Z0),
	(memberchk(Y, C) -> U=[C|Z0]
	; find_cluster(Y, Z0, C0, Z1),
	  append(C,C0, C1),
	  U=[C1|Z1]
	).


% ?- find_cluster(a, [[a,b],[c,d]], C, X).
find_cluster(X,[],[X],[]).
find_cluster(X,[Y|Z],Y,Z):- memberchk(X,Y).
find_cluster(X,[Y|Z],U,[Y|V]):- find_cluster(X,Z,U,V).


	  /***************************
	  *     Reverse Automaton    *
	  ***************************/

% Reversing coalgebra
% ?- am_flip([a-[b-c], d-[b-c], e-[b-c], x-[y-z]], R).
%@ R = [c-[b-[a, d, e]], z-[y-[x]]] .
% ?- am_flip([a-[b-c], d-[b-c], e-[m-c], x-[y-z]], R).
%@ R = [c-[b-[a, d], m-[e]], z-[y-[x]]] .
% ?- am_flip([a-[[], b-c, d-e, e-c], x-[y-z]], R).
%@ R = [c-[b-[a], e-[a]], e-[d-[a]], z-[y-[x]]] .
% ?- am_flip([a-[d-e, e-c], x-[y-z]], R).
%@ R = [c-[e-[a]], e-[d-[a]], z-[y-[x]]] .
% ?- am_flip([1-[a-1]], R).
% ?- am_flip([], X).

am_flip --> am_flip_triples, triples_to_coa.

%
am_flip_triples(X, Y):- maplist(am_flip_triples_aux, X, X0), 	append(X0, Y).

%
am_flip_triples_aux(I-As, B):-
			fold_right(am_flip_triples_aux_aux(I), As, B, []).

%
am_flip_triples_aux_aux(I, C-J, [J-(C-I)|W], W).
am_flip_triples_aux_aux(_, _,  W, W).
%
fold_right(_, [], P, P).
fold_right(F, [A|As], P, Q):- call(F, A, Q0, Q), fold_right(F, As, P, Q0).


% ?- am_reverse(coa([1-[[]]], 1), R).
%@ R = coa([1-[[]]], 1) .
% ?- am_reverse(coa([1-[a-2], 2-[b-3], 3-[[]]], 1), R).
%@ R = coa([1-[[]], 2-[a-1], 3-[b-2]], 3) .

am_reverse(coa(E, I), Coa) :- am_flip(E,  E0),
	once((select(I-W, E0, E1);  W=[], E1=E0)),
	once(am_finals(coa(E, _), Is)),
	multi_entry_to_single(Is, [I-[[]|W]|E1], Coa0),
	am_fresh(Coa0, Coa1),
	am_clean(Coa1, Coa2),
	am_minimum(Coa2, Coa).

%
multi_entry_to_single(Is, E, coa(D, Js))  :-
	am_clean(coa(E, _), coa(F, _)),
	sort(Is, Js),
	once(power_state_closure([Js], F, [], D)).

% ?- power_state_closure([[1]], [1-[a-[1]]],  [], X).
% ?- power_state_closure([[1]], [1-[a-[2]], 2-[a-[1,2], b-[1]]], [], X).

power_state_closure([], _, P, P).
power_state_closure([S|R], E, P, Q):- memberchk(S-_, P),
	power_state_closure(R, E, P, Q).
power_state_closure([S|R], E, P, Q):-
	power_state_closure_step(S, R, R0, E, P, P0),
	power_state_closure(R0, E, P0, Q).

%
power_state_closure_step(S, R, R0, E, P, [S-A|P]):-
	merge_goto(S, E, [], A),
	foldl(power_state_closure_step_aux, A, R, R0).

%
power_state_closure_step_aux(_-G, U, V):- memberchk(G, U), V=U; V=[G|U].
power_state_closure_step_aux([], U, U).

% ?- merge_goto([1,2], [1-[a-[2,3]], 2-[a-[1,2], b-[3]]], [], X).
%@ X = [a-[1, 2, 3], b-[3]].
merge_goto([], _, A, A).
merge_goto([X|Xs], E, A, B):- memberchk(X-R, E),
	once(gdict_ord_merge(R, A, A0)),
	merge_goto(Xs, E, A0, B).

% ?- gdict_ord_merge([[], a-[1,2], c-[3]], [a-[2,3], b-[2]], R).
gdict_ord_merge([], X, X).
gdict_ord_merge(X, [], X).
gdict_ord_merge([I-A|X], [I-B|Y], [I-C|Z]):- ord_union(A, B, C),
	gdict_ord_merge(X, Y, Z).
gdict_ord_merge([U|X], [U|Y], [U|Z]):-  gdict_ord_merge(X, Y, Z).
gdict_ord_merge([U|X], [V|Y], [U|Z]):-  U@<V,
	gdict_ord_merge(X, [V|Y], Z).
gdict_ord_merge([U|X], [V|Y], [V|Z]):-  gdict_ord_merge([U|X], Y, Z).

%?- triples_to_coa([1-(a-4), 3-(c-3), 2-(b-1), 1-(a-2)], R).
%@ R = [1-[a-[2, 4]], 2-[b-[1]], 3-[c-[3]]].
triples_to_coa(X, Y):-sort(X, X0),
	pairs_to_assoc(X0, Assoc, []),
	maplist(triples_to_coa_aux, Assoc, Y).
%
triples_to_coa_aux(I-U, I-V):- pairs_to_assoc(U, V, []).

%
am_to_reversed_dag(E, D):- once(am_to_reversed_dag(E, [], Ps)),
	sort(Ps, Ps0),
	once(pairs_to_assoc(Ps0, D, [])).

%
am_to_reversed_dag([I-A|R], X, Y):- once(am_to_reversed_dag(A, I, X, X0)),
	am_to_reversed_dag(R, X0, Y).
am_to_reversed_dag([], X, X).

%
am_to_reversed_dag([_-G|R], I, X, Y):- am_to_reversed_dag(R, I, [G-I|X], Y).
am_to_reversed_dag([_|R], I, X, Y):-  am_to_reversed_dag(R, I, X, Y).
am_to_reversed_dag([], _, X, X).

% ?- am_conflict_pairs([a-[[],1-a], b-[1-b], c-[[],1-a]], S).
% ?- coa:am_conflict_pairs([a-[1-a], b-[1-b]], S).
% ?- coa:am_conflict_pairs([a-[[]], b-[[]]], S).
am_conflict_pairs(Coa, S):-
	am_arity_dict(Coa, Dict),
	once(arity_conflict_pairs(Dict, IniCon, [])),
	sort(IniCon, IniCon0),
	am_flip(Coa, RCoa),
	once(propagate_conflict(IniCon0, RCoa, [], S)).

% by agenda programming
propagate_conflict([], _, P, P).
propagate_conflict([U|R], Coa, P, Q):- memberchk(U, P),
	propagate_conflict(R, Coa, P, Q).
propagate_conflict([I-J|R], Coa, P, Q):-  memberchk(I-A, Coa),
	memberchk(J-B, Coa),
	fiber_product(A, B, F),
	append(F, R, R0),
	propagate_conflict(R0, Coa, [I-J|P], Q).
propagate_conflict([U|R], Coa, P, Q):-
	propagate_conflict(R, Coa, [U|P], Q).

% ?- fiber_product([a-[7,8,9], b-[1,2,3]], [a-[6,7,8], b-[1,2,3]],  P).
%@ P = [1-2, 1-3, 2-3, 6-7, 6-8, 6-9, 7-8, 7-9, 8-9] .

fiber_product(A, B, P):- once(fiber_product(A, B, [], P0)),
	sort(P0, P).

fiber_product([], _,  P, P).
fiber_product([X|Y], N,  P, Q):-
	fiber_product_one(X, N, P, R),
	fiber_product(Y, N, R, Q).
%
fiber_product_one([], _, P, P).
fiber_product_one(_, [], P, P).
fiber_product_one(X, [Y|Z], P, Q):-
	fiber_product_one_one(X, Y, P, R),
	fiber_product_one(X, Z, R, Q).
%
fiber_product_one_one(_, [], P, P).
fiber_product_one_one(A-G, A-H, P, Q):-
	s_product(G, H, P, Q).
fiber_product_one_one(_, _, P, P).

% ?- module(pac_word).
% ?- pairs_to_assoc([1-[], 1-2, 1-3, 2-3, 2-4], X, []).
%@ X = [1-[2, 3, []], 2-[3, 4]].
pairs_to_assoc([], X, X).
pairs_to_assoc([I-A|R], [I-As|V], V0):-
	once(pairs_to_assoc(I, R, R0, [A], As)),
	pairs_to_assoc(R0, V, V0).

%
pairs_to_assoc(I, [I-A|R], R0,  S, T):-
	once(contract_insert(A, S, S0)),
	pairs_to_assoc(I, R, R0, S0, T).
pairs_to_assoc(_, R, R,  S, S).

% ?- contract_insert(3, [2, 5, 7], R).
%@ R = [2, 3, 5, 7].

contract_insert(X, [], [X]).
contract_insert(X, [X|S], [X|S]).
contract_insert(X, [Y|S], [X, Y|S]):- X@<Y.
contract_insert(X, [Y|S], [Y|R]):- contract_insert(X, S, R).


% ?- all_states([1-[[],a-b], 2-[c-d]], R),
all_states(X, S):- maplist(all_states_aux, X, S0), sort(S0, S).
%
all_states_aux(A-_, A).

%
arity(X, Y):- maplist(arity_aux, X, Y0), sort(Y0, Y).
%
arity_aux([], []).
arity_aux(A-_, A).

%
am_arity_dict(Coa, Dict):- maplist(am_arity_dict_aux, Coa, Dict).
%
am_arity_dict_aux(I-X, I-M):- arity(X, M).

% ?- arity_conflict_pairs([1-a, 2-b, 3-a, 4-b], A, []).
arity_conflict_pairs([I-A|As], P, Q):- arity_conflict_pairs(I, A, As, P, P0),
	arity_conflict_pairs(As, P0, Q).
arity_conflict_pairs([], A, A).

%
arity_conflict_pairs(_, _, [], P, P).
arity_conflict_pairs(I, A, [J-B|Bs], [I-J|P], Q):- A\==B,
	arity_conflict_pairs(I, A, Bs, P, Q).
arity_conflict_pairs(I, A, [_|Bs], P, Q):-
	arity_conflict_pairs(I, A, Bs, P, Q).

% ?-s_product([1,2,3],[1,2,3], P).
%@ P = [1-2, 1-3, 2-3] .

s_product(X, Y, P):- s_product(X, Y, [], Q), sort(Q, P).
%
s_product([], _, P, P).
s_product([X|Xs], Y, P, Q):-
	s_product(X, Xs, Y, P, R),
	s_product(Xs, Y, R, Q).
%
s_product(X, Xs, Y, P, Q):-
	foldl(s_product_aux(X), Y, P, R),
	s_product(Xs, Y, R, Q).

%
s_product_aux(X, A, L, L):- A==X, !.
s_product_aux(X, A, L, [W|L]):- s_pair(X, A, W).

% ?- assoc_product([1,2,3], [1,2,3], R, []).
%@ R = [1-[2, 3], 2-[3]] .

assoc_product([], _, P, P).
assoc_product(_, [], P, P).
assoc_product([X|Xs], [Y|Ys], [X-[Y|Ys]|P], Q):- Y@>X,
	assoc_product(Xs, [Y|Ys], P, Q).
assoc_product(Xs, [_|Ys], P, Q):-assoc_product(Xs, Ys, P, Q).

%  subtraction on assoc lists.
% ?- assoc_subtract([1-[2,3,4], 2-[3,4], 3-[4,5,6]], [1-[3], 2-[4], 5-[6,7,8]], R).
%@ R = [1-[2, 4], 2-[3], 3-[4, 5, 6]] .

assoc_subtract([], _, []).
assoc_subtract(A, [], A).
assoc_subtract([I-A|Ps], [I-B|Qs], R):- once(ord_subtract(A, B, C)),
	(	C == []
	->      R = Rs
	;	R = [I-C|Rs]
	),
	once(assoc_subtract(Ps, Qs, Rs)).
assoc_subtract([I-A|Ps], [J-B|Qs], [I-A|Rs]):-  I@<J,
	once(assoc_subtract(Ps,[J-B|Qs], Rs)).
assoc_subtract([I-A|Ps], [_|Qs], [I-A|Rs]):-
	once(assoc_subtract(Ps, Qs, Rs)).

% ?-assoc_complement([1,2,3], [1-[2], 2-[3]], R).
%@ R = [1-[3]] .

assoc_complement(A, C, D):- assoc_product(A, A, AConf, []),
	once(assoc_subtract(AConf, C, D)).

%
s_pair(A, B, A-B):- A @< B.
s_pair(A, B, B-A).

%
am_char([], coa([1-[[]]], 1)).
am_char(Is, coa([1-[[]|G], 2-[[]]], 1)):- maplist(am_char_aux, Is, G).

am_char_aux(I, I-2).

am_char_simple([], coa([1-[[]]], 1)).
am_char_simple(Is, coa([1 - G, 2-[[]]], 1)):-
	maplist(am_char_simple_aux, Is, G).
%
am_char_simple_aux(I, I-2).


			/***************************************
			*     automaton sysnthesis basics 2    *
			***************************************/

%%  am_cap(+X:coa, +Y:coa, -Z:coa) is det
%
%   True if Z is unified with a coa intersection of X and Y.

% digraph traversal by Agenda Programming.
%
am_cap(coa(E0, I0), coa(E1, I1), Coa):-
	once(am_cap([I0*I1], [], E0, E1, [], E2)),
	am_fresh(coa(E2, I0*I1), Coa0),
	once(am_minimum(Coa0, Coa)).

%
am_cap([], _, _, _,  P, P).
am_cap([U|R], H, A, B, P0, P):- memberchk(U, H),
	am_cap(R, H, A, B, P0, P).
am_cap([I*J|R], H, A, B,  P0, P):- memberchk(I-GI, A),
	memberchk(J-GJ, B),
	once(product_cap(GI, GJ, R0, R, Q, [])),
	am_cap(R0, [I*J|H], A, B, [I*J-Q|P0], P).

%
product_cap([], _, R, R, Q, Q).
product_cap(_, [], R, R, Q, Q).
product_cap([[]|G], [[]|H], R0, R, [[]|Q0], Q):-
	product_cap(G, H, R0, R, Q0, Q).
product_cap([[]|G], H, R0, R, Q0, Q):-
	product_cap(G, H, R0, R, Q0, Q).
product_cap(G, [[]|H], R0, R, Q0, Q):-
	product_cap(G, H, R0, R, Q0, Q).
product_cap([U-I|G], [U-J|H], [I*J|R0], R, [U-I*J|Q0], Q):-
	product_cap(G, H, R0, R, Q0, Q).
product_cap([U-_|G], [V-I|H], R0, R, Q0, Q):- U@<V,
	product_cap(G, [V-I|H], R0, R, Q0, Q).
product_cap([U-I|G], [_|H], R0, R, Q0, Q):-
	product_cap([U-I|G], H, R0, R, Q0, Q).

%%  am_minus(+X:coa, +Y:coa, -Z:coa) is det
%
%   True if Z is unified with a coa subtraction of X with Y.

am_minus(coa(E0,I0), coa(E1, I1), Coa):-
	once(am_minus([I0*I1], [], E0, E1, [], E2)),
	append(E0, E2, E3),
	am_determine(coa(E3, I0*I1), NCoa),
	once(am_minimum(NCoa, Coa)).

%
am_minus([], _, _, _,  P, P).
am_minus([U|R], H, A, B, P0, P):- memberchk(U, H),
	am_minus(R, H, A, B, P0, P).
am_minus([I*J|R], H, A, B,  P0, P):- memberchk(I-GI, A),
	memberchk(J-GJ, B),
	once(product_minus(GI, GJ, R0, R, Q, [])),
	once(am_minus(R0, [I*J|H], A, B, [I*J-Q|P0], P)).

%
product_minus([], _, R, R, Q, Q).
product_minus([W|G], [], R0, R, [W|P], Q):-
	product_minus(G, [], R0, R, P, Q).
product_minus([[]|G], [[]|H], R0, R, Q0, Q):-
	product_minus(G, H, R0, R, Q0, Q).
product_minus([[]|G], H, R0, R, [[]|Q0], Q):-
	product_minus(G, H, R0, R, Q0, Q).
product_minus(G, [[]|H], R0, R, Q0, Q):-
	product_minus(G, H, R0, R, Q0, Q).
product_minus([U-I|G], [U-J|H], [I*J|R0], R, [U-I*J|Q0], Q):-
	product_minus(G, H, R0, R, Q0, Q).
product_minus([U-I|G], [V-J|H], R0, R, [U-I|Q0], Q):- U@<V,   % bug fix [2014/08/23]
	product_minus(G, [V-J|H], R0, R, Q0, Q).
product_minus([U-I|G], [_|H], R0, R, Q0, Q):-
	product_minus([U-I|G], H, R0, R, Q0, Q).


%%  am_cup(+X:coa, +Y:coa, -Z:coa) is det
%
%   True if Z is unified with a coa union of X and Y.

am_cup(coa(A,I), Coa0,  Coa):-
	length(A, N),
	am_shift(N, Coa0, coa(C, K)),
	append(A, C, D),
	once(am_determine([[I,K]], [], D,  E, [])),
	once(am_fresh(coa(E, [I,K]), Coa1)),
	once(am_minimum(Coa1, Coa)).


%%  am_concat(+X:coa, +Y:coa, -Z:coa) is det
%
%   True if Z is unified with a concatenation of X and Y.

% ?- am_concat(coa([1-[[], a-2], 2-[[]]],1),  coa([1-[a-2], 2-[[]]],1), R).
% ?- am_concat(coa([1-[[], a-1]],1),  coa([1-[[], b-1]],1), R).

am_concat(coa(E0, I0), Coa1, Coa):-once((
	length(E0, N),
	am_shift(N, Coa1, coa(E3, I3)),
	memberchk(I3-G, E3),
	maplist(am_replace_null(G), E0, E1),
	append(E1, E3, E4),
	am_slim(coa(E4,I0), E5),
	am_determine(E5, E),
	am_minimum(E, Coa))).

%%  am_power(+N:integer, +X:coa, -Y:coa) is det
%
%   True if Y is unified with a N times coa concatenation of X.

am_power(N, X, Y):- poly:bits(N, [], Bs),
	am_unit(U),
	once(am_power(Bs, X, U, Y)).

%
am_power([], _, X, X).
am_power([0|Bs], A, X, Y):-
	am_concat(X, X, X2),
	am_power(Bs, A, X2, Y).
am_power([_|Bs], A, X, Y):-
	am_concat(X, X, X2),
	am_concat(A, X2, AX2),
	am_power(Bs, A, AX2, Y).

%
am_power_lower(N, E, U):- am_unit(I),
	am_cup(I, E, E0),
	am_power(N, E0, U).

am_power_upper(N, Coa, Coa0):-
	am_power(N, Coa, Coa1),
	am_star(Coa, Coa2),
	am_concat(Coa1, Coa2, Coa0).

am_power_between(I, J, Coa, Coa0):-
	am_power(I, Coa, Coa1),
	K is J - I,
	am_power_lower(K, Coa, Coa2),
	am_concat(Coa1, Coa2, Coa0).
%
am_unit(coa([1-[[]]], 1)).
%
am_empty(coa([], 0)).

			/**************************
			*     special automata    *
			**************************/

% ?- coa:am_shift(3, coa([1-[a-2, b-3]],1), R).
%@ R = coa([4-[a-5, b-6]], 4).

am_shift(N, coa(E,I), coa(E0, I0)):-
	I0 is I+N,
	maplist(state_id_shift(N), E, E0).

%
state_id_shift(N, I-A, J-B):- J is I + N,
	maplist(goto_id_shift(N), A, B).

%
goto_id_shift(N, A-I, A-J):- J is I+N.
goto_id_shift(_, X, X).

%  ?- am_replace_null([a-2], 2-[[]], X).
am_replace_null(A, I-[[]|B], I-C):-  ord_union(A, B, C).
am_replace_null(_, P, P).


%%  am_kleene(+X:am, -Y:am) is det
%
%   True if Y is unified with a kleene closure of X.

% ?- am_minimum(coa([1-[a-2, a-3, b-3], 2-[b-3], 3-[a-1]], 1), R).
% ?- am_plus(coa([1-[a-2], 2-[[]]], 1), R).
% ?- am_star(coa([1-[a-2], 2-[[]]], 1), R).
% ?- am_star(coa([1-[a-2, b-3], 3-[[]], 2-[[]]], 1), R).
% ?- am_star(coa([1-[a-2, a-3, b-3], 3-[[]], 2-[[]]], 1), R).
% ?- am_star(coa([1-[a-2], 2-[[]]], 1), R).
%@ R = coa([1-[[], a-1]], 1) .
% ?- am_star(coa([1-[a-2, b-2], 2-[[]]], 1), R).
%@ R = coa([1-[[], a-1, b-1]], 1) .
% ?- am_plus(coa([1-[a-2, b-2], 2-[[]]], 1), R).
%@ R = coa([1-[a-2, b-2], 2-[[], a-2, b-2]], 1) .
% ?- am_copy(coa([1-[a-2], 2-[[]]], 1), D).


am_kleene(coa(E, I), Coa):-
	once(select(I-R, E, E0)),
	once(drop_null(R, NR)),
	maplist(am_replace_null([[]|NR]), E0, E1),
 	once(am_determine(coa([I-R|E1], I), Coa)).

%
am_star(C, D):- am_kleene(C, coa(E0, I)),
	once(select(I-R0, E0, E)),
	(	R0=[[]|_]
	->	R = R0
	;	R = [[]|R0]
	),
	once(am_minimum(coa([I-R|E], I), D)).

%
am_plus(X, Y):- once((am_kleene(X, Z), am_minimum(Z, Y))).

%
am_fresh(coa([], _), C):- am_empty(C).
am_fresh(coa(E, I), coa(E0, I0)):-
	length(E, L),
	numlist(1, L, Ns),
	zip_hyphen(M, _, E),
	zip_hyphen(M, Ns, S),
	subst_coa(E, S, E0),
	memberchk(I-I0, S).

%
am_fresh(Coa0, coa(E0, _), Coa):-  length(E0, N), am_shift(N, Coa0, Coa).

%
am_size(coa(E, _), N):- length(E, N).

am_initial_state(coa(_, I), I).

am_equations(coa(E, _), E).

% ?- am_finals(coa([1-[[]], 2-[a-3], 3-[[]]], 1), R).
%@ R = [1, 3] .
% am_finals(coa(E, _), A):-
% 	foldl(pred(	([I-H, [I|U], U] :- once(is_final_state(H)))
% 		  &	[_, U, U]),
% 		  E, A, []).

am_finals(coa(E, _), A):- foldl(am_finals_aux, E, A, []).

%
am_finals_aux(I-H, [I|U], U) :- once(is_final_state(H)).
am_finals_aux(_, U, U).

%
is_final_state(H):- memberchk([], H).
is_final_state(H):- member(special(Spe), H),
	final_special_pred(Spe).

%  ?- am_states(coa([1-[[]], 2-[a-3], 3-[[]]], 1), R).
%@ R = [1, 2, 3].
am_states(coa(E, _), A):- maplist(am_states_aux, E, A).
%
am_states_aux(I-_, I).
%
am_copy(C,  D):-  C = coa(E0, _),
	length(E0, N),
	am_shift(N, C, D).
%
am_clean(coa(E, I),  coa(D, I)):- maplist(am_clean_aux,  E, E0),
	sort(E0, D).

am_clean_aux(J-A, J-B):-sort(A, B).

%
drop_null([[]|A], A).
drop_null(A, A).

%
add_one(X, Y, Z):-  ( memberchk(X, Y)  ->  Z=Y ;   Z = [X|Y]).

% ?-subst_coa([1-[b-2]], [1-9, 2-10], X).
%@ X = [9-[b-10]].
% ?-subst_coa([1-[[], b-2]], [1-9, 2-10], X).
%@ X = [9-[[], b-10]].
% ?- am_remove_useless_states(coa([1-[[], a-1, a-2], 2-[b-3], 3-[c-3]], 1), R).
%@ R = coa([1-[[], a-1]], 1) .

am_remove_useless_states(coa(E, I), coa(E0, I)):-
	am_live_dead(coa(E,I), _, Dead),
	am_remove_states(E, Dead, E0).

am_live_dead(Coa, Live, Dead):- am_finals(Coa, Fs),
	am_equations(Coa, E),
	am_to_reversed_dag(E, Assoc),
	once(dg_path_find(Fs, Assoc,  [], Live, Dead)).

% ?- dg_path_find([], [1-[2], 2-[1]], [], X, Y).
%@ X = [],
%@ Y = [1, 2].
% ?- dg_path_find([1], [1-[2], 2-[1], 3-[1,2]], [], X, Y).
%@ X = [2, 1],
%@ Y = [3].

dg_path_find([], E, Live, Live, Dead):- maplist(dg_path_find_aux, E, Dead).
dg_path_find([X|R], E,  Z,  Live, Dead):- memberchk(X, Z),
	dg_path_find(R, E, Z, Live, Dead).
dg_path_find([X|R], E, Z, Live, Dead):- select(X-G, E, E0),
	union(G, R, R0),
	dg_path_find(R0, E0, [X|Z], Live, Dead).
dg_path_find([X|R], E, Z, Live, Dead):- dg_path_find(R, E, [X|Z], Live, Dead).

dg_path_find_aux(I-_, I).

% ?- am_connected_region(coa([1-[[]]], 1), C).
% ?- am_connected_region(coa([1-[[], a-2], 2-[[]], 3-[a-3]], 1), C).

am_connected_region(coa(E, I), coa(E0, I)):-
	once(am_path_find([I], [], E, E0, [])).

%
am_path_find([],_H, _E, R, R).
am_path_find(_, _H, [],  R, R).
am_path_find([X|R], H, STS,  E0, E1):- memberchk(X, H),
	am_path_find(R, [X|H], STS, E0, E1).
am_path_find([X|R], H, STS,  [X-G|E0], E1) :- select(X-G, STS, STS0),
	once(am_path_find_next(G, R0, R)),
	am_path_find(R0, [X|H], STS0, E0, E1).
am_path_find([X|R], H, STS,  [X|E0], E1) :- am_path_find(R, [X|H], STS, E0, E1).

%
am_path_find_next([_-G|R], [G|X], Y):- am_path_find_next(R, X, Y).
am_path_find_next([_|R], X, Y):- am_path_find_next(R, X, Y).
am_path_find_next([], X, X).

% naive method.

% ?- am_remove_empty_states(coa([], 1), C).
%@ C = coa([], 1).
% ?- am_remove_empty_states(coa([1-[[],a-2], 2-[]], 1), C).
%@ C = coa([1-[[]]], 1) .
% ?- am_remove_empty_states(coa([1-[[],a-2, a-3], 3-[b-2], 2-[]], 1), C).
%@ C = coa([1-[[]]], 1) .

am_remove_empty_states(coa(E, I), coa(E0,I)):-	am_elim_empty_states(E, E0).

%
am_elim_empty_states(E, E0):- am_elim_empty_states(E, S, [], E1, []),
	(	S == []
	->	E0 = E1
	;   	once(am_elim_link(S, E1, E2)),
		once(am_elim_empty_states(E2, E0))
	).

%
am_elim_empty_states([], X, X, Y, Y).
am_elim_empty_states([I-[]|R], [I|X], X0, Y, Y0):-
	am_elim_empty_states(R, X, X0, Y, Y0).
am_elim_empty_states([U|R], X, X0, [U|Y], Y0):-
	am_elim_empty_states(R, X, X0, Y, Y0).

%
am_elim_link([], E, E).
am_elim_link(Is, E, E0):- once(am_elim_link(Is, E, E0, [])).

%
am_elim_link(_, [], E, E).
am_elim_link(Is, [I-A|R], [I-B|E], F):- once(elim_link_step(Is, A, B, [])),
	am_elim_link(Is,  R, E, F).

%
elim_link_step(_,  [], X, X).
elim_link_step(Is, [[]|R], [[]|X], Y):- elim_link_step(Is, R, X, Y).
elim_link_step(Is, [_-S|R], X, Y):-  memberchk(S, Is),
	elim_link_step(Is, R, X, Y).
elim_link_step(Is, [I-S|R], [I-S|X], Y):- elim_link_step(Is, R, X, Y).

% ?- am_remove_states([1-[a-2,b-2]], [1,2], R).
% ?- am_remove_states([1-[[],a-2,b-2], 2-[b-1]], [1], R).
am_remove_states([], _, []).
am_remove_states([I-_|R], Ds, E):- memberchk(I, Ds),
	am_remove_states(R, Ds, E).
am_remove_states([I-A|R], Ds, [I-B|E]):- once(am_remove_goto(A, Ds, B)),
	am_remove_states(R, Ds, E).

%
am_remove_goto([], _, []).
am_remove_goto([[]|R], Ds, [[]|S]):- am_remove_goto(R, Ds, S).
am_remove_goto([_-G|R], Ds, S):- memberchk(G, Ds),
	am_remove_goto(R, Ds, S).
am_remove_goto([U|R], Ds, [U|S]):- am_remove_goto(R, Ds, S).

%
subst_coa([], _, []).
subst_coa([I-G|R], S, [I0-G0|R0]):- memberchk(I-I0, S),
	subst_am_goto(G, S, G0),
	subst_coa(R, S, R0).

%
subst_am_goto([A-I|R], S, [A-J|R0]):-  memberchk(I-J, S),
	subst_am_goto(R, S, R0).
subst_am_goto([X|R], S, [X| R0]):- subst_am_goto(R, S, R0).
subst_am_goto([], _, []).

			/****************************
			*     characters classes    *
			****************************/

:- meta_predicate cook_char_dict(?, ?, 1, ?).
% ?- cook_char_dict(0, 127, atom, Dict),  maplist(([X]:- writeq(X), nl),  Dict).
cook_char_dict(N0, N, Filter, 	Dict):- code_types(N0, N, Us),
	collect(Filter, Us, Ts),
	maplist(cook_char_dict_aux(N0, N), Ts, Dict).

%
cook_char_dict_aux(N0, N, T, T-CT):- collect_chars(N0, N, T, CT).

%
code_types(N0, N, S):- setof(T, C^ ( between(N0, N, C), code_type(C, T)), S).

% ?- collect_chars(0, 10000, alpha, X).
collect_chars(N0, N, CC, X):-
	setof(C, (code_type(C, CC), between(N0, N, C)), S0),
	zip_hyphen(S0, S0, S),
	iboole:i_merge(S, Z0),
	maplist(collect_chars_aux, Z0, X).

collect_chars_aux(V0-W0,  V-W):- char_code(V, V0), char_code(W, W0).

%
reserved_char_class([
	 alnum - ['0'-'9','A'-'Z',a-z],
	 alpha - ['A'-'Z',a-z],
	 ascii - ['\000\'-'\177\'],
	 cntrl - ['\000\'-'\037\','\177\'-'\177\'],
	 csym - ['0'-'9','A'-'Z','_'-'_',a-z],
	 csymf - ['A'-'Z','_'-'_',a-z],
	 digit - ['0'-'9'],
	 end_of_line - ['\n'-'\r'],
	 graph - [(!)-(~)],
	 lower - [a-z],
	 newline - ['\n'-'\n'],
	 period - [(!)-(!), ('.')-('.'), (?)-(?)],
	 prolog_atom_start - [a-z],
	 prolog_identifier_continue - ['0'-'9','A'-'Z','_'-'_',a-z],
	 prolog_symbol - [(#)- ($), (&)- (&), (*)- (+),
			 (-)-(/), (:)-(:), (<)-(@),
			 (\)-(\), (^)-(^), (~)-(~)],
	 prolog_var_start - ['A'-'Z','_'-'_'],
	 punct - [(!)-(/), (:)-(@),'['-('`'),'{'-(~)],
	 quote - ['"'-'"','\''-'\'', ('`')-('`')],
	 space - ['\t'-'\r',' '-' '],
	 upper - ['A'-'Z'],
	 white - ['\t'-'\t',' '-' '],
	 utf8 - ['\200\' -  sup ],		% '
	 hiragana - ['ぁ'-'ゔ'],
	 katakana - ['ァ'-'ヺ'],
	 kanji	- ['一'-'龠']
]).

utf8_byte_char_class([
	 utf8   - ['\200\' - '\377\'],
	 utf8c  - ['\200\' - '\277\'],
	 utf8b  - ['\300\' - '\377\']
	       ]).

% ?-  C is 0xc0,  char_code(X, C).

%
look_ahead(C, [C|X], [C|X]).


			/**************************************
			*  compiling coalgebra into clause    *
			**************************************/

%%  expand_coa(+K:arity, +C:coa, +I:int, +M:atom, -G:goal,
%              +P:list, -Q:list) is det
%
%   True if G  is unified with a compiled
%   version of C with (K+2)-ary state predicates in P (minus Q).
%
%   A sample use:
%	?- expand_coa(0, [1-[[], [10-10]-1]], 1, user, G, P, []).
%	G = user:'nt#52',
%	P = [	user: ('nt#52'(_G6838, _G6838):-true),
%		user: ('nt#52'([_G6868|_G6875], _G6877):-
%					_G6868==10,
%					'nt#52'(_G6875, _G6877))] .

%	?- expand_coa(0, [1-[[], [10-10]-1]], 1, user, G, P, []).
expand_coa(K, C, I, M, M_G, P, Q):-
	all_states(C, Nts),
 	maplist(expand_coa_aux, Nts, Assoc),
	memberchk(I-GI, Assoc),
	void_states(C, Vs),
	remove_void_state(C, I, C0),
	compile_coa(C0, Vs, K, M, P, Q, Assoc),
	prefix_convention(M, GI, M_G).

expand_coa_aux(J, J-N):- new_nt_name(N).

%
remove_void_state([], _, []).
remove_void_state([J-[[]]|C], Init, R):- J\==Init, !,
	remove_void_state(C, Init, R).
remove_void_state([X|C], Init, [X|R]):-
	remove_void_state(C, Init, R).

% ?- new_nt_name(X).
new_nt_name(N):- var(N), nt_name_prefix(Prefix),
				 gensym(Prefix, N).
new_nt_name(_).

%
nt_name_prefix(N):- nb_current(nt_name_prefix, N),
					 N \== [],
					 !.
nt_name_prefix('nt#').

%
insert_semicolon([X], X).
insert_semicolon([X, Y|Z], X;U):- insert_semicolon([Y|Z], U).

%
compose_relation([X, Y], [X, X0], [X0, Y]).
compose_relation([A, B, X, Y], [A, A0, X, X0], [A0, B, X0, Y]).

prefix_convention([], P, P).
prefix_convention(M, P, M:P).

%
void_states([], []).
void_states([I-[[]]|R], [I|R0]):- void_states(R, R0).
void_states([_|R], R0):- void_states(R, R0).

%%  compile_coa(+K:arity,+Vs:list,  +C:coa, +I:int, +M:atom, -G:goal,
%	       +P:list, -Q:list, +A:assoc) is det
%
%   True if G  is unified with a compiled  version of C
%   with (K+2)-ary state predicates in P (minus Q). State
%   id numbers are replaced by A.
%
% ?- compile_coa([1-[[], [10-20, 22-30]-1]], [], 2, user, P, [], [1- a]).
compile_coa([],_, _, _, P, P, _).
compile_coa([I-A|R], Vs, N, M,  P, Q, Assoc):-
	memberchk(I-I0, Assoc),
	once(compile_coa_state(I0, Vs, A, N, M, P, P0, Assoc)),
	compile_coa(R, Vs, N, M, P0, Q, Assoc).
%
compile_coa_state(I, Vs,[X|A], N, M, P, Q, Assoc):-
	once(compile_coa_state_one(I, Vs, X, N, M, P, P0, Assoc)),
	compile_coa_state(I, Vs, A, N, M, P0, Q, Assoc).
compile_coa_state(_, _, [], _, _, P, P, _).

%
compile_coa_state_one(I, _, [],  N, M, [R|P], P, _):-
	compile_coa_epsilon_rule(N, I, R0),
	once(prefix_convention(M, R0, R)).
compile_coa_state_one(I, _, special(X), N, M, [Pred|P], P, _):-
	once(compile_special_pred(X, [I, _, N, M], Pred)).
compile_coa_state_one(I, Vs, A-J, N, M, P,  Q, Assoc):- is_list(A),
	(	memberchk(J, Vs)
	->	J0 = void_state
	;  memberchk(J-J0, Assoc)
	),
	compile_coa_rule(A, I, J0, N, M, P, Q).

%
compile_coa_epsilon_rule(0, I, Head:- true):-
	Head=..[I,X,X].
compile_coa_epsilon_rule(2, I, Head:- true):-
	Head=..[I, A, A, X, X].

% ?- compile_coa_rule([10-20, 22-30], a, a, 2, user, P, []).
compile_coa_rule([], _, _, _, _,P, P).
compile_coa_rule([x(X, U)|A], I, J, N, M, P, Q):-
	once(compile_special_char(X, U, N, I, J, M, P, P0)),
	compile_coa_rule(A, I, J, N, M, P0, Q).
compile_coa_rule(A, I, J, N, M, P, Q) :-
	once(consecutive_intervals(A, Rest, B)),
	compile_coa_intervals(B, I, J, N, M, P, P0),
	compile_coa_rule(Rest, I, J, N, M, P0, Q).

% ?- compile_coa_intervals([1-3], a, b, 0, user, P, []).
% ?- compile_coa_interval(1, 3, a, b, 0, user, P, []).
compile_coa_intervals([], _, _, _, _, P, P).
compile_coa_intervals([L-U|A], I, J, N, M, P, Q):-
	once(compile_coa_interval(L, U, I, J, N, M, P, P0)),
	compile_coa_intervals(A, I, J, N, M, P0, Q).

%
narrow_interval(L, U):- integer(L),
						integer(U),
						L + 5  >  U.

compile_coa_interval(L, U, I, J, N, M, P, Q):- narrow_interval(L, U), !,
	compile_interval_smart(L, U, I, J, N, M, P, Q).
compile_coa_interval(L, U, I, J, N, M, [R|T], T):-
	once(generate_code_test([L-U], V, Cond)),
	(	N==0, Head =..[I, [V|Z], Z0]
	;	N==2, Head =..[I, [V|W], W0, [V|Z], Z0]
	),
	(	J==void_state, Call = true
	;	N==0, Call=..[J, Z, Z0]
	;	N==2, Call=..[J, W, W0, Z, Z0]
	),
	(	J==void_state, Z0=Z, W0=W
	; true
	),
	slim_goal((Cond, Call), Body),
	prefix_convention(M, Head:- Body, R).

% ?- compile_interval_smart(1, a, b, 2, user, P, []).
% ?- compile_interval_smart(1, a, void_state, 2, user, P, []).
% ?- compile_interval_smart(1, a, b, 0, user, P, []).
% ?- compile_interval_smart(1, a, void_state, 2, user, P, []).

compile_interval_smart(D, I, J, N, M, [Cls|P], P):-
		(	N==0, Head =..[I, [D|Z], Z0]
		;	N==2, Head =..[I, [D|W], W0, [D|Z], Z0]
		),
		(	J==void_state, Call = true
		;	N==0, Call=..[J, Z, Z0]
		;	N==2, Call=..[J, W, W0, Z, Z0]
		),
		(	J==void_state, Z0=Z, W0=W
		; true
		),
		prefix_convention(M, Head:- Call, Cls).

% ?- compile_interval_smart(1, 2, a, void_state, 2, user, P, []).
% ?- compile_interval_smart(1, 2, a, b, 2, user, P, []).

compile_interval_smart(L, U, I, J, N, M, P, Q):- L=<U, !,
		compile_interval_smart(L, I, J, N, M, P, P0),
		L0 is L + 1,
		compile_interval_smart(L0, U, I, J, N, M, P0, Q).
compile_interval_smart(_, _, _, _, _, _, P, P).

% ?-  consecutive_intervals([a-b, b-c, c], X, Y).
consecutive_intervals([], [], []).
consecutive_intervals([A-B|R], U, [A-B|V]):- consecutive_intervals(R, U, V).
consecutive_intervals([X|R], [X|R], []).

%
goto_argument_form(0, V, [[V|R], S], [R, S]).
goto_argument_form(2, V, [[V|A], B, [V|R], S], [A, B, R, S]).

% ?- generate_code_test([1-1], V, C).
% ?- generate_code_test([1-1, 4-4], V, C).
% ?- generate_code_test([(1-4)\2], V, C).
generate_code_test(A-B, V, C):- generate_code_test([A-B], V, C).
generate_code_test(Is, V, C):- adjacent_interval(Is, Js, []),
	maplist(generate_code_test_aux(V), Js, Ks),
	insert_semicolon(Ks, C).

%
generate_code_test_aux(V, X, Y):- once(generate_code_between(X, V, Y)).


% ?- adjacent_interval([1-2, 4-5, 7-9, 11-13], P, []).
% ?- adjacent_interval([1-2, 4-5, 7-9, 11-13, 15-17], P, []).

adjacent_interval([], X, [X|P], P).
adjacent_interval([A-B|R], C-D, [(C-B)\X|P], Q):-  C\==D,
	succ_transitive(D, X, A),
	adjacent_interval(R,  P, Q).
adjacent_interval(X, U, [U|P], Q):-
	adjacent_interval(X, P, Q).

%
adjacent_interval([U|R], P, Q):- adjacent_interval(R, U, P, Q).
adjacent_interval([], P, P).

%
succ_transitive(A, X, B):-   plus(A, 2, B), plus(A, 1, X).

%
generate_code_between(inf-sup, _,	true).
generate_code_between(inf-A, V,		V @=< A).
generate_code_between(A-sup, V,		A @=< V).
generate_code_between(A-A,  V,		V == A).
generate_code_between(A-B,  V,		(A @=< V, V @=<B)).
generate_code_between((A-B)\W,  V,	BCode0):-
	once(generate_code_between(A-B, V, BCode)),
	( BCode == true, BCode0 = (V\== W)
	; BCode0 = (V\==W, BCode)).

			/*************
			*     sed    *
			*************/

% ?- expand_phrase(sed(@(s/"a"/"b")), [], G, L, []).
% ?- phrase(sed(w("[afl]") >> [[]]), `abcdefghijklmn`,  V), basic:smash(V).
% ?- phrase(sed(w(".+", A)>> pred(A, [[A,A]])), `ab`,  V), basic:smash(V).
% ?- phrase(sed((w(".", A), w(".", B))>> pred([A,B],([C]:- append(B, A, C)))), `ab12`,  V), basic:smash(V).
% ?- phrase(sed((w(".", A), w(".", B))>> append(B, A)), `abcdefghijklmn`,  V), basic:smash(V).
% ?- phrase(sed(w("[afl]")>> pred([[]])), `abcdefghijklmn`,  V), basic:smash(V).
% ?- phrase(sed(wl("\\\\[a-zA-Z]*")>> pred([[]])), `\\abcdefghijklmn`,  V), basic:smash(V).
% ?- phrase(sed(w("a")>>(user:pred([`x`]))), `abaaa`,  V), basic:smash(V).
% ?- show(phrase(sed(w("a")>>pred([`x`])))).
% ?- show_phrase(sed((w(".", A), w(".", B))>>append(B, A))).
% ?- show_phrase(sed(w("a")>>pred([`x`]))).
% ?- show_phrase(w("a*")).
% ?- show_am("\\\\[a-zA-Z]*").
% ?- show_am("[a-zA-Z]*").

expand_recognize_act(Vars, Words, Words_in_pac,  Mod,  Rec_Act, List, List3):-
	recognize_act(Vars, Words, Words_in_pac,  Mod,  R_A, List, List1),
    expand_core( rec(Sed_name, Vars,
				 	 ([[acc(Acc)|M], M, acc(Acc,[]), []])
				 &	 ([M, M, [], []])
				 &	 ( [L, M, P, Q]  :-  call(R_A, L, L0, P, P0), !,
                                         call(Sed_name, L0, M, P0, Q))
	             &	 ( [[C|Z0], Z, [C|P], Q]:- !, call(Sed_name, Z0, Z, P, Q))
	             &	 ( [[C|Z0], Z, acc(V,[C|P]), Q]:-
						   call(Sed_name, Z0, Z, acc(V, P), Q))
					),
                Mod, Sed_main, List1, List2),
    expand_core( pred(Vars, [U,V]:- call(Sed_main, V, [], U, [])),
                 Mod, Rec_Act, List2, List3).

% The prefix 'user:' is for get_acc to be found in runtime .
user:get_acc(V, acc(V,L), L):-!.
user:get_acc([], L, L).
%
user:get_acc(V, _, acc(V,L), L):-!.
user:get_acc(I, I, L, L).
%
user:set_acc(V, L, acc(V, L)).

% ?- recognize_act([], "a",  "b", user, A, P, []).
recognize_act(Vars, Words, Action,  Mod,  Recog_Act, List, List3):-
	once(expand_sed_act(Action, Mod, Action_expanded, List, List0)),
	lazy_prepend(Action_expanded, Action_slim, L, L0),
	expand_phrase(Words, Mod, Phrase, List0, List1),
	phrase_to_pred(Phrase, Mod, ([P, P0]:- Expanded_phrase), List1, List2),
	expand_core(pred(Vars, ([L, L0, P, P0]:-
				    once(Expanded_phrase), once(Action_slim))),
		    Mod, Recog_Act, List2, List3).

%
expand_sed_act(X, _, X, L, L):- (var(X); number(X); string(X); is_list(X)), !.
expand_sed_act(X + Y, M, X0 + Y0, L, L0):-
	expand_sed_act(X, M, X0, L, L1),
	expand_sed_act(Y, M, Y0, L1, L0).
expand_sed_act(X, M, Y, L, L0):- expand_arg(X, M, Y, L, L0).

% ?- expand_head_sed([], ("a"), "b", user, G, List, []), pac:show(G, List).
% ?- expand_head_sed([], ("a"), "b", user, G, List, []), !, maplist(assert, List), call(G, X, `a`, []).
% ?- expand_head_sed([A], ("a"), "b", user, G, List, []), !, maplist(assert, List), call(G, X, `a`, []).


expand_head_sed(Vars, Words, Action,  Mod,  G_callable, List, List3):-
	expand_phrase(Words, Mod, Phrase, List, List0),
	expand_sed_act(Action, Mod, Action_expanded, List0, List1),
	lazy_prepend_slim(Action_expanded, Action_slim, L, []),
	phrase_to_pred(Phrase, Mod,
		       [P, P0]:- Expanded_phrase, List1, List2),
	expand_core(pred(Vars, ([L, P, P0]:- Expanded_phrase, Action_slim)),
		    Mod, G_callable, List2, List3).

% ?- word((=([a,b]), =([c,d])), X, []).
word((A, B), X, Y):-  word(A, X, Z), word(B, Z, Y).
word(A, X, Y):- string(A), string_codes(A, C), append(C, Y, X).
word(A, X, Y):- call(A, S), append(S, Y, X).

%%
% ?- lazy_prepend_slim("abc", A, V, V0).
% ?- lazy_prepend_slim([abc], A, V, V0).
% ?- lazy_prepend_slim(abc, A, V, V0).
% ?- lazy_prepend_slim(abc+xyz, A, V, V0).
% ?- lazy_prepend_slim(X+Y, A, V, V0).

lazy_prepend_slim(E, A, V, V0):- once(lazy_prepend(E, B, V, V0)),
			  slim_goal(B, A).
%
lazy_prepend(X, basic:prepend(X, Y, Z), Y, Z):- var(X), !.
lazy_prepend(X, true, P, Q):- is_list(X), !, basic:prepend(X, P, Q).
lazy_prepend(X, true, [X|Y], Y):- string(X), !.
lazy_prepend(X, true, [X0|Y], Y):- number(X), !, number_string(X,X0).
lazy_prepend(X + Y, (A, B), P, Q):- lazy_prepend(X, A, P, P0),
				    lazy_prepend(Y, B, P0, Q).
lazy_prepend(X, (C, basic:prepend(Z, P, Q)),  P, Q):- complete_args(X, [Z], C).

% ?- phrase(sed(b/"a"/"hello"), ` a a `, R), basic:smash(R).
% ?- phrase(sed(a/"a"/"hello"), ` a a `, R), basic:smash(R).
% ?- phrase(sed(w/"a"/"hello "/" world"), ` a a `, R), basic:smash(R).
% ?- phrase(sed(wl/"a"/"hello "/" world"), ` a a `, R), basic:smash(R).
% ?- sed((wl/"a"/"hello "/" world"), ` a a `, R), basic:smash(R).
% ?- sed((wl/("a", X)/"hello "/" world"), ` a a `, R), basic:smash(R).  % << error test.



			/***************
			*     regex    *
			***************/

%%  expand_w(+X:regex, +M:atom, -G:goal, -P:list, -Q:list) is det
%
%   True if G is unifed with a goal and helper predicates are
%   created in P (minsu Q) so that  calling G  behaves
%   as X in the shortest-match-mode.


% ?- phrase((w(*(.), X), w(*(.), Y)),  [a,b,c], []).
% ?- phrase((wl(*(.), X), wl(*(.), Y)),  [a,b,c], []).

% ?- expand_w("a*$", [], G, P, []).
% ?- phrase(w("a*"), `aaaaab`, R).
% ?- expand_w("a*", A, [], [],  G, P, []).

expand_w(X, _, call(X), P, P):-var(X).	% @ meta call
expand_w(X, M, G, P, Q):-
	regex_coalgebra_code(X, coa(C, I, _)),
	expand_coa(0, C, I, M, G, P, Q).

%%  expand_w(+X:regex, -A:var, -B:var, +M:atom, -G:goal, +P:list, -Q:list) is det
%
%   True if G is unifed with a goal and helper predicates are
%   created in P (minsu Q) so that  calling G  behaves
%   as X in the shortest-match-mode with d-list (A, B).

expand_w(X, A, B, _, call(X, A, B), P, P):- var(X).  % @ meta call
expand_w(X, A, B, M, G, P, Q):-
	regex_coalgebra_code(X, coa(C, I, _)),
	expand_coa(2, C, I, M, G0, P, Q),
	complete_args(G0, [A, B], G).

%%  expand_wl(+X:regex, +M:atom, -G:goal, -P:list, -Q:list) is det
%
%   True if G is unifed with a goal and helper predicates are
%   created in P (minsu Q) so that  calling G  behaves
%   as X in the longest-match-first mode.

expand_wl(X, _, call(X), P, P):-var(X).	% @ meta call
expand_wl(X, M, G, P, Q):-
	regex_coalgebra_code(X, coa(C, I, _)),
	null_last_coa(C, C0),
	expand_coa(0, C0, I, M, G, P, Q).

%%  expand_wl(+X:regex, A:var, B:var, +M:atom, -G:goal, -P:list, -Q:list) is det
%
%   True if G is unifed with a goal and helper predicates are
%   created in P (minsu Q) so that  calling G  behaves
%   as X in the longest-match-first mode with d-list (A, B)

expand_wl(X, A, B, _, call(X, A, B), P, P):- var(X).  % @ meta call
expand_wl(X, A, B, M, G, P, Q):-
	regex_coalgebra_code(X, coa(C, I, _)),
	null_last_coa(C, C0),
	expand_coa(2, C0, I, M, G0, P, Q),
	complete_args(G0, [A, B], G).

%
null_last_coa(X, Y):- maplist(null_last_coa_state, X, Y).

%
null_last_coa_state(A-B, A-C):- reverse(B, C).

	  /***********************
	  *     compare regex    *
	  ***********************/

% ?- regex_compare(C, "a", "a").
% ?- regex_compare(C, "a", "b").
% ?- regex_compare(C, "a*", "a").
regex_compare(C, X, Y):- regex_normal_am_list([X,Y], [X0, Y0]),
	once(am_compare(C, X0, Y0)).

% ?-  regex_normal_am_list(["a", "b"], R).
% ?-  regex_normal_am_list(["(ab)*", "b"], R).

regex_normal_am_list(X, Y):- regex_word(regexlist(X), regexlist(X0)),
	maplist(word_normal_am, X0, Y).


	  /***************************
	  *     bisimilarity test    *
	  ***************************/

% ?- am_bisimilar(coa([], 0), coa([], 1)).
% ?- am_bisimilar(coa([1-[[]]], 1), coa([1-[[]]], 1)).
% ?- (regex_am("a", R), am_bisimilar(R, R)).
% ?- (regex_am(a+ (+a), R), regex_am((+a)+a, S), !,  am_bisimilar(R, S)).
% ?- (regex_am(a+ (+a), R), regex_am((+a)+a+a, S), !,  am_bisimilar(R, S)).

am_bisimilar(coa([], _), coa([], _)).
am_bisimilar(coa(E, I), coa(F, J)):-
	once(bisimilar(I, J, [], _, E, F)).

%
bisimilar(I, J, Done, Done, _, _):- memberchk(I-J, Done).
bisimilar(I, J, Done , Done0, E, F):-
	memberchk(I-G, E),
	memberchk(J-H, F),
	bisimilar_down(G, H, [I-J|Done], Done0, E, F).
%
bisimilar_down([], [], Done, Done, _, _).
bisimilar_down([[]|R], [[]|S], Done, Done0, E, F):-
	bisimilar_down(R, S, Done, Done0, E, F).
bisimilar_down([C-I|R], [C-J|S], Done, Done0, E, F):-
	bisimilar(I, J, Done, Done1, E, F),
	bisimilar_down(R, S, Done1, Done0, E, F).

	  /*********************************
	  *     one-way similarity test    *
	  *********************************/

% ?- (regex_am("a", A), regex_am("b", B), am_compare(C, A, B)).
% ?- (regex_am("a", A), regex_am("a", B), am_compare(C, A, B)).
% ?- (regex_am("a", A), regex_am("a*", B), am_compare(C, A, B)).
% ?- (regex_am("a*", A), regex_am("a", B), am_compare(C, A, B)).
am_compare(C, A, B):- once(am_similar(A, B)),
	(	am_similar(B, A)
	->	C = (=)
	;	C = (<)
	).
am_compare(C, A, B):- am_similar(B, A) -> C = (>); C = incomparable.

%
am_similar(coa([], _), coa([], _)).
am_similar(coa(E, I), coa(F, J)):-
	similar(I, J, [], _, E, F).

%
similar(I, J, Done, Done, _, _):- memberchk(I-J, Done).
similar(I, J, Done , Done0, E, F):-
	memberchk(I-G, E),
	memberchk(J-H, F),
	once(similar_down(G, H, [I-J|Done], Done0, E, F)).
%
similar_down([], _, Done, Done, _, _).
similar_down([[]|R], [[]|S], Done, Done0, E, F):-
	similar_down(R, S, Done, Done0, E, F).
similar_down([C-I|R], [C-J|S], Done, Done0, E, F):-
	similar(I, J, Done, Done1, E, F),
	similar_down(R, S, Done1, Done0, E, F).
similar_down(R, [_|S], Done, Done0, E, F):-
	similar_down(R, S, Done, Done0, E, F).
