:- module(meta,
	  [foldr/4, iterate/3,
	   for/3, foldnum/4,
	   fold/3, fold/4, fold/5,
	   mapterm_rec/3,
	   maplist_opp/3,
	   repeat/2, monad/3, time/3,
	   until/4, while/4,
	  iff/2, cputime/1, cputime/0]
	 ).

:- use_module(pac(op)).

term_expansion('FLIP_CLAUSES'(N), Cs):- numlist(0, N, Ns),
	maplist(create_flip_clause, Ns, Cs).

% if-and-only-if condition test.

% ?- meta:iff(true, member(X, [1,2])).
iff(C0, C1) :- \+( (C0,\+C1) ; (C1, \+C0)).

% ?- enable_pac_query.
% ?- nopac(meta:flip([2,1], is, 1+1, R)).
% ?- nopac(meta:flip([], true)).
% ?- nopac(meta:flip([], writeln(hello))).
% ?- nopac(meta:flip([1,3,2], append, [a,b], C, [c, d])).
% ?- nopac(meta:flip([1, 2, 3], plus, 1, 2, X)).
% ?- nopac(meta:flip([1, 3, 2], plus, 1, 2, X)).
% ?- meta:call(flip([1,3,2], append), [a,b], C, [c, d]).
% ?- meta:flip([1,3,2], append, [a,b], C, [c, d]).
create_flip_clause(I, Clause) :- must_be(nonneg, I),
	Clause = (H :- pac:flip_list(Permu, As, Bs),  Call),
	length(As, I),
	length(Bs, I),
	H =..[flip, Permu, F|As],
	Call =..[call, F|Bs].

% Example of expanded clauses.
% flip(A, D, B, C) :-
%     pac:flip_list(A, [B, C], [E, F]),
%     call(D, E, F).

'FLIP_CLAUSES'(6).

:- meta_predicate foldr(3, ?, ?, ?).
foldr(_, [], X, X):-!.
foldr(F, [A|L], X, Y):- call(F, A, Y, Y0), foldr(F, L, X, Y0).


% ?- meta:foldR(cons, [a,b,c,d], [], R).
% cons(X, Y, [X|Y]).
% foldR(F, List, X, Y):-  foldl(swap_args(F), List, Y, X).
% swap_args(F, X, Y, Z):- call(F, X, Z, Y).

% ?- foldl(cons, [a,b,c,d], [], R).
% ?- foldl(pred([X, Y, [X|Y]]), [1,2,3,4], [], R).
% ?- foldr(pred([X, [X|Y], Y]), [1,2,3,4], R, []).
% ?- foldr(flip([1,3,2], pred([X, [X|Y], Y])), [1,2,3,4], [], R).
% ?- foldl(flip([1,3,2], pred([X, Y, [X|Y]])), [1,2,3,4], R, []).
% ?- foldr(pred([X, Y, [X|Y]]), [1,2,3,4], [], R).
% ?- fold_paths_of_term(pred([A,[A|B], B]), f(1,2), X, []).
fold_paths_of_term(G, T, X, Y):- fold_paths_of_term([[T]], [], X, Y, G).
%
fold_paths_of_term([], _, X, X, _):-!.
fold_paths_of_term([Ts|L], P, X, Y, G):-
	(	Ts==[]
	->	fold_paths_of_term(L, P, X, Y, G)
	;	Ts=[T|Rs],
		fold_paths_of_term(T, [Rs|L], P,  X, Y, G)
	).
%
fold_paths_of_term(T, Ls, P, X, Y, G):- atomic(T), !,
	call(G, [T|P], X, Xtp),
	fold_paths_of_term(Ls, P, Xtp, Y, G).
fold_paths_of_term(T, Ls, P, X, Y, G):- T=..[F,A|As],
	fold_paths_of_term(A, [As|Ls], [F|P], X, Y, G).

% ?- fold_args(plus, f(1,2,3,4), 0, S).
% ?- fold_args(pred([X, Y, Z]:- Z is X + Y), f(1,2,3,4), 0, S).
fold_args(F, V, A, B):-
	fold_args(1, F, V, A, B).
%
fold_args(I, F, V, A, B):-	arg(I, V, Vi), !,
	call(F, Vi, A, Ai),
	I1 is I+1,
	fold_args(I1, F, V, Ai, B).
fold_args(_I, _F, _V, _A, _B).

% ?- help(mapterm).
% ?- help(maparg).

% ?- mapargs(pred([A, [A, A]]), (a,b), Out).
% ?- mapterm(pred([A, [A, A]]), (a,b), Out).
% ?- mapsubterms(pred([A, [A, A]]), (a,b), Out).
% ?- mapsubterms(pred([A, [A, A]]), f(h(a),g(b)), Out).
%@ Out = [f(h(a), g(b)), f(h(a), g(b))].


% mapterm(F, A, B):-
% 	functor(A, Fa, Na),
% 	functor(B, Fa, Na),
% 	mapterm(1, F, A, B).
% %
% mapterm(I, F, A, B):- arg(I, A, Ai), !,
% 	call(F, Ai, Bi),
% 	arg(I, B, Bi),
% 	J is I+1,
% 	mapterm(J, F, A, B).
% mapterm(_, _, _, _).


% ?- callargs(write, (a,b)).
callargs(F, A):- callargs(1, F, A).
%

callargs(I, F, A):- arg(I, A, Ai), !,
	call(F, Ai),
	J is I+1,
	callargs(J, F, A).
callargs(_, _, _).


% ?- Y = f(1, 2), trace, nopac((meta:mapargs(=, count(2, 1, 1), f(a,b), Y))).
%@    Call: (11) meta:mapargs(=, count(2, 1, 1), f(a, b), f(1, 2)) ? no debug
%@ Y = f(a, b).

% ?- Y = f(1, 2), meta:mapargs(=, count(1, 2, 2), f(a,b), Y).
% ?- Y = g(1, 2), mapargs(=, count(1, 1, 2), f(a,b), Y).
% mapargs(F, count(N, I, J), A, B):- mapargs(N, I, J, F, A, B).
% %
% mapargs(0, _, _, _, _, _):-!.
% mapargs(N, I, J, F, A, B):- arg(I, A, Ai),
% 	call(F, Ai, Bj),
% 	setarg(J, B, Bj),
% 	N1 is N-1,
% 	I1 is I+1,
% 	J1 is J+1,
% 	mapargs(N1, I1, J1, F, A, B).

% ?- help(mapterm).


% ?- meta:scanargs(pred([I, X, I-X]), f(a,b), Z).
scanargs(F, A, B):- ( var(A)
					->	functor(B, C, N),
						functor(A, C, N)
					;	functor(A, C, N),
						functor(B, C, N)
					),
	scanargs(1, N, F, A, B).
%
scanargs(I, N, _, _, _):- I > N, !.
scanargs(I, N, F, A, B):- arg(I, A, Ai),
	arg(I, B, Bi),
	call(F, I, Ai, Bi),
	J is I +1,
	scanargs(J, N, F, A, B).

% ?- meta:scanargs(pred([I, I]), f(A, B)).
scanargs(F, A):- functor(A, _,  N),
	scanargs_(1, N, F, A).
%
scanargs_(I, N, _, _):- I > N, !.
scanargs_(I, N, F, A):- arg(I, A, Ai),
	call(F, I, Ai),
	J is I + 1,
	scanargs_(J, N, F, A).


% ?- Z = g(x,y), meta:setargs(pred([I, J, I-J]), f(a,b), Z).
setargs(F, A, B):-  functor(A, C, N),
		( var(B)
		->	functor(B, C, N)
		;	true
		),
		setargs(1, N, F, A, B).
%
setargs(I, N, _, _, _):- I > N, !.
setargs(I, N, F, A, B):- arg(I, A, Ai),
	call(F, I, Ai, Bi),
	setarg(I, B, Bi),
	J is I + 1,
	setargs(J, N, F, A, B).

% ?- A = f(a,b), meta:setargs(pred([I, X, I-X]), A).
setargs(F, A):- functor(A, _,  N),
	setargs_(1, N, F, A).
%
setargs_(I, N, _, _):- I > N, !.
setargs_(I, N, F, A):- arg(I, A, Ai),
	call(F, I, Ai, Bi),
	setarg(I, A, Bi),
	J is I + 1,
	setargs_(J, N, F, A).

% ?-  maprows(zdd:inner_prod(f(1,2)), m(f(1,2), f(3,4)), B).
%@ B = f(7, 10).
% ?-  maprows(zdd:inner_prod(f(1,2)), m(f(1,2,3), f(3,4,5)), B).
%@ B = f(7, 10, 13).
maprows(_, A, A):- atom(A), !.
maprows(F, A, B):- arg(1, A, A1),
	functor(A1, Fa, Na),
	functor(B, Fa, Na),
	maprows(Na, F, A, B).
%
maprows(0, _, _, _):-!.
maprows(I, F, A, B):- arg(I, B, Bi),
	call(F, I, A, Bi),
	J is I - 1,
	maprows(J, F, A, B).

%
until(F, S, X, Y):- call(S, X, X0),
	(	call(F, X0) -> Y = X0
	;   until(F, S, X0, Y)
	).
%
iterate(_, stop(X), X):-!.
iterate(S, X, Y):-	call(S, X, X0),
	iterate(S, X0, Y).

while(F, S, X, Y):-
	(	call(F, X)
	->	call(S, X, X0),
		while(F, S, X0, Y)
	;   Y = X
	).

%	process(+Delta/2, +Final/1, ?S:state, ?S0:state) is semidet.
%	Starting with initial state S, repeat applying Delta
%	to the successive states until the state
%	satisfies the condition Final/1, with which S0 is unified.

% :- meta_predicate process(1, 2, +, -).
% process(Final, _, S, S):- call(Final, S).
% process(Final, Delta, S, S0):-
% 		call(Delta, S, S1),
% 		process(Final, Delta, S1, S0).

:- meta_predicate for(?, 1).
for(I..J, F):-!, for(I, J, F).
for(I-J, F):-for(I, J, F).

%
for(I, J, _):- I>J, !.
for(I, J, F):- call(F, I), !,
		I0 is I + 1,
		for(I0, J, F).

% ?- F = plus, meta:foldnum(F, 1-2, 0, R).
% ?- foldnum(pred([X,Y,Z]:-plus(X, Y, Z)), 1-2, 0, R).
% ?- foldnum(plus, 1-2, 0, R).
% ?- let(X, pred([I,[I|U], U])), foldnum(X, 1-10, R, []).

:- meta_predicate foldnum(3, ?, ?, ?).
foldnum(F, I-J, U, V):-!, foldnum(I, J, U, V, F).
foldnum(F, I..J, U, V):- foldnum(I, J, U, V, F).
%
foldnum(I, J, U, U, _):- I>J, !.
foldnum(I, J, U, V, F):- call(F, I, U, U0),
	K is I + 1,
	foldnum(K, J, U0, V, F).

% REMARK. ":- meta_predicate maplist_opp(:, ?, ?)" NOT work.
maplist_opp([F|Fs], X, [FX|Y]):- call(F, X, FX), !,
	maplist_opp(Fs, X, Y).
maplist_opp([], _, []).

%
:- meta_predicate map(2, ?, ?).
map(P) --> maplist(phrase(P)).
%
:- meta_predicate phrase_list(2, ?, ?).
phrase_list(P) --> maplist(phrase(P)).

% ?- repeat(20, write(.)).
% ?- repeat(between(1, 20, I), writeln(I)).
:- meta_predicate repeat(?, 0).
repeat(N, G):- integer(N), !, (between(1, N, _), G, fail; true).
repeat(P, G):- (call(P), call(G), fail) ; true.

%%% for statistics
% ?- time(10^8, X is 1 + 2, N).

:- meta_predicate time(?, 0, ?).
time(N, G, T) :- N0 is N,  time0(repeat(N0, G), T).

time0(G, T):-  time0(G, T0, T1), T is T1-T0.

time0(G, T, T0):-  statistics(cputime, T),
	call(G),
	statistics(cputime, T0).

% ?- cputime, A=(X=1), cputime(100, A, T), cputime(Total).
% ?- cputime, A=(X=1), cputime(10000000, A, T), cputime(Total).
% ?- cputime, cputime(10000000, X=1, T), cputime(Total).
% ?- cputime, cputime(10000000, (b_setval(x,1), b_getval(x, Y)), T), cputime(Total).
% ?- cputime, cputime(10000000, (X=1, Y=X), T), cputime(Total).
% ?- cputime, cputime(1000000, X=1, T), cputime(Total).
:- meta_predicate cputime(?, 0, ?).
%
cputime(N, G, T):- writeln("Running pac runtime library cputime/3...  "),
				   cputime(N, G, 0.00, T).
%
cputime(0, _G,  T, T).
cputime(N, G, T, T0):-  succ(N0, N),
						cputime_for_step(G, S),
						T1 is T + S,
						cputime(N0, G, T1, T0).
%
cputime_for_step(G, T):-
	statistics(cputime, T0),
	call(G),
	statistics(cputime, T1),
	T is T1-T0.
%
cputime:- statistics(cputime, T), b_setval(cputime, T).
cputime(T):- statistics(cputime, Stop), b_getval(cputime, Start), T is Stop - Start.

% ?- cputime(10000000, true, T).
% ?- meta:cputime(1000, (A=a, B=A), T).
% ?- meta:cputime(1000, (b_setval(a, b), b_getval(a, B)), T).

% monad/3
:- meta_predicate monad(:, ?, ?).
monad(G, X, Y):- once(ml:bind_context(G, (X, []), (Y, _))).

% recursive mapterm.
:- meta_predicate mapterm_rec(2, ?, ?).
mapterm_rec(F, A, B):- 	functor(A, Fa, Na),
						functor(B, Fa, Na),
						mapterm_rec(F, Na, A, B).
%
mapterm_rec(_, 0, _, _):-!.
mapterm_rec(F, I, A, B):- arg(I, A, Ai), arg(I, B, Bi),
	(	call(F, Ai, Bi) -> true
	;	atomic(Ai)	-> Bi = Ai
	;	mapterm_rec(F, Ai, Bi)
	),
	J is I - 1,
	mapterm_rec(F, J, A, B).

% recursive maplist.
:- meta_predicate maplist_rec(2, ?, ?).
maplist_rec(_, [], []):-!.
maplist_rec(F, [X|Xs], [Y|Ys]):-
	(	X = [_|_]
	->  maplist_rec(F, X, Y)
	;	call(F, X, Y)
	),
	maplist_rec(F, Xs, Ys).

% ?- fold(M, member(M, [1,2,3]), cons, [], X).
%@ X = [3, 2, 1].
cons(X, Y, [X|Y]).

:- meta_predicate fold(?, :, :, ?, ?).
fold(V, G, A, X, Y):-	Acc = '$ACC'(X),
	(	call(G),
		arg(1, Acc, U),
		call(A, V, U, W),
		nb_setarg(1, Acc, W),
		fail
	;	arg(1, Acc, Y)
	).

% ?- Sum = acc(0), fold(J, between(1, 3, J), test_add_to_acc, Sum).
test_add_to_acc(X, Acc):- arg(1, Acc, V),
	V0 is X + V,
	nb_setarg(1, Acc, V0).

:- meta_predicate fold(?, :, :, ?).
fold(V, G, Fun, X):-
		(	call(G),
			call(Fun, V, X),
			fail
		;	true
		).

% ?- fold(I, between(1, 3, I), writeln).
% ?- fold(I, between(1, 3, I), X^(Y is X*X, writeln(Y))).
:- meta_predicate fold(?, :, ?).
fold(V, G, V^Act):-!,
	(	call(G),
		call(Act),
		fail
	;	true
	).
fold(V, G, A):-
	(	call(G),
		call(A, V),
		fail
	;	true
	).

% ?- N=100, time(repeat(N, det_foreach(between(1,5,J), member(J, L)))).
%@ % 7,201 inferences, 0.001 CPU in 0.001 seconds (94% CPU, 8277011 Lips)
%@ N = 100.
% ?- det_foreach(between(1,5,N),  member(r(N),L)).
%@ L = [r(1), r(2), r(3), r(4), r(5)|_].
% exists M exists L forall J  (between(1,5, J) -> (member(J, L), member(J, M))).
% ?- det_foreach(between(1,5, J), (member(J, L), member(J, M))).
%@ L = [1, 2, 3, 4, 5|_],
% exists L forall J  (between(1,5, J), between(3, 7, J)  -> member(J, L))
% ?- det_foreach((between(1,5, J), between(3, 7, J)), member(J, L)).
%@ L = [3, 4, 5|_].
% exists L forall J  (between(1,5, J); between(3, 7, J))  -> member(J, L))
% ?- foreach((between(1,5, J); between(3, 7, J)), member(J, L)).
%@ L = [1, 2, 3, 4, 5, 6, 7|_].
% ?- foreach(between(1, 5, J), dif(L, J)), L = 3.
% ?- time(repeat(10^5, det_foreach(between(1, 5, J), (member(J, L),
%	det_foreach(between(6, 8, K), (member(K, L),
%	det_foreach(between(9, 10, M), member(M, L)))))))).
%@ % 99,800,001 inferences, 13.491 CPU in 14.710 seconds (92% CPU, 7397365 Lips)
%@ true.
% ?- time(repeat(10, det_foreach(between(1, 5, J), member(J, L)))).
% ?- time(repeat(10, foreach(between(1, 5, J), member(J, L)))). % stack limit.

det_foreach(Gen, Con):-	term_variables(Gen, Vs),
	sort(Vs, Vs0),
	term_variables(Con, Ws),
	sort(Ws, Ws0),
	ord_subtr_var(Ws0, Vs0, Ws1),
	once(foreach_by_copy(Gen, Con, Vs0, Ws1)).
%
foreach(Gen, Con, Vs, Ws):-	once(foreach_by_copy(Gen, Con, Vs, Ws)).

% ?- N=200, K = 5, time((repeat(N, once((foreach_by_copy(between(1,5, J), member(J, L), [J], [L]), writeln(L)))))).
foreach_by_copy(A, B, Vs, Ws):-
	findall(Vs, A, Sol),
	copy_of_goal(Sol, Vs, Ws, B, Cs),
	maplist(call, Cs).
%
copy_of_goal([], _, _, _, []).
copy_of_goal([A|As], Vs, Ws, B, [G|Gs]):-
	copy_term(Vs+Ws+B, A+Ws+G),
	copy_of_goal(As, Vs, Ws, B, Gs).
%
ord_subtr_var([], _, []):-!.
ord_subtr_var(X, [], X):-!.
ord_subtr_var([X|Xs], [Y|Ys], Zs):- X==Y, !,
	ord_subtr_var(Xs, Ys, Zs).
ord_subtr_var([X|Xs], [Y|Ys], [X|Zs]):- X@<Y, !,
	ord_subtr_var(Xs, [Y|Ys], Zs).
ord_subtr_var(Xs, [_|Ys], Zs):- ord_subtr_var(Xs, Ys, Zs).

		/**************************************************************
		*     foreach by assert/compile_aux: SLOWer than COPY_TERM    *
		**************************************************************/

% ?- foreach_by_assert(between(1,2,J), member(J, L), [J], [L]).
% foreach_by_assert(Gen, Con, Vs, Ws):-
% 	retractall('$CONSUMER'(_,_)),
% 	assert(( '$CONSUMER'(Vs, Ws):- Con)),
% 	Stash = '$STASH'(Ws),
% 	(	call(Gen),
% 		arg(1, Stash, U),
% 		once('$CONSUMER'(Vs, U)),
% 		nb_setarg(1, Stash, U),
% 		fail
% 	;   arg(1, Stash, Ws)
% 	).

% foreach_by_assert(Gen, Con, Vs, Ws):-
% 	context_module(M),
% 	pac:expand_core(pred([Vs, Ws]:- Con), M, G, P, []),
% 	maplist(assert, P),
% 	Stash = '$STASH'(Ws),
% 	(	call(Gen),
% 		arg(1, Stash, U),
% 		once(call(G, Vs, U)),
% 		nb_setarg(1, Stash, U),
% 		fail
% 	;   arg(1, Stash, Ws)
% 	).
