:- module(naot, [prop_to_naot/2, prop_to_naot/3,
				subst_aot/4,
				eval_ground_aot/2]).

 :- op(1060, xfy, ~).	% equivalence
 :- op(1060, xfy, <->).  % equivalence
 :- op(1060, xfy, <=> ). % equivalence
 :- op(1040, xfy, \/).	% OR
 :- op(1030, xfy, /\).	% AND
 :- op(1020, fy, \).	% NOT

		/**********************
		*     Basic tests.    *
		**********************/

% ?- prop_to_naot((A; B;C; D;E;F;G;H;I;J;K;L;M;N;O;P;Q;R;S;T;U;V;W;X;Y;Z), RR).
%@ RR = d([1, 2, 3, 4, 5, 6, 7, 8|...], [], []).

% ?- prop_to_aot((A; B; C), RR).
% ?- prop_to_aot((A, B, C), RR).
% ?- module(naot).

		/*************************
		*     Basic operatins    *
		*************************/
%
zip([], [], []).
zip([X|Xs], [Y|Ys], [X-Y|Zs]):-zip(Xs, Ys, Zs).

% ?- meet([1,2],[3,4]).
meet([X|R], S):- (memberchk(X, S)->true
				 ; meet(R, S)
				 ).
%
negation(true, false).
negation(false, true).
%
complement_literal(not(I), I).
complement_literal(I, not(I)).
%
no_complement(X):- \+ has_complement(X).
%
has_complement([X|R]):- complement_literal(X, X0),
				   memberchk(X0, R), !.
has_complement([_|R]):- has_complement(R).

%
literal(true).
literal(false).
literal(not(X)):- must_be(integer, X).
literal(X):- integer(X).

%
sort_aot(c(A1,A2,A3),c(B1,B2,B3)):-
	sort(A1,B1),sort(A2,B2),sort(A3,B3).
sort_aot(d(A1,A2,A3),d(B1,B2,B3)):-
	sort(A1,B1),sort(A2,B2),sort(A3,B3).

%
make_aot(X, Y):- once(make_aot(X, Y, root)).

make_aot(and(X,Y), Z, _):-
	once(make_aot(X, X0, and)),
	once(make_aot(Y, Y0, and)),
	once(make_and_node(X0, Y0, Z)).
make_aot(or(X,Y), Z, _):-
	once(make_aot(X, X0, or)),
	once(make_aot(Y, Y0, or)),
	once(make_or_node(X0, Y0, Z)).
make_aot(not(X), Y, SW):-
	( SW = and -> Y= c([],[X],[])
	; SW = or  -> Y= d([],[X],[])
	; Y  = not(X)
	).
make_aot(X, Y, SW):- integer(X), !,
	( SW = and -> Y= c([X],[],[])
	; SW = or  -> Y= d([X],[],[])
	; Y = X
	).
make_aot(X, X, _).

%
make_and_node(true, X, X).
make_and_node(false, _, false).
make_and_node(X, true, X).
make_and_node(_, false, false).
make_and_node(c(A1,A2,A3), c(B1, B2, B3), c(C1,C2,C3)):-
		union(A1,B1,C1),
		union(A2,B2,C2),
		union(A3,B3,C3).
make_and_node(c(A1,A2,A3),  D, c(A1, A2, [D|A3])):-
	D=d(_,_,_).
make_and_node(D, c(A1,A2,A3),  c(A1, A2, [D|A3])):-
	D=d(_,_,_).
make_and_node(X, Y, c([], [], [X, Y])):-
	X=d(_,_,_), Y=d(_,_,_).

%
make_or_node(true, _, true).
make_or_node(false, X, X).
make_or_node(_, true, true).
make_or_node(X, false, X).
make_or_node(d(A1,A2,A3), d(B1, B2, B3), d(C1,C2,C3)):-
		union(A1,B1,C1),
		union(A2,B2,C2),
		union(A3,B3,C3).
make_or_node(d(A1,A2,A3), C, d(A1,A2,[C|A3])):-
		C=c(_,_,_).
make_or_node(C, d(A1,A2,A3), d(A1,A2,[C|A3])):-
		C=c(_,_,_).
make_or_node(A, B, d([], [],[A, B])):-
		A=c(_,_,_), B=c(_,_,_).



% test1
% ?- prop_to_naot(true, X).
% ?- prop_to_naot(false, X).
% ?- prop_to_naot(X, RR).
% ?- prop_to_naot(X<->X, RR).
% ?- prop_to_naot(\(X<->X), RR).
% ?- prop_to_naot(A;B, X).
% ?- prop_to_naot(A, X).
% ?- prop_to_naot(not(A), X).
% ?- prop_to_naot(A;B;C;D,X).
% ?- prop_to_naot(\(A;B;C;D), X).
% ?- problem(1, P), prop_to_naot(P, X).
% ?- problem(2, P), prop_to_naot(P, X).
% ?- problem(3, P), prop_to_naot(P, X).
% ?- problem(4, P), prop_to_naot(P, X).
% ?- problem(5, P), prop_to_naot(P, X).
% ?- (time(problem(30, P)), !, time(prop_to_naot(P, X)), !), fail; true.
%@ % 14,571 inferences, 0.004 CPU in 0.004 seconds (90% CPU, 4012944 Lips)
%@ % 2,028,635 inferences, 3.622 CPU in 3.634 seconds (100% CPU, 560014 Lips)
% ?- (time(problem(40, P)), !, time(prop_to_naot(P, X)), !), fail; true.
%@ % 33,825 inferences, 0.009 CPU in 0.010 seconds (88% CPU, 3757498 Lips)
%@ % 4,717,045 inferences, 20.056 CPU in 20.114 seconds (100% CPU, 235195 Lips)
% ?- prop_to_naot((A+B)*(C+D)*(E+F)*(G+H)*(I+J)*(K+L)*(M+N)*
% (O+P)*(Q+R)*(S+T)*(U+V)*(W+X)*(Y+Z), RR).
% ?- prop_to_naot(\ ((A+B)*(C+D)*(E+F)*(G+H)*(I+J)*(K+L)*(M+N)*
% (O+P)*(Q+R)*(S+T)*(U+V)*(W+X)*(Y+Z)), RR, Zip).

prop_to_naot(X, Y):- once(prop_to_aot(X, Z, _Zip)),
		 once(reduce_aot(Z, Y, a([],[]))).

prop_to_naot(X, Y, Zip):- once(prop_to_aot(X, Z, Zip)),
		 once(reduce_aot(Z, Y, a([],[]))).

%
reduce_literal(true, true, _).
reduce_literal(false, false, _).
reduce_literal(not(X), Y, Anc):- !, Anc=a(P, N),
			(	memberchk(X, P)->  Y = false
			;   memberchk(X, N)->  Y = true
			;	Y = c([],[X],[])
			).
reduce_literal(X, Y, Anc):- Anc=a(P, N),
			(	memberchk(X, P)->  Y = true
			;   memberchk(X, N)->  Y = false
			;	Y = c([X],[],[])
			).

%
reduce_aot(X, Y, Anc):- literal(X), !,	reduce_literal(X, Y, Anc).
reduce_aot(c(A1,A2,_), false, _):- meet(A1, A2), !.
reduce_aot(c(A1,A2,_), false, a(P,N)):- (meet(A1, N); meet(A2, P)), !.
reduce_aot(d(A1,A2,_), true, _):- meet(A1, A2), !.
reduce_aot(d(A1,A2,_), true, a(P,N)):- (meet(A1, P); meet(A2, N)), !.
reduce_aot(c(A1,A2,A3), B,  a(P,N)):- subtract(A1, P, B1),
					  subtract(A2, N, B2),
					  union(B1, P, P0),
					  union(B2, N, N0),
					  sort(A3, B3),
					  reduce_aot_conj(B3, c(B1, B2,[]), B0, a(P0,N0)),
					  reduce_singleton(B0, B).

reduce_aot(d(A1,A2,A3), B,  a(P,N)):- subtract(A1, N, B1),
					  subtract(A2, P, B2),
					  sort(A3, B3),
					  reduce_aot_disj(B3, d(B1, B2,[]), B0, a(P, N)),
					  reduce_singleton(B0, B).
%
reduce_singleton(d([], [], []), false).
reduce_singleton(d([], [], [X]), Y):- reduce_singleton(X, Y).
reduce_singleton(c([], [], []), true).
reduce_singleton(c([], [], [X]), Y):- reduce_singleton(X, Y).
reduce_singleton(d(A1,A2,_), true):-	meet(A1, A2), !.
reduce_singleton(d(A1,A2,A3), R):-	select(d(B1,B2,B3), A3, A4), !,
	union(A1, B1, C1),
	union(A2, B2, C2),
	union(B3, A4, A5),
	reduce_singleton(d(C1, C2, A5), R).
reduce_singleton(c(A1,A2,_), false):-	meet(A1, A2), !.
reduce_singleton(c(A1,A2,A3), R):-	select(c(B1,B2,B3), A3, A4), !,
	union(A1, B1, C1),
	union(A2, B2, C2),
	union(B3, A4, A5),
	reduce_singleton(c(C1, C2, A5), R).
reduce_singleton(c(A1,A2,[d([X],[],[])]),c([X|A1], A2, [])).
reduce_singleton(c(A1,A2,[d([],[X],[])]),c(A1, [X|A2], [])).
reduce_singleton(c(A1,A2,[d([],[],[X])]),R):- !,
	reduce_singleton(c(A1, A2, [X]), R).
reduce_singleton(d(A1,A2,[c([X],[],[])]),d([X|A1], A2, [])).
reduce_singleton(d(A1,A2,[c([],[X],[])]),d(A1, [X|A2], [])).
reduce_singleton(d(A1,A2,[c([],[],[X])]),R):- !,
	reduce_singleton(d(A1, A2, [X]), R).
reduce_singleton(X, X).

%
push_aot_disj(d(A1,A2,A3), d(B1,B2,B3), d(C1,C2,C3)):-
	union(A1,B1,C1),
	union(A2,B2,C2),
	union(A3,B3,C3).
push_aot_disj(A, d(B1,B2,B3), d(B1,B2,[A|B3])).
%
reduce_aot_disj([], X, Y, _):-sort_aot(X, Y).
reduce_aot_disj([A|As], X, Y, Anc):- reduce_aot(A, A0, Anc),
		  ( A0=false -> reduce_aot_disj(As, X, Y, Anc)
		  ; A0=true ->  Y=true
		  ; push_aot_disj(A0, X, X0),
			reduce_aot_disj(As, X0, Y, Anc)
		  ).
%
push_aot_conj(c(A1,A2,A3), c(B1,B2,B3), c(C1,C2,C3)):-
	union(A1,B1,C1),
	union(A2,B2,C2),
	union(A3,B3,C3).
push_aot_conj(A, c(B1,B2,B3), c(B1,B2,[A|B3])).

%
reduce_aot_conj([], X, Y, _):- sort_aot(X, Y).
reduce_aot_conj([A|As], X, Y, Anc):- reduce_aot(A, A0, Anc),
		  ( A0=true -> reduce_aot_conj(As, X, Y, Anc)
		  ; A0=false ->  Y=false
		  ; push_aot_conj(A0, X, X0),
			reduce_aot_conj(As, X0, Y, Anc)
		  ).
% ?- module(naot).


% ?- trace, prop_to_naot(X->true, R), subst_aot(R, V, 1, true).
% ?- trace, prop_to_naot(X=Y, R), subst_aot(R, V, 1, true),
%	subst_aot(V, U, 2, true), eval_ground_aot(U, W).

% ?- trace, prop_to_naot(X=Y, R), subst_aot(R, V, 1, true),
% ?- prop_to_naot(X=Y, R).
% ?- prop_to_naot(X=Y, R), subst_aot(R, V, 1, true), subst_aot(V, V0, 2, true).
% ?- prop_to_naot(X=Y, R), subst_aot(R, V, 1, true), trace, subst_aot(V, V0, 2, true).
% ?- prop_to_naot(X=Y, R), subst_aot(R, V, 1, false), trace, subst_aot(V, V0, 2, false).
% ?- prop_to_naot(X=Y, R), subst_aot(R, V, 1, true), trace, subst_aot(V, V0, 2, false).
% ?- prop_to_naot(X=Y, R), subst_aot(R, V, 1, false), trace, subst_aot(V, V0, 2, true).

subst_aot(true, true, _, _).
subst_aot(false, false, _,_).
subst_aot(c(X, Y, Z), U, I, V):-
	subst_aot_c_p(X, X0,  I, V),
	(	X0 = false -> U=false
	;	subst_aot_c_n(Y, Y0, I, V),
		( Y0 =false -> U=false
		; subst_aot_list_c(Z, Z0, I, V),
			(	Z0=false -> U = false
			;	X0=[], Y0=[], Z0=[] ->  U = true
			;	U=c(X0,Y0,Z0)
			)
		)
	).
subst_aot(d(X,Y,Z), U, I, V):-
	subst_aot_d_p(X, X0, I, V),
	(	X0= true -> U=true
	;	subst_aot_d_n(Y, Y0, I, V),
		( Y0=true -> U=true
		; subst_aot_list_d(Z, Z0, I, V),
			(	Z0=true -> U = true
			;	X0=[], Y0=[], Z0=[] ->  U = false
			;	U=d(X0,Y0,Z0)
			)
		)
	).

%
subst_aot_c_p(Xs, Ys, X, V):- select(X, Xs, Zs), !,
	(	V = false -> Ys =false
	;	Ys = Zs
	).
subst_aot_c_p(Xs, Xs, _, _).
%
subst_aot_c_n(Xs, Ys, X, V):- select(X, Xs, Zs), !,
	(	V = true -> Ys =false
	;	Ys = Zs
	).
subst_aot_c_n(Xs, Xs, _, _).
%
subst_aot_d_p(Xs, Ys, X, V):- select(X, Xs, Zs), !,
	(	V = true -> Ys =true
	;	Ys = Zs
	).
subst_aot_d_p(Xs, Xs, _, _).
%
subst_aot_d_n(Xs, Ys, X, V):- select(X, Xs, Zs), !,
	(	V = false -> Ys =true
	;	Ys = Zs
	).
subst_aot_d_n(Xs, Xs, _, _).
%
subst_aot_list_c(Ms, R, X, V):- subst_aot_list_c(Ms, Ns, Ball, X, V),
	( var(Ball)	-> 	R=Ns
	; R=Ball
	).
%
subst_aot_list_c([], [], _, _, _).
subst_aot_list_c([M|Ms], Ps, Ball, X, V):- subst_aot(M, N, X, V),
	(  N = true   -> subst_aot_list_c(Ms, Ps, Ball, X, V)
	;  N = false  -> Ball = false
	;  Ps=[N|Ns],
	   subst_aot_list_c(Ms, Ns, Ball, X, V)
	).
%
subst_aot_list_d(Ms, R, X, V):- subst_aot_list_d(Ms, Ns, Ball, X, V),
	( var(Ball)	-> 	R=Ns
	; R=Ball
	).
%
subst_aot_list_d([], [], _, _, _).
subst_aot_list_d([M|Ms], Ps, Ball, X, V):- subst_aot(M, N, X, V),
	(  N = false   -> subst_aot_list_d(Ms, Ps, Ball, X, V)
	;  N = true  -> Ball = true
	;  Ps = [N|Ns],
	   subst_aot_list_d(Ms, Ns, Ball, X, V)
	).
%
prop_to_aot(X,Y):- prop_to_aot(X,Y,_).


prop_to_aot(X, Y, Zip):- var_to_int(X, IX, Zip),
					once(expand_macro(IX, X0)),
					once(demorgan(X0, X1)),
					once(make_aot(X1, Y)).

%
var_to_int(X, Y, Zip):- term_variables(X, Vs),
						copy_term((Vs,X), (Us,Y)),
						numlist(1, Us),
						zip(Us, Vs, Zip).
%
numlist(_, []).
numlist(N, [N|R]):- succ(N, N0), numlist(N0, R).

% ?- literal(false).
eval_ground_aot(true, true).
eval_ground_aot(false, false).
eval_ground_aot(c(P,N,L), R):-
	( (	memberchk(false, P)
	  ;	memberchk(false, N)
	  ;	member(M, L),
		eval_ground_aot(M, R0),
		R0 = false
	  ) ->  R = false
	; R = true
	).
eval_ground_aot(d(P,N,L), R):-
	( (	memberchk(true, P)
	  ;	memberchk(true, N)
	  ;	member(M, L),
		eval_ground_aot(M, R0),
		R0 = true
	  ) ->  R = true
	; R = false
	).

% ?- prop_to_aot(X, R).
% ?- prop_to_aot((X<->Y)=true, A, Z).
% ?- prop_to_aot(true<->true, A, Z).
% ?- prop_to_aot(true->true, R).
% ?- prop_to_aot(true<->true, R).
% ?- prop_to_aot(true<->false, R).

macro(\+(X),	not(X)).
macro(\(X),		not(X)).
macro(X\=Y,		not(X=Y)).
macro((X,Y),	and(X,Y)).
macro(/\(X, Y), and(X,Y)).
macro(X*Y,		and(X,Y)).
macro((X;Y),	or(X,Y)).
macro(\/(X,Y),	or(X,Y)).
macro(X+Y,		or(X,Y)).
macro(X->Y,		or(not(X), and(X,Y))).
macro(X=Y,		and(X->Y, Y->X)).
macro(X<->Y,	X=Y).
macro(X<=>Y,	X=Y).
macro(X~Y,		X=Y).
macro(++(X,Y),	not(X=Y)).
macro(exor(X, Y), not(X=Y)).
macro(nand(X, Y), not(and(X,Y))).

%
expand_macro(X, X):- integer(X), !.
expand_macro(X, Y):- macro(X, X0), !,
	expand_macro(X0, Y).
expand_macro(A,  B):- A=..[F|As],
	maplist(expand_macro, As, Bs),
	B=..[F|Bs].

%
demorgan(and(X, Y), and(X0, Y0)):- demorgan(X, X0),
								   demorgan(Y, Y0).
demorgan(or(X, Y), or(X0, Y0)):- demorgan(X, X0),
								   demorgan(Y, Y0).

demorgan(not(not(X)), Y):- demorgan(X, Y).
demorgan(not(and(X, Y)), or(X0, Y0)):- demorgan(not(X), X0),
									   demorgan(not(Y), Y0).
demorgan(not(or(X,Y)), and(X0, Y0)):- demorgan(not(X), X0),
									demorgan(not(Y), Y0).
demorgan(not(true), false).
demorgan(not(false), true).
demorgan(X, X).


		/*****************************
		*     a la Burse's sample    *
		*****************************/
% ?-  problem(4, Prop), build_robdd(V0s, Prop, N, Zip), writeln(done).
% ?-  problem(5, Prop), build_robdd(V0s, Prop, N, Zip), writeln(done).
% ?-  problem(5, Prop), build_robdd(V0s, Prop, N, Zip),

matrix(N, M):- length(M, N), maplist(col(N), M).

col(N, L):- length(L, N).

problem(N, P):- matrix(N, M), constraint(M, P).

inner_product([X], [Y], X*Y).
inner_product([X|Xs], [Y|Ys], X*Y + Z):- inner_product(Xs, Ys, Z).
%

constraint([V], E=true):- inner_product(V, V, E).
constraint([V|Vs], (E=true)*C*D):- inner_product(V, V, E),
	constraint(Vs, D),
	inner_constraint_one(V, Vs, C).
%
inner_constraint_one(V, [U], C=false):- inner_product(V, U, C).
inner_constraint_one(V, [U|Us], (C=false)*D):-
	inner_product(V, U, C),
	inner_constraint_one(V, Us, D).
