:- module(cgi, [matrix_rule/3, cgi_matrix/2]).

:- use_module(pac(op)).
:- use_module(pac(basic)).
:- use_module(util(misc)).
:- use_module(util(coalgebra)).
:- use_module(util('fol-prover')).
:- use_module(util(html)).
:- use_module(util(latex)).
:- use_module(util(math)).
:- use_module(util(seqcal4)).
:- use_module(util(npuzzle)).
:- use_module(util(turing)).
:- qcompile(util(matrix)).		% kind rules are included.
:- use_module(util(langsem)).
:- use_module(util(ptq)).
:- use_module(util(models)).
:- use_module(util(hmodel)).
:- use_module(util(prooftree)).
:- use_module(util(ifmap)).
:- use_module(zdd(zdd)).
:- use_module(zdd('zdd-plain')).
:- use_module(pac('pac-listing')).
% :- use_module(util(leantap)).

term_expansion --> pac:expand_pac.

:- op(670, yfx, \).
:- op(200, xfy, ^).

% :- use_module(util(zdd)).
% :- use_module(util('ptq-fragment')).
% :- use_module(util('emacs-handler')).
% :- use_module(util(interval)).
% :- use_module(gb('gb-top')).

chat80(X, Y) :- ensure_loaded(lib('chat80-cgi')),
				chat80:cgi(X, Y).
show_env(X, Y):- getenv(X, Y).

primrec_demo(X, Y):- primrecfast:eval_primrec_demo(X, Y).

term_to_goal(X, such_that(V, Y), [], [], true):-
	term_to_goal(X, Y, V).

setfont(X, Y):- flatten(["<font color=red size=6>", X, "</font>"], Y).

query_ans_html	-->	html:query_ans_html(pac:eval).

query_ans_html(X)  -->  html:query_ans_html(X).


		/**************************************
		*     First-order prover exercises    *
		**************************************/

%;; (setq module-query  "qcompile(util(cgi)), module(cgi).")
% ?- qcompile(util(cgi)), module(cgi).

web_proof_default_options([ clausal(false),
			    center([]),
			    initial_depth(300),
			    iteration_limit(5)]).

%%	valid_tex(+X, -R) is det.
%	Prove formula X, and unify R the result.

% ?- cgi:valid_test(a->a, R).
% ?- cgi:valid_test(a->(b->a), R).
valid_test(X, Result):-fol_prover:fol_linear_prove(X, Result, 10).


%%	valid_test_clausal(+P:pair, -R) is det.
%	Prove clausal form P of FOL formulae P,  and
%	unify R with result.

% ?- cgi:valid_test_clausal(([[a,~b], [a,b], [~a,b], [~a,~b]], []), Result),
%	smash(Result).
valid_test_clausal((X, Y), Result):-
	rename_vars((X, Y), (X0, Y0)),
	valid_test_clausal(Y0, X0, Result).


% ?- valid_test_clausal([], [[a],[a, -a]], Result).
% ?- leantap:apply_options([rm_tautology], [[a, -a]], X).
%
valid_test_clausal([], X, Result):-!,
		leantap:apply_options(
		[	pure_literal_law,
			rm_tautology,
			cnf_normal,
			map_factoring,
			cnf_variant_normal
		 ],
		X, Y),
		fol_prover:linear_solve(Y, Result).
valid_test_clausal(As, Ds, Result):- leantap:linear_solve_center(As, Ds, Result).


% ?- rename_vars([p(x,x), q(X, X)], R).
% ?- rename_vars([p(x,y,x), p(1,2), q(X, X)], R).
rename_vars(X, Y):- rename_vars(X, Y, [], _).
%
rename_vars(X, Y, A, B):-  % only u,v,w,z,y,z are Variables
	(	atom(X), u @=< X, X @=<z ->
			(	memberchk(X-Y, A) -> B = A
			; 	B = [X-Y|A]
			)
	;	is_list(X)	-> 	rename_vars_list(X, Y, A, B)
	;	compound(X)	->
		X =..[F|Xs],
		rename_vars_list(Xs, Ys, A, B),
		Y =..[F|Ys]
	;	Y = X,
		B = A
	).
%
rename_vars_list([], [], A, A).
rename_vars_list([X|Xs], [Y|Ys], A, B):-
	rename_vars(X, Y, A, C),
	rename_vars_list(Xs, Ys, C, B).

%%	web_proof(+X:options, -V) is det.
%	Prove formula in X, and unify V with the rusult.
web_proof(Options, V):- web_proof_default_options(D),
	completing_options(D, Options, Options0),
	memberchk(input(X), Options0),
	once(fol:old_prove(X, S, Options0)),
	(	memberchk(clausal(true), Options0)
	->  	(	integer(S),
			number_codes(S, S0),
			R = ["depth of search = ", S0, "\n<br/>"]
		;	R = S
		),
		memberchk(trace(Trace), Options0),
		maplist(copy_term, Trace, Trace0),
		maplist(rename_variable, Trace0),
		insert("\n<br/>", Trace0, Trace1),
		V0 = [R|Trace1]
	;	(	integer(S),
			V0 = 'proved.'
		;	V0 = 'no proof found.'
		)
	),
	smash_codes(V0, V).


% ?-  cgi:rename_variable(f(A,B,C,A)).
rename_variable(T):- term_variables(T, Vs),
	( Vs == []
	->	Zip = []
	;	length(Vs, N),
		numlist(1, N, Ns),
		maplist(pred([I, V, I-V]), Ns, Vs, Zip)
	),
	maplist(pred([I - Name]:- fol:number_to_var_name(I, Name)), Zip).

% ?- cgi_dnf(a, Dnf).
% ?- cgi_dnf(a+b, X).
% ?- cgi_dnf(-(a+b), X).
% ?- cgi_dnf(-(a), X).

cgi_dnf(Boole, Z):- zdd((zdd:dnf(Boole, X), sets(X, Z))).

%?- cgi:demo_sed((@(s/"e"/"x"), "ea"), R).
%?- cgi:demo_sed((w("e")/"x", "ea"), R).

demo_sed((Exp, String), Val):- string_codes(String, Codes),
	once(let_sed(G, Exp)),
	once(call(G,  Codes, Out)),
	string_codes(Val, Out).

% ?- cgi_matrix([x, "1", "2"], R).
% ?- cgi_matrix(["x+y", "1", "2"], R).
cgi_matrix([E|Ms], A):-
	term_string(E0, E),
	maplist(parse_matrix, Ms, Ns),
	matrix_calc(Ns, E0, A).
%
parse_matrix(S, M):-
	split_string(S, "\n", "\s\n\t", U),
	remove([], U, V),
	maplist(parse_raw, V, M).
%
parse_raw(S, R):- split_string_by_word(S, "\s\s", U),
	maplist(pred([X,Y]:- term_string(Y, X)), U, V),
	remove(end_of_file, V, W),
	maplist(pred([X, Y]:- Y is X), W, R).
%
matrix_html -->
	maplist(maplist(cgi_tag(td))),
	maplist(cgi_tag(tr)),
	insert("\n"),
	cgi_tag(table),
	cgi_tag(html),
	term_smash0.
%.
cgi_tag(T, X, ["<", T, ">", X, "</", T, ">"]).

% rational_codes(rdiv(X,Y), X/Y):-!.
% rational_codes(X, X).

% javascript からの使い方
% act(`cgi:csv_demo`, `[@columns, @csv]`, `columns;csv`,
%      `ここを押すと指定されたカラムがすぐ下に表示される`,
%      CGI, setInnerHTML, `div`);
csv_demo --> peek([E, X], X),
	{herbrand(E, IndexList)},
	csv:columns(IndexList, `,`, `<br>`),
	smash.

%- cgi:matrix_rule([1, set::pow([1,2])], y, R).
matrix_rule(L,x(N),A1):-!,
	nth1(N,L,A2),
	matrix_rule(L, A2, A1).
matrix_rule([A|A1], x, A2):-!,
	matrix_rule([A|A1],A,A3),
	matrix_rule([A|A1],A3,A2).
matrix_rule([A1,A|A2],y,A3):-!,
	matrix_rule([A1,A|A2],A,A4),
	matrix_rule([A1,A|A2],A4,A3).
matrix_rule([A1,A2,A|A3],z,A4):-!,
	matrix_rule([A1,A2,A|A3],A,A5),
	matrix_rule([A1,A2,A|A3],A5,A4).


%@ true.
%@ cgi.pl qcompiled.

%@ Context module: cgi.
%@ true.
