:- module(dev, []).
term_expansion --> pac:expand_pac.
:- use_module(pac(op)).

%%%%  Sequential Editing Commands
%%%%%    by  Kuniaki Mukai  SFC   2006 September ~
%%%%%
:- encoding(utf8).
:- set_prolog_flag(allow_variable_name_as_functor, true).

% ?- dev:collect_ascii(`Hello World !`,  A), smash(A).
%@ Hello---World---

collect_ascii --> sed(wl("[^a-zA-Z\\\n]+") / "---").


%%%%
ec--> edit_hawking_with_comment.

eo--> edit_hawking_without_comment.

ed--> edit_hawking_delete.

ei--> edit_hawking_insert.

es--> edit_hawking_short.

ecs--> edit_hawking_with_comment_short.

edit_hawking_with_comment --> region, peek(A, ["[[[[ ", A, " ====>\n",  A, " ]]]]\n{{{{ }}}}\n"]), overwrite.

edit_hawking_short --> region, peek(A, ["[[[[ ", A, " ====> ",  A, " ]]]]"]), overwrite.

edit_hawking_with_comment_short --> region, peek(A, ["[[[[ ", A, " ====> ",  A, " ]]]]　{{{{ }}}}"]), overwrite.

edit_hawking_without_comment --> region, peek(A, ["[[[[ ", A, " ====> ",  A, " ]]]]"]), overwrite.

edit_hawking_delete --> region, peek(A, ["[[[[ ", A, " ====>  ]]]]"]), overwrite.

edit_hawking_insert --> peek(["[[[[ ====> ]]]]"]), overwrite.


nkf(D, S, Options) :- eh:directory(D, nkf(S, Options)).

nkf(S, Options) :- eh:directory_filter(S, Fs0),
	insert(" ", Fs0, Fs),
	eh:sh(nkf(Options, Fs)).

suffix(L, Y):- member(X, L), sub_atom(Y,_,_,0,X), !.

%%%%

align --> region, env_align, overwrite.

env_equation(Label) -->
	peek(A, [ "\\begin{equation}", Label, "\n",  A,  "\\end{equation}\n"]).

env_align --> split,
	remove([]),
	maplist(phrase((split(`=`), align_row))),
	insert("\\\\\n"),
	peek(A, ["\\begin{align}\n", A, "\n\\end{align}\n"]).

align_row([X|Y], [X, "=\\;& "|Y]).

adj_space --> region, eh:sed_word(`  ` -> ` `), overwrite.

noindent --> region, eh:sed_word((`\n` + (*(` `)))  -> `\n`), overwrite.

number_order(E) :- number_order(E, Z), 	maplist(writeln, Z).

tails([X],[[X]]).
tails(X,[X|Xs]):- X=[_|X0], tails(X0, Xs).

convolute([],[], []).
convolute(As, Bs, Cs):-
	reverse(Bs, B0s),
	tails(B0s, TBs),
	tails(As, TAs),
	maplist(matrix:innerproduct, TAs, TBs, Cs).

contract_white --> wor_sed(
       (code(end_of_line), +(code(white);`　`))  ->  `\n`;
       +(code(white); `　`)                      ->  ` `).

tex_tex(A) --> region, texparse, eval(A), tex_codes, overwrite.

try_slide(env(try, L), env(slide, [group(`問題`), env(try,L)]), [], [], true).
try_slide(L, M, L, M, true):- listp(L).

%% description environment

env_description --> elisp:paragraph, maplist(parse_description),
	peek(A, ["\\begin{description}\n", A, "\\end{description}\n"]).

parse_description(P, ["\\item[", X, "] ", Y, "\n"]):-
	phrase(parse_description(X, Y), P).

parse_description(X, Y) --> expr(*code(white)),
	expr(X),
	expr(+code(white)),
	expr(rest,Y).

%%%%
edit_math(R) --> region, texparse,
	eval(in_math_mode(R)),
	tex_codes,
	overwrite.

do(X,Y):- conv_math(simple_greek, X, Y).

dom --> region, texparse, eval(in_math_mode(simple_greek)), tex_codes, 	overwrite.

a2g --> conv_math(map_alpha_greek).

mtr --> region, texparse, eval(in_math_mode(map_alpha_greek)), tex_codes, overwrite.

conv_math(Table) --> region, texparse, 	eval(conv_math(Table)), tex_codes, overwrite.

conv_math(Table, dol(L0), dol(L), [], [], eval(in_math_mode(Table), L0, L)).
conv_math(Table, ddol(L0), ddol(L), [], [], eval(in_math_mode(Table), L0, L)).
conv_math(X, Y, X, Y, true):- listp(X).

in_math_mode(T, X, Y, [], [], true):- call(T, X, Y).
in_math_mode(_, group(X), group(Y),  X, Y, true).
in_math_mode(T, env(E, L0), env(E, L), [], [], eval( conv_math(T), L0, L) ).
in_math_mode(_, X, Y, X, Y, true):- listp(X).

map_alpha_greek(0'R, cs(alpha)).
map_alpha_greek(0'U, cs(beta)).
map_alpha_greek(0'J, cs(gamma)).
map_alpha_greek(0'A, cs(alpha)).
map_alpha_greek(0'B, cs(beta)).
map_alpha_greek(0'C, cs(gamma)).
map_alpha_greek(0'D, cs(delta)).
map_alpha_greek(0'X, cs(xi)).
map_alpha_greek(0'Y, cs(eta)).
map_alpha_greek(0'Z, cs(zeta)).
map_alpha_greek(0'a, cs(alpha)).
map_alpha_greek(0'b, cs(beta)).
map_alpha_greek(0'c, cs(gamma)).

simple_greek(cs(greekA), cs(alpha)).
simple_greek(cs(greekB), cs(beta)).
simple_greek(cs(greekC), cs(gamma)).
simple_greek(cs(greekD), cs(delta)).
simple_greek(cs(greekX), cs(xi)).
simple_greek(cs(greekY), cs(eta)).
simple_greek(cs(greekZ), cs(zeta)).

idx --> region, sed(idx), overwrite.

idx([`\\index{x@`, Body, `}`, `\\`, Com, `{`, Body, `}`]) -->
    `\\`, expr(*(code(alpha)), Com), `{`, expr(Body), `}`.

idy --> region, make_index, overwrite.

make_index(Body, [`\\index{x@`, Body, `}`, Body]).

alpha_greek --> region, {alpha_greek_map(S)}, listsubst(S).

alpha_greek_map(
[(`A`,`\\alpha`), (`B`,`\\beta`), (`C`,`\\gamma`), (`D`,`\\delta`),
(`F`,`\\phi`), (`L`,`\\lambda`), (`M`,`\\mu`), (`N`,`\\nu`),
(`P`,`\\pi`), (`S`,`\\sigma`), (`T`,`\\tau`), (`X`,`\\xi`),
(`Y`,`\\eta`), (`Z`,`\\zeta`), (`a`,`\\alpha`), (`b`,`\\beta`),
(`c`,`\\gamma`), (`d`,`\\delta`), (`f`,`\\phi`), (`l`,`\\lambda`),
(`m`,`\\mu`), (`n`,`\\nu`), (`p`,`\\pi`), (`s`,`\\sigma`),
(`t`,`\\tau`), (`x`,`\\xi`), (`y`,`\\eta`), (`z`,`\\zeta`)]).

%%%%%%

mkrepsed(X, Dir) -->
     current(A),
     mkrepdcg(X, Dir),
     cons([`\n`, X, `--> region, `, `sed(`, X, `), overwrite.\n\n`]),
     cons(A),
     phrase_on_car(commentout).

commentout --> peek(A,  [ `/***********\n`, A, `\n************/\n`]).

%  r,mkrepdcg(sample, id) :   for normal direction
%  r,mkrepdcg(sample, swap) :  for opposite direction
mkrepdcg(Pred,Direction) --> split,
    maplist(split(+` `)),
    maplist(remove([])),
    remove([]),
    maplist(phrase((Direction, repdcg(Pred)))).

repdcg(Pred, [X,Y], Z):- quote(X,QX),  quote(Y,QY),
    format(codes(Z), `~w(~s) --> ~s.~n`, [Pred,QX,QY]).

ajaxCodes --> qstring, wrap_before_nl(`<script>smash(`,  `);</script>`), overwrite.

find --> sed(eh:replace).

select_id_name([X,_,_,Y|_],[X,`,`,Y,`\n`]).
select_id_name(_,`unallowed line`).

op    --> {elisp:send_to_lisp(`(shell-command \`open .\`)`)}.
ot    --> {elisp:send_to_lisp(`(mac-open-terminal)`)}.

path  --> peek(`(concat default-directory (buffer-name (current-buffer)))`),
	lisp.

ts --> {ensure_loaded(lib('convert-dcg'))}, region,sed(dcg_ts), overwrite.
st --> {ensure_loaded(lib('convert-dcg'))}, region,sed(dcg_st), overwrite.

t2h   -->  texparse, eval(tex2html).

% texuncomment --> texparse, eval(texuncomment), tex_codes.

% try2problem  -->  texparse, eval(try2problem), tex_codes.

rmnl  --> region, elisp:paragraph, remove([]), maplist(sed(rmnl)),
          insert(`\n\n`), overwrite.

rmnl([]) --> `\n`.

linesort --> split, sort, overwrite.

adj --> region, zenkaku_hankaku, overwrite.

zenkaku_hankaku --> listsubst([ (`。`,`．`), (`、`, `，`), (`　`, ` `), (`（`,`(`),
     (`）`,`)`),  (`｛`, `{`),  (`｝`, `}`) ]).

r(F)--> region, phrase(F).

ro(F)--> region, phrase(F), overwrite.

adj_punc    --> {ensure_loaded(lib('convert-dcg'))}, region,sed(punct),overwrite.

adjopp  -->{ensure_loaded(lib('convert-dcg'))}, region,sed(punct_opp),overwrite.

cr       --> convert_rule.

spaceproper  --> expr(+(` `)).

anc     --> region, anc0, overwrite.

anc0    --> expr(*(` `)), expr(X), spaceproper, expr(rest, Y),
	peek([`<a href=\``, X, `\`>` , Y, `</a>\n`]).

% a b
% c d
% ========>
% \begin{description}
% \item[a] b
% \item[c] d
% \end{description}


des  --> region, env_description, overwrite.

program --> region, peek(X, [`\\begin{program}\n`, X, `\\end{program}\n`]), overwrite.

% deit --> region, environment(itemize), dsnap,  overwrite.
% deit --> region, peek("hello\nworld!"), dsnap,  overwrite.
% deit --> region, peek(`hello\nworld!`), dsnap,  overwrite.

pre  -->  region, peek(A, [`<pre>\n`, A, `</pre>\n`]),  overwrite.

opt  -->  region, peek(A, [`<option>`, A, `</option>`]),  overwrite.

sli  --> region, slide(slide, ``), overwrite.
slif  --> region, slide(slide, `[method=file]`), overwrite.
wsli  --> region, slide(wideslide, ``), overwrite.
wslif  --> region, slide(wideslide, `[method=file]`), overwrite.

slide(E, Opts) --> expr(X  + (+(`\n`))),
	 peek(B,[`\\begin{`, E, `}`, Opts, `{`, X, `}\n`, B, `\\end{`, E, `}\n`]).

listcs --> region, listcs0, sort, insert(`\n`).

listcs0 --> texparse, eval(cs), flatten.

yacs([X]) --> tex_cs([cs(X)|Y],Y).
yacs([])  --> [_].

eol_space  --> listsubst([(`\n`, ` `)]).

named_env(Atom) --> expr(Opt), expr(`\n`), {atom_codes(Atom, Name)},
	{Beg = [`\\begin{` , Name , `}[` , Opt , `]\n`]},
	{End = [`\\end{`, Name , `}\n`]},
	peek(Body, [Beg, Body, End]).

% erase text
% ?- erase(`<option>`;`</option>`, `<option>hello</option>`, X), smashout(X).

tokenize(A) --> split(A), maplist(inverse(atom_codes)).

section_slide --> region,
    expr(_),
    (	`\\subsection`; `\\section`),
    sed(convert_slide),
    peek(Body,  [`\\begin{slide}`, Body, `\\end{slide}\n`]),
    overwrite.

convert_slide(`\n\\end{slide}\n\\begin{slide}`)-->
    expr(*`\n`), ( `\\subsection`;   `\\section`).

% CSV
cols(Ids) --> {maplist(aplam(#(x,nth1(x))), Ids, Call_list)},
	maplist(dual(mapdual, Call_list)).

mapdual(Data, Funs, Vals):- maplist(dual(Data), Funs, Vals).

aplam(X,Y,Z):- simplify(@(X,Y),Z).

aplam(X,Y,Z,U):- simplify(@(@(X,Y),Z),U).

sort_col(I) -->  predsort( compare_col(I) ).

compare_col(I,D,L1,L2) :- nth1(I,L1,A1), nth1(I,L2,A2), compare(D,A1,A2).

numbering(A,  (N,[[CodesOfN, ` `, A]|X] ),  (N1, X) ) :-
    plus(1, N, N1), atom_codes(N, CodesOfN).

numbering(X,Y):- foldl(numbering, X, (1,Y),  (_,[])).

fold_col(F,I,X,V,V1):- nth1(I,X,A), name(A1,A), call(F,A1,V,V1).

fold_col(F,I,X,Y):- foldl(fold_col(F,I), X, 0, Y).

fold_col(F,I) --> region, in, fold_col(F,I), atom_codes.

% Numbering lines of a region incuding blank lines.
line_number_nl --> region, split, current(X),
	{length(X, N),
	 numlist(1, N, Ns),
	 maplist([I, Line, [C, `. `, Line, `\n`]] :- number_codes(I, C),
		 Ns, X, R)
	},
	peek(R),
	overwrite.

% Numbering lines of a region skipping blank lines.
line_number --> line_number_l.	% line_number --> line_number_r.

line_number_l --> region, split, current(X),
	{ fold([Line, (J, A), (K, B)]:-
	      ( Line==[] -> A=[`\n`|B], K=J
	      ; K is J+1,
	        number_codes(K, C),
		A=[[C, `. `, Line, `\n`]|B]
	      ),
		X,  (0, R), (_, []))
	},
	peek(R),
	overwrite.

line_number_r --> region, split, current(X),
	{ foldr([Line, (K, B), (J, A)]:-
	      ( Line==[] -> A=[`\n`|B], K=J
	      ; K is J+1,
	        number_codes(K, C),
		A=[[C, `. `, Line, `\n`]|B]
	      ),
		X,  (_, []), (0, R))
	},
	peek(R),
	overwrite.


make_stable_marriage_problem(N, WP, MQ):- numlist(1, N, L),
	maplist(X\ inverse(=.., [w, X]), L, W),
	maplist(X\ inverse(=.., [m, X]), L, M),
	maplist(X\ shuffle(M), W, P),
	maplist(X\ shuffle(W), M, Q),
	zip([X,Y]\ inverse(=.., [-, X, Y]), W, P, WP),
	zip([X,Y]\ inverse(=.., [-, X, Y]), M, Q, MQ).
