:- module(dirfil, []).

% ?- working_directory(W, '/Users/cantor/local/lib/pacpl7').
% ?- dirfil:process(depth_first_walk, leaf_node, node((.), (.)), Y).

% ?- dirfil:process(replace_top_for_next, empty_list, [d((.), (.), S)], _).

% :- meta_predicate process(2, 1, +, -).
% process(_, Final, S, S):- call(Final, S).
% process(Delta, Final, S, S0):-
% 		call(Delta, S, S1),
% 		process(Delta, Final, S1, S0).

%
ignore_special((.)).
ignore_special((..)).
ignore_special('.git').
ignore_special('.DS_Store').


		/***************************************
		*     file_under/2 by process.         *
		***************************************/

% ?- file_under_by_fold('/path/to/directory/', Y).
file_under_by_fold(W, Y):-
		process(depth_first_walk, leaf_node, node(W, W), Y).

%
depth_first_walk(node((.), N), A):- !,
				working_directory(_, .),
				working_directory(W, W),
				depth_first_walk(node(W, N), A).
depth_first_walk(node(W, (.)), A):- !,
				depth_first_walk(node(W, W), A).
depth_first_walk(Node, A):- child_nodes(Node, Ss),
				member(A, Ss).

%
leaf_node(node(_, '.git')).
leaf_node(node(W, N)):- working_directory(_, W),
				  exists_file(N).

% ?-  child_nodes(node('/path/to/directory/', directory), L).
child_nodes(node((.), N), Ss):- !, working_directory(W, W),
	child_nodes(node(W, N), Ss).
child_nodes(node(W, (.)), Ss):- !,
	child_nodes(node(W, W), Ss).
child_nodes(node(W, N), Ss):-
	working_directory(_,W),
	exists_directory(N),
	directory_files(N, Fs),
	working_directory(W,N),
	working_directory(W0, W0),
	child_nodes(Fs, W0, Ss).

%
child_nodes([], _, []).
% child_nodes([X|Y], W, Z):- (X=(.);X=(..)), !,
child_nodes([X|Y], W, Z):- ignore_special(X), !,
	child_nodes(Y, W, Z).
child_nodes([X|Y], W, [node(W, X)|Z]):-
	child_nodes(Y, W, Z).

		/****************************************
		*     dict_of_files by process.		    *
		****************************************/

% ?- dict_of_files_by_fold('/path/to/directory/', S).
dict_of_files_by_fold(W, S):-
			process(replace_top_for_next, empty_list,
						 [d(W, W, S)], _).

%%  dict_of_files_by_fold(-D:<dict-like structure>) is det.
%   List files recursively under the working directory, and
%   unify D with the result in the form of the dict with
%	<subdirectory name> - <sub list> for subdirectories.
%
% ?- dict_of_files_by_fold(D).

dict_of_files_by_fold(S):-  working_directory(W,W),
			dict_of_files_by_fold(W, S).

%
empty_list([]).

%
replace_top_for_next([d(W, N, S)|X], Z):-
	working_directory(_, W),
	child_directories(N, S, Z, X).

%
child_directories(N, S, A, B):- exists_directory(N),
						 directory_files(N, S0),
						 working_directory(W, N),
						 working_directory(W0, W0),
						 residue_for_next(S0, W0, S, A, B),
						 working_directory(_, W).

%
residue_for_next([], _, [], A, A).
residue_for_next([I|X], W, Y, A, B):- ignore_special(I),!,
		residue_for_next(X, W, Y, A, B).
residue_for_next([I|X], W, [I|Y], A, B):- exists_file(I), !,
		residue_for_next(X, W, Y, A, B).
residue_for_next([I|X], W, [I-U|Y], [d(W,I,U)|A], B):-
		exists_directory(I),
		!,
		residue_for_next(X, W, Y, A, B).
residue_for_next([I|X], W, [I|Y], A, B):- residue_for_next(X, W, Y, A, B).
