:- module(pac_regex_prepare, []).

char_types(S) :-
	 Types =  [alnum, alpha, csym, csymf,digit,lower,cntrl,quote,white],
	 setof(T, X^(member(T, Types), char_type(X, T)), S0),
	 setof(paren(X), A^char_type(A, paren(X)), S1),
	 union(S0, S1, S).


types_intervals(X):- char_types(S),
	maplist(type_intervals, S, Y),
	maplist([T, Is, T-Is], S, Y, X).

type_intervals(T, Is):-
	setof(J, (between(0, 255, J), code_type(J, T)), Cs),
	maplist(code_char, Cs, Js),
	codes_intervals(Js, Is).

code_char(I, A):- char_code(A, I).

char_type_pairs(T0, T1):-
	char_types(S),
	member(T0, S),
	member(T1, S),
	T0 @< T1,
	char_type(X, T0),
	char_type(X, T1).

char_type_pairs(S):- setof(T0-T1, char_type_pairs(T0, T1), S).

%
type_interval_table([alnum-['0'-'9', 'A'-'Z', a-z],
		     alpha-['A'-'Z', a-z],
		     cntrl-['\000\'-'\037\', '\177\'],
		     csym-['0'-'9', 'A'-'Z', '_', a-z],
		     csymf-['A'-'Z', '_', a-z],
		     digit-['0'-'9'],
		     lower-[a-z],
		     quote-['"', '\'', '`'],
		     white-['\t', ' '],
		     paren(')')-['('],
		     paren(']')-['['],
		     paren('}')-['{']]).

codes_intervals([], []).
codes_intervals([A|Chars], Is):- codes_intervals(Chars, Js),
	codes_intervals(A, Js, Is).

%
codes_intervals(A, [], [A]).
codes_intervals(A, [X-Y|R], [A-Y|R]):- next_char(A, X).
codes_intervals(A, [X-Y|R], [A, X-Y|R]).
codes_intervals(A, [X|R], [A-X|R]):- atom(X),
	next_char(A, X).
codes_intervals(A, [X|R], [A, X|R]):- atom(X).
