:- module(gblex, [gb/2, gb/3, gb/4, rat/2, term_poly/2, poly_term/2]).
:- use_module(util(misc)).
% :- expects_dialect(pac).
term_expansion --> pac:expand_pac.
:- use_module(pac(op)).


% ?- gb([c,f, a,b,d,e, y, x],a*x+b*y-c; d*x+e*y-f, R).
%@ R = (y*e*a+ -1*y*d*b+d*c+ -1*a*f;x*a+y*b+ -1*c;x*d+y*e+ -1*f).
% ?- gb([x, c,f, a,b,d,e, y],a*x+b*y-c; d*x+e*y-f, R).
%@ R = (e*a*x-e*c-d*b*x+b*f;y*b+a*x-c;y*e+d*x-f) .
% ?- gb([c,f,a,b,d,e,x,y],a*x+b*y-c; d*x+e*y-f, R).
%@ R = (x*e*a-x*d*b-e*c+b*f;y*b+x*a-c;y*e+x*d-f) .
% ?- gb(x - t; y - t, A).
% ?- trace, gb(1, A).
% ?- gblex:gb([x,y], -x + y + 1;  -y + 2 * x, X).
% ?- gb(x^2+y^2-1; x+y-z, X).
%@ X = (-1+y^2+x^2;z-y-x).

% ?- gblex:gb_rev(x^2+y-3;x*y+y^2-1, X).
% ?- gb(x^2+y^2-1; x+y-k, X).
%@ X = (x^2-x*k+1/2*k^2-1/2;y+x-k)
% ?- gblex:gb_rev(x^2+y^2-1; x+y-k, X).
%@ X = (x^2+y^2-1;k-x-y) .
% ?- gb(x+y-z;x-y+3*z-2; x^2+y^2+z^2-k^2, X).
% ?- gblex:gb_ord([a,b,x,y, k], a*b-1; a*x-a*y; x-y-k, X).
%@ X = (b*a-1;y-x;k) .
% ?- F1=x^2-x*y-x-y^2+y, F2=x^3-x^2*y-x^2+x*y-y^3, gbtotal:gb([y, x], F1;F2, X).
% ?- gblex:gb(x*y*z;  y-x; y-2, R).
%@ R = (-2+x;-2+y;z).
% ?- gblex:gb([x, i, u], x^2-3; i^2+1; (x*a+ i*b)^2-u, R).
%@ R = (x^2-3;i^2+1;b^2-2*b*a*i*x-3*a^2+u).
% ?- gblex:gb([x(j), i, u], x(j)^2-3; i^2+1; (x(j)*a+ i*b)^2-u, R).
%@ R = (x(j)^2-3;i^2+1;b^2-2*b*a*i*x(j)-3*a^2+u).
% ?- gblex:gb([x(j), i, u], x(j)^3;x(j)^2-3; i^2+1; (x(j)*a+ i*b)^2-u, R).
% ?- gblex:gb([x(j), i, u], x(j)^2-3; i^2+1; (x(j)*a+ i*b)^2-u, R).
%@ R = (x(j)^2-3;i^2+1;b^2-2*b*a*i*x(j)-3*a^2+u)
% ?- gblex:gb((a+b+3+7)^3, R),write(R).
%@ b^3+3*b^2*a+30*b^2+3*b*a^2+60*b*a+300*b+a^3+30*a^2+300*a+1000
% ?- gblex:gb((a+b)^100, R),write(R).
% ?- gb([i, a], i^2+1; i^3*a+i*b, X).
% ?- gb(x-y; y-z; z-u; x-1; u,  X).
% ?- gb([r(3), i, a, b], r(3)^2-1; i^2+1; i^3*a+r(3)^3*i*b, X).
% ?- gb([r(3), i, b, a], r(3)^2-1; i^2+1; i^3*a+ i*b, X).
% ?- gb(r(3)^2-1; i^2+1; i^3*a+ i*b, X).
% ?- gb([ans,r(3),i], r(3)^2+1; i^2+1; (r(3)+i +a)^3; r(3)^4+ i^3-ans, X).
%@  ans^2-2*ans+2
% ?- gb([ans,r(3),i], r(3)^2+1; i^2+1; (r(3)+i +a)^3; r(3)^4+ i^4-ans, X).
%@ ans-2.
% ?- gb([ans,r(3),i], r(3)^2-3; i^2+1; (r(3)+i +a)^3; r(3)^4+ i^4-ans, X).
%@  ans-10;
% ?- trace, gb([], [], r(3)^2-1; i^2+1; i^3*a+ i*b, X).
%@ X = (b-a;i^2+1;r(3)^2-1) .

gb --> groebner_base.

gb(Ord, X, Y):- math:indeterminates(X, V),
	subtract(V, Ord, V0),
	sort(V0, V1),
	append(Ord, V1, U),
	groebner_base(U, X, Y).

gb(Pre, Post) --> gb_in, phrase(Pre), buchberger, phrase(Post), gb_out.

%
gb_rev(X, Y):- math:indeterminates(X, As),
	reverse(As, Bs),
	groebner_base(As, Bs, X, Y).

%  ?- gblex:groebner_base(x-y; y-z; z-u; x-1; u,  X).
groebner_base --> gb_in, !, buchberger, !, gb_out.
%
groebner_base(Ord) --> {sort(Ord, Ord0)}, groebner_base(Ord, Ord0).
%
groebner_base(Ord, Ord0, F, G):-
	zip(Ord, Ord0, A),
	subst(A, F, F0),
	groebner_base(F0, G0),
	zip(Ord0, Ord, B),
	subst(B, G0, G).
%
gb_in --> flatten(;),
	maplist(term_poly),
	maplist(canonical),
	remove([]).
%
gb_out --> predsort(compare_poly),
	maplist(poly_term),
	join_with_semicolon.

% ?- gblex:linear([],a, R).
% ?- gblex:linear([a],1, R).
% ?- gblex:linear([a], a*b, R).
% ?- gblex:linear([x,y], a*x+b*y+c;d*x+e*y+f, R), write(R).
%@ R = (x*e*a^2-x*d*b*a-f*b*a+e*c*a;y*e*a-y*d*b+f*a-d*c) .
% ?- E= (a*x+b*y+c;d*x+e*y+f), S = [ (a,1), (b,-1), (c,1), (d,1), (e,1), (f,3)], subst(S, E, E0), gblex:gb(E0, R).

%  Printing polynomial, base, affine form.
% ?- gblex:print_affine([x*[1*[]], y*[1*[]], [1*[]]]).
% ?- gblex:print_affine_base([[x*[1*[]], y*[1*[]],[1*[]]], [x*[1*[]], y*[1*[]],[1*[]]]]).
% ?- gblex:print_poly_base([[1*[x^1], 2*[y^2], 1*[]], [1*[x^1], 2*[y^2], 1*[]]]).
% ?- gblex:print_poly_base_exp(1+x+2*y^2;1+x+2*y^2).

%
print_affine([]):- !, writeln(0).
print_affine(B):- affine_to_poly(B, P),
	poly_term(P, T),
	writeln(T).

print_affine_base(B):- maplist(print_affine, B).

%
print_poly([]):- !, writeln(0).
print_poly(P):- poly_term(P, P0), writeln(P0).

print_poly_base([]):- !, writeln(0).
print_poly_base(B):- maplist(print_poly, B).

%
print_exp(E):- term_poly(E, E0), print_poly(E0).

print_poly_base_exp(B):- flatten((;), B, B1),
	maplist(print_exp, B1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% [2013/09/24] buchberger: Fourth version.
% policy: keep  head unique.
% clearly correct based on Dickson's theorem on mono ideal.
% The Dickson theorem implies that no infinite sequence of
% incomparables head terms exists.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
buchberger(X, Y):-  pairs(X, D),
	buchberger(D, X, Y0),
	normal_base(Y0, Y1),
	maplist(canonical, Y1, Y).

%
buchberger([], G, G).
buchberger([(A,B)|R], G, H):- s_poly(A, B, C),
	reduce_head(C, G, D0),
	canonical(D0, D),
	buchberger(D, R, G, H).

%
buchberger(C, R, G, H)	:- zero_poly(C), !, buchberger(R, G, H).
buchberger(C, _, _, [C]):- unit_poly(C), !.
buchberger(C, R, G, H)	:- foldr(C^[U, X, [(C,U)|X]], G, R, R0),
		buchberger(R0, [C|G], H).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% ?- gblex:s_poly([1*[x^1], 1*[a^1]],  [1*[y^1], 1*[b^1]], S).
%@ S = [1*[y^1,a^1],-1*[x^1,b^1]].

s_poly([C0*V0|P0], [C1*V1|P1], S):-
      mono_lcm(V0, V1, _, U0, U1),
      mul_poly_poly([C1*U0], P0, Q0),
      mul_poly_poly([C0*U1], P1, Q1),
      subtract_poly_poly(Q0, Q1, S0),
      canonical(S0, S).

% ?- gblex:normal_base([[1*[x^2], 1*[x^1], 1*[a^1]], [1*[x^1]], [1*[x^2]]], R).
% ?- gblex:normal_base([[1*[x^2]], [1*[]], [1*[x^2]]], R).
normal_base(X, Y):- select(P, X, Y0), reduce_poly_one(P, Y0, Q), !,
	normal_base([Q|Y0], Y).
normal_base(X, Y):- zero_poly(Z), remove(Z, X, Y).

% ?- gblex:normal_poly([1*[x^1], 1*[a^1]],  [[1*[x^1]], [1*[a^1]]], R).
normal_poly(P, G, Q):- reduce_poly_one(P, G, P0), !,
	normal_poly(P0, G, Q).
normal_poly(P, _, P).

% ?- gblex:normal_poly([1*[x^1], 1*[a^1]],  [[1*[x^2]], [1*[y^1]]]).
normal_poly(P, G):- reduce_poly_one(P, G, _), !, fail.
normal_poly(_, _).

% ?- gblex:normal_poly_one([1*[b^1, a^1], 5*[b^1], 3*[a^1]], [1*[b^1]], U).
%@ U = [1*[b^1],3 rdiv 5*[a^1]].
%@ U = [1*[b^1],3 rdiv 5*[a^1]].

normal_poly_one([1*P|R], [1*Q|S], U) :- div_mono_mono_v(P, Q, A),
	mul_mono_poly((-1)*A, S, S0),
	add_poly_poly(S0, R, U0),
	canonical(U0, U).

% ?-  gblex:reduce_head([1*[x^2], 3*[a^1]], [[1*[x^1], 4*[]]], R).
%@ R = [3*[a^1],16*[]].

reduce_head([C*M|X], G, P):- member([_C*N|Y], G), % Assuming G is canonical.
	div_mono_mono_v(M, N, Q),
	!,
	mul_mono_poly(C*Q, Y, Y0),
	subtract_poly_poly(X, Y0, P0),
	reduce_head(P0, G, P).
reduce_head(P, _, P).


% ?- gblex:reduce_mono(1*[x^1], [1*[x^1],2*[a^1]], P).
%@ P = [-2*[a^1]].
% ?- gblex:reduce_mono(1*[x^2], [1*[x^1],2*[a^1]], P).
%@ P = [-2*[x^1,a^1]].
%
reduce_mono(M, [M0|Ms], P):- div_mono_mono(M, M0, D),
	mul_mono_poly(D, Ms, P0),
	mul_scalar_poly(-1, P0, P).

% ?- gblex:reduce_poly_one([1*[x^2], 1*[x^1]], [[1*[x^1]]], R).
%@ R = [1*[x^1]] .
% ?- gblex:reduce_poly_one([1*[x^2], 1*[x^1]], [[1*[x^1],1*[a^1]]], R).
%@ R = [1*[x^1,a^1],-1*[x^1]] .
% ?- gblex:reduce_poly_one([1*[x^2], 1*[x^1]], [[1*[x^1]], [1*[a^1]]], R).
%@ R = [1*[x^1]] .
% ?- gblex:reduce_poly_one([1*[a^2]], [[1*[x^1]]], R).

reduce_poly_one(P, Ps, R) :-
	select(M, P, P0),
	member(Q, Ps),
	reduce_mono(M, Q, C),
	add_poly_poly(C, P0, R).

% ?- gblex:affine([x], [1*[x^1,b^1], 1*[x^1, a^1], [3*[]]], R).
% ?- gblex:affine([x,y], [1*[y^1], 1*[x^1,b^1], 1*[x^1, a^1], [3*[]]], R).
% ?- gblex:affine([x,y], [1*[y^1], 1*[x^1,b^1], 1*[x^1, a^1]], R).

affine([], Poly, [Poly]):-!.
affine([V|Vs], Poly, [V*A|L]):- extract_var(V, Poly, A, R), A\==[], !,
	affine(Vs, R, L).
affine([_|Vs], Poly, L):- affine(Vs, Poly, L).

% ?- trace, gblex:affine([x,y], [1*[y^1], 1*[x^1,b^1], 1*[x^1, a^1]], R), gblex:affine_to_poly(R, S).
%@    Call: (8) gblex:affine([x, y], [1*[y^1], 1*[x^1, b^1], 1*[x^1, a^1]], _G805) ? no debug
%@ R = [x*[1*[b^1],1*[a^1]],y*[1*[]],[]],
%@ S = [1*[x^1,b^1],1*[x^1,a^1],1*[y^1]] .

affine_to_poly(L, X):- ditribute_var(L, X0),
	append(X0, X1),
	merge_poly(X1, X).

ditribute_var([], []).
ditribute_var([[]|R], X):- !, ditribute_var(R, X).
ditribute_var([V*A|R], [B|X]):- mul_mono_poly(1*[V^1], A, B),
	ditribute_var(R, X).
ditribute_var([A|R], [A|Ps]):- ditribute_var(R, Ps).

% ?- gblex:extract_var(x, [1*[x^1,b^1], 1*[x^1, a^1], [3*[]]], A, R).
%@ A = [1*[b^1],1*[a^1]],
%@ R = [[3*[]]].

extract_var(_, [], [], []):-!.
extract_var(X, [C*M|P], [C*M0|A], R):- select(X^1, M, M0), !,
	extract_var(X, P, A, R).
extract_var(X, [Q|P], A, [Q|R]):- extract_var(X, P, A, R).


% Gauss's sweep method.
linear(Vs) --> gb_in, !,
	gauss_sweep(Vs),
	maplist(affine_to_poly),
	gb_out.

% ?- gblex: elim_linear_var([x*[2*[a^1]], [1*[]]], x, [1*[a^1]], [[1*[]]], R).
elim_linear_var(F, X, B, G0, H):- select(X*A, F, F0), !,
	sweep(A, B, F0, G0, H).
elim_linear_var(F, _, _, _, F).

%
sweep(A, B, P, Q, R):- div_poly_poly(A, B, J, []), !,
	mul_scalar_poly(-1, J, J0),
	mul_poly_affine(J0, Q, Q0),
	add_affine_affine(P, Q0, R).
sweep(A, B, P, Q, R):- div_poly_poly(B, A, J, []), !,
	mul_scalar_poly(-1, Q, Q0),
	mul_poly_affine(J, P, P0),
	add_affine_affine(P0, Q0, R).
sweep(A, B, P, Q, R):-
	mul_poly_affine(B, P, P0),
	mul_scalar_poly(-1, A, A0),
	mul_poly_affine(A0, Q, Q0),
	add_affine_affine(P0, Q0, R).

% ?- gblex: mul_scalar_affine(2, [x*[1*[b^1]], y*[1*[]]], R).
%@ R = [x*[2*[b^1]],y*[2*[]]] .
mul_scalar_affine(C, F, G):- mul_poly_affine([C*[]], F, G).

% ?- gblex: mul_poly_affine([1*[a^1]],[x*[1*[b^1]], y*[1*[]]], R).
%@ R = [x*[1*[b^1,a^1]],y*[1*[a^1]]] .
mul_poly_affine(_, [], []).
mul_poly_affine(A, [X*B|P], [X*C|Q]):- !, mul_poly_poly(A, B, C),
	mul_poly_affine(A, P, Q).
mul_poly_affine(A, [B|P], [C|Q]):- mul_poly_poly(A, B, C),
	mul_poly_affine(A, P, Q).

% ?- gblex: add_affine_affine([x*[1*[a^1]]],[x*[1*[b^1]], y*[1*[]]], R).
%@ R = [x*[1*[b^1],1*[a^1]],y*[1*[]]].
% ?- gblex: add_affine_affine([x*[-1*[a^1]]],[x*[-1*[b^1]], y*[1*[]]], R).
%@ R = [x*[-1*[b^1],-1*[a^1]],y*[1*[]]].
% ?- gblex: add_affine_affine([x*[-1*[a^1]]],[x*[1*[a^1]], y*[1*[]]], R).
%@ R = [x*[],y*[1*[]]].

add_affine_affine(P, Q, [X*C|R]):- select(X*A, P, P0), select(X*B, Q, Q0), !,
	add_poly_poly(A, B, C),
	add_affine_affine(P0, Q0, R).
add_affine_affine(P, Q, [X*A|R]):- select(X*A, P, P0), !,
	add_affine_affine(P0, Q, R).
add_affine_affine(P, Q, [X*A|R]):- select(X*A, Q, Q0), !,
	add_affine_affine(P, Q0, R).
add_affine_affine([], P, P):- !.
add_affine_affine(P, [], P):- !.
add_affine_affine([P], [Q], [R]):- add_poly_poly(P, Q, R).

% ?- E = (a*x+b*y+c;d*x+e*y+f), gblex:linear([x,y], E, R).
%@ R = (x*e*a^2-x*d*b*a-f*b*a+e*c*a;y*e*a-y*d*b+f*a-d*c) .

gauss_sweep(Vs, Ps, S):-  maplist(affine(Vs),  Ps, F),
	gauss_sweep([], F, Up, Low),
	append(Up, Low, S0),
	remove([[]], S0, S).

%
gauss_sweep(Up, Low, Up0, Low0):- select(P, Low, Low1), select(X*A, P, P0), !,
	maplist(pred([X, A, P0], [V, W]:- elim_linear_var(V, X, A, P0, W)), Low1, Low2),
	maplist(pred([X, A, P0], [V, W]:- elim_linear_var(V, X, A, P0, W)), Up, Up1),
	gauss_sweep([P|Up1], Low2, Up0, Low0).
gauss_sweep(U, L, U, L).

% [2013/09/23]
% some tiny

canonical([]).
canonical([1*_|_]).

canonical([],[]).
canonical([1*A|X], [1*A|X]):- !.
canonical([C*M|Ps], [1*M|Qs]):- C0 is 1 rdiv C,
	maplist(pred(C0, [D*M, D0*M]:- D0 is C0*D), Ps, Qs).


% ?- gblex:join_with_semicolon([a,b], R).
%@ R = (a;b).
% ?- gblex:join_with_semicolon([0,0], R).
%@ R = 0.

join_with_semicolon(X, Y):- remove(0, X, X0), join_with_semicolon_(X0, Y).

join_with_semicolon_([], 0):-!.
join_with_semicolon_([X], X):-!.
join_with_semicolon_([X, Y], (X;Y)):-!.
join_with_semicolon_([X, Y|Z], (X;U)):- join_with_semicolon_([Y|Z], U).

% ?- join_term(+, [a,b,c,d], Z).
% Z = a+b+c+d

join_term(F, [X, Y|Z], U):- !, join_term(F, [Y|Z], U0),
	U =.. [F,X,U0].
join_term(_, [X], X).

%
zero_poly([]).
unit_poly([1*[]]).
constant_poly([_*[]]).

%% read term -> internal form of polynomial
% ?- gblex:term_poly((a-b)^3, X), poly_term(X, Y).
%
term_rat(T, R):- rat(T, R0), rat_rat(R0, R).

rat_poly(r(X,[1*[]]), X):-!.
rat_poly(r(X,[C*[]]), Y):- C\==0, C0 is 1 rdiv C,
	mul_scalar_poly(C0, X, Y).

term_poly(L=R, P):- !, term_poly(L-R, P).
term_poly([], [[]]):- !.
term_poly(T, P):- term_rat(T, R), rat_poly(R, P).

rat_rat(r(P, Q), r(P0, Q0)):- poly(P, P0),
	poly(Q, Q0).

% internal polynomial -> term
% ?- gblex:poly_term([1*[a],1*[b],1*[c]], X).
% X = a+b+c

poly_term([], 0):-!.
poly_term(P, T) :- maplist(mono_term, P, Ts),
	join_term(+, Ts, T0),
	reduce:term_rewrite(gblex:normal_exp_left, [], T0, T1),
	reduce:term_rewrite(gblex:intro_prefix_minus_right, [], T1, T2),
	reduce:term_rewrite(gblex:intro_infix_minus, [], T2, T).

%
mono_term(1*M, T) :- !, mono_term_(M, T).
mono_term(C*M, C*T) :- mono_term_(M, T).

mono_term_([], 1):-!.
mono_term_(M, T):- join_term(*,  M, T).

%%%%
rat_term(R, T):- rat_poly(R, P), !, poly_term(P, T).
rat_term(r(X, Y), /(X0, Y0)):- poly_term(X, X0), poly_term(Y, Y0).

% ?- gblex: rat_pow( -1 rdiv 3, -3, R).
% R = -27.
rat_pow(X, J, Q) :- J >= 0, !, rat_pow(X, J, 1, Q).
rat_pow(X, J, Q) :- J0 is -J, rat_pow(X, J0, 1, Q0),
	Q is 1 rdiv Q0.

rat_pow(_, 0, P, P):-!.
rat_pow(X, J, P, Q):- J0 is J-1, P0 is P*X,
	rat_pow(X, J0, P0, Q).

%
term_term --> term_rat, rat_term.

%%%
%  ?- gblex:poly([1*[x^1]] - [1*[y^1]], P).
% P = [1*[x^1], -1*[y^1]].


basic_poly(0,[]):-! .
basic_poly(X,[X*[]]):-rational(X),! .
basic_poly(X,[1*[X^1]]):-! .


% ?- gblex:poly(3*x*y, X).
%@ X = [3*[y^1,x^1]] .


poly(0,[]):-! .
poly([X|Y],[X|Y]):-! .
poly(X,[X*[]]):-rational(X),! .
poly(X-Y,A1):-!,(poly(X,A2),poly(Y,A3)),subtract_poly_poly(A2,A3,A1) .
poly(X+Y,A1):-!,(poly(X,A2),poly(Y,A3)),add_poly_poly(A2,A3,A1) .
poly(X*Y,A1):-!,(poly(X,A2),poly(Y,A3)),mul_poly_poly(A2,A3,A1) .
poly(X^N,A1):-!,poly(X,A2),power_poly(N,A2,A1) .
poly(X,[1*[X^1]]):-! .


% ?-gblex:rat((1+x)/y, R).
%@ R = r((1*1+x*1)*1, 1*1*y)


rat(+X,A1):-!,rat(X,A1) .
rat(-X,A1):-!,rat(-1*X,A1) .
rat(X-Y,A1):-!,rat(X+ -1*Y,A1) .
rat(X+Y,A1):-!,(rat(X,A2),rat(Y,A3)),'pac#431'(A2,A3,A1) .
rat(X*Y,A1):-!,(rat(X,A2),rat(Y,A3)),'pac#432'(A2,A3,A1) .
rat(X/Y,A1):-!,(rat(X,A2),rat(Y,A3)),'pac#433'(A2,A3,A1) .
rat(X^N,A1):-!,rat(X,A2),'pac#434'(A2^N,A1) .
rat(X,r(X,1)):-! .

'pac#431'(r(A,B),r(C,D),r(A*D+C*B,B*D)):-true .
'pac#432'(r(A,B),r(C,D),r(A*C,B*D)):-true .
'pac#433'(r(A,B),r(C,D),r(A*D,B*C)):-true .
'pac#434'(r(B,C)^N, r(B^N,C^N)):-true .

% :- bekind(rat,[]).
% +X	=  X.
% -X	=  (-1)* X.
% X-Y	=  X + (-1)*Y.
% X+Y	=  #([r(X0, X1), r(Y0, Y1)]\ `r(X0*Y1+Y0*X1, X1*Y1))@X@Y.
% X*Y	=  #([r(X0, X1), r(Y0, Y1)]\ `r(X0*Y0, X1*Y1))@X@Y.
% X/Y	=  #([r(X0, X1), r(Y0, Y1)]\ `r(X0*Y1, X1*Y0))@X@Y.
% X^N	=  #(N^[r(X0, X1)]\ `r(X0^N, X1^N))@X.
% X	=  `r(X, 1).
% :-ekind.

% ?- listing(gblex:rat).

% ?- gblex:mono_merge([c^1, a^1], [d^1, b^2, a^1], X).

mono_merge([], X, X):-!.
mono_merge(X, [], X):-!.
mono_merge([X^N|Xs], [X^N0|Xs0], [X^N1|Zs]):- !,
	N1 is N+N0, mono_merge(Xs, Xs0, Zs).
mono_merge([X^N|Xs],[X0^N0|Xs0], [X^N|Zs]):-  X @> X0, !,
	mono_merge(Xs, [X0^N0|Xs0], Zs).
mono_merge([X^N|Xs], [X0^N0|Xs0], [X0^N0|Zs]):-
	mono_merge([X^N|Xs], Xs0, Zs).

% ?- gblex:compare_mono(X, [y^1, x^2],[y^2, x^1]).
%@ X = (<).

compare_mono(C, [], []):-!, C=(=).
compare_mono(C, [], _):-!, C=(<).
compare_mono(C, _, []):-!,  C=(>).
compare_mono(C0, [X^N|Xs], [Y^N0|Xs0]):- X==Y, !, compare(C, N, N0),
	(  C == (=)  ->  compare_mono(C0, Xs, Xs0) ;  C0 = C  ).
compare_mono(C, [X^_|_], [Y^_|_]):- X @< Y, !, C= (<).
compare_mono((>), _, _).


% ?- gblex:compare_poly(X, [1*[y^2, x^2]],[1*[y^1, x^2]]).
%@ X = (>).
compare_poly(C, [], []):-!, C=(=).
compare_poly(C, [], _):-!,  C=(<).
compare_poly(C, _, []):-!,  C=(>).
compare_poly(C0, [_*M|Ms], [_*N|Ns]):-
	compare_mono(C, M, N),
	(	C == (=)
	->	compare_poly(C0, Ms, Ns)
	;       C0 = C
	).

% ?- gblex:sort_poly_list([[1*[a^1]], [1*[a^1]], [1*[b^1]]], X).
%@ X = [[1*[b^1]],[1*[a^1]],[1*[a^1]]].
sort_poly_list(X, Y):-
	predsort( pred(( [A, [_*M|_], [_*M0|_]]:-
			gblex:compare_mono(A0,  M, M0),
			(	A0 == (>)
			->	A  = (<)
			;	A  = (>))
		  )),
		  X, Y).

% ?- gblex:sort_poly([1*[a^1], 1*[b^1], 1*[a^1]], X).
%@ X = [1*[b^1],1*[a^1],1*[a^1]].
sort_poly(X, Y):-
	predsort(pred( ([A, _*M, _*M0]:-
			gblex:compare_mono(A0,  M, M0),
			(	A0 == (>)
			->	A  = (<)
			;	A  = (>))
		  )),
		  X, Y).

merge_poly([], []).
merge_poly([0*_|Xs], Xs0):- !, merge_poly(Xs, Xs0).
merge_poly([X], [X]).
merge_poly([C*M,C0*M|Xs], Xs0):- !, C1 is C+C0,
	merge_poly([C1*M|Xs], Xs0).
merge_poly([X|Xs], [X|Xs0]):- merge_poly(Xs, Xs0).

% ?- gblex:mul_mono_mono(1*[c^1, a^1], 2*[c^1, b^2], X).
%@ X = 2*[c^2,b^2,a^1].
mul_mono_mono(C*M, C0*M0, C1*M1):- C1 is C0*C, mono_merge(M, M0, M1).

mul_scalar_poly(C, P, Q):- maplist(mul_scalar_mono(C), P, Q).

mul_scalar_mono(C, C0*M, C1*M):- C1 is C0*C.

% ?- gblex:div_mono_mono(-1*[b^2, c^1, a^3], 1*[b^1, a^1], D).
div_mono_mono(C*M, C0*M0, C1*M1):- C1 is C rdiv C0,
	div_mono_mono_v(M, M0, M1).

div_mono_mono_v(M, [], M):- !.
div_mono_mono_v([X^I|Ys], [X^J|Ys0], Zs):- !,
	( I == J -> div_mono_mono_v(Ys, Ys0, Zs)
	; I> J, K is I-J, Zs=[X^K|Zs0], div_mono_mono_v(Ys, Ys0, Zs0)
	).
div_mono_mono_v([X^I|Ys], [X0^J|Ys0], [X^I|Zs]):- compare((>), X, X0),
	div_mono_mono_v(Ys, [X0^J|Ys0], Zs).

% div_poly_mono_v(A, B, Q, R)

% % ?- gblex:div_poly_mono([1*[x^2], 2*[x^1], 1*[]], [2*[x^1], 2*[]], X, []).

% div_poly_mono(A, C*B, Q, R):- div_poly_mono_v(A, B, Q0, R),
% 	C0 is 1 rdiv C,
% 	mul_scalar_poly(C0, Q0, Q).

% %
% div_poly_mono_v([], _, [], []).
% div_poly_mono_v([C*M|A], N, [C*N0|Q], R):- div_mono_mono_v(M, N, N0, []), !,
% 	div_poly_mono_v(A, N, Q, R).
% div_poly_mono_v([W|A], N, Q, [W|R]):-
% 	div_poly_mono_v(A, N, Q, R).

% ?- gblex:div_poly_poly([1*[x^2], 2*[x^1], 1*[]], [2*[x^1], 2*[]], X, []).
%@ X = [1 rdiv 2*[x^1],1 rdiv 2*[]].
div_poly_poly([], _, [], []):-!.
div_poly_poly([M|A], [N|B], [N0|Q], R):- div_mono_mono(M, N, N0), !,
	mul_mono_poly(N0, B, B0),
	subtract_poly_poly(A, B0, P),
	div_poly_poly(P, [N|B], Q, R).
div_poly_poly(P, _, [], P).

% ?- gblex:div_poly_poly([1*[x^2], 2*[x^1], 1*[]], [2*[x^1], 1*[]], X).
div_poly_poly(P, Q, R):- div_poly_poly(P, Q, R, []).

% ?- gblex:div_poly_poly([1*[x^2], 2*[x^1], 1*[]], [1*[x^1], 1*[]]).
%@ true.
div_poly_poly(P, Q):- div_poly_poly(P, Q, _, []).

% ?- gblex:add_poly_poly([1*[b^1], 2*[a^1]], [1*[a^1]], X).
%@ X = [1*[b^1],3*[a^1]].
add_poly_poly([], X, X):-!.
add_poly_poly(X, [], X):-!.
add_poly_poly([C*M|Xs],[C0*M0|Xs0], Zs):-  compare_mono(B, M, M0),
	(   B==(=)
	->  C1 is C+C0,
		(	C1 == 0
		->	Zs=Zs0
		;	Zs=[C1*M|Zs0]
		),
	    add_poly_poly(Xs, Xs0, Zs0)

	;   B==(>)
	->  Zs=[C*M|Zs0],
	    add_poly_poly(Xs, [C0*M0|Xs0], Zs0)

	;   Zs=[C0*M0|Zs0],
	    add_poly_poly([C*M|Xs], Xs0, Zs0)
	).

% ?- gblex:mul_mono_poly(1*[c^1], [1*[b^1], 1*[a^1]], X).
%@ X = [1*[c^1,b^1],1*[c^1,a^1]].
mul_mono_poly(M, P, P0):- maplist(mul_mono_mono(M),P, P0).

% ?-numlist(1, 10000, L), maplist([I, I*[x^I]], L, P), gblex:mul_scalar_poly(-1, P, Q), time(100, gblex:subtract_poly_poly(P, Q, S), T).
%@ T = 1.155068.

subtract_poly_poly([], X,  Y):-	!, mul_scalar_poly(-1, X, Y).
subtract_poly_poly(X, [],  X):-	!.
subtract_poly_poly([C*M|Xs],[C0*M0|Xs0], Zs):-  compare_mono(B, M, M0),
	(   B==(=)
	->  C1 is C-C0,
		(	C1 == 0
		->	Zs=Zs0
		;	Zs=[C1*M|Zs0]
		),
	    subtract_poly_poly(Xs, Xs0, Zs0)

	;   B==(>)
	->  Zs=[C*M|Zs0],
	    subtract_poly_poly(Xs, [C0*M0|Xs0], Zs0)

	;   C1 is  -C0,
	    Zs=[C1*M0|Zs0],
	    subtract_poly_poly([C*M|Xs], Xs0, Zs0)
	).

% ?- gblex: (mul_poly_poly([1*[b^1], 1*[a^1]],[1*[b^1], 1*[a^1]], X), mul_poly_poly(X, X, Y)).
mul_poly_poly([], _, []):- !.
mul_poly_poly(_, [], []):- !.
mul_poly_poly([1*[]], P, P):- !.
mul_poly_poly(P, [1*[]], P):- !.
mul_poly_poly(P, Q, R):- mul_poly_poly(P, Q, [], R).

mul_poly_poly([], _, P, P):- !.
mul_poly_poly([M|Ms], P, Q, R):- mul_mono_poly(M, P, P0),
	add_poly_poly(P0, Q, Q0), mul_poly_poly(Ms, P, Q0, R).

% ?- gblex:power_poly(3, [2*[a^2]], U).
power_poly(0, _, U):- !, unit_poly(U).
power_poly(N, X, U):- N0 is N>>1, M is N mod 2,	power_poly(N0, X, U0),
	mul_poly_poly(U0, U0, U1),
	(M==1 -> mul_poly_poly(X, U1, U); U=U1).

% ?-gblex:mono_lcm([b^2, a^1], [b^1, a^2], LCM, Y, Z).
mono_lcm([], Y, Y, Y, []):-!.
mono_lcm(X, [], X, [], X):-!.
mono_lcm([V^I|X],[V^J|Y], [V^M|Z], P, Q):- !, compare(C, I, J),
	( C == (=) -> M=I, mono_lcm(X, Y, Z, P, Q)
	; C == (>) -> M=I, K is I-J, Q=[V^K|Q0], mono_lcm(X, Y, Z, P, Q0)
	; M =J, K is J-I, P=[V^K|P0], mono_lcm(X, Y, Z, P0, Q)
	).
mono_lcm([U^I|X],[V^J|Y], [U^I|Z], P, [U^I|Q]):- V @< U,!,
	mono_lcm(X, [V^J|Y], Z, P, Q).
mono_lcm(X,[B|Y], [B|Z], [B|P], Q):- mono_lcm(X, Y, Z, P, Q).


% basic laws  for simplifying normal_exp_left internal expressions
% ?- term_rewrite(gblex:normal_exp_left, [], T0, T).
% ?- term_rewrite(gblex:normal_exp_left, [], (1*x+0*x)* (1*x), T).
%@ T = x*x.
% ?- term_rewrite(gblex:normal_exp_left, [], (1*x - 0*x)* (-1*x), T).
%@ T = -1*x*x.

% ?- poly_term([1*[y^1], -1*[x^1]], X).
%@ X = y-x.

% Stage 1
normal_exp_left([],	0).
normal_exp_left(0+X,	X).
normal_exp_left(+X,	X).
normal_exp_left(1*X,	X).
normal_exp_left(0*_,	0).
normal_exp_left(X^1,	X).
normal_exp_left(_^0,	1).
normal_exp_left(A*B,	C):- rational(A), rational(B), C is A*B.
normal_exp_left(A+B,	C):- rational(A), rational(B), C is A+B.
normal_exp_left(A^B,	C):- rational(A), rational(B), C is A^B.
normal_exp_left(-A,	C) :- rational(A), C is -A.
normal_exp_left(-A,	(-1)*A).
normal_exp_left(A*(B*C),	(A*B)*C).
normal_exp_left(A+(B+C),	(A+B)+C).
normal_exp_left(A*B,	B*A)	:- rational(B).
normal_exp_left(A+B,	B+A)	:- rational(B).
normal_exp_left(X*(Y+Z),	X*Y + X*Z).
normal_exp_left((X+Y)*Z,	X*Z + Y*Z).
normal_exp_left(X - Y,		X+ (-1)*Y).
normal_exp_left(rdiv(A,B),	A/B).

% Stage 2
intro_prefix_minus_right(1*A, A).
intro_prefix_minus_right(A+B, A+C):- extend_prefix_minus_scope(B, C).
%
extend_prefix_minus_scope(A,  -(B)):- rational(A), A<0, B is -A.
extend_prefix_minus_scope(A*B, C*B):- extend_prefix_minus_scope(A, C).
extend_prefix_minus_scope(-(A)*B, -(A*B)).

% Stage 3 (final)
intro_infix_minus(A + -(B), A - B).

% ?- gblex:plain_geo_test(X).
plain_geo_test(X):- F =
(a*a1-1; b*b1-1; c*c1-1; d*d1-1; e*e1-1;
 f*f1-1; g*g1-1; h*h1-1; i*i1-1; j*j1-1;
 k*k1-1; l*l1-1; m*m1-1;
e*(a+b)-f*(c+d);
b*l-m*(c+d);
b*(f+e)-a*(c+d);
i*e-k*(c+d);
g*d-h*(f+e);
h*(a+b)-g*c;
d*(a+b)-c*(f+e);
k*(a+b)-i*f;
m*(f+e)-l*a;
m+l-(i+k);
(a+b)-(f+e)-x),
gb([    x,
	a,  b,  c,  d,  e,  f,  g,  h,  j,  m,  l,  i,  k,
   	a1, b1, c1, d1, e1, f1, g1, h1, j1, m1, l1, i1, k1
       ], F, X),
writeln(X).
