:- module(gb, []).

:-use_module(pac('expand-pac')).
:-use_module(gb('gb-vector')).
:-use_module(gb('bdfm-generic')).

term_expansion --> pac:expand_pac.
:- use_module(pac(op)).

% ?- trace, gb:test(X).
%
default_option([
		order(lexical),
		prime(0),				% 2147483647,	17
		remove_content(24),		% timing bit length
		fraction_free(true),
		vars([]),
		homogeneous(false),
		modular(false),
		minimal(true),
		trace(false),
		vector(false),
		initial_vector_size(8),
		gc(false),
		trim(false),
		exp(true),
		verbose(false)
		]).

%
generic_normal_poly(true, Prime, Poly, R):- !,
	(	Prime == 0
	->	once(poly:remove_content_poly(Poly, R))
	;	once(poly:galois_poly(Prime, Poly, R))
	).
generic_normal_poly(_, _, Poly, R):- once(poly:canonical_form(Poly, R)).

% ?- completing_option([], Y).
completing_option(X,  Y):- default_option(Ds),
 		completing_option(Ds, X, Y).

%
completing_option([], X, X).
completing_option([A| As], X,  Y):-
	(	check_option(A, X)
	->	X0 = X
	;	X0 = [A|X]
	),
	completing_option(As, X0, Y).

check_option(A, X):- functor(A, F, 1),
	functor(B, F, 1),
	memberchk(B, X).

% ?- gb:help.
help :-  default_control(X), writeln(X).

% ?- gb:option_to_zip([a(1), b(2)], Z),
option_to_zip(X, Y):- maplist(pred([A, K-V]:- A =..[K,V]), X, Y).

%
% ?- gb:gb((a;b), Y, []).
% ?- gb:gb(a, Y, []).
% ?- gb:gb((a+b; a-b), Y, [vars([a,b])]).
% ?- gb:gb((a+b; a-b), Y, [vars([b,a])]).
% ?- gb:gb((a^2+b^2-1; a-b), Y, []).
% ?- gb:gb((a^2+b^2-1; a-b), Y, [vars([b,a])]).
% ?- gb:gb((a+b)^100, Y, []).
% ?- gb:gb((a+b-c)^20, Y, []).
% ?- gb:gb(x^2+y^2-4; x*y-1,  Y, []).
% ?- gb:gb((a^2+b^2-1; a-b), Y, [homogeneous(true)]).
% ?- gb:gb((a^2+b^2-1; a-b), Y, [homogeneous(true), vector(true)]).
% ?- gb:gb(a, Y, [vector(true)]).
% ?- gb:gb(a;b, Y, [vector(true)]).
% ?- gb:gb(a;b;a, Y, [vector(true)]).
% ?- gb:gb(a-b;b+a, Y, [vector(true)]).
%@ Y = [b, a] .

% ?- gb:gb((a+b)^100, Y, [vector(true)]).
% ?- gb:gb((a+b-c)^20, Y, [vector(true)]).
% ?- gb:gb((a^2+b^2-1; a-b), Y, [vector(true)]).
% ?- gb:gb((a^2+a-1; a), Y, [vector(true)]).
% ?- gb:gb((a^2+a-1; a-1), Y, [vector(true)]).

% ?- gb:gb(r(3)^2+1; i^2+1; (r(3)+i +a)^3; r(3)^4+ i^3-ans, X, [vars([ans,r(3),i])] ).
% ?- gb:gb(r(3)^2+1; i^2+1; (r(3)+i +a)^3; r(3)^4+ i^3-ans, X, [vars([ans,r(3),i]), vector(true)] ).

% ?- F1 = x^5+y^4+z^3-1, F2=x^3+y^3+z^2-1, time(gb:gb(F1;F2, X, [vars([z,y,x])])).
%@ % 123,369,311 inferences, 14.531 CPU in 14.596 seconds (100% CPU, 8490174 Lips)
% ?- F1 = x^5+y^4+z^3-1, F2=x^3+y^3+z^2-1, time(gb:gb(F1;F2, X, [vars([z,y,x]),vector(true)])).
%@ % 255,189,436 inferences, 44.758 CPU in 44.867 seconds (100% CPU, 5701514 Lips)
%@ F1 = x^5+y^4+z^3-1,
%@ F2 = x^3+y^3+z^2-1,
%@ X = [... + ... + ... * ... - 1617* ... ^ ... - 3228*z^7-1614*z^6+378*z^5+872*z^4+432*z^3+80*z^2, ... + ... + ... * ... - 27268337* ... ^ ... - 14641598*z^4+1214234*z^3+2619854*z^2+139968*z-69984, ... + ... - ... * ... - 26735354* ... ^ ... - 7526542*z^3+650492*z^2+34992*z-34992, ... - ... - ... * ... - 9673136* ... ^ ... - 95162*z^2+31104*z-31104, x^3+y^3+z^2-1, ... + ... + ... ^ ... + z^3-1, ... - ... + ... * ... - 1, ... + ... - ... * ...] .

% ?- F1 = x^5+y^4+z^3-1, F2=x^3+y^3+z^2-1, time(gb:gb(F1;F2, X, [order(total), vars([z,y,x]),vector(true)])).
% ?- F1 = x^5+y^4+z^3-1, F2=x^3+y^3+z^2-1, time(gb:gb(F1;F2, X, [vars([x,y, z])])).
%@ % 406,849 inferences, 0.039 CPU in 0.040 seconds (98% CPU, 10311461 Lips)
% ?- F1 = x^5+y^4+z^3-1, F2=x^3+y^3+z^2-1, time(gb:gb(F1;F2, X, [vars([x,y, z]), vector(true)])).
%@ % 1,938,943 inferences, 0.225 CPU in 0.228 seconds (99% CPU, 8600858 Lips)


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  Do not try the following two, which seem to take     %
%  unexpectedly long time. bug ?                        %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% ?- F1 = x^5+y^4+z^3-1, F2=x^3+y^3+z^2-1, gb:gb(F1;F2, X, [order(total), vars([x,y, z]), vector(true)]).
% ?- F1 = x^5+y^4+z^3-1, F2=x^3+y^3+z^2-1, gb:gb(F1;F2, X, [order(total), vars([x, y, z])]).

gb(X, Y, Opt):- gb_setup(X, Opt, U, Opt0), !,
				gb_BDFM(U, V, Opt0), !,
				gb_clean(V, Y, Opt0).

gb_setup(X, Opt, Y, [assoc(Assoc)|Opt0]):-
		completing_option(Opt, Opt0),

		if( memberchk(verbose(true), Opt),
			(	writeln("Options:"),
				maplist(pred([X]:-format("\t~w\n", [X])), Opt)
			),
			true),
		memberchk(vars(Vars),  Opt0),
		poly:poly_in(Vars, Assoc, X, X0),

		if( memberchk(homogeneous(true), Opt0),
			maplist(poly:make_poly_homogeneous(0), X0, X1),
			X1 = X0),

		if( memberchk(fraction_free(true), Opt0),
			Y = X1,
			maplist(poly:canonical_form, X1, Y)).

%
gb_clean(X, Y, Opt):-
	if( memberchk(vector(true), Opt),
	    vector:list_to_vector(X0, X),
	    X0 = X),

	if( memberchk(gc(true),	Opt), garbage_collect),

	if( memberchk(trim(true), Opt), trim_stacks),

	if( memberchk(minimal(true), Opt),
	    minimal_base(X0, X1, Opt),
	    X1 = X0),

	if( memberchk(homogeneous(true), Opt),
		(	memberchk(order(Ord), Opt),
			maplist(poly:make_poly_non_homogeneous(Ord), X1, X2)
		),
	    X2 = X1),

	if( memberchk(exp(true), Opt),
		(	memberchk(assoc(Assoc), Opt),
			poly:poly_out(Assoc, X2, Y)
		),
		Y = X2).

gb_BDFM(X, Y, Opt):-
	memberchk(modular(Mod), Opt),
	memberchk(vector(Vec),  Opt),
	if(Mod,
	   gbtrace:gb_BDFM(X,  Y,  Opt),    % <=  Not yet implemented.
	   if(Vec,
			gvec:gb_BDFM(X, Y, Opt),    % <= using vectors (experimental)
			bdfm:gb_BDFM(X, Y, Opt))).	% <= usual list programming

%  only for debugging
compare_gb_methods(C, Opt1, Opt2, X):-
	gb([exp(false)|Opt1], X, Y1, CK1),
	gb([exp(false)|Opt2], X, Y2, CK2),
	memberchk(order-Ord1, CK1),
	memberchk(order-Ord2, CK2),
	maplist(pred(Ord1, ([A, B]:- poly:sort_poly(Ord1, A, A0),
				poly:canonical_form(A0, B))), Y2, Z2),
	maplist(pred(Ord2, ([A, B]:- poly:sort_poly(Ord2, A, A0),
				poly:canonical_form(A0, B))), Y1, Z1),
	maplist(pred([Y1,Ord1],
		     ([U, V]:- reduce_by_polyset(U, Y1, V, Ord1))), Z2, R2),
	maplist(pred([Y2,Ord2],
		     ([U, V]:- reduce_by_polyset(U, Y2, V, Ord2))), Z1, R1),
	compare_gb(C, R1, R2).

%
compare_gb(C, R1, R2):- maplist(=([]), R1), !,
	(	maplist(=([]), R2)
	->	C = (=)
	;	C = (<)
	).
compare_gb(C, _, R2):-
	(	maplist(=([]), R2)
	->	C = (>)
	;	C = (incomparable)
	).

% ?- gb: ideal_membership(a, a, []).
% ?- gb: ideal_membership(a, a;b, []).

ideal_membership(P, Q, Opt0):-
	gb([exp(false)|Opt0], Q, GB, Opt),
	memberchk(order(Ord), Opt),
	memberchk(assoc(Assoc), Opt),
	poly:number_vars(P, Assoc, P0),
	poly:poly_in(P0, Ps),
	maplist(poly:sort_poly(Ord), Ps, Ps0),
	maplist(pred([GB,Ord],
		     ([U, R]:- once(reduce_by_polyset(U, GB, R, Ord)))),
		Ps0, Rs),
	!,
	maplist(=([]), Rs).

% ?- gb:minimal_base([[1*[x^2], 1*[x^1], 1*[a^1]], [1*[x^1]], [1*[x^2]]], R,

minimal_base(X, Y, Opt):-
	subset([order(Ord), fraction_free(FF), prime(BL)], Opt),
	reduce_poly_head(X, Y0, Ord, FF, BL, _Trace, []), % <== indispensable  ([2014/07/20])
	normalize_base(Y0, Y, Ord, FF, BL).
%
reduce_poly_head(X, Y, Ord, FF, BL, [P|T0], T):-
	select(P, X, Y0),
	reduce_head_by_polyset_one(P, Y0, Q, Ord, FF, T0, T1),
	!,			% <== indispensable ([2014/07/21])
	(   poly:zero_poly(Q)
	->	reduce_poly_head(Y0, Y, Ord, FF, BL, T1, T)
	;	once(generic_normal_poly(FF, BL, Q, Q0)),
		reduce_poly_head([Q0|Y0], Y, Ord, FF, BL, T1, T)
	).
reduce_poly_head(X, X, _, _, _, T, T).

%
normalize_base(X, Y, Ord, FF, BL):- select(P, X, Y0),
	reduce_by_polyset_one(P, Y0, Q, Ord, FF, _Trace, []),
	!,
	(   poly:zero_poly(Q)
	->	normalize_base(Y0, Y, Ord, FF, BL)
	;	once(generic_normal_poly(FF, BL, Q, Q0)),
		normalize_base([Q0|Y0], Y, Ord, FF, BL)
	).
normalize_base(X, X, _, _,_).

%
reduce_head_by_polyset(X, Ps, R, Ord, FF, T0, T):-
	reduce_head_by_polyset_one(X, Ps, Z, Ord, FF, T0, T1), !,
	reduce_head_by_polyset(Z, Ps, R, Ord, FF, T1, T).
reduce_head_by_polyset(X, _, X, _, _, T, T).

%
% reduce_head_by_polyset_one([M|X0], Ps, R, Ord, FF):-
% 	poly:rev_member([N|Y], Ps),
% %	member([N|Y], Ps),
% 	poly:div_mono_mono(M, N, D),
% 	!,
% 	poly:mul_mono_poly(D, Y, Q),
% 	poly:subtract_poly_poly(X0, Q, R, Ord).

reduce_head_by_polyset_one(X, Ps, R, Ord, true, [P|T], T):- !,
	poly:rev_member(P, Ps),
%	member(P, Ps),
	once(reduce_head_by_poly_z(X, P, R, Ord)).
reduce_head_by_polyset_one(X, Ps, R, Ord, _, [P|T0], T):-
	poly:rev_member(P, Ps),
%	member(P, Ps),
	once(reduce_head_by_poly(X, P, R, Ord, T0, T)).

reduce_by_polyset(X, Ps, R, Ord, FF):-
	reduce_by_polyset_one(X, Ps, Z, Ord, FF, _Trace, []), !,
	reduce_by_polyset(Z, Ps, R, Ord, FF).
reduce_by_polyset(X, _, X, _, _).

%
reduce_by_polyset_one(X, Ps, R, Ord, true, [P|T], T):- !,
	poly:rev_member(P, Ps),
%	member(P, Ps),
	once(reduce_by_poly_z(X, P, R, Ord)).
reduce_by_polyset_one(X, Ps, R, Ord, _, [P|T], T):-
	poly:rev_member(P, Ps),
%	member(P, Ps),
	once(reduce_by_poly(X, P, R, Ord)).

% reduce_by_polyset_one(X, Ps, R, Ord, FF, BL):- select(M, X, X0),
% 	poly:rev_member([N|Y], Ps),
% %	member([N|Y], Ps),
% 	poly:div_mono_mono(M, N, D),
% 	!,
% 	poly:mul_mono_poly(D, Y, Q),
% 	poly:subtract_poly_poly(X0, Q, R, Ord).

% ?- gb:reduce_by_poly([1*[x^1], 1*[a^1]], [1*[x^1], -1*[]], R, lexical).
%@ R = [1*[a^1], 1*[]] .

reduce_head_by_poly([M|X], [N|Y], R, Ord):-
	poly:div_mono_mono(M, N, D),
	!,
	poly:mul_mono_poly(D, Y, Q),
	poly:subtract_poly_poly(X, Q, R, Ord).

% ?- gb:reduce_by_poly_z([1*[x^1], 1*[a^1]], [1*[x^1], -1*[]], R, lexical).
%@ R = [1*[a^1], 1*[]] .
% ?- gb:reduce_by_poly_z([4*[x^1], 1*[a^1]], [6*[x^1], -1*[]], R, lexical).
%@ R = [3*[a^1], 2*[]] .

reduce_head_by_poly_z([A*M|X], [B*N|Y], R, Ord):-
	poly:div_mono_mono_term(M, N, N0),
	!,
	poly:lcm(A, B, L),
	A0  is L // A,
	B0  is L // B,
	poly:mul_scalar_poly(A0, X, Q),
	poly:mul_mono_poly(B0*N0, Y, Y0),
	poly:subtract_poly_poly(Q, Y0, R, Ord).

% ?- gb:reduce_by_poly([1*[x^1], 1*[a^1]], [1*[x^1], -1*[]], R, lexical).
%@ R = [1*[a^1], 1*[]] .
% ?- gb:reduce_by_poly([1*[x^1], 1*[a^1]], [1*[x^2], -1*[]], R, lexical).
%@ false.

reduce_by_poly(X, [N|Y], R, Ord):- select(M, X, X0),
	poly:div_mono_mono(M, N, D),
	!,
	poly:mul_mono_poly(D, Y, Q),
	poly:subtract_poly_poly(X0, Q, R, Ord).

% ?- gb:reduce_by_poly_z([1*[x^1], 1*[a^1]], [1*[x^1], -1*[]], R, lexical).
%@ R = [1*[a^1], 1*[]] .
% ?- gb:reduce_by_poly_z([4*[x^1], 1*[a^1]], [6*[x^1], -1*[]], R, lexical).
%@ R = [3*[a^1], 2*[]] .
% ?- gb:reduce_by_poly_z([4*[x^1], 1*[a^1]], [6*[x^2], -1*[]], R, lexical).
%@ false.

reduce_by_poly_z(X, [B*N|Y], R, Ord):- select(A*M, X, X0),
	poly:div_mono_mono_term(M, N, N0),
	!,
	poly:lcm(A, B, L),
	A0  is L // A,
	B0  is L // B,
	poly:mul_scalar_poly(A0, X0, Q),
	poly:mul_mono_poly(B0*N0, Y, Y0),
	poly:subtract_poly_poly(Q, Y0, R, Ord).

normal_poly(true, Prime, Poly, R):- !,
	(	Prime == 0
	->	once(poly:remove_content_poly(Poly, R))
	;	once(poly:galois_poly(Prime, Poly, R))
	).
normal_poly(_, _, Poly, R):- once(poly:canonical_form(Poly, R)).
