:- module(gvec, []).
:- use_module(gb(vector)).
term_expansion --> pac:expand_pac.
:- use_module(pac(op)).

% %
% init_base(G, true) :- !, new_vector(G, 6).
% init_base([], _).

%
add_to_base(Obj, G0, Index, G1, true):- new_index(Index, Obj, G0, G1).
add_to_base(R, G, R, [R|G], _).

%
spoly(I, J, Spoly, G, Ord, FF, Vec):-
	if(FF,	s_poly_z(I, J, Spoly, G, Ord, Vec),
		s_poly(I, J, Spoly, G, Ord, Vec)).

%
gb_BDFM([P|F], G, Opt):-
	subset([
			order(Ord),
			trace(Trace),
			assoc(Assoc),
			fraction_free(FF),
    		prime(Prime),
			vector(Vec)
		   ], Opt),
	if(Vec, SizePred = (vector:v_size), length),
	Debug =debug(Trace, Assoc, SizePred),
	memberchk(initial_vector_size(Size), Opt),
	new_vector(G0, Size),
%	init_base(G0, Vec),
	add_to_base(P, G0, _, G1, Vec),
	poly_pairs(F, [], D, G1, G2, Ord, Vec),
	gb_spoly(D, G2, G, Debug, Ord, FF, Prime, Vec).

%
poly_pairs([], D, D, G, G, _, _).
poly_pairs([P|F], D0, D, G0, G, Ord, Vec):-
	add_to_base(P, G0, M, G1, Vec),
	update_spoly_agenda(P, M, D0, D1, G1, Ord, Vec),
	poly_pairs(F, D1, D, G1, G, Ord, Vec).

%
gb_spoly([], G, G, _, _Ord, _, _, _).
gb_spoly([I-J|D], G0, G, Debug, Ord, FF, BL, Vec):-
	spoly(I, J, Spoly, G0, Ord, FF, Vec),
	once(reduce_head_by_polyset(Spoly, G0, S0, Ord, FF, _Trace, [])),
	once(gb:normal_poly(FF, BL, S0, R)),
%	gb:log(Debug, D, G0, R),
	(  R == []
	->	gb_spoly(D, G0, G, Debug, Ord, FF, BL, Vec)
	;	add_to_base(R, G0, M, G1, Vec),
		update_spoly_agenda(R, M, D, D0, G1, Ord, Vec),
		gb_spoly(D0, G1, G, Debug, Ord, FF, BL, Vec)
	).

s_poly(I, J, Spoly, G, Ord, true):- !, v_ref(I, G, P),
	v_ref(J, G, Q),
	poly:s_poly(P, Q, Spoly, Ord).
s_poly(I, J, Spoly, _, Ord, false):-
	poly:s_poly(I, J, Spoly, _,  Ord).

%
s_poly_z(I, J, Spoly, G, Ord, true):- !, v_ref(I, G, P),
	v_ref(J, G, Q),
	poly:s_poly_z(P, Q, Spoly, Ord).
s_poly_z(I, J, Spoly, _, Ord, _):-
	poly:s_poly_z(I, J, Spoly, Ord).

%
head_term(G, I, X):-  v_ref(I, G, [_*X|_]).

%
head_term(G, I, X, true):-  head_term(G, I, X).
head_term(_, I, I, _).
%

head_term_lcm(G, I, J, X):- head_term_lcm(G, I, J, X, true).

head_term_lcm(G, I, J, X, true):- !, head_term(G, I, Y),
	head_term(G, J, Z),
	poly:mono_lcm(Y, Z, X).
head_term_lcm(_, I, J, X, _):-  poly:mono_lcm(I, J, X).

% head_term_lcm(G, I, J, X, true):- !, head_term(G, I, Y),
% 	head_term(G, J, Z),
% 	poly:mono_lcm(Y, Z, X).
% head_term_lcm(_, I, J, X, _):-  poly:mono_lcm(I, J, X).

%%
update_spoly_agenda(R, M, D, D0, G, Ord, Vec):-
	if( (Vec, R=[_*H|_]),
		update_spoly_agenda(D, M-H, D0, G, Ord, Vec),
		update_spoly_agenda(D, M, D0, G, Ord, Vec)).

%
update_spoly_agenda(D, M, D0, G, Ord, Vec):-
	update_spoly_agenda_B(D, U, [], M, G, Ord, Vec),
	update_spoly_agenda_DFM(M, V, G, Ord, Vec),
	sort_spoly(V, V0, G),
	merge_spoly_agenda(U, V0, D0, G, Vec).

%
update_spoly_agenda_B([], D, D, _, _, _, _).
update_spoly_agenda_B([I-J|R], P, Q, M, G, Ord, Vec):-
	b_cond(M, I, J, G, Vec), !,
	update_spoly_agenda_B(R, P, Q, M, G, Ord, Vec).
update_spoly_agenda_B([A|R], [A|P], Q, M, G, Ord, Vec):-
	update_spoly_agenda_B(R, P, Q, M, G, Ord, Vec).

%
update_spoly_agenda_DFM(M, D, G, Ord, true):-  head_term(G, 1, T1),
	update_spoly_agenda_DFM_vec(1-T1, M, D, [], G, Ord).
update_spoly_agenda_DFM(M, V, G, Ord, false):-
	update_spoly_agenda_DFM_list(G, G, M, V, [], Ord).

%
update_spoly_agenda_DFM_vec(I-_, J-_, P, P, _, _):- I >= J, !.
update_spoly_agenda_DFM_vec(V, W, P, Q, G, Ord):-
	V = I-_,
	( d_cond(V, W, true);
	  f_cond(V, W, G, true);
	  m_cond(V, W, G, true) ),
	!,
	K is I + 1,
	head_term(G, K, Tk),
	update_spoly_agenda_DFM_vec(K-Tk, W, P, Q, G, Ord).
update_spoly_agenda_DFM_vec(I-_, W, [I-J|P], Q, G, Ord):- K is I+1,
	W = J-_,
	head_term(G, K, Tk),
	update_spoly_agenda_DFM_vec(K-Tk, W, P, Q, G, Ord).

%
update_spoly_agenda_DFM_list([I|Gi], G, J, P, Q, Ord):-
	( f_cond(I, J, Gi, false);
	  m_cond(I, J, G, false);
	  d_cond(I, J, false) ),
	!,
	update_spoly_agenda_DFM_list(Gi, G, J, P, Q, Ord).
update_spoly_agenda_DFM_list([I|Gi], G, J, [I-J|P], Q, Ord):- !,
	update_spoly_agenda_DFM_list(Gi, G, J, P, Q, Ord).
update_spoly_agenda_DFM_list(_, _,  _, P, P, _).

%
b_cond(K-Tk, I, J, G, true):- !, K > J,
	head_term(G, I, Ti),
	head_term(G, J, Tj),
	poly:mono_lcm(Ti, Tj, Tij),
	poly:div_mono_mono_term(Tij, Tk),
	poly:mono_lcm(Ti, Tk, Tik),
	Tik \== Tij,
	poly:mono_lcm(Tj, Tk, Tjk),
	Tjk \== Tij.
b_cond([_*Tk|_], [_*Ti|_], [_*Tj|_], _, false):-
	poly:mono_lcm(Ti, Tj, Tij),
	poly:div_mono_mono_term(Tij, Tk),
	poly:mono_lcm(Ti, Tk, Tik),
	Tik \== Tij,
	poly:mono_lcm(Tj, Tk, Tjk),
	Tjk \== Tij.

%
d_cond(_-Ti, _-Tj, true):- !,
	poly: merge_mono_mono(Ti, Tj, T),
        poly: mono_lcm(Ti, Tj, T).
d_cond([_*Ti|_], [_*Tj|_], false) :-
	poly: merge_mono_mono(Ti, Tj, T),
        poly: mono_lcm(Ti, Tj, T).

%
f_cond(I-Ti, _-Tj, G, true) :- !,
	poly:mono_lcm(Ti, Tj, Tij),
	I0 is I-1,
	between(1, I0, K),
	head_term(G, K, Tk),
	poly:mono_lcm(Tk, Tj, Tij).
f_cond([_*Ti|_], [_*Tj|_], G, false) :-
	poly:mono_lcm(Ti, Tj, T),
%	poly:rev_member([_*Tk|_], G),
	v_member([_*Tk|_], G),
	poly:mono_lcm(Tk, Tj, T).

m_cond(I-Ti, J-Tj, G, true):- !, M is J-1,
	poly:mono_lcm(Ti, Tj, Tij),
	between(1, M, K),
	K\== I,
%	call(G, K, Tk),
	v_ref(K, G, Tk),
	poly:div_mono_mono_term(Tij, Tk),
	poly:mono_lcm(Tj, Tk, Tjk),
	Tjk \== Tij.
m_cond(Pi, Pj, G, false) :-   Pi = [_*Ti|_],
	Pj = [_*Tj|_],
	poly:mono_lcm(Ti, Tj, Tij),
%	poly:rev_member(Pk, G),
	v_member(Pk, G),
	Pk \== Pi,
	Pk = [_*Tk|_],
	poly:div_mono_mono_term(Tij, Tk),
	poly:mono_lcm(Tj, Tk, Tjk),
	Tjk \== Tij.

%
merge_spoly_agenda([], X, X, _, _).
merge_spoly_agenda(X, [], X, _, _).
merge_spoly_agenda([P|R], [Q|S], [P|T], G, Vec):-
	if(Vec, compare_spoly(G, C, P, Q),
		compare_spoly(C, P, Q)),
	C = (<),
	!,
	merge_spoly_agenda(R, [Q|S], T, G, Vec).
merge_spoly_agenda(R, [Q|S], [Q|T], G, Vec):-
	merge_spoly_agenda(R, S, T, G, Vec).

%
compare_spoly(G, C, I-J, I0-J0):- head_term_lcm(G, I, J, LCM),
	head_term_lcm(G, I0, J0, LCM0),
	poly:compare_total_order(C0, LCM, LCM0),
	( C0 == (=) ->  C = (<);  C = C0).

% ?- bdfm:compare_spoly(C, [1*[a^2]]-[1*[b^2]], [1*[b^1]]-[1*[a^2]]).
%@ C =  (>).
compare_spoly(C, [_*M|_]-[_*N|_], [_*M0|_]-[_*N0|_]):-
	poly:mono_lcm(M,  N,  A),
	poly:mono_lcm(M0, N0, A0),
	poly:compare_total_order(C0, A, A0),
	(C0 == (=) -> C = (<); C = C0 ).
%
sort_spoly(X, Y, G):- predsort(compare_spoly(G), X, Y).
%
sort_spoly(X, Y):- predsort(compare_spoly, X, Y).

%
reduce_head_by_polyset(X, Ps, R, Ord, FF, T0, T):-
	reduce_head_by_polyset_one(X, Ps, Z, Ord, FF, T0, T1), !,
	reduce_head_by_polyset(Z, Ps, R, Ord, FF, T1, T).
reduce_head_by_polyset(X, _, X, _, _, T, T).

reduce_head_by_polyset_one(X, Ps, R, Ord, true, [P|T], T):- !,
	v_member(P, Ps),
	once(reduce_head_by_poly_z(X, P, R, Ord)).
reduce_head_by_polyset_one(X, Ps, R, Ord, _, [P|T0], T):-
%	poly:rev_member(P, Ps),
	v_member(P, Ps),
	once(reduce_head_by_poly(X, P, R, Ord, T0, T)).

reduce_by_polyset(X, Ps, R, Ord, FF):-
	reduce_by_polyset_one(X, Ps, Z, Ord, FF, _Trace, []), !,
	reduce_by_polyset(Z, Ps, R, Ord, FF).
reduce_by_polyset(X, _, X, _, _).

%
reduce_by_polyset_one(X, Ps, R, Ord, true, [P|T], T):- !,
%	poly:rev_member(P, Ps),
	v_member(P, Ps),
	once(reduce_by_poly_z(X, P, R, Ord)).
reduce_by_polyset_one(X, Ps, R, Ord, _, [P|T], T):-
%	poly:rev_member(P, Ps),
	v_member(P, Ps),
	once(reduce_by_poly(X, P, R, Ord)).


% reduce_by_polyset_one(X, Ps, R, Ord, FF, BL):- select(M, X, X0),
% 	poly:rev_member([N|Y], Ps),
% %	member([N|Y], Ps),
% 	poly:div_mono_mono(M, N, D),
% 	!,
% 	poly:mul_mono_poly(D, Y, Q),
% 	poly:subtract_poly_poly(X0, Q, R, Ord).


% ?- gb:reduce_by_poly([1*[x^1], 1*[a^1]], [1*[x^1], -1*[]], R, lexical).
%@ R = [1*[a^1], 1*[]] .

reduce_head_by_poly([M|X], [N|Y], R, Ord):-
	poly:div_mono_mono(M, N, D),
	!,
	poly:mul_mono_poly(D, Y, Q),
	poly:subtract_poly_poly(X, Q, R, Ord).

% ?- gb:reduce_by_poly_z([1*[x^1], 1*[a^1]], [1*[x^1], -1*[]], R, lexical).
%@ R = [1*[a^1], 1*[]] .
% ?- gb:reduce_by_poly_z([4*[x^1], 1*[a^1]], [6*[x^1], -1*[]], R, lexical).
%@ R = [3*[a^1], 2*[]] .

reduce_head_by_poly_z([A*M|X], [B*N|Y], R, Ord):-
	poly:div_mono_mono_term(M, N, N0),
	!,
	poly:lcm(A, B, L),
	A0  is L // A,
	B0  is L // B,
	poly:mul_scalar_poly(A0, X, Q),
	poly:mul_mono_poly(B0*N0, Y, Y0),
	poly:subtract_poly_poly(Q, Y0, R, Ord).

% ?- gb:reduce_by_poly([1*[x^1], 1*[a^1]], [1*[x^1], -1*[]], R, lexical).
%@ R = [1*[a^1], 1*[]] .
% ?- gb:reduce_by_poly([1*[x^1], 1*[a^1]], [1*[x^2], -1*[]], R, lexical).
%@ false.

reduce_by_poly(X, [N|Y], R, Ord):- select(M, X, X0),
	poly:div_mono_mono(M, N, D),
	!,
	poly:mul_mono_poly(D, Y, Q),
	poly:subtract_poly_poly(X0, Q, R, Ord).

% ?- gb:reduce_by_poly_z([1*[x^1], 1*[a^1]], [1*[x^1], -1*[]], R, lexical).
%@ R = [1*[a^1], 1*[]] .
% ?- gb:reduce_by_poly_z([4*[x^1], 1*[a^1]], [6*[x^1], -1*[]], R, lexical).
%@ R = [3*[a^1], 2*[]] .
% ?- gb:reduce_by_poly_z([4*[x^1], 1*[a^1]], [6*[x^2], -1*[]], R, lexical).
%@ false.

reduce_by_poly_z(X, [B*N|Y], R, Ord):- select(A*M, X, X0),
	poly:div_mono_mono_term(M, N, N0),
	!,
	poly:lcm(A, B, L),
	A0  is L // A,
	B0  is L // B,
	poly:mul_scalar_poly(A0, X0, Q),
	poly:mul_mono_poly(B0*N0, Y, Y0),
	poly:subtract_poly_poly(Q, Y0, R, Ord).

normal_poly(true, Prime, Poly, R):- !,
	(	Prime == 0
	->	once(poly:remove_content_poly(Poly, R))
	;	once(poly:galois_poly(Prime, Poly, R))
	).
normal_poly(_, _, Poly, R):- once(poly:canonical_form(Poly, R)).
