:- module(html, [quote/2, tag/3,tag_nl/3, html/4, html/5,
				 double_quote/2]).

:- use_module(util(misc)).
:- use_module(pac(basic)).
:- use_module(pac('expand-pac')).
:- use_module(pac(op)).
:- use_module(util(obj)).
:- use_module(util(misc)).
:- use_module(util('emacs-handler')).

term_expansion --> pac:expand_pac.

:- encoding(utf8).
:- set_prolog_flag(allow_variable_name_as_functor, true).

%%	buffer_html(+A:phrase, +X:any, -F:<file name>) is det.
%	Apply phrase A to X to get Y, and, by getting the buffer information,
%	write the result Y to the buffer file F, and open F.

buffer_html(A, X, F) :- phrase(A, X, Y),
	get_buffer_info_codes(L),
 	obj_get([buffer(N), dir(D)], L, _),
	map_path_html(D, N, F),
  	html_write_me(F, Y),
  	html_open(F).

%%	get_buffer_info_codes(-X:obj) is det.
%	Get the buffer info  in codes, and unify X with it.

% ?- html:get_buffer_info_codes(X).

get_buffer_info_codes([buffer(N), dir(D), whoami(U), hostname(H)]):-
	elisp:current_buffer_name(N),
 	elisp:default_directory(D),
 	getinfo_codes(whoami, U),
 	getinfo_codes(hostname, H).

%%	get_buffer_info(-X:obj) is det.
%	Get the buffer info, and unify X with it.

% ?- html:get_buffer_info(X).

get_buffer_info([buffer(N), dir(D), whoami(U), hostname(H)]):-
	elisp:cur_buf(N),
 	elisp:cur_dir(D),
 	getinfo(whoami, U),
 	getinfo(hostname, H).

%
html(F, Top) --> region, html_export(F, Top).

html_export(F, Top) --> paragraph,
	remove([]),
	phrase(Top),
	html_write_me(F),
	html_open(F),
	clear.

html(F, Top, Sub) --> region,
	paragraph,
	maplist(phrase(Sub)),
	act(Top),
	html_write_me(F),
	html_open(F),
	clear.

tex2html(env(N, Body),
	["<", N, ">" , Body1, "</", N, ">"],
	[Body],  [Body1], true).
tex2html(X, X, [], [], true).

% [2009/10/18]
% # question_button_list([list_type(ul), label(`ここを押せ`)]), browse, clear
% $a+b+c+d$の値は
% $a+b+c$と$d$の和である.
question_button_list(L) --> obj_init(L), question_button_list.

question_button_list -->
	obj_merge([label("==&gt;"),
		   list_type(ol),
		   mid_symbol_codes("==>"),
		   file_name(test('deldel.html'))
		  ]),
	obj_get([list_type(O)]),
	current(Env),
	obj_act((split,
		maplist(remove(0'\n)),   %'
		remove([]),
		sed(question_button_list_item(Env)),
		tag(O))).

question_button_list_item(E, QA) --> [Line],
	{
	 obj_get([label(L),  mid_symbol_codes(M)], E),
	 append([X, M, Y], Line),
	 mimetex_question_button([question(X), button(Y), label(L)], Z),
	 obj_get([acc(Z0)], Z),
	 tag(li, Z0, QA)
	}.

prolog_answer_button(E, ["\n", QA]) --> [Line],
	{ text_parse_solve(Line, Y),
	  quote(Y, Y0),
	  obj_get([label(Label)], E),
	  quote(Label, Label0),
	  LI = ["?- ", Line, ".  ",
		     "<script>ans(",
		      Y0,
		      "," ,
		      Label0,
		      ");</script>"],
	 obj_put([acc(LI)], E, _E),
	 tag(li, LI, QA)
	}.

% all solved forms html.
text_parse_solve(X, Y):-
	( catch(herbrand(V, X, G), _, fail) ->  findall(V, call(G), Y0)
	; Y0= "Syntax Error"
	),
	solved_form_html(Y0, Y).

solved_form_html([], "false"):-!.
solved_form_html([[]|_], "true"):-!.
solved_form_html(X, [X]):- atom(X), !.
solved_form_html(X, Y):-  % do not add `\n` after <br>
        maplist(maplist(herbrand_opp), X, S1),
	maplist(insert("<br>"), S1, S2),
	insert(" ;<br><br>", S2, S3),
	flatten(["<br>", S3], Y).

% ?- html:parse_mimetex((`$a$`, [mimetex_name(fff)]), (Y, _)),  smash(Y).
%@ <script>fff(`a`);</script>

parse_mimetex((X,E),(Y,E)):- parse_mimetex(E, X, Y).

% ?- html:parse_mimetex_qa([], `$a$ ==> $b$`, R), smash(R).
% ?- html:parse_mimetex_qa([], `$a$ ==> $\\b$`, R), basic:smash(R).

parse_mimetex(E, X, Y) :- obj_put([acc(X)], E, E0),
	mimetex(E0, E1),
	obj_get([acc(Y)], E1).

% ?- html:at_or_qa([], `$a$ ==> $b$`, R), smash(R).
%@ <li>$a$<script>ans("$b$","==&gt;");</script></li>

at_or_qa(_, [0'@ |X], [X, "\n"]):- !.	%'
at_or_qa(E, X, Y):-   parse_mimetex_qa(E, X, Y0),
	tag_nl(li, Y0, Y).


parse_mimetex_qa(E, X, Y) :- once(parse_qa_line(Q, A, X, [])),
	obj_put([question(Q), button(A)], E, E0),
	mimetex_question_button(E0, E1),
	!,
	obj_get([acc(Y)], E1).

% ?- html:parse_qa_line(Q, A, `  a  ==>  b  `, []).
parse_qa_line(Q, A) --> wl("[ \t]*"),
	w(".*", Q),
	wl("[ \t]*==>[ \t]*"),
	w(".*", A),
	w("[ \t]*$").

%
boole_qa(E, [Q, A], Y) :-
	atom_codes(A, A0),
	atomic_list_concat([$, Q, $], Q0),
	atom_codes(Q0, Q1),
 	obj_put([question(Q1), button(A0)], E, E0),
	mimetex_question_button(E0, E1),
	!,
	obj_get([acc(Y)], E1).

%
dummy_qa(_, [_, _], "Hello world").


% ?- html:mimetex_question_button([question(`a$\\pi$`), button(`x$\\sigma$`)], X), obj_get([acc(V)], X), smash(V).

mimetex_question_button -->
	obj_get([question(X)]),
	obj_put([acc(X)]),
	mimetex_question,
	obj_get([acc(X0)]),
	obj_get([button(Y)]),
	obj_put([acc(Y)]),
	mimetex_button,
	obj_get([acc(Y0)]),
	obj_put([acc([X0, Y0])]).

% ?- html:mimetex([acc(`a $b+c$ `)], X), smash(X).
mimetex_question(X, X).

%
mimetex --> obj_merge([mimetex_name(mimetex), insert([])]),
	obj_get([mimetex_name(N), insert(Insert)]),
	obj_act(html:mimetex(N, Insert)).

%
mimetex(N, Insert) --> mime_parse_me,
	flatten,
	maplist(mimetex_(N, Insert)).

%
mimetex_(N, Insert)--> peek(dol(X), [Insert, X]), !,
	tex_quote,
	peek(Y, ["<script>", N, "(", Y, ");</script>"]).
mimetex_(N, Insert)--> peek(ddol(X), [Insert, X]), !,
	tex_quote,
	peek(Y, ["<center><script>", N, "(", Y, "); </script></center>"]).
mimetex_(_, _) --> [].

%
tex_quote --> tex_codes, flatten, quote.

% ?- html:mimetex_button([acc(`a`)], R), obj_get([acc(V)], R), smash(V).
mimetex_button --> obj_merge([label("==&gt;"),
      			      parse(html:parse_mimetex_quote),
			      ans(ans)]),
	obj_get([label(Label),  ans(A),  parse(P)]),
	obj_act(P),
	obj_get([acc(X)]),
	{quote(X, X0)},
	{quote(Label, Label0)},
	obj_put([acc(["<script>", A, "(", X0, "," , Label0,  ");</script>"])]).

% ?- html:parse_mimetex_quote(f, `$a$ and b`, X), smash(X).
% ?- html:parse_mimetex_quote(f, `$a+\\emptyset$`, X), smash(X).
parse_mimetex_quote --> parse_mimetex_quote(mimetexcode).

%
parse_mimetex_quote(_,X,X).

%
join_neighbor_codes([], []).
join_neighbor_codes([X|Y], R):- join_neighbor_codes(Y, R0),
	join_neighbor_codes(X, R0, R).

join_neighbor_codes(dol(X), Y, [dol(X)|Y]):- !.
join_neighbor_codes(ddol(X), Y, [ddol(X)|Y]):- !.
join_neighbor_codes(X, [], [[X]]):- !.
join_neighbor_codes(X, [L|R], [[X|L]|R]):- listp(L), !.
join_neighbor_codes(X, Y, [[X]|Y]).

% [2010/11/03]
% ?- html:mimetex_onclick(`a$\\sigma$b`, X), smash(X).
% ?- html:mimetex_onclick(`a$\\sigma$b`, X), quote(X, Y), smash(Y).
mimetex_onclick --> mime_js, js_plus, quote.

mime_js --> texparse, flatten, split_mime, maplist(js_func_call).

%
split_mime([], [[]]):- !.
split_mime([X|R], [[X|Y]|M]):-  integer(X), !, split_mime(R, [Y|M]).
split_mime([X|R], [[],X|S]):- split_mime(R, S).

%
js_func_call(dol(X), dol(Y)):- !, tex_js(X, Y).
js_func_call(ddol(X), ddol(Y)):- !, tex_js(X, Y).
js_func_call(X, just(Y)):- flatten(X, X0), quote(X0, Y).

% tex_js --> tex_codes, flatten, quote.

% js_plus(X, Y):- maplist(smash_functor, X, X0),
% 	insert("+", X0, X1),
% 	flatten(X1, Y).

% smash_functor(X, [F,"(", A, ")"]):- X=..[F|A].

% % ?- html:quote_math(f, dol(`abc`), X), smash(X).
% %@ f("abc")
% quote_math(N, dol(X), Y):- !, tex_codes(X, X0), flatten(X0, X1),
% 	quote(X1, X2),
% 	Y = [N, "(", X2, ")"].
% quote_math(N, ddol(X), Y):- !, tex_codes(X, X0), flatten(X0, X1),
% 	quote(X1, X2),
% 	Y=["\"<center>\"+", N, "(", X2, ")+\"</center>\""].
% quote_math(_, X, Y):- flatten(X, X0), quote(X0, Y).

% %%%% Simple quoting
% % ?- quote(`\\\`abc`, X), smash(X).
% % ?- html:single_quote(`\\\`abc`, X), smash(X).

% quote --> double_quote.

% % ?- html:double_quote(`abc`, X), smash(X).
% % ?- html:double_quote(`abc`, X), atom_codes(X0, X).
% double_quote(X, [0'\"|X0]):- escape(0'\", X, X0, [0'\"]).

% single_quote(X, [0'\'|X0]):- escape(0'\', X, X0, [0'\']).

% % ?- html:back_quote(`a\`bc`, R), smash(R).
% %@ `a\`bc`
% back_quote(X, [0'\`|X0]):- escape(0'\`, X, X0, [0'\`]).

% % ?- html:escape(0'", `"ab"`, X, []), smash(X).
% %@ \"ab\"
% %@ X = [92, 34, 97, 98, 92, 34].

% escape(X, Y) --> escape_quasi_string(Y, X), !.


% escape_quasi_string([], _)	--> [].
% escape_quasi_string([A|B], M)	-->
% 	escape_quasi_string(A, M),
% 	escape_quasi_string(B, M).
% escape_quasi_string(S,    M) --> { string(S), string_codes(S, S0)}, !,
% 	escape_quasi_string(S0, M).
% escape_quasi_string(S,    M) --> { atom(S), atom_codes(S, S0)}, !,
% 	escape_quasi_string(S0, M).
% escape_quasi_string(M,    M, [0'\\ , M    | X],	X).
% escape_quasi_string(0'\\, _, [0'\\ , 0'\\ | X],	X).
% escape_quasi_string(0'\n, _, [0'\\ , 0'n  | X],	X).
% escape_quasi_string(0'\t, _, [0'\\ , 0't  | X],	X).
% escape_quasi_string(C,    _, [C           | X], X).



			/****************************
			*	simple escape       *
			****************************/

% ?- html:escape(`abc`, X, []), atom_codes(X0, X).
% ?- html:escape(``, X, []), atom_codes(X0, X).
% ?- html:escape(`\n`, X, []), atom_codes(X0, X).
% ?- html:escape(`\\\n`, X, []), atom_codes(X0, X), writeln(X0).
% ?- html:escape(`\\\s`, X, []), atom_codes(X0, X), writeln(X0).

% escape([], X, X) :- !.
% escape([0'\\ |R], [0'\\ , 0'\\ |X], Y):- !, escape(R, X, Y).
% escape([[A|B]|R])	--> !, escape(A), escape(B), escape(R).
% escape([C|R], [C|X], Y) :- escape(R, X, Y).

% ?- html:mime_parse_me(`$a$+$b$`, R).
% R = [dol([97]),43,dol([98])].

mime_parse_me(X,Y):- mime_parse_me(Y,X,[]), !.

mime_parse_me([]) --> [].
mime_parse_me([Y|Z]) --> tex:tex_math(X,[]), {tex_env(X,[Y])}, mime_parse_me(Z).
mime_parse_me([A|X]) --> [A], mime_parse_me(X).

		/************************
		*     open HTML page    *
		************************/

oh --> region, {open_current_html}, clear.

open_current_html :-
	get_html_buffer_info(L),
	open_current_html(L).

open_current_html(L) :-
	obj_get([buffer(N), path_html(P), whoami(U), hostname(H)], L, _),
	sh(open("http://"+ H+ "/~"+ U+ "/"+ P+ N)).

get_html_buffer_info(L):- get_buffer_info_codes(L0),
	obj_get([dir(D)], L0, _),
	once(local_apache_path(D, P)),
	obj_push([path_html(P)], L0, L).
%
local_apache_path -->
	{	getenv(home, H),
		atomics_to_string([H, public_html], /, P),
		atom_codes(P, PCodes)
	},
	residue(PCodes),
	!.
local_apache_path -->
    {	elisp:message(
		[	"fDCG: Current directory",
			"is not under 'Sites' directory."
		])
    }.

%
tex_js --> tex_codes, flatten, quote.

%
js_plus(X, Y):- maplist(smash_functor, X, X0),
	insert("+", X0, X1),
	flatten(X1, Y).

smash_functor(X, [F,"(", A, ")"]):- X=..[F|A].

% ?- html:quote_math(f, dol(`abc`), X), smash(X).

quote_math(N, dol(X), Y):- !, tex_codes(X, X0), flatten(X0, X1),
	quote(X1, X2),
	Y = [N, "(", X2, ")"].
quote_math(N, ddol(X), Y):- !, tex_codes(X, X0), flatten(X0, X1),
	quote(X1, X2),
	Y=["\"<center>\"+", N, "(", X2, ")+\"</center>\""].
quote_math(_, X, Y):- flatten(X, X0), quote(X0, Y).

%%%% Simple quoting
% ?- quote(`\\\`abc`, X), smash(X).
%@ "\\`abc"
%@ X = [34, 92, 92, 96, 97, 98, 99, 34].
% ?- quote(`\\\`abc`, X), smash(X).
%@ `\\\`abc`
%@ X = [34,92,92,92,34,97,98,99,34].
% ?- html:single_quote(`\\\`abc`, X), smash(X).
%@ '\\`abc'
%@ X = [39,92,92,34,97,98,99,39].

quote --> double_quote.

% ?- html:double_quote(`abc`, X), smash(X).
% ?- html:double_quote(`abc`, X), atom_codes(X0, X).

double_quote(X, [0'\"|X0]):- escape(0'\", X, X0, [0'\"]). %'

single_quote(X, [0'\'|X0]):- escape(0'\', X, X0, [0'\']).

% ?- html:back_quote(`a\`bc`, R), smash(R). %'
%@ `a\`bc`
back_quote(X, [0'\`|X0]):- escape(0'\`, X, X0, [0'\`]).  %'

% ?- html:escape(0'", `"ab"`, X, []), smash(X).
%@ \"ab\"

escape(X, Y) --> escape_quasi_string(Y, X), !.

%
escape_quasi_string([A|B], M)	-->
	escape_quasi_string(A, M),
	escape_quasi_string(B, M).
escape_quasi_string([], _)	--> [].
escape_quasi_string(S,    M) --> { string(S), string_codes(S, S0)},
	escape_quasi_string(S0, M).
escape_quasi_string(S,    M) --> { atom(S), atom_codes(S, S0)},
	escape_quasi_string(S0, M).
escape_quasi_string(M,    M, [0'\\ , M    | X],	X).		%'
escape_quasi_string(0'\\, _, [0'\\ , 0'\\ | X],	X).		%'
escape_quasi_string(0'\n, _, [0'\\ , 0'n  | X],	X).		%'
escape_quasi_string(0'\t, _, [0'\\ , 0't  | X],	X).		%'
escape_quasi_string(0'\s, _, [0'\\ , 0'\s | X],X).		%'
escape_quasi_string(C,    _, [C           | X], X).		%'


			/****************************
			*	simple escape       *
			****************************/

% ?- html:escape(`abc`, X, []), atom_codes(X0, X).
% ?- html:escape(``, X, []), atom_codes(X0, X).
% ?- html:escape(`\n`, X, []), atom_codes(X0, X).
% ?- html:escape(`\\\n`, X, []), atom_codes(X0, X), writeln(X0).
% ?- html:escape(`\\\s`, X, []), atom_codes(X0, X), writeln(X0).

escape([], X, X) :- !.
escape([0'\\ |R], [0'\\ , 0'\\ |X], Y):- !, escape(R, X, Y).
escape([[A|B]|R])	--> !, escape(A), escape(B), escape(R).  %'
escape([C|R], [C|X], Y) :- escape(R, X, Y).


%?- html:drop_angle_suffix(`abc`, X).
%?- html:drop_angle_suffix(`abc<2>`, X).

drop_angle_suffix([], []):-!.
drop_angle_suffix([0'<|_], []):-!.  % ' drop <...>
drop_angle_suffix([X|Y], [X|Z]):- drop_angle_suffix(Y,Z).

%
html_list(F) --> region, paragraph, remove([]), maplist(F), overwrite.

		/************************************************
		*     create query & answers table html page    *
		************************************************/

html_qa(M) --> query_ans_html(M).
html_qa(M, Tag) --> query_ans_html(M, Tag).

mimetex_qa(Label) --> drill(Label).

%
query_ans_html(S) -->
	{	getenv(home_html_root, H0),
		concat_atom([H0, '/sample'], Dir0),
		getenv(host_html_root, H1),
		concat_atom([H1, '/sample'], Dir1)
	},
	query_ans_html(S, ol),
	current(SOL),
	obj_put([directory(Dir0),
		 counter_name(sample_counter)]),
	phrase((eh:counter(update), obj_get([count(C)]))),
	initLoad_must_exist,
	{ atomics_to_string([Dir0, /, sample, C, '.html'], F0),
	  html_write_me(F0, SOL),
	  atomics_to_string([Dir1, /, sample, C, '.html'], F1)
	},
	peek([	"<a href=\"", F1, "\">",
			"Click to show the generated page</a>"]),
	term_smash0.

%
query_ans_html(M, Tag) -->  query_ans_html_body(M), tag(Tag).

%
query_ans_html(File, M, Tag)--> region,
				query_ans_html_export(File, M, Tag).

%
query_ans_html_export(File, M, Tag) -->
	query_ans_html(M, Tag),
	html_write_me(File),
	html_open(File),
	clear.

initLoad_must_exist --> obj_get([directory(D)]),
	{ working_directory(Old, D),
		(	exists_file('initLoad.js') -> true
		; 	getenv(exercise_js_name, P),
			getenv(home_cgi_bin, CB),
			atomics_to_string([CB, P], /, P0),
			sh(ln(-fs, P0, 'initLoad.js'))
    	  ),
    	  working_directory(_, Old)
	}.


%;; (setq module-query  "qcompile(util(html)), module(html).")
% ?- qcompile(util(html)), module(html).

% ?- html:query_ans_html(pac:eval, ol, `(:id)@(3)`, X), basic:smash(X).


query_ans_html_body(M) --> split, remove([]),
		maplist(html_li(M)).

html_li(M, X, Y):- catch(html_li_(M, X, Y),  _,
			 (Y = [" <li> ", X, " ==> Error !  </li>"])).

%
html_li_(M) --> current(X), parse_run(M), html_li(X, ans).

html_li(X, J) --> { atom_codes(J, J0) },
	smash_codes,
	qstring,
    peek(Y, ["<li>", X, "<script>", J0, "(", Y, ");</script></li>\n"]).

% ?- html:parse_run(pac:eval, `:(=(a))`, R).
% ?- html:parse_run(pac:eval, `(X\\X)@ quote(a)`, R).

% ?- herbrand_in_context(`:(=(a))`, H), nopac(pac:eval(H, R)).

% (setq module-query  "qcompile(util(html)), module(html).")
% ?- qcompile(util(html)), module(html).
% ?- parse_run(pac:eval, `:(=(a))`, R), smash(R).


% ?- qcompile(util(html)), module(html).
% ?- eval_string(":(=(a))").
% ?- eval_string("append@[a,b]@[c,d]").
% ?- eval_string(":append@[a,b]@[c,d]").
% ?- eval_string("pred([X,Y,Z]:-append(Y,X,Z))@[1,2,3]@[a,b,c]").
% ?- eval_string(":pred([X,Y,Z]:-append(Y,X,Z))@[1,2,3]@[a,b,c]").
% ?- eval_string("(#([X,Y]\\ append(Y,X)))@[1,2,3]@[a,b,c]").
% ?- eval_string(":fun([X,Y] >> append(Y,X))@[1,2,3]@[a,b,c]").
% ?- eval_string("append @ append([1,2],[3,4])@ append([a,b], [c,d])").
% ?- eval_string(":append @ :append([1,2],[3,4])@ :append([a,b], [c,d])").
% ?- eval_string("pred([X, X])@5").
% ?- eval_string(":pred([X, X])@5").
% ?- eval_string("maplist @ fun([X]>>(append(X)@ append([a],[b]))) @ [[1,2],[3,4]]").
% ?- eval_string(":maplist @ fun([X]>>(:append(X)@ :append([a],[b]))) @ [[1,2],[3,4]]").
%@ [[1,2,a,b],[3,4,a,b]]

eval_string(X):- string_codes(X, S),
	parse_run(pac:eval, S, U),
	smash(U).

parse_run(M) --> herbrand_in_context,
	call(M),
	herbrand_opp.

html_result(value(X), Y) :- !, numbervars(X, 0, _),
	with_output_to(codes(Y), print(X)).
html_result(L, Y) :- solutions_codes(L, Y).

%
solutions_codes([], "false"):-!.
solutions_codes([[]|_], "true"):-!.
solutions_codes --> maplist(herbrand_opp), insert("<br>\n").

%
html_table  --> region, matrix:html_table.

html_table_th  --> region, matrix:html_table_th.

html_table_var --> region, matrix:html_table_.

%%%%  handling html files

%%	browse(+P:obj, -Q:obj) is det.
%
%	The object P is supposed to contain [file_name(FN), acc(X)]
%	then write the html data X to the file FN, and open it with
%	the local web server. Q extends P so as to contain
%	[anchor_tag_codes(A)], where A is the codes of html  anchor-tag
%	to the written html file for X.

browse --> obj_get([file_name(FN), acc(X)]),
	{ expand_file_search_path(html(FN), Posix) },
	obj_act(file_name, anchor_tag_codes, html:anchor_tag),
	{ html_write_me(Posix, X), html_open(FN) }.

%	html_write_me(+F, +X) is det.
%	write a html text data given in X to the file F.
html_write_me(F, X):-
	A = [ file(other('html-begin-end'/htmlbegin0)),
	      file(other('html-begin-end'/htmlbegin1)),
	      file(other('html-begin-end'/htmlbegin2)),
	      text(X),
	      file(other('html-begin-end'/htmlend))  ],
	 assemble(A, F).

% ?- listing(html_write_me).
%@ html:html_write_me(C, A) :-
%@ 	B=[file(pac_home('html-begin-end'/htmlbegin0)), file(pac_home('html-begin-end'/htmlbegin1)), file(pac_home('html-begin-end'/htmlbegin2)), text(A), file(pac_home('html-begin-end'/htmlend))],
%@ 	assemble(B, C).
%@
%@ html:html_write_me(B, A, A) :-
%@ 	html_write_me(B, A).
%@
%@ true.


%
html_write_me(F, X, X):- html_write_me(F, X).

%	html_open(+F) is dete.
%	Open the html file F on a local web server
%	with a default browser.

html_open(F):-  atom_codes(F, F0),
	 local_apache_path(F0, P),
	 getinfo_codes(hostname, H),
	 getinfo_codes(whoami, U),
	 sh(open("http://" + H + "/~"+ U + "/"+ P)).

%	base_of_path(+X:string, -Y:string) is det.
% 	Y is the base of the posix path X.

base_of_path(X, Y) :-  split_string(X, "/", "", L),
		       last(L, Y).


%
anchor_tag --> base_of_path,  peek(X, ["<a href=\"", X, "\"> </a>"]).

% ?- qa(f, `abc\n\nxyz`, X).

qa(X) --> region, paragraph, sed(qa(X)), flatten, overwrite.

%
qa(N, ["<li>", X, "<script>", N1, "(", Y1, "); </script></li>\n"]) --> [X, Y],
    {atom_codes(N, N1), webstring(Y,Y1)}.

button(A, T, ["<script>", A0, "(", T0, ");</script>\n"]):-
	atom_codes(A, A0),
	webstring(T, T0).

wqa(N) --> w(".*", X), wl("\n\n(\n)*"),
	peek(Y, [X,Y]),
	sed(wqa(N)),
	flatten.

wqa(N, ["<li> ",  X,
		"<script>",
			N1, "(`<pre>", Y1, "</pre>`); ",
		"</script>", "
	</li>\n"])	-->  [X, Y],
			     {	atom_codes(N,N1),
				webencode(Y, Y1)}.

webencode --> sed(js_char),	% sed/3 is a pac macro (see pac-etc.pl)
	flatten,						% So the prefix html is necessary:
	sed(js_char2),				% meta_predicate for sed does not help.
	flatten.

erase(L) --> dsed(erase(L)).
erase(L,X,X)--> phrase(L).

%
webstring --> qstring,
	flatten,
	sed(js_char2).

% ?- html:qstring(`abc`, R).

qstring --> sed(js_char),
	peek(A,  ["\"", A, "\""]),
	flatten.

%
esc --> sed(js_char), flatten.

atomic_list_prolog_lisp(L, P):-
	maplist(atom_codes, L, C0),
	maplist(qstring, C0, C1),
	insert(` `, C1, P).

% copy a buffer into the mirror directory.
map_mirror(['Cabinet', lecture, class2009b] - [public_html]).
map_mirror(['infomath2009', _] - ['2009-infomath']).
map_mirror(['infomath2009'] - ['2009-infomath']).
map_mirror(['semi-proof'] - ['2009-seminar-proof']).

%%	copy_mirro_pdf(+M:
copy_mirror_pdf(M) :- get_buffer_info(Buffer_info),
	obj_get([buffer(F), dir(D)], Buffer_info, _),
	atom_concat(D, '_region_.pdf', P),
	map_path_pdf(M, D, F, Q),
	sh(cp(P, Q)).

%
map_path_pdf(M, D, F, P):- file_path_map(M, '.tex', '.pdf', D, F, P).

% ?- html:file_path_map(map_mirror, '.tex', '.pdf', '/x/a/y/', 'cd', X).
% X = '/x/b/y/cd.pdf'.

file_path_map(M, Ext, Ext0, D, F, P):- file_path_map(M, D, X),
	change_extension(Ext, Ext0, F, F0),
	atomic_list_concat(['', X, F0], /, P).

file_path_map(M, D, P):- atomic_list_concat(D0, / , D),
	remove('', D0, X0),
	path_map(M, X0, X1),
	atomic_list_concat(X1, /,  P).

%  rewriting path in a context-free way (i.e. ad hoc)
path_map(M, X, Y):- path_map_one(M, X, Z), !, path_map(M, Z, Y).
path_map(_, X, X).

%
path_map_one(M, X, Y):- call(M, A-B), append(A, C, X), append(B, C, Y).
path_map_one(M, [X|Y], [X|Z]):- path_map_one(M, Y, Z).

% ?- change_extension('.tex', '.pdf', 'abc.tex', X).
% X = 'abc.pdf'.

change_extension(E, E0, X, Y):- atom_concat(H, E, X), !, atom_concat(H, E0, Y).
change_extension(_, E0, X, Y):- atom_concat(X, E0, Y).

%%	copy_mirror_buffer(-M) is det.
%
% copy  the current buffer to a file
% the path of which is computed from the buffer info,
% and a mapping table to get the target path.

copy_mirror_buffer(M) :- get_buffer_info(Buffer_info),
	obj_get([buffer(F), dir(D)], Buffer_info, _),
	atom_concat(D, F, P),
	file_path_map(M, D, Q0),
	atomic_list_concat(['', Q0, F], /, Q),
	sh(cp(P, Q)).

%
set_target(X):- atom_codes(T, X), nb_setval(target_file, T).
get_target(X):- nb_getval(target_file, X).

%	Wrap with a tag mark
tag(Tag) --> peek(Body, ["<", Tag, ">", Body, "</", Tag, ">"]).

tag_nl(Tag) --> peek(Body, ["<", Tag, ">", Body, "</", Tag, ">\n"]).

tag_front(Tag) -->peek(Body, ["<", Tag, ">", Body]).

tag_back(Tag) --> peek(Body, [Body, "<", Tag, ">"]).


%%	set_tex(+X, -Y) is det.
%	Convert a term X consisting of set operations
%	into a  Latex text Y.


set_tex([],cs(emptyset)):-! .
set_tex(X,A1):-listp(X),!,(maplist(set_tex,X,A2),insert(',',A2,A3)),'pac#435'(A3,A1) .
set_tex(X,A1):-integer(X),!,number_codes(X,A1) .
set_tex(p(X,Y),['(',A1,',',A2,')']):-!,set_tex(X,A1),set_tex(Y,A2) .
set_tex(X+Y,[A1,cs(cup),A2]):-!,set_tex(X,A1),set_tex(Y,A2) .
set_tex(X&Y,[A1,cs(cap),A2]):-!,set_tex(X,A1),set_tex(Y,A2) .
set_tex((X\Y),[A1,cs(setminus),A2]):-!,set_tex(X,A1),set_tex(Y,A2) .
set_tex(X*Y,[A1,cs(times),A2]):-!,set_tex(X,A1),set_tex(Y,A2) .
set_tex(pow(X),[cs(mathrm),group(pow),'(',A1,')']):-!,set_tex(X,A1) .

'pac#435'(A,[cs('{'),A,cs('}')]):-true .

%%	subset_tex(+X, -Y) is det.
%	Convert a tex term X consisting of subset relations into
%	a latex codes Y.


subset_tex(X=Y,[A1,=,A2]):-!,subset_tex(X,A1),subset_tex(Y,A2) .
subset_tex(X=<Y,[A1,cs(subseteq),A2]):-!,subset_tex(X,A1),subset_tex(Y,A2) .
subset_tex(X>=Y,[A1,cs(supseteq),A2]):-!,subset_tex(X,A1),subset_tex(Y,A2) .
subset_tex(X<Y,[A1,cs(subset),A2]):-!,subset_tex(X,A1),subset_tex(Y,A2) .
subset_tex(X>Y,[A1,cs(supset),A2]):-!,subset_tex(X,A1),subset_tex(Y,A2) .
subset_tex(true,group([cs(mathtt),' true'])):-! .
subset_tex(false,group([cs(mathtt),' false'])):-! .
subset_tex(in(X,Y),[A1,cs(in),A2]):-!,subset_tex(X,A1),subset_tex(Y,A2) .

%
% js code conversion

js_char("\\\\") --> "\\".
js_char("\\`") --> "`".
js_char("\\n")	--> "\n".
js_char("\\/")	--> "/".

js_char2("&amp;amp;")	--> "&".
js_char2("&amp;prime;") --> "'".
js_char2("&amp;lt;")	--> "<".
js_char2("&amp;gt;")	--> ">".

tex_codes(cs(N),["\\",N]):-! .
tex_codes(comment(X),["%",X,"\n"]):-! .
tex_codes(env(N,X),["\\begin{",A1,"}",A2,"\\end{",A1,"}"]):-!,tex_codes(N,A1),tex_codes(X,A2) .
tex_codes(group(G),A1):-!,tex_codes(@@(["{",G,"}"],:maplist@(::)),A1) .
tex_codes(math(X),A1):-!,tex_codes(@@(["$",X,"$"],:maplist@(::)),A1) .
tex_codes(dmath(X),A1):-!,tex_codes(@@(["$$",X,"$$"],:maplist@(::)),A1) .
tex_codes([A1|A2],A3):-!,maplist(tex_codes,[A1|A2],A3) .
tex_codes(X,X):-atomic(X),! .
