:- module(int_part, []).

:- use_module(library(clpfd)).
:- use_module(util(misc)).
:- use_module(pac(basic)).
% :- expects_dialect(pac).
term_expansion --> pac:expand_pac.
:- use_module(pac(op)).



% ?- prolog_stack_property(X, Y).
% ?- set_prolog_stack(global, limit(2*10**9)).

% ?- fold_poly_zip_bounded([], [], 1, [], X).
% ?- fold_poly_zip_bounded([poly(1,[1,1])], [poly(1,[1,1])], 3, [], X).
% ?- fold_poly_zip_bounded([poly(1,[1,1])], [poly(1,[1,1])], 5, [], X).
% ?- fold_poly_zip_bounded([poly(1,[1,1])], [poly(1,[1,1])], 6, [], X).

fold_poly_zip_bounded(Ps, Qs, N, X, Y):-
	fold_poly_zip(pred(N, ([K, K, P, Q, R]
							:- math:poly_mul_bd_poly(P, Q, N, R))),
				  Ps, Qs, 0, X, Y).

% ?- scm(int_part).
partition_number(N, P, Opt):-
	(	memberchk(formula, Opt) -> partition_number(N, P)
	;	memberchk(rec,  Opt)	-> integer_partition(N, Ps),
					length(Ps, P)
	;  	memberchk(history, Opt) -> course_of_partition(N, [A|_]),
					length(A, P)
  	;  	memberchk(history2, Opt) -> partition_history_tr(N, [A|_]),
					length(A, P)
	;	memberchk(clpfd, Opt)	-> setof(Vs, partition_clpfd(N, Vs), All),
					length(All, P)
	;	partition_number(N, P)
	).

% ?- between(1, 52, N), partition_number(N, X, [history]), writeln(#('Y'(N)) = X), fail; true.

% ?- partition_number(500, P, [formula]).
% ?- partition_number(500, P, []).
% ?- partition_number(40, P, []).
% ?- partition_number(40, P, [formula]).
% ?- partition_number(40, P, [rec]).
% ?- partition_number(40, P, [history]).
% ?- partition_number(40, P, [history2]).
% ?- partition_number(20, P, [history]).
% ?- partition_number(20, P, [clpfd]).
% ?- time(partition_number(52, X, [formula])).
% ?- time(partition_number(52, X, [history])).
%@ % 47,461,267 inferences, 5.356 CPU in 5.489 seconds (98% CPU, 8861529 Lips)
%@ X = 281589 .
% ?- time(partition_number(52, X, [history2])).
%@ % 47,461,222 inferences, 5.722 CPU in 5.833 seconds (98% CPU, 8295086 Lips)
%@ X = 281589 .
% ?- time(partition_number(53, X, [history])).
%@ % 56,645,575 inferences, 6.636 CPU in 6.749 seconds (98% CPU, 8536582 Lips)
%@ X = 329931 .
% ?- time(partition_number(53, X, [history2])).
%@ % 56,645,538 inferences, 6.478 CPU in 6.598 seconds (98% CPU, 8743664 Lips)
% ?- time(partition_number(54, X, [history2])).
%@ % 67,495,266 inferences, 7.389 CPU in 7.528 seconds (98% CPU, 9135086 Lips)
%@ X = 386155 .
% ?- time(partition_number(55, X, [history2])).
%@ % 80,293,565 inferences, 9.192 CPU in 9.345 seconds (98% CPU, 8734916 Lips)
%@ X = 451276 .
% ?- time(partition_number(60, X, [history2])).
%@ % 187,006,232 inferences, 20.617 CPU in 20.900 seconds (99% CPU, 9070386 Lips)
%@ X = 966467 .


% ?- time(partition_number(52, X, [rec])). % Out of global stack
% ?- time(partition_number(53, X, [])).
% ?- time(partition_number(53, X, [history])).  % max
% ?- time(partition_number(53, X, [history2])).  % Out of Global Stack

% Conclusion: induction with `history' is a little bit
% better than the simple recursion on the integer.
%
% ?- time(partition_number(53, X, [rec])).		% out of stack.
% ?- time(partition_number(54, X, [history])).	% out of stack.


		/******************************************************
		*     ?- set_prolog_stack(global, limit(2*10**9)).    *
		******************************************************/

% ?- time(partition_number(65, X, [history])).
%@ X = 2012558 .
% ?- time(partition_number(65, X, [history2])).
%@ X = 2012558 .

	/********************************************************************
	*    (F) Fast partion number based on the regression formula:       *
	*                                                                   *
	*     $\sum_{j\geq0} p(n-\frac{j(3j\pm 1)}2)(-1)^j = 0$             *
	********************************************************************/

% ?- partition_number(1, X, [formula]).
% ?- partition_number(2, X, [formula]).
% ?- partition_number(3, X, [formula]).
% ?- partition_number(4, X, [formula]).
% ?- partition_number(6, X, [formula]).
% ?- partition_number(7, X, [formula]).
% ?- partition_number(20, X, [formula]).
% ?- time(partition_number(10000, X, [formula])).
%@ % 9,075,313 inferences, 1.560 CPU in 1.578 seconds (99% CPU, 5816379 Lips)
%@ X = 36167251325636293988820471890953695495016030339315650422081868605887952568754066420592310556052906916435144.

%
partition_number(N, P):- succ(N, N1),
		  fast_partition(N1, V),
		  arg(N1, V, P).

% ?- fast_partition(1, X).
% ?- fast_partition(2, X).
% ?- fast_partition(3, X).
% ?- fast_partition(4, X).
% ?- fast_partition(5, X).
% ?- fast_partition(6, X).
% ?- fast_partition(7, X).
% ?- fast_partition(8, X).
% ?- fast_partition(9, X).
% ?- fast_partition(10, X).
% ?- fast_partition(11, X).
% ?- fast_partition(21, X).

%
fast_partition(N,  Vector):- functor(Vector, p, N),
	arg(1, Vector, 1),
	fast_partition(2, N, Vector).

%
fast_partition(K, N, _):- K> N, !.
fast_partition(K, N, V):- fast_partition(1, K, V, 0, S),
	arg(K, V, S),
	succ(K, K1),
	fast_partition(K1, N, V).
%
fast_partition(J, K, V, S, S0):-
	sign_index(J, Sign, X, Y),
	K1 is K-X,
	K2 is K-Y,
	( K1=<0, K2=<0 	->	S0 = S
	; 	( K1 > 0 -> arg(K1, V, A1); A1 = 0 ),
		( K2 > 0 -> arg(K2, V, A2); A2 = 0 ),
		S1 is S + Sign * (A1 + A2),
		succ(J, J1),
		fast_partition(J1, K, V, S1, S0)
	).

%
sign_index(J, Sign,  X, Y):-
	( J mod 2 =:= 0 ->  Sign = -1; Sign = 1 ),
    A is 3 * J * J,
	X is (A - J) // 2,
	Y is (A + J) // 2.

% ?- time(partition_number_list(10000, X)).
%@ % 1,462,095,308 inferences, 96.689 CPU in 96.820 seconds (100% CPU, 15121631 Lips)
%@ X = 36167251325636293988820471890953695495016030339315650422081868605887952568754066420592310556052906916435144.

partition_number_list(N, P):- succ(N, N1),
		  fast_partition_list(N1, V),
		  nth1(N1, V, P).

% ?- fast_partition_list(10, X).
fast_partition_list(N,  Vector):- length(Vector, N),
	nth1(1, Vector, 1),
	fast_partition_list(2, N, Vector).

%
fast_partition_list(K, N, _):- K> N, !.
fast_partition_list(K, N, V):- fast_partition_list(1, K, V, 0, S),
	nth1(K, V, S),
	succ(K, K1),
	fast_partition_list(K1, N, V).
%
fast_partition_list(J, K, V, S, S0):-
	sign_index(J, Sign, X, Y),
	K1 is K-X,
	K2 is K-Y,
	( K1=<0, K2=<0 	->	S0 = S
	; 	( K1 > 0 -> nth1(K1, V, A1); A1 = 0 ),
		( K2 > 0 -> nth1(K2, V, A2); A2 = 0 ),
		S1 is S + Sign * (A1 + A2),
		succ(J, J1),
		fast_partition_list(J1, K, V, S1, S0)
	).

		/****************************************
		*     integer partition by recursion    *
		****************************************/

% ?- scm(int_part).
% ?- partition_number(1, X, [rec]).
% ?- partition_number(2, X, [rec]).
% ?- partition_number(3, X, [rec]).
% ?- partition_number(10, X, [rec]).

integer_partition(1, [[1-1]]).
integer_partition(N, Ps):- N>1,
	succ(N0, N),
	integer_partition(N0, Qs),
	bump_shift_list(Qs, P0s, []),
	sort(P0s, Ps).

%
reverse([], X, X).
reverse([A|R], X, Y):- reverse(R, [A|X], Y).

%
bump_shift_list([], X, X).
bump_shift_list([A|B], X, Y):-
	bump_shift(A, X, X0),
	bump_shift_list(B, X0, Y).

% ?- bump_shift([1-2], X, Y).
% ?- bump_shift([2-1, 3-1], X, Y).
bump_shift([], X, X).
bump_shift([1-K|U], [[1-K0|U]|X], Y):- !, succ(K, K0),
    bump_shift([], [1-K|U], X, Y).
bump_shift(U, [[1-1|U]|X], Y):-  bump_shift([], U, X, Y).

% ?- bump_shift([],[1-2], X, Y).
%@ X = [[1-1, 2-1]|Y] .
bump_shift(_Rev, [],  X, X).
bump_shift(Rev, [I-J|V], X, Y):- succ(I, I0),
	succ(J0, J),
	bump_shift(Rev, I, J0, I0, V, X, X0),
	bump_shift([I-J|Rev], V, X0, Y).
%
bump_shift(Rev, _I, 0, I0, V, [U|X], X):- !, bump_shift_(I0, V, U0),
	reverse(Rev, U0, U).
bump_shift(Rev, I, J, I0, V, [U|X], X):- bump_shift_(I0, V, U0),
	reverse(Rev, [I-J|U0], U).

bump_shift_(I, [I-K|V], [I-K0|V]):- !, succ(K, K0).
bump_shift_(I, V, [I-1|V]).

		/**************************************
		*		Inversion of Polynomials.     *
		**************************************/

% For a given  polynimial p(x) = ax + ... (a is non zero rational),
% there is a unique polynomial q(x) of a specified degree
% such that  p(q(x) + r(x))  = x   as a formal power series
% for some power series r(x) whose lowest degree is
% larger than the degree of p(x). q(x) is considered
% as a polynomial approximation to the inverse of
% of the  polynomial function p(x) around zero.

% This program is written as an application of the integer partition
% just for fun.

%  A list [1, 0, 1, 0, 0, 0], for example,  means
%  a polynimial x + x^3.

% ?- scm(int_part).
% ?- inverse_poly([3, 1], R, [verify]).
% ?- inverse_poly([2, 1], R, [verify]).
% ?- inverse_poly([1, 0, 0], R, [verify]).
% ?- inverse_poly([1, 1, 0, 0, 0, 0], R,[verify]).
%@ [1,0,0,0,0,0,-132,165,-308,616,-1176,1764,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]
%@ R = [1, -1, 2, -5, 14, -42] .
% ?- inverse_poly([1, 0, 1], R,[verify]).
% ?- inverse_poly([1, 1, 0, 0, 0, 0], R,[verify]).
% ?- inverse_poly([3, 0, 1, 0, 0, 0], R,[verify]).
% ?- inverse_poly([1, 0, 1, 0, 0, 0], R,[verify]).
% ?- inverse_poly([1, 1, 0, 0, 0, 0, 0, 0, 0, 0], R,[verify]).
% ?- inverse_poly([1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0], R,[verify]).
% ?- inverse_poly([1,1], R, [degree(20), verify]).
% ?- inverse_poly([1,1], R, [degree(30), verify]).
% ?- inverse_poly([1,1], R, [degree(35), verify]).
% ?- inverse_poly([1,1], R, [degree(35), verify]).
% ?- inverse_poly([1,1], R, [degree(36), verify]).
% ?- inverse_poly([1,1], R, [degree(37), history, verify]).
% ?- inverse_poly([1,1], R, [degree(37), rec, verify]).
%@ ERROR: Out of local stack
%@ ERROR: Out of local stack

% ?- inverse_poly([1,1], R, [degree(37), power_table, verify]).
% ?- scm(int_part).
% ?- inverse_poly([-2], R, []).
% ?- inverse_poly([-2], R, []).
% ?- inverse_poly([-2], R).
% ?- inverse_poly([1, 0], R).
% ?- inverse_poly([1, 1], R, []).
% ?- inverse_poly([1, 1, 0], R, [verify]).
% ?- inverse_poly_by_power_increment(120, [1, 1], R).
% ?- inverse_poly_by_power_increment(121, [1, 1], R).
% ?- inverse_poly_by_power_increment(120, [1, 1], R).
% ?- inverse_poly([1, 1], R, [degree(50), power_table]).
% ?- inverse_poly([1, 1], R, [degree(50), rec]).
% ?- inverse_poly([1, 1], R, [degree(50)]).
% ?- inverse_poly([1, 1], R, [degree(120), power_table]).
%@ R = [1, -1, 2, -5, 14, -42, 132, -429, 1430|...] .

% ?- inverse_poly([1, 1], R, [degree(121), power_table]).
% ?- inverse_poly([1, 1], R, [degree(122), power_table]).
%@ R = [1, -1, 2, -5, 14, -42, 132, -429, 1430|...] .

% ?- inverse_poly([1, 1], R, [degree(123), power_table]).
%@ ERROR: Out of local stack
% ?- inverse_poly([1, 1], R, [degree(122), power_table, verify]).

% ?- inverse_poly([1, 1], R, [degree(11), rec]).
%@ R = [1, -1, 2, -5, 14, -42, 132, -429, 1430|...] .
%
inverse_poly(Poly, R, Opt0):-
	(	memberchk(degree(Deg), Opt0)
	->	Opt = Opt0
	;	length(Poly, Deg),
		Opt = [degree(Deg)|Opt0]
	),
	length(Poly, Lp),
	N is max(Deg, Lp),
	length(Poly0, N),
	append(Poly, Q, Poly0),
	maplist(=(0), Q),
	Vec=..[u|Poly0],
	length(R, N),
	IVec=..[v|R],
	Poly=[C1|_],
	D1 is rdiv(1, C1),
	R = [D1|_],
	(	memberchk(rec, Opt)
	->		integer_partition(2, Ps),
			inverse_poly(2, N, Vec, IVec, Ps)
	;		course_of_partition(2, Hs),
 			inverse_poly_history(2, N, Vec, IVec, Hs)
	),
	(	memberchk(verify, Opt0)
	-> 	math:poly(##([0|Poly], [0|R]), [_|U]),
		% polynimial composition.
		writeln(U)
	;	true
	).

% ?- poly_to_list(poly(3, [1]), R).
%@ R = [0, 0, 0, 1].
% ?- poly_to_list(poly(3, [0]), R).
poly_to_list(poly(0, L), L):-!.
poly_to_list(poly(E, L), M):- !, succ(E0, E),
	poly_to_list(poly(E0, [0|L]), M).
poly_to_list(L, L).

%
inverse_poly(I, N, _,  _, _):- I > N, !.
inverse_poly(I, N, Vec, IVec, Ps):-
	nth_coefficient(I, Vec, IVec, C, Ps),
	arg(I, IVec, B),
	arg(1, Vec, A1),
	B is  - (C rdiv A1),
	succ(I, I0),
	bump_shift_list(Ps, Q0s, []),
	sort(Q0s, Qs),
	inverse_poly(I0, N, Vec, IVec, Qs).

%
inverse_poly_history(I, N, _,  _, _):- I > N, !.
inverse_poly_history(I, N, Vec, IVec, Hs):-
	Hs=[Ps|_],
	nth_coefficient(I, Vec, IVec, C, Ps),
	arg(I, IVec, B),
	arg(1, Vec, A1),
	B is  - (C rdiv A1),
	succ(I, I0),
	extend_partition_list(Hs, I0, Qs, []),
	sort(Qs, Q0s),
	inverse_poly_history(I0, N, Vec, IVec, [Q0s|Hs]).

% ?- fold_partition([1-1, 2-2], 6, v(5, 6, 1), V).
%@ V = 540
% ?- fold_partition([3-1, 2-2], 120,  v(5, 6, 1), V).
%@ V = 2160 .
% ?- X is (120*(5**3)*(6**2)) // (6*2).
%@ X = 45000.

fold_partition(P, Jfac, Cs, D):-
	foldl(pred(Cs, ([N-K, U, V]
				   :- arg(N, Cs, B),
					  math:scalar_exp(B, K, P0),
					  V is P0 * U)),
		  P, 1, D0),
	foldl(pred(([_-K, U, V]
			   :- factorial(K, Fk),
				  V is Fk*U)),
		  P, 1, F),
	D is (Jfac rdiv F)*D0.

% ?- fold_partition_list([[1-1, 2-2],[1-1, 2-2]], 3, v(5, 6), V).
%@ V = 1080 .

fold_partition_list(Ps, J, Cs, E):-
	factorial(J, Jfac),
	foldl(pred([Cs, Jfac], ( [P, U, V]
					 :- fold_partition(P, Jfac, Cs, A),
						V is A+U)),
		  Ps, 0, E).

%
nth_coefficient(I, U, V, C, Ps):-
	nth_coefficient(I, 2, Ps, U, V, 0, C).
%
nth_coefficient(I, J, _, _, _, C, C):- J > I, !.
nth_coefficient(I, J, Ps, U, V, C, C0):-
	collect_by_partition_size(Ps, J, Qs),
	fold_partition_list(Qs, J, V, A),
	arg(J, U, M),
	C1 is C + M*A,
	succ(J, J0),
	nth_coefficient(I, J0, Ps, U, V, C1, C0).

%
collect_by_partition_size([A|R], K, [A0|R0]):-
	( A = p(_, A0);  A =A0 ),
	partition_size(A0, K),
	!,
	collect_by_partition_size(R, K, R0).
collect_by_partition_size([_|R], K, R0):- !,
	collect_by_partition_size(R, K, R0).
collect_by_partition_size([], _, []).

%
partition_size([], 0).
partition_size([_-I|R], K):- K >= 0, K0 is K -I,
	partition_size(R, K0).

%
factorial(N, F):- factorial(N, 1, F).
%
factorial(1, N, N).
factorial(K, N, N0):- N1 is K*N, succ(K0, K),
					  factorial(K0, N1, N0).
%
mul_list(L, M):- foldl(pred([X,Y,Z]:-  Z is X*Y), L, 1, M).

% ?- num_of_bags(4, [1,2,3], V).

num_of_bags(N, L, V):- factorial(N, F),
					mul_list(L, M),
					V is F//M.


		/********************************************************
		*     inverse a polynomial by incrementing exponents    *
		********************************************************/

% [2016/04/20] Drastic improvent without integer partitions !!
% A drastic improvent for inversing polynomials
% by decreasing multiplications on polynomials.
% It is based on a familiar and well-known binomial formula:
%
%	 (a + b)^n = Sum of nCr a^(n-r)b^r
%
% with i running from 0 to n.
%
% For examples:
%		(a + b)^2 = a^2 +2ab + b^2
%		(a + b)^3 = a^3 +3a^2b + 3ab^2 + b^3

% combinations
% ?- combination(4, 0, X).
% ?- combination(4, 1, X).
% ?- combination(4, 2, X).
% ?- combination(4, 3, X).
% ?- combination(4, 4, X).

% ?- scm(int_part).
% ?- poly_coeff(1, poly(1, [2]), V).
%@ V = 2.
% ?- poly_coeff(1, poly(1, []), V).
%@ V = 0.
% ?- poly_coeff(1, [], V).
%@ V = 0.
% ?- poly_coeff(1, [1,2], V).
%@ V = 2.

poly_coeff(I, poly(J, L), V):- !,
	plus(K, J, I),
	(	nth0(K, L, V)
			-> true
			; V=0
	).
poly_coeff(I, L, V):-
	(	nth0(I, L, V)
			-> true
			; V=0
	).

% ?- scm(int_part).
% ?- fold_poly_coeff(1,[3,5], [poly(1, [1]), poly(1, [2])], X).
%@ X = 13 .
% ?- fold_poly_coeff(1,[3], [poly(1, [1]), poly(1, [2])], X).
%@ X = 3 .
% ?- fold_poly_coeff(1,[3,5], [poly(1, [1])], X).
%@ X = 3.

fold_poly_coeff(I, Ws, Ps, X):- fold_poly_coeff(I, Ws, Ps, 0, X).
%
fold_poly_coeff(I, [W|Ws], [P|Ps], X, Y):-!,
	poly_coeff(I, P, C),
	X0 is W*C+X,
	fold_poly_coeff(I, Ws, Ps, X0, Y).
fold_poly_coeff(_, [], _, X, X).
fold_poly_coeff(_, _, [], X, X).

% ?- scm(int_part).
% ?- inverse_poly_by_power_increment(3, [1], R).
% ?- inverse_poly([1], R, [degree(3), power_table]).
% ?- inverse_poly([1, 1], R, [degree(10), power_table]).
% ?- inverse_poly([1, 1], R, [degree(20), power_table]).
% ?- inverse_poly([1, 1], R, [degree(100), power_table]).
% ?- inverse_poly([1, 1], R, [degree(120), power_table]).
% ?- inverse_poly([1, 1], R, [degree(10), power_table,verify]).
% ?- inverse_poly([1, 1], R, [degree(20), power_table]).
% ?- inverse_poly([1, 1], R, [degree(100), power_table]).
% ?- inverse_poly([1, 1], R, [degree(120), power_table]).
% ?- inverse_poly([1, 1], R, [degree(121), power_table]).
%@ R = [1, -1, 2, -5, 14, -42, 132, -429, 1430|...] .

% ?- set_prolog_stack(global, limit(2*10**9)).
%@ true.

inverse_poly_by_power_increment(N, [A1|As], R):-
	B1 is 1 rdiv A1,
	update_power(1, poly(1, [B1]), [[]], N, Ps),
	repeat_update_power(2, N,  A1, As, Ps, Qs),
	last(Qs, R0),
	once(trim_constant_term(R0, R)).

% For compatibility with other methods
trim_constant_term(poly(0, [_|R]), R).
trim_constant_term(poly(1, R), R).
trim_constant_term(poly(J, R), S):- succ(J0, J),
	trim_constant_term(poly(J0, [0|R]), S).
trim_constant_term([_|R], R).

%
repeat_update_power(I, N, _, _, Ps, Ps):- I > N, !.
repeat_update_power(I, N, A, Ws, Ps, Qs):-
	reverse(Ps, [_|Us]),
 	fold_poly_coeff(I, Ws, Us, B),
	C is - (B rdiv A),
	succ(I, I0),
	update_power(I, poly(I, [C]), Ps, N, Rs),
	repeat_update_power(I0, N, A, Ws, Rs, Qs).

% ?- update_power(1, poly(1,[1]), [[]], 3, X).
%@ X = [poly(0, [0, 0, 1]), poly(0, [0, 1])] .
% ?-  update_power(2, poly(3,[1]), [poly(2,[1]), poly(1,[1])], 10, P).
%@ P = [poly(0, [0, 0, 0, 1, 0, 3, 0, 3, 0, 1]), poly(0, [0, 0, 1, 0, 2, 0, 1]), poly(0, [0, 1, 0, 1])] .

update_power(N, A, Ps, Max, [QH|Qs]):-
	update_power(N, A, Ps, Max, Qs, []),
	last(Qs, Q),
	Qs = [H|_],
	math:poly_mul_bd_poly(Q, H, Max, QH).
%
update_power(0, _, _, _, X, X):-!.
update_power(I, A, Ps, Max, [P0|X], Y):-
	fold_poly_power_combi(I, A, Ps, Max, P0),
	succ(I0, I),
	Ps=[_|Qs],
	update_power(I0, A, Qs, Max, X, Y).

% ?-  fold_poly_power_combi(2, poly(3,[1]), [poly(2,[1]), poly(1,[1])], 10, P).
fold_poly_power_combi(N, A, Ps, Max, P):-
	fold_poly_power_combi(Ps, 0, N, A, poly(0,[1]), Max, [], P).

%
fold_poly_power_combi([], _, _, _, V, _, P, Q):- !,
	math:poly_add(V, P, Q).
fold_poly_power_combi([U|Us], I, N, A, V, Max, P, Q):-
	combination(N, I, C),
	math:poly_mul_bd_poly(V, U, Max, VU),
	math:poly_scalar(C, VU, CVU),
	math:poly_add(CVU, P, P0),
	succ(I, I0),
	math:poly_mul_bd_poly(A, V, Max, AV),
	fold_poly_power_combi(Us, I0, N, A, AV, Max, P0, Q).

% combi(N, R, C) <=>
% Let X, R be a set  s.t. |X| = N, a number R, respectively,
% then C =  |{Y in powerset(X) | |Y| = R}|

combination(_, 0, 1):- !.
combination(N, N, 1):- !.
combination(N, R, C):- N >= R,
	NR is N - R,
	(  R < NR
		-> combination(N, R, NR, C)
		;  combination(N, NR, R,C)
	).

%
combination(N, R, S, C):- succ(S, S1),
	nat_prod(S1, N, 1, M),
	factorial(R, FR),
	C is M rdiv FR.

%
nat_prod(K, N, X, X):-  K>N, !.
nat_prod(K, N, X, Y):-  X0 is K*X,
	succ(K, K1),
	nat_prod(K1, N, X0, Y).

% A bench mark of two methods (A) and  (B) on integer partitions.

		/************************************************************
		*     (A) integer partition by course-of-value induction    *
		************************************************************/

% ?- partition_number(40, X, [history]).
%@ X = 37338 .
% ?- partition(10, X), maplist(writeln, X).

partition(N, Ps):- course_of_partition(N, [Ps|_]).

% ?- course_of_partition(4, X).

incremental_partition(Ps, N, [Qs|Ps]):-
		extend_partition_list(Ps, N, Q0s, []),
		sort(Q0s, Qs).

% ?- scm(int_part).
% ?- course_of_partition(2, X).
% ?- course_of_partition(3, X).
% ?- course_of_partition(4, X).
% ?- course_of_partition(1,X), incremental_partition(X,2,Y).
% ?- course_of_partition(1,X), incremental_partition(X,2,Y), incremental_partition(Y,3,Z).

course_of_partition(0,[[p(0, [])]]).
course_of_partition(N, Ps) :- N > 0,
	succ(N1, N),
	course_of_partition(N1, Qs),
	incremental_partition(Qs, N, Ps).

% ?- scm(int_part).
extend_partition_list([], _, P, P).
extend_partition_list([A|R], N, P, Q):-
	extend_partition_level(A, N, P, P0),
	extend_partition_list(R, N, P0, Q).

%
extend_partition_level([], _, P, P).
extend_partition_level([X|R], N, P, Q):-
	extend_canonical(X, N, P, P0),
	!,
	extend_partition_level(R, N, P0, Q).

% Increasing order on block size
% (decreasing order on block size does not work.
extend_canonical(p(_, []), N, [p(N, [N-1])|D], D).
extend_canonical(p(I, [Block_length-K|R]), N, [P|D], D):-
	N1 is N-I,
	(	N1 >  Block_length
	->  P = p(N, [N1-1, Block_length-K|R])
	;   N1 == Block_length
	->  succ(K, K1),
		P = p(N, [Block_length-K1|R])
	).
extend_canonical(_, _, D, D).


		/******************************
		*     partition_history_tr    *
		******************************/
% Tail recursion version.
% Despite of tail-recursion, it shows weaker performance ! Why ?

% ?- scm(int_part).
%@ true.
% ?- partition_history_tr(0, X).
% ?- partition_history_tr(1, X).
% ?- partition_history_tr(2, X).
% ?- partition_history_tr(3, X).
% ?- partition_history_tr(4, X).
% ?- partition_history_tr(40, X).
% ?- partition_history_tr(50, X).
% ?- partition_history_tr(52, X).	% stack overflow
% ?- partition_history_tr(53, X).	% stack overflow
% ?- partition_history_tr(52, X), flatten(X, Y), length(Y, L).
% ?- partition_history_tr(40, X), flatten(X, Y), length(Y, L).
% ?- partition_number(40, N, [formula]).
%@ N = 37338.

partition_history_tr(0, [[p(0, [])]]).
partition_history_tr(N, Ps) :- N > 0,
		partition_history_tr(0, P0s),
		partition_history_tr(1, N, P0s, Ps).

%
partition_history_tr(I, N, Ps, Ps):- I > N, !.
partition_history_tr(I, N, Ps, Qs):-
 	extend_partition_list(Ps, I, A, []),
 	sort(A, B),
 	succ(I, J),
 	partition_history_tr(J, N, [B|Ps], Qs).

		/*************************************************
		*     (B)  integer partition in Finite Domain    *
		*************************************************/

% ?- time(partition_clpfd_count(20, X)).
% ?- time(partition_clpfd_count(30, X)).
% ?- time(partition_clpfd_count(40, X)).
% ?- time(partition_clpfd_count(45, X)).
% ?- time(partition_clpfd_count(50, X)).
%@ % 2,800,717,455 inferences, 222.620 CPU in 223.016 seconds (100% CPU, 12580703 Lips)
%@ X = 204226.
% ?- partition_clpfd_count(51, X).
%@ ERROR: Out of global stack

% ?- partition_clpfd(10, X).
partition_clpfd_count(N, N0):- setof(Vs, partition_clpfd(N, Vs), All),
					  length(All, N0).

%
partition_clpfd(N, Vs):-
	length(Vs, N),
	ins(Vs, ..(0, N)),
	numlist(1, N, Ns),
	inner_product(Vs, Ns, 0, N),
	label(Vs).
%
inner_product([], [], A, A).
inner_product([X|R], [I|S], A, B):-
	#=(I*X + A, A0),
	inner_product(R, S, A0, B).

		/***************************************
		* (C) naive, but inefficient version   *
		***************************************/

% ?- naive_young(2, Ps).
% ?- naive_young(3, Ps).
% ?- naive_young(4, Ps).
% ?- naive_young(5, Ps).
% ?- time(naive_young(6, Ps)).
%@ % 3,098,266 inferences, 0.328 CPU in 0.367 seconds (89% CPU, 9441874 Lips)
%@ Ps = [ ... ].

% ?- time(naive_young(7, Ps)).
%@ % 42,369,834 inferences, 3.568 CPU in 3.833 seconds (93% CPU, 11874930 Lips)
%@ ERROR: Out of local stack

naive_young(N, Ps):- n_powered_n(N, Series),
	collect(pred(N, [V] :-
					 fold_vector(V, N)),
			Series, Ps).

% ?- n_powered_n(8, P).
% ?- n_powered_n(2, P).
% ?- n_powered_n(4, P).

% ?- util:choices([[1,2],[1,2]], R).
n_powered_n(N, P):- numlist(0, N, L),
				  length(Q, N),
				  maplist(=(L), Q),
				  choices(Q, P).

% ?- fold_vector([1,2,3], S).
%@ S = 14.
fold_vector(R, S):- fold_vector(R, 1, 0, S).

%
fold_vector([], _,  A, A).
fold_vector([X|R], J, A, B):-
	C is J*X + A,
	succ(J, J0),
	fold_vector(R, J0, C, B).
