:-module(interval,[]).
% :- use_module('pac-core').
:- use_module(pac('expand-pac')).
% :- expects_dialect(pac).
term_expansion --> pac:expand_pac.
:- use_module(pac(op)).

% ?- eval(interval:interval, i(1)+i(1), X).
% X = [i(1)]
% ?- eval(interval:interval, i(1)+i(1,2), X).
% X = [i(1)]
% ?- eval(interval:interval, i(1,1)+i(2,3), X).
% X = [i(1, 3)]
% ?- eval(interval:interval, i(2,3)+i(1,1), X).
% X = [i(1, 3)]
% ?- eval(interval:interval, i(3, 4)+i(1,1), X).
% X = [i(1, 1), i(3, 4)]
% ?- eval(interval:interval, i(3, 4) & i(1,1), X).
% X = []
% ?- eval(interval:interval, -( i(3, 4) & i(1,1)), X).
% X = [i(0)]

%% rules for interval calculation

interval(A + B, C, [A, B], [A0, B0], interval:merge_interval2(A0, B0, C)).
interval(&(A, B), C, [-( (- A) + (- B) )], [C], true).  % de Morgan law
interval(-(A), C, [A], [A0], interval:complement(A0,C)).
interval(i(A), [i(A0)], [is::A], [A0], true).
interval(i(A,B), [i(A0,B0)], [is::A, is::B], [A0, B0], true).
interval(X, X, [], [], true).

merge_interval2(X, Y, Z):- append(X, Y, Z0),
	sort_interval(Z0, Z1),
	merge_interval(Z1, Z).

%%% merge intervals
% ?- interval:merge_interval([i(0, 9), i(10,12)], X).
% X = [i(0, 12)]

merge_interval([], []).
merge_interval([I|Is0], Is):- merge_interval(Is0, Is1),
	merge_interval(I, Is1, Is).

merge_interval(I, [], [I]).
merge_interval(I, [J|Js0], Js):- merge_one(I, J, I0), !,
	merge_interval(I0, Js0, Js).
merge_interval(I, Is0, [I|Is]):- merge_interval(Is0, Is).

merge_one(i(P0, Q0), i(P1, Q1), i(P, Q)) :-
	(P0 =< P1 -> P1-Q0 =< 1;  P0 - Q1 =< 1),
	P is min(P0, P1),  Q is max(Q0, Q1).
merge_one(i(P0, Q0), i(P1), i(P)):-
	(P0 =< P1 -> P1-Q0 =< 1; true),
	P is min(P0, P1).
merge_one(i(P0), i(P1,Q1), i(P)):-
	(P0 =< P1 -> true; P0-Q1=<1),
	P is min(P0, P1).
merge_one(i(P0), i(P1), i(P)):- P is min(P0,P1).

% ?- interval:sort_interval([i(3,3), i(2)], X).
% X = [i(2), i(3, 3)]
% ?- interval:complement([i(0, 3), i(6, 8), i(10)], X).
% X = [i(4, 5), i(9, 9)]

% sort_interval(X, Y):- sort:predsort(interval:interval_order, X, Y).
sort_interval(X, Y):- predsort(interval_order, X, Y).

complement(Is, Js):-  complement(0, Is, Ks), remove_empty(Ks, Js).

complement(L, [], [i(L)]).
complement(L, [i(P)], [i(L, P0)]):- P0 is P-1.
complement(L, [i(P,Q)|Is], [i(L, P0)|Js]):-
	 P0 is P-1,
	 Q0 is Q+1,
         complement(Q0, Is, Js).

remove_empty(X, Y):- act(cons_unless(empty), X, Y, []).

cons_unless(P, X, Y, Y):- call(P, X), !.
cons_unless(_, X, [X|Y], Y).

empty(i(P,Q)):- Q < P.

interval_order(D, X, Y):- arg(1, X, P), arg(1, Y, Q), compare(D0, P, Q),
	(D0= '=' -> compare(D, X, Y); D = D0).
