:- module(langsem,[]).
:- use_module(util(math)).
:- use_module(pac(op)).
:- use_module(pac(basic)).

%;; (setq module-query  "qcompile(util(langsem)), module(langsem).")
% ?- qcompile(util(langsem)), module(langsem).

val(X, Y):- pac:eval(X, Y).

% ?- montague_basic(j, X).
% ?- montague_basic(powerset([a,b,c]), X).
% ?- montague_basic(filter([a,b,c], [c]), X).
% ?- montague_basic(currify([(a,b)]), X).
% ?- montague_basic(a, A), length(A, L).
% ?- montague_basic(inverse([(1, b), (3, a), (2, b)], [b]), A).
% ?- montague_basic(value(a, [(a,b)]), A).
% ?- montague_basic(value(a, [(a,b), (a,c)]), A).
% ?- montague_basic(member(2, [1,2]), R).
%
individuals([j,m,u]).
%
montague_basic(Ind, Ind):-individuals(D), memq(Ind, D), !.
montague_basic(unicorn, [u]):-!.
montague_basic(find, [(j, u)]):-!.
montague_basic(powerset(X), P) :- !, math:powerset(X, P).
montague_basic(filter(A, B), F):- !, math:filter(B, A, F).
montague_basic(currify(A), C):- !, currify(A, C).
montague_basic(a, A):-!, individuals(D), a(D, A).
montague_basic(value(R, X), V):- !, value(R, X, V).
montague_basic(inverse(F, Y), X):- !, inverse(F, Y, X).
montague_basic(member(X, Y), R):- !, member(X, Y, R).

% ?- trace, run_cgi(`(peek([@phrase2, @words2,@frame2]), maplist(herbrand), apply(models:semantics))`, [words2 = `[a, man]`, phrase2 =`np`, frame2 =`[man(j),man(b),woman(m),unicorn(u), kick(u,j), find(m,u), walk(j),walk(b),walk(m)]`]).
% ?- test.

test :- repeat,
	(sample(X), writeln(X), write('=> '),
	 (montague(X, Y)-> writeln(Y); writeln( '** fail **')),
	 fail
	; true ), !.

% ?- qcompile(util(langsem)), module(langsem).
% ?- montague(value(currify(a([j,u])), unicorn), R).
% ?- montague(value @ currify(a([j,u])) @ unicorn, R).
% ?- montague(value(currify(a([j,u])), [j]), R).

sample(filter([j], [j,u])).
sample(currify([(a,b)])).
sample(a([j,u])).
sample(unicorn).
sample(find).
sample(currify(a([j,u]))).
sample(value(currify(a([j,u])), unicorn)).
sample(currify(find)).
sample(value(currify(a([j,u])), unicorn)).
sample(value @ currify(a([j,u])) @ unicorn).
sample(inverse(currify(find), (value @ currify(a([j,u])) @ unicorn))).
sample(member(inverse(currify(find), (value @currify(a([j,u])) @ unicorn)), filter([j], [j, u]))).
sample(member(inverse(currify(find), value(currify(a([j,u])), unicorn)), filter([j], [j, u]))).

% ?- test.

test(Query, YesNo):- D = [j,u],
	J = filter([j],D),
	A = a(D),
	AC = currify(A),
	U = unicorn,
	AUnicorn = (@(AC,U)),
	Love = love,
	LoveC = currify(Love),
	Inv = inverse(LoveC,AUnicorn),
	Query = member(Inv,J),
	val(Query, YesNo).

%
member(X, Y, true):-member(X,Y),!.
member(_, _, false).

% ?- qcompile(util(langsem)), module(langsem).
% ?- langsem:field([(a,b),(b,c)], X).
% ?- langsem:currify([(a,b), (b,c)], X).

field(R, F):- setof(A, B^(member((A,B), R); member((B, A), R)), F), !.
field(_, []).

% ?- langsem:value([(a,b)], a, X).
% X = b

value(F, A, V):- memberchk((A,V), F).
%
values(X, R, Vs):- setof(V, member((X, V), R), Vs), !.
values(_, _, []).
%
currify(R, C):-	field(R, S),
	(   setof((X, Ys), (member(X, S), values(X, R, Ys)), C)
	->	true
	;	C = []
	).
%
compose(X, Y, V):-  join(X, Y, V).
%
join(X, Y, V):-
	(	setof((A, B), C^(member((A,C),X), member((C,B),Y)), V)
	->	true
	;	V = []
	).

% ?- inverse([(1, a),(2, b), (3, a)], [a], X).
% X = [1, 3].

inverse(F, R, S):-
	(	setof(X, Y^(member(Y, R),member((X, Y), F)), S)
	->	true
	;	S = []
	).

%
meet(X,Y):- member(A,X), member(A,Y).
%
have(X, Y, true):- member(Y, X), !.
have(_, _, false).

%
a(D, R):-( powerset(D, PowD),
	   setof((X,Y),
		(member(X,PowD),member(Y,PowD), meet(X,Y)),
		 R)
	-> true
	;  R=[]
	).
%
love([(j,u)]).
%
find([(j,u)]).
%
unicorn([u]).

% ?- montague(1, X).
% ?- montague(j, X).
% ?- montague(k, X).	% error.
% ?- montague(a, X).   % default for a
% ?- montague(a([1,2]), X).
% ?- montague(a@[1,2], X).
% ?- montague([1,2], X).
% ?- montague(A, X).
% ?- montague(powerset@powerset([1]), X), length(X, L).
% ?- montague(powerset@powerset([1,2,3,4]), X), length(X, L).

montague(X, X):-  (var(X); integer(X); is_list(X)), !.
montague(X, Y):- elim_atmark(X, X0),
	montague_args(X0, U),
	(	montague_basic(U, Y) -> true
	;	call(U, Y)
	).
%
montague_args([F|As], V):-
	montague_args(F, F0),
	maplist(montague, As, Bs),
	reduce:complete_args(F0, Bs, V).

%
montague_args(F, G):- nonvar(F),
	F=..[H|As],
	maplist(montague, As, Bs),
	G=..[H|Bs].

% ?- elim_atmark(a@(b@c), X).
% ?- elim_atmark(a@b@c, X).
elim_atmark(X, [X]):- var(X), !.
elim_atmark(X@Y, Z):- !, elim_atmark(X, U),
	append(U, [Y], Z).
elim_atmark(X, [X]).
