:- module(latex,[latexit/2, latexit/3, latex_math/2, latex_math/3]).

:- meta_predicate letex_math(2,?,?).

:- set_prolog_flag(allow_variable_name_as_functor, true).
:- use_module(util(prooftree)).
:- use_module(pac(op)).
:- use_module(pac(basic)).
:- use_module(util(file)).
:- use_module(util(obj)).
:- use_module(util('emacs-handler')).
% :- use_module(configs(cgiconfig)).


% "asy -f pdf  xxx.asy;  open xxx.pdf"
asy_current_buffer_file :- cur_dir(D),
	cur_buf(B),
	atom_trim_tail(B, '.asy', B0),
	boomerang(D,
		     (asy(-f(pdf), B0);
			 open(-a('Preview'), B0+'.pdf'))).

%%	pdflatex_current_buffer_file is det.
%  With current buffer file, "platex; dvipdfm; open -a Preview".
pdflatex_current_buffer_file :- cur_dir(D),
	cur_buf(B),
	atom_trim_tail(B, '.tex', B0),
	boomerang([D], (platex(B0);
		    dvipdfmx(B0);
		    open(-a('Preview'), B0+ '.pdf'))).

%%	platex_buffer is det.
%  platex; dvipdfmx;  open.

platex_buffer :- cur_dir(D), cur_buf(B),
	atom_trim_tail(B, '.tex', B0),
	boomerang( D,
	  ( platex(B0);
	    dvipdfmx(B0);
	    open(-a('Preview'), B0+ '.pdf'))).

%
% ?- atom_trim_tail(abcd, cd, X).
% ?- atom_trim_tail(abcd, xx, X).
atom_trim_tail(Atom, Tail, Head):-
	(  sub_atom(Atom, P, _, 0, Tail)
	-> sub_atom(Atom, 0,  P, _, Head)
	;  Head = Atom ).


		/***************
		*     latex    *
		***************/

%%	latex_math(+Ext:text, +X:obj, -Y:obj) is det.
%	Open the file whose directory, base, and extention are
%	given in X.

latex_open_file(Ext) --> obj_get([directory(D), base(B)]),
	{ pshell(open(D+ (/)+ B+ (.)+ Ext)) }.

%%	latexit(M:atom, +X:obj, -Y:obj) is det.
%	Latex it with mode M. Necessary information is in X.

latexit(prooftree) --> prooftree, latex_math.
latexit(math) --> latex_math.

%;; (setq module-query  "qcompile(util(latex)), module(latex).")
% ?- qcompile(util(latex)), module(latex).
% ?- minimal_math_pdf(`$$\\sqrt{a+b}$$`, A).
% ?- minimal_math_pdf(`$$\\sqrt{a_b}$$`, A).

minimal_math_pdf -->
	latex_math(
	(	generate_tex,
    	platex,
    	get_size,
    	dvipdfmx,
		cleaning_tmp_file
	)),
	obj_get([base(B)]),
	{  pshell(open(-a('Skim'), '~/local/cgi-bin/images/' + B +'.pdf'))}.

%
latex_math --> latex_math(
	(	generate_tex,
    	platex,
    	get_size,
    	dvipdfmx,
		cleaning_tmp_file,
		loc(pdf),
		img_src_tag,
		smash
	) ).
%
latex_math(Ph) -->
	{	getenv(home_html_root, HR),
		atomics_to_string([HR, images], /, D),
		qshell(mkdir(-p, D)),
		getenv(host_html_root, UR),
		atomics_to_string([UR, images], /, E)
	},
	peek(X, []),
	obj_put([text(X)]),
	obj_put([directory(D)]),
	obj_put([url_directory(E)]),
	obj_put([counter_name(pdfcounter), stem(math)]),
	{ working_directory(OldD, D) },
	new_file,
	!,
 	phrase(Ph),
    { working_directory(_, OldD) }.

% ?- phrase(((=), {write(a)}, (=), {write(b)}), 1, X).

loc(Ext) --> obj_get([url_directory(T), base(B)]), peek([T, (/), B, (.), Ext]).

img_src_tag --> peek(X, ["<img src=""", X, """/>"]).


%%	trim_expr(+B: expr, +E:expr, +X:text, -Y:text) is det.
%	Trim X by patterns B and E, and unify Y with the result.

trim_expr(B, E) --> expr(B), expr(X), expr(E), peek([], X).

%%	tex_pdf(+X:obj, -Y:url) is det.
%	PDFLatex a given text, whose information is in X,
%	and write the html that has a link to the pdf,  and unify Y with the
%	url of the html file.

tex_pdf -->
    generate_tex,
    platex,
    get_size,
    dvipdfmx,
    generate_html,
    cleaning_tmp_file.

%
cleaning_tmp_file --> obj_get([base(B)]),
   { maplist(delete_file(B), [dvi,out,log,aux,size]) }.

%
delete_file(B, Ext):-
	concat_atom([B, (.) ,Ext], F),
	(	exists_file(F)
         ->	delete_file(F)
	 ;	true
	).

%%	generate_tex(+X:obj, -Y:obj) is det.
%	Generate a latex text in X supplying  preamble,
%	and unify Y with updated X.

generate_tex --> obj_get([base(B), text(X)]),
      { getenv(home, H),
		concat_atom([H, '/texmf/tex/latex/ptex/preambles/jsarticle-minimal.tex'], C),
		concat_atom([H, '/texmf/tex/latex/ptex/preambles/minimalmath.tex'], M),
		S = ["\\def\\boxfilename{", B, ".size}\n"],
		D = "\\begin{document}\\minimalpdf{",
		E = "}\\end{document}",
		concat_atom([B, '.', tex], Ftex),
		combine_file([eh:include_text(C), S, eh:include_text(M), D, X, E], Ftex)
	}.

%%	platex(+X:obj, -Y:obj) is det.
%	Latex X into Y.

platex --> obj_get([base(B)]),
	{ getenv(platex, Tex) },
	{ sh([Tex, ' -halt-on-error',
		      ' -interaction=nonstopmode ',
		      B])}.

%%	pdf2ps(+X:obj, -Y:obj) is det.
%	Pdf2ps X into Y.

pdf2ps --> obj_get([base(B)]), { sh(pdf2ps(B+ ".pdf",  B+ ".ps")) }.

%%	ps2jpg(+X:obj, -Y:obj) is det.
%	Ps2jpg  X into Y.

ps2jpg --> obj_get([base(B)]), { sh(convert(B + ".ps",  B + ".jpg")) }.

%%	get_size(+X:obj, -Y:obj) is det.
%	Get the size of X  into Y.

get_size --> obj_get([base(F)]),
  { atomic_list_concat([F, '.size'], Fsize),
    file(Fsize, read, getstring(S)),
    split(S, [W, H|_])
  },
    obj_put([width(W), height(H)]).

%%	dvipdfmx(+X:obj, -Y:obj) is det.
%	Dvipdfmx  X into Y.

dvipdfmx --> obj_get([base(B), width(W), height(H)]),
    {   P   = """"+ W+ ","+ H+ """",
        M   = "" ,
		PDF =  B+ "."+ pdf,
		getenv(dvipdfmx, DVIPDFMX),
		SH  =.. [DVIPDFMX, -o(PDF), -p(P), M, -x('0pt'), -y('0pt'), B],
		sh(SH)
    }.

% ?-  sh(open('/Users/cantor/tmp')).

%%	generate_html(+X:obj, -Y:obj) is det.
%	Generate HTML file from X into Y.

generate_html --> obj_get([base(B)]),
    {   H = [ "<html><body>", "<img src=""", B, ".pdf""/>", "</body></html>" ],
	atomic_list_concat([B, '.' , html],  HTML),
        file(HTML, write, smash(H))
    }.

%%	open_pdf(+X:obj, -Y:obj) is det.
%	Open pdf file X  in X, and unify Y with X.

open_pdf --> obj_get([directory(D), base(B)]),
	{ pshell(open(D + (/) + B+ '.pdf')) }.

%%	math_view(+X:obj, -Y:obj) is det.
%	Latex X into Y, and preview it.

math_view --> latex_math(
	(generate_tex,
    	platex,
    	get_size,
    	dvipdfmx,
	cleaning_tmp_file,
	obj_put([target('~/public_html/cgi-bin/images')]),
	loc(pdf))),
	current(P),
	{send_off([ "(find-file-existing \"", P, "\")"])}.

%%	latexit(X, X) is det.
%	Latex the current buffer content, and preview it.

latexit --> {latexit}.

latexit :-  send_off("(command-execute 'latex-it)"),
    lisp_atom("default-directory", D),
	dvipdfmx(D),
	send_off("(preview-it)").
