:- module(math, [boole/2,
		 pairs/2,
		 choices/2,  product/3,
		 powerlist/3, powerset/2,
		 string_to_ground_term/2,
		 subst/3,
		 nCr/3, combi/3, combination/4,
		 fun_to_rel/2, rel_to_fun/2,  principal_filter/3,
		 all_maps/3,
		 filter/3, inverse_image/3, image/3,
		 bigcap/2,
		 raise_list/3, distribute_cons/3,
		 fact/2, factorial/2, factorial/3]).

:- use_module(pac(basic)).
:- use_module(util(meta2)).
:- use_module(pac('expand-pac')).
term_expansion --> pac:expand_pac.
:- use_module(pac(op)).

% ?- factorial(54, X).
% ?- X is (10000)^2.		%億
% ?- X is (10000)^3.		%兆
% ?- X is (10000)^4.		%京
% ?- X is (10000)^5.		%
% ?- X is (10000)^6.
% ?- X is (10)^68.			% 無量大数

% ?- X is (10)^68.			% 無量大数
% ?- factorial(54, F), X is 10^68, R is rdiv(F, X).
%@ F = 230843697339241380472092742683027581083278564571807941132288000000000000,
%@ X = 100000000000000000000000000000000000000000000000000000000000000000000,
%@ R = 839804478670853788470022637495376624731865611727r363797880709171295166015625000000000000000000 .

% ?- factorial(54, F), X is 10^68, Q is F//X.
%@ F = 230843697339241380472092742683027581083278564571807941132288000000000000,
%@ X = 100000000000000000000000000000000000000000000000000000000000000000000,
%@ Q = 2308 .

% ?- list_to_brace([[]], X).
% ?- list_to_brace([[],[]], X).
% ?- list_to_brace([[],[], []], X).

:- bekind(list_to_brace, []).
[] = `({}).
[X] = `({})@X.
[X|Y] = `({})@ (`(',')@X@Y).
:- ekind.

% ?- str("ab" + "cd" + "ef", X).
%@ X = "abcdef".
% ?- eval(str::("ab" + "cd" + "ef"), X).
%@ X = "abcdef".

:- bekind(str, []).
X+Y =  :string_concat@X@Y.
% X+Y =  :pred([X, Y, Z]:- string_concat(X, Y, Z))@X@Y.  %% OK
X = `X.
:- ekind.

% ?- binomial(10, [a,b], E), maplist(val([a=(1/2), b=(1/2)]), E, F).

		/**********************************
		*     Eval Boolean expressions    *
		**********************************/

% ?-eval(math:cond(boolify, (true->true)), R).
% ?-eval(math:cond(boolify, (1=1->1=2)), R).
% ?-eval(math:cond(boolify, (1=2->1=2)), R).
% ?-eval(math:cond(boolify, (1=1; 2=3)), X).
% ?-eval(math:cond(boolify, (1=2 -> 2=3)), X).
% ?-eval(math:cond(boolify, (2=2 -> 2=2)), X).
% ?-eval(math:cond(boolify, (1=2 -> 2=2)), X).
% ?-eval(math:cond(boole, true), R).
% ?-eval(math:cond(boole,(-(true*false) +false)), R).
% ?-eval(math:cond(boole, (-(true+false) +false)), R).
% ?-eval(math:cond(boole, -(true + false ) +false), R).
% ?-eval(math:cond(boole, -(true * false ) +false), R).

cond(A1,(t->X),A2):-!,cond(A1,X,A2) .
cond(_A1,(nil-> _A2),nil):-! .
cond(A1,(X->Y),A2):-!,cond(A1,X,A3),cond(A1,(A3->Y),A2) .
cond(_A1,(t;_A2),t):-! .
cond(A1,(nil;X),A2):-!,cond(A1,X,A2) .
cond(A1,(X;Y),A2):-!,cond(A1,X,A3),cond(A1,(A3;Y),A2) .
cond(_A1,(nil,_A2),nil):-! .
cond(A1,(true,X),A2):-!,cond(A1,X,A2) .
cond(A1,(X,Y),A2):-!,cond(A1,X,A3),cond(A1,(A3,Y),A2) .
cond(A1,-t,A2):-!,cond(A1,nil,A2) .
cond(_A1,-nil,t):-! .
cond(A1,-X,A2):-!,cond(A1,X,A3),cond(A1,-A3,A2) .
cond(_A1,t,t):-! .
cond(_A1,nil,nil):-! .
cond(_A1,true,t):-! .
cond(_A1,false,nil):-! .
cond(S,X,A1):-call(S,X,A1) .

% ?- listing(cond).
% ?- boolify(a=a, X).

boolify(X,true):-once(call(X)),! .
boolify(_A1,false).


% ?- scm(math).
% ?- boole(true*false, X).

boole(true*A,A1):-!,boole(A,A1) .
boole(false* _A1,false):-! .
boole(X*Y,A1):-!,(boole(X,A2),boole(Y,A3)),canonical_boole(*,A2,A3,A1) .
boole(true+ _A1,true):-! .
boole(false+A,A1):-!,boole(A,A1) .
boole(X+A,A1):-!,(boole(X,A2),boole(A,A3)),canonical_boole(+,A2,A3,A1) .
boole((X->Y),A1):-!,boole(-X+Y,A1) .
boole(-true,false):-! .
boole(-false,true):-! .
boole(-X,A1):-!,boole(X,A2),canonical_boole(-,A2,A1) .
boole(not(X),A1):-!,boole(X,A2),canonical_boole(-,A2,A1) .
boole(true,true):-!.
boole(false,false).


canonical_boole(+, false, X, X).
canonical_boole(+, X, false, X).
canonical_boole(+, true, _, true).
canonical_boole(+, _, true, true).
canonical_boole(+, X, Y, X+Y).
canonical_boole(*, true, X, X).
canonical_boole(*, X, true, X).
canonical_boole(*, false, _, false).
canonical_boole(*, _, false, false).
canonical_boole(*, X, Y, X*Y).
%
canonical_boole(-, true, false).
canonical_boole(-, false, true).
canonical_boole(-, X, -(X)).

% ?- module(math).
% ?- set_query_context(math).   % <== this is inserted by handle([sqc]).
% ?- canonical_boole(A,B,C).
% ?- eval(math:truth_table(false->true), R).
% ?- eval(math:truth_table(false + true -> false), R).

truth_table(-true,false):-! .
truth_table(-false,true):-! .
truth_table(not(true),false):-! .
truth_table(not(false),true):-! .
truth_table(+true,true):-! .
truth_table(+false,false):-! .
truth_table(false+false,false):-! .
truth_table(_A1+ _A2,true):-! .
truth_table(true*true,true):-! .
truth_table(_A1* _A2,false):-! .
truth_table((true->false),false):-! .
truth_table((_A1-> _A2),true).

% 		/***********************************
% 		*     invering  power series       *
% 		***********************************/

% % ?- trace, math:power_series_inverse([1, 1 rdiv 2, 1 rdiv 6], Xs).
% % Xs = [1, -1 rdiv 2, 1 rdiv 3]


% % ?- trace, math:power_series_inverse([0,1], Xs).

% power_series_inverse(Ds, Xs):- power_series(Ds, As, Ts),
%     reverse([As|Ts], Ws),
%     Ds =[A1|_],
%     Rdiv_A is 1 rdiv A1,
%     find_unknowns(Ws, [Rdiv_A], Ys),
%     reverse(Ys, Xs).

% %
% power_series([A1|Ds], As, Ts) :-
% 	power_series_loop(Ds, [A1], [], As, Ts).

% %
% power_series_loop([], As, Ts, As, Ts).
% power_series_loop([D|Ds], As, Ts, Bs, Us):-
% 	power_series_step(D, As, Ts, Cs, Ws),
% 	power_series_loop(Ds, Cs, Ws, Bs, Us).

% power_series_step(A, As, Ts, [A|As], Us):-
%     reverse(As, As_rev),
% 	innerproduct(As_rev, As, B),
% 	power_series_step(As_rev, B, Ts, Us).
% %
% power_series_step(_, B, [], [[B]]).
% power_series_step(As, B, [T|Ts], [[B|T]|Us]):-
% 	innerproduct(As, T, C),
% 	power_series_step(As, C, Ts, Us).

% :- meta_predicate inverse_binary(3, ?, ?, ?).
% %
% inverse_binary(F, X, Y, Z):- call(F, Y, Z, X).
% %
% find_unknowns([_], Xs, Xs):-!.
% find_unknowns([[A]|Ts], Xs, [X|Ys]):-
% 	maplist(inverse_binary(pred([U,V,[U|V]), Ts, Cs, Us),
% 	find_unknowns(Us, Xs, Ys),
% 	innerproduct(Ys,Cs,W),
% 	X is - (W rdiv A).

% inner product of two "vectors" adjusting to the shorter
% ?- innerproduct([1,2], [1,2,3], X).

innerproduct(X,Y,Z):- innerproduct(X, Y, 0, Z).

innerproduct([], _, X, X).
innerproduct(_, [], X, X).
innerproduct([A|As], [B|Bs], X, Y):-
	X0 is X+ A * B,
	innerproduct(As, Bs, X0, Y).

% ?- poly([1,2]+[2,3], X).
% ?- poly([1,2]+[2,3,4], X).
% ?- poly(##(poly(1, [1,0,1]), poly(2, [1,1])), X).
%@ X = poly(2,[1,1,0,0,1,3,3,1]) .
% ?- poly([0,0,1]*[0,0,1], X).
%@ X = [0,0,0,0,1] .
% ?- poly([0,1]^8, X).
%@ X = [0,0,0,0,0,0,0,0,1] .
% ?- poly([0,1]^0 mod [1,0,1], X).
% X = [1]
% ?- poly([0,1]^1 mod [1,0,1]), X).
%@ X = [0, 1] .
% ?- poly([0,1]^2 mod [1,0,1], X).
% ?- poly([0,1]^3 mod [1,0,1], X).
% ?- poly([0,1]^4 mod [1,0,1], X).
% ?- poly(##([0, 0, 0, 1], [1, 1]), X).
% ?- poly(##([0, 0, 0, 0, 1], [1, 1]), X).
% ?- poly(##([0, 0, 0, 0, 0, 1], [1, 1]), X).


%%% "Little Endian" polynomial calculus
% ?- poly( poly(2, [1,1]) + poly(3, [1,1]), X).
% ?- poly( poly(2, [1,1])^2 + poly(3, [1,1])^3, X).
%@ X = poly(4, [1, 2, 1, 0, 0, 1, 3, 3, 1]) .
% ?- poly(poly(2, [1,1])*poly(3, [1,1]), X).
%@ X = poly(5, [1, 2, 1]) .
% ?- poly(poly(2, [1,1])^5, X).
% ?- poly((- (1*2*3))^2, X).
%@ X = [36]
% ?- poly((- (1*2*3)), X).
%@ X = [-6].
% ?- poly((- (rdiv(1,2)*3)), X).
%@ X = [-3 rdiv 2].
% ?- poly((- (rdiv(2,4)*3)), X).
%@ X = [-3 rdiv 2].

% ?- listing(poly).

poly(X+Y,A1):-!,(poly(X,A2),poly(Y,A3)),poly_add(A2,A3,A1) .
poly(X-Y,A1):-!,poly(X+[-1]*Y,A1) .
poly(-X,A1):-!,poly([-1]*X,A1) .
poly(X*Y,A1):-!,(poly(X,A2),poly(Y,A3)),poly_mul(A2,A3,A1) .
poly(X^Y,A1):-!,poly(X,X0),poly_exp(X0,Y,A1) .
poly(##(X,Y),A1):-!,(poly(X,A2),poly(Y,A3)),poly_compose(A2,A3,A1) .
poly(inverse(P),A1):-!,poly(P,A2),power_series_inverse(A2,A1) .
poly(P mod Q,A1):-!,(poly(P,A2),poly(Q,A3)),poly_mod_little(A2,A3,A1) .
poly(rev(P),A1):-!,poly(P,A2),reverse_endian(A2,A1) .
poly(X,[X]):-rational(X),! .
poly(X,X).

% ?- poly_mul_bd_poly(poly(0, [1]), poly(0, []), 1, Z).
% ?- poly_mul_bd_poly(poly(0, [1]), poly(0, [2]), 1, Z).
% ?- poly_mul_bd_poly(poly(0, [1, 2, 1]), poly(0, [1, 1]), 3, Z).
% ?- poly_mul_bd_poly(poly(0, [1, 2, 1]), poly(0, [1, 1]), 4, Z).
% ?- poly_mul_bd_poly(poly(1, [1, 2, 1]), poly(2, [1, 1]), 7, Z).
% Degree of the returned polynomial is less than N.

poly_mul_bd_poly(U, V, N, poly(K, Z)):-
	(U = poly(I, X); I=0, X=U), !,
	(V = poly(J, Y); J=0, Y=V), !,
	K is I + J,
	N0 is max(N - K, 0),
	poly_mul_bd(X, Y, N0, Z).

% ?- poly_scalar_bd(1, [1], 2, X).
% ?- poly_scalar_bd(1, [1, 2, 3], 5, X).
% ?- poly_scalar_bd(1, [1, 2, 3, 5], 3, X).

poly_scalar_bd(X, Y, Max, Z):-
	poly_scalar_bd(X, Y, Max, Z, []).

%
poly_scalar_bd(_, [], _, U, U).
poly_scalar_bd(_, _, 0, U, U).
poly_scalar_bd(C, [A|B], J, [D|U], V):- D is A*C,
	succ(J0, J),
	poly_scalar_bd(C, B, J0, U, V).
%
reverse_endian(poly(X, Y), poly(X, Y0)):- !, reverse(Y, Y0).
reverse_endian(X, X0):-  reverse(X, X0).

%
list_plus(A+B,A1):- list_plus(A,A2),
	list_plus(B,A3),
	append(A2,A3,A1).


% ?- poly_compose(poly(1, [1,0,1]), poly(2,[1,1]), X).
% ?- poly_compose(poly(1, [1,0,1]), poly(2,[1,1]), X).
% ?- poly_compose(poly(2, [0,0,0,1]), [1,1], X).
% ?- poly_compose([1, 1, 1], [1,1], X).
% ?- poly_compose([1, 1, 1], [0,1], X).
% ?- poly_compose([0, 1], [2, 3], X).

poly_compose(poly(E0, X0), P, Q):-
	poly_exp(P, E0, P0),
	poly_compo0(X0, P, P1),
	poly_mul(P0, P1, Q).

poly_compose(X0, P, Q):- poly_compose_exact(X0, P, Q).

% ?- scm(math).
% ?- poly_compose_exact([], poly(1,[1]), X).
% ?- poly_compose_exact([1], poly(1,[1]), X).
% ?- poly_compose_exact([1, 1], poly(1,[1,2]), X).
% ?- poly_compose_exact([1, 1, 1], poly(0,[1,1]), X).

poly_compo0([], _, []).
poly_compo0([A|As], P, Q):-
	poly_compo0(As, P, P0),
	poly_mul(P, P0, P1),
	poly_add([A], P1, Q).

poly_exp(poly(E0,X0), N, poly(E, X)):- bits(N, Bs),
	poly_exp0(X0, Bs, [1], X),
	E is E0*N.
poly_exp(X0, N, X):- bits(N, Bs), poly_exp0(X0, Bs, [1], X).

poly_exp0(_, [], X, X).
poly_exp0(X, [B|R], Y, Z):-
	poly_mul_tr(Y, Y, Y0),
       (B==0 -> Y1=Y0; poly_mul_tr(Y0, X, Y1) ),
	poly_exp0(X, R, Y1, Z).

poly_scalar(A, poly(E,X), poly(E, Y)):- poly_scalar0(A, X, Y).
poly_scalar(A, X, Y):- poly_scalar0(A, X, Y).

poly_scalar0(0, _, []).
poly_scalar0(A, X, Y):- maplist(scalar_mul(A), X, Y).

scalar_mul(A, X, Y) :- Y is A*X.

% ?- math:scalar_exp(1 rdiv 3, 4, R), writeln(R).
% ?- math:scalar_exp(1 rdiv 3, 3, R), writeln(R).
% ?- math:scalar_exp(1 rdiv 3, 2, R), writeln(R).
% ?- math:scalar_exp(1 rdiv 3, 1, R), writeln(R).
% ?- math:scalar_exp(1 rdiv 3, 0, R), writeln(R).

scalar_exp(X, N, Y):- bits(N, Bs),
	scalar_exp(Bs, X, 1, Y).

scalar_exp([], _, Y, Y).

scalar_exp([B|R], X, Y, Y0):-  Y1 is Y*Y,
	( B==0 -> Y2 = Y1;  Y2 is X*Y1	),
	scalar_exp(R, X, Y2, Y0).

poly_mul(X, Y, Z):- number(X), poly_scalar(X, Y, Z).
poly_mul(poly(E0,X0),poly(E1,X1), poly(E2,X2)):-
	poly_mul_tr(X0,X1,X2), E2 is E0+E1.
poly_mul(poly(E0,X0), X1, poly(E0,X2)):-poly_mul_tr(X0,X1,X2).
poly_mul(X0, poly(E1,X1), poly(E1,X2)):-poly_mul_tr(X0,X1,X2).
poly_mul(X0, X1, X2):-poly_mul_tr(X0,X1,X2).

% ?- scm(math).
% ?- poly_mul_tr([], [], X).
% ?- poly_mul_tr([1], [2], X).
% ?- poly_mul_tr([1], [1,1], X).
% ?- poly_mul_tr([1,1], [1,1], X).
% ?- poly_mul_tr([1,0,1], [1, 0,1], X).
% ?- poly_mul_tr([1,1,1], [1, 1, 1], X).

poly_mul_tr(A, B, C):- poly_mul_tr(A, B, [], [], C).

poly_mul_tr([], _, W, S, X):- reverse(W, S, X).
poly_mul_tr([A|As], B, W, S, X):-
	poly_scalar(A, B, C),
	poly_add(C, S, V),
	( V == [] ->  U=0, S0=S; V=[U|S0]),
	poly_mul_tr(As, B, [U|W], S0, X).

% ?- scm(math).

% ?- poly_mul_bd([], [], 3, X).
% ?- poly_mul_bd([1], [2], 3, X).
% ?- poly_mul_bd([1], [1,1], 3, X).
% ?- poly_mul_bd([1,1], [1,1], 2, X).
% ?- poly_mul_bd([1,1,1], [1, 1, 1], 3, X).
% ?- poly_mul_bd([1,1,1], [1, 1, 1], 4, X).
% ?- poly_mul_bd([1,1,1], [1, 1, 1], 5, X).
% ?- poly_mul_bd([1,1,1], [1, 1, 1], 6, X).
% ?- poly_mul_bd([1,1,1], [1, 1, 1], 7, X).

poly_mul_bd(A, B, Max, C):- poly_mul_bd(A, B, Max, [], [], C).

%
poly_mul_bd(_, _, 0, W,  _, R):- !, reverse(W, R).
poly_mul_bd([], _, N, W, S, X):- !,
	head(N, S, S0),
	reverse(W, S0, X).
poly_mul_bd([A|As], B, N, W, S, X):-
	poly_scalar_bd(A, B, N, C),
	poly_add(C, S, V),
	succ(N0, N),
	( V = [] -> U = 0, S0 = S;  V=[U|S0]),
	poly_mul_bd(As, B, N0, [U|W], S0, X).

%
head(_, [], []):-!.
head(0, _, []):-!.
head(N, [X|Y], [X|H]):- succ(N0, N),
	head(N0, Y, H).

poly_add(poly(E0,X0), poly(E1,X1), poly(E, X)):- poly_add(E0,X0,E1,X1,E,X).
poly_add(poly(E0,X0), X1, poly(E, X)):- poly_add(E0,X0, 0,X1,E,X).
poly_add(X0, poly(E1,X1), poly(E, X)):- poly_add(0,X0,E1,X1,E,X).
poly_add(X0, X1, X):- poly_add0(X0,X1,X).

poly_add(E0,X0,E1,X1,E,X):-
	E is min(E0, E1),
	D0 is E0-E,
	D1 is E1-E,
	shift(D0, X0, Y0),
	shift(D1, X1, Y1),
	poly_add0(Y0, Y1, Y2),
	(maplist(=(0), Y2) -> X=[]; X=Y2).

poly_add0([],X, X).
poly_add0(X,[], X).
poly_add0([X|Xs], [Y|Ys], [Z|Zs]):- Z is X+Y, poly_add0(Xs,Ys,Zs).

poly_reverse0(poly(E, X), Y) :- repeat(E, cons(0), [], P), reverse(X, P, Y).
poly_reverse0(X, Y) :- reverse(X, Y).

cut_prefix(_, [], []).
cut_prefix(X, [X|Xs], Y):- !, cut_prefix(X, Xs, Y).
cut_prefix(_, Y, Y).

shift(0, X, X):-!.
shift(_, [], []):-!.
shift(N, X, [0|Y]):- N1 is N-1, shift(N1, X, Y).

reverse([], X, X).
reverse([A|As], X, Y):- reverse(As, [A|X], Y).

%  Big endian version
% ?- poly_mod([1,0,0,-1], [1,1,1], X).
%@ X = [0,0] .
% ?- poly_mod([1,0], [1,0], X).
%@ X = [0].
% ?- canonical_mpoly([0, 1,0], A, B).
%@ A = 1,
%@ B = [0].

poly_mod(P, Q, R):- canonical_mpoly(Q, A, Q0),
	length(Q0, L0),
	length(R0, L0),
	maplist(=(0), R0),
	poly_mod(P, Q0, R0, R1),
	poly_scalar0(A, R1, R).

poly_mod([], _, R, R).
poly_mod([C|P], Q, [R0|Rs], R):-
	poly_mod(R0, C, Q, Rs, R1),
	poly_mod(P, Q, R1, R).

poly_mod(A, C, [X|Xs], [Y|Ys], [Z|Zs]):- Z is Y+A*X, poly_mod(A, C, Xs, Ys, Zs).
poly_mod(A, C, [B], [], [D]):- D is A*B + C.

canonical_mpoly([0|P], A, Q):- !, canonical_mpoly(P, A, Q).
canonical_mpoly([A|P], A, P0):- A0 is -1 rdiv A, poly_scalar0(A0, P, P0).

% little-endinan poly_mod based on big-endian
%
% ?- poly_add([1,0], [1,0], R).
poly_mod_little(P, Q, R):- big_endian(P, P0), big_endian(Q, Q0),
	poly_mod(P0, Q0, R0),
	little_endian(R0, R).

big_endian(X, Y):- poly_reverse0(X, X0), cut_prefix(0, X0, Y).

little_endian(X, Y):- cut_prefix(0, X, X0), reverse(X0, Y).

const(F, _, L, V):- call(F, L, V).

arith_curry(F, Y, Z):- build_arith_expr(F, Y, E), !, Z is E.

build_arith_expr(+(X), Y, +(X,Y)).
build_arith_expr(*(X), Y, *(X,Y)).
build_arith_expr(^(X), Y, ^(X,Y)).
build_arith_expr(mod(X), Y, mod(X,Y)).
build_arith_expr(rdiv(X), Y, rdiv(X,Y)).
build_arith_expr(E0, Y, E):- E0=..[F|A0], append(A0,[Y], As), E=..[F|As].

%  arith/3
arith(+, [X, Y|_], Z) :- Z is X+Y.
arith(*, [X, Y|_], Z) :- Z is X*Y.
arith(^, [X, Y|_], Z) :- Z is X^Y.
arith(mod, [X, Y|_], Z) :- Z is X mod Y.
arith(rdiv, [X, Y|_], Z) :- Z is X rdiv Y.
arith(unary(OP), [X|_], V):- E=..[OP, X], V is E.
arith(binary(OP), [X,Y|_], V):- E=..[OP, X, Y], V is E.
arith(ternary(OP), [X,Y,Z|_], V):- E=..[OP, X, Y, Z], V is E.

% ?- val(arith(1+2), X).
% ?- val(arith(1 rdiv 2), X).
% ?- val(arith(1 mod 2), X).

%%%% Endo-map algebra
%%%% endo map / permutation algebra %%%
% *  : group operation
% ^  : integer power
% ?- endo(([1,3,2] ^ -103)*([3,1,2]^102) , X).

perm(E, V) :- endo(E, V), (_V0= l(V); _V0=p(V); V= _V0).


endo_default([],A1):-!,endo_default(l([]),A1) .
endo_default([X|Y],A1):-!,endo_default(l([X|Y]),A1) .


%%% rule of endo-map algebra


endo(id(N),A1):-N0 is N,!,endo_id(v(N0),A1) .
endo(shift(X),A1):-!,endo(X,A2),endo_cycle(A2,A1) .
endo(shift(J,P),A1):-J0 is J,!,endo(P,A2),endo_cycle_v(J0,A2,A1) .
endo(cycle(N),A1):-!,endo(shift(1,N),A1) .
endo(X*Y,A1):-!,(endo(X,A2),endo(Y,A3)),endo_prod(A2,A3,A1) .
% endo(X^Y,A1):-Y0 is Y,!,endo(X,A2),endo_aux_1(A2,A1) .
endo(perms(X),A1):-!,list(X,A2),model:permutations(A2,A1) .
endo(rand(N),A1):-(N0 is N,numlist(1,N0,L),shuffle(L,P)),!,endo(l(P),A1) .
endo(cast(P),A1):-!,endo(P,A2), endo_aux_2(A2,A1) .
endo(cast(F,P),A1):- endo(P,A2),cast(F,A2,A1) .

endo_aux_1(A^[B],C):-endo_pow(A,B,C) .
endo_aux_2(A,B):-'[|]'(v,A,B) .

% ?- cast(v, l([1,2,3]), X).
% X = v(1, 2, 3).
cast(v, l(X), Y):- !, Y=..[v|X].
cast(v, p(L), Y):- !, sort(L, L0), maplist(snd,L0,P), Y=..[v|P].
cast(v, X, X):-!.
cast(p, l(X), p(Y)):- !, length(X, N), numlist(1, N, Z), zip(Z,X, Y).
cast(p, p(X), p(X)):-!.
cast(p, X, p(Y)):- !, functor(X,_,N), X=..[_|X0], numlist(1, N, Z), zip(Z, X0, Y).
cast(l, p(L), l(P)):- !, sort(L, L0), maplist(snd, L0, P).
cast(l, l(X), l(X)):- !.
cast(l, X, l(Y)):- X=..[_|Y].

%% id
endo_id(p, p([])).
endo_id(l, l([])).
endo_id(v(N), Id):- functor(Id, v, N), endo_id_v(N, Id).

endo_id_v(0, _):-!.
endo_id_v(J, Id):- arg(J, Id, J), J0 is J-1, endo_id_v(J0, Id).

%% prod
% ?-  zip([a,b,c],[b,c,a], Z), eval(endo, p(Z)^2, X).
% Z = [ (a, b), (b, c), (c, a) ],
% X = p([ (a, c), (b, a), (c, b) ])

endo_prod(l(X), l(Y), l(Z)):- endo_prod_l(X, Y, Z).
endo_prod(p(X), p(Y), p(Z)):- endo_prod_p(X, Y, Z).
endo_prod(X, Y, Z):- endo_prod_v(X, Y, Z).

endo_prod(F, X, Y, Z):- F==l-> endo_prod_l(X, Y, Z); F==p-> endo_prod_p(X, Y, Z); endo_prod_v(X, Y, Z).

endo_prod_l(F, G, H):- endo_prod_l(1, F, G, H).

endo_prod_l(_, _, [], []) :- !.
endo_prod_l(N, F, [X|Xs], [Y|Ys]) :- (nth1(X, F, Y)->true; Y=X),  N0 is N-1, endo_prod_l(N0, F, Xs, Ys).

endo_prod_p([], Y, Y).
endo_prod_p(X, [], X).
endo_prod_p(F, G, H):- maplist(endo_prod_p_(F), G, H).

endo_prod_p_(F, (A,B), (A,C)):- memberchk((B,C), F) -> true; C=B.

endo_prod_v(X, Y, Z):- functor(X, v, N), functor(Z, v, N), endo_prod_v(N, X, Y, Z).

endo_prod_v(0, _, _,_):-!.
endo_prod_v(J, X, Y, Z):- arg(J, X, K), arg(K, Y, L), arg(J, Z,  L),
	J0 is J-1, endo_prod_v(J0, X, Y, Z).

%%% inverse
endo_inverse(l(X), l(Y)):- !, endo_inverse_l(X, Y).
endo_inverse(p(X), p(Y)):- !, endo_inverse_p(X, Y).
endo_inverse(X, Y):- endo_inverse_v(X, Y).

endo_inverse(F, X, Y):- F==l -> endo_inverse_l(X, Y); F==v -> endo_inverse_v(X, Y); endo_inverse_p(X, Y).

endo_inverse_l(L, M):- length(L, N), numlist(1, N, L0), zip(L, L0, L1),
	lists:msort(L1, L2), unzip(L2, _, M).

endo_inverse_v(P, P0):- functor(P, F,  N), functor(P0, F, N), endo_inverse_v(N, P, P0).

endo_inverse_v(0, _, _):- !.
endo_inverse_v(J, P, P0):- arg(J, P, A), arg(A, P0, J), J0 is J-1, endo_inverse_v(J0, P, P0).

endo_inverse_p([], []).
endo_inverse_p([(X,Y)|F], [(Y,X)|F0]):- endo_inverse_p(F, F0).

%%  cyclic
% ?- endo_cycle(l([3,2,1]), X).
% X = p([ (3, 2), (2, 1), (1, 3)])
% ?- endo_cycle(v(3,2,1), X).
% X = v(1, 3, 2).
endo_cycle(l(L), p(L0)):- endo_cycle_l_p(L, L0).
endo_cycle(X, Y):- endo_cycle_v(1, X, Y).

% ?- endo_cycle_v(1, v(1,2,3), X).
% X = v(2, 3, 1).
endo_cycle_v(J, P, Q):- functor(P, F, N),
						functor(Q, F, N),
						endo_cycle_v(N, N, J, P, Q).

endo_cycle_v(0, _, _, _, _):- !.
endo_cycle_v(K, N, J, P, Q):- arg(K, P, V),
							  V0 is ((V+J-1) mod N)+1,
							  arg(K, Q, V0), K0 is K-1,
							  endo_cycle_v(K0, N, J, P, Q).
% ?- scm(math).
% ?- endo_cycle_l_p([3, 1, 2], X).
endo_cycle_l_p([],  []).
endo_cycle_l_p(Dom, Z):- Dom=[X|Y],
						 append(Y,[X], Dom0),
						 maplist(pred([X,Y,(X,Y)]), Dom, Dom0,Z).

%% power
% ?- endo_pow(l([2,1]), 2, X).
% X = l([1, 2])
% ?- endo_pow(p([(2,1), (1,2)]), 2, X).
% X = p([ (2, 2), (1, 1)])
% ?- endo_pow(v(2,1), 2, X).
% X = v(1, 2)

endo_pow(l(L), N, l(Y)):- !, endo_pow(l, L, N, [], Y).
endo_pow(p(L), N, p(Y)):- !, endo_pow(p, L, N, [], Y).
endo_pow(X, N, Y):- functor(X, _, A), endo_id(v(A), Id), endo_pow(v, X, N, Id, Y).

endo_pow(F, X, N, Id, Y):- N<0, !, N0 is -N, bits(N0, Bits), endo_pow_pos_b(F, X, Bits, Id, X0),
	endo_inverse(F, X0, Y).
endo_pow(F, X, N, Id, Y):- bits(N, Bits), endo_pow_pos_b(F, X, Bits, Id, Y).

endo_pow_pos_b(_, _, [], X, X).
endo_pow_pos_b(F, X, [D|R], Y, Z):- endo_prod(F, Y, Y, Y0),
      ( D==0 -> Y1=Y0; 	endo_prod(F, Y0, X, Y1) ),
	endo_pow_pos_b(F, X, R, Y1, Z).

%%%%% conjugate
% ?- conjugate(v(2,3,1), v(3,1,2)).
% ?- conjugate(l([2,3,1]), l([1,2,3])).

singleton(X, [X]).

conjugate(X, Y):- endo_type(X, T), endo_type(Y, T).

% ?- conjugate_class([1,2,3], X), maplist(writeln, X).
%@ X = [[[1,2,3]],[[1,3,2],[2,1,3],[3,2,1]],[[2,3,1],[3,1,2]]] .
% [[1, 2, 3]]
% [[1, 3, 2], [2, 1, 3], [3, 2, 1]]
% [[2, 3, 1], [3, 1, 2]]

conjugate_class(L, C):- eval(endo, perms(L), Ps),
	maplist(pred([X, l(X)]), Ps, Qs),
	quotient_set(Qs, endo_type, C).

quotient_set(S, F, Q):-  maplist(F, S, T), zip(T, S, Z),
	sort(Z, Z0), merge_pairs(Z0, Q0),
	maplist(maplist(pred([X,Y]:-arg(1, X, Y))), Q0, Q).

%% type
% ?- endo_type(l, [2,1,3], X).
% X = [1, 2].
% ?- endo_type(v(2,1,3), X).
% X = [1, 2].
% ?- endo_type(p([(1,2), (2,1), (3,3)]), X).
% X = [1, 2]

endo_type(l(X), T):- !, endo_type_l(X, T).
endo_type(p(X), T):- !, endo_type_p(X, T).
endo_type(X, T):-  X=..[v|L], endo_type_l(L, T).

endo_type_l(L, T):- maplist(singleton, L, L0),
	union_find_l(1, L, L0, Ks),
	maplist(length, Ks, Ls),
	msort(Ls, T).

endo_type_p(X, T):- field(X, F),
	maplist(singleton, F, L0),
	union_find_p(X, L0, Ks),
	maplist(length, Ks, Ls),
	msort(Ls, T).

% Union-find
% ?- union_find_l(1, [2,1], [[1],[2]], X).
% X = [[1, 2]].
union_find_l(_, [], Bs, Bs):- !.
union_find_l(X, [Y|E], Bs0, Bs):- find_cluster(X, Bs0, CX, Bs1),
   (	memberchk(Y, CX)  ->  Ps = Bs0
   ;    find_cluster(Y, Bs1, CY, Bs2),
        append(CX, CY, U),
	Ps = [U|Bs2]
   ),
  X0 is X+1, union_find_l(X0, E, Ps, Bs).


% ?- union_find_p([(1,2),(2,1)], [[1],[2]], X).
% X = [[1, 2]].
union_find_p([], Bs, Bs):- !.
union_find_p([(X,Y)|E], Bs0, Bs):- find_cluster(X, Bs0, CX, Bs1),
   (	memberchk(Y, CX)  ->  Ps = Bs0
   ;    find_cluster(Y, Bs1, CY, Bs2),
        append(CX, CY, U),
	Ps = [U|Bs2]
   ),
  union_find_p(E, Ps, Bs).

%%% some tiny
meet(A, B):- member(X, A), member(X, B).

find_cluster(A, B, C, D):- select(C, B, D), memberchk(A, C), !.

field(L, L):- listp(L), !.
field(l(L), F):- !, length(L, N), numlist(1, N, F).
field(p(L),  F):- !, length(L, N), numlist(1, N, F).
field(X, F):- X=..[v|F].

add_pair((X,Y), L, L0):- sort([X,Y], P), union(P, L, L0).

% merge
% ?- merge_pairs([(a,b), (a,c)], X).
% X = [[b, c]].

merge_pairs(X, Y):- merge_pairs_(X, X0), maplist(snd, X0, Y).

merge_pairs_([], []).
merge_pairs_([(A,B)|R], X):- merge_pairs_(R, Y), merge_pair(A, B, Y, X).

merge_pair(A, B, [(A,X)|R], [(A,[B|X])|R]):-!.
merge_pair(A, B, R, [(A,[B])|R]).

% apply endo map
% ?- endo_apply(l([2,1,3]), 2, X).
% X = 1.

endo_apply(l(F), A, B):- !, nth1(A, F, B).
endo_apply(p(F), A, B):- !, memberchk((A,B), F).
endo_apply(F, A, B):- arg(A, F, B).

% group action orbits
% ?-  group_orbits(l([2,1,3]), X).
% X = [[1, 2], [3]].
% ?- group_orbits(l([2,1,3]), [1,2,3], X).
% X = [[1, 2], [3]].

group_orbits(F, Y):- field(F, D), group_orbits(F, D, Y).

group_orbits(F, X, Y):- group_orbits(F, X, [[]], Y0),
	maplist(reverse, Y0, Y1), remove([], Y1, Y2),
	sort(Y2, Y).

group_orbits(_, [], O, O):-!.
group_orbits(F, [X|Xs],[[]|R], O):- !, group_orbits(F, Xs, [[X]|R], O).
group_orbits(F, Xs, [H|R], O):- H=[A|_], endo_apply(F, A, A0),
	(  memberchk(A0, H) -> group_orbits(F, Xs, [[], H|R], O)
	;  remove(A0, Xs, Ys), group_orbits(F, Ys, [[A0|H]|R], O) ).

%%%
eval_set_class--> eval([set, subset, ps]).
set_class  -->  eval(set).

setcalc(X, Y, Z, U, V):- set(X, Y, Z, U, V).  % for compatibility

shorter(<, X, Y):- length(X, L), length(Y, M), L<M, !.
shorter(>, _, _).

seq(X, Y) :- maplist(eval, X, Y).

seq(S, X, Y) :- maplist(eval(S), X, Y).

%%  Polymorphic Boolean Algebra

boole_complement(true, false).
boole_complement(false, true).

% ?- boole(fff, a+b, R).
% R = true
% ?- boole(fff, a*b, R).
% R = false

fff(a, true).
fff(_, false).

% :- edit(set/2).


% ?- time(repeat(1000000, eval([boole(prolog)], 1==1, X))).
% is faster than this:
% ?- time(repeat(1000000, eval([boole, prolog], 1==1, X))).

% The index directive :- idnex(boole(0,1,0,0,0,0)) shows
% no significant effect.

% ?- eval(subset(set), [1,2]+[2,1] < [1,2]+[3], X).
% ?- eval(subset(set), [1,2]+[2,1]+[3,2] < [1,2]+[3], X).
% ?- eval(subset(set), [3]+[1]+[2] =< [1,2]+[3], X).
% ?- eval([boole(subset(set))], ([1]=<[1,2]) * ([2,1]>=[1,2,3]), X).
% ?- eval([boole(subset(set))], pow([1]+[2]) < pow([1,2]+[3]), X).
% ?- eval([boole(subset(set))], [1]+[2] < [1,2], X).
% ?- eval([boole(prolog)], 1==1, X).
% ?- eval([boole(frprolog)], (between(1, 5, 3)* member(a,[b,a])), X).
% ?- eval([boole(prolog)], 1==1 -> 2==1, X).
% ?- listing(boole).

% ?- cond(prolog, (true->fail;fail); fail, R).
% R = nil

% ?- binomial(10, [a,b], E).
nil_complement(true, nil).
nil_complement(nil, true).

%
proper_subset([], [_|_]):-!.
proper_subset([X|Y], [X|Z]):- proper_subset(Y,Z).

% powerset(X, Y):- once(foldl(powerset, X,  Y, [[]])).

% powerset(_,[], []).
% powerset(X,[Y,[X|Y]|U],[Y|Z]):- powerset(X, U, Z).

%  powerset(X, Y)  is true if Y is the powerset of X.
powerset(X, Y):- powerset(X, [[]], Y).

powerset([], X, X).
powerset([A|R], X, Y):-
	powerset(X, A, X, X0),
	powerset(R, X0, Y).

%
powerset([], _, X, X).
powerset([X|R], A, S, Y):- powerset(R, A, [[A|X]|S], Y).

%?- ya_powerset([], X).
%@ X = [[]].
%?- ya_powerset([1], X).
%@ X = [[1], []].
%?- ya_powerset([1,2], X).
%@ X = [[2], [2, 1], [1], []].
%?- ya_powerset([1,2,3,4], X), sort(X, X0).
%@ X = [[4], [4, 1], [4, 2, 1], [4, 2], [4, 3, 2], [4, 3, 2|...], [4, 3|...], [4
%?- powerset([1,2,3,4], X).
%@ X = [[], [1], [2], [1, 2], [3], [1, 3], [2, 3], [1|...], [...]|...].

% What is the last argument for ?
% xreverse(Xs, Ys) :-
% 	xreverse(Xs, [], Ys, Ys).

% xreverse([], Ys, Ys, []).
% xreverse([X|Xs], Rs, Ys, [_|Bound]) :-
% 	xreverse(Xs, [X|Rs], Ys, Bound).

% ?- trace, xreverse([1,2,3|R], X).

% ?-  sublist(X, [a,b,c]).
sublist([], []).
sublist(X, [_|As]):- sublist(X, As).
sublist([A|X], [A|As]):- sublist(X, As).

% ?- permutations([a,b,c],X).
% X = [[a, b, c], [b, a, c], [b, c, a], [a, c, b], [c, a, b], [c, b, a]]
permutations([], [[]]).
permutations([X|Xs], Ps):- permutations(Xs, Qs),
	permutations(X, Qs, Ps).

permutations(_, [], []).
permutations(X, [P|Ps], L):- insert_list(X, P, P0),
	permutations(X, Ps, Qs),
	append(P0, Qs, L).

% ?- disjoint_pairs([1,2], X).
disjoint_pairs(X, S) :- bagof((L, R), disjoint_pair(X, L, R), S).

disjoint_pair(X, L, R):-  sublist(L, X),
						  subtract(X, L, L0),
						  sublist(R, L0).

% ?- insert_list(a, [1,2], X).
% X = [[a, 1, 2], [1, a, 2], [1, 2, a]]
insert_list(X, [], [[X]]).
insert_list(X, [A|As], [[X,A|As]|L]):-insert_list(X, As, L0),
	maplist(cons(A), L0, L).

% ?- disjoint_family([[a], [], [b, c]]).
% ?- disjoint_family([[a], [b], [b, c]]).
% [2014/09/08]
disjoint_family([]).
disjoint_family([A|R]):- disjoint(R, A),
	disjoint_family(R).

disjoint([], _).
disjoint([A|_], B):- member(X, A), member(X, B), !, fail.
disjoint([_|R], B):- disjoint(R, B).


% %%% Interesting !!
% ?- powerset_rev([a,b], X).
% X = [[], [b], [a], [b, a]]

% powerset_rev(X, Y):- forlist(powerset_rev, X,  [[]], Y).

% powerset_rev(_,[], []).
% powerset_rev(X,[Y|Z],[Y,[X|Y]|U]):-powerset_rev(X, Z, U).

%
graded_powerset(X, P):-
	foldr(const(inverse(cons([]))), X, P0, [[[]]]),
	foldl(carp_up, X, P0, P).

graded_powerset(N, X, P):- repeat(N, cons([]), [[[]]], P0),
	foldl(carp_up, X, P0, P).

%
carp_up(X, P, Q):- maplist(maplist(cons(X)), P, [_|P0]),
       zip_append(P, P0, Q).

%
zip_append([], A, A).
zip_append(A, [], A).
zip_append([A|X], [B|Y], [C|Z]):- append(A, B, C), zip_append(X, Y, Z).

% % ?- trace, math:cartesian([a,b],[c,d], R).
% cartesian(Xs,Ys,Zs) :- prod(basic:ordered_pair, Xs, Ys, Zs).

% cartesian(P, Xs, Ys, Zs) :- prod(P, Xs, Ys, Zs).


% % ?- product(ordered_pair, [1,2], [a,b,c], X).
% :- meta_predicate product(:, ?, ?, ?).
% product(F, Xs, Ys, Zs):- prod(F, Xs, Ys, Zs, []).


% :- meta_predicate prod(:, ?, ?, ?).
% prod(F, X, Y, Z):- prod(F, X, Y, Z, []).

% prod(_, [], _, Z, Z).
% prod(F, [A|As], Y, Z, U):- prod_(F, A, Y, Z, V),
% 	prod(F, As, Y, V, U).

% prod_(_,_,[],Z,Z).
% prod_(F, A, [B|Bs], [C|Cs],U):- call(F, A, B, C),
% 	prod_(F, A, Bs, Cs, U).
%@
%@ true.
%@ math.pl qcompiled.

%@ Context module: math.
%@ true.

%  ?- pairs([a, b, c], R).
%@ R = [(a, b),  (a, c),  (b, c)] .

pairs(A, P):- pairs(A, P, []).

pairs([], X, X).
pairs([A|X], L, N):- foldl(pred(A, [B, [(A,B)|U], U]), X, L, M),
	pairs(X, M, N).

%%	choices(+X:list, -Y:list) is det.
%	Unify Y with the set of `choice sequences' of X.

%	choices/2 is much faster (3 time faster) than that by foldl/foldr
%	below.

% ?- choices([[1], [1]], X).
% ?- choices([[1]], X).
% ?- choices([[]], X).
% ?- choices([], X).
% ?- choices([[1,2],[a,b,c],[3,4]], R), maplist(writeln, R).

choices([], [[]]).
choices([A|As], X):- choices(As, X0),
	choices(A, X0, X, []).
%
choices([], _, Y, Y).
choices([A|As], Ch, X, Y):-
	cons_choice(Ch, A, X, X0),
	choices(As, Ch, X0, Y).
%
cons_choice([], _, X, X).
cons_choice([B|Bs], A, [[A|B]|X], Y):- cons_choice(Bs, A, X, Y).


% ?- product([], [1,2,3], X).
% ?- product([1],[a], X).
% ?- product([1],[], X).
% ?- product([1,2],[a,b], X).

product(X, Y, Z):- product(X, Y, Z, []).

product([], _, P, P).
product([X|R], Y, P, Q):- pairing(Y, X, P, P0),
	product(R, Y, P0, Q).
%
pairing([], _, P, P).
pairing([Y|Ys], X, [X-Y|P], Q):-pairing(Ys, X, P, Q).



% ?- trace, pairs([a,b,c], X).
% X = [ (a, b), (a, c), (b, c)]

% ?- mapset([a,b,c],[1,2,3,4], X), length(X, L).
% ?- mapset([],[1], X), length(X, L).
% ?- mapset([1],[], X), length(X, L).

mapset(Xs, Ys, Fs):- mapset(Xs, Ys, [[]], Fs).

mapset([], _, X, X):-!.
mapset([X|Xs], Ys, Gs, Fs) :- mapset_extend(Ys, X, Gs, [], Us),
	mapset(Xs, Ys, Us, Fs).

mapset_extend([], _, _, Fs, Fs).
mapset_extend([Y|Ys], X, Zs, Gs, Fs):-
	maplist(cons((X,Y)), Zs, Us),
	append(Us,Gs,Hs),
	mapset_extend(Ys, X, Zs, Hs, Fs).

% cons(X, Y, [X|Y]).

%% direct sum
% ?- direct_sum([a,b,c], [a,b,c], X).
direct_sum(X, Y, Z):- direct_sum(0, 1, X, Y, Z).

direct_sum(A, B, X, Y, Z):-maplist(ordered_pair(A), X, X1),
	maplist(ordered_pair(B), Y, Y1),
	append(X1, Y1, Z).

% ?- nlist([a,b], 3, X).
%@ X = [[], [a], [a, a], [a, a, a], [a, a, b], [a, b], [a, b, a], [a, b, b], [b], [b, a], [b, a, a], [b, a, b], [b, b], [b, b, a], [b, b, b]] .

nlist(A, N, L):- powerlist_sum(A, N, L0), append(L0, L1), sort(L1, L).

% ?- powerlist_sum([a,b],2,X).
% X = [[[a, a], [b, a], [a, b], [b, b]], [[a], [b]], [[]]]

powerlist_sum(A, N, L):- repeat(N, powerlist_cons(A), [[[]]], L).

powerlist_cons(A, [X0|Xs], [X, X0|Xs]):- product(cons, A, X0, X).

powerlist(A, N, L):- meta:repeat(N, product(cons, A), [[]], L).


% ?- N=50,  numlist(1, N, Ns), time(raise_list(Ns, 5,X)), length(X, L).
%@ % 318,878,060 inferences, 24.608 CPU in 28.858 seconds (85% CPU, 12958082 Lips)

%@ true.
%@ math.pl qcompiled.

% ?- N=3,  numlist(1, N, Ns), time(raise_list(Ns, 2, X)), length(X, L).

% ?- N=100,  numlist(1, N, Ns), time(raise_list(Ns, 3,X)), length(X, L).
raise_list(_, 0, [[]]):-!.
raise_list(L, N, P):- N0 is N-1,
	raise_list(L, N0, P0),
	diteribute_cons(L, P0, P, []).

% ?- distribute_cons([a,b], [[1, 2], [3, 4]], R).
distribute_cons(X, Y, P) :- diteribute_cons(X, Y, P, []).
%
diteribute_cons([], _, P, P).
diteribute_cons([A|As], Ls, P, P0):-
	diteribute_cons_aux(Ls, A, P, Q),
	diteribute_cons(As, Ls, Q, P0).
%
diteribute_cons_aux([], _, P, P).
diteribute_cons_aux([X|Xs], A, [[A|X]|P], Q):-
	diteribute_cons_aux(Xs, A, P, Q).

%
keta(N, R, C) :- keta(N, R, 0, C).

keta(N, R, C, C1):- N mod R =:= 0, !, C2 is C + 1, N1 is N// R,
	keta(N1, R, C2, C1).
keta(_, _, C, C).

append(&(X,Y), Z, [X,Y], [U,V], append(U,V,Z)).
append(X, X, [], [], true) :- listp(X).
append(X,[X],[], [], true).

bits(N, X) :- bits(N, X, []).

bits(0, X, X):-!.
bits(N, X, Y):- B is N mod 2, N0 is N>>1,
	bits(N0, X, [B|Y]).

boole_rule_OR_and_over_or(X, Y):-
	boole_rule(X, Y)
	;
	and_over_or(X, Y).

% for compatibility

conjunctive_normal_form--> boole_cnf, simple_cnf.

disjunctive_normal_form--> boole_dnf, simple_dnf.

% ?- boole_cnf(or(and(a,b), and(c,d)), X).
%@ X = [[a, c], [a, d], [b, c], [b, d]] .

% ?- boole_dnf(or(and(a,b), and(c,d)), X).

% ?- boole_cnf(not(or(and(a,b), and(c,d))), X).
%@ X = [[not(a),not(b)],[not(c),not(d)]] .
% ?- boole_cnf(not(or(not(and(a,b)), and(c,d))), X).
%@ X = [[a],[b],[not(c),not(d)]] .

% ?-  trace, boole_cnf(not(and(a,b)), X).
% ?- listing(boole_cnf).



boole_cnf(and(A,B),A1):-!,(boole_cnf(A,A2),boole_cnf(B,A3)),union(A2,A3,A1) .
boole_cnf(or(A,B),A1):-!,(boole_cnf(A,A2),boole_cnf(B,A3)),union_product(A2,A3,A1) .
boole_cnf(not(and(A,B)),A1):-!,boole_cnf(or(not(A),not(B)),A1) .
boole_cnf(not(or(A,B)),A1):-!,boole_cnf(and(not(A),not(B)),A1) .
boole_cnf(not(not(A)),A1):-!,boole_cnf(A,A1) .
boole_cnf(not(false),A1):-!,boole_cnf(true,A1) .
boole_cnf(not(true),A1):-!,boole_cnf(false,A1) .
boole_cnf(true,[]):-! .
boole_cnf(false,[[]]):-! .
boole_cnf(A,[[A]]):-! .

boole_dnf(X, Y):- boole_cnf(not(X), Z),
	maplist(maplist(pred([not(A),A]&[B,not(B)])), Z, Y).

% ?- boole_dnf(and(a,b), Y).
% ?- boole_cnf(and(a,b), Y).
% ?- boole_cnf(or(and(a,b), and(c, d)), Y).
% ?- boole_cnf(and(or(a,b), or(c, d)), Y).
% ?- boole_dnf(true, Y).
% ?- boole_dnf(false, Y).

% :- bekind(boole_dnf).
% or(A,B) = :union@A@B.
% and(A,B) =  :union_product@A@B.
% not(and(A,B)) = or(not(A), not(B)).
% not(or(A,B)) = and(not(A), not(B)).
% not(not(A)) = A.
% not(false) = true.
% not(true) = false.
% A  = `[[A]].
% :- ekind.

union_product([], _, []).
union_product([A|R], X, P):- maplist(union(A), X, Q),
	union_product(R, X, Q0),
	union(Q, Q0, P).

% ?- simple_cnf([[a],[b], [a], [not(c),not(d)]], X).
%@ X = [[a],[b],[not(c),not(d)]].

% ?- simple_cnf([[a],[b], [a], [not(c),not(d), c]], X).
%@ X = [[a],[b]].

simple_cnf(X, Y):- simple_cnf_one(X, X0), !, simple_cnf(X0, Y).
simple_cnf(X, Y):- maplist(sort, X, X0),
	sort(X0, Y).

simple_cnf_one(X, Y):- select(C, X, X0), cnf_rule(C, X0, Y).

cnf_rule([], X, [[]]):- X\==[].
cnf_rule(C, X, X):- memberchk(true, C).
cnf_rule(C, X, [C0|X]):- select(false, C, C0).
cnf_rule(C, X, X):- select(A, C, C0),
	complementary(A, B),
	memberchk(B, C0).

complementary(not(A), A):-!.
complementary(A, not(A)).

% ?- simple_dnf([[a],[b], [a], [not(c),not(d), c]], X).
%@ X = [[a],[b]].

simple_dnf(X, Y):- simple_dnf_one(X, X0), !, simple_dnf(X0, Y).
simple_dnf(X, Y):- maplist(sort, X, X0),
	sort(X0, Y).

simple_dnf_one(X, Y):- select(C, X, X0), dnf_rule(C, X0, Y).

dnf_rule([], X, X).
dnf_rule(C, X, X):- memberchk(false, C).
dnf_rule(C, X, [C0|X]):- select(true, C, C0).
dnf_rule(C, X, X):- select(A, C, C0),
	complementary(A, B),
	memberchk(B, C0).


flatten_boole(X,A1):-(functor(X,F,2),member(F,[or,and,&]),flatten(F,X,L0),maplist(flatten_boole,L0,L1),Y=..[F,L1]),!,flatten_boole(Y,A1) .
flatten_boole(-X,not(A1)):-!,flatten_boole(X,A1) .
flatten_boole(not(X),not(A1)):-!,flatten_boole(X,A1) .


boole_flat_paramod(and(L), and([X0|L0]), X, X0):- select(X, L, L0).
boole_flat_paramod(or(L), or([X0|L0]), X, X0):- select(X, L, L0).

% boole_rule/2
boole_rule(not(not(L)), L).
boole_rule(not(true), false).
boole_rule(not(false), true).
boole_rule(not(or(L)), and(L0)):- maplist(unary(not), L, L0).
boole_rule(not(and(L)), or(L0)):- maplist(unary(not), L, L0).
boole_rule(or([]), false).
boole_rule(or([X]), X).
boole_rule(or(L), true):- member(true, L).
boole_rule(or(L), or(L0)):- select(false, L, L0).
boole_rule(or(L), or(L0)):- sort(L, L0), L\==L0.
boole_rule(or(L), true):-  memq(not(X), L), memq(X, L).
boole_rule(or(L), or(L0)):- select(or(X), L, L1), append(X, L1, L0).
boole_rule(and([]), true).
boole_rule(and([X]), X).
boole_rule(and(L), false):- member(false, L).
boole_rule(and(L), and(L0)):- select(true, L, L0).
boole_rule(and(L), and(L0)):- sort(L, L0), L\==L0.
boole_rule(and(L), false):- memq(not(X), L), memq(X,L).
boole_rule(and(L), and(L0)):- select(and(X), L, L1), append(X, L1, L0).

and_over_or(and(L), or(L1)):- distribute(and, or, L, L1).

or_over_and(or(L), and(L1)):- distribute(or, and, L, L1).

% ?- term_rewrite:term_rewrite(desugar, ^, (a-b)^(-x), X).
% ?- reduce:term_rewrite(desugar, +, (a-b)^(-x), X).
% ?- reduce:term_rewrite(desugar, (^,[1]), (a-b)^(-x), X).
% ?- reduce:term_rewrite(desugar, (^,[2]), (a-b)^(-x), X).
% ?- reduce:subtree((+);(*);(/);(^,[2]), (a*b)^(2/3), X, Y, Z).

% rule for reduce
% ?- reduce:term_rewrite(ring:desugar, [], -(a-b), X).
%@ X = -1* (a+ -1*b).

desugar(X-Y, X+(-1)*Y).
desugar(+X, X).
desugar(-X, (-1)*X).
desugar(X/Y, rdiv(X, Y)).

% ?- scm(math).
% ?- ya_desugar(1, X).
% ?- nb_setval(context_module, math).
%
% ?- set(pow([1,2]), R).

ya_desugar(+X,A1):-!,ya_desugar(X,A1) .
ya_desugar(X-Y,A1+(-)*A2):-!,ya_desugar(X,A1),ya_desugar(Y,A2) .
ya_desugar(-X,-1*A1):-!,ya_desugar(X,A1) .
ya_desugar(X/Y,A1 rdiv A2):-!,ya_desugar(X,A1),ya_desugar(Y,A2) .
ya_desugar(X,X):-! .
prolog(A,true):-(call(A),!),! .
prolog(_A1,false):-! .
%
% ?- zip_hyphen([a,2], [3,4], X).
%@ Correct to: "pac_word:zip_hyphen([a,2],[3,4],X)"? yes
%@ X = [a-3, 2-4].

% ?- math:subst([1-true, 0-false], f(1,0,1), R).
%@ R = f(true, false, true).
subst_atomic(M, X, Y)	:- atomic(X), !,
					   ( key_val_check(M, X, Y) -> true
					   ;	Y = X
					   ).
subst_atomic(M, X, Y)	:-
        functor(X, F, N),
		functor(Y, F, N),
        subst_atomic_args(N, M, X, Y).
%
subst_atomic_args(0, _, _, _) :- !.
subst_atomic_args(J, M, X, Y) :-
        arg(J, X, A),
        arg(J, Y, B),
        subst_atomic(M, A, B),
        K is J-1,
        subst_atomic_args(K, M, X, Y).
%
key_val_check([A|As], X, Y) :-
	(	(A = (X-Y); A = (X,Y); A = (X:Y) ) -> true
	;	key_val_check(As, X, Y)
	).

%% subst(M,X,Y)   apply the substitution M to X to get Y.
subst(M, X, Y):- subst_atomic(M, X, Y).

%  ?- time(fibo(100000, N)).
%%%%
fibo(0, 1, _).
fibo(1, 1, 1).
fibo(N, F, F0):- N>1, N0 is N-1, fibo(N0, F0, G), F is F0+G.

fibo(N, F) :- fibo(N, F, _).

%% tail recursion version
%  ?- time(fibo_tr(100000, N)).

fibonacci(N, F):- fibo_tr(N, F).

fibo_tr(0, _, Y, Y).
fibo_tr(N, X, Y, Z):- N>0, N0 is N-1, Z0 is X + Y,
	fibo_tr(N0, Y, Z0, Z).

fibo_tr(0, 1):-!.
fibo_tr(1, 1):-!.
fibo_tr(N, F):- N > 1, N0 is N-1, fibo_tr(N0, 1, 1, F).

% ?-arrange([1,2,3],L).
% ?-arrange([1,2,3],[X,Y]).
arrange(_, []).
arrange(L, [X|R]):- select(X, L, L0),
	arrange(L0, R).

% %%% shuffling
% ?- select_random(A, [taro, jiro, hanako], X).
% A = hanako,
% X = [taro, jiro]
% ?- shuffle([taro,jiro,hanako], X).
% X = [hanako, taro, jiro]

shuffle(X, Y):- shuffle(X, [], Y).
shuffle([], A,  A):-!.
shuffle(L, A,  B):- length(L, N),
	N0 is random(N),
	select_nth(N0, X, L, L0),
	shuffle(L0, [X|A], B).

select_random(X, L, L0):- length(L, N),
	I is random(N),
	select_nth(I, X, L, L0).

select_nth(0, X, [X|L], L).
select_nth(N, X, [Y|L], [Y|L0]):- N>0, N0 is N-1,
	select_nth(N0, X, L, L0).

random_sublist(L, L0):- length(L, N), N0 is N+1,
	R is random(N0),
	length(L0, R),
	random_sublist_(L0, L).

random_sublist_([], _).
random_sublist_([X|Xs],  L):- select_random(X, L, L0),
	random_sublist_(Xs, L0).

%%%
modulo(M, X, V):-  V is X mod M.

%
g_base(lex, X, Y):- !, ensure_loaded('gb-lex'), gblex:gb(X, Y).
g_base(total, X, Y):-  ensure_loaded('gb-total'), gbtotal:gbase(X, Y).

% split(F, X, Y, [X0, X1], [Y0,  Y1], append(Y0, Y1, Y)):-
% 	functor(X, F, 2),
% 	arg(1, X, X0),
% 	arg(2, X, X1).
% split(_, X, [X], [], [], true).

%?- rel_to_fun([(a,b),(a,c),(b,d)], X).
%@ X = [b-[d], a-[c, b]].

rel_to_fun(X, Y):-rel_to_fun(X, [], Y).

%
rel_to_fun([], X, X).
rel_to_fun([P|R], X, Y):- (P=(A-B); P=(A,B)), !,
	(	select(A-M, X, X0)
	->	rel_to_fun(R, [A-[B|M]|X0], Y)
	;	rel_to_fun(R, [A-[B]|X], Y)
	).

% ?- rel_to_fun([(a,b),(a,c),(b,d)], X), fun_to_rel(X, Y).
%@ X = [b-[d], a-[c, b]],
%@ Y = [b-d, a-c, a-b].

fun_to_rel(X, Y):- fun_to_rel(X, Y, []).
%
fun_to_rel([], Y, Y).
fun_to_rel([A|X], Y, Z):-  (A = (U,L); A = (U-L)), !,
	fun_to_rel(L, U, Y, Y0),
	fun_to_rel(X, Y0, Z).
%
fun_to_rel([], _, Z, Z).
fun_to_rel([V|Vs], U, [U-V|Y], Z):-
	fun_to_rel(Vs, U, Y, Z).

% ?- all_maps([1,2],[1,2,3], Asg).
% ?- all_maps([1,2,3,4,5,6,7,8],[1,2,3], Asg), length(Asg, Len),
%	Len =:= 3^8.
all_maps([], _, [[]]).
all_maps([X|Xs], Range, Fs):- all_maps(Xs, Range, Gs),
	all_maps(Range, X, Gs, Fs, []).
%
all_maps([], _,  _, Fs, Fs).
all_maps([Y|Ys], X, Fs, Gs, Hs ):- extend_functions(Fs, X-Y, Gs, Ms),
	all_maps(Ys, X, Fs, Ms, Hs).
%
extend_functions([], _, Hs, Hs).
extend_functions([F|Fs], A, [[A|F]|Hs], Gs):-
	extend_functions(Fs, A, Hs, Gs).


%?- math:filter([a,b,c],[c], X).
%@ X = [[c],[c,a],[c,b],[c,a,b]].
%?- filter([c], [a,b,c],X).

filter(X, D, F):- subtract(X, D, Y),
	powerset(Y, P),
	maplist(pred(D, [X,Y] :- append(D,X,Y)),
		P, F).



% ?- principal_filter([a,b,c], a, X).
principal_filter(D, A, PF):- select(A, D, D0), !,
	powerset(D0,  PD),
	maplist(pred(A, [X, [A|X]]), PD, PF).
principal_filter(_, _, []).

% ?- principal_filter_set([a,b,c], [a], X).
%@ X = [[[a, c], [a, c, b], [a, b], [a]]].
% ?- principal_filter_set([a,b,c], [a,b], X).
%@ X = [[[a, c], [a, c, b], [a, b], [a]], [[b, c], [b, c, a], [b, a], [b]]].

principal_filter_set(D, As, PFs):-
	maplist(pred(D, [A, PF]:- principal_filter(D, A, PF)),
		As, PFs).

% ?- scm(math).

%?- inverse_image([(a,b),(b,c)], [a], M).
%?- inverse_image([(a,b),(b,c)], [c], M).

inverse_image(F, R, S):- inverse_image(F, R, [], S).

inverse_image([(A,B)|C], R, S, S0):- memberchk(B, R), !,
	inverse_image(C, R, [A|S], S0).
inverse_image([_|C], R, S, S0):- !, inverse_image(C, R, S, S0).
inverse_image([], _, S, S).


%?-  image([(a,1),(b,2)], [b,a], I).
image(F, D, I):- fun_image(F, D, [], I).

fun_image([(A,B)|C], D, I, I0):- memberchk(A, D), !,
	fun_image(C, D, [B|I], I0).
fun_image([_|C], D, I, I0):- !, fun_image(C, D, I, I0).
fun_image([], _, I, I).

%?- b_subset(X, [a,b]).

b_subset([], []).
b_subset(S, [A|R]) :- b_subset(S0, R),
	(	S = S0
	;	S = [A|S0]
	).

%
% ?- time(repeat(10000, eval(list::((set::pow(1..2))+(set::pow(1..2))), X))).
% % 890,003 inferences, 0.150 CPU in 0.157 seconds (96% CPU, 5933353 Lips)
% true.
% ?- time(repeat(10000, eval(ss, list::((set::pow(1..2))+(set::pow(1..2))), X))).
% % 760,003 inferences, 0.140 CPU in 0.134 seconds (105% CPU, 5428593 Lips)
% true.
% ?- time(repeat(100000, eval(ss, list::((set::pow(1..2))+(set::pow(1..2))), X))).
% % 7,600,003 inferences, 1.310 CPU in 1.311 seconds (100% CPU, 5801529 Lips)
% true.

% ?- set(pow([1]+[2]), V).
% V = [[],[1],[2],[1,2]].
% ?- eval(set::pow([]), X).
% ?- set(cup([[1,2],[3,4]]), X).
% X = [1,2,3,4]
% ?- set(cap([[1,2],[2,3]]), X).
% X = [2]
% ?-set(cup(pow([1,2,3])), X).
% X = [1,2,3]
% ?- set(cap(pow([1,2,3])), X).
% X = []
% ?- set(cup(map([pow([1,2]), pow([a,b])])), R).
% R = [[],[1],[1,2],[2],[a],[a,b],[b]]

bigcap([A|B], C):- bigcap(B, A, C).

bigcap(_, [], []):-!.
bigcap([], A, A):-!.
bigcap([X|Y],A,B):- intersection(X, A, A0),
	bigcap(Y, A0, B).

%
poly_in(Ps, X, Y, A, B, Q):- member(P, Ps), call(P, X, Y, A, B, Q), !.
poly_in(_, X, Y, [], [], true):- poly_in(X, Y).

% ?- poly_in((a-b)^(1+3), R).
% R = (a+ -1*b)^4
% Converting polynomials into internal canonical forms.

poly_in(-X,A1):-!,poly_in(-1*X,A1) .
poly_in(X-Y,A1):-!,poly_in(X+ -1*Y,A1) .
poly_in(X+Y,A1+A2):-!,poly_in(X,A1),poly_in(Y,A2) .
poly_in(X*Y,A1*A2):-!,poly_in(X,A1),poly_in(Y,A2) .
poly_in(X/Y,A1/A2):-!,poly_in(X,A1),poly_in(Y,A2) .
poly_in(X^N,A1^N0):-N0 is N,!,poly_in(X,A1) .
poly_in($(X),X):-! .
poly_in(X mod N,A1 mod N0):-N0 is N,!,poly_in(X,A1) .

:-meta_predicate each(1,2,?,?) .

each(_A1,_A2,X,X):-(\+compound(X),!),! .
each(T,D,X,A1):-(call(T,X),!,X=..[F|Xs]),!,(maplist(each(T,D),Xs,A2),rev(=..,F.A2,A3)),call(D,A3,A1) .
each(A1,A2,X,A3):-X=..[F|Xs],!,maplist(each(A1,A2),Xs,A4),rev(=..,F.A4,A3) .
ari_val(_A1,A,A):-number(A),! .
ari_val(Assoc,A,B):-memberchk(A=B,Assoc),! .
ari_val(A1,A*B,A2):-!,(ari_val(A1,A,A3),ari_val(A1,B,A4)),rev(is,A3*A4,A2) .
ari_val(A1,A+B,A2):-!,(ari_val(A1,A,A3),ari_val(A1,B,A4)),rev(is,A3+A4,A2) .
ari_val(A1,A-B,A2):-!,(ari_val(A1,A,A3),ari_val(A1,B,A4)),rev(is,A3-A4,A2) .
ari_val(A1,A^B,A2):-!,(ari_val(A1,A,A3),ari_val(A1,B,A4)),rev(is,A3^A4,A2) .
ari_val(A1,-B,A2):-!,ari_val(A1,B,A3),rev(is,-A3,A2) .
ari_val(A1,+B,A2):-!,ari_val(A1,B,A3),rev(is,+A3,A2) .
si(E,Z):-Z is E .

polish_codes(com(F,A),A1):-!,(polish_codes(A,A2),maplist(polish_codes,A2,A3)),phrase(F,A3,A1) .
polish_codes([A1|A2],A3):-!,maplist(polish_codes,[A1|A2],A3) .

% combinations
% ?- scm(ring).
% ?- combi(4,3, X).
%@ X = 4 .
% ?- combi(20,10, X).
% ?- fact(20, X).

% nCr(N, R, C) <=>
% Let X, R be a set  s.t. |X| = N, a number R, respectively,
% then C =  |{Y in powerset(X) | |Y| = R}|

% ?- time(math:nCr(200,100, R)).
%@ % 713 inferences, 0.000 CPU in 0.000 seconds (85% CPU, 4512658 Lips)
%@ R = 90548514656103281165404177077484163874504589675413336841320 .
% ?- time(math:nCr(10000,5000, R)).

combi(N, R, C) :- nCr(N, R, C).

nCr(N, R, C):-  N >= 0, R>=0, NR is N - R, NR>=0, !,
	(  R < NR -> combination(N, R, NR, C)
	;  combination(N, NR, R,C)
	).
nCr(_,_,0).

combination(N, R, S, C):- S1 is S+1,
	nat_prod(S1, N, 1, M),
	fact(R, FR),
	C is M rdiv FR.

nat_prod(K, N, X, X):-  K>N, !.
nat_prod(K, N, X, Y):-  X0 is K*X, K1 is K+1,
	nat_prod(K1, N, X0, Y).

%
fact(N, F):- N >= 0, factorial(N, 1, F).
%
factorial(N, F):- N >= 0, factorial(N, 1, F).
%
factorial(0, X, X).
factorial(N, X, Y):- N > 0, X0 is N*X,
	succ(N1, N),
	factorial(N1, X0, Y).

% ?- binomial(3,X).
% ?- binomial(10,X).
% ?- binomial(100,X).

% ?- binomial(3,X).
binomial(N, E) :-
 	do(for(0..N, pred(N, [J, B, [C|B]]:-combi(N, J, C))), [], E).


% ?- toy_foreach(between(1,5,J), member(J, L), [J], [L]).
%@ L = [1, 2, 3, 4, 5|_] .

toy_foreach(A, B, Vs, Ws):-
	findall(Vs, A, Sol),
	copy_of_goal(Sol, Vs, Ws, B, Cs),
	maplist(call, Cs).

% ?- yaforeach(between(1,5,J), member(J, L), [J], [L]).
yaforeach(A, B, Vs, Ws):-
	findall(Vs, A, Sols),
	maplist(pred([Vs, Ws, B],
				 ([A]:- copy_term(Vs+Ws+B, A+Ws+G), call(G))),
			 Sols).

%
copy_of_goal([], _, _, _, []).
copy_of_goal([A|As], Vs, Ws, B, [G|Gs]):-
	copy_term(Vs+Ws+B, A+Ws+G),
	copy_of_goal(As, Vs, Ws, B, Gs).


toy_foreach_numbervars(A, B, Vs, Ws):- numbervars(Vs),
	findall(B, A, Bs),
	hzip(Vs, Ws, Zip),
	subst_numbervars(Zip, Bs, Cs),
	maplist(call, Cs).

%
hzip([],[],[]).
hzip([X|Xs],[Y|Ys],[X-Y|Zs]):-hzip(Xs, Ys, Zs).
%
subst_numbervars(_, X, Y):- (var(X); atomic(X)), !, Y = X.
subst_numbervars(M, '$VAR'(I), A):-!, memberchk('$VAR'(I)-A, M).
subst_numbervars(M, X, Y):- functor(X, F, N), functor(Y, F, N),
	subst_numbervars(N, M, X, Y).
%
subst_numbervars(0, _, _, _):-!.
subst_numbervars(J, M, X, Y):- I is J - 1,
	arg(J, X, A),
	arg(J, Y, B),
	subst_numbervars(M, A, B),
	subst_numbervars(I, M, X, Y).

% ?- string_to_ground_term("f(X, Y)", T).s
string_to_ground_term(X, Y):-
	term_string(Y, X, [variable_names(Eqs)]),
	maplist(call, Eqs).

% ?- egcd(3, 5, X, Y).
% ?- findall((I,J),
%	(between(-100, 100, I), between(-100, 100, J), 3*I + 5*J =:= 1),  X),
%	length(X, L), maplist(writeln, X).

% ?- egcd(377, 493, N, M), G is N*377+M*493.
% ?- X is gcd(377,493).

% egcd(+Integer, +Integer, -Integer, -Integer)  J.Burse
egcd(_, 0, 1, 0) :- !.
egcd(A, B, X, Y) :-
    divmod(A, B, Q, R),
    egcd(B, R, S, X),
    Y is S - Q*X.
