:- module(matrix, [matrix_calc/3, matrix/2]).
:- use_module(library(lists)).
:- use_module(util(html)).
:- use_module(util(misc)).
:- use_module(util('meta2')).
:- use_module(util('gb-lex-old')).
:- use_module(pac(basic)).
:- use_module(pac(reduce)).

term_expansion --> pac:expand_pac.

:- use_module(pac('expand-pac')).
:- use_module(pac(op)).

equation_separator(;).
:- use_module(pac(basic)).

% ?- qcompile(util(matrix)), module(matrix).
% ?- matrix:power_mod_table(3, X).
% ?- matrix:power_mod_table(3, X).
% ?- matrix:eqs_matrix((x+y=1;x-y=1), A, B, C).
% ?- matrix:eqs_matrix((x=1), A, B, C).
% ?- gblex:term_poly(x-1, _G3008).

% ?- trace, matrix:matrix_calc([[[1, 2, 1], [1, 3, 1], [-2, -1, 0]], [[1, 2, 1], [1, 3, 1], [-2, -1, 0]], [[1, 0, 0], [0, 1, 0], [0, 0, 1]]], x+y, _3664).

algebra(X, Y, Z) --> matrix(X, Y, Z).

is_matrix([_|_]).

%%%
subst_rdiv(X, Y) :- subst_functor([((rdiv/2), (/))], X, Y).

% ?- subst_functor([(f/1,  g)], f(g(f(1))), R).
% R = g(g(g(1))).

subst_functor(_, X, X):- (number(X); var(X)), !.
subst_functor(S, X, Y):- X =..[F|A],
	maplist(subst_functor(S), A, B),
	functor(X, F, N),
	(  memberchk((F/N, G), S)
	->	H = G
	;	H = F
	),
	Y =..[H|B].

% ?- disable_pac_query.
% ?- enable_pac_query.

% ?- qcompile(util(matrix)).
% ?- module(matrix).
% ?- matrix(inverse(square([1,2,3])),  R).
% ?- matrix(square([1,2,3]),  R).
% ?- matrix(:square@[1,2,3],  R).
%
% ?- eval(matrix:: (inverse @ square([1,2,3])),  R).
% ?- eval((matrix:matrix):: (inverse @ square([1,2,3])),  R).
% ?- eval((:matrix):: (inverse @ square([1,2,3])),  R). % << Error.
% ?- eval(matrix@inverse@square@[1,2,3],  R).
% ?- eval(matrix@(inverse@(square@[1,2,3])),  R).
% ?- eval(matrix@(inverse@(square@[1,2,3])),  R), call(R, V).
% ?- eval((user:matrix):: inverse@(square@[1,2,3]),  R).  % << Error.
% ?- eval(matrix:matrix:: unit(3), X).
% ?- eval(matrix:matrix:: (scalar(2+3)@unit(2)), X).
% ?- eval(matrix:matrix:: (unit(2) + scalar(2, unit(2))), X).
% ?- eval(matrix:matrix:: (unit(2) - scalar(2, unit(2))), X).
% ?- eval(matrix:matrix::(unit@ (:length([1,2]))), R).
% ?- eval(matrix:matrix::(unit(3) * unit(3)), X).
% ?- eval(matrix:matrix::(unit(3)+unit(3)), X).
% ?- eval((matrix:matrix)::scalar(1+2, unit(3)+unit(3)), X).
% ?- eval((matrix:matrix)::diag(square(vector(3, 1))), X).

:- bekind(matrix, []).
square(L)	= :vector_to_square@L.
diag(M)		= :diagonal@M.
unit(N)		= :unit_matrix@N0 :- N0 is N.
inverse(M)	= xargs([M0]>>M0^(-1))@M.
vector(N, A)	= V	:- length(V, N), maplist(=(A), V).
t(A)		= :transpose_matrix@A.
tr(A)		= :matrix_trace@A.
-A		 = scalar(-1)@A.
A - B		= A + scalar(-1, B).
A + B		= :add_m_m@ A@B.
A * B		= :mul_m_m@ A@B.
s_add(A,B)	= :add_s_m(A0)@B :- A0 is A.
s_mull(A,B)	= :mul_s_m(A0)@B :- A0 is A.
A^N		= :exp_m(N0)@ A  :-  N0 is N.
max(A, B)	=  `C :-  ( A< B -> C=B; C=A).
max(A, B)	=  `C :-  ( A< B -> C=A; C=B).
scalar(A, X)	= :mul_s_m(A0)@ X :- A0 is A.
subst(X, P)	= :poly_subst@X@ (math:poly(P)).
X = `X.
:- ekind.

% ?- matrix_calc([[[1,2,1],[1,3,1],[-2,-1,0]],[[1,2,1],[1,3,1],[-2,-1,0]],[[1,0,0],[0,1,0],[0,0,1]]], x+y, R).

:- bekind(matrix_calc//1,[]).
x		= with([A|_], `A).
y		= with([_,A|_], `A).
z		= with([_,_,A|_], `A).
x(K)		= with(L, :nth1(K, L)).
square(L)	= :vector_to_square@L.
diag(M)		= :diagonal@M.
unit(N)		= :unit_matrix@N0 :- N0 is N.
inverse(M)	= xargs([M0] >> M0^(-1))@M.
vector(N, A)	= V	:- length(V, N),
 			   maplist(=(A), V).
t(A)		= :transpose_matrix@A.
tr(A)		= :matrix_trace@A.
-A		= scalar(-1)@A.
A - B		= A + scalar(-1, B).
A + B		= :add_m_m@ A@B.
A * B		= :mul_m_m@ A@B.
s_add(A,B)	= :add_s_m(A0)@B :- A0 is A.
s_mull(A,B)	= :mul_s_m(A0)@B :- A0 is A.
A^N		= :exp_m(N0)@ A  :-  N0 is N.
max(A, B)	= `C :-  ( A< B -> C=B; C=A).
max(A, B)	= `C :-  ( A< B -> C=A; C=B).
scalar(A, X)	= :mul_s_m(A0)@ X :- A0 is A.
subst(X, P)	= :poly_subst@ X @ (math:poly(P)).
X = `X.
:- ekind.

%%	transpose_matrix(?X:list_of_list, _Y:list_of_list) is det.
%	Transpose_Matrix as matrix tranposing. (Bi-directional.)

% ?- transpose_matrix([[1,2],[3,4]], R), transpose_matrix(S, R).
transpose_matrix(X, Y) :- zip(X, Y).

% ?- zip([[a,b,c],[1,2,3]], R), zip(R, S).
% ?- zip(R, [[a,b,c],[1,2,3]]), zip(S, R).

zip([], []):- !.
zip(X, Y):- nonvar(Y), !, zip_(Y, X).
zip(X, Y):- zip_(X, Y).
%
zip_([[]|_], []):- !.
zip_(X, [Hs|Z]):- heads(X, Hs, Y), zip_(Y, Z).
%
heads([[X|Xs]|R], [X|S], [Xs|U]) :- heads(R, S, U).
heads([], [], []).

% ?- poly_subst([[1]], [1,1], X).
% ?- maplist(maplist(pred([_, 0])), [[1,2],[3,4]], Z0).
% ?- poly_subst([[1]], [1,2], X).
poly_subst(X, Y, Z):- maplist(maplist(pred([_,0])), X, Z0),
	poly_subst(X, Y, Z0, Z).

poly_subst(_, [], Z, Z).
poly_subst(X, [A|R], U, Z):- mul_m_m(X, U, U0),
	maplist(maplist(pred(A, [_,A])), X, C),
	add_m_m(C, U0, V),
	poly_subst(X, R, V, Z).

% ?- unit_matrix(3, M).
%@ M = [[1,0,0],[0,1,0],[0,0,1]] .
% ?- unit_matrix(5, 3, M).
%@ M = [[5,0,0],[0,5,0],[0,0,5]] .

unit_matrix(N, M):- unit_matrix(1, N, M).

unit_matrix(_, 0, []).
unit_matrix(K, N, [[K|A]|X1]):-
	N>0, N0 is N-1,
	unit_matrix(K, N0, X0),
	maplist(pred([C, [0|C]]), X0, X1),
	length(A, N0),
	maplist(=(0), A).

generic(F, F0, X, Y, Z):-
	rational(X)-> call(F0, X, Y, Z); call(F, X, Y, Z).

%%%% polynomial over matrix
% ?- eval(:subst(scalar(3, unit(3))^(-1), [1,2,3]), X).
% ?- eval(:subst(scalar(3, unit(3))^(-1), poly(3, [0,1])), X).
% X = [[1 rdiv 81, 0, 0], [0, 1 rdiv 81, 0], [0, 0, 1 rdiv 81]].

% matrix_polynomial(Q, X, Y):- (Q = poly(N, As) -> true; N=0, As = Q),
% 	reverse(As, As0), matrix_poly(As0, X, P), eval(:P*(X^N), Y).

% matrix_poly(As, X, Y):- length(X, D), eval(:unit(D), Unit),
% 	eval(:scalar(0, Unit), Zero),
% 	matrix_poly(As, X, Unit, Zero, Y).

matrix_poly([], _, _, X, X).
matrix_poly([A|As], M, U, X, Y):- matrix_poly(As, M, U, M*X + scalar(A, U), Y).

:- meta_predicate product_table(3, ?, ?, ?).

% ?- product_table(plus, [1,2,3], [4,5,6], R).
%@ R = [[5, 6, 7], [6, 7, 8], [7, 8, 9]].
% ?- product_table(pred([X, Y, Z]:- Z is mod(X*Y, 3)), [1,2,3], [4,5,6], R).
%@ R = [[1, 2, 0], [2, 1, 0], [0, 0, 0]] .

product_table(Fun, Vec1, Vec2, Matrix):- product_table_(Vec1, Vec2, Matrix, Fun).
%
product_table_([], _, [], _):-!.
product_table_([A|As], Vec, [Va|Out], Fun):-
	product_table_(As, Vec, Out, Fun),
	maplist(pred([A,Fun], [X, Y]:- call(Fun, A, X, Y)), Vec, Va).
%
product_table_([], _, [], _):-!.
product_table_([A|As], Vec, [Va|Out], Fun):-
	product_table_(As, Vec, Out, Fun),
	maplist(pred(A,[X, Y]:- call(Fun, A, X, Y)), Vec, Va).

%
mult(P, X, Y, Z):-  Z is (X * Y) mod P.

add(P, X, Y, Z):-  Z is (X + Y) mod P.

%%%%%%%%%%%%%%%
% ?- matrix:latex_prod_table(mult(5), 5, Res).
% ?- matrix:latex_prod_table_list([3,4,5], X).

latex_prod_table_list(L,Y) :-
   maplist( pred([X, mult(X)]),  L, As),
   maplist(latex_prod_table, As, L, X),
   maplist(eh:wrap("$$", "$$\n"), X, Y).
%
latex_prod_table(A, N, Res):- N1 is N-1,
	numlist(0, N1, I),
    product_table(A, I, I, X),
	array_tex(I, atom_codes, X, Res).

array_tex(I, F) --> attach_index(I),
    peek([UpperIndex|R],R),
	{length(I, N)},
	{tabular_preamble0(N, PreAmb0)},
	{PreAmb=["c|"|PreAmb0]},
        maplist(maplist(F)),
        maplist(nsert(" & ")),
        insert("\\\\\n"),
	{maplist(atom_codes, UpperIndex, UpperIndex0)},
	{insert(" & ", UpperIndex0, UpperIndex1)},
	peek(A, ["\\begin{tabular}{", PreAmb, "}\n",
	        UpperIndex1, "\\\\\\hline\n",
	        A, "\n",
		"\\end{tabular}\n"]).

attach_index(I) --> cons(I),
	transpose_matrix,
	cons([' '|I]),
	transpose_matrix.

maxnum([],0).
maxnum([X|R],M):- maxnum(R,M1), M is max(X,M1).

tabular_preamble(0,"|"):-!.
tabular_preamble(N,["|c"|P]):- N1 is N-1,  tabular_preamble(N1,P).

tabular_preamble0(0,""):-!.
tabular_preamble0(N,["c"|P]):- N>0, N1 is N-1,  tabular_preamble0(N1,P).

%
power_mod_table(P, M):-
	P1 is P-1,
	numlist(1, P1, Xs),
	numlist(1, P1, Ys),
	html_table(power_mod(P), Xs, Ys, M).

%
% ?- qcompile(util(matrix)), module(matrix).


% ?- power_mod(3, 2, 2, R).
power_mod(M, X, Y, R):-   R is powm(X, Y, M).

%%%%%% power_mod_matrix:  (not necessary, just for fun)
% ?- power_mod_matrix(7,X).

% power_mod_matrix(P, Rs):- P1 is P-1,
% 	numlist(1, P1, Ns),
% 	numlist(2, P1, As),
% 	foldl(power_mod_matrix(P,Ns), As, Rs, []).

% power_mod_matrix(P, Ns, A, [Col|Rs], Rs):-
% 	maplist(\(N,  \(R, {R is powm(A, N, P)})), Ns, Col).

% ?- matrix:html_prod_table(3, X).
html_prod_table(N, Table) :- N1 is N-1,
	numlist(0, N1, I),
	html_table(mult(N), I, I, Table).

html_prod_table(N, _, Table) :- html_prod_table(N, Table).

:- meta_predicate html_table(3, ?, ?, ?).
html_table(Fun, Col, Row, Table) :-
	product_table(Fun, Col, Row, Table0),
	html_table_(Col, Row, Table0, Table).
%								%
html_table_(X, Y):-
	array(X, [A|X0]),
	cars(X0, B, X1),
	html_table_(A, B, X1, Y).
%
html_table_(A, B, X, Y):-
	maplist(tag(th), [[]|A], Ath),
	maplist(tag(th), B, Bth),
	maplist(maplist(tag(td)), X, Xtd),
	cars(X1, Bth, Xtd),
	maplist(tag_nl(tr), [Ath|X1], X2),
	tag_nl(table, X2, Y).


%%%%  convert text to html table.
array --> array("\s\s").

array(Code) --> split,
	maplist(split(plus, Code)),
	maplist(remove([])),
	remove([]).

html_table --> array,
	maplist(phrase((maplist(tag(td)), tag_nl(tr)))),
	tag_nl(table).

html_table_th(X, Y):- array(X, [A|X0]),
	maplist(tag(th), A, A0),
	maplist(maplist(tag(td)), X0, X1),
	maplist(tag_nl(tr), [A0|X1], X2),
	tag_nl(table, X2, Y).

%%%%%%%
% ?- foldl(matrix, matrixinverse),"0  1\n1  0\n\n", A).
% ?- eval(matrix, ps(ap), [], (matrix, matrixinverse), ["0  2\n1  0\n\n"], X).

% ?- qcompile(util(matrix)), module(matrix).

% ?- matrix(=, `1  2\n`, X).
%@ X = [[1], [2]] .
% ?- matrix(=, `1  2\n3  4\n`, X).
%@ X = [[1, 3], [2, 4]] .

parse --> matrix.

parse(X) --> matrix(X).

matrix(X) --> array,
	maplist(maplist(text_number)),
%	transpose_matrix,
	phrase(X).
%	transpose_matrix.

% text_number --> herbrand, flip(is).
text_number --> herbrand, pred([X,Y]:- Y is X).

namex --> text_number.


%%%%%%%%%% nametest for matrix operations %%%%%%%%%%%
%     1 2 1
%     1 3 1
%     -2 -1 0
% ?- M=[[1,1,-2],[2,3,-1],[1,1,0]], (matrixinverse(M,Y,X),mul_m_m(M,X,E),mul_m_m(X,M,E)).
% ?- matrixinverse([[1,0],[0,1]], _, X).

%%%% Some Tiny %%%
add(X, Y, Z):- Z is X + Y.

reverse([],X,X).
reverse([A|B],X,Y):- reverse(B, [A|X],Y).

zeroclear(X,Y):- maplist(peek(0),X,Y).

%
inverse --> matrixinverse.

matrixinverse(M,Y):- matrixinverse(M,X,Y), X=[].

matrixinverse(M,X,Y):-
	transpose_matrix(M,M1),
	maplist(peek([]), M1, A),
	exp_m(0,M1,E),
	matrixinverse_loop(m([],M1), n([],[],A,E), P,Q),
	matrixout(P,X),
	matrixout(Q,Y).

matrixout(c(A,B), C):- reverse(A,B,C).
matrixout(v(A,B), C):- reverse(A,B,C).
matrixout(m(X,Y), Z):- reverse(X,Y,Z1), transpose_matrix(Z1,Z).
matrixout(n(X,Y,Z,U), V):- maplist(reverse, X,Y,Y1), maplist(reverse,Z,U,U1),
	reverse(Y1,U1,U2), transpose_matrix(U2,V).

matrixinverse_loop(M,N,M1,N1):- pivot(M,N,M2,N2), !,
	pivotnormalize(M2,N2,M3,N3),
	pivotsweep(M3,N3,M4,N4),
	matrixinverse_next(M4,N4,M5,N5),
	matrixinverse_loop(M5,N5,M1,N1).
matrixinverse_loop(M,N,M,N).

matrixinverse_next(m(X,Y), n(A,B,C,D), m(X1,Y1), n(A1,B1,C1,D1)):-
	maplist(cdr, X, X2),
	maplist(cdr, Y, Y2),
	matrixinverse_next(m(X2, Y2), m(X1, Y1)),
	maplist(car, B, B2),
	maplist(cdr, B, B3),
	maplist(car, D, D2),
	maplist(cdr, D, D3),
	maplist(cons, B2, A, A2),
	maplist(cons, D2, C, C2),
	matrixinverse_next(n(A2,B3,C2,D3), n(A1,B1,C1,D1)).

matrixinverse_next(m(X,[Y|Y1]),m([Y|X],Y1)).
matrixinverse_next(n(A,B,[C1|C2],[D1|D2]), n([C1|A],[D1|B],C2,D2)).

pivot(X,Y,Z,U):- pivotrow(X,Y,Z,U), !.
pivot(X,Y,Z,U):- pivotcolumn(X,Y,Z1,U1), !,  pivot(Z1,U1,Z,U).

pivotrow(M, N, M, N):- M= m(_,[[A|_]|_]), A=\=0,!.
pivotrow(m(X,[Y|Z]), n(A,B,[C|D],[E|F]), m(X1, [Z1,Y|Z2]), n(A1,B1,[D1,C|D2], [F1,E|F2])):- pivotrow(m(X,Z), n(A,B,D,F), m(X1,[Z1|Z2]), n(A1,B1, [D1|D2],[F1|F2])).

pivotcolumn(m(X,Y), n(A,B,C,D), m(X1,Y1), n(A1,B1,C1,D1)):-
	iszerocars(Y),
	!,
	maplist(car,X,Xcar),
	maplist(cdr,X,Xcdr),
	maplist(car,Y,Ycar),
	maplist(cdr,Y,Ycdr),
        maplist(car,B,Bcar),
	maplist(cdr,B,Bcdr),
	maplist(car,D,Dcar),
	maplist(cdr,D,Dcdr),
	pivotcolumn(m(Xcdr,Ycdr), n(A,Bcdr,C,Dcdr), m(X2,Y2), n(A1,B2,C1,D2)),
	insertaftercar(Xcar,X2,X1),
	insertaftercar(Ycar,Y2,Y1),
	insertaftercar(Bcar,B2,B1),
	insertaftercar(Dcar,D2,D1).
pivotcolumn(X,Y,X,Y).

iszerocars([]).
iszerocars([[0|_]|X]):- iszerocars(X).

insertaftercar([],[],[]).
insertaftercar([X|X1],[[Y|Y1]|Y2],[[Y,X|Y1]|Y3]):-
	   insertaftercar(X1,Y2,Y3).

pivotnormalize(m(X,[[Y|Z]|U]), n(A,B,[C|D],[E|F]),
	       m(X,[[1|Z1]|U]), n(A,B,[C1|D],[E1|F])):-
	Q is 1 rdiv Y,
	mul_v_s(Z,Q,Z1),
	mul_v_s(C,Q,C1),
	mul_v_s(E,Q,E1).

pivotsweep(m(X,[Y|Z]),n(A,B,[C|D],[E|F]), m(X1,[Y|Z1]),
		   n(A1,B1,[C|D1],[E|F1])):-
	maplist(car, X, Xcar),
	maplist(sweep(Y), Xcar, X, X1),
	maplist(sweep(C), Xcar, A, A1),
	maplist(sweep(E), Xcar, B, B1),
	maplist(car, Z, Zcar),
	maplist(sweep(Y), Zcar, Z, Z1),
	maplist(sweep(C), Zcar, D, D1),
	maplist(sweep(E), Zcar, F, F1).

sweep(Y, A, V, W):- A1 is -A, mul_v_s(Y, A1, Y1), add_v_v(Y1, V, W).

sweep1(Y, A, V, W):- A1 is -A, W is A1*Y+V.

firstcolumn(X) --> zip_step(X).

% 列ベクトルを対角行列に変換する．
% [1,2,3]  => [[1, 0, 0], [0, 2, 0], [0, 0, 3]]

% ?- matrix:vector_to_square([1,2,3], Y).
% ?- trace, matrix:matrix(square([1,2,3]), R).

vector_to_square([],[]).
vector_to_square([X|Y], [[X|Zeros]|Z]):- vector_to_square(Y,Y1),
	maplist(cons(0), Y1, Z),
	maplist(peek(0), Y, Zeros).

% ?- vector_to_square([1,2,3], X), diagonal(X, Y).

diagonal([],[]).
diagonal([[X|_]|Vs], [X|Xs]):- maplist(pred([[_|A],A]), Vs, Ws),
	diagonal(Ws, Xs).

matrix_trace(X, T):- diagonal(X, D), foldl(pred([X,Y,Z]:- Z is X+Y), D, 0, T).

add_s_m(_, [], []).
add_s_m(A, [[X|Xs]|Y], [[X0|Xs]|Y0]):-
	X0 is X+A,
	zip_cons(Z, Z0, Y),
	add_s_m(A, Z0, Z1),
	zip_cons(Z, Z1, Y0).

zip_cons(X,Y,Z):- maplist(pred([A,B,[A|B]]), X,Y,Z).

% 列ベクトルの足し算
% vector addition; add_v_v(X,Y,Z)   Z = X+Y.

add_v_v(X,Y,Z):- maplist(add,X,Y,Z).

% 行列の足し算
% matrix addition;add_m_m(X,Y,Z)   Z = X+Y. (matrix addition)


add_m_m(X,Y,Z):- number(X), number(Y), !,
	Z is X + Y.
add_m_m(X,Y,Z):- number(X), !,
	length(Y, N),
	unit_matrix(X, N, X0),
	add_m_m(X0, Y, Z).
add_m_m(X,Y,Z):- number(Y), !,
	length(X, N),
	unit_matrix(Y, N, Y0),
	add_m_m(X, Y0, Z).
add_m_m(X,Y,Z):- maplist(maplist(add),X,Y,Z).

% 列ベクトルの内積
% inner product

innerproduct(X,Y,Z):- innerproduct(X,Y,0,Z).

innerproduct([],[],S,S).
innerproduct([X|X1],[Y|Y1], S0, S) :-
	S1 is S0 + (X * Y),
	innerproduct(X1, Y1, S1, S).

% 列ベクトルの左から行列をかける．
% linear action on vector; mul_m_v(X,Y,Z)  Z= XY   (X: matrix, Y: vector)
mul_m_v(X,Y,Z):- transpose_matrix(X,X1), maplist(innerproduct(Y),X1,Z).

% 行列を右からスカラー倍する．
% scalar multiply; mul_v_s(X,A,Y)   Y=XA  (X: vector, A: scalar)

mul_v_s(X, A, Y) :-  maplist(mul_v_s_aux(A), X, Y).

mul_s_v(S, V, V0) :- mul_v_s(V, S, V0).

mul_v_s_aux(A, X, Y):- Y is  X*A.

% 行列のかけ算
% matrix multipy; mul_m_m(X,Y,Z)   Z=XY  (X:matrix Y:matrix)

mul_m_m(X,Y,Z):- number(X), number(Y), !, Z is X*Y.
mul_m_m(X,Y,Z):- number(X), !,
	length(Y, N),
	unit_matrix(X, N, X0),
	mul_m_m(X0, Y, Z).
mul_m_m(X,Y,Z):- number(Y), !,
	length(X, N),
	unit_matrix(Y, N, Y0),
	mul_m_m(X, Y0, Z).
mul_m_m(X,Y,Z):- transpose_matrix(X,X1), maplist(dual_innerproduct(X1), Y, Z).

%
dual_innerproduct(X, Y, Z):-  maplist(innerproduct(Y), X, Z).

%  行列のスカラー倍
mul_s_m(S) --> maplist(mul_s_v(S)).

% 行列のべき乗

exp_m(E, M, M0) :- E >=0, !, exp_m_(E, M, M0).
exp_m(E, M, M0) :- matrixinverse(M, M1), E0 is -E, exp_m_(E0, M1, M0).

exp_m_(0, M, Id):- !, maplist(peek(1),M,Ones), vector_to_square(Ones,Id).
exp_m_(1, M, M) :- !.
exp_m_(N, M, X) :- N1 is N // 2, exp_m_(N1,M,A),  mul_m_m(A,A,A1),
	(N mod 2 =:= 0  ->  X=A1;   mul_m_m(M,A1,X)).

% exp_m(0, M, Id):- !, maplist(set(1),M,Ones), vector_to_square(Ones,Id).
% exp_m(N, M, X):- N1 is N // 2, exp_m(N1,M,A),  mul_m_m(A,A,A1),
% 	(N mod 2 =:= 0  ->  X=A1;   mul_m_m(M,A1,X)).

% ?- with_order([y,x], gauss, (x=1);(y=2), S).
% S = ([x, y]=[1, 2])
% ?-  gauss((x = 1); (y=2), X).
% X = ([y, x]=[2, 1])
%  solving systems of linear equations
% ?- gauss(x = 1, X).
% X = ([x]=[1])
% ?- gauss([[0, 1], [1, 0]]*[x, y]= [-1, 1], X).
% X = ([x, y]=[-1, 1])
% ?- solve_equations(2*x+y+z=15; 4*x+2*y+5*z=39;8*x+8*y+9*z=83, Sol)
% Sol = ([z, y, x]=[3, 2, 5])
% solve_equations(2*x+y+z=15; 4*x+2*y+5*z=39;8*x+4*y+4*z=83, Sol)
% Sol = 'no solution'
% solve_equations(2*x+y+z=15; 4*x+6*y+3*z=41;2*x+9*y+3*z=37, Sol).
% Sol = ([y, x]=[11/4, 49/8]+z*[-1/4, -3/8])
% ?- solve_equations(x+y=5; 2*x+4*y=30, Sol).
% ?- eqs_matrix(x+y=5 ; 2*x+4*y=30 , X, Y, Z).
% ?- solve_equations(x=y ; y=z ; z=z, X).
% X = ([y, z]=[0, 0]+x*[1, 1])
% ?- solve_equations(x=y; y=z; z=z; x=1, X).
% X = ([x, y, z]=[1, 1, 1])
% ?- solve_equations(2*x+y+3*z=1;2*x+6*y+8*z=3;6*x+8*y+18*z=5, X).
% X = ([z, y, x]=[0, 2 rdiv 5, 3 rdiv 10])
% ?- solve_equations(x+z=1; x+y+z=2; x-y+z=1, X).
% X = 'no solution'
% ?- solve_equations(3*x+y-6*z= -10; 2*x+y-5*z= -8; 6*x-3*y+3*z=0, X).

% X = ([y, x]=[-4, -2]+z*[3, 1])

gauss((M * X ) = C, P) :- listp(C), !,
	phrase((transpose_matrix, reverse, transpose_matrix), M, M0),
	maplist(reverse, [X, C], [X0, C0]),
	maplist(multiply(-1), C0, C1),
	solve_AX_C(M0, X0, C1, P).
gauss --> solve_equations.

multiply(-1, X, Y):- Y is (-1) * X.

%%

solve_equations(Eqs, P):- solve_equations_(Eqs, P).

solve_equations_(Eqs, P):- eqs_matrix(Eqs, X, Y, Vs),
	transpose_matrix(X, X0),
	maplist(peek([]), X0, U),
	solve(m(U, [], [], X0, Y, Vs, []), P0),
	solution(P0, P1),
	subst_rdiv(P1, P).

solution(m(_, _, _,_ , D, _, _), 'no solution'):- member(N, D), \+ N=0, !.
solution(m(_, C, X,_ , _, _, []), X = C):-!.
solution(m(_, C, X,_ , _, _, S), X= Sol):-
	length(X, L0),
	maplist(zeropad(L0), S, S0),
	foldl(plusterm, S0, C, Sol).

zeropad(L, A*X, A*Y):- length(X, L0), N is L-L0, repeat(N, cons(0), X, Y).


plusterm(X, Y, Y+X).

eqs_matrix(X, Y, Z, Vs):- eqs_rows(X, X0), make_matrix(X0, Y, Z, Vs).

eqs_rows --> {equation_separator(P)}, flatten(P),
	maplist(elim_eq),
	maplist(gblex:term_poly),
	maplist(poly_affine).

poly_affine(Poly, (C, S)):- select(C*[], Poly, P0)
	-> maplist(var_scalar_form, P0, S)
	;  C=0, maplist(var_scalar_form, Poly, S).

var_scalar_form(C*[A^1], (A,C)).

elim_eq(X=Y, X-Y).

make_matrix(Eqs, M, C, Vs):-
	maplist(fst, Eqs, C),
	maplist(snd, Eqs, Ls),
	vars(Ls, Vs),
	maplist(make_row(Vs), Ls, M).

vars --> maplist(maplist(fst)), append, sort.

make_row(Vs, Ls, As):- maplist(coeff(Ls), Vs, As).

coeff(Eqs, X, Y):- memberchk((X,Y), Eqs).
coeff(_, _, 0).

elim_mul([], (0, [])).
elim_mul([A|B], (A, B0)):- rational(A), !, maplist(mul_pair, B, B0).
elim_mul(A, (0, A0)):- maplist(mul_pair, A, A0).

%%  veryfying solutions by substitution.
%
% % ?- justify(x=1, A, R).
% % ?- justify(x=x, A, R).
% % ?- justify((x=1);(x=2), A, R).
% %  A = 'no solution'
% ?-justify((x-y=1);(x+y=2), A, R).
%
justify(Eqs, Ans, R) :- solve_equations(Eqs, Ans),
	once(justify0(Eqs, Ans, R)).

justify0(Eqs, (Xs = Vs), justified):-
      vec_expr_vec(Vs, V),
      zip(Xs, V, Assoc),
      beta_subst(Assoc, Eqs, Eqs0),
      flatten(;, Eqs0, E),
      forall(member(L=R, E), equal_zero(L-R)).
justify0(_, _ = _, 'Failed').
justify0(_, _, _).

vec_expr_vec(X, Y):- eval(vec_expr_vec, X, Y).

vec_expr_vec(A+B, C, [A, B], [A1, B1], ip(\(X,\(Y, $(X+Y))), A1, B1, C)).
vec_expr_vec(A*V, C, [], [],  maplist(\(X,  $(A*X)), V, C)).
vec_expr_vec(X, X, [],  []).

equal_zero(X):- make_affine_form(X, X0), forall(member(A, X0), (A = 0; A = (_*0))).

affine_linear --> ~((A,L), [A|L]),
	maplist(pair_mul_expr),
	remove_redundant,
	arithmetic_term(+).

pair_mul_expr((_,0), 0).
pair_mul_expr((X,1), X).
pair_mul_expr((X,C), C*X).
pair_mul_expr(X, X).

remove_redundant([],[]).
remove_redundant([0|X],Y):- !, remove_redundant(X, Y).
remove_redundant([A|X],[A|Y]):- remove_redundant(X, Y).

arithmetic_term(+, [], 0):-!.
arithmetic_term(*, [], 0):-!.
arithmetic_term(_, [X], X):-!.
arithmetic_term(F, [X,Y|Z], U):- arithmetic_term(F, [Y|Z], U0), U =.. [F, X, U0].

%%%%% Gauss'es Sweeping Method (engine)
%% configuration:
%% m(<upper right matrix>, <upper vector>, <solved variables>
%%   <lower right matrix>, <lower vector>, <unsolved variables>
%%   <generators>)

solve(P, P):- final(P),!.
solve(P, Q):- find_pivot(P, P0, R), !,
	R = p([A|As], C),
	A0 is -A,
	C0 is C rdiv A0,
	maplist(rdiv(A0), As, Bs),
	sweep_matrix(Bs, C0, P0, P1),
	push_row(Bs, C0, P1, P2),
	solve(P2, Q).
solve(P, Q):- push_generator(P, P0), !, solve(P0, Q).
solve(P, P).

final(m([],_,_,_,_,_,_)).

rdiv(X, Y, Z):- Z is Y rdiv X.

push_row(A, B, m(P,C,X0, Q,D,X,Z), m(P0,[B|C],X0,Q,D,X,Z)):-
	maplist(cons, A, P, P0).

%  ?- push_generator(m([[1,2],[3,4]],[5,6],[u],[[7,8], [9,10]],[11,12],[x, y],[]), P).

% P = m([[3, 4]], [5, 6], [u], [[9, 10]], [11, 12], [y], [x*[1, 2]])

push_generator(
      m([A|As], B, Xs, [_|Cs], D, [Y|Ys],        Z),
      m(As,     B, Xs, Cs    , D, Ys    ,  [Y*A|Z])).

% ?- find_pivot(m([],[],[],[[1,2], [3,4]],[5,6],[x,y],[]), P, Q).
% P = m([], [], [x], [[2], [4]], [6], [y], []),
% Q = p([1, 3], 5)

% ?- find_pivot([[1,2], [3,4]],[5,6], X, Y, A,B).
% X = [[2], [4]], Y = [6], A = [1, 3], B = 5

find_pivot(m(A,B,F0,D,E,[Z|F],G), m(A,B,[Z|F0], D0, E0,F,G), p(X,Y)):-
	find_pivot(D,E,D0,E0,X,Y).

find_pivot([[A|As]|Bs], [C|Cs], D, E, F, G):-
	cars(Bs, BsCars, BsCdrs),
      (  A \== 0
      -> D=[As|BsCdrs], E=Cs, F=[A|BsCars], G=C
      ;
	find_pivot([As|BsCdrs], Cs,  D0, E0,  F, G),
	maplist(cons, [A|BsCars], D0, D),
	E=[C|E0]
      ).

sweep_matrix(A, B, m(P,C,Y0,Q,D,Y,Z), m(P0,C0,Y0,Q0,D0,Y,Z)):-
	sweep_matrix(A,B,P,C,P0,C0),
	sweep_matrix(A,B,Q,D,Q0,D0).

sweep_matrix(_, B, [], C, [], C0):-!, C0 is B+C.
sweep_matrix(_, _, [[]|M], [], M0, []):-!, maplist(peek([]), M, M0).
sweep_matrix(A, B, M, C, M0, [Q0|Q]):-
	cars(M, Mcars, Mcdrs),
	C=[C0|C1],
	sweep_matrix(A, B, Mcdrs, C1, P,  Q),
	sweep_vector(A, B, Mcars, C0, P0, Q0),
	maplist(cons, P0, P, M0).

sweep_vector(P, Q, [A0|As], C, Bs, D):-
	maplist(multiply_minus_add(A0), P, As, Bs),
	D is (A0 * Q) +  C.

multiply_minus_add(A, B, C, D):- D is (A*B) + C.


cars(X, Y, Z):- var(X), (Y==[]; Z==[]), !, X = [].
cars([], [], []).
cars([[A|B]|Cs], [A|As], [B|Ds]):- cars(Cs, As, Ds).

car([X|_], X).
cdr([_|X], X).

%%%% solving system of linear equations
% ?- osolve_AX_C([[1]],[x],[2],X).
% ?- solve_AX_C([[1,0],[0,1]],[x,y],[2,3],X).
% ?- solve_AX_C([[0,1],[1,0]],[x,y],[2,3],X).
% ?- solve_AX_C([[0,1/3],[1/5,0]],[x,y],[2/3,3/4],X).
% ?-  M = [[1, 1, -2], [2, 3, -1], [1, 1, 0]], C=[5,7,13], V=[x,y,z], solve_AX_C(M,V,C, P).
% ?-  M = [[1, 1, -2], [2, 3, -1], [1, 1, 0]], C=[5,7,13], V=[x,y,z], solve_AX_C(M,V,C, P).
% ?-  M = [[1,1], [0,0]], C= [3, 3], V=[x,y], solve_AX_C(M,V,C, P).
% ?-  M = [[1,1], [0,0]], C= [3, 0], V=[x,y], solve_AX_C(M,V,C, P).
% %%%%%%%%%  end of test of matrix operations %%%%%

solve_AX_C(M, X, C, P):-
        maplist(maplist(rational_rational), M, M0),
	maplist(rational_rational, C, C0),
	maplist(peek([]), M0, U),
	solve(m(U, [], [], M0, C0, X, []), P0),
	solution(P0, P1),
	subst_rdiv(P1, P).

rational_rational(E, V):- gblex:term_poly(E, V0),
	(V0=[]-> V=0; V0=[V*[]]).

% %%% traceing  solve_equations
% ?- trace, solve_equations(x+y= 1; x-y=0, S).
% eqs_matrix((x+y=1;x-y=0),  [[1, 1], [1, -1]], [-1, 0], [x, y]) ?
% zip([[1, 1], [1, -1]], [[1, 1], [1, -1]]) ?
% solve(
%  m([[], []], [], [], [[1, 1], [1, -1]], [-1, 0], [x, y], []),
%  m([], [1 rdiv 2, 1 rdiv 2], [y, x], [], [], [], [])) ?
% solution(m([], [1 rdiv 2, 1 rdiv 2], [y, x], [], [], [], []),
%         [y, x]=[1 rdiv 2, 1 rdiv 2]) ?
% subst_rdiv([y, x]=[1 rdiv 2, 1 rdiv 2],
%            [y, x]=[1/2, 1/2])
% solve_equations((x+y=1;x-y=0), [y, x]=[1/2, 1/2]) ?
% S = ([y, x]=[1/2, 1/2])

%% display matrix in text
% /***
% # flatten, (matrix,matrixinverse, padding)
% -2   0
% 0   2
% ****/

rational_codes(-rdiv(X,Y), Z):- !, X0 is -X, herbrand_opp(X0/Y, Z).
rational_codes(rdiv(X,Y), Z):- !, herbrand_opp(X/Y, Z).
rational_codes(X, Y):- herbrand_opp(X, Y).

% padding/2
padding --> maplist(maplist(rational_codes)),
	padding_,
	transpose_matrix,
	maplist(insert("  ")),
	insert("\n").

% padding_/2
padding_(X, Y):-
	maplist(maplist(length), X, L),
	maplist(maximum, L, M),
	maplist(padding_, M, X, Y).

% padding/3
padding(N, X, X0):- length(X, J),  J0 is N - J,
	length(Ps, J0),
	maplist(=(0'\s), Ps),  %'
	append(Ps, X, X0).

% padding_/3
padding_(N, X, Y):- maplist(padding(N), X, Y).

maximum(L, M):- extremal(max, L, M).

minimum(L, M):- extremal(min, L, M).

extremal(_, [X], X):-!.
extremal(M, [X|Y], Z):- extremal(M, Y, Y0),
	( M==max -> Z is max(Y0, X); Z is min(Y0, X)).

%%
% ?- latex_prod_table(mult(5), 5, Res).
% ?- latex_prod_table_list([3,4,5], X).

latex_table --> latex_table(id).

latex_table(F) --> current(X),
	maplist(length), sort, reverse, current([MaxLen|_]),
	peek(X),
	maplist(maplist(F)),
	{ length(Pre, MaxLen),  maplist(=('c|'), Pre) },
	maplist(insert(" & ")),
    maplist(fun([X]-> [X, "\\\\\\hline\n"])),
	peek(A, ["\\begin{tabular}{", "|", Pre, "}\\hline\n",
	        A,
			"\\end{tabular}\n"]).
