:- module(pac_meta, [act_tree/3, times/3,
					 pac_product/4, scons/3,
					 paths_of_term/2]).

:- set_prolog_flag(optimise, true).
term_expansion --> pac:expand_pac.
:- use_module(pac(op)).
paths_of_term(T, X):-
	fold_paths_of_term(	pred([A, [Ra|B], B]:- reverse(A, Ra)),
						T, X, []).

% ?- repeat(10, writeln(hello)).
% ?- repeat(10000, pred([X,Y]:- Y is X+1), 0, X).
% ?- maplist(pred([], [[X|_], X], true), [[1,2],[a,b]], V).
% ?- map([]^[[X|_], X]:-true, [[1,2],[a,b]], V).
% ?- map([[X|_], X], [[1,2],[a,b]], V).
% ?- map([[X|_], X]\\true, [[1,2],[a,b]], V).
% ?- map((=,=),[1,2],X).

% ?- meta:phrase_list(pred([], [[X|_], X]), [[1,2],[a,b]], V).
% ?- phrase_list([]^[[X|_], X]:-true, [[1,2],[a,b]], V).
% ?- phrase_list([[X|_], X], [[1,2],[a,b]], V).
% ?- phrase_list([[X|_], X]\\true, [[1,2],[a,b]], V).
% ?- phrase_list((=,=),[1,2],X).

% ?- meta:map([[X|_], X], [[1,2],[a,b]], V).
% ?- meta:map([]^[[X|_], X]:-true, [[1,2],[a,b]], V).
% ?- meta:map([[X|_], X]:- true, [[1,2],[a,b]], V).
% ?- meta:map((=,=),[1,2],X).

:- meta_predicate map(2, ?, ?).
map(P) --> maplist(phrase(P)).
%
:- meta_predicate phrase_list(2, ?, ?).
phrase_list(P) --> maplist(phrase(P)).

% Fold a tree (t(a, [t1, t2,...,tn]), where Default t = (,) .
% similar folding lists.
% ?- act_tree( pred([[X,Y],X=Y]), [(a, [1,2]), (b, [3,4])], R).
% R = [4=b,3=b,2=a,1=a].
% ?- act_tree( pred([[A,B], (A, B)]), [(a, [1,2]), (b, [3,4])], X).
% X = [ (1, a), (2, a), (3, b), (4, b)]
% ?- act_tree(pred([[Y,Z], [(Y,Z)|A], A]), [(a, [1,2]), (b, [3,4])], X,[]).
% X = [ (4, b), (3, b), (2, a), (1, a)].
% ?- r_act_tree(pred([A, [A|B], B]), [(a, [1,2]), (b, [3,4])], X, []).
% X = [[4,b],[3,b],[2,a],[1,a]].
:- meta_predicate l_act_tree(3,?,?,?),
	r_act_tree(3,?,?,?),
	act_tree(2,?,?),
	act_tree(3,?,?,?).
%%
act_tree(F, T, X, Y):- act_tree(l, F, T, X, Y).

r_act_tree(F, T, X, Y):- act_tree(r, F, T, X, Y).

%% l_act_tree/4 must be before act_tree/3.
l_act_tree(F, T, X, Y):- act_tree(l, F, T, X, Y).

act_tree(F, Tree, V):-
	l_act_tree(pred(F, [A, [B|X],X]:- call(F, A, B)),
		   Tree, V, []).
%
act_tree(M, F, T, X, Y):- act_tree(M, ',', F, [], T, X, Y).

act_tree(M, I, F, T, X, Y):- act_tree(M, I, F, [], T, X, Y).

act_tree(_,_, _, _, [], X, X):-!.
act_tree(l,S, F, As, [T0|T], X, Y):- !,
	act_tree(l, S, F, As, T0, X, X0),
	act_tree(l, S, F, As, T, X0, Y).
act_tree(r, S, F, As, [T0|T], X, Y):- !,
	act_tree(r, S, F, As, T0, X0, Y),
	act_tree(r, S, F, As, T, X, X0).
act_tree(M, S, F, As, T, X, Y):- compound(T), functor(T, S, 2), !,
        arg(1, T, A),
	arg(2, T, B),
	act_tree(M, S, F, [A|As], B, X, Y).
act_tree(_, _, F, As, T, X, Y):- call(F, [T|As], X, Y).

% ?- times(2, [a,b], B).
% B = [[a,a],[a,b],[b,a],[b,b]]
times(N, A, B):- times(N, A, [[]], B).

times(0, _, A, A).
times(N, A, B, C):- N>0, scons(A, B, D),
	N0 is N-1,
	times(N0, A, D, C).

% ?- meta:scons([1,2,3],[[a,b],[c,d]], R).
%@ R = [[1, a, b], [1, c, d], [2, a, b], [2, c, d], [3, a, b], [3, c, d]].
scons(X, Y, Z):-
     foldl(pred(Y, [A, U, V]
		:- foldl(pred([Y,A], [B, [[A|B]|R], R]),
			      Y, U, V)),
	   X, Z, []).

:- meta_predicate pac_product(3,?,?,?).
% ?-meta:pac_product(pred([X, Y, [X|Y]]), [1,2,3],[[a,b],[c,d]], R).
% ?-meta:pac_product(pred([X, Y, [X, X|Y]]), [1,2,3],[[a,b],[c,d]], R).
% ?-meta:pac_product(pred([X, Y, X-Y]), [1,2,3],[a,b,c,d], R).

pac_product(F, X, Y, Z):-
	foldl(pred([F, Y], [A, U, V]:-
		 foldl(pred([F, A], [B, [C|R], R]:-call(F, A, B, C)),
		       Y, U, V)),
	      X, Z, []).
%
map_cons(X,Y,Z):-  maplist(pred([A,B,[A|B]]), X, Y, Z).

% ?- pac_meta:cartesian([1,2,3],[a,b,c], Z).
cartesian(X, Y, Z):- pac_product(pred([A,B,[A-B]]), X, Y, Z).

%;; (setq module-query  "qcompile(util(meta2)), module(meta2).")
% ?- time(repeat(1000, integral(0, 1, 0.1, pred([X, 1]), R))).
% ?- pac_meta:integral(0, 1, 0.1, pred([X, 1]), 0, R).
% ?- qcompile(util(meta2)), module(meta2).
% ?- meta2:integral(0, 1, 0.1, pred([X, 1]), 0, R).
% R = 1.0999999999999999.
% ?- integral(0, 1, 0.1, pred([X, X]), 0, R).
% R = 0.55.
% ?- integral(0, 1, 0.001, pred([X, X^2]), 0, R).
% R = 0.33283350000000095.
% ?- integral(-10, 10, 0.001, pred([X, X^2]), 0, R).
% R = 666.7666700000648.
% ?- integral(-10, 10, 0.001, pred([X, X^2]), 0, R).
% R = 666.7666700000648.
% ?- integral(-10, 10, 0.001, pred([X, X^2]), 0, R).
% ?- time(integral(-20, 20, 0.0001, pred([X, exp(-(X^2)/2)/sqrt(2*pi)]), 0, R)).
% ?- time(slow_integral(-100, 100, 0.0001, pred([X, exp(-(X^2)/2)/sqrt(2*pi)]), 0, R)).
%@ % 6,000,002 inferences, 1.460 CPU in 1.471 seconds (99% CPU, 4110176 Lips)
%@ R = 0.9999999999999251.

%
% :- meta_predicate integral(?,?,?,2,?).
integral(L, R, S, F, V):-
	(		F = [X, Exp]
	->		assert('INTFUN'(X, V):- V is Exp),
			F0 = 'INTFUN'
	;		F0 = F
	),
	integral(L, R, S, F0, 0, V).

:- meta_predicate integral(?,?,?,2,?,?).
integral(L, R, _S, _F, V, V):- R =< L, !.
integral(L, R, S, F, V, V0):- call(F, L, H),
	M is S*H+V,
	L0 is L+S,
	integral(L0, R, S, F, M, V0).

% For comparison benchmark test only
slow_integral(L, R, _S, _F, V, V):- R =< L, !.
slow_integral(L, R, S, F, V, V0):- call(F, L, H),
	M is S*H+V,
	L0 is L+S,
	slow_integral(L0, R, S, F, M, V0).

% poly_sort(X, Y):-
% 	predsort([A, _*M, _*M0]:-
% 			(	gblex:compare_mono(A0,  M, M0),
% 				(A0 == (>) -> A = (<); A=(>))
% 			),
% 		  X, X0),
% 	poly_merge(X0, Y).

% a la Mathematica's Table function.
% ?- table(reverse, [[1,2],[a,b]], X).
% ?- table(=, [[1,2],[a,b]], X).
% X = [[[1,a],[1,b]],[[2,a],[2,b]]].
% ?- table([As,T]:- T=..[f|As], [[1,2],[a,b]], X).
% ?- table([As,T]:- sumlist(As, T), [[1,2],[3,4],[5,6,7],[8,9,10]], X).

:- meta_predicate table(2,?,?), table(?,2,?,?).

table(F, Is, M):- table(Is, F, [], M).

table([], F, As, V):-!, call(F, As, V).
table([I|Is], F, As, M):- maplist(table(Is, F, As), I, M).

table(Is, F, As, A, M):- table(Is, F, [A|As], M).

		/***************************************
		*     file_under/2 by process macro    *
		***************************************/

ignore_special((.)).
ignore_special((..)).
ignore_special('.git').
ignore_special('.DS_Store').
ignore_special(worktrees).

%
leaf_node(node(_, X)):- ignore_special(X), !.
leaf_node(node(W, N)):- working_directory(_, W),
				  exists_file(N).

% ?-  children(node('/path/to/directory/', directory), L).
% ?-  pac_meta:children(node('/Users/', cantor), L).

children(X, []):- leaf_node(X), !.
children(node((.), N), Ss):- !, working_directory(W, W),
	children(node(W, N), Ss).
children(node(W, (.)), Ss):- !,
	children(node(W, W), Ss).
children(node(W, N), Ss):-
	working_directory(_,W),
	exists_directory(N),
	directory_files(N, Fs),
	working_directory(W,N),
	working_directory(W0, W0),
	children(Fs, W0, Ss).

%
children([], _, []).
children([X|Y], W, Z):- (X=(.);X=(..)), !,
	children(Y, W, Z).
children([X|Y], W, [node(W, X)|Z]):-
	children(Y, W, Z).

% ?- pac_meta:file_under('/Users/cantor/public_html/', Y).
file_under(W, Node):-
		do(process(parent, leaf_node), node(W, W), Node).
%
parent(node((.), N), A):- !,
				working_directory(_, .),
				working_directory(W, W),
				parent(node(W, N), A).
parent(node(W, (.)), A):- !,
				parent(node(W, W), A).
parent(Node, A):- children(Node, Ss),
				member(A, Ss).

		/****************************************
		*     dict_of_files by process macro.   *
		****************************************/

%%  dict_of_files(+W:directory, -D:<dict>) is det.
%   List files recursively under the directory W, and
%   unify D with the result in the form of the dict with
%	<subdirectory name> - <sub list> for subdirectories.
%
% sample call:
% ?- pac_meta:dict_of_files((.), D), length(D, N).

dict_of_files(W, S):-
	do(process(push_children, pred([[]])), [d(W, W, S)], _).

%
dict_of_files(S):-  working_directory(W,W),
			dict_of_files(W, S).

%
push_children([d(W, N, S)|X], Z):-
	working_directory(_, W),
	child_directories(N, S, Z, X).

%
child_directories(N, S, A, B):- exists_directory(N),
			 directory_files(N, S0),
			 working_directory(W, N),
			 working_directory(W0, W0),
			 residue_for_next(S0, W0, S, A, B),
			 working_directory(_, W).

%
residue_for_next([], _, [], A, A).
residue_for_next([I|X], W, Y, A, B):- ignore_special(I),!,
		residue_for_next(X, W, Y, A, B).
residue_for_next([I|X], W, [I|Y], A, B):- exists_file(I), !,
		residue_for_next(X, W, Y, A, B).
residue_for_next([I|X], W, [I-U|Y], [d(W,I,U)|A], B):-
		exists_directory(I),
		!,
		residue_for_next(X, W, Y, A, B).
residue_for_next([I|X], W, [I|Y], A, B):- residue_for_next(X, W, Y, A, B).
