:- module(ml,
	  [parse_bind_context/2, bind_context/3,
	   parse_markup_body/4, parse_markup_text/4, eval_markup_text/2,
	   parse_phrase_save/1,
	   act/3, act/4
	   ]).

% ?- [util(ml)].

:- use_module(util(snippets)).
% :- use_module('../pac-core').

		/***************************
		*     parse_markup_text    *
		***************************/

% [2013/06/14]

expand_arg(X,Y,Z,U,V):- pac:expand_arg(X, Y, Z, U, V).

parse_markup_text(X, A, B) :- once(parse_markup_text(X, [], A, B)).

% ?- ml:parse_markup_text(X, `@ (=).\n==\n`, V).
% ?- ml:parse_markup_text(X, `@a,b,c,\nd.\n123\n@f.\n456\n==\n==`, R).
% ?- ml:parse_markup_text(X, `@a,b,c,\nd.\n123\n@f.\n456\n`, R).
% ?- ml:parse_markup_text(X, `@ (=).\naaa\n==\n`, V).
% ?- ml:parse_markup_text(X, `@ (=).\n@ (=).\naaa\n==\n`, V).
% ?- ml:parse_markup_text(X, `@ (=).\n@ (=).\naaa\n==\nbbb\n`, V).
% ?- ml:parse_markup_text(X, `@ (=).\na\n`, V).
% ?- ml:parse_markup_text(X, `@ (=).\n==\n`, V).
% ?- ml:parse_markup_text(X, `@ pred([X,X]).\n==\n`, V).
% ?- ml:parse_markup_text(X, `@ pred([X,X]).\n1`, V).
%@ X = [com('pac#10', [49])],
% ?- ml:parse_markup_text(L, `@ pred([X,[X,X]]).\nhello`, Y), basic:smash(L).

parse_markup_text(P, P, [], []).
parse_markup_text([G|P], Q) --> parse_tag(G),
	parse_markup_text(P, Q).
parse_markup_text(P, Q) --> line_feed(P, R),
	parse_markup_text(R, Q).

% ?- ml:parse_tag(X, `@ (=).\n1`, R).
parse_tag(com(G, Body)) --> "@", skip_markup_open(Tag_codes),
	parse_markup_body(Body, []),
	{
   	 basic:herbrand_in_context(web, Tag_codes, G0), % <==== & の宣言なし
	 expand_arg(G0, [], G, Aux, []),
	 maplist(assert, Aux)
	}.


% [2013/09/09]
% parse_phrase_save(X) :- herbrand(_, X, G0)
%
eval_markup_text(X, Y) :- parse_markup_text(L, X, _),
	once(bind_context((L, []),  (Y, _))).

%
act(G) --> act([], G).

% ?- trace, ml:parse_phrase_save(`a`).
parse_phrase_save(X) :- herbrand_in_context(web, X, G0),
	 % prolog_load_context(module, M),
	 % expand_arg(G0, M, G, Aux, []),
	 expand_arg(G0, [], G, Aux, []),
	 maplist(assert, Aux),
	 nb_setval(phrase_tag, G).

%
parse_markup_body(X, A, B):- once(parse_markup_body(X, [], A, B)).

%
parse_markup_body(X, X)	--> markup_close.
parse_markup_body([C|X], Y) --> parse_tag(C),
	parse_markup_body(X, Y).
parse_markup_body(X, Y) --> line_feed(X, Z),
	parse_markup_body(Z, Y).

% skip to the next line.
line_feed([0'\n|X], X)--> "\n".     % ' for the following not getting red.
line_feed([A|X], Y) --> [A], line_feed(X, Y).
line_feed(X, X)--> [].

%
skip_markup_open(X, A, B):- once(skip_markup_open(X, [], A, B)).
%
skip_markup_open(X, X) --> ".\n".   % full stop.
skip_markup_open([A|X], Y) --> [A], skip_markup_open(X, Y).
skip_markup_open(X, X) --> [].

%
markup_close --> "==", line_feed(_,[]).
markup_close --> "@end", line_feed(_,[]).
markup_close([], []).	% implicit multiple closure

% ?- parse_bind_context(`split, pred([X,X]).\nh\ne\n`, Y), smash(Y)).
% ?- parse_bind_context(`pred([X,X]).\nhello`, Y), smash(Y).
% ?- parse_bind_context(`pred([X,[X,X]]).\nhello`, Y), smash(Y).
% ?- parse_bind_context(`(=).\nhello`, Y), smash(Y).
% ?- parse_bind_context(` (=).\nhello`, Y), smash(Y) .
% ?- parse_bind_context(`pred([X,X]).\n[1,2,3]\n`, V), smash(V).
% ?- parse_bind_context(`herbrand, pred([X,X]), term_codes.\n[1,2,3]\n`, V), smash(V).
% ?- parse_bind_context(`herbrand, maplist([X, f(X)]), term_codes.\n[1,2,3]\n`, V), smash(V).

% ?- parse_bind_context(`dcl([a(1)]).\nhello`, Y), smash(Y).
% ?- eval_markup_text(`@dcl([a(1)]).\nhello`, Y), smash(Y).
% ?- eval_markup_text(`@   dcl([a(1)]).\nhello`, Y), smash(Y).
% ?- eval_markup_text(`@  pred([X,X]).\nhello`, Y), smash(Y).

parse_bind_context(X, Y) :-  append(Tag, [0'., 0'\n | Rest], X), !,
   	 herbrand_in_context(web, Tag, G0),
	 expand_arg(G0, [], G, Aux, []),
	 maplist(assert, Aux),
	 once(bind_context((com(G, Rest), []),  (Y, _))).

act(E, G, X, Y):-  once(bind_context(G, (X, E), (Y, _))).

:- meta_predicate use_context(?, 2, ?).
use_context(C, F, (X, C), (Y, C)):- call(F, X, Y).

% bind_context/2
% ?- bind_context(([com(id, [])], []), X).
% ?- bind_context((com(obj(put([f(1)])),  []), []), X).

bind_context(([A|B], E), ([A0|B0], F)):-
	bind_context((A, E), (A0, E0)),
	bind_context((B, E0), (B0, F)).
bind_context((com(F, A), E), (B, D)):-
	bind_context((A, E), (A0, E0)),
	bind_context(F, (A0, E0), (B, D)).
bind_context(X, X).

% ?- bind_context(obj_put([a(hello)]), ([],[]), Y).
% ?- bind_context((obj_put([a(hello)]), obj_get([a(X)])), ([],[]), Y).
% ?- bind_context((get([a(A)]), peek(A)),([], [a(hello)]), Env).

:- meta_predicate  bind_context(:, ?, ?).

bind_context(M:G) --> m_bind_context(G, M).
bind_context(G)	  --> m_bind_context(G, []).

%
m_bind_context(M:G, _)	--> m_bind_context(G, M).
m_bind_context((F,G), M)--> m_bind_context(F, M),
	m_bind_context(G, M).
m_bind_context({F}, []) --> {once(F)}.
m_bind_context({F}, M)	--> {M:once(F)}.
m_bind_context(&(F), [])--> phrase(F).
m_bind_context(&(F), M)	--> phrase(M:F).
m_bind_context(dcl(D), M) --> m_bind_context(obj(push(D)), M).
m_bind_context(context(E), _, (X, E), (X, E)).
m_bind_context(context(C, F), M, (X, C), (Y, C)):- once(x_phrase(F, M, X, Y)).
m_bind_context(obj(F), _, (X, C), (X, D)) :- obj:obj(F, C, D).
m_bind_context(get(A), _, (X, C), (X, D)) :- obj:obj_get(A, C, D).
m_bind_context(put(A), _, (X, C), (X, D)) :- obj:obj_put(A, C, D).
m_bind_context(save, _, (X, C), ([], D))  :- obj:obj_put([acc(X)], C, D).
m_bind_context(restore, _, (_, C), (X, D)) :- obj:obj_pull([acc(X)], C, D).
m_bind_context(F, M, (X, C), (Y, C)):- 	once(x_phrase(F, M, X, Y)).

%
x_phrase((A,B), M, X, Y):- x_phrase(A, M, X, X0),
	x_phrase(B, M, X0, Y).
x_phrase(A;B, M, X, Y):-
	(	x_phrase(A, M, X, Y)
	;	x_phrase(B, M, X, Y)).
x_phrase(\+(A), M, X, Y):- \+ x_phrase(A, M, X, Y).
x_phrase(once(A), M, X, Y):- once(x_phrase(A, M, X, Y)).
x_phrase(M:A, _, X, Y):- x_phrase(A, M, X, Y).
x_phrase(F, [], X, Y):- call(F, X, Y).
x_phrase(F, M, X, Y):- call(M:F, X, Y).
