:- module(obj,
	  [obj/3, obj_clear/2, obj_get/2, obj_get/3, obj_act/3, obj_act/5,
	  obj_init/2, obj_init/3, obj_init/4, obj_merge/3,
	  obj_merge_add/3, obj_pull/3, obj_pull_one/3, obj_push/3,
	  obj_put/3, obj_update/3, obj_update_one/3,
	  obj_default/3
	  ]).

		/*************
		*     obj    *
		*************/

% ?- module(obj).
% ?- qcompile(util(obj)), module(obj).
%%% object manager
% ?- obj([a(X), b(Y),c(Z)], [c(3), b(2), a(1)], U).
% ?- obj(get([a(X)]), [c(3), b(2), a(1)], U).
% ?- obj(put([a(4)]), [c(3), b(2), a(1)], U).
% ?- obj(pull([a(X), b(Y)]), [c(3), b(2), a(1)], U).
% ?- obj(push([a(4)]), [c(3), b(2), a(1)], U).
% ?- obj(append([a(4)]), [c(3), b(2), a(1)], U).
% ?- obj(select(a(X)), [c(3), b(2), a(1)], U).
% ?- obj_pull([a(1)], [a(1),a(2), a(1)], X).
% ?- obj_update([a(1),b(2)], [b(4),c(5), a(3)], R).
% ?- obj(act(a, b, =), [a(hello)], X).
% ?- obj(act(a, a, =), [a(hello)], X).
% ?- obj_act(a, =, [a(hello)], X).
% ?- obj_act(a, pred([A,A]:- writeln(A)), [a(hello)], X).

%
obj([], Y, Y):-!.
obj([A|B], Y, Y):-!, subset([A|B], Y).
obj((A,B), Y, Z):- !, obj(A, Y, Y0), obj(B, Y0, Z).
obj(get(X), Y, Z):-!, obj_get(X, Y, Z).
obj(put(X), Y, Z):-!, obj_put(X, Y, Z).
obj(push(X), Y, Z):-!, obj_push(X, Y, Z).
obj(pull(X), Y, Z):-!, obj_pull(X, Y, Z).
obj(update(X), Y, Z):-!, obj_update(X, Y, Z).
obj(merge(X), Y, Z):-!, obj_merge(X, Y, Z).
obj(default(X), Y, Z):-!, obj_default(X, Y, Z).
obj(act(A, B, G), X, Y):-!, obj_act(A, B, G, X, Y).
obj(act(A, G), X, Y):-!, obj_act(A, G, X, Y).
obj(act(G), X, Y):-!, obj_act(G, X, Y).
obj(X, Y, Z):- call(X, Y, Z).

% meta
obj_get(X, Y):- subset(X, Y).
obj_get(X, Y, Y):- subset(X, Y).
obj_put(X, Y, Z):- obj_update(X, Y, Z).
obj_push(X, Y, Z):- append(X, Y, Z).
obj_pull(X, Y, Z):- foldl(obj_pull_one, X, Y, Z).
obj_update(X, Y, Z):- foldl(obj_update_one, X, Y, Z).
obj_default(X, Y, Z):- foldl(obj_default_one, X, Y, Z).
obj_clear(_, []).

% obj helpers

obj_pull_one(A, X, Y):- (select(A, X, Y) -> true; Y=X).
%
obj_update_one(A, B, [A|B0]) :- functor(A, F, 1),
	functor(A0, F, 1),
	select(A0, B, B0),
	!.
obj_update_one(A, B, [A|B]).

% ?- obj:obj_default([a(2)], [a(1)], X).
% ?- obj:obj(default([a(1)]), [], R).
% ?- obj:obj(default([a(X)]), [], R).

obj_default_one(A, B, C):- functor(A, F, 1),
	functor(A0, F, 1),
	(	memberchk(A0, B) ->
		(	A0 = A -> C = B
		;	C = B
		)
	; 	C = [A|B]
	).

% ?- obj_merge([a(2)], [a(1)], X).
% ?- obj_merge([b(2)], [a(1)], X).
obj_merge([])-->[].
obj_merge([X|Y]) --> obj_merge_add(X), obj_merge(Y).
%
obj_merge_add(X, Y, Z):- functor(X, A, N),
	functor(X0, A, N),
	(	memberchk(X0, Y) ->
		X0 = X,
		Z = Y
	;	Z=[X|Y]
	).

% ?- obj_act(a, b, =, [a(hello)], X).
% ?- obj_act(a, a, =, [a(hello)], X).
% ?- obj_act(a, write, [a(hello)], X).
% ?- obj_act(a, =, [a(hello)], X).
% ?- obj_act(a, pred([A,A]:- writeln(A)), [a(hello)], X).

:- meta_predicate obj_act(?,?,2,?,?).
obj_act(A, B, G, X, Y):- A0=..[A, V],
	obj_pull([A0], X, X0),
	solve(G, V, W),
	B0=..[B, W],
	obj_push([B0], X0, Y).
%
solve(true, A, A).
solve((X,Y), A, B):- !, solve(X, A, A0), solve(Y, A0, B).
solve(X, A, B):- call(X, A, B).

:- meta_predicate obj_act(?,2,?,?).
%
obj_act(A, G, X, Y):- obj_act(A, A, G, X, Y).

% :- meta_predicate obj_act(2,?,?).
obj_act(G, X, Y):- obj_act(acc, acc, G, X, Y).

%
obj_init(X, [acc(X)]).
obj_init(Objs, X, [acc(X)|Objs]).
obj_init(A, Objs, X, [B|Objs]):- functor(B, A, 1), arg(1, B, X).
