:- module(ptq, []).

%% (setq module-query  "qcompile(util(ptq)), module(ptq).")
%
%  run lisp for this s-expression, and type "jj" to get a qcompling query,
%  then run the query.

% ?- qcompile(util(ptq)), module(ptq).

:- use_module(pac(reduce)).
:- use_module(util(misc)).
:- use_module(util(models)).
:- use_module(util('emacs-handler')).
:- use_module(util(beta)).
:- use_module(util(langsem)).
:- use_module(pac(op)).
:- use_module(pac(basic)).


:- op(670, xfy, \).
% ?- qcompile(util(ptq)), module(ptq).
% ?- ptq:demo(A,B), smash(B).
% ?- ptq:ptq_set(s, [john,walk], V), smash(V).
% ?- ptq(np, [a, unicorn], R), smash(R).


demo(A,B):-demo_sentence(A), once(ptq(A, B)).

%
demo_sentence([john,walk]).
demo_sentence([john,find, a, unicorn]).
demo_sentence([john,seek, a, unicorn]).
demo_sentence([every, unicorn, walk]).
demo_sentence([a, unicorn, find, john]).

%% de Brujin indexing
debrujin(T0, T) :- debrujin(T0, T, _).
%
debrujin(T, T1, N):- debrujin([], 1, T, N, T1).
%
debrujin(N, T, N1, T1):- debrujin([], N, T, N1, T1).
%
debrujin(Assoc, N, T, N, T1):- assocq(Assoc, T, T1), !.
debrujin(Assoc, N, X, N1, Y):- scope(X, V, B, Y, #(N), B1), !,
	N2 is N+1,
	debrujin([(V, #(N))|Assoc], N2, B, N1, B1).
debrujin(Assoc, N, T, N1, T1):- eh:termrec(ptq:debrujin_(Assoc), T, T1, N, N1).
%
debrujin_(A, X, Y, Z, U):- debrujin(A, X, Z, Y, U).

scope(\(V,B), V, B, \(V0,B0), V0, B0).
scope(iota(V, B), V, B, iota(V0, B0), V0, B0).
scope(forall(V, B), V, B, forall(V0, B0), V0, B0).
scope(exists(V, B), V, B, exists(V0, B0), V0, B0).

shift(N, #(M), #(M1)):- !, M1 is M + N.
shift(N, T, T1) :- mapterm(shift(N), T, T1).

alpha_equiv(X, Y):- debrujin(X, Z, N), debrujin(Y, Z, N).

% ptq_sgn([(\)/2, (\\)/2, iota/2, forall/2, exists/2
%%%% for html %%%
% ?- ptq:ptq_set(s, [john,walk], X), smash(X).
% ?- ptq:ptq_set([john,walk], X), smash(X).

ptq_set(X, S):- ptq_set(s, X, S).
%
ptq_set(C, X, S):- ptq_all_(C, X, S0),
	term_smash0(["<font color=\"red\">", S0, "</font>"], S).

ptq_all_(C, X, S):- setof(Y, ptq(C, X, Y), S1), !,
	maplist(term_smash0,S1,S2), insert("<br>",S2,S).
ptq_all_(_, _, "** Category  mismatch ?").


%%% for cgi  %%%
% ?- ptq:ptq([john, walk], X).
% ?- ptq:ptq(s, [john, find, a, unicorn], X), smash(X).
% ?- ptq:ptq(s, [john, seek, a, unicorn], X), smash(X).
% ?- ptq:webchurch(p(a, b), X), smash(X).

ptq --> ptq(s).

ptq(C) -->  parse(C), !,
	simplify, !,
	sharp_to_print_name, !,
	pretty_print_html, !.

webchurch --> simplify,
	sharp_to_print_name, !,
	pretty_print_html, !.


%%% parse %%%
% ?- ptq:parse(s, [john,walk], X).
parse(C, X, Y):- call(C, Y, X, []), normal(Y).

normal(error(_,_,_)):- !, fail.
normal(_).

%%% DCG for a Tiny PTQ  (Lambda version)

s(@(NP2,int(\(X,@(NP1,int(VP)))))) --> np(NP1), vp(VP,X,NP2).
% for 'specific' reading: Rule S14, T14
s(@(NP, int(VP))) --> np(NP), vp(VP).
% Rule S5, T5

np(@(A, int(CN))) --> determiner(A), cn(CN).
np(PN) --> pn(PN).

vp(ITV)  --> itv(ITV).
vp(@(TV, int(NP))) --> tv(TV), np(NP).

vp(@(TV, int(\(P, @@(P,X)))), X, NP) --> tv(TV), np(NP).

itv(walk)-->[walk].

tv(find_star) --> [find].
tv(seek) --> [seek].
tv(\(P, \(X, @@(P, int(\(Y, X=Y)))))) --> [is].

pn(\(P, @@(P, j))) --> [john].
pn(\(P, @@(P, b))) --> [bill].

cn(unicorn) --> [unicorn].
cn(man) --> [man].

% ?- qcompile(util(ptq)), module(ptq).
determiner(X\ (Y\ (exists(Z, @@(X,Z) & @@(Y,Z)))))-->[a].
determiner(X\ (Y\ (forall(Z, @@(X,Z)-> @@(Y,Z)))))-->[every].

%%% simplify expressions

% ?- qcompile(util(ptq)), module(ptq).
% ?- trace, ptq:simplify(exists(#1, man(#1) & (#1=john)), E).
% ?- trace, ptq:simplify(true(first, second), E).
% ?- trace, ptq:simplify(true, E).

% ?- current_op(X, Y, \).
% ?- write_canonical(a\b\c).
%
variable(#(_)).
%
signature( \+ ((#)/1) ).

:- meta_predicate ptq_reduce(:,?,?).
ptq_reduce(R) --> {signature(S)}, reduce:reduce(R, subtree(S)).

% ?- qcompile(util(ptq)), module(ptq).

% ?- ptq: simplify( ycombinator @ factorial @ cn(2), X).
% ?- ptq: simplify( @(@(ycombinator,  factorial), cn(2)), X).

simplify --> {b_setval(brujin_index, 0)},
	ptq_reduce(expand_lambda_macro),
	debrujin,
	ptq_reduce(ptq_rule),
	ptq_reduce(redex_to_term),
	ptq_reduce(elim_redundant),
	ptq_reduce(paramodulation),
	ptq_reduce(identity),
	ptq_reduce(elim_redundant),
	ptq_reduce(boole),
	debrujin.
%
ptq_rule(@@(P,X), ext(P)@X).
ptq_rule(ext(int(X)), X).
ptq_rule((X\Y)@Z, U):- b_getval(brujin_index, I),
	debrujin(I, X\Y, J, X0\Y0),
	b_setval(brujin_index, J),
	beta_subst([(X0, Z)], Y0, U).
ptq_rule(X, Y):- ptq_postulate(X,Y).

% PTQ表示規約 f@a= f(a).   (f@y)@x = f(x, y)
% ?- redex_to_term(f@y, E).
% ?- reduce(ptq:redex_to_term, (f@y)@x, E).

redex_to_term(X@Y, Z):-
	L = [exists(_,_), forall(_,_), #(_), ext(_), int(_),
	     @@(_,_), @(_,_)],
	\+ member(X, L),
	X=..[F|X0],
	Z=..[F,Y|X0].

%
paramodulation(exists(X, B), exists(X, B0)):-
	select_eq_conj(B, X=R, B1),
	beta:beta_subst([(X,R)], B1, B0).
%
select_eq_conj(L=R, L=R, true):- variable(L).
select_eq_conj(L=R, R=L, true):- variable(R).
select_eq_conj(A&B, E, A0&B):- select_eq_conj(A, E, A0).
select_eq_conj(A&B, E, A&B0):- select_eq_conj(B, E, B0).

identity(X = Y, true):- X == Y.

% ?- qcompile(util(ptq)), module(ptq).
% ?-  reduce:reduce(ptq:boole, true & false, X).
% ?-  reduce:reduce(ptq:boole, true & false, X).
% ?-  reduce:reduce(ptq:boole, true & true, X).
% ?-  reduce:reduce(ptq:boole, or(true, false), X).

boole((A&B)&C, A&(B&C)).
boole(true&A, A).
boole(A&true, A).
boole(_&false, false).
boole(false&_, false).
boole(not(not(X)), X).
boole(or(_,true), true).
boole(or(true,_), true).
boole(or(false,A), A).
boole(or(A, false), A).
boole(not(true), false).
boole(not(false), true).

elim_redundant(exists(X, Y), Y):- \+ appear(X, Y).
elim_redundant(forall(X, Y), Y):- \+ appear(X, Y).

% ?- appear( a, f(a, b)).
appear(X, X).
appear(X, Y):- Y=..[_|Y0], member(Z, Y0), appear(X, Z).

% %
expand_ptq_macro(X, X):- var(X), !.
expand_ptq_macro(X, Y):- def(X, Def), !,  expand_ptq_macro(Def, Y).
expand_ptq_macro(X, Y):- X=..[F|Xs],
	(	def(F, D)
	->  expand_ptq_macro(D, E),
		maplist(expand_ptq_macro, Xs, Ys),
		beta:lambda_application(E, Ys, Y)
	;	maplist(expand_ptq_macro, Xs, Ys),
		Y =..[F|Ys]
	).

% #(1) -> x; #(2) -> y; ....
% ?- ptq:sharp_to_print_name(f(#1, #3), X).
%@ X = f(x, z).

sharp_to_print_name(X, X):- var(X), !.
sharp_to_print_name(#(I), X):-
	nth1(I, [x,y,z,u,v,w,l,m,n,p,q,r,s,t,u,a,b,c,d,e,f,g,h,i,j,k], X),
	!.
sharp_to_print_name(#(I), #(J)):- !, number_codes(I,J).
sharp_to_print_name(T, T0):- T=..[F|A],
	maplist(sharp_to_print_name, A, A0),
	T0 =.. [F|A0].

% ?- num_num(3, C).
% num_num(I, J) :- format(codes(C), "~w", [I]), number_codes(J,C).

% Meaning postulate 4 (意味公準)

ptq_postulate(find_star@P@X, @@(P, int(Y\ find@Y@X))).

% ?- qcompile(util(ptq)), module(ptq).

%% term to lambda expression
% ?- ptq:expand_lambda_macro(cn(0), X), ptq:sharp_to_print_name(X, Y).
% ?- ptq:expand_lambda_macro(cn(1), X), ptq:sharp_to_print_name(X, Y).
% ?- ptq:expand_lambda_macro(cn(2), X), ptq:sharp_to_print_name(X, Y).

expand_lambda_macro(M, D):-
	(	def(M, E)	->	debrujin(E, D)
	;	M=..[F|As],
		def(F, M0),
		debrujin(M0, M1),
		beta_redex(M1, As, D)
	).

%
beta_redex(M, [], M).
beta_redex(M, [A|As], N):- beta_redex(M@A, As, N).

%%%% 関数定義 (将来拡張用)
def(cn(X), M):- church_numeral(X, M).
def(p, x\y\x).
def(q, x\y\ y).
def(pi1, x\y\ x).
def(qi2, x\y\ y).
def(true, x\y\ x).
def(false, x\y\ y).
def(and, x\y\ (x @ y @ false)).
def(not, x\ (x @ false @ true)).
def(if, x\y\z\ (x @ y @ z)).
def(succ, n\f\x\ (f @ (n @ f @ x))).
def(pred, n\f\x\ (n @ (g \ h \ ( h @ (g @ f))) @ (u \ x) @ (u \ u))).
def(plus, m\n\f\x\ ( m @ f @ (n @ f @ x))).
def(mult, m\n\ (m @ (plus @n) @ cn(0))).
def(iszero, n\ (n@ (x \ false) @ true)).
def(zero, cn(0)).
def(one, cn(1)).
def(two, cn(2)).
def(three, cn(3)).
def(cons, f\s\b\ (b @ f @s)).
def(car, p\ (p @ true)).
def(cdr, p\ (p @ false)).
def(ycombinator,  f\ ((x\ (f @ (x @ x))) @ (x\ (f @ (x @ x))))).
def(factorial, f\n\ if(iszero(n), cn(1), mult(n, f @ pred(n)))).
def(test, f\n\ cn(0)).

% ?- qcompile(util(ptq)).
% ?- module(ptq).

% ?- church_numeral(0, CN).
% ?- church_numeral(1, CN).
% ?- church_numeral(2, CN).

church_numeral(N, CN) :- church_numeral(N, CN0, f, x),
	fresh_bind(CN0, CN).

%
church_numeral(0, F\X\ X, F, X).
church_numeral(N, F\X\ (F @ M), F, X):- N > 0,
	N0 is N-1,
	church_numeral(N0, F\X\ M, F, X).

%% pretty printing
% ?- sharp_to_print_name(exists(#1, man(#1)), X), pretty_print_html(X, Y), smash(Y).

% pretty_print_html --> eval(#m(ev, pretty_print_html, ptq)).
% pretty_print_html(X, Y):- once(eval([ptq:pretty_print_html], X, Y)).

pretty_print_html(A, [A]):- atomic(A), !.
pretty_print_html(X, Y):- X=..[F|A],
			length(A, N),
			maplist(pretty_print_html, A, B),
			pretty_print_html0(F, N, B, Y).

pretty_print_html0(exists, 2, [A, B], ["&exist;", A, B]).
pretty_print_html0(forall, 2, [A, B], ["&forall;", A, B]).
pretty_print_html0((\), 2, [A, B], ["&lambda;", A, B]).
pretty_print_html0((&), 2, [A, B], ["(", A, "&and;", B,")"]).
pretty_print_html0((or), 2, [A, B], ["(", A, "&or;", B,")"]).
pretty_print_html0((not), 1, [A], ["(","&not;", A, ")"]).
pretty_print_html0((->), 2, [A,B], ["(",A, "&rarr;", B, ")"]).
pretty_print_html0(F, 2, [A, B], ["(", A, F, B, ")"]):- declared_op(F, 2).
pretty_print_html0(F, 1, [A], ["(", F, A, ")"]):- declared_op(F, 1).
pretty_print_html0(F, _, L0, [F, "(", L , ")" ]):- insert(",", L0, L).

%
declared_op(X,N):- current_op(_,T,X),
	( N=2, memberchk(T, [xfx,xfy,yfx,yfy])
	; N=1, memberchk(T, [fx, fy])
	).
