:- module(webcla, []).

%%	 Web version

:- use_module(pac('expand-pac')).
:- use_pac(web).
% :- expects_dialect(pac).
term_expansion --> pac:expand_pac.
:- use_module(pac(op)).

:- set_prolog_flag(unknown, fail).%
%%%% Code for Printing  %%%%%

dir_path(deldel). % directory for temporary files  .pdf .png .bb  etc

file_name(cla).  % file name without extension.


% ?- Text= [command=`theory([bird, fly], [[bird] =>	[fly]])`, radioButton=`pdf`],
%  web:run_cgi( `(peek([@command, @radioButton]), maplist(herbrand_in_context),(eh:apply3(ifmap:theory_to_cla_html)), (basic:smash), through)`,
%	Text).
%% theory to cla [2015/10/30]
theory_to_cla_html(X,T,URL):- dir_path(D),  % T: output type
	file_name(F),
	phrase((eh:counter(check), eh:counter(update), obj_get([count(C)])),
	       [directory(D), counter_name(atm_cur_id)],  _),
	format(codes(OutHtml), "~w/~w~w.html", [D, F, C]),
	snap(c),
	tyeory_to_cla_html(F, D, C, X, T, Body),
	snap(d),
	flatten(["<html><body>\n", Body, "</body></html>"], H1),
	atom_codes(OutNameAtom, OutHtml),
	create_file(OutNameAtom, H1),
	eh:expand_cgi_path(OutHtml, URL).


%
theory_to_cla_html(F,D,C,R,T,H) :-
	regex_am(R, coa(A, Ini)),
	am_finals(coa(A, _), Fin),
	length(A, Num),
 	coalgebra_triples(A, M),
 	automaton_quasi_string(am(M, Ini, Fin), Quasi_String),
	maplist(pred([U-_, U]), A, S),
	H1 = (format_codes(`<p>Regular expression = ~w</p>`,[R]) &
	format_codes(`<p> The number of states = ~d</p>`,[Num]) &
	format_codes(`<p> Initial state = ~w</p>`,[Ini]) &
	format_codes(`<p> Final states = ~w</p>`,[Fin]) &
	format_codes(`<p> All states = ~w</p>`,[S]) &
	format_codes(`<p> State Transitions:</p>~n`,[])),
	atomic_list_concat([D, '/', F, C, '.dot'], DotName),
 	file(DotName, write, smash(Quasi_String)),
	formatForHtml(F,D,C,T,Format,Args),			% call sh from inside
	hybrid_print_moves(M, M1),
	format_codes_list(H1 & format_codes(Format,Args) & M1, L, []),
	!,
	append(L, H).

formatForHtml(F,D,C,X,Format,[F,C,X]) :-
        img_frame(Format),
        atomic_list_concat([D, (/), F, C], Base),
	once(option_table(X, Opt, Ext0, Ext1)),
     (  X == pdf
     -> Com = ps2pdf(`-sOutputFile=`+ Base+ `.`+ Ext1,
		     Base+ `.`+ Ext0)
     ;  Com = 'DUMMY=1'
     ),
         once(eh:sh(dot(-'T'(Opt), Base+ `.` + dot,
		  -o(Base + `.` + Ext0)); Com)).

img_frame(X) :- flatten([ `<p><div `,
	  `id='diagram' `,
	  `style='border : solid 2px #ff0000; `,
	  `width : 1600px; `,			% was 600px
	  `height : 500px; `,			% was 300px
	  `overflow : auto; '><br/>`,
	  `<img src="~w~w.~w"/>`,
	  `</div></p>~n`
	 ], X).

% ?- coalgebra:hybrid_print_moves([(1, [97-97], 2)], R).

hybrid_print_moves(M, H):-
	maplist(pred([(X, A, Y), (X, B, Y)]:-
	       maplist(interval_code_char, A, B)),
		M, M0),
	print_moves(M0, H).

%
print_moves(M, H):-
	maplist(print_moves_x, M, T),
	flatten(T, T1),
	flatten(["<pre>\n", T1, "</pre>\n"], H).

print_moves_x((X,A,Y), H):-
	format_codes(`   ~w----~w--->~w~n` , [X,A,Y], H).
