:-module(web, [  op(8,  fy, '`'),
	cgi_main/0, cgi_normal/0, cgi_debug/0,
	cgi_phrase/2,cgi_phrase/3,
	create_file/2, decode_pair/2, echo/2, equation_pair/2,
	eval_query/2, expandref/2,
	form_encode/2, format_codes/3, get_eof/1, get_request/1, getcodes/2,
	parse_form/2, parse_query/2, parse_query/3,
	query_to_answer/2, query_to_html_codes/2, response_form/2,
	sendmail/1,
	writeBR/1, write_html/1]).

:- use_module(pac(basic)).
:- use_module(pac(op)).
:- use_module(pac('expand-pac')).
:- use_module(util('emacs-handler')).
:- op(670, yfx, \).
:- op(200, xfy, ^).

:- use_module(util(file)).
:- use_module(util(cgi)).

:- set_prolog_flag(allow_variable_name_as_functor, true).
:- nb_setval(webform, []).
:- nb_setval(query_op, []).

		/***********************************
		*     Predicates for Prolog CGI    *
		***********************************/

%%	cgi_main  is det.
%	main entry for Prolog CGI.
cgi_main:- catch(cgi_normal, Ball, (response_form(Ball, R), write_html(R))).

%%	cgi_normal is det.
%	Top level process of Prolog CGI.
cgi_normal :- prompt(_,''),
	once(get_request(Form)),
	nb_setval(webform, Form),
	getenv(home, H),
	atomic_list_concat([H,'Desktop', http_request], /, Save),
	setup_call_cleanup(
			open(Save, write, S),
			(write(S, Form), write(S, ".\n")),
			close(S)),
	query_to_answer(Form, Ans),
	write_html(Ans),
	notify_cgi_access(Form).

% Clicking a button on browser, then run the query.
cgi_debug:- catch(cgi_debug0, Ball, (response_form(Ball, R), write_html(R))).
%
cgi_debug0 :- getenv(home, H),
	atomic_list_concat([H,'Desktop', http_request], /, Input),
	get_request(Input, Form),
	nb_getval(webform, Form),
	query_to_answer(Form, Ans),
    write_html(Ans).

%%	query_to_answer(+X:codes, -Y:codes) is det.
%	Eval query X in codes to cgi response form Y.
query_to_answer(X, Y):-
	memberchk(expr = C, X),
	query_to_html_codes(C, Y0),
	response_form(Y0, Y).

%%	notify_cgi_access(+F:codes) is det.
%	Email F to the $USER when notify_cgi_access_by_mail is true.
notify_cgi_access(Form):- getenv(notify_cgi_access_by_mail, true), !,
	memberchk(expr = Query, Form),
	getenv(user, Name),
	ignore(sendmail(["From: ", Name,	"\n",
		  "Subject: cgi in prolog access\n\n",
		  Query, "\n"])).
notify_cgi_access(_).

%
sendmail(X):- getenv(scratchfile, File_tmp), !,
	 getenv(user, User),
	file(File_tmp, write,
	       (  current_output(Out),
	          set_stream(Out, encoding(utf8)),
	          smash(X),
		  flush_output)),
	sh( cat(File_tmp) + " | " +  '/usr/sbin/sendmail'(User)).
sendmail(_).

%%	query_to_html_codes(+X, -Y) is det.
%	Eval query X into html codes Y.
% ?- sqc(web).
% ?- query_to_html_codes(`(peek([b]), append([a]))`, X), smash(X).
% ?- query_to_html_codes(`?-append([a],[b], A)`, X), smash(X).
% ?- parse_query((?- append([a], [b], _G2170)), ['A'=_G2170], _G2288).
% ?- parse_query(`?- append([a], [b], X)`, Y).
% ?- parse_query(`(peek([1]), append([2]), set::pow)`, X).
% ?- parse_query(`(?-append([a],[b], X))`, V).
% ?- parse_query(`append([a],[b])`, V).
% ?- parse_query(`(??- member(X, [1,2,3]))`, V).
% ?- parse_query(`(peek([b]), append([a]))`, V).
% ?- eval(misc:set::pow([1,2]), V).
% ?- parse_query(`(peek([a,b]), ([X]\\set::pow(X)))`, V).
% ?- expand_basic_phrase(fun([X], misc:set::pow(X)), user, F, P, []).
%  parse_query(X, _, _):- var(X), !, throw

query_to_html_codes(X, Y):-
	herbrand_deref(X, X0, V),
	parse_query(X0, V, Y).


%%	parse_query(+X, -Y) is det
%	Parse query X into a term Y.
parse_query(X,  Y):-
	herbrand_deref(X, X0, V),
	parse_query(X0, V, Y).

%%	parse_query(+E, -V, -Y) is det.
%	Parse query E to a term Y and variable_names V.
parse_query((?-X), V, Y):- !,
	must_be(callable, X),	% check_act(X, 0),
	once(expand_goal(X, [], G, P, [])),
	maplist(assert, P),
	(call(G) -> Y = V ; Y = fail).
parse_query(??-(X), V, Y):- !,	must_be(callable, X), % check_act(X, 0),
	once(expand_goal(X,[],G,P,[])),
	maplist(assert, P),
	(findall(V, call(G), Y) -> true; Y = []).
parse_query(X, _, Y):-	eval_query(X, Y).

%%	eval_query(+E, -V) is det.
%	Eval query E to a term V.
% ?- let(F, ([X]\ (set::pow@ X))), web:eval_query((peek([[1], [2]]), eh:apply(append), F), X).
eval_query(X, Y):-
	phrase_to_pred(X, [], H:-G, P, []),
	maplist(assert, P),
	H = [X,Y],
	call(G).

%%	cgi_phrase(+X, -V) is det.
%	Eval cgi-phrase  X to a term V.
cgi_phrase(X,  V):- cgi_phrase(X, Head, Rest), !,
	phrase_to_pred(Rest, [], H:-G, P, []),
	maplist(assert, P),
	pac:eval(Head, V0),
	H = [V0, V],
	call(G).
cgi_phrase(X,  V):- call(X, V).

% ?- cgi_phrase(((a,b),c), X, Y).
cgi_phrase(((X,Y),Z), U, V):- !, cgi_phrase((X,(Y,Z)), U, V).
cgi_phrase((X,Y), X, Y).

%%	response_form(+X, -Y) is det.%
%	Make term X into webform Y required by httprequest (Ajax)
%	interface.
response_form(X, Y):-
	(	nb_current(webform, W); W = [] ), !,
	(	member(buttonid = B, W); B = `dummybuttonid` ), !,
	(	member(targetid = R, W); R = `dummytargetid` ), !,
	atom_codes(B0, B),
	atom_codes(R0, R),
	term_smash0(X, X0),
	uri_encoded(path, X0, E),
	Y = [B0, "\n", R0, "\n", E].

%%	form_encode(+X, -Y) is det.
%	Bidirectional form encoding/decoding.
form_encode(X, Y):- atomic(X), !,
	www_form_encode(X, Y0),
	atom_codes(Y0, Y).
form_encode(X, Y):- atomic(Y), !,
	www_form_encode(X0, Y),
	atom_codes(X0, X).
form_encode(X, Y):-
	(	nonvar(X) -> atom_codes(A, X),
		www_form_encode(A, B),
		atom_codes(B, Y)
	;	atom_codes(B, Y),
		www_form_encode(A, B),
		atom_codes(A, X)
	).

%%	get_request(-Form:codes) is det.
%	Get CGI form, and parse it to Form.
get_request(Form):- get_eof(S), parse_form(S, Form).

%
get_request(RFile, Form):-
	setup_call_cleanup(
		open(RFile, read, U),
		read(U, Form),
		close(U)),
	nb_setval(webform, Form).

%%	herbrand_web(+X, -Y) is det.
%	parse cgi query X to a term Y.
% ?- herbrand_web(`a(A)`, X).
herbrand_web(X, Y):-  herbrand(web, _ , X, Y).

%%	herbrand_web(?B, +X, -Y) is det.
%	Parse cgi query X to a term Y with variable_names B.
% ?- herbrand_web(B, `a(A)`, X).
herbrand_web(Bindings, X, Y):- herbrand(web, Bindings, X, Y).

%%	herbrand_deref(+X, -Y, -B) is det.
%	Parse cgi query X to a term Y with variable_names B,
%	expanding @-symbol by deref-ing.

herbrand_deref(X, Y, Binds):-
	herbrand_web(Binds, X, X1),
	expandref(X1, Y).

%%	expandref(+X, Y) is det.
%	Expand all @-symbol in X referring to an assoc list,
%	which was in a global variable 'webform'.

expandref(X, Y):-  (var(X); atomic(X)), !, Y = X.
expandref(@(X), Y):- !, getcodes(X, Y).
expandref(X, Y):-  X=..[F|A], maplist(expandref, A, B), Y=..[F|B].

%%	getcodes(+X, -Y) is det.
%	Get codes of @-symbol X, and unify Y with the value.

% ?- b_setval(webform, [a= hello]), getcodes(a, X).
% X = hello
getcodes(X, Y):- nb_getval(webform, Env), member(X = Y, Env), !.
getcodes(X, _):- throw(error(no_data_for(X))).

%%	write_html(+R) is det.
%       write R back to browsers with some header info.

write_html(R):-
	(	string(R) -> E = R
	;	term_smash0(R, E)
	),
	string_length(E, N),
	format("Content-type: text/plain; charset=utf-8~n~n"),
	format("Content-Length: ~w~n", [N]),
	write(E).

%%	get_eof(-S) is det.
%	get an input from browsers into S.
get_eof(S):- get_code(C),
	(C < 0 -> S = []; S = [C|S1], get_eof(S1)).

% ?- get_eof('/Users/cantor/Desktop/http_request', R).
get_eof(U, S):- get_code(U, C),
	(	C < 0 -> S = []
	;	S = [C|S1],
		get_eof(U, S1)
	).

%%	parse_form(+X, -Y) is det.
%	Parse web form text codes X into Y.
parse_form(X, Y):- basic:split(`&`, X, L),
	maplist(equation_pair, L, Y1),
	maplist(decode_pair, Y1, Y2),
	maplist(name2atom, Y2, Y).

		/***********************
		*     tiny helpers.    *
		***********************/

decode_pair((X, Y), (X0, Y0)):- form_encode(X0, X), form_encode(Y0, Y).
%
name2atom((A,B), A1 = B) :- atom_codes(A1, A).
%
writeBR(W):- format("~w<br>\n", [W]).
%
equation_pair(E, (A, B)):- append(A, [0'=|B], E), !.   %'
%
format_codes(F, A, X):-  format(codes(X), F, A).
%
echo --> [].
%
create_file(F, C):- file(F, write, smash(C)).
