:- module(add_node_by_one_one, []).

:- use_module(zdd('zdd-array')).
:- use_module(zdd(zdd)).
:- use_module(zdd('minato-r5')).
:- use_module(pac(op)).
:- use_module(zdd('mate-indirect')).


		/*******************
		*   Tiny helpers   *
		*******************/

% ?- arrow_symbol(_->_, F).
% ?- arrow_symbol(a->b, F, X, Y).
arrow_symbol( _ -> _).
%
arrow_symbol(A, A0):- functor(A, A0, 2).
arrow_symbol(A, A0, A1, A2):- functor(A, A0, 2),
		arg(1, A, A1),
		arg(2, A, A2).

%
mate_less_than(_ - A, B -_):- A @< B.

% One of the most basic helpers.
composable_pairs(A-B, A-C, B, C).
composable_pairs(A-B, C-A, B, C).
composable_pairs(B-A, A-C, B, C).
composable_pairs(B-A, C-A, B, C).
%
normal_pair(A-B, U-V):-!, ( B @< A -> U=B, V=A; U=A, V=B ).
normal_pair(A->B, U->V):- ( B @< A -> U=B, V=A; U=A, V=B ).


		/****************************************************
		*     replace_end/bypass:  Most basic operations.   *
		****************************************************/

% ?- zdd((X<< +[[c-d]], @(subst_node([a->c], c, a, X, Y)), psa(X), psa(Y))).
subst_node(_, _, _, X, 0, _):- X<2, !.
subst_node(Es, A, P, X, Y, S):-			% replace A with P
	cofact(X, t(U, L, R), S),
	arrow_symbol(U, Arrow, Lu, Ru),
	(	Arrow = (->) ->	Y = 0
	;	A @< Lu ->	Y = 0
	;	subst_node(Es, A, P, L, L0, S),
		(	Ru = A	->
			normal_pair(Lu-P, V),
			zdd_ord_insert([V|Es], R, R0, S)
		;	Lu = A	->
			normal_pair(Ru-P, V),
			zdd_ord_insert([V|Es], R, R0, S)
		;	subst_node(Es, A, P, R, R1, S),
			zdd_insert(U, R1, R0, S)
		),
		zdd_join(L0, R0, Y, S)
	).

% ?- rect_nodes(rect(0,2), Ns).
% ?- rect_nodes(rect(10,10), Ns), length(Ns, L).
rect_nodes(rect(W, H), Ns):-
	findall(p(I,J),
			 (	between(0, W, I),
				between(0, H, J)
			 ),
			 Ns).

% ?- rect_udg(rect(1,1), UDG).
% ?- rect_udg(rect(10,10), UDG),length(UDG, L).
rect_udg(rect(W, H), UDG):-
	findall( p(I,J)-Suc,
				 (	between(0, W, I),
					between(0, H, J),
					findall( p(K,L),
						(  L=J, K is I + 1, K =< W
						;  K=I, L is J + 1, L =< H
						),
						Suc)
				 ),
				 UDG).

% ?- zdd((@udg_mate([], 1, X))).
% ?- zdd((@udg_mate([b-[]], 1, X), psa(X))).
% ?- rect_udg(rect(1,1), UDG), reverse(UDG, U),
%	zdd((@udg_mate(U, 1, X), psa(X))).
% ?- rect_mate(rect(1,0), X, S), psa(X, S).
% ?- rect_mate(rect(1,0), X, S), psa(X, S).
% ?- rect_mate(rect(1,1), X, S), psa(X, S).
% ?- rect_mate(rect(2,2), X, S), psa(X, S), card(X, C, S).
% ?- rect_mate(rect(3,3), X, S), card(X, C, S).
% ?- rect_mate(rect(4,4), X, S), card(X, C, S).
% ?- rect_mate(rect(5,5), X, S), card(X, C, S).
% ?- time((rect_mate(rect(6,6), X, S), card(X, C, S))).
%@ % 58,616,972 inferences, 6.255 CPU in 6.295 seconds (99% CPU, 9371439 Lips)
%@ X = 756594,
%@ S = ..,
%@ C = 575780564.

% ?- time((rect_mate(rect(7,7), X, S), card(X, C, S))).
% ?- time((rect_mate(rect(7,7), X, S), card(X, C, S))).
%@ % 264,554,753 inferences, 27.795 CPU in 28.319 seconds (98% CPU, 9517899 Lips)
%@ X = 3591236,
%@ S = ..,
%@ C = 789360053252.
% ?- time((rect_mate(rect(8,8), X, S), card(X, C, S))).
%@ % 2,503,775,609 inferences, 232.253 CPU in 232.940 seconds (100% CPU, 10780371 Lips)
%@ X = 110190,
%@ S = ..,
%@ C =

% ?- call_with_time_limit(2000, time((rect_mate(rect(9,9), X, S), card(X, C, S)))).
%@ % 11,947,268,675 inferences, 1053.032 CPU in 1057.017 seconds (100% CPU, 11345590 Lips)
%@ X = 377107,
%@ S = ..,
%@ C = 41044208702632496804.

% ?- call_with_time_limit(36000, time((rect_mate(rect(10,10), X, S), card(X, C, S)))).
%@ GC at node  p(10,0)
%@ GC at node  p(9,0)
%@ GC at node  p(8,0)
%@ GC at node  p(7,0)
%@ GC at node  p(6,0)
%@ GC at node  p(5,0)
%@ GC at node  p(4,0)
%@ GC at node  p(3,0)
%@ GC at node  p(2,0)
%@ GC at node  p(1,0)
%@ GC at node  p(0,0)
%@ % 29,396,165,381 inferences, 3825.651 CPU in 3841.534 seconds (100% CPU, 7683964 Lips)
%@ X = 1268341,
%@ S = ..,
%@ C = 1568758030464750013214100.

% [2022/10/01]
% ?- call_with_time_limit(36000, time((rect_mate(rect(11,11), X, S), card(X, C, S)))).
%@ GC at node  p(11,0)
%@ GC at node  p(10,0)
%@ GC at node  p(9,0)
%@ GC at node  p(8,0)
%@ GC at node  p(7,0)
%@ GC at node  p(6,0)
%@ GC at node  p(5,0)
%@ GC at node  p(4,0)
%@ GC at node  p(3,0)
%@ GC at node  p(2,0)
%@ GC at node  p(1,0)
%@ GC at node  p(0,0)
%@ % 129,787,391,715 inferences, 14036.125 CPU in 14092.162 seconds (100% CPU, 9246668 Lips)
%@ X = 4207535,
%@ S = ..,
%@ C = 182413291514248049241470885236.

% ?- call_with_time_limit(500, time((rect_mate(rect(12,12), X, [slim(all)], S), card(X, C, S)))).

% ?- rect_mate(rect(3,3), X, [gc(0)], S), card(X, C, S).
% ?- rect_mate(rect(3,3), X, [gc(all)], S), card(X, C, S).
% ?- rect_mate(rect(3,3), X, [], S), card(X, C, S).
% ?- rect_mate(rect(3,3), X, [gc(no)], S), card(X, C, S).

reverse_successors(X-A, X-B):- reverse(A, B).

% slim(K);  slim(no); slim(all).
rect_mate(Rect, X, S):-rect_mate(Rect, X, [slim(0)], S).

%
rect_mate(Rect, X, InOpts, S):-
	(	memberchk(gc(CtrlSlim), InOpts) -> true
	;	CtrlSlim= 0
	),
	Rect = rect(W,H),
	rect_udg(Rect, Dg0),
	maplist(reverse_successors, Dg0, Dg1),
	reverse(Dg1, Dg),
	open_state(S),
	set_key(max_node, p(W, H), S),
	udg_mate(CtrlSlim, Dg, 1, X0, S),
	get_key(max_node, Max, S),
	prune_final(p(0,0), Max, X0, X, S).
%
udg_mate(_, [], X, X, _).
udg_mate(Ctrl, [N-Ps|UDG], X, Y, S):-
	add_node(N, Ps, X, X0, S),
	get_key(max_node, Max, S),
	prune_by_classify_link(N, Max, X0, X1, S),
	ctrl_slim(Ctrl, N, X1, X2, S),
	udg_mate(Ctrl, UDG, X2, Y, S).

%
ctrl_slim(all, P, X, Y, S):-!,
	format("GC at ~w \n", [P]),
	zdd_slim(X, Y, S),
	garbage_collect.
ctrl_slim(no, _, X, X, _):-!.
ctrl_slim(K, P, X, Y, S):-integer(K),
	P = p(_, J),
	(	J = K ->
		format("GC at ~w \n", [P]),
		zdd_slim(X, Y, S),
		garbage_collect
	;	Y = X
	).


% ?- zdd((@add_node(a, [], 1, X), psa(X))).
% ?- zdd((@add_node(b, [], 1, X), @add_node(a, [b], X, Y), psa(Y))).
% ?- zdd((@add_node(b, [], 1, X), @add_node(a, [b], X, Y),
%	@add_node(c, [a,b], Y, Z), psa(Z))).

add_node(N, [], X, Y, S):-!, zdd_insert(N-N, X, Y, S).
add_node(N, [P|Ps], X, Y, S):- normal_pair(N->P, Q),
	subst_node([Q], P, N, X, X0, S),
	zdd_insert(N-N, X, X1, S),
	zdd_join(X0, X1, X2, S),
	add_neighbor_links(N, Ps, X2, Y, S).
%
add_neighbor_links(_, [], X, X, _):-!.
add_neighbor_links(N, [P|Ps], X, Y, S):- normal_pair(N-P, Link),
	add_link(Link, X, X0, S),
	zdd_join(X0, X, X1, S),
	add_neighbor_links(N, Ps, X1, Y, S).

% ?- zdd((X<<{[a-b, c-d]}, psa(X), @(add_link(b-c, X, Y)), psa(Y))).
% ?- zdd((X<< +[[a-b, c-d], [e-f]], psa(X), @(add_link(b-c, X, Y)), psa(Y))).
% ?- zdd((X<< +[[a-b, c-d], [a-a, b-b], [e-f]], psa(X), @(add_link(b-c, X, Y)), psa(Y))).
% ?- zdd((X<< +[[a-a, b-b, c-c]], psa(X), @(add_link(b-c, X, Y)), psa(Y))).

add_link(_, X, 0, _):- X<2, !.
add_link(U, X, Y, S):- % memo is useless here
	cofact(X, t(A, L, R), S),
	(	A = (_->_) -> Y = 0
	;	add_link(U, L, L0, S),
		U = Ul-Ur,
		(	mate_less_than(U, A) -> R0 = 0
		; 	A = U -> R0 = 0
		; 	composable_pairs(U, A, U0, V0) ->
			subst_node([Ul->Ur], U0, V0, R, R0, S)
		;	add_link(U, R, R1, S),
			zdd_insert(A, R1, R0, S)
		),
		zdd_join(L0, R0, Y, S)
	).

% ?- zdd((X<< +[[a-b, a->b]], @prune_final(a, b, X, Y), psa(Y))).
prune_final(_, _, X, X, _):- X<2, !.
prune_final(P, Q, X, Y, S):- cofact(X, t(A, L, R), S),
	(	A = (_->_) -> Y = X
	;	prune_final(P, Q, L, L0, S),
		A = A1-A2,
		(	 A1 = P ->
			 (	A2 = Q
			 -> prune_final(P, Q, R, R0, S)
			 ;	R0 = 0
			 )
		;	A1 = A2 ->
			prune_final(P, Q, R, R0, S)
		;	R0 = 0
		),
		zdd_join(L0, R0, Y, S)
	).


% ?- classify_link(p(0,1), p(3,3), p(1,0)-p(3,3), X).
% ?- classify_link(p(0,1), p(3,3), p(0,2)-p(3,3), X).
% ?- classify_link(p(0,1), p(3,3), p(0,1)-p(3,3), X).

classify_link(_, _, _->_, id):-!.
classify_link(P, End, A-B,  Case):- on_frontier(P, A), !,
   	(	on_frontier(P, B) -> Case = keep
	;	B = End -> Case = keep
	;	Case = 0
	).
classify_link(_, _, A-A, ignore):-!.
classify_link(_, _, _, 0).
%
on_frontier(p(0, J), p(0, J)):-!.
on_frontier(p(0, J), p(K, L)):-!, K = 1, L< J.
on_frontier(p(I, J), p(I, K)):-  K >= J, !.
on_frontier(p(I, J), p(I0, K)):- I0 is I + 1, K < J.
%
prune_by_classify_link(_, _, X, X, _):- X<2, !.
prune_by_classify_link(P, End, X, Y, S):- cofact(X, t(A, L, R), S),
	classify_link(P, End, A, Case),
	(	Case = id ->   Y = X
	;  	prune_by_classify_link(P, End, L, L0, S),
		(	Case = keep ->
			prune_by_classify_link(P, End, R, R1, S),
			zdd_insert(A, R1, R0, S)
		;	Case = ignore ->
			prune_by_classify_link(P, End, R, R0, S)
		;	R0 = 0
		),
		zdd_join(L0, R0, Y, S)
	).

% ?- open_state(S), open_state(W),
%	mate_ord_insert([a, <>(b)], 1, X, S, W),
%	mate_ord_insert([c, <>(d)], 1, Y, S, W),
%	mate_join(X, Y, Z, S, W),
%	mate_card(Z, C, S, W).
