:- module(mate_order, []).

:- use_module(zdd('zdd-array')).
:- use_module(zdd(zdd)).
:- use_module(pac(op)).

		/**********************
		*   sample queries    *
		**********************/

% ?- #S, rectangular_benchimark(rect(0,0), Z, S), card(Z, C, S).
% ?- #S, rectangular_benchimark(rect(0,1), Z, S), card(Z, C, S).
% ?- #S, rectangular_benchimark(rect(1,0), Z, S), card(Z, C, S).
% ?- #S, rectangular_benchimark(rect(1,1), Z, S), card(Z, C, S).
% ?- #S, rectangular_benchimark(rect(2,2), Z, S), card(Z, C, S).
% ?- #S, rectangular_benchimark(rect(2,3), Z, S), card(Z, C, S).
% ?- #S, rectangular_benchimark(rect(3,3), Z, S), card(Z, C, S).
% ?- #S, rectangular_benchimark(rect(4,4), Z, S), card(Z, C, S).
% ?- #S, rectangular_benchimark(rect(5,5), Z, S), card(Z, C, S).
% ?- #S, rectangular_benchimark(rect(6,6), Z, S), card(Z, C, S).
% ?- #S, rectangular_benchimark(rect(7,7), Z, S), card(Z, C, S).
% ?- #S, time((rectangular_benchimark(rect(7,7), Z, S), card(Z, C, S))).
%@ % 658,675,265 inferences, 49.654 CPU in 49.817 seconds (100% CPU, 13265356 Lips)
%@ S = ..,
%@ Z = 94322,
%@ C = 789360053252 .
% ?- #S, time((rectangular_benchimark(rect(8,8), Z, S), card(Z, C, S))).
%@ % 7,729,793,561 inferences, 728.353 CPU in 731.728 seconds (100% CPU, 10612696 Lips)
%@ S = ..,
%@ Z = 390004,
%@ C = 3266598486981642 .

rectangular_benchimark(R, Z, S):- R = rect(W, H),
		rect_links(R, Links),
		path_count_by_simple_frontier(Links, p(0,0)-p(W, H), Z, S).

% ?- #S, X<<pow([a,b]), max_length(X, I, S).
%% Qcompile: /Users/cantor/devel/zdd/prolog/zdd/build-frontier-table-mate-order.pl
% max_length for rectangular grid graphs.

% ?- N = 7, time(forall(between(1, N, K),
%	(	open_state(S),
%		rectangular_benchimark(rect(K,K), Z, S),
%		max_length(Z, Len, S),
%		close_state(S),
%		format("max length for rect(~w,~w) = ~w\n", [K, K, Len])
%	))).
%@ max length for rect(1,1) = 2
%@ max length for rect(2,2) = 8
%@ max length for rect(3,3) = 14
%@ max length for rect(4,4) = 24
%@ max length for rect(5,5) = 34
%@ max length for rect(6,6) = 48
%@ max length for rect(7,7) = 62
%@ % 815,391,154 inferences, 64.787 CPU in 65.033 seconds (100% CPU, 12585767 Lips)
%@ N = 7.
%@ max length for rect(1,1) = 2
%@ max length for rect(2,2) = 8
%@ max length for rect(3,3) = 14
%@ max length for rect(4,4) = 24
%@ max length for rect(5,5) = 34
%@ max length for rect(6,6) = 48
%@ max length for rect(7,7) = 62
%@ % 815,391,154 inferences, 65.269 CPU in 65.520 seconds (100% CPU, 12492762 Lips)
%@ N = 7.

%  [2022/11/01]
% ?- #S, time(rectangular_test(rect(10,10), Y, S)), card(Y, C, S).
%@ % 99,075,935,814 inferences, 11358.338 CPU in 13619.622 seconds (83% CPU, 8722750 Lips)
%@ S = ..,
%@ Y = 6467210,
%@ C = 1568758030464750013214100 .

% ?- #S, X<<pow([a,b,c]), zdd_rand_path(X, P, S).
% ?- #S, N=[a,b,c, d], power_links(N, B, S), choose_random_paths(B, 2, C, S),
%	maplist(writeln, C), map_count_path(C).
% ?- #S, K=2, N=5, numlist(1, N, Ns), power_links(Ns, B, S),
% choose_random_paths(B, K, C, S), close_state(S),  maplist(writeln, C),
%	map_count_path(C).
map_count_path([]).
map_count_path([A|As]):- count_path(A), !, map_count_path(As).

% ?- count_path(st([a-b], a-b)).
% ?- count_path(st([a-b,b-c,a-c], a-c)).
count_path(st(L,ST)):- open_state(S),
%		time(path_count_by_simple_frontier(L, ST, Z, S)),
		path_count_by_simple_frontier(L, ST, Z, S),
		card(Z, C, S),
		format("path count = ~w\n", [C]),
		close_state(S).

		/*******************************************
		*     Random different pair from a list    *
		*******************************************/

% ?- random_dif_pair([a,a,b,b], A, B).
% ?- random_dif_pair([], A, B).		% false
% ?- random_dif_pair([a], A, B).	% false
random_dif_pair(P, A, B):- sort(P, P0), length(P0, L),
	L > 1,
	random_ord_dif_pair(P0, A, B, L).

%
random_ord_dif_pair(P, A, B, L):-
	I is random(L),
	J is random(L),
	(	I\==J ->
		nth0(I, P, A),
		nth0(J, P, B)
	;	random_ord_dif_pair(P, A, B, L)
	).

% ?- elim_node([a,b,c, d], [a-b, a-c, b-c], R).
% ?- elim_node([a,b,c, d], [a-b], R).
elim_node([], _, []):-!.
elim_node(Ns, [], Ns):-!.
elim_node(Ns, [A-B|As], Ns0):-
	elim_node_one(Ns, [A,B], Ns1),
	elim_node(Ns1, As, Ns0).
%
elim_node_one([], _, []):-!.
elim_node_one(Ns, [], Ns):-!.
elim_node_one([A|Ns], Us, Ns0):-
	(	select(A, Us, Us1) ->
		elim_node_one(Ns, Us1, Ns0)
	;	elim_node_one(Ns, Us, Ns1),
		Ns0=[A|Ns1]
	).

% ?- #S, X<<pow([a-b]), choose_random_paths(X, 2, Y, S).
% ?- #S, X<<pow([a-b, c-d]), choose_random_paths(X, 10, Y, S),
%	maplist(writeln, Y).
choose_random_paths(_, 0, [], _):-!.
choose_random_paths(X, I, [st(P, A-B)|Y], S):- X>1,
	zdd_rand_path(X, P, S),
	P\==[],
	choose_random_LR(P, A, B),
	!,
	J is I - 1,
	choose_random_paths(X, J, Y, S).
choose_random_paths(X, I, Y, S):- X>1,
	choose_random_paths(X, I, Y, S).
%
choose_random_LR(P, A, B):- length(P, L), L > 0,
	I is random(L),
	J is random(L),
	nth0(I, P, A-_),
	nth0(J, P, _-B).

% ?- #S, power_links([a,b], B, S), psa(B, S), card(B, C, S).
% ?- #S, power_links([a,b,c,d], B, S), psa(B, S), card(B, C, S).
power_links(A, B, S):-
	findall(X-Y, (member(X, A), member(Y, A), X@<Y), L),
	<<(B, pow(L), S).

% ?- normal_mate_list([1-2], X).
% ?- normal_mate_list([2-1, 1-2], X).
normal_mate_list([], []).
normal_mate_list([P|R], [P0|R0]):- P=I-J,
	(	J@<I -> P0= J-I
	;	P0 = P
	),
	normal_mate_list(R, R0).


		/**************
		*     main    *
		**************/

% ?- #S, path_count_by_simple_frontier([a-b], a-b, X, S), card(X, C, S).
% ?- #S, path_count_by_simple_frontier([a-b, b-c, a-c], a-c, X, S), card(X, C, S).
% ?- #S, path_count_by_simple_frontier([a-b, a-c, b-d, c-d], a-d, X, S), card(X, C, S).
% ?- #S, path_count_by_simple_frontier([b-a, c-a, d-b, d-c], a-d, X, S), card(X, C, S).
% ?- #S, path_count_by_simple_frontier([b-a, b-c, c-a, d-b, d-c], a-d, X, S), card(X, C, S).
% ?- #S, path_count_by_simple_frontier([a-d, b-a, b-c, c-a, d-b, d-c], a-d, X, S), card(X, C, S).
path_count_by_simple_frontier(Links, ST, X, S):-
		normal_links_with_st(ST, Links, ST0, Links0, D),
		path_count_by_simple_frontier(Links0, D, ST0, X, S).

path_count_by_simple_frontier(Links, D, ST, Z, S):- Ctrl=[gc(link)],  % for default
	path_count_by_simple_frontier(Ctrl, Links, D, ST, Z, S).
%
path_count_by_simple_frontier(Ctrl, Links, D, ST, Z, S):- ST = I-J,
		findall(K-K, member(K, D), Init),
		zdd_append(Init, 1, X, S),
		Ctrl0=[end(J), start(I)|Ctrl],
		add_links(Ctrl0, Links, X, Y, S),
		prune_final(I, J, Y, Z, S).

% ?- normal_links_with_st(a-d, [a-b, b-c, c-d], R, L, D).
% ?- normal_links_with_st(a-c, [a-b, b-c, a-c], R, L, D).
% ?- normal_links_with_st(a-d, [a-b, a-c, d-b, d-c], R, L, D).
% ?- N=100, rect_links(rect(N,N), Links),
%	time((normal_links_with_st(p(0,0)-p(N,N), Links, ST, R, D))).
% ?- N=1, rect_links(rect(N,N), Links),
%	time((normal_links_with_st(p(0,0)-p(N,N), Links, ST, R, D))).
% ?- N=2, rect_links(rect(N,N), Links),
%	time((normal_links_with_st(p(0,0)-p(N,N), Links, ST, R, D))),
%	maplist(writeln, R).

normal_links_with_st(I-J, Links, I0-J0, Links0, D):-
	open_state(S),
	build_node_layers([J], Links, N),
	reverse(N, N0),
	number_node_layers(N0, 0, _, S),
	number_links(Links, Links1, S),
	normal_mate_list(Links1, Links2),
	predsort(mate_compare, Links2, Links3),
	reverse(Links3, Links4),
	map_put_fr(Links4, Links0),
	memo(number_node(J)-J0, S),
	memo(number_node(I)-I0, S),
	domain_of_links([I0-J0|Links4], D),
	close_state(S).
%
map_put_fr([], []).
map_put_fr([I-J|R], [fr(I-J, J)|R0]):- map_put_fr(R, R0).

		/****************************
		*	 build layers of links  *
		****************************/

% ?- domain_of_links([a-b, b-c, a-c], Y).
domain_of_links(X, Y):-
	findall( A, (	member(L, X),
					( L = (A - _)
					; L = (_ - A)
					)
				),
		   Y0),
   sort(Y0, Y).

% ?- build_node_layers([d], [c-d, b-c, a-b], N).
%@ N = [[d], [c, d], [b, c], [a, b]].
build_node_layers(Ns, X, L):- build_link_node_layers(Ns, X, _, _, L, []).

% ?- build_link_node_layers([d], [a-b, a-c, b-d, c-d], Unused, Layers, N, []).
% ?- build_link_node_layers([d], [c-d, b-c, a-b], Unused, Layers, N, []).
build_link_node_layers([], X, X, [], N, N):-!.
build_link_node_layers(Ns, X, Y, [L|Ls], [Ns|N], N0):-
	layer_links(Ns, X, X0, L),
	domain_of_links(L, Ns0),
	subtract(Ns0, Ns, Ns1),
	build_link_node_layers(Ns1, X0, Y, Ls, N, N0).

% ?- #S, number_layers_frontier([[a-b]], L, 0, C, S).
% ?- #S, number_layers_frontier([[a-b], [b-c, a-c]], L, 0, C, S).
number_layers_frontier([], [], C, C, _).
number_layers_frontier([L|Ls], [L0|Ls0], C, C0, S):-
	number_links_with_frontier(L, L0, C, C1, S),
	number_layers_frontier(Ls, Ls0, C1, C0, S).

% ?- #S, number_node_layers([[a]], 0, C, S).
% ?- #S, number_node_layers([[a,b], [b,c,a], []], 0, C, S).
number_node_layers([], C, C, _).
number_node_layers([Ns|R], C, C0, S):-
	number_node_list(Ns, C, C1, S),
	number_node_layers(R, C1, C0, S).

% ?- #S, number_node_list([a], 0, C, S).
% ?- #S, number_node_list([a,b, b,c,a], 0, C, S).
number_node_list([], C, C, _):-!.
number_node_list([N|Ns], C, C0, S):- number_node(N, C, C1, S),
	number_node_list(Ns, C1, C0, S).

% ?- layer_links([a,b], [a-b], L0, L1).
% ?- layer_links([a,b], [a-b, c-d], L0, L1).
% ?- layer_links([a,b], [a-b, b-c], L0, L1).
layer_links(_, [], [], []):-!.
layer_links(Ns, [A-B|Links], Links0, [A-B|Layer]):-
	(	member(C, Ns),
		(A = C; B = C)
	),
	!,
	layer_links(Ns, Links, Links0, Layer).
layer_links(Ns, [L|Links], [L|Links0], Layer):-
	layer_links(Ns, Links, Links0, Layer).

% ?- #S, number_node(a, 0, C, S).
number_node(N, C, C0, S):- number_node(N, _, C, C0, S).

% ?- #S,  time((numlist(1, 100000, Ns), foldl(pred(S, ([I, C, C0]:-
%	number_node(st(I), K, C, C0, S))), Ns, 0, R))).
number_node(N, I, C, C0, S):- memo(number_node(N)-I, S),
	(	nonvar(I) -> C0 = C
	;	C0 is C+1,
		I = C0
	).

% ?- #S, number_node_layers([[a],[b]], 0, C, S), number_links([a-b, b-a, a-a, b-a, a-b], L, S).
number_links([], [], _).
number_links([A-B|L], [A0-B0|L0], S):-
	memo(number_node(A)-A0, S),
	memo(number_node(B)-B0, S),
	number_links(L, L0, S).

% ?- #S, number_links([a-b], L, 0, C, S).
% ?- #S, number_links([a-b, b-c], L, 0, C, S).
number_links([], [], C, C, _).
number_links([A-B|L], [A0-B0|L0], C, C0, S):-
	number_node(A, A0, C, C1, S),
	number_node(B, B0, C1, C2, S),
	number_links(L, L0, C2, C0, S).

		/*******************
		*	 Helpers       *
		*******************/

% ?- mate_compare(C, a-c, b-a).
% ?- predsort(mate_compare, [1-1, 1-2, 1-3, 2-2, 2-3, 2-4, 3-3, 3-4], X).
%@ X = [1-1, 1-2, 2-2, 1-3, 2-3, 3-3, 2-4, 3-4].

% ?- findall(A-B, (between(0, 10, B), between(0, B, A)), R),
%  predsort(mate_compare, R, R0), maplist(writeln, R0).
% assumeing first @=< second.
mate_compare(C, A-B, X-Y):- compare(C0, B, Y),
	(	C0=(=) -> compare(C, A, X)
	;	C = C0
	).

% ?- arrow_symbol(_->_, F).
% ?- arrow_symbol(a->b, F, X, Y).
arrow_symbol( _ -> _).
%
arrow_symbol(A, A0):- functor(A, A0, 2).
arrow_symbol(A, A0, A1, A2):- functor(A, A0, 2),
		arg(1, A, A1),
		arg(2, A, A2).

% One of the most basic helpers.
composable_pairs(A-B, A-C, B, C).
composable_pairs(A-B, C-A, B, C).
composable_pairs(B-A, A-C, B, C).
composable_pairs(B-A, C-A, B, C).
%
normal_pair(A-B, U-V):-!, ( B @< A -> U=B, V=A; U=A, V=B ).
normal_pair(A->B, U->V):- ( B @< A -> U=B, V=A; U=A, V=B ).

% ?- rect_nodes(rect(0,2), Ns).
% ?- rect_nodes(rect(10,10), Ns), length(Ns, L).
rect_nodes(rect(W, H), Ns):-
	findall(p(I,J),
			 (	between(0, W, I),
				between(0, H, J)
			 ),
			 Ns).

% ?- rect_links(rect(1,1), Links).
% ?- rect_links(rect(10,10), Links),length(Links, L).
rect_links(rect(W, H), Links):-
	findall( p(I,J)-p(K,L),
				 (	between(0, W, I),
					between(0, H, J),
					(  L=J, K is I + 1, K =< W
					;  K=I, L is J + 1, L =< H
					)
				 ),
				 Links).

		/************************
		*     core predicates   *
		************************/
%
add_links(_, [], X, X, _).
add_links(Ctrl, [FR|Ls], X, Y, S):- FR=fr(U, F),
	memberchk(end(End), Ctrl),
	add_link(F-End, U, X, X1, S),
	zdd_join(X, X1, X2, S),
	(	( Ls = [] ; Ls = [fr(_, G)|_], G \== F ) ->   % step of adjacent frontiers found.
		prune_by_classify_link(F, End, X2, X3, S)
	;	X3 = X2		% Redundant pruning skipped.
	),
	(	memberchk(gc(link), Ctrl) ->
%		format("at ~w with link ~w\n", [X3, U]),
		zdd_slim(X3, X4, S),
		garbage_collect
	;	X4 = X3
	),
	add_links(Ctrl, Ls, X4, Y, S).

%
add_link(_, _, X, 0, _):- X<2, !.
add_link(FE, U, X, Y, S):- FE = F-E,
	cofact(X, t(A, L, R), S),
	add_link(FE, U, L, L0, S),
	classify_link(F, E, A, Case),
	(	( Case = 0; Case = arrow ) ->	R0 = 0			% many hits.
	;	Case = ignore -> add_link(FE, U, R, R0, S)		% no hits.
	;	U = Ul-Ur,
		(	A = U -> R0 = 0		%,  write(.)			% so so hits
		; 	composable_pairs(U, A, U0, V0) ->
			subst_node(FE, [Ul->Ur], U0, V0, R, R0, S)
		;	add_link(FE, U, R, R1, S),
			zdd_insert(A, R1, R0, S)
		)
	),
	zdd_join(L0, R0, Y, S).
%
subst_node(_, _, _, _, X, 0, _):- X < 2, !.
subst_node(FE, Es, A, P, X, Y, S):-	 FE = Fr-End,		% replace A with P
	cofact(X, t(U, L, R), S),
 	subst_node(FE, Es, A, P, L, L0, S),
	classify_link(Fr, End, U, Case),
	arrow_symbol(U, _, Lu, Ru),
	(	( Case = 0 ; Case = arrow ) -> R0 = 0
	;	Case = ignore ->
		subst_node(FE, Es, A, P, R, R0, S)
	;	(	Ru = A	->
			normal_pair(Lu-P, V),
			zdd_ord_insert([V|Es], R, R0, S)
		;	Lu = A	->
			normal_pair(P-Ru, V),
			zdd_ord_insert([V|Es], R, R0, S)
		;	subst_node(FE, Es, A, P, R, R1, S),
			zdd_insert(U, R1, R0, S)
		)
	),
	zdd_join(L0, R0, Y, S).

		/********************
	    *     prune mates   *
		********************/

% ?- #(S), X<< +[[a-b, a->b]], prune_final(a, b, X, Y, S).
prune_final(P, P, _, 1, _):-!.
prune_final(_, _, X, 0, _):- X<2, !.
prune_final(P, Q, X, Y, S):- cofact(X, t(A, L, R), S),
	prune_final(P, Q, L, L0, S),
	(	A = (_->_) -> R0 = 0
	;  	A = P-Q -> prune_final0(R, R0, S)
	;	A = V-V -> prune_final(P, Q, R, R0, S)
	;	R0 = 0
	),
	zdd_join(L0, R0, Y, S).
%
prune_final0(X, X, _):- X<2, !.
prune_final0(X, Y, S):- cofact(X, t(A, L, R), S),
	prune_final0(L, L0, S),
	(	A = (_->_) -> zdd_insert(A, R, R0, S)
  	;	A = (B-B) -> prune_final0(R, R0, S)
	;	R0 = 0
	),
	zdd_join(L0, R0, Y, S).

		/***********************************
		*     classify_link by frontier    *
		***********************************/

%  "A node P is on frontier" means that P may be touched by a remaining link in the future.

% ?- on_frontier(3, 4). % true
% ?- on_frontier(4, 3). % false
on_frontier(P, F):- P @=< F.
%
classify_link(_, _, _->_, arrow):-!.
classify_link(F, End, A-B,  Case):- on_frontier(A, F), !,
   	(	on_frontier(B, F) -> Case = keep
	;	B = End -> Case = keep
	;	Case = 0
	).
classify_link(_, E, E-E, 0):-!.
classify_link(_, _, A-A, ignore):-!.
classify_link(_, _, _, 0).

%
prune_by_classify_link(_, _, X, X, _):- X<2, !.
prune_by_classify_link(F, End, X, Y, S):- cofact(X, t(A, L, R), S),
	prune_by_classify_link(F, End, L, L0, S),
	classify_link(F, End, A, Case),
	(	Case = arrow -> zdd_insert(A, R, R0, S)
	;	Case = keep ->					% many hits.
		prune_by_classify_link(F, End, R, R1, S),
		zdd_insert(A, R1, R0, S)
	;	Case = ignore ->				% many hits.
		prune_by_classify_link(F, End, R, R0, S)
	;	R0 = 0							% many bits.
	),
	zdd_join(L0, R0, Y, S).


		/*******************************
		*  printing mate for  debug    *
		*******************************/

% ?- #(S), X<< +[[a-b],[b-c]],  pmate(X, S).
pmate(X, S):- setup_call_cleanup(
				open_state(M),
				(	drop_path(X, Y, S, M),
					psa(Y, M)
				),
				close_state(M)).

%
drop_path(X, Y, _, _):- X<2, !, Y=X.
drop_path(X, Y, S, M):- cofact(X, t(A, L, R), S),
	drop_path(L, L0, S, M),
	(	A=(_-_) ->
		drop_path(R, R1, S, M),
		zdd_insert(A, R1, R0, M)
	;	R0 = 1
	),
	zdd_join(L0, R0, Y, M).
