:- module(udg_path_count, []).

:- use_module(zdd('zdd-array')).
:- use_module(zdd(zdd)).
:- use_module(zdd('minato-r5')).
:- use_module(pac(op)).

		/*******************
		*   Tiny helpers   *
		*******************/

% ?- arrow_symbol(_->_, F).
% ?- arrow_symbol(a->b, F, X, Y).
arrow_symbol( _ -> _).
%
arrow_symbol(A, A0):- functor(A, A0, 2).
arrow_symbol(A, A0, A1, A2):- functor(A, A0, 2),
		arg(1, A, A1),
		arg(2, A, A2).

%
mate_less_than(_ - A, B -_):- A @< B.

% One of the most basic helpers.
composable_pairs(A-B, A-C, B, C).
composable_pairs(A-B, C-A, B, C).
composable_pairs(B-A, A-C, B, C).
composable_pairs(B-A, C-A, B, C).
%
normal_pair(A-B, U-V):-!, ( B @< A -> U=B, V=A; U=A, V=B ).
normal_pair(A->B, U->V):- ( B @< A -> U=B, V=A; U=A, V=B ).


		/****************************************************
		*     replace_end/bypass:  Most basic operations.   *
		****************************************************/

% ?- zdd((X<< +[[c-d]], @(subst_node([a->c], c, a, X, Y)), psa(X), psa(Y))).
subst_node(_, _, _, X, 0, _):- X<2, !.
subst_node(Es, A, P, X, Y, S):-			% replace A with P
	cofact(X, t(U, L, R), S),
	arrow_symbol(U, Arrow, Lu, Ru),
	(	Arrow = (->) ->	Y = 0
	;	compare(<, A, Lu) ->	Y = 0
	;	subst_node(Es, A, P, L, L0, S),
		(	Ru = A	->
			normal_pair(Lu-P, V),
			zdd_ord_insert([V|Es], R, R0, S)
		;	Lu = A	->
			normal_pair(Ru-P, V),
			zdd_ord_insert([V|Es], R, R0, S)
		;	subst_node(Es, A, P, R, R1, S),
			zdd_insert(U, R1, R0, S)
		),
		zdd_join(L0, R0, Y, S)
	).

		/**********************************
		*     simple count path by mate   *
		**********************************/

% ?- rect_udg(rect(1,1), Ls,  M), udg_path_count_with_sort(Ls, M, C, _).
% ?- rect_udg(rect(2,1), Ls,  M), udg_path_count_with_sort(Ls, M, C, _).
% ?- rect_udg(rect(3,1), Ls,  M), udg_path_count_with_sort(Ls, M, C, _).
% ?- rect_udg(rect(4,1), Ls,  M), udg_path_count_with_sort(Ls, M, C, _).
% ?- rect_udg(rect(1,2), Ls,  M), udg_path_count_with_sort(Ls, M, C, _).
% ?- rect_udg(rect(1,3), Ls,  M), udg_path_count_with_sort(Ls, M, C, _).
% ?- rect_udg(rect(1,4), Ls,  M), udg_path_count_with_sort(Ls, M, C, _).
% ?- rect_udg(rect(2,2), Ls,  M), udg_path_count_with_sort(Ls, M, C, _).
% ?- time((rect_udg(rect(3,3), Ls,  M), udg_path_count_with_sort(Ls, M, C, _))).
% ?- rect_udg(rect(4,4), Ls,  M),
%	call_with_time_limit(600, udg_path_count_with_sort(Ls, M, C, _)).
% ?- rect_udg(rect(5,5), Ls,  M),
%	call_with_time_limit(100, udg_path_count_with_sort(Ls, M, C, _)).
% ?- bfs_rect_path_count(rect(1,1), C, _).
% ?- bfs_rect_path_count(rect(2,2), C, _).
% ?- bfs_rect_path_count(rect(3,3), C, _).
% ?- bfs_rect_path_count(rect(4,4), C, _).

bfs_rect_path_count(R, C, Opts):- % Links is used as it is.
	rect_udg(R, Ls,  M),
	bfs_path_count(Ls, M, C, Opts).

% ?- bfs_path_count([a-b, b-c,  a-c], a-c, C, _).
% ?- bfs_path_count([a-b, a-c, b-d, c-d, c-e, e-f, d-f, a-f], a-f, C, _).
bfs_path_count(Links, A-B, C, Opts):- % Links is used as it is.
	memberchk(state(S), Opts),
	memberchk(path(Y), Opts),
	memberchk(mate(X), Opts),
	open_state(S),
	findall(W, (member(M, Links), (M = (W-_); M=(_-W))), Ps),
	sort([A,B|Ps], Qs),
	sort([A,B], [N0, N1]),
	set_key(final, N0-N1, S),
	findall(U-U, member(U, Qs), M),
	sort(M, M0),
	zdd_insert_atoms(M0, 1, Mates, S),
	sort(Links, Links0),
	bfs_path_count(Links0, Mates, X, 0, Y, S),
	card(Y, C, S).

%
bfs_path_count([], X, X, Y, Y, _).
bfs_path_count([L|Ls], X, X0, Y, Y0, S):-
	bfs_path_count(Ls, X, X1, Y, Y1, S),
	add_links([L], X1, X0, Y1, Y0, S).

% sort give links, and pass it.
% ?- rect_udg(rect(1, 3), Ls,  M),
%	call_with_time_limit(100, udg_path_count_with_sort(Ls, M, C, _)).
%@ C = 8.
% ?- rect_udg(rect(2,2), Ls,  M),
%	call_with_time_limit(100, udg_path_count_with_sort(Ls, M, C, _)).
% ?- rect_udg(rect(2,2), Ls,  M),
% ?- rect_udg(rect(3,3), Ls,  M),
%	call_with_time_limit(100, udg_path_count_with_sort(Ls, M, C, _)).
% ?- rect_udg(rect(4,4), Ls,  M),
%	call_with_time_limit(100, udg_path_count_with_sort(Ls, M, C, _)).
% ?- rect_udg(rect(2,2), X, Y), udg_path_count(X, Y, C, _).

% sort give links, and pass it.
udg_path_count_with_sort(Links, Final, C, Opts):-
	sort(Links, Links0),
	reverse(Links0, Links1),
	udg_path_count(Links1, Final, C, Opts).
%
udg_path_count(Links, A-B, C, Opts):- % Links is used as it is.
	memberchk(state(S), Opts),
	memberchk(path(Y), Opts),
	memberchk(mate(X), Opts),
	open_state(S),
	findall(W, (member(M, Links), (M = (W-_); M=(_-W))), Ps),
	sort([A,B|Ps], Qs),
	sort([A,B], [N0, N1]),
	set_key(final, N0-N1, S),
	findall(U-U, member(U, Qs), M),
	sort(M, M0),
	zdd_insert_atoms(M0, 1, Mates, S),
	bfs_sort_links([N1], Links, [], L_links),
	reverse(L_links, L_links_r),
	add_layered_links(L_links_r, Mates, X, 0, Y, S),
	card(Y, C, S).
%
add_layered_links([], X, X, Y, Y, _).
add_layered_links([L|Ls], X, X0, Y, Y0, S):-
	add_links(L, X, X1, Y, Y1, S),
	add_layered_links(Ls, X1, X0, Y1, Y0, S).

% ?- zdd((X<<{[a-a, b-b]}, set_key(final, a-b), @(add_links([a-b], X, X0, 0, Y)), psa(Y))).
% ?- zdd((@(set_key(final, c-d)), X<<{[a-a, b-b]}, @(add_links([a-b], X, X0, 0, Y)), psa(X0))).
% ?-rect_udg(rect(3,3), Ls, E), findall(U-U, (member(D, Ls), (D=(U-_); D=(_-U))), M), sort(M, M0),
% zdd((X<<{M0}, set_key(final,E), @(add_links(Ls, X, X0, 0, Y)), card(Y, Card))).
%@ Ls = [p(0, 0)-p(0, 1), p(0, 0)-p(1, 0), p(0, 1)-p(0, 2), p(0, 1)-p(1, 1), p(0, 2)-p(0, 3), p(0, 2)-p(1, 2), p(0, 3)-p(1, 3), p(..., ...)-p(..., ...), ... - ...|...],
%@ E = p(0, 0)-p(3, 3),
%@ M = [p(0, 0)-p(0, 0), p(0, 1)-p(0, 1), p(0, 0)-p(0, 0), p(1, 0)-p(1, 0), p(0, 1)-p(0, 1), p(0, 2)-p(0, 2), p(0, 1)-p(0, 1), p(..., ...)-p(..., ...), ... - ...|...],
%@ M0 = [p(0, 0)-p(0, 0), p(0, 1)-p(0, 1), p(0, 2)-p(0, 2), p(0, 3)-p(0, 3), p(1, 0)-p(1, 0), p(1, 1)-p(1, 1), p(1, 2)-p(1, 2), p(..., ...)-p(..., ...), ... - ...|...],
%@ X = 17,
%@ X0 = 15726402,
%@ Y = 14297372,
%@ Card = 184.

add_links([], X, X, C, C, _).
add_links([L|Ls], X, X0, C, C0, S):-
	add_link(L, X, X1, S),
	get_key(final, E, S),
	separate_finals(X1, Finals, X2, E, S),   % experimental
	zdd_join(C, Finals, C1, S),
	zdd_join(X, X2, X3, S),
	add_links(Ls, X3, X0, C1, C0, S).


% ?- <<(X, +[[a-a, b-b, c-c]], S),
%	simple_add_links([a-b,a-c,b-c], X, X0, S), psa(X0, S).

simple_add_links([], X, X, _).
simple_add_links([L|Ls], X, X0, S):-
	add_link(L, X, X1, S),
	zdd_join(X, X1, X2, S),
	simple_add_links(Ls, X2, X0, S).


% ?- zdd((X<<{[a-b, c-d]}, psa(X), @(add_link(b-c, X, Y)), psa(Y))).
% ?- zdd((X<< +[[a-b, c-d], [e-f]], psa(X), @(add_link(b-c, X, Y)), psa(Y))).
% ?- zdd((X<< +[[a-b, c-d], [a-a, b-b], [e-f]], psa(X), @(add_link(b-c, X, Y)), psa(Y))).
% ?- zdd((X<< +[[a-a, b-b, c-c]], psa(X), @(add_link(b-c, X, Y)), psa(Y))).

add_link(_, X, 0, _):- X<2, !.
add_link(U, X, Y, S):- % memo is useless here
	cofact(X, t(A, L, R), S),
	(	A = (_->_) -> Y = 0
	;	add_link(U, L, L0, S),
		U = Ul-Ur,
		(	mate_less_than(U, A) -> R0 = 0
		; 	A = U -> R0 = 0
		; 	composable_pairs(U, A, U0, V0) ->
			subst_node([Ul->Ur], U0, V0, R, R0, S)
		;	add_link(U, R, R1, S),
			zdd_insert(A, R1, R0, S)
		),
		zdd_join(L0, R0, Y, S)
	).

% ?- zdd((X<<{[a-b, c-c, x->y]}, @(separate_finals(X, Y, Z, a-b)), psa(X), psa(Z), psa(Y))).
% ?- zdd((X<<{[a-b, c-c, x->y], [c-d, a->b], [a-b, c-d], [a-b, d-d]},
%	@(separate_finals(X, Y, Z, a-b)), psa(X), psa(Z), psa(Y))).

separate_finals(X, Fin, Z, E, S):- zdd_divisible_by_list(X, [E], X0, S),
	separate_non_finals(X0, Non, E, S),
	zdd_subtr(X0, Non, Fin, S),
	zdd_subtr(X, X0, Z, S).
%
separate_non_finals(X, 0, _, _):- X<2, !.
separate_non_finals(X, Y, E, S):- cofact(X, t(A, L, R), S),
	separate_non_finals(L, L0, E, S),
	(	A = (U-V) ->
		(	(A = E; U=V) -> separate_non_finals(R, R0, E, S)
		;	R0 = R
		)
	;	R0 = 0
	),
	cofact(Y, t(A, L0, R0), S).

%
collect_closed_mate(X, 0, _, _):- X<2, !.
collect_closed_mate(X, Y, E, S):- cofact(X, t(A, L, R), S),
	(	A = (_->_)	-> Y = 0
	;	A = E ->
		collect_mate_free(R, R0, S),
		cofact(Y, t(A, 0, R0), S)
	; 	collect_closed_mate(L, L0, E, S),
		(	A = (U-U) ->
			collect_closed_mate(R, R0, E, S)
		;	R0 = 0
		),
		cofact(Y, t(A, L0, R0), S)
	).
%
collect_mate_free(X, Y, _):- X<2, !, Y=X.
collect_mate_free(X, Y, S):- cofact(X, t(A, L, R), S),
	(	A = (_->_) -> Y = X
	;	collect_mate_free(L, L0, S),
		(	A = (U-U) ->
			collect_mate_free(R, R0, S)
		;	R0 = 0
		),
		cofact(Y, t(A, L0, R0), S)
	).

		/***********************
		*     layered links    *
		***********************/

% ?- bfs_sort_links([a], [a-b, b-c, d-e, a-c], [],  Out).
% ?- time((rect_udg(rect(1,1), Ls, A-B), bfs_sort_links([B], Ls, [], Y))).
% ?- time((rect_udg(rect(100,100), Ls, A-B),
%	bfs_sort_links([B], Ls, [], Y))), length(Y, Len).
%
add_new(X, Y, Y):- memberchk(X, Y), !.
add_new(X, Y, [X|Y]).

bfs_sort_links([], _, Y, Y).
bfs_sort_links(X, M, Y, Z):-
	bfs_layer_list(X, [], L, [], K, M, M0),
	bfs_sort_links(K, M0, [L|Y], Z).

% ?- bfs_layer_list([a], [], L0, [], K0, [a-b], M0).
% ?- bfs_layer_list([a, b], [], L0, [], K0, [a-b, b-c, a-c], M0).
% ?- bfs_layer_list([a, b], [], L0, [], K0, [a-b, b-c, d-e, a-c], M0).
%
bfs_layer_list([], L, L, K, K, M, M).
bfs_layer_list([N|Ns], L, L0, K, K0, M, M0):-
	bfs_layer(N, L, L1, K, K1, M, M1),
	bfs_layer_list(Ns, L1, L0, K1, K0, M1, M0).
%
bfs_layer(N, L, L0, K, K0, M, M0):-
	select(X, M, M1),
	(X = (P-N); X = (N-P)),
	!,
	add_new(X, L, L1),
	add_new(P, K, K1),
	bfs_layer(N, L1, L0, K1, K0, M1, M0).
bfs_layer(_, L, L, K, K, M, M).

		/*************************
		*     linear grid udg    *
		*************************/

% ?- zdd((@rect_col_bridge(rect(0,0), Cols, Brs))).
% ?- zdd((@rect_col_bridge(rect(0,1), Cols, Brs))).
% ?- zdd((@rect_col_bridge(rect(1,1), Cols, Brs))).
% ?- time(zdd((@rect_col_bridge(rect(20,20), Cols, Brs)))).
%@ % 4,157,825,846 inferences, 452.733 CPU in 468.459 seconds (97% CPU, 9183838 Lips)

% ?- time(zdd((@rect_col_bridge(rect(20,20), Cols, fBrs)))).
%@ % 4,157,825,846 inferences, 551.569 CPU in 559.546 seconds (99% CPU, 7538185 Lips)

rect_col_bridge(Rect, Cols, Brs, S):-
	rect_col_list(Rect, Cols, S),
	rect_bridge(Rect, Brs).
%
rect_bridge(rect(W, H), Brs):-
	(	W > 0 ->
		W0 is W-1,
		mk_bridge_list(0-W0, 0-H, Brs)
	;	Brs = []
	).


% ?- zdd((@rect_col_list(rect(1,1), Vs), @s_map(psa, Vs))).
% ?- time(zdd(@rect_col_list(rect(15,15), Vs))).

rect_col_list(rect(W, H), Vs, S):-
	mk_column_list(rect(p(0,0), p(W, H)), Cols),
	s_map(linear_grid_udg, Cols, Vs, S).

% ?- mk_column_list(rect(p(0,0), p(1,1)), Cols).
mk_column_list(rect(p(J0, K0), p(J1, K1)), Cols):-
	numlist(J0, J1, Js),
	numlist(K0, K1, Ks),
	maplist(mk_column(Ks), Js, Cols).
%
mk_column(Ks, J, Col):- findall(p(J, K), member(K, Ks), Col).

% ?- zdd((@(linear_grid_udg([a], X)), psa(X))).
% ?- zdd((@(linear_grid_udg([a,b], X)), psa(X))).
% ?- zdd((@(linear_grid_udg([a,b,c,d], X)), psa(X))).

linear_grid_udg([], 1, _).
linear_grid_udg([J|Js], X, S):- linear_grid_udg(Js, X0, S),
	add_node_grid_udg(J, Js, X0, X, S).
%
add_node_grid_udg(J, Js, X, Y, S):- zdd_insert(J-J, X, Y0, S),
	add_links_grid_udg(J, Js, X, 0, Y1, S),
	zdd_join(Y0, Y1, Y, S).
%
add_links_grid_udg(_, [], _, Y, Y, _):-!.
add_links_grid_udg(J, [K|_], X, Y, Y0, S):-add_link_grid_udg(J-K, X, Y1, S),
	zdd_join(Y, Y1, Y0, S).

%
add_link_grid_udg(_, 0, 0, _):-!.
add_link_grid_udg(J-K, 1, Y, S):-!, zdd_ord_insert([J-K, J->K], 1, Y, S).
add_link_grid_udg(J-K, X, Y, S):- % memo is useless here.
	cofact(X, t(U, L, R), S),
	U=A-B,
	add_link_grid_udg(J-K, L, L0, S),
	(	K = A ->
		zdd_ord_insert([J-B, J->K], R, R0, S)
	;   throw(unexpected_link)
	),
	zdd_join(L0, R0, Y, S).

% ?- mk_bridge(2, 0-3, Bridge).
mk_bridge(I, J-K, Bridge):- findall(p(I,B)-p(I0,B), ( between(J, K, B), succ(I, I0) ), Bridge).

% ?- mk_bridge_list(0-1, 0-3, Bs), maplist(writeln, Bs).
%@ [p(0,0)-p(1,0),p(0,1)-p(1,1),p(0,2)-p(1,2),p(0,3)-p(1,3)]
%@ [p(1,0)-p(2,0),p(1,1)-p(2,1),p(1,2)-p(2,2),p(1,3)-p(2,3)]
mk_bridge_list(Min-Max, J-K, Bs):-
	findall(B, (between(Min, Max, I), mk_bridge(I, J-K, B)), Bs).


% ?- test(rect(0,0), X, S), psa(X, S).
% ?- test(rect(1,0), X, S), psa(X, S).
% ?- test(rect(2,0), X, S), psa(X, S).
% ?- test(rect(30,0), X, S), psa(X, S).
% ?- test(rect(1,1), X, S), psa(X, S).
% ?- test(rect(1,1), X, S), card(X, C, S).
% ?- test(rect(2,2), X, S), card(X, C, S).
% ?- test(rect(3,3), X, S), card(X, C, S).
% ?- test(rect(4,4), X, S), card(X, C, S).
% ?- time((test(rect(5,5), X, S), card(X, C, S))).
%@ % 698,911,122 inferences, 87.332 CPU in 88.565 seconds (99% CPU, 8002940 Lips)
%@ X = 7971232,
%@ S = ..,
%@ C = 1262816.

test(rect(W, H), X, S):- open_state(S),
	set_key(max_node, p(W, H), S),
	W0 is W-1,
	set_key(bridge_id, W0, S),
	rect_col_bridge(rect(W, H), Cols, Brs, S),
	reverse(Cols, RCols),
	maplist(reverse, Brs, Brs0),
	reverse(Brs0, RBrs),
	RCols=[Col|Cols0],
	fold_columns(RBrs, Cols0, Col, X0, S),
	prune_final(p(0, 0), p(W, H), X0, X, S).

%
fold_columns([], [], X, X, _).
fold_columns([B|Bs], [C|Cs], X, Y, S):-
	bridge_prune(B, C, X, X0, S),
	get_key(bridge_id, K, S),
	K0 is K-1,
	set_key(bridge_id, K0, S),
	fold_columns(Bs, Cs, X0, Y, S).

% ?- path_count(rect(0,0), X, S), psa(X, S), get_compare(F, S).
% ?- path_count(rect(1,0), X, S), card(X, C, S).
% ?- path_count(rect(2,0), X, S), card(X, C, S).
% ?- path_count(rect(30,0), X, S), card(X, C, S).
% ?- path_count(rect(1,1), X, S), card(X, C, S).
% ?- path_count(rect(2,2), X, S), card(X, C, S).
% ?- path_count(rect(3,3), X, S), card(X, C, S).
% ?- time((path_count(rect(4,4), X, S), card(X, C, S))).
% ?- time((path_count(rect(5,5), X, S), card(X, C, S))).
%@ % 659,736,449 inferences, 56.014 CPU in 56.567 seconds (99% CPU, 11778141 Lips)
%@ % 659,736,449 inferences, 78.308 CPU in 78.596 seconds (100% CPU, 8424911
%@ % 659,704,102 inferences, 61.864 CPU in 62.239 seconds (99% CPU, 10663748 Lips)
%@ C = 1262816.

path_count(rect(W, H), X, S):- open_state(S),
	set_key(max_node, p(W, H), S),
	set_key(bridge_id, W, S),
	rect_bridge(rect(W, H), Brs),
	maplist(reverse, Brs, Brs0),
	reverse(Brs0, RBrs),
	next_column(Col, S),
	fold_columns(RBrs, Col, X0, S),
	prune_final(p(0, 0), p(W, H), X0, X, S).

% ?- open_state(S), set_key(bridge_id, 2, S), set_key(max_node, p(4,4), S), next_column(X, S), psa(X, S).
next_column(X, S):- get_key(bridge_id, I, S),
	get_key(max_node, N, S),
	N = p(_, H),
	next_column(I, H, X, S).
%
next_column(I, H, X, S):- findall(p(I, J), between(0, H, J), Ps),
	linear_grid_udg(Ps, X, S).

%
fold_columns([], X, X, _).
fold_columns([B|Bs], In, Y, S):-
	zdd_slim(In, X, S),
	get_key(bridge_id, K, S),
	K0 is K-1,
	set_key(bridge_id, K0, S),
	next_column(C, S),
	bridge_prune(B, C, X, X0, S),
	fold_columns(Bs, X0, Y, S).

%
bridge_prune(Bs, X, Y, Z, S):- bridge(Bs, X, Y, 0, Z0, S),
	get_key(max_node, P, S),
	get_key(bridge_id, I, S),
	prune(I, P, Z0, Z1, S),
	zdd_join(X, Z1, Z, S).

% ?- zdd((X<< +[[a-b]], Y<< +[[x-y]], @bridge([a-x], X, Y, 0, Z), psa(Z))).
% ?- zdd((X<< +[[a-b],[a-c]], Y<< +[[x-y], [x-z]], @bridge([a-x], X, Y, 0, Z), psa(Z))).
% ?- zdd((X<< +[[a-b],[a-c]], Y<< +[[x-y], [x-z]], @bridge([a-x, b-y], X, Y, 0, Z), psa(Z))).
bridge([], _, _, Z, Z, _):-!.
bridge([U|Us], X, Y, Z, Z0, S):-
	add_link(U, Z, Z1, S),
	bridge(U, X, Y, W, S),
	zdd_join_list([W, Z1, Z], Z2, S),
	bridge(Us, X, Y, Z2, Z0, S).

% ?- zdd((X<< +[[a-b]], Y<< +[[x-y]],  @bridge(a-x, X, Y, Z), psa(Z))).
% ?- zdd((X<< +[[a-b], [a-c]], Y<< +[[x-y]],  @bridge(a-x, X, Y, Z), psa(Z))).
% ?- zdd((X<< +[[a-b], [a-c]], Y<< +[[x-y], [x-u]],  @bridge(a-x, X, Y, Z), psa(Z))).
bridge(_, X, Y, 0, _):- (X<2; Y<2), !.
bridge(U, X, Y, Z, S):- cofact(X, t(A, L, R), S),
		(	A = (_->_) -> Z = 0
		;	A = A0-A1,
			U = J-K,
			bridge(U, L, Y, L0, S),
			(	J @< A0 ->  R0 = 0
			;	J = A0 ->
				subst_node([J->K], K, A1, Y, Y1, S),
				zdd_merge(Y1, R, R0, S)
			;	J = A1 ->
				subst_node([J->K], K, A0, Y, Y1, S),
				zdd_merge(Y1, R, R0, S)
			;	bridge(U, R, Y, Y0, S),
				zdd_insert(A, Y0, R0, S)
			),
			zdd_join(L0, R0, Z, S)
		).

% ?- zdd((X<< *[[p(1, 0)-p(2,2)]], @prune(1, p(2,2), X, Y), psa(Y))).
% ?- zdd((X<< +[[p(1, 0)-p(2,1)], [p(1,1)-p(1,2), p(1,0)-p(2,2)], [p(1,1)-p(2,2)]],
%	psa(X), @prune(1, p(2,2), X, Y), psa(Y))).
%
prune(_, _, X, X, _):-X<2, !.
prune(I, P, X, Y, S):-cofact(X, t(A, L, R), S),
	(	A = (_->_) -> Y = X
	;	prune(I, P, L, L0, S),
		A = A0-A1,
		A0 = p(J, _),
		A1 = p(K, _),
		(	J = I ->
			(	( K = I; A1 = P) ->
				prune(I, P, R, R1, S),
				zdd_insert(A, R1, R0, S)
			;	R0 = 0
			)
		;	A0 \== A1 -> R0 = 0
		;	prune(I, P, R, R0, S)
		),
		zdd_join(L0, R0, Y, S)
	).

% ?- zdd((X<< +[[a-b, a->b]], @prune_final(a, b, X, Y), psa(Y))).
prune_final(_, _, X, X, _):- X<2, !.
prune_final(P, Q, X, Y, S):- cofact(X, t(A, L, R), S),
	(	A = (_->_) -> Y = X
	;	prune_final(P, Q, L, L0, S),
		A = A1-A2,
		(	 A1 = P ->
			 (	A2 = Q
			 -> prune_final(P, Q, R, R0, S)
			 ;	R0 = 0
			 )
		;	A1 = A2 ->
			prune_final(P, Q, R, R0, S)
		;	R0 = 0
		),
		zdd_join(L0, R0, Y, S)
	).
