:- module(count_algebra, [bin_symm_funs/3]).
:- use_module(library(apply)).
:- use_module(library(apply_macros)).
:- use_module(library(clpfd)).
:- use_module(library(statistics)).
:- use_module(zdd('zdd-array')).
:- use_module(util(meta2)).
:- use_module(pac(basic)).
:- use_module(pac(meta)).
:- use_module(util(math)).
:- use_module(pac('expand-pac')). % For the kind block.
:- use_module(zdd('zdd-misc')).
:- use_module(zdd(zdd)).
:- use_module(pac(op)).

:- set_prolog_flag(stack_limit, 10_200_147_483_648).

 :- op(1060, xfy, ~).		% equivalence
 :- op(1060, xfy, #).		% exor
 :- op(1060, xfy, <->).		% equivalence
 :- op(1050, yfx, <-).
 :- op(1060, xfy, <=> ).	% equivalence
 :- op(1040, xfy, \/).		% OR
 :- op(1030, xfy, /\).		% AND
 :- op(1020, fy, \).		% NOT
 :- op(700, xfx, :=).		% Assignment
 :- op(1000, xfy, &).

% For pac query.
 :- pac:op(1000, xfy, &).
 :- pac:op(700, xfx, :=).

term_expansion --> pac:expand_pac.

% :- meta_predicate(@(:)).
% @(X):- shift(X).

join(X, Y, Z):- @(join(X, Y, Z)).
%
join(X, Y, Z, S):- zdd_join(X, Y, Z, S).

%
hyphen([],[],[]).
hyphen([X|Xs],[Y|Ys],[X-Y|Zs]):- hyphen(Xs, Ys, Zs).
%
mk_pair(X, Y, [X,Y]).
% ?- prod_pair([1,2],[a,b,c], X).
prod_pair(A, B, C):- product(mk_pair, A, B, C).


% ?- Oprs=[[1,1]-1,[1,2]-2,[1,3]-3,[1,4]-4,[1,5]-5,[1,6]-6,[2,1]-2,[2,2]-6,[2,3]-5,[2,4]-3,[2,5]-4,[2,6]-1,[3,1]-3,[3,2]-4,[3,3]-1,[3,4]-2,[3,5]-6,[3,6]-5,[4,1]-4,[4,2]-5,[4,3]-6,[4,4]-1,[4,5]-2,[4,6]-3,[5,1]-5,[5,2]-3,[5,3]-2,[5,4]-6,[5,5]-1,[5,6]-4,[6,1]-6,[6,2]-1,[6,3]-4,[6,4]-5,[6,5]-3,[6,6]-2],
%	alg_der(Oprs, Der), maplist(writeln, Der).

% Sample for 1*(2*3) not equal to    (1*2)*3
% ?-   X = [[2,3]-4, [1,4]-5, [1,2]-6, [6,3]-7], alg_der(X, Y).
% => false
% ?-   X = [[2,3]-6, [1,6]-6, [1,2]-2, [2,3]-6],
%	alg_der(X, Y).

merge_xalg_der([], X, X).
merge_xalg_der([P|Ps], X, Y):- merge_in_der(P, X, X0),
	merge_xalg_der(Ps, X0, Y).

% ?- merge_in_der(a-b*c, [a-[b*c]], R).
% ?- merge_in_der(a-c*d, [a-[b*c]], R).
% ?- merge_in_der(a-c*d, [a-[c*e]], R).  % false
merge_in_der(X, Y, Z):- merge_in_der_(Y, X, Z).
%
merge_in_der_([], A-P, [A-[P]]).
merge_in_der_([A-L|R], A-P, [A-L0|R]):-!,
	add_ord_keyval(P, L, L0).
merge_in_der_([A0-L|R], A-P, [A0-L|R0]):- A0 @< A,!,
	merge_in_der_(R, A-P, R0).
merge_in_der_(R, A-P, [A-[P]|R]).

% ?- add_ord_keyval(a*1, [], L).
% ?- add_ord_keyval(a*1, [a*2], L).  % false
% ?- add_ord_keyval(b*1, [a*1, c*3], L).
% ?- add_ord_keyval(d*1, [a*1, c*3], L).
% ?- add_ord_keyval(a*1, [b*2, c*3], L).
add_ord_keyval(K*V, [K*U|L], [K*V|L]):-!, U=V.
add_ord_keyval(P, [Q|L], [Q|L0]):- P = K*_, Q = K0*_,
	K0 @< K, !,
	add_ord_keyval(P, L, L0).
add_ord_keyval(P, L, [P|L]).

% ?- check_ord_keyval(a*1, [], L).
% ?- check_ord_keyval(a*V, [], L).  % false
% ?- check_ord_keyval(a*1, [a*2], L).  % false
% ?- check_ord_keyval(b*1, [a*1, c*3], L).
% ?- check_ord_keyval(d*1, [a*1, c*3], L).
% ?- check_ord_keyval(a*1, [b*2, c*3], L).
check_ord_keyval(K*V, L, L):- var(V), !, check_ord_keyval(K*V, L).
check_ord_keyval(P, L, L0):- add_ord_keyval(P, L, L0).

% ?- check_ord_keyval(b*2, [a*1, b*2, c*3]).
% ?- check_ord_keyval(b*3, [a*1, b*2, c*3]).  % false
% ?- check_ord_keyval(b*X, [a*1, b*2, c*3]).
% ?- check_ord_keyval(b*X, [a*1, c*3]).	% false
check_ord_keyval(K*V, [K*U|_]):-!, U=V.
check_ord_keyval(P, [K0*_|L]):- P=K*_, K0 @< K,
	check_ord_keyval(P, L).

% ?- time(find_model(2, B, S)).
% ?- time(find_model(3, B, S)).
% ?- time(find_model(4, B, S)).
% ?- time(find_model(5, B, S)).
% ?- time(find_model(6, B, S)), psa(B, S).
% ?- time(find_model(6, B, S)), psa(B, S). % [2021/12/24]
%@ % 896,018,635,494 inferences, 56245.613 CPU in 110023.200 seconds (51% CPU, 15930463 Lips)
%@  zid = 5677
%@ 	[[1,1]-1,[1,2]-2,[1,3]-3,[1,4]-4,[1,5]-5,[1,6]-6,[2,1]-2,[2,2]-6,[2,3]-5,[2,4]-3,[2,5]-4,[2,6]-1,[3,1]-3,[3,2]-4,[3,3]-1,[3,4]-2,[3,5]-6,[3,6]-5,[4,1]-4,[4,2]-5,[4,3]-6,[4,4]-1,[4,5]-2,[4,6]-3,[5,1]-5,[5,2]-3,[5,3]-2,[5,4]-6,[5,5]-1,[5,6]-4,[6,1]-6,[6,2]-1,[6,3]-4,[6,4]-5,[6,5]-3,[6,6]-2]
%@ B = 5677,
%@ S = ..
% ?- time(find_model(7, B, S)).  % Not terminate ?

find_model(N, B, S):-
	open_state(S),
	numlist(1, N, D),
	filter_left_unit(D, A, S),
	filter_non_commutative(2, 3, A, A0, S),
	filter_inversible(D, A0, A1, S),
	find_associative(A1, [], [], B, S).
%
find_associative(A, _, _, A, _):- A<2, !.
find_associative(A, Alg, Der, B, S):- cofact(A, t(Eqn, L, R), S),
	find_associative(L, Alg, Der, L0, S),
	(	L0 > 1 -> R0 = 0
	;	(	merge_in_der(Eqn, Alg, Der, Der0)
		-> 	find_associative(R, [Eqn|Alg], Der0, R0, S)
		;	R0 = 0
		)
	),
	cofact(B, t(Eqn, L0, R0), S).

%
filter_associative(A, B, S):- filter_associative(A, [], [], B, S).
%
filter_associative(A, _, _, A, _):- A<2, !.
filter_associative(A, Oprs, Der, B, S):- cofact(A, t(Opr, L, R), S),
	filter_associative(L, Oprs, Der, L0, S),
	(	merge_in_der(Opr, Oprs, Der, Der0)
	-> 	filter_associative(R, [Opr|Oprs], Der0, R0, S)
	;	R0 = 0
	),
	cofact(B, t(Opr, L0, R0), S).

% ?- eqn_der([1,1]-1, [], P), sort(P, P0).
% ?- eqn_der([1,2]-1, [[1,3]-1], P), sort(P, P0).
eqn_der(L, Ls, P):-	apply_eqn_one(L, [L|Ls], D, []),
	sort(D, P).
%
apply_eqn_one(_, [], P, P):-!.
apply_eqn_one(L, [M|Ms], P, Q):- basic_der(L, M, P, P0),
	apply_eqn_one(L, Ms, P0, Q).

% ?- basic_der([1,1]-1,[1,1]-1, P, []).
% ?- basic_der([1,2]-3,[3,3]-4, P, []).
% ?- basic_der([1,2]-2,[2,1]-1, P, []).
% ?- basic_der([1,2]-3,[2,1]-4, P, []).
basic_der(P, Q, M, N):- der_rule(P, Q, L),
	collect_der(L, M, N).
%
collect_der([], V, V):-!.
collect_der([X=X, A|U], [A|V], V0):-!, collect_der(U, V, V0).
collect_der([_, _|U], V, V0):- collect_der(U, V, V0).
%
der_rule([A, B]-C, [U, V]-W,
	[	C = U, B-[A,V]*W,
		C = V, A-[U,B]*W,
		W = A, V-[U,B]*C,
		W = B, U-[A,V]*C
	]).

%
alg_der(Alg, Der):- merge_xalg_der(Alg, [], [], Der).
%
merge_xalg_der([], _, Der, Der):-!.
merge_xalg_der([Eqn|Alg], Alg0, Der, Der0):-
	merge_in_der(Eqn, Alg0, Der, Der1),
	merge_xalg_der(Alg, [Eqn|Alg0], Der1, Der0).
%
merge_in_der(Eqn, Alg, Der, Der0):-
	eqn_der(Eqn, Alg, P),
	sort(P, Q),
	merge_xalg_der(Q, Der, Der0).

% ?- left_identy_rule(1, [1,2,3], E).
left_identy_rule(_, [], []):-!.
left_identy_rule(E, [N|Ns], [[1, N]-N |Rest]):-
	left_identy_rule(E, Ns, Rest).

% ?- N = 3, open_state(S), numlist(1, N, D), binary_algebra(D, A, S),
%	filter_associative(A, B, S), card(B, C, S).
% ?- N = 4, open_state(S), numlist(1, N, D), binary_algebra(D, A, S),
%	filter_associative(A, B, S), card(B, C, S).
%
binary_algebra(D, A, S):- raise_list(D, 2, D2),
	zdd_funs(D2, D, A, S).

% ?- open_state(S), filter_left_unit([1], A, S), psa(A, S).
% ?- open_state(S), filter_left_unit([1,2,3], A, S), psa(A, S).
% ?- open_state(S), numlist(1, 3, D), filter_left_unit(D, A, S).

filter_left_unit(D, A, S):- D=[E|D0],
	prod_pair(D0, D, D2),
	zdd_funs(D2, D, A0, S),
	left_identy_rule(E, D, U),
	zdd_append(U, A0, A, S).

% ?- open_state(S), numlist(1, 4, D),
%	filter_left_unit(D, A, S),
%	filter_associative(A,[],[], B, S).
%
filter_inversible([], A, A, _):-!.
filter_inversible(_, A, 0, _):- A<2, !.
filter_inversible(D, A, B, S):-  memo(filter_inversible(D, A)-B, S),
	(	nonvar(B) -> true	%, write(.) % many hits.
	;
	cofact(A, t(P, L, R), S),
	(	P = [_, X]-1
	->  (	select(X, D, D0) -> true
		;	D0 = D
		)
	;	D0 = D
	),
	filter_inversible(D, L, L0, S),
	filter_inversible(D0, R, R0, S),
	cofact(B, t(P, L0, R0), S)
	).
%
filter_non_commutative(_, _, X, 0, _):- X<2, !.
filter_non_commutative(I, J, X, Y, S):- cofact(X, t(P, L, R), S),
	filter_non_commutative(I, J, L, L0, S),
	(	P = [I, J]-K
	->	filter_non_commutative(J, I, K, R, R0, S)
	;	filter_non_commutative(I, J, R, R0, S)
	),
	cofact(Y, t(P, L0, R0), S).

%
filter_non_commutative(_, _, _, X, 0, _):- X < 2, !.
filter_non_commutative(I, J, K,  X, Y, S):- cofact(X, t(Q, L, R), S),
	filter_non_commutative(I, J, K, L, L0, S),
	(	Q = [I,J]-N
	->	(	N\==K -> R0 = R
		;	R0 = 0
		)
	;	filter_non_commutative(I, J, K, R, R0, S)
	),
	cofact(Y, t(Q, L0, R0), S).


		/***********************************************
		*     old version count_A_algebras_with_zdd    *
		***********************************************/

% ?-zdd(( count_A_algebras_with_zdd(1, C))).
% ?-zdd(( count_A_algebras_with_zdd(2, C))).
% ?-time(zdd(( count_A_algebras_with_zdd(3, C)))).
% ?-time(zdd(( count_A_algebras_with_zdd(4, C)))).
% ?-call_with_time_limit(1800, zdd(( count_A_algebras_with_zdd(5, C)))).
%@ ERROR: Unhandled exception: Time limit exceeded

count_A_algebras_with_zdd(N, C):-
	prepare_count_A_algebras(N, Track, Etree),
	count_A_algebras([h(Etree, [], Track)], 0, C).
%
prepare_count_A_algebras(N, Track, Etree):-
	numlist(1, N, Dom),
	raise_list(Dom, 2, Dom2),
	distribute_cons(Dom, Dom2, Dom3),
	maplist(pred([U, a(U, [], [])]), Dom3, Track),
	set_pred(mk_pair, equational_mapsto),
	zdd:zdd_funs(Dom2, Dom, Etree).
%	zdd:map_space(Dom2, Dom, Etree, symmetric_mapsto).
%
count_A_algebras(X, Y, Z):- @(count_A_algebras(X, Y, Z)).
%
count_A_algebras([], C, C, _).
count_A_algebras([h(0, _, _)|Hs], C, C0, S):-!,
	count_A_algebras(Hs, C, C0, S).
count_A_algebras([h(_, _, 0)|Hs], C, C0, S):-!,
	count_A_algebras(Hs, C, C0, S).
count_A_algebras([h(1, _, [])|Hs], C, C0, S):-!, C1 is C+1,
	count_A_algebras(Hs, C1, C0, S).
count_A_algebras([h(1, _, _)|Hs], C, C0, S):-!,
	count_A_algebras(Hs, C, C0, S).
count_A_algebras([h(Etree, Es, Tr)|Hs], C, C0, S):-
	cofact(Etree, t(E, L, R), S),
	repeat_law([E|Es], Tr, Tr0),		% for early detection.
	(	Tr0 = 0
	->	count_A_algebras([h(L, Es, Tr)|Hs], C, C0, S)
	;	count_A_algebras([h(L, Es, Tr), h(R, [E|Es], Tr0)|Hs], C, C0, S)
	).		% no difference order of h(L)  and h(R).

% ?-zdd(( count_AC_algebras_with_zdd(1, C))).
% ?-zdd(( count_AC_algebras_with_zdd(2, C))).
% ?-zdd(( count_AC_algebras_with_zdd(3, C))).
% ?-time(zdd(( count_AC_algebras_with_zdd(4, C)))).
% ?-call_with_time_limit(100, time(zdd(( count_AC_algebras_with_zdd(5, C))))).
%@ % 347,069,373 inferences, 99.855 CPU in 100.001 seconds (100% CPU, 3475729 Lips)
%@ ERROR: Unhandled exception: Time limit exceeded

count_AC_algebras_with_zdd(N, C):-
	prepare_count_AC_algebras(N, Track, Etree),
	count_AC_algebras([h(Etree, [], Track)], 0, C).
%
prepare_count_AC_algebras(N, Track, Etree):-
	numlist(1, N, Dom),
	raise_list(Dom, 2, Dom2),
	distribute_cons(Dom, Dom2, Dom3),
	maplist(pred([U, a(U, [], [])]), Dom3, Track),
	set_pred(mk_pair, equational_mapsto),
	bin_symm_funs(Dom2, Dom, Etree).

equational_mapsto([A,B], C, A+B = C).

%	zdd:map_space(Dom2, Dom, Etree, symmetric_mapsto).
%
count_AC_algebras(X, Y, Z):- @(count_AC_algebras(X, Y, Z)).

%
count_AC_algebras([], C, C, _).
count_AC_algebras([h(0, _, _)|Hs], C, C0, S):-!,
	count_AC_algebras(Hs, C, C0, S).
count_AC_algebras([h(_, _, 0)|Hs], C, C0, S):-!,
	count_AC_algebras(Hs, C, C0, S).
count_AC_algebras([h(1, _, [])|Hs], C, C0, S):-!, C1 is C+1,
	count_AC_algebras(Hs, C1, C0, S).
count_AC_algebras([h(1, _, _)|Hs], C, C0, S):-!,
	count_AC_algebras(Hs, C, C0, S).
count_AC_algebras([h(Etree, Es, Tr)|Hs], C, C0, S):-
	cofact(Etree, t(E, L, R), S),
	symm_repeat_law([E|Es], Tr, Tr0),		% for early detection.
	(	Tr0 = 0
	->	count_AC_algebras([h(L, Es, Tr)|Hs], C, C0, S)
	;	count_AC_algebras([h(L, Es, Tr), h(R, [E|Es], Tr0)|Hs], C, C0, S)
	).		% no difference order of h(L)  and h(R).


% ?- I=3, J=2, K=1, L=1,
%	numlist(1, I, X), numlist(1, K, Y),
%	raise_list(X, J, Xj),
%	raise_list(Y, L, Yl),
%	time( ( call_with_time_limit(200, (
%		zdd((bin_symm_funs(Xj, Yl, F, test_equational_form),
%			psa(F), card(F, C))))))).

test_equational_form([X, Y], [Z], X+Y=Z).

%
bin_symm_funs(D, R, F):-
	bin_symm_funs(D, R, F, equational_mapsto).
%
:- meta_predicate bin_symm_funs(?,?,?,3).
bin_symm_funs(D, R, F, Pred):- @(bin_symm_funs(D, R, F, Pred)).
%
bin_symm_funs([], _, 1, _, _).
bin_symm_funs([[U, V]|As], Bs, F, Pred, S):- U @> V, !,
	bin_symm_funs(As, Bs, F, Pred, S).
bin_symm_funs([A|As], Bs, F, Pred, S):-
	bin_symm_funs(As, Bs, F0, Pred, S),
	bin_symm_extend_funs_by_one(A, Bs, F0, 0, F, Pred, S).
%
bin_symm_extend_funs_by_one(_, [], _, F, F, _, _):-!.
bin_symm_extend_funs_by_one(A, [B|Bs], H, F0, F, Pred, S):-
	call(Pred, A, B, Pair),
	zdd_insert(Pair, H, H0, S),
	zdd_join(H0, F0, F1, S),
	bin_symm_extend_funs_by_one(A, Bs, H, F1, F, Pred, S).

% ?- count_associative_algebra(1, X).
% ?- count_associative_algebra(2, X).
% ?- time(count_associative_algebra(3, X)).
% ?- time(count_associative_algebra(4, X)).
% ?- call_with_time_limit(180, time(count_associative_algebra(4, X))).
%@ % 984,263,663 inferences, 157.935 CPU in 198.803 seconds (79% CPU, 6232073 Lips)
%@ ERROR: Unhandled exception: Time limit exceeded

count_associative_algebra(N, NumOfAssocAlgebras):-
	numlist(1, N, Dom),
	raise_list(Dom, 2, Dom2),
	distribute_cons(Dom, Dom2, Dom3),
	maplist(pred([U, a(U, [], [])]), Dom3, Track0),
	sort(Track0, Track1),
	all_funs(Dom2, Dom, Maps, default_mapsto),
	maplist(maplist( pred([[A,B]-C, A+B=C]) ), Maps, FamEs),
	foldl( pred(Track1, (	[Es, C, C0]	:-
							repeat_law(Es, Track1, Track),
							(		Track = [] -> C0 is C+1
							;		C0 = C
						    ))
					   ),
		   FamEs,  0,  NumOfAssocAlgebras).

%
repeat_law(E, [a(_,[A],[A])|X], Y):-!, repeat_law(E, X, Y).
repeat_law(_, [a(_,[_],[_])|_], 0):-!.	% 0 means NON Associativitty.
repeat_law(Eqs, X, X0):- member(E, Eqs),
	apply_law_once(E, X, X1),
	!,
	repeat_law(Eqs, X1, X0).
repeat_law(_, X, X).

%
apply_law_once(E, X, [A0|X0]):-
	select(A, X, X0),
	law(E, A, A0).

% law/3
law( A+B=C,	a([X, A, B], U, []),
			a([X, A, B], U, [X, C])).
law( A+B=C,	a(T, U, [A, B]),
			a(T, U, [C])).
law( A+B=C,	a(T, [A, B], U),
			a(T, [C], U)).
law( A+B=C,	a([A, B, X], [], U),
			a([A, B, X], [C, X], U)).

% ?-  symm_count_associative_algebra(1, X).
% ?- symm_count_associative_algebra(2, X).
% ?- time(symm_count_associative_algebra(3, X)).
% ?- call_with_time_limit(180, time(symm_count_associative_algebra(4, X))).
%@ % 1,290,894,487 inferences, 161.322 CPU in 180.000 seconds (90% CPU, 8001973 Lips)
%@ ERROR: Unhandled exception: Time limit exceeded

symm_count_associative_algebra(N, NumOfAssocAlgebras):-
	numlist(1, N, Dom),
	raise_list(Dom, 2, Dom2),
	distribute_cons(Dom, Dom2, Dom3),
	maplist(pred([U, a(U, [], [])]), Dom3, Track0),
	sort(Track0, Track1),
	all_funs(Dom2, Dom, Maps, ordered_mapsto),
	sort(Maps, Maps0),
	maplist(maplist( pred([[A,B]-C, A+B=C]) ), Maps0, FamEs),
	foldl( pred(Track1, (	[Es, C, C0]	:-
							symm_repeat_law(Es, Track1, Track),
							(		Track = [] -> C0 is C+1
							;		C0 = C
						    ))
					   ),
		   FamEs,  0,  NumOfAssocAlgebras).

%
symm_repeat_law(E, [a(_,[A],[A])|X], Y):-!, symm_repeat_law(E, X, Y).
symm_repeat_law(_, [a(_,[_],[_])|_], 0):-!.	% 0 means NON Associativitty.
symm_repeat_law(Eqs, X, X0):- member(E, Eqs),
	symm_apply_law_once(E, X, X1),
	!,
	symm_repeat_law(Eqs, X1, X0).
symm_repeat_law(_, X, X).

%
symm_apply_law_once(E, X, [A0|X0]):-
	select(A, X, X0),
	symm_law(E, A, A0).

symm_law( A+B=C,	a([X, A, B], U, []), a([X, A, B], U, [X, C])).
symm_law( A+B=C,	a([X, B, A], U, []), a([X, B, A], U, [X, C])).
symm_law( A+B=C,	a(T, U, [A, B]), a(T, U, [C])).
symm_law( A+B=C,	a(T, U, [B, A]), a(T, U, [C])).
symm_law( A+B=C,	a(T, [A, B], U), a(T, [C], U)).
symm_law( A+B=C,	a(T, [B, A], U), a(T, [C], U)).
symm_law( A+B=C,	a([A, B, X], [], U), a([A, B, X], [C, X], U)).
symm_law( A+B=C,	a([B, A, X], [], U), a([B, A, X], [C, X], U)).

%
:- meta_predicate all_funs(?, ?, ?, 3).

all_funs(Xs, Y, Z):-
	all_funs(Xs, Y, Z, default_mapsto).

% ?- N is 2, M is 2, time(( numlist(1, N, X), numlist(1, M, Y),
%	raise_list(X, 2, X2),
%	all_funs(X2, Y, F, ordered_mapsto), length(F, C))).

all_funs([], _, [[]], _).
all_funs([X|Xs], Y, Z, M):- all_funs(Xs, Y, Z0, M),
	all_funs(Y, X, Z0, Z, [], M).
%
all_funs([], _, _, Z, Z, _).
all_funs([Y|Ys], X, Z, U, V, M):-
	all_funs_extend(Z, Y, X, U, U0, M),
	all_funs(Ys, X, Z, U0, V, M).
%
all_funs_extend([], _, _, U, U, _).
all_funs_extend([Z|Zs], Y, X, U, V, M):-
	(	call(M, X, Y, P) -> U =[[P|Z]|U0]
	;	U = [Z|U0]
	),
	all_funs_extend(Zs, Y, X, U0, V, M).
%
ordered_mapsto([A, B], Y, [A, B]-Y):- A@=<B, !.

%
default_mapsto(X, Y, X-Y).
