:- module(minato_r4, []).

% ( Revision of copy of minato-r2)

:- use_module(util(math)).
:- use_module(util(meta2)).
:- use_module(pac(basic)).
:- use_module(pac(meta)).
:- use_module(util(misc)).
:- use_module(pac('expand-pac')).
:- use_module(zdd('zdd-array')).
:- use_module(zdd(zdd)).
:- use_module(zdd('zdd-misc')).
:- use_module(zdd('zdd-plain')).
:- use_module(pac(op)).
:- op(1060, xfy, #).		% exor
:- op(1000, xfy, &).

% term_expansion --> pac:expand_pac.
term_expansion --> expand_pac.

		/******************
		*     Switches    *
		******************/
% ORTHOGONAl
% active_node(X, Y):- active_node_orth(X, Y).
% rect_nodes(X, Y):- rect_nodes_orth(X, Y).

% SKEW.
active_node(X, Y):- active_node_skew(X, Y).
rect_nodes(X, Y):- rect_nodes_skew(X, Y).

% ?- count_rect_path(rect(0,0), C, _).
% ?- count_rect_path(rect(1,0), C, _).
% ?- count_rect_path(rect(0,1), C, _).
% ?- count_rect_path(rect(1,1), C, _).
% ?- count_rect_path(rect(1,2), C, _).
% ?- count_rect_path(rect(2,1), C, _).
% ?- count_rect_path(rect(2,2), C, _).
% ?- count_rect_path(rect(1,10), C, _).
% ?- count_rect_path(rect(10,1), C, _).
% ?- count_rect_path(rect(100,1), C, _).
% ?- count_rect_path(rect(1,100), C, _).

% ?- time(count_rect_path(rect(6,6), C, _)).
%@ % 28,666,943 inferences, 2.073 CPU in 2.107 seconds (98% CPU, 13827602 Lips)
%@ C = 575780564.
%@ % 25,642,240 inferences, 3.558 CPU in 4.465 seconds (80% CPU, 7207373 Lips)
%@ C = 575780564.

% ?- time(count_rect_path(rect(7,7), C, _)).
%@ % 128,031,121 inferences, 10.589 CPU in 10.664 seconds (99% CPU, 12090872 Lips)
%@ C = 789360053252.
%@ % 106,041,498 inferences, 12.348 CPU in 12.454 seconds (99% CPU, 8587850 Lips)
%@ C = 789360053252.

% [2022/08/10]   cleanup old hash vector element with 0
% ?- time(count_rect_path(rect(8,8), C, _)).
%@ % 597,174,366 inferences, 66.803 CPU in 67.310 seconds (99% CPU, 8939312 Lips)
%@ C = 3266598486981642.
%@ % 597,174,366 inferences, 64.201 CPU in 64.572 seconds (99% CPU, 9301613 Lips)
%@ C = 3266598486981642.
%@ % 597,174,366 inferences, 59.440 CPU in 59.864 seconds (99% CPU, 10046731 Lips)
%@ C = 3266598486981642.
%@ % 608,401,418 inferences, 66.125 CPU in 66.439 seconds (100% CPU, 9200834 Lips)
%@ C = 3266598486981642.
%@ % 608,401,418 inferences, 60.675 CPU in 61.180 seconds (99% CPU, 10027244 Lips)
%@ C = 3266598486981642.
% [2022/08/10]
%@ % 597,174,366 inferences, 59.526 CPU in 59.907 seconds (99% CPU, 10032085 Lips)
%@ C = 3266598486981642.
% ?- time(count_rect_path(rect(8,8), C, _)).
%@ % 436,367,556 inferences, 52.415 CPU in 52.981 seconds (99% CPU, 8325235 Lips)
%@ C = 3266598486981642.

% ?- time(count_rect_path(rect(9,9), C, _)).
%@ % 2,028,259,194 inferences, 194.275 CPU in 195.632 seconds (99% CPU, 10440121 Lips)
%@ C = 2449629600675855540670.

% ?- call_with_time_limit(1800, time(count_rect_path(rect(10,10), C, S))).
%@ % 12,006,266,932 inferences, 1229.766 CPU in 1239.202 seconds (99% CPU, 9763047 Lips)
%@ C = 1568758030464750013214100,
%@ S = [state(..), out(3998810)|_].
%@ % 12,782,653,306 inferences, 1164.030 CPU in 1174.499 seconds (99% CPU, 10981375 Lips)
%@ C = 1568758030464750013214100,

% ?- call_with_time_limit(36000, count_rect_path(rect(11,11), C, _)).
%@ C = 182413291514248049241470885236.

% ?- convert_skew(p(1,1), Q), convert_skew(P, Q), convert_skew(P, R).
%@ Q = R, R = p(2, 1),
%@ P = p(1, 1).
orth_skew(p(I, J), Y):- var(Y), !,
	Y = p(K, I),
	K is I + J.
orth_skew(p(I, J), p(K, I)):- J is K - I.

% ?- arrow_symbol(_->_, F).
% ?- arrow_symbol(a->b, F, X, Y).
atmark(X):- functor(X, @, 2).
change_symbol(A-B, A->B).
arrow_symbol( _ -> _).
arrow_symbol(A, A0):- functor(A, A0, 2).
arrow_symbol(A, A0, A1, A2):- functor(A, A0, 2),
		arg(1, A, A1),
		arg(2, A, A2).
%
is_link(->(_, _)).
%
convert_normal_pair(P, Q, S):- orth_skew_pair(P, Q0),
	normal_pair(Q0, Q, S).
%
normal_arrow(X, Y):- @(normal_arrow(X, Y)).
%
normal_arrow(A->B, Arr, S):-!,
	(	zdd_compare(>, A, B, S) -> Arr = (B->A)
	;	Arr = (A->B)
	).
normal_arrow(A-B, Arr, S):-normal_arrow(A->B, Arr, S).
%

lt_pair(_-A, B -_, S):- zdd_compare(<, A, B, S).


		/*************************************
		*     basics of incrementing DG      *
		*************************************/

composable_pairs(A-B, A-C, B, C).
composable_pairs(A-B, C-A, B, C).
composable_pairs(B-A, A-C, B, C).
composable_pairs(B-A, C-A, B, C).
%
normal_pair(P, U):- @(normal_pair(P, U)).
%
normal_pair(A-B, B-A, S):- zdd_compare(>, A, B, S), !.
normal_pair(A->B, B->A, S):- zdd_compare(>, A, B, S), !.
normal_pair(P, P, _).

% ?- orth_skew_pair(p(1,1)-p(2,2), X), orth_skew_pair(Y, X).
%@ X = p(2, 1)-p(4, 2),
%@ Y = p(1, 1)-p(2, 2).
orth_skew_pair(P, Q):-
	(	P = (A-B) -> Q = (A0-B0)
	;	P = (A->B),
		Q = (A0->B0)
	),
	orth_skew(A, A0),
	orth_skew(B, B0).

link_sort_opp(X, Y):- sort(X, X0), reverse(X0, Y).

% ?- node_ends(p(0,0), En, rect(1,1)).
% ?- node_ends(p(0,0), En, rect(0,0)).
% ?- node_ends(p(1,1), En, rect(2,2)).

node_ends(P, En, rect(W,H)):- P = p(I,J),
	(	I < W ->
		K is I + 1,
		En0 = [p(K,J)]
	;	En0 = []
	),
	(	J < H ->
		L is J + 1,
		En = [p(I, L)|En0]
	;   En = En0
	).

% ?- node_ends_skew(p(0,0), En, rect(1,1)).
% ?- node_ends_skew(p(0,0), En, rect(0,0)).
% ?- node_ends_skew(p(1,1), En, rect(2,2)).
% ?- node_ends_skew(p(1,1), En, rect(1,2)).
node_ends_skew(P, En, rect(W,H)):- P = p(I,J),
	(	J < W ->
		K is J + 1,
		M is I + 1,
		En0 = [p(M, K)]
	;	En0 = []
	),
	(	I < H + J ->
		L is I + 1,
		En = [p(L, J)|En0]
	;   En = En0
	).

% ?- R=rect(2,2), repeat(in_rect(P, R),
%	(node_ends_shadows(P, En, Sh, R), writeln(P-En-Sh))).
in_rect(p(I, J), rect(W, H)):- between(0, W, I), between(0, H, J).

% ?- rect_nodes_orth(rect(1,1), PES), maplist(writeln, PES).
% ?- rect_nodes_orth(rect(3,3), PES), maplist(writeln, PES).
% ?- rect_nodes_orth(rect(10,10), PES), maplist(writeln, PES).
% ?- rect_nodes_orth(rect(30,30), PES), length(PES, L).
% ?- rect_nodes_orth(rect(100,100), PES), length(PES, L).
% ?- time((rect_nodes_orth(rect(1000,1000), PES), length(PES, L))).

rect_nodes_orth(R, PES):-
	findall(P-E, (in_rect(P, R), node_ends(P, E, R)), PES).

% ?- rect_nodes_skew(rect(100,100), PES), length(PES, L).
% ?- time((rect_nodes_skew(rect(1000,1000), PES), length(PES, L))).
rect_nodes_skew(R, PES_skew):- rect_nodes_orth(R, PES_orth),
						  convert_orth_skew(PES_orth, PES_skew).
%
convert_orth_skew([], []).
convert_orth_skew([A-B|C], [A0-B0|C0]):-
	orth_skew(A, A0),
	maplist(orth_skew, B, B0),
	convert_orth_skew(C, C0).


		/****************************************************
		*     replace_end/bypass:  Most basic operations.   *
		****************************************************/

% ?- zdd((X<<{[c-d]}, @(subst_node([a->c], c, a, X, Y)), psa(X), psa(Y))).
subst_node(_, _, _, X, 0, _):- X < 2, !.
subst_node(Es, A, P, X, Y, S):- get_key(max, Max, S), A==Max, !,
	subst_max_node(Es, Max, P, X, Y, S).
subst_node(Es, A, P, X, Y, S):- % memo is useless here
	/* replace A with P */
	cofact(X, t(U, L, R), S),
	arrow_symbol(U, Arrow, Lu, Ru),
	(	Arrow = (->) ->	Y = 0
	;	zdd_compare(<, A, Lu, S) ->	Y = 0
	;	subst_node(Es, A, P, L, L0, S),
		(	Ru = A	->
			normal_pair(Lu-P, V, S),
			zdd_ord_insert([V|Es], R, R0, S)
		;	Lu = A	->
			normal_pair(Ru-P, V, S),
			zdd_ord_insert([V|Es], R, R0, S)
		;	subst_node(Es, A, P, R, R1, S),
			zdd_insert(U, R1, R0, S)
		),
		zdd_join(L0, R0, Y, S)
	).

% ?- count_rect_path(rect(8,8), C, S).

% ?- zdd((X<<{[c-c]}, @(subst_max_node([], c, b, X, Y)), psa(Y))).
% ?- zdd((X<<{[a-b]}, @(subst_max_node([], c, b, X, Y)), psa(Y))).
% ?- zdd((X<<{[a-b, c-c]}, @(subst_max_node([], c, d, X, Y)), psa(Y))).

subst_max_node(_, _, _, X, 0, _):- X < 2, !.
subst_max_node(Es, Max, P, X, Y, S):-  	/* replace Max with P */
	cofact(X, t(U, L, R), S),
	arrow_symbol(U, Arrow, Lu, Ru),
	(	Arrow = (->) ->	Y = 0
	;	subst_max_node(Es, Max, P, L, L0, S),
		(	Max @=< Lu ->
			zdd_ord_insert([P-Max|Es], R, R0, S)
		;	Max @=< Ru -> R0 = 0
		;	subst_max_node(Es, Max, P, R, R1, S),
			zdd_insert(U, R1, R0, S)
		),
		zdd_join(L0, R0, Y, S)
	).

% ?- zdd((X<<{[a-b, c-d]}, @(bypass_via_node(b-c, X, Y)), psa(X), psa(Y))).
bypass_via_node(E, X, Y, S):- bypass_via_node([], E, X, Y, S).

% ?- zdd((X<<{[a-b, c-d]}, @(bypass_via_node([b->c], b-c, X, Y)), psa(X), psa(Y))).
bypass_via_node(_, _, X, 0, _):- X < 2,!.
bypass_via_node(Bypass, E, X, Y, S):- % memo is useless here
	cofact(X, t(P, L, R),  S),
	(	P = (_->_)  ->  Y = 0
	;	lt_pair(E, P, S) -> Y = 0
	;  	bypass_via_node(Bypass, E, L, L0, S),
		(	E = P ->	R0 = 0
		;  	composable_pairs(E, P, U, V) ->
			subst_node(Bypass, U, V, R, R0, S)
		;	bypass_via_node(Bypass, E, R, R1, S),
			zdd_insert(P, R1, R0, S)
		),
		zdd_join(L0, R0, Y, S)
	).

		/*************************
		*     count_rect_path    *
		*************************/

% ?- test(rect(1,1), Y).
test(X, Y):- count_rect_path(X, Y, _, _).

% ?- N = 3,  count_rect_path(rect(0, N), C).
% ?- N = 3, count_rect_path(rect(1, N), C).
% ?- N = 10, count_rect_path(rect(1, N), C).
% ?- count_rect_path(rect(3,1), C, S).
% ?- count_rect_path(rect(2,2), C, S).
% ?- count_rect_path(rect(3,3), C, S).
% ?- count_rect_path(rect(4,4), C, S).

% ?- call_with_time_limit(200, time(count_rect_path(rect(8,8), C, _))).
%@ % 908,479,618 inferences, 89.804 CPU in 91.150 seconds (99% CPU, 10116254 Lips)
%@ C = 3266598486981642.

% ?- count_rect_path(rect(1,1), C).
% ?- count_rect_path(rect(1,1), C, X).
% ?- count_rect_path(rect(1,1), C, X, S), psa(X, S).
% ?- count_rect_path(rect(2,2), C, X, S), psa(X, S).
%@ 	[p(0,0)-p(2,2),(p(0,0)->p(1,0)),(p(1,0)->p(2,0)),(p(2,0)->p(2,1)),(p(2,1)->p(2,2))]
%@ 	[p(0,0)-p(2,2),(p(0,0)->p(1,0)),(p(1,0)->p(2,0)),(p(1,1)->p(1,2)),(p(1,1)->p(2,1)),(p(1,2)->p(2,2)),(p(2,0)->p(2,1))]
%@ 	[p(0,0)-p(2,2),(p(0,0)->p(1,0)),(p(1,0)->p(1,1)),(p(1,1)->p(2,1)),(p(2,1)->p(2,2))]
%@ 	[p(0,0)-p(2,2),(p(0,0)->p(1,0)),(p(1,0)->p(1,1)),(p(1,1)->p(1,2)),(p(1,2)->p(2,2))]
%@ 	[p(0,0)-p(2,2),(p(0,0)->p(1,0)),(p(0,1)->p(0,2)),(p(0,1)->p(1,1)),(p(0,2)->p(1,2)),(p(1,0)->p(2,0)),(p(1,1)->p(2,1)),(p(1,2)->p(2,2)),(p(2,0)->p(2,1))]
%@ 	[p(0,0)-p(2,2),(p(0,0)->p(1,0)),(p(0,1)->p(0,2)),(p(0,1)->p(1,1)),(p(0,2)->p(1,2)),(p(1,0)->p(1,1)),(p(1,2)->p(2,2))]
%@ 	[p(0,0)-p(2,2),(p(0,0)->p(0,1)),(p(0,1)->p(1,1)),(p(1,1)->p(2,1)),(p(2,1)->p(2,2))]
%@ 	[p(0,0)-p(2,2),(p(0,0)->p(0,1)),(p(0,1)->p(1,1)),(p(1,1)->p(1,2)),(p(1,2)->p(2,2))]
%@ 	[p(0,0)-p(2,2),(p(0,0)->p(0,1)),(p(0,1)->p(1,1)),(p(1,0)->p(1,1)),(p(1,0)->p(2,0)),(p(2,0)->p(2,1)),(p(2,1)->p(2,2))]
%@ 	[p(0,0)-p(2,2),(p(0,0)->p(0,1)),(p(0,1)->p(0,2)),(p(0,2)->p(1,2)),(p(1,2)->p(2,2))]
%@ 	[p(0,0)-p(2,2),(p(0,0)->p(0,1)),(p(0,1)->p(0,2)),(p(0,2)->p(1,2)),(p(1,1)->p(1,2)),(p(1,1)->p(2,1)),(p(2,1)->p(2,2))]
%@ 	[p(0,0)-p(2,2),(p(0,0)->p(0,1)),(p(0,1)->p(0,2)),(p(0,2)->p(1,2)),(p(1,0)->p(1,1)),(p(1,0)->p(2,0)),(p(1,1)->p(1,2)),(p(2,0)->p(2,1)),(p(2,1)->p(2,2))]

count_rect_path(Rect, C):- count_rect_path(Rect, C, _).
%
count_rect_path(Rect, C, Opts):-
	memberchk(state(S), Opts),
	memberchk(out(X), Opts),
	open_state(S),
	set_key(rect, Rect, S),
	rect_nodes(Rect, A),
	sort(A, A1),
	reverse(A1, B),
	B =[Max-_|_],
	set_key(max, Max, S),
	arg(1, Max, K),
	set_key(prune_index, K, S),
	node_one_by_one(B, 1, X, S),
	card(X, C, S).

% ?- count_rect_path(rect(0,0), C, S).
% ?- count_rect_path(rect(0,1), C, S).
% ?- count_rect_path(rect(1,0), C, S).
% ?- count_rect_path(rect(1,1), C, S).
% ?- count_rect_path(rect(1,2), C, S).
% ?- count_rect_path(rect(100,1), C, S).
% ?- count_rect_path(rect(2,1), C, S).
% ?- count_rect_path(rect(2,2), C, S).
% ?- count_rect_path(rect(2,3), C, S).
% ?- count_rect_path(rect(3,2), C, S).
% ?- count_rect_path(rect(3,3), C, S).
% ?- count_rect_path(rect(4,4), C, S).
% ?- count_rect_path(rect(5,5), C, S).
% ?- count_rect_path(rect(6,6), C, S).
% ?- count_rect_path(rect(7,7), C, S).
% ?- count_rect_path(rect(8,8), C, S).
% ?- time(count_rect_path(rect(9,9), C, S)).

%@ % 2,028,259,251 inferences, 179.414 CPU in 180.716 seconds (99% CPU, 11304895 Lips)
%@ C = 2449629600675855540670,
%@ S = [state(..), out(271190)|_].

%@ % 2,763,116,175 inferences, 246.356 CPU in 248.711 seconds (99% CPU, 11215959 Lips)
%@ C = 41044208702632496804,
%@ S = [state(..), out(980774)|_].

% ?- zdd(@(node_one_by_one([], 1, X))).
node_one_by_one([], X, X, _).
node_one_by_one([P-En|R], X, Y, S):-
	writeln(P),
	add_node_one(P, En, X, Z, S),
	get_key(max, Max, S),
	prune_inactive_node(P, Z, U, Max, S),
	get_key(prune_index, I, S),
	arg(1, P, J),
	(	J = I ->  U0 = U
	;	set_key(prune_index, J, S),
		writeln("Slimming..."),
		zdd_slim(U, U0, S),
		garbage_collect
	),
	node_one_by_one(R, U0, Y, S).


% ?- zdd((X<<{[b-c, b->c]}, psa(X),
%  @(add_node_one(a, [b], X, Y)), psa(Y))).
% ?- time(count_rect_path(rect(8,8), C, S)).
%@ % 608,835,325 inferences, 54.280 CPU in 54.650 seconds (99% CPU, 11216634 Lips)
%@ C = 3266598486981642,
%@ S = [state(..), out(238658)|_].

%@ % 436,367,556 inferences, 35.323 CPU in 36.801 seconds (96% CPU, 12353625 Lip
%@ C = 3266598486981642,
%@ S = [state(..), out(4921865)|_].

% ?- time(count_rect_path(rect(1,1), C, S)).
add_node_one(P, [], X, Y, S):-!, zdd_insert(P-P, X, Y, S).
add_node_one(P, [A], X, Y, S):-!,
	subst_node([P->A], A, P, X, Y0, S),
	zdd_insert(P-P, X, Y1, S),
	zdd_join(Y0, Y1, Y, S).
add_node_one(P, [A, B], X, Y, S):-
	subst_node([P->A], A, P, X, Y0, S),
	subst_node([P->B], B, P, X, Y1, S),
	bypass_via_node([P->A, P->B], B-A, X, Y2, S),
	zdd_insert(P-P, X, Y3, S),
	zdd_join_list([Y0, Y1, Y2, Y3], Y, S).

% ?- zdd((X<<{[p(1)-p(2)]}, @(prune_inactive_node(p(1), X, Y, p(2))))).
% ?- zdd((X<<{[p(1)-p(2)]}, @(prune_inactive_node(p(0), X, Y, p(3))))).
% ?- zdd((X<<{[p(1)-p(2), p(3)-p(4)]}, @(prune_inactive_node(p(2), X, Y, p(5))))).
% ?- zdd((X<<{[p(1)-p(5), p(3)-p(4)]}, @(prune_inactive_node(p(2), X, Y, p(5))))).
% ?- zdd((X<<{[p(1)-p(5)], [p(3)-p(4)]}, @(prune_inactive_node(p(2), X, Y, p(5))), psa(Y))).
% ?- zdd((X<<{[p(1)-p(5), p(2)-p(2)]}, @(prune_inactive_node(p(2), X, Y, p(5))), psa(Y))).
% ?- zdd((X<<{[p(1)-p(4), p(5)-p(5)]}, @(prune_inactive_node(p(2), X, Y, p(5))), psa(Y))).
% ?- zdd((X<<{[p(1)-p(4)], [p(5)-p(5)]}, @(prune_inactive_node(p(2), X, Y, p(5))), psa(Y))).
% ?- zdd((X<<{[p(1)-p(4)], [p(5)-p(5)]}, @(prune_inactive_node(p(5), X, Y, p(5))), psa(Y))).

% ?- count_rect_path(rect(10, 1), C, S).
% ?- count_rect_path(rect(1,1), C, S).
% ?- R=rect(1,1), findall(on(P,Q), (in_rect(P, R), in_rect(Q, R), active_node(P, Q)), S),
%  maplist(writeln, S).

% ?- R=rect(10,10), findall(on(P,Q), (in_rect(P, R), in_rect(Q, R),
%	active_node(P, Q)), S),
%   maplist(writeln, S), fail; true.

% ?- active_node_orth(p(0,0),p(0,0)).  % true (conentionally)
% ?- active_node_orth(p(0,0),p(0,1)).
% ?- active_node_orth(p(0,0),p(1,0)).
% ?- active_node_orth(p(0,0),p(1,1)).
% ?- active_node_orth(p(0,1),p(1,0)).  % true.
% ?- active_node_orth(p(0,1),p(1,1)).
% ?- active_node_orth(p(1,0),p(1,1)).  % true
%% active_node_orth(P, Q) is true if Q is directly linked from some P' < P.

active_node_orth(p(0, J), p(0, L)):-!, L = J.
active_node_orth(p(I, J), p(I, L)):-!, J =< L.
active_node_orth(p(I, J), p(K, L)):- K is I+1, !, L < J.

% ?- rect_nodes(rect(10,10), Ns), findall(Q, ( member(P-_, Ns),
%	orth_skew(P, Q)), Sk), sort(Sk, SSk),
%	findall(active(U, V), (member(U, SSk), member(V, SSk), active_node_skew(U, V)), W), maplist(writeln, W), fail; true.

% ?- active_node_skew(p(0,0), p(0, 0)).  % true
% ?- active_node_skew(p(0,0), p(0, 1)).  % false (vacuously)
% ?- active_node_skew(p(1,1), p(2, 1)).  % true
% ?- active_node_skew(p(1,0), p(1, 1)).  % true
% ?- active_node_skew(p(1,1), p(2, 2)).  % false
% ?- active_node_skew(p(1,1), p(2, 1)).  % true.
% ?- active_node_skew(p(1,0), p(2, 1)).  % false
% ?- active_node_skew(p(1,1), p(2, 0)).  % true
active_node_skew(p(I, J), p(I, L)):- !, J =< L. % Actually not ncessary clause ?
active_node_skew(p(I, J), p(K, L)):- K is I + 1,
			(	 L = J ->  J > 0
			;	 L < J
			).
%
prune_inactive_node(_, X, X, _, _):- X<2, !.
prune_inactive_node(P, X, Y, Max, S):- % memo is useless here
	cofact(X, t(A, L, R), S),
	(	A = (_->_) -> Y = X
	; 	prune_inactive_node(P, L, L0, Max, S),
		A = (U - V),
		(	active_node(P, U) ->
			(	( active_node(P, V); V = Max )	 ->
				prune_inactive_node(P, R, R1, Max, S),
				zdd_insert(A, R1, R0, S)
			;	R0 = 0
			)
		;  U = Max -> R0 = 0
		;  U = V -> prune_inactive_node(P, R, R0, Max, S)
		;  R0 = 0
		),
		zdd_join(L0, R0, Y, S)
	).
