:- module(path_count4, [
	]).

:- use_module(library(apply)).
:- use_module(library(apply_macros)).
:- use_module(library(clpfd)).
:- use_module(library(statistics)).
:- use_module(zdd('zdd-array')).
:- use_module(util(math)).
:- use_module(util(meta2)).
:- use_module(pac(basic)).
:- use_module(pac(meta)).
:- use_module(util(misc)).
:- use_module(pac('expand-pac')). % For the kind block.
:- use_module(zdd('zdd-misc')).
:- use_module(pac(op)).
:- use_module(zdd(zdd)).
:- use_module(zdd('update-link')).

:- set_prolog_flag(stack_limit, 10_200_147_483_648).

 :- op(1060, xfy, ~).		% equivalence
 :- op(1060, xfy, #).		% exor
 :- op(1060, xfy, <->).		% equivalence
 :- op(1050, yfx, <-).
 :- op(1060, xfy, <=> ).	% equivalence
 :- op(1040, xfy, \/).		% OR
 :- op(1030, xfy, /\).		% AND
 :- op(1020, fy, \).		% NOT
 :- op(700, xfx, :=).		% Assignment
 :- op(1000, xfy, &).

% for pac query.
 :- pac:op(1000, xfy, &).
 :- pac:op(700, xfx, :=).

term_expansion --> pac:expand_pac.


% ?- zdd((dfs_dg_path_count(dg([a,b], [a-b]), C, a-b))).
% ?- zdd((dfs_dg_path_count(dg([a,b], [a-b, b-a]), C, a-b))).
% ?- zdd((dfs_dg_path_count(dg([a,b,c], [a-b]), C, a-b))).
% ?- zdd((dfs_dg_path_count(dg([a,b,c], [a-b, b-c]), C, a-c))).
% ?- zdd((dfs_dg_path_count(dg([a,b,c,d], [a-b, b-c, b-d, a-c, c-b]), C, a-d))).
% ?- R = rect(1,1),  zdd((rect_dg(R, N, E), dfs_dg_path_count(dg(N, E), C, p(0,0)-p(1,1)))).

% ?-zdd((dfs_rect_path_count(rect(1,2), C))).
%@ C = 4 .
% ?-time(zdd((dfs_rect_path_count(rect(15,1), C)))).
%@ % 106,749,564 inferences, 11.306 CPU in 12.134 seconds (93% CPU, 9441435 Lips)
%@ C = 32768 .
% ?-time(zdd((dfs_rect_path_count(rect(2,2), C)))).
%@ % 17,867 inferences, 0.031 CPU in 0.057 seconds (54% CPU, 583412 Lips)
%@ C = 12 .

% ?-time(zdd((dfs_rect_path_count(rect(3,3), C)))).
%@ % 484,761 inferences, 0.196 CPU in 0.340 seconds (58% CPU, 2468610 Lips)
%@ C = 184 .
% ?-time(zdd((dfs_rect_path_count(rect(4,4), C)))).
%@ % 33,077,014 inferences, 3.574 CPU in 4.034 seconds (89% CPU, 9254250 Lips)
%@ C = 8512 .
% ?-time(zdd((dfs_rect_path_count(rect(5,5), C)))).
%@ % 7,143,092,797 inferences, 898.272 CPU in 962.381 seconds (93% CPU, 7952034 Lips)
%@ C = 1262816 .

dfs_rect_path_count(R, C):- @(dfs_rect_path_count(R, C)).
%
dfs_rect_path_count(R, C, S):-	R = rect(W, H),
	rect_dg(R, N, E),
	dfs_dg_path_count(dg(N, E), C, p(0,0)-p(W, H), S).

%
dfs_dg_path_count(DG, C, ST):- @(dfs_dg_path_count(DG, C, ST)).
%

dfs_dg_path_count(DG, C, ST, S):- DG = dg(N, E),
	ST = A-_,
	build_suc(dg(N, E), S),
	findall(X-X, member(X, N), Id),
	linear_zdd(Id, Zid, S),
	dfs_build_qp(Zid, A, ST, S),
	count_child_path(Zid, C, S).
%
dfs_build_qp(X, A, ST, S):- memo(visited(X, A)-B, S),
	(	nonvar(B) -> true
	;	B = visited,
		memo(suc(A)-M, S),
		dfs_build_qp(M, X, A, ST, S)
	).
%
dfs_build_qp([], _, _, _, _).
dfs_build_qp([N|Ns], X, A, ST, S):-
	dfs_qp_add_child(A-N, X, ST, S),
	dfs_build_qp(Ns, X, A, ST, S).

% ?- zdd((X<<{[a-a, b-b]}, dfs_qp_add_child(a-b, X, a-b),
%	memo(child(X)-U),
%	dfs_qp_add_child(b-a, X, a-b),
%	memo(child(X)-V))).

dfs_qp_add_child(E, X, ST):- @(dfs_qp_add_child(E, X, ST)).
%

dfs_qp_add_child(E, X, ST, S):- update_qp(E, X, Y, S),
	(	Y = 0 ->  true
	;	(	final_qp(Y, ST, S) ->  Y0 = 1
		;	Y0 = Y
		),
		add_child(child(X), E-Y0, S),
		E = _- R,
		dfs_build_qp(Y0, R, ST, S)		% Depth-first search.
	).

% ?- zdd((X << {[1-2]}, final_qp(X, 1-2))).
% ?- zdd((X << {[1-2,3-4]}, final_qp(X, 1-2))).
% ?- zdd((X << {[1-1,2-2]}, final_qp(X, 1-2))).
% ?- zdd((X << {[]}, final_qp(X, 1-2))).

final_qp(X, ST):- @(final_qp(X, ST)).
%
final_qp(X, ST, S):- X>1,
	cofact0(X, A, Y, S),
	(	A = (I-I) -> final_qp(Y, ST, S)
	;	A = ST -> qp_id(Y, S)
	).
%
qp_id(1, _):-!.
qp_id(X, S):- X>1,
	cofact0(X, I-I, Y, S),
	qp_id(Y, S).

% ?- zdd((build_suc(dg([a,b], [a-b])), memo(suc(a)-U))).
% ?- zdd((build_suc(dg([a,b,c], [a-b, a-c, a-b, b-a])), memo(suc(a)-U),
%	memo(suc(b)-V), memo(suc(c)-W))).

build_suc(R):- @(build_suc(R)).
%
build_suc(dg(Ns, Rel), S):-
	maplist(pred(S, [X]:- memo(suc(X)-[], S)), Ns),
	rel_to_coa(suc, Rel, S).

% ?- zdd((rel_to_coa(suc, [a-b]), memo(suc(a)-U))).
% ?- zdd((rel_to_coa(suc, [a-b, a-c, a-b, b-a]), memo(suc(a)-U),
%	memo(suc(b)-V), memo(suc(c)-W))).

rel_to_coa(Key, R):- @(rel_to_coa(Key, R)).
%
rel_to_coa(_, [], _):-!.
rel_to_coa(Key, [A-B|R], S):-functor(U, Key, 1),
	arg(1, U, A),
	add_child(U, B, S),
	rel_to_coa(Key, R, S).
%
count_child_path(X, C):- @(count_child_path(X, C)).
%
count_child_path(X, X, _):- X<2, !.
count_child_path(X, C, S):- memo(count_child_path(X)-C, S),
	(	nonvar(C) -> true
	;	memo(child(X) - M, S),
		count_child_path(M, 0, C, S)
	).
%
count_child_path([], C, C, _).
count_child_path([_-X|R], C, D, S):- count_child_path(X, U, S),
	C0 is U + C,
	count_child_path(R, C0, D, S).

		/**************************************
		*     breadth first search version    *
		**************************************/


% ?-time(zdd((bfs_rect_path_count(rect(15,0), C)))).
% ?-time(zdd((bfs_rect_path_count(rect(0, 120), C)))).
% ?-time(zdd((bfs_rect_path_count(rect(1,1), C)))).
% ?-time(zdd((bfs_rect_path_count(rect(2,2), C)))).
%@ % 17,728 inferences, 0.002 CPU in 0.002 seconds (97% CPU, 8330827 Lips)
%@ C = 12 .
% ?-time(zdd((bfs_rect_path_count(rect(3,3), C)))).
%@ % 483,059 inferences, 0.044 CPU in 0.045 seconds (97% CPU, 11082639 Lips)
%@ C = 184 .
% ?-time(zdd((bfs_rect_path_count(rect(4,4), C)))).

%@ % 33,025,049 inferences, 3.659 CPU in 3.785 seconds (97% CPU, 9024986 Lips)
%@ C = 8512 .

% ?-profile(time(zdd((bfs_rect_path_count(rect(4,4), C))))).
%@ % 33,025,163 inferences, 4.752 CPU in 4.915 seconds (97% CPU, 6950417 Lips)
%@ C = 8512.




% ?-time(zdd((bfs_rect_path_count(rect(5,5), C)))).
%@ % 5,517,389,364 inferences, 816.623 CPU in 842.814 seconds (97% CPU, 6756348 Lips)
%@ C = 1262816 .

bfs_rect_path_count(R, C):- @(bfs_rect_path_count(R, C)).
%
bfs_rect_path_count(R, C, S):-	R = rect(W, H),
	rect_dg(R, N, E),
	bfs_dg_path_count(dg(N, E), C, p(0,0)-p(W, H), S).
%
bfs_dg_path_count(DG, C, ST):- @(bfs_dg_path_count(DG, C, ST)).
%
bfs_dg_path_count(DG, C, ST, S):- DG = dg(N, E),
	ST= A-_,
	build_suc(dg(N, E), S),
	findall(X-X, member(X, N), Id),
	linear_zdd(Id, Zid, S),
	bfs_build_qp([A-Zid], [], ST, S),
	count_child_path(Zid, C, S).
%
% bfs_build_qp(X, Y, _, _):- length(X, L), length(Y, M), writeln(L-M), fail.
bfs_build_qp([], [], _, _):-!.
bfs_build_qp([], X, ST, S):-!, bfs_build_qp(X, [], ST, S).
bfs_build_qp([A-X|R], R0, ST, S):- memo(visited(X, A)-B, S),
	(	nonvar(B)
	->	% write(.),  % many hits.
		bfs_build_qp(R, R0, ST, S)
	;	B = visited,
		memo(suc(A)-M, S),
		bfs_build_qp(M, X, A, U-R0, ST, S),
		bfs_build_qp(R, U, ST, S)
	).

%
bfs_build_qp([], _, _, U-U, _, _).
bfs_build_qp([N|Ns], X, A, U-V, ST, S):-
	bfs_qp_add_child(A-N, X, Y, ST, S),
	(	( Y = 0; Y = 1) -> U0 = U
	;	U = [N-Y|U0]
	),
	bfs_build_qp(Ns, X, A, U0-V, ST, S).

% ?- zdd((X<<{[a-a, b-b]}, qp_add_child(a-b, X, a-b),
%	memo(child(X)-U),
%	qp_add_child(b-a, X, a-b),
%	memo(child(X)-V))).

bfs_qp_add_child(E, X, Y, ST):- @(bfs_qp_add_child(E, X, Y, ST)).
%
bfs_qp_add_child(E, X, Y, ST, S):- update_qp(E, X, Y0, S),
	(	Y0 = 0 ->  Y = 0
	;	(	final_qp(Y0, ST, S) ->  Y = 1
		;	Y = Y0
		),
		add_child(child(X), E-Y, S)
	).


		/****************
		*     bridge    *
		****************/

% ?- zdd((count_path_line(p(1,1),p(2,1), C))).
% ?- zdd((count_path_line(p(1,2), p(3,2), C))).
% ?- zdd((count_path_line(p(1,2),p(10,2), C))).
% ?- zdd((count_path_line(p(1,2),p(100,2), C))).
% ?- time(zdd((count_path_line(p(1,2),p(1000, 2), C)))).
%@ % 4,691,349 inferences, 0.982 CPU in 1.154 seconds
% ?- time(zdd((count_path_line(p(1,2),p(10000,2), C)))).
%@ % 61,745,569 inferences, 59.803 CPU in 60.290 seconds

count_path_line(X, Y, C):- @(count_path_line(X, Y, C)).
%
count_path_line(X, Y, C, S):-
	qp_line(X, Y, R, S),
	memo_final_line(R, X, Y, S),
	qp_list(Q, [X-X], S),
	path_tree(Q, Tree, S),
	card(Tree, C, S).
%
memo_final_line([], _, _, _).
memo_final_line([A|R], X, Y, S):-
	qp_list(A, List, S),
	(	List = [X-Y]
	->  memo(qp_final(A)-true, S)
	;  true
	),
	memo_final_line(R, X, Y, S).

%
qp_line(X, Y, R):- @(qp_line(X, Y, R)).
%
qp_line(X, X, [Q], S):-!,
	zdd_insert_atoms([X-X], 1, Q, S),
	memo(qp_suc(Q)-List, S),
	( var(List)-> List=[]; true).
qp_line(X, Y, R, S):- X = p(I, K), Y = p(J, _),
	M is (I+J)//2,
	qp_line(X, p(M, K), R1, S),
	M0 is M+1,
	qp_line(p(M0,K), Y, R2, S),
	qp_joint([p(M,K)-p(M0,K), p(M0,K)-p(M,K)], R1, R2, R3, S),
	prune_qp_line_local(R3, X, Y, R, S).
%
prune_qp_line_local([], _, _, [], _).
prune_qp_line_local([Q|Qs], X, Y, Rs, S):-
	X = p(I,K),
	Y = p(J,K),
	qp_list(Q, List, S),
	(	member(p(A,_)-p(B,_), List),
		A \== B,
		(	I<A, A<J
		;	I<B, B<J
		)
	->  Rs0 = Rs
	;   Rs = [Q|Rs0]
	),
	prune_qp_line_local(Qs, X, Y, Rs0, S).

%% test basics.
% ?- zdd((
%	qp_list(X, [a-a]), qp_list(Y, [b-b]),
%	qp_joint([a-b, b-a], [X], [Y], Z),
%	maplist(pred(([U]:-qp_list(U, List), writeln(List))), Z))).
% ?- zdd((
%	qp_list(X, [a-a]), qp_list(Y, [b-b]),
%	qp_joint([a-b, b-a], [X], [Y], Z),
%	zdd_join(X, Y, A),
%	memo(qp_suc(A)-L), writeln(qp_suc(A)-L),
%	maplist(pred(([U]:-qp_list(U, List), writeln(List))), Z))).
% ?- zdd((
%	qp_list(X, [a-a]), qp_list(Y, [b-b]),
%	qp_joint([a-b, b-a], [X], [Y], Z),
%	zdd_join(X, Y, A),
%	memo(qp_suc(A)-L), writeln(qp_suc(A)-L),
%	maplist(pred(([U]:-qp_list(U, List), writeln(List))), Z))).

qp_joint(X, Y, Z, U):-@(qp_joint(X, Y, Z, U)).
%
qp_joint(X, Y, Z, U, S):- map_prod_sum(Y, Z, V, [], S),
	qp_bridge_links(X, Y, Z, U, V, S).

% ?- zdd((qp_list(X, [a-a]), qp_list(Y, [b-b]),
%	qp_bridge(a-b, X, Y, Z), qp_list(Z, U))).
% ?- zdd((qp_list(X, [a-a]), qp_list(Y, [b-b]),
%	qp_bridge(b-a, X, Y, Z), qp_list(Z, U))).

qp_bridge(E, X, Y, Z):- @(qp_bridge(E, X, Y, Z)).
%
qp_bridge(A-B, X, Y, Z, S):- % assuming qp  X < qp Y
	qp_list(X, L, S),
	qp_list(Y, M, S),
	(	select(U-A, L, L0),
		select(B-V, M, M0)
	;	select(B-V, L, L0),
		select(U-A, M, M0)
	),
	!,
	ord_union(L0, M0, N0),
	ord_union([U-V], N0, N),
	qp_list(Z, N, S).
qp_bridge(_, _, _, 0, _).

%
qp_bridge_links(X, Y, Z, U, V):- @(qp_bridge_links(X, Y, Z, U, V)).
%
qp_bridge_links([], _, _, Z, Z, _).
qp_bridge_links([E|Es], X, Y, Z, U, S):-
	qp_bridge_link(X, E, Y, Z, Z0, S),
	qp_bridge_links(Es, X, Y, Z0, U, S).
%
qp_bridge_link([], _, _, Z, Z, _).
qp_bridge_link([X|Xs], E, Y, Z, U, S):-
	qp_bridge_basic(Y, X, E, Z, Z0, S),
	qp_bridge_link(Xs, E, Y, Z0, U, S).
%
qp_bridge_basic([], _, _, Z, Z, _).
qp_bridge_basic([Y|Ys], X, E, Z, U, S):- qp_bridge(E, X, Y, V, S),
	(	V = 0
	->	Z0 = Z
	;	Z  = [V|Z0],
		select_goto(E, X, Y, V, S)
	),
	qp_bridge_basic(Ys, X, E, Z0, U, S).
%
select_goto(E, X, Y, Z):- @(select_goto(E, X, Y, Z)).
%
select_goto(A-B, X, Y, Z, S):-
	G = (A-B)-Z,
	( A@<B -> W = X ; W = Y ),
	getmemo(qp_suc(W)-U, V, S),
	(	var(V) -> V = [] ; true ),
	add_new(G, V, U).

		/***********************
		*     end of bridge    *
		***********************/

% ?-  belong(3, 3, 4).
% ?-  belong(3, 4, 1).
belong(X, I, J):- I @=< J, !, I @=< X, X @=< J.
belong(X, I, J):- J @=< X, X @=< I.
%
sub_interval(I, J, X, Y):- belong(I, X, Y), belong(J, X, Y).
%
disjoint_interval(I, J, X, Y):- \+ belong(I, X, Y), \+ belong(J, X, Y).

% ?- compatible_interval(1, 2, 3, 4).
% ?- compatible_interval(1, 3, 2, 4). % false
% ?- compatible_interval(2, 4, 5, 7).
% ?- compatible_interval(7, 5, 2, 4).

compatible_interval(I, J, X, Y):- disjoint_interval(I, J, X, Y), !.
compatible_interval(I, J, X, Y):- sub_interval(I, J, X, Y), !.
compatible_interval(I, J, X, Y):- sub_interval(X, Y, I, J).

% ?- rect_dg(rect(1,1), X, Y).
% ?- rect_dg(rect(0,1), X, Y).
% ?- rect_dg(rect(0,0), X, Y).

rect_dg(rect(W, H), Nodes, Links):-
	rect_nodes(W, H, Nodes),
	rect_links(W, H, Nodes, Links).

% ?- rect_nodes(4,5,Ns), length(Ns, L).
rect_nodes(W, H, Nodes):-
	findall(p(I, J),
			(	between(0, W, I),
				between(0, H, J)
			),
			Nodes0),
	sort(Nodes0, Nodes).

%
rect_links(W, H, Ns, Links):-
	findall(P-Q,
			(	member(P, Ns),
				P = p(I, J),
				(	Q = p(I, J1),
					(	J1 is J-1,
						J1 >= 0
					;	J1 is J+1,
						J1 =< H
					)
				;   Q = p(I1, J),
					(	I1 is I-1,
						I1 >= 0
					;	I1 is I+1,
						I1 =< W
					)
				)
			),
			Links0),
	sort(Links0, Links).

		/********************
		*   frontier based  *
		********************/

% ?- zdd(path_count(rect(0,0), C)).
% ?- zdd(path_count(rect(1,0), C)).
% ?- zdd(path_count(rect(0,1), C)).
% ?- zdd(path_count(rect(1,1), C)).
% ?- zdd(path_count(rect(2, 1), C)).
% ?- zdd(path_count(rect(3, 1), C)).
% ?- zdd(path_count(rect(1, 3), C)).
% ?- zdd(path_count(rect(4, 1), C)).
% ?- zdd(path_count(rect(2, 3), C)).
% ?- time(zdd(path_count(rect(4,3), C))).
% ?- time(zdd(path_count(rect(4,4), C))).
%@ C= 8512
% ?- call_with_time_limit(360, zdd(path_count(rect(5, 5), C))).

path_count(Rect, C):- @(path_count(Rect, C)).

%
path_count(rect(0,0), 1, _):-!.
path_count(Rect, C, S):-
	rect_dg(Rect, Ns, Links),
	findall(A-A, member(A, Ns), Ids),
	Rect = rect(W, H),
	set_key(final, p(W, H), S),
	qp_list(Q, Ids, S),
	memo(qp_suc(Q)-[], S),
	frontier(Links, [Q], FinalQs, S),
	fold_qp_memo_final(FinalQs, S),
	path_tree(Q, Tree, S),
	card(Tree, C, S).

%
frontier([], Qs, Qs, _).
frontier([E|Es], Qs, Qs0, S):-
	frontier_link(E, Qs, Qs1, Qs, S),
	prune_frontier(Qs1, E, Qs2, S),
	frontier(Es, Qs2, Qs0, S).

%
frontier_link(_, [], Qs, Qs, _):-!.
frontier_link(E, [X|Qs], Qs0, Qs1, S):- X<2, !,
 	frontier_link(E, Qs, Qs0, Qs1, S).
frontier_link(E, [Q|Qs], Next, Qs1, S):-
	qp_update(E, Q, Q0, S),
	(	Q0 < 2
	->	Next = Qs0
	;	qp_memo_final(Q0, S),
		Next = [Q0|Qs0]
	),
	qp_update_succs(E-Q0, Q, S),
	frontier_link(E, Qs, Qs0, Qs1, S).

% ?- zdd((qp_list(Q, [a]), qp_list(Q0, [b]), qp_list(Q1, [c]),
%	map_prod_sum([Q, Q0], [Q1], Z),
%	maplist(pred([A, M]:-qp_list(A, M)), Z, W))).
%
map_prod_sum(X, Y, Z):- map_prod_sum(X, Y, Z, []).
%
map_prod_sum(X, Y, Z, U):- @(map_prod_sum(X, Y, Z, U)).

map_prod_sum([], _, U, U, _).
map_prod_sum([Q|Y], Z, U, V, S):- qp_list(Q, L, S),
	map_qp_sum(Z, L, U, U0, S),
	map_prod_sum(Y, Z, U0, V, S).
%
map_qp_sum([], _, U, U, _).
map_qp_sum([Q|As], L, [Q0|U], V, S):-qp_list(Q, M, S),
	ord_union(L, M, N),
	qp_list(Q0, N, S),
	map_qp_sum(As, L, U, V, S).

%
prune_frontier(F, Link, F0):- @(prune_frontier(F, Link, F0)).
%
prune_frontier([], _, [], _).
prune_frontier([Q|F], Link, [Q|F0], S):- admissible_qp(Q, Link, S), !,
	prune_frontier(F, Link, F0, S).
prune_frontier([_|F], Link, F0, S):- prune_frontier(F, Link, F0, S).

%
admissible_qp(Q, Link, S):- qp_list(Q, List, S),
	admissible_qp_list(List, Link).

% ?- admissible_qp_list([p(0,0)-p(3, 1), p(3, 2)-p(3, 0)], p(0,0)-p(0,2)).
admissible_qp_list([], _):-!.
admissible_qp_list([X-X|List], Link):-!, admissible_qp_list(List, Link).
admissible_qp_list([_-X|_], P-_):- X @< P, !, fail.
admissible_qp_list([_|List], Link):- admissible_qp_list(List, Link).

% ?- zdd((path_count(rect(1,0), C))).
% ?- zdd((path_count(rect(0,1), C))).
% ?- zdd((path_count(rect(1,1), C))).
% ?- zdd((path_count(rect(3,1), C))).
% ?- time(zdd((path_count(rect(1,8), C)))).   % take time.
%@ % 86,018,490 inferences, 10.604 CPU in 11.323
% ?- time(zdd((path_count(rect(8,1), C)))).
%@ % 9,254,518 inferences, 3.089 CPU in 4.850
% ?- zdd((path_count(rect(2,3), C))).
% ?- zdd((path_count(rect(3,2), C))).
% ?- time(zdd((path_count(rect(3,3), C)))).
%@ % 50,584,360 inferences, 7.362 CPU in 9.480
%@ C = 184 .
% ?- time(zdd((path_count(rect(4,4), C)))).
%@ % 9,474,429,287 inferences, 1256.779 CPU in 1433.734 seconds (88% CPU, 7538660 Lips)
%@ C = 8512 .

qp_update(E, X, Y):- @(qp_update(E, X, Y)).
%
qp_update(_, X, 0, _):- X < 2.
qp_update(E, X, Y, S):- qp_list(X, U, S),
	memo(qp_update(E, X)-Y, S),
	(	nonvar(Y)  ->	true
	;	select_touch(E, [A, B], U, U0)
	->	connect_pairs(E, [A, B], W),
		ord_union([W], U0, V),
		qp_list(Y, V, S)
	;	Y = 0
	).
qp_update(_, _, 0,  _).

% ?- zdd((qp_list(X, [a-a, b-b]), memo(qp_suc(X)-[]))).
qp_update_succs(G, Q):- @(qp_update_succs(G, Q)).

% ?- zdd(qp_update_succs(a, 1)).
qp_update_succs(G, Q, S):- getmemo(qp_suc(Q)-U, V, S),
	G = E-Q0,
	(	var(V) -> U = [E-Q0]
	;	add_new(E-Q0, V, U)
	).

qp_admissible(L):-
	forall((select(A-B, L, L0), select(C-D, L0, _)),
		   compatible_interval(A, B, C, D)).

% ?- compatible_interval(1, 2, 3, 4).
% ?- compatible_interval(1, 3, 2, 4). % false
% ?- compatible_interval(2, 4, 5, 7).
% ?- compatible_interval(7, 5, 2, 4).

% compatible_interval(I, J, X, Y):- disjoint_interval(I, J, X, Y), !.
% compatible_interval(I, J, X, Y):- sub_interval(I, J, X, Y), !.
% compatible_interval(I, J, X, Y):- sub_interval(X, Y, I, J).
%@ true.
%@ path-count4.pl qcompiled.

% ?- select_touch(a-b, [A, B], [a-a, b-b], X).
% ?- select_touch(a-b, [A, B], [a-a, b-b], X),
%	connect_pairs(a-b, [A, B], Y).
%@ false.

% ?- select_touch(a-b, [A, B], [a-a, b-c], X),
%	connect_pairs(a-b, [A, B], Y).
% ?- select_touch(a-b, [A, B], [a-a, c-b], X).

select_touch(_, [], V, V).
select_touch(E, [A|As], [A|U], V):- touch(E, A), !,
	select_touch(E, As, U, V).
select_touch(E, As, [A|U], [A|V]):-select_touch(E, As, U, V).

% ?- connect_pairs(a-b, [X-Y, U-V], W).
connect_pairs(X-Y, [U-X, Y-V], U-V).
connect_pairs(X-Y, [Y-V, U-X], U-V).

% connect_pair(X-Y, U-X, U-Y).
% connect_pair(X-Y, Y-V, X-V).

touch(A-_, _-A).
touch(_-A, A-_).

% ?- add_new(a,[b,c,a], X).
add_new(X, Y, Y):- memberchk(X, Y), !.
add_new(X, Y, [X|Y]).

% ?- zdd((add_child(a, 1), memo(a-X))).
% ?- zdd((add_child(a, 1), add_child(a, 2), memo(a-X))).
% ?- zdd((add_child(a, 1), add_child(a, 2), add_child(a, 1), memo(a-X))).
% ?- numlist(1, 100000, Ns),
%	time(zdd(( maplist(pred([Child]:- add_child(a, Child)), Ns), memo(a-X)))).
%@ % 2,600,319 inferences, 81.730 CPU in 81.817

% ?- zdd((qp_list(Q, [p(0,0)-p(1,0)]), qp_list(Q, L))).
%@ L = [p(0, 0)-p(1, 0)].
%
qp_list(Q, List):- @(qp_list(Q, List)).
%
qp_list(0, 0, _):-!.
qp_list(1, [], _):-!.   %% ???
qp_list(Q, List, S):- var(Q), !,
	zdd_insert_atoms(List, 1, Q, S),
	memo(qp_suc(Q)-U, S),
	(	var(U) -> U = []
	;	true			% Already registered.
	).
qp_list(Q, [A|As], S):- cofact(Q, t(A, 0, R), S),
	qp_list(R, As, S).

%
qp_memo_final(X):- @(qp_memo_final(X)).
%
qp_memo_final(Q, S):-  memoq(qp_final(Q)-true, S), !.
qp_memo_final(Q, S):- qp_list(Q, List, S),
	get_key(final, F, S),
	(	qp_final(List, F)
	->	memo(qp_final(Q)-true, S)
	;	true
	).
%
fold_qp_memo_final([], _).
fold_qp_memo_final([Q|Qs], S):- qp_memo_final(Q, S),
	fold_qp_memo_final(Qs, S).

%
path_tree(Q, T):- @(path_tree(Q, T)).
%

path_tree(X, 0,  _):- X<1, !.
path_tree(Q, 1, S):- memoq(qp_final(Q)-true, S), !.
path_tree(Q, T, S):- memo(path_tree(Q)-T, S),
	(	nonvar(T) -> true
	;	memo(qp_suc(Q) - L, S),		% writeln(qp_suc(Q)-L),
		path_tree_list(L, 0, T, S)
	).
%
path_tree_list([], C, C, _).
path_tree_list([E-Q|Qs], C0, C, S):-
	path_tree(Q, A, S),
	zdd_insert(E, A, A0, S),
	zdd_join(C0, A0, C1, S),
	path_tree_list(Qs, C1, C, S).

% ?- qp_final([p(0,0)-p(2,3),p(0,3)-p(0,3),p(1,3)-p(1,3)], p(2,3)).

qp_final([p(0,0)-F|R], F):-!, qp_all_id(R).
qp_final([A-A|R], F):- qp_final(R, F).

%
qp_all_id([]).
qp_all_id([A-A|R]):- qp_all_id(R).

%
dag_card(Q, C):- @(dag_card(Q, C)).
%
dag_card(Q, C, S):- open_state(M),
	dag_card(Q, C, S, M),
	close_state(M).
%
dag_card(Q, 0,  _, _):- Q<2, !.
dag_card(Q, 1, S, _):- memoq(qp_final(Q)-true, S), !.
dag_card(Q, C, S, M):- memo(dag_card(Q)-C, M),
	(	nonvar(C) -> true
	;	memo(qp_suc(Q) - L, S),
		dag_card_list(L, 0, C, S, M)
	).
%
dag_card_list([], C, C, _, _).
dag_card_list([_-Q|Qs], C0, C, S, M):-
	dag_card(Q, A, S, M),
	C1 is C0 + A,
	dag_card_list(Qs, C1, C, S, M).


		/***************
		*     debug    *
		***************/
%
snap_qp_update(E, X, Y, S):- qp_update(E, X, Y, S),
	qp_list(X, X0, S),
	qp_list(Y, Y0, S),
	format("~w\n~w\n~w\n\n", [qp_update(E, X, Y), X0, Y0]).
%
dump_frontier(Qs, S):- maplist(pred(S,
	([Q]:- qp_list(Q, U, S), writeln(U))), Qs).
%
dump_qp(X, _):-  X<2, !, writeln(X).
dump_qp(X, S):-  qp_list(X, L, S),
	writeln(X=L).

%
delta_star([], Q, Q, _).
delta_star([E|Es], Q,  Q0, S):-  memo(qp_suc(Q)-L, S),
	writeln(E),
	memberchk(E-Q1, L),
	dump_qp(Q1, S),
	delta_star(Es, Q1, Q0, S).

%
mirror(X, Y):- @(mirror(X, Y)).
%
mirror(X, X, _):- X<2, !.
mirror(X, Y, S):- cofact(X, t(p(I,J)-p(K,L), Lx, Rx), S),
	mirror(Lx, MLx, S),
	mirror(Rx, MRx, S),
	zdd_insert(p(J, I)-p(L, K), MRx, MRx0, S),
	zdd_join(MLx, MRx0, Y, S).

%
rect_path_tree(Rect, T):-@(rect_path_tree(Rect, T)).

rect_path_tree(Rect, T, S):-
	sort_links_by_rank(Rect, Rs, Ids),
	Rect = rect(W, H),
	set_key(final, p(W, H), S),
	qp_list(Q, Ids, S),
	repeat_frontier(Rs, [Q], FinalQs, S),
	fold_qp_memo_final(FinalQs, S),
	path_tree(Q, T, S).

% ?- spy(remove_end_links).
% ?- remove_end_links([1-[c-a]], R0, a-d).
% ?- remove_end_links([1-[c-a, d-c, a-d]], R0, a-d).

remove_end_links(R, R0, ST):- maplist(pred(ST,
	(	[J-Es, J-Es0]:- foldl(
		pred([ST, Es, Es0],
			 ( [P-Q, U, V]:- ST = S-T,
					(	( Q==S; P==T)
					->  V = U
					;	U = [P-Q|V]
					))),  Es,  Es0, []))),  R, R0).

%
rest(Q, X, Y, S):- memo(qp_suc(Q)-L, S),
	(	select(Q, X, X0)-> true
	;	X0 = X
	),
	rest_list(L, X0, Y, S).
%
rest_list([], X, X, _).
rest_list([_-Q|Qs], X, Y, S):-
	rest(Q, X, X0, S),
	rest_list(Qs, X0, Y, S).


% ?- zdd((dbg_qp_update(p(0,0)-p(1,0), [p(0,0)-p(0,0),p(1,0)-p(1,0)], L))).
% ?- spy(dbg_qp_update).
% ?- zdd((dbg_qp_update(p(2,2)-p(2,3), [p(0,0)-p(2,2),p(0,3)-p(0,3), p(1,3)-p(1,3), p(2,3)-p(2,3)], L))).

dbg_qp_update(E, X, Y):- @(dbg_qp_update(E, X, Y)).
%
dbg_qp_update(E, X, Y, S):- qp_list(Q, X, S),
	qp_update(E, Q, Q0, S),
	qp_list(Q0, Y, S).


% ?- zdd((rect_path_tree(rect(3,2), T), card(T, C),
%   zdd_insert_atoms([p(0,0)-p(2,3),p(0,3)-p(0,3),p(1,3)-p(1,3)], 1, W),
%	mirror(T, MT), card(MT, CMT),
%	rect_path_tree(rect(2,3), T0), card(T0, C0), mirror(T0, MT0),
%	card(MT0, CM0),
%	zdd_subtr(T, MT0, D), card(D, CD), psa(D)
%)).


% Test to show getmemo/2,3 works well.
% ?- time((numlist(1, 10000, Ns),
%	zdd((maplist( pred([A]:- getmemo(a(A)-A, _)), Ns),
%		 maplist( pred(([A]:- getmemo(a(A)-B, Old), B is 2*Old)),  Ns),
%		 foldl( pred(([I, B, C]:- getmemo(a(I)-J, J),
%							 C is B + J)), Ns, 0, Sum))))).
%@ % 938,719 inferences, 0.074 CPU in 0.075 seconds (99% CPU, 12701698 Lips)
%@ Ns = [1, 2, 3, 4, 5, 6, 7, 8, 9|...],
%@ Sum = 100010000.
