:- module(path_count5, [
	]).

:- use_module(library(apply)).
:- use_module(library(apply_macros)).
:- use_module(library(clpfd)).
:- use_module(library(statistics)).
:- use_module(zdd('zdd-array')).
:- use_module(util(math)).
:- use_module(util(meta2)).
:- use_module(pac(basic)).
:- use_module(pac(meta)).
:- use_module(util(misc)).
:- use_module(pac('expand-pac')). % For the kind block.
:- use_module(zdd('zdd-misc')).
:- use_module(pac(op)).
:- use_module(zdd(zdd)).

:- set_prolog_flag(stack_limit, 10_200_147_483_648).

 :- op(1060, xfy, ~).		% equivalence
 :- op(1060, xfy, #).		% exor
 :- op(1060, xfy, <->).		% equivalence
 :- op(1050, yfx, <-).
 :- op(1060, xfy, <=> ).	% equivalence
 :- op(1040, xfy, \/).		% OR
 :- op(1030, xfy, /\).		% AND
 :- op(1020, fy, \).		% NOT
 :- op(700, xfx, :=).		% Assignment
 :- op(1000, xfy, &).

% for pac query.
 :- pac:op(1000, xfy, &).
 :- pac:op(700, xfx, :=).

term_expansion --> pac:expand_pac.

% some tiny.
x(p(X, _), X).
y(p(_, Y), Y).
link_symbol(_->_).
change_symbol(A-B, A->B).
%
touch(A-_, _-A):-!.
touch(_-A, A-_).

% ?- connect_pairs(a-b, [b-c, d-a], W).
% ?- connect_pairs(a-b, [a-c, a-b], W).  % false

connect_pairs(X-Y, [U-X, Y-V], U-V):-!.
connect_pairs(X-Y, [Y-V, U-X], U-V).

% ?- both_ends_on_frontier(p(3, 2)-p(3,5), p(2, 1)-p(3, 1)).
% ?- both_ends_on_frontier(p(2, 2)-p(3,5), p(2, 1)-p(3, 1)).  %false
% ?- both_ends_on_frontier(p(2, 2)-p(2,1), p(2, 1)-p(3, 1)).  %false
both_ends_on_frontier(p(S, _)-p(S, _), p(J, _)-p(K, _)):- S is max(J, K).

% ?- compare_hybrid(C, c-d, c->d).
% ?- compare_hybrid(C, c->d, c->d).
% ?- compare_hybrid(C, c->d, c-d).
% ?- compare_hybrid(C, c-d, c-d).
% ?- predsort(compare_hybrid, [a->b, b->a], R).
% ?- predsort(compare_hybrid, [a-b, b-a], R).
% ?- predsort(compare_hybrid, [b->a, a->b, a-b, a->b, c-b, a-b, b-a], R).
% ?- zdd((set_compare(compare_hybrid), zdd_compare(C, a-b, c-d))).
% ?- zdd((set_compare(compare_hybrid), zdd_compare(C, c-d, c->d))).
% ?- zdd((set_compare(compare_hybrid), zdd_compare(C, c->d, c-d))).
% ?- zdd((set_compare(compare_hybrid), zdd_insert_atoms([a-b, c-d], 1, X))).
% ?- zdd((set_compare(compare_hybrid), zdd_insert_atoms([a->b, c-d], 1, X), psa(X))).

compare_hybrid(C, X, Y):- compare(C0, X, Y),
	(	C0 = (=)  -> C = (=)
	;	functor(X, Fx, _),
		functor(Y, Fy, _),
		(	Fx == Fy
		-> 	(	Fx = (-)
			->  ( C0 = (<) -> C = (>) ; C = (<) )
			;	C = C0
			)
		;	(	Fx = (-) -> C = (<) ; C = (>) )
		)
	).

:- meta_predicate rect(:).

rect(X):- zdd((set_compare(compare_hybrid), X)).

% ?- rect((X<<{[p(0,0)-p(2,1)], [p(1,1)-p(2,3), p(2,2)-p(2,2)]},
%	prune_frontier(X, 2, p(0,0), Y), psa(Y))).
% ?- rect((X<<{[p(0,0)-p(2,1)], [p(1,1)-p(1,1), p(2,2)-p(2,2)]},
%	prune_frontier(X, 2, p(0,0), Y), psa(Y))).
% ?- rect((X<<{[p(0,0)-p(2,1)], [p(1,1)-p(1,1), p(2,2)-p(2,3)]},
%	prune_frontier(X, 2, p(0,0), Y), psa(Y))).

prune_frontier(F, I, P, F0):- @(prune_frontier(F, I, P, F0)).
%
prune_frontier(F, _, _, F, _):- F<2, !.
prune_frontier(F, I, P, F0, S):- cofact(F, t(A, L, R), S),
	prune_frontier(L, I, P, L0, S),
	(	A = (U-V)
	->	(	(	U = V
			;	x(U, I),
				x(V, I)
			;	U = P,
				x(V, I)
			)
		->	prune_frontier(R, I, P, R0, S)
		;	R0 = 0
		)
	;	R0 = R
	),
	cofact(F0, t(A, L0, R0), S).


		/**********************
		*     bridge block    *
		**********************/

fold_bridge_block(Es, X, Y, Z):- @(fold_bridge_block(Es, X, Y, 0, Z)).

fold_bridge_block(Es, X, Y, Z0, Z):- @(fold_bridge_block(Es, X, Y, Z0, Z)).

%
fold_bridge_block(_, _, [], _, Z, Z, _):- !.
fold_bridge_block(E, As, X, Y, Z0, Z, S):- cofact(X, t(A, L, R), S),
	(  link_symbol(A) ->   Z = Z0
	;  bridge_block(E, As, L, Y, Z0, Z1, S),
	   (	touch(E, A)
	   ->	bridge_block(A, E, As, R, Y, Z1, Z, S)
	   ;	bridge_block(E, [A|As], R, Y, Z1, Z, S)
	   )
	).


% ?- zdd((
%	E = p(1,0)-p(0,0),
%	X << {[p(0,0)-p(0,0)]},
%	Y << {[p(1,0)-p(1,0), p(1,1)-p(1,1), p(3,1)->p(3,2)]},
%	bridge_block(E, X, Y, Z),
%	psa(Z))).

% ?- zdd((
%	E = p(1,0)-p(0,0), E0=p(0,0)-p(1,0),
%	X << {[p(0,0)-p(0,0), p(0,1)-p(0,1)]},
%	Y << {[p(1,0)-p(1,0), p(1,1)-p(1,1)]},
%	bridge_block(E, X, Y, 0, Z0),
%	zdd_join(Z0, X, X0),
%	bridge_block(E0, X0, Y, Z0, Z1),
%	psa(Z0), psa(Z1))).

bridge_block(E, X, Y, Z):-  % assuming block order X < Y
	@(bridge_block(E, [], X, Y, 0, Z)).

bridge_block(E, X, Y, Z0, Z):- @(bridge_block(E, [], X, Y, Z0, Z)).

%
bridge_block(_, _, X, _, Z, Z, _):- X<2, !.
bridge_block(E, As, X, Y, Z0, Z, S):- cofact(X, t(A, L, R), S),
	(  link_symbol(A) ->   Z = Z0
	;  bridge_block(E, As, L, Y, Z0, Z1, S),
	   (	touch(E, A)
	   ->	bridge_block(A, E, As, R, Y, Z1, Z, S)
	   ;	bridge_block(E, [A|As], R, Y, Z1, Z, S)
	   )
	).
%
bridge_block(_, _, _, _, Y, Z, Z, _):- Y<2, !.
bridge_block(A, E, As, X, Y, Z0, Z, S):- cofact(Y, t(B, L, R), S),
	(	link_symbol(B) ->  Z = Z0
	;	bridge_block(A, E, As, X, L, Z0, Z1, S),
		(	touch(E, B)
		->	connect_pairs(E, [A, B], C),
			zdd_merge(X, R, U, S),
			change_symbol(E, E0),
			zdd_insert_atoms([E0, C|As], U, V, S),
			zdd_join(V, Z1, Z, S)
		;	bridge_block(A, E, [B|As], X, R, Z1, Z, S)
		)
	).


easy_bridge_block(Es, X, Y, Z):-  @(easy_bridge_block(Es, X, Y, Z)).

easy_bridge_block(Es, X, Y, Z, S):-  s_zdd_merge(X, Y, Z0, S),
	mqp_update_links(Es, Z0, Z, S).

		/*************
		*     end    *
		*************/

		/****************
		*     bridge    *
		****************/


% ?- zdd((count_path(p(0,0), p(0, 1), C))).
% ?- zdd((count_path(p(0,0), p(1, 0), C))).
% ?- zdd((count_path(p(0,0), p(1,1), C))).
% ?- zdd((count_path(p(0,0), p(2,2), C))).
% ?- zdd((count_path(p(0,0), p(3,3), C))).
% ?- zdd((count_path(p(1,1), p(2,1), C))).
% ?- zdd((count_path(p(1,2), p(3,2), C))).
% ?- zdd((count_path(p(1,2),p(10,2), C))).
% ?- zdd((count_path(p(1,2),p(12,2), C))).
% ?- zdd((count_path(p(1,2),p(13,2), C))).
% ?- zdd((count_path(p(1,2),p(14,2), C))).

% ?- zdd((count_path(p(1,2),p(100,2), C))).
% ?- time(zdd((count_path_line(p(1,2),p(1000, 2), C)))).
%@ % 4,691,349 inferences, 0.982 CPU in 1.154 seconds
% ?- time(zdd((count_path_line(p(1,2),p(10000,2), C)))).
%@ % 61,745,569 inferences, 59.803 CPU in 60.290 seconds

count_path(X, Y, C):- @(count_path(X, Y, C)).
%
count_path(X, Y, C, S):-
	qp_block(X, Y, R, S),
	memo_final(R, X, Y, S),
	qp_list(Q, [X-X], S),
	path_tree(Q, Tree, S),
	card(Tree, C, S).
%
memo_final([], _, _, _).
memo_final([A|R], X, Y, S):-
	qp_list(A, List, S),
	(	qp_final(List, X, Y)
	->  memo(qp_final(A)-true, S)
	;  true
	),
	memo_final(R, X, Y, S).
%
qp_block(X, Y, R):- @(qp_block(X, Y, R)).
%
qp_block(X, X, [Q], S):-!,
	zdd_insert_atoms([X-X], 1, Q, S),
	memo(qp_suc(Q)-List, S),
	( var(List) -> List=[]; true).
qp_block(p(I, J), p(I, K), R, S):-!,
	M is (J+K)//2,
	qp_block(p(I, J), p(I, M), R1, S),
	M0 is M+1,
	qp_block(p(I, M0), p(I, K), R2, S),
	qp_joint([p(I, M)-p(I, M0), p(I, M0)-p(I, M)], R1, R2, R3, S),
	prune_block_local(R3, p(I, J), p(I, K), R, S).
qp_block(p(I, J), p(K, L), R, S):-
	M is (I+K)//2,
	qp_block(p(I, J), p(M, L), R1, S),
	M0 is M+1,
	qp_block(p(M0,J), p(K, L), R2, S),
	findall(p(M, H)-p(M0, H), between(J, L, H), U),
	findall(p(M0, H0)-p(M, H0), between(J, L, H0), V),
	append(U, V, W),
	sort(W, Es),
	qp_joint(Es, R1, R2, R3, S),
	prune_block_local(R3, p(I, J), p(K, L), R, S).
%
prune_block_local(X, _, _, X, _).

% ?- zdd(rect_bundle(p(0,0), p(0,0), Bs, Cs)), length(Bs, Len).
% ?- zdd(rect_bundle(p(0,0), p(1,1), Bs, Cs)), length(Bs, Len).
% ?- zdd(rect_bundle(p(0,0), p(1,0), Bs, Cs)), length(Bs, Len).
% ?- time(zdd(rect_bundle(p(0,0), p(10,10), Bs, Cs))),
%	length(Bs, Len), length(Cs, LenC).

rect_bundle(LL, UR, Bridges, Column_qps):-
	bridge_bundle(LL, UR, Bridges),
	column_qp_bundle(LL, UR, Column_qps).
%
column_qp_bundle(X, Y, Z):- @(column_qp_bundle(X, Y, Z)).
%
column_qp_bundle(p(I, J), p(K, L), Z, S):-
	qp_along_line(J, L, List),
	convert_to_col(List, Q, I, S),
	U is K - I,
	column_qp_bundle(U, [Q], Z0, S),
	reverse(Z0, Z).

%
column_qp_bundle(0, Z, Z, _):- !.
column_qp_bundle(K, [Q|Z], Z0, S):-
	map_h_shift(Q, Q0, S),
	K0 is K-1,
	column_qp_bundle(K0, [Q0, Q|Z], Z0, S).

%
zdd_repeat(0, _, X, X, _):-!.
zdd_repeat(N, F, X, Y, S):- N>0,  call(F, X, U, S),
	N0 is N-1,
	zdd_repeat(N0, F, U, Y, S).


% ?- bridge_bundle(p(0,0), p(0,0), S).
% ?- bridge_bundle(p(0,0), p(1,0), S).
% ?- bridge_bundle(p(0,0), p(1,1), S).
% ?- bridge_bundle(p(0,0), p(3,1), S), maplist(writeln, S).
% ?- time((bridge_bundle(p(0,0), p(100,2), S), length(S, Len))).

bridge_bundle(p(I, _), p(I, _), []):-!.
bridge_bundle(p(I, J), p(K, L), Bundle):-
	findall(Bridge, (	between(I, K, U),
						U0 is U-1,
						U0 >= I,
						findall( H,
								(  between(J, L, V),
								   (  H = p(U0, V)-p(U, V)
								   ;  H = p(U, V)-p(U0, V)
								   )
								),
								Bridge)
					),
			Bundle).

% ?- bridge_shift([[p(1,1)-p(1,1)]], X).
bridge_shift(X, Y):-maplist(maplist(h_shift_link), X, Y).

%
h_shift_link(p(X, Y)-p(Z, U), p(X0, Y)-p(Z0, U)):-
	X0 is X+1,
	Z0 is Z+1.
%
v_shift_link(p(X, Y)-p(Z, U), p(X, Y0)-p(Z, U0)):-
	Y0 is Y+1,
	U0 is U+1.

% ?- zdd(convert_to_row([[1-2], [3-4]], Row, 0)).
convert_to_row(X, Y, Z):- @(convert_to_row(X, Y, Z)).
%
convert_to_row([], [], _, _).
convert_to_row([X|Y], [X0|Y0], H, S):-
	convert_to_row_(X, X1, H),
	zdd_insert_atoms(X1, 1, X0, S),
	convert_to_row(Y, Y0, H, S).

% ?- convert_to_row_([1-2, 3-4], R, 5).
convert_to_row_([], [], _).
convert_to_row_([I-J|X], [p(I, H)-p(J, H)|X0], H):-
	convert_to_row_(X, X0, H).

% ?- zdd(convert_to_col([[1-2], [3-4]], Col, 0)).
convert_to_col(X, Y, Z):- @(convert_to_col(X, Y, Z)).
%
convert_to_col([], [], _, _).
convert_to_col([X|Y], [X0|Y0], H, S):-
	convert_to_col_(X, X1, H),
	zdd_insert_atoms(X1, 1, X0, S),
	convert_to_col(Y, Y0, H, S).

% ?- convert_to_col_([1-2, 3-4], R, 5).
convert_to_col_([], [], _).
convert_to_col_([I-J|X], [p(H, I)-p(H, J)|X0], H):-
	convert_to_col_(X, X0, H).

% ?- zdd((convert_to_col([[1-2], [3-4]], Col, 0),
%	map_h_shift(Col, Col0))).
map_h_shift(X, Y):- @(map_h_shift(X, Y)).
%
map_h_shift([], [], _).
map_h_shift([X|Y], [X0|Y0], S):- h_shift(X, X0, S),
	map_h_shift(Y, Y0, S).

% ?- zdd((convert_to_row([[1-2], [3-4]], Row, 0),
%	map_v_shift(Row, Row0))).
map_v_shift(X, Y):- @(map_v_shift(X, Y)).
%
map_v_shift([], [], _).
map_v_shift([X|Y], [X0|Y0], S):- v_shift(X, X0, S),
	map_v_shift(Y, Y0, S).

% ?- zdd((qp_along_line(1, 2, X), convert_to_col(X, Y, 3),
%	maplist(h_shift, Y, Y0),
%	forall(member(M, Y0), (qp_list(M, L), writeln(L))))).
%
h_shift(X, Y):- @(h_shift(X, Y)).
%
h_shift(X, X, _):- X < 2, !.
h_shift(X, Y, S):- cofact(X, t(p(I, J)-p(I, K), L, R), S),
	I0 is I + 1,
	h_shift(L, L0, S),
	h_shift(R, R0, S),
	cofact(Y, t(p(I0, J)-p(I0, K), L0, R0), S).

% ?- zdd((qp_along_line(1, 2, X), convert_to_row(X, Y, 3),
%	maplist(v_shift, Y, Y0),
%	forall(member(M, Y0), (qp_list(M, L), writeln(L))))).

v_shift(X, Y):- @(v_shift(X, Y)).
%
v_shift(X, X, _):- X < 2, !.
v_shift(X, Y, S):- cofact(X, t(p(I, J)-p(K, J), L, R), S),
	v_shift(L, L0, S),
	v_shift(R, R0, S),
	J0 is J + 1,
	cofact(Y, t(p(I, J0)-p(K, J0), L0, R0), S).


% ?- spy(solve_rect).
% ?- zdd((solve_rect(p(0,0),p(1,0), Z))).
% ?- zdd((solve_rect(p(0,0),p(1,1), Z))).
% ?- zdd((solve_rect(p(0,0),p(2,2), Z))).
% ?- zdd((solve_rect(p(0,0),p(3,3), Z))).
% ?- call_with_time_limit(300, zdd((solve_rect(p(0,0),p(4,4), Z)))).
%@ n(3,41)
%@ n(2,1681)
%@ n(1,68921)
%@ n(0,2825761)
%@ ERROR: Unhandled exception: Time limit exceeded


% ?- zdd((solve_rect(p(0,0),p(4,4), Z))).
%@ n(3)
%@ n(2)
%@ n(1)
%@ Z = [4996364, 4996368, 4996372, 4996376, 4996380, 4996384, 4996388, 4996392, 4996396|...] .


% ?- zdd((solve_rect(p(0,0),p(15,15), Z))).


solve_rect(X, Y, Z):- @(solve_rect(X, Y, Z)).
%
solve_rect(p(I, J), p(I, L), Z, S):-!, initial_column(I, J, L, Z, S).
solve_rect(p(I, J), p(K, L), Z, S):-
	initial_col_bridge(I, J, L, Col, B, S),
	N is (K-I)-1,
	map_h_shift(Col, Col0, S),
	repeat_bridge(N, Col, B, Col0, Z, S).

%
repeat_bridge(0, X, B, C, Y, S):-!, length(X, Len), writeln(n(0,Len)),
	qp_joint(B, X, C, Y, S).
repeat_bridge(N, X, B, C, Y, S):- length(X, Len), writeln(n(N, Len)),
	qp_joint(B, X, C, Y0, S),
	bridge_shift(B, B0),
	map_h_shift(C, C0, S),
	N0 is N-1,
	repeat_bridge(N0, Y0, B0, C0, Y, S).


% ?- zdd((initial_col_bridge(0, 0, 3, C, B))).
% ?- time(zdd((initial_col_bridge(0, 0, 15, C, B)))).
initial_col_bridge(I, L, H, C, B):-	@(initial_col_bridge(I, L, H, C, B)).
%
initial_col_bridge(I, L, H, C, B, S):- initial_column(I, L, H, C, S),
	initial_bridge(I, L, H, B).


% ?- zdd((initial_column(1, 2, 4, Col),
%   forall(member(M, Col), (qp_list(M, L), writeln(L))))).
initial_column(I,Low,Hi, Col):- @(initial_column(I, Low, Hi, Col)).
%
initial_column(I,Low,Hi, Col, S):-
	qp_along_line(Low, Hi, X),
	convert_to_col(X, Col, I, S).

% ?- initial_bridge(0, 1, 4, Bridge), maplist(writeln, Bridge).
initial_bridge(I, Low, Hi, Bridge):- J is I + 1,
	findall([p(I, V)-p(J, V), p(J, V)-p(I, V)],
			between(Low, Hi, V),
			Bridge).


		/****************
		*     prune     *
		****************/

% ?- zdd((qp_along_line(1, 13, X), length(X, L), convert_to_col(X, Y, 3),
%	prune_by_column(Y, 3, p(3,1)-p(3,1), Y0), length(Y, Ly))).

prune_by_column(F, X, P, F0):- @(prune_by_column(F, X, P, F0)).
%
prune_by_column([], _, _, [], _).
prune_by_column([Q|F], X, P, [Q|F0], S):- admissible_qp(Q, X, P, S), !,
	prune_by_column(F, X, P, F0, S).
prune_by_column([_|F], X, P, F0, S):-
	prune_by_column(F, X, P, F0, S).
%
admissible_qp(Q, X, P):- @(admissible_qp(Q, X, P)).
%
admissible_qp(Q, X, P, S):- qp_list(Q, List, S),
	admissible_qp_list(List, X, P).

% ?- admissible_qp_list([p(0,0)-p(2,1), p(3,1)-p(3,1)], 2, p(0,0)).
admissible_qp_list([], _, _):-!.
admissible_qp_list([P-p(X, _)|List], X, P):-!,
	admissible_qp_list(List, X, P).
admissible_qp_list([p(X, _)-p(X, _)|List], X, P):- !,
	admissible_qp_list(List, X, P).
admissible_qp_list([p(Z, _)-_|_], X, _):- Z>X, !.
admissible_qp_list([P-P|List], X, P):- admissible_qp_list(List, X, P).

% ?- rect((qp_along_line(1,1, Ls), maplist_qp(Ls, Q))).
% ?- rect((qp_along_line(1,10, Ls), maplist_qp(Ls, Q))).
% ?- time(rect((qp_along_line(1, 15, Ls), maplist_qp(Ls, Q)))).
%@ % 99,784,517 inferences, 8.136 CPU in 8.371 seconds (97% CPU, 12264944 Lips)

%
maplist_qp(X, Y):- @(maplist_qp(X, Y)).
%
maplist_qp([], [], _).
maplist_qp([L|Ls], [Q|Qs], S):-
	qp_list(Q, L, S),
	maplist_qp(Ls, Qs, S).

% simple line quasi path
qp_line(I, I, Q, S):-!, zdd_singleton(I, Q, S).
qp_line(I, J, Q, S):-
	M is (I+J)//2,
	M0 is M+1,
	qp_line(I, M, R, S),
	qp_line(M0, J, R0, S),
	bridge_line(M, M0, R, R0, Q0, S),
	bridge_line(M, M0, Q0, R0, Q1, S),
	zdd_join(R, R0, R1, S),
	zdd_join(R1, Q1, Q, S).
%
bridge_line(I, J, X, Y, Z):- @(bridge_line(I, J, X, Y, 0, Z)).

bridge_line(_, _, X, 0, Z, Z, _):- X < 2, !.
bridge_line(_, _, 0, X, Z, Z, _):- X < 2, !.
bridge_line(I, J, X, Y, Z0, Z, S):- cofact(X, t(J0-K, L, R), S),
	( J0 < I
	->	Z = Z0
	;  	bridge_line(I, J, L, Y, Z0, Z1, S),
		(	J0 = J
		->	zdd_insert(I-K, R, R0, S),
			zdd_join(R0, Z1, Z, S)
		;	bridge_line(I, J, R, Y, 0, Z2, S),
			zdd_insert(J0-K, Z2, Z3, S),
			zdd_join(Z3, Z1, Z, S)
		)
	).


% ?- qp_along_line(1,1,X).
% ?- I = 6, qp_along_line(0,I,X), length(X, L), N is 2^(I+1).
% ?- I = 7, qp_along_line(0,I,X), length(X, L), N is 2^(I+1).
% ?- I = 20, qp_along_line(0,I,X), length(X, L), N is 2^(I+1).
%@ I = 20,
%@ X = [[0-20], [0-19, 20-20], [0-16, 17-20], [0-16, 17-19, 20-20], [0-17, 18-20], [0-17, 18-19, ... - ...], [0-16, ... - ...|...], [... - ...|...], [...|...]|...],
%@ L = 54608393,
%@ N = 2097152 .
% ?- qp_along_line(0,3,X), maplist(writeln, X).

% simple line quasi path
qp_along_line(I, I, [[I-I]]).
qp_along_line(I, J, Q):-
	M is (I+J)//2,
	M0 is M+1,
	qp_along_line(I, M, R),
	qp_along_line(M0, J, R0),
	simple_prod_union(R, R0, R1, []),
	simple_bridge_links([M-M0, M0-M], R, R0, Q, R1).


%
map_qp_list([], X, X, _).
map_qp_list([L|Ls], X, Y, S):- qp_list(Q, L, S),
	zdd_join(Q, X, X0, S),
	map_pq_list(Ls, X0, Y, S).

% ?- simple_bridge(1-2, [1-1], [2-2], R).
% ?- simple_bridge(2-1, [1-1], [2-2], R).
% ?- simple_bridge(2-3, [1-1], [2-2], R).
simple_bridge(A-B, L, M, N):-	% assuming qp  X < qp Y
	(	select(B-V, M, M0),
		select(U-A, L, L0)
	;	select(B-V, L, L0),
		select(U-A, M, M0)
	),
	!,
	(	U @> V
	-> 	ord_union(L0, [U-V|M0], N)
	; 	ord_union([U-V], M0, M1),
		ord_union(L0, M1, N)
	).
simple_bridge(_, _, _, []).

%
simple_bridge_links([], _, _, Z, Z).
simple_bridge_links([E|Es], X, Y, Z, U):-
	simple_bridge_link(X, E, Y, Z, Z0),
	simple_bridge_links(Es, X, Y, Z0, U).
%
simple_bridge_link([], _, _, Z, Z).
simple_bridge_link([X|Xs], E, Y, Z, U):-
	simple_bridge_basic(Y, X, E, Z, Z0),
	simple_bridge_link(Xs, E, Y, Z0, U).
%
simple_bridge_basic([], _, _, Z, Z).
simple_bridge_basic([Y|Ys], X, E, Z, U):- simple_bridge(E, X, Y, V),
	(	V = []
	->	Z0 = Z
	;	Z  = [V|Z0]
	),
	simple_bridge_basic(Ys, X, E, Z0, U).

% ?- Q = [a-a], Q0=[b-b], Q1 = [c-c],
%	simple_prod_union([Q, Q0], [Q1], Z).

%
simple_prod_union(X, Y, Z):- simple_prod_union(X, Y, Z, []).
%
simple_prod_union([], _, U, U).
simple_prod_union([L|Y], Z, U, V):-
	simple_union(Z, L, U, U0),
	simple_prod_union(Y, Z, U0, V).
%
simple_union([], _, U, U).
simple_union([M|As], L, [N|U], V):-
	ord_union(L, M, N),
	simple_union(As, L, U, V).

%% test basics.
% ?- zdd((
%	qp_list(X, [a-a]), qp_list(Y, [b-b]),
%	qp_joint([a-b, b-a], [X], [Y], Z),
%	maplist(pred(([U]:-qp_list(U, List), writeln(List))), Z))).
% ?- zdd((
%	qp_list(X, [a-a]), qp_list(Y, [b-b]),
%	qp_joint([a-b, b-a], [X], [Y], Z),
%	zdd_join(X, Y, A),
%	memo(qp_suc(A)-L), writeln(qp_suc(A)-L),
%	maplist(pred(([U]:-qp_list(U, List), writeln(List))), Z))).
% ?- zdd((
%	qp_list(X, [a-a]), qp_list(Y, [b-b]),
%	qp_joint([a-b, b-a], [X], [Y], Z),
%	zdd_join(X, Y, A),
%	memo(qp_suc(A)-L), writeln(qp_suc(A)-L),
%	maplist(pred(([U]:-qp_list(U, List), writeln(List))), Z))).

qp_joint(X, Y, Z, U):-@(qp_joint(X, Y, Z, U)).
%
qp_joint(X, Y, Z, U, S):- map_prod_sum(Y, Z, V, [], S),
	qp_bridge_links(X, Y, Z, U, V, S).

% ?- zdd((qp_list(X, [a-a]), qp_list(Y, [b-b]),
%	qp_bridge(a-b, X, Y, Z), qp_list(Z, U))).
% ?- zdd((qp_list(X, [a-a]), qp_list(Y, [b-b]),
%	qp_bridge(b-a, X, Y, Z), qp_list(Z, U))).

qp_bridge(E, X, Y, Z):- @(qp_bridge(E, X, Y, Z)).
%
qp_bridge(A-B, X, Y, Z, S):- % assuming qp  X < qp Y
	qp_list(X, L, S),
	qp_list(Y, M, S),
	(	select(U-A, L, L0),
		select(B-V, M, M0)
	;	select(B-V, L, L0),
		select(U-A, M, M0)
	),
	!,
	ord_union(L0, M0, N0),
	ord_union([U-V], N0, N),
	qp_list(Z, N, S).
qp_bridge(_, _, _, 0, _).

%
qp_bridge_links(X, Y, Z, U, V):- @(qp_bridge_links(X, Y, Z, U, V)).
%
qp_bridge_links([], _, _, Z, Z, _).
qp_bridge_links([E|Es], X, Y, Z, U, S):-
	qp_bridge_link(X, E, Y, Z, Z0, S),
	qp_bridge_links(Es, X, Y, Z0, U, S).
%
qp_bridge_link([], _, _, Z, Z, _).
qp_bridge_link([X|Xs], E, Y, Z, U, S):-
	qp_bridge_basic(Y, X, E, Z, Z0, S),
	qp_bridge_link(Xs, E, Y, Z0, U, S).
%
qp_bridge_basic([], _, _, Z, Z, _).
qp_bridge_basic([Y|Ys], X, E, Z, U, S):- qp_bridge(E, X, Y, V, S),
	(	V = 0
	->	Z0 = Z
	;	Z  = [V|Z0],
		select_goto(E, X, Y, V, S)
	),
	qp_bridge_basic(Ys, X, E, Z0, U, S).
%
select_goto(E, X, Y, Z):- @(select_goto(E, X, Y, Z)).
%
select_goto(A-B, X, Y, Z, S):-
	G = (A-B)-Z,
	( A@<B -> W = X ; W = Y ),
	getmemo(qp_suc(W)-U, V, S),
	(	var(V) -> V = [] ; true ),
	add_new(G, V, U).

		/***********************
		*     end of bridge    *
		***********************/

% ?-  belong(3, 3, 4).
% ?-  belong(3, 4, 1).
belong(X, I, J):- I @=< J, !, I @=< X, X @=< J.
belong(X, I, J):- J @=< X, X @=< I.
%
sub_interval(I, J, X, Y):- belong(I, X, Y), belong(J, X, Y).
%
disjoint_interval(I, J, X, Y):- \+ belong(I, X, Y), \+ belong(J, X, Y).

% ?- compatible_interval(1, 2, 3, 4).
% ?- compatible_interval(1, 3, 2, 4). % false
% ?- compatible_interval(2, 4, 5, 7).
% ?- compatible_interval(7, 5, 2, 4).

compatible_interval(I, J, X, Y):- disjoint_interval(I, J, X, Y), !.
compatible_interval(I, J, X, Y):- sub_interval(I, J, X, Y), !.
compatible_interval(I, J, X, Y):- sub_interval(X, Y, I, J).

% ?- rect_dg(rect(1,1), X, Y).
% ?- rect_dg(rect(0,1), X, Y).
% ?- rect_dg(rect(0,0), X, Y).

rect_dg(rect(W, H), Nodes, Links):-
	rect_nodes(W, H, Nodes),
	rect_links(W, H, Nodes, Links).
% ?- rect_nodes(4,5,Ns), length(Ns, L).

rect_nodes(W, H, Nodes):-
	findall(p(I, J),
			(	between(0, W, I),
				between(0, H, J)
			),
			Nodes0),
	sort(Nodes0, Nodes).

%
rect_links(W, H, Ns, Links):-
	findall(P-Q,
			(	member(P, Ns),
				P = p(I, J),
				(	Q = p(I, J1),
					(	J1 is J-1,
						J1 >= 0
					;	J1 is J+1,
						J1 =< H
					)
				;   Q = p(I1, J),
					(	I1 is I-1,
						I1 >= 0
					;	I1 is I+1,
						I1 =< W
					)
				)
			),
			Links0),
	sort(Links0, Links).

		/********************
		*   frontier based  *
		********************/

% ?- zdd(path_count(rect(0,0), C)).
% ?- zdd(path_count(rect(1,0), C)).
% ?- zdd(path_count(rect(0,1), C)).
% ?- zdd(path_count(rect(1,1), C)).
% ?- zdd(path_count(rect(2, 1), C)).
% ?- zdd(path_count(rect(3, 1), C)).
% ?- zdd(path_count(rect(1, 3), C)).
% ?- zdd(path_count(rect(4, 1), C)).
% ?- zdd(path_count(rect(2, 3), C)).
% ?- zdd(path_count(rect(4,3), C)).
%@ C = 976 .
% ?- zdd(path_count(rect(4,4), C)).
%@ C= 8512
% ?- call_with_time_limit(360, zdd(path_count(rect(5, 5), C))).

path_count(Rect, C):- @(path_count(Rect, C)).

%
path_count(rect(0,0), 1, _):-!.
path_count(Rect, C, S):-
	rect_dg(Rect, Ns, Links),
	findall(A-A, member(A, Ns), Ids),
	Rect = rect(W, H),
	set_key(final, p(W, H), S),
	qp_list(Q, Ids, S),
	memo(qp_suc(Q)-[], S),
	frontier(Links, [Q], FinalQs, S),
	fold_qp_memo_final(FinalQs, S),
	path_tree(Q, Tree, S),
	card(Tree, C, S).

%
frontier([], Qs, Qs, _).
frontier([E|Es], Qs, Qs0, S):-
	frontier_link(E, Qs, Qs1, Qs, S),
	prune_frontier(Qs1, E, Qs2, S),
	frontier(Es, Qs2, Qs0, S).

%
frontier_link(_, [], Qs, Qs, _):-!.
frontier_link(E, [X|Qs], Qs0, Qs1, S):- X<2, !,
 	frontier_link(E, Qs, Qs0, Qs1, S).
frontier_link(E, [Q|Qs], Next, Qs1, S):-
	qp_update(E, Q, Q0, S),
	(	Q0 < 2
	->	Next = Qs0
	;	qp_memo_final(Q0, S),
		Next = [Q0|Qs0]
	),
	qp_update_succs(E-Q0, Q, S),
	frontier_link(E, Qs, Qs0, Qs1, S).

% ?- zdd((qp_list(Q, [a]), qp_list(Q0, [b]), qp_list(Q1, [c]),
%	map_prod_sum([Q, Q0], [Q1], Z),
%	maplist(pred([A, M]:-qp_list(A, M)), Z, W))).
%
map_prod_sum(X, Y, Z):- map_prod_sum(X, Y, Z, []).
%
map_prod_sum(X, Y, Z, U):- @(map_prod_sum(X, Y, Z, U)).

map_prod_sum([], _, U, U, _).
map_prod_sum([Q|Y], Z, U, V, S):- qp_list(Q, L, S),
	map_qp_sum(Z, L, U, U0, S),
	map_prod_sum(Y, Z, U0, V, S).
%
map_qp_sum([], _, U, U, _).
map_qp_sum([Q|As], L, [Q0|U], V, S):-qp_list(Q, M, S),
	ord_union(L, M, N),
	qp_list(Q0, N, S),
	map_qp_sum(As, L, U, V, S).


% ?- zdd((path_count(rect(1,0), C))).
% ?- zdd((path_count(rect(0,1), C))).
% ?- zdd((path_count(rect(1,1), C))).
% ?- zdd((path_count(rect(3,1), C))).
% ?- time(zdd((path_count(rect(1,8), C)))).   % take time.
%@ % 86,018,490 inferences, 10.604 CPU in 11.323
% ?- time(zdd((path_count(rect(8,1), C)))).
%@ % 9,254,518 inferences, 3.089 CPU in 4.850
% ?- zdd((path_count(rect(2,3), C))).
% ?- zdd((path_count(rect(3,2), C))).
% ?- time(zdd((path_count(rect(3,3), C)))).
%@ % 50,584,360 inferences, 7.362 CPU in 9.480
%@ C = 184 .
% ?- time(zdd((path_count(rect(4,4), C)))).
%@ % 9,474,429,287 inferences, 1256.779 CPU in 1433.734 seconds (88% CPU, 7538660 Lips)
%@ C = 8512 .

qp_update(E, X, Y):- @(qp_update(E, X, Y)).
%
qp_update(_, X, 0, _):- X < 2.
qp_update(E, X, Y, S):- qp_list(X, U, S),
	memo(qp_update(E, X)-Y, S),
	(	nonvar(Y)  ->	true
	;	select_touch(E, [A, B], U, U0)
	->	connect_pairs(E, [A, B], W),
		ord_union([W], U0, V),
		qp_list(Y, V, S)
	;	Y = 0
	).
qp_update(_, _, 0,  _).

% ?- zdd((qp_list(X, [a-a, b-b]), memo(qp_suc(X)-[]))).
qp_update_succs(G, Q):- @(qp_update_succs(G, Q)).

% ?- zdd(qp_update_succs(a, 1)).
qp_update_succs(G, Q, S):- getmemo(qp_suc(Q)-U, V, S),
	G = E-Q0,
	(	var(V) -> U = [E-Q0]
	;	add_new(E-Q0, V, U)
	).

qp_admissible(L):-
	forall((select(A-B, L, L0), select(C-D, L0, _)),
		   compatible_interval(A, B, C, D)).

% ?- compatible_interval(1, 2, 3, 4).
% ?- compatible_interval(1, 3, 2, 4). % false
% ?- compatible_interval(2, 4, 5, 7).
% ?- compatible_interval(7, 5, 2, 4).


% compatible_interval(I, J, X, Y):- disjoint_interval(I, J, X, Y), !.
% compatible_interval(I, J, X, Y):- sub_interval(I, J, X, Y), !.
% compatible_interval(I, J, X, Y):- sub_interval(X, Y, I, J).
%@ true.
%@ path-count4.pl qcompiled.

% ?- select_touch(a-b, [A, B], [a-a, b-b], X).
% ?- select_touch(a-b, [A, B], [a-a, b-b], X),
%	connect_pairs(a-b, [A, B], Y).
%@ false.

% ?- select_touch(a-b, [A, B], [a-a, b-c], X),
%	connect_pairs(a-b, [A, B], Y).
% ?- select_touch(a-b, [A, B], [a-a, c-b], X).

select_touch(_, [], V, V).
select_touch(E, [A|As], [A|U], V):- touch(E, A), !,
	select_touch(E, As, U, V).
select_touch(E, As, [A|U], [A|V]):-select_touch(E, As, U, V).

% ?- add_new(a,[b,c,a], X).
add_new(X, Y, Y):- memberchk(X, Y), !.
add_new(X, Y, [X|Y]).

% ?- zdd((qp_list(Q, [p(0,0)-p(1,0)]), qp_list(Q, L))).
%@ L = [p(0, 0)-p(1, 0)].
%
qp_list(Q, List):- @(qp_list(Q, List)).
%
qp_list(0, 0, _):-!.
qp_list(1, [], _):-!.   %% ???
qp_list(Q, List, S):- var(Q), !,
	zdd_insert_atoms(List, 1, Q, S),
	memo(qp_suc(Q)-U, S),
	(	var(U) -> U = []
	;	true			% Already registered.
	).
qp_list(Q, [A|As], S):- cofact(Q, t(A, 0, R), S),
	qp_list(R, As, S).

%
qp_memo_final(X):- @(qp_memo_final(X)).
%
qp_memo_final(Q, S):-  memoq(qp_final(Q)-true, S), !.
qp_memo_final(Q, S):- qp_list(Q, List, S),
	get_key(final, F, S),
	(	qp_final(List, F)
	->	memo(qp_final(Q)-true, S)
	;	true
	).
%
fold_qp_memo_final([], _).
fold_qp_memo_final([Q|Qs], S):- qp_memo_final(Q, S),
	fold_qp_memo_final(Qs, S).

%
path_tree(Q, T):- @(path_tree(Q, T)).
%

path_tree(X, 0,  _):- X<1, !.
path_tree(Q, 1, S):- memoq(qp_final(Q)-true, S), !.
path_tree(Q, T, S):- memo(path_tree(Q)-T, S),
	(	nonvar(T) -> true
	;	memo(qp_suc(Q) - L, S),		% writeln(qp_suc(Q)-L),
		path_tree_list(L, 0, T, S)
	).
%
path_tree_list([], C, C, _).
path_tree_list([E-Q|Qs], C0, C, S):-
	path_tree(Q, A, S),
	zdd_insert(E, A, A0, S),
	zdd_join(C0, A0, C1, S),
	path_tree_list(Qs, C1, C, S).

% ?- qp_final([p(0,0)-p(2,3),p(0,3)-p(0,3),p(1,3)-p(1,3)], p(2,3)).

qp_final([X-Y|R], X, Y):-!, qp_all_id(R).
qp_final([A-A|R], X, Y):- qp_final(R, X, Y).

qp_final([p(0,0)-F|R], F):-!, qp_all_id(R).
qp_final([A-A|R], F):- qp_final(R, F).

%
qp_all_id([]).
qp_all_id([A-A|R]):- qp_all_id(R).

%
dag_card(Q, C):- @(dag_card(Q, C)).
%
dag_card(Q, C, S):- open_state(M),
	dag_card(Q, C, S, M),
	close_state(M).
%
dag_card(Q, 0,  _, _):- Q<2, !.
dag_card(Q, 1, S, _):- memoq(qp_final(Q)-true, S), !.
dag_card(Q, C, S, M):- memo(dag_card(Q)-C, M),
	(	nonvar(C) -> true
	;	memo(qp_suc(Q) - L, S),
		dag_card_list(L, 0, C, S, M)
	).
%
dag_card_list([], C, C, _, _).
dag_card_list([_-Q|Qs], C0, C, S, M):-
	dag_card(Q, A, S, M),
	C1 is C0 + A,
	dag_card_list(Qs, C1, C, S, M).


		/***************
		*     debug    *
		***************/
%
snap_qp_update(E, X, Y, S):- qp_update(E, X, Y, S),
	qp_list(X, X0, S),
	qp_list(Y, Y0, S),
	format("~w\n~w\n~w\n\n", [qp_update(E, X, Y), X0, Y0]).
%
dump_frontier(Qs, S):- maplist(pred(S,
	([Q]:- qp_list(Q, U, S), writeln(U))), Qs).
%
dump_qp(X, _):-  X<2, !, writeln(X).
dump_qp(X, S):-  qp_list(X, L, S),
	writeln(X=L).

%
delta_star([], Q, Q, _).
delta_star([E|Es], Q,  Q0, S):-  memo(qp_suc(Q)-L, S),
	writeln(E),
	memberchk(E-Q1, L),
	dump_qp(Q1, S),
	delta_star(Es, Q1, Q0, S).

%
mirror(X, Y):- @(mirror(X, Y)).
%
mirror(X, X, _):- X<2, !.
mirror(X, Y, S):- cofact(X, t(p(I,J)-p(K,L), Lx, Rx), S),
	mirror(Lx, MLx, S),
	mirror(Rx, MRx, S),
	zdd_insert(p(J, I)-p(L, K), MRx, MRx0, S),
	zdd_join(MLx, MRx0, Y, S).

%
rect_path_tree(Rect, T):-@(rect_path_tree(Rect, T)).

rect_path_tree(Rect, T, S):-
	sort_links_by_rank(Rect, Rs, Ids),
	Rect = rect(W, H),
	set_key(final, p(W, H), S),
	qp_list(Q, Ids, S),
	repeat_frontier(Rs, [Q], FinalQs, S),
	fold_qp_memo_final(FinalQs, S),
	path_tree(Q, T, S).

% ?- spy(remove_end_links).
% ?- remove_end_links([1-[c-a]], R0, a-d).
% ?- remove_end_links([1-[c-a, d-c, a-d]], R0, a-d).

remove_end_links(R, R0, ST):- maplist(pred(ST,
	(	[J-Es, J-Es0]:- foldl(
		pred([ST, Es, Es0],
			 ( [P-Q, U, V]:- ST = S-T,
					(	( Q==S; P==T)
					->  V = U
					;	U = [P-Q|V]
					))),  Es,  Es0, []))),  R, R0).

%
rest(Q, X, Y, S):- memo(qp_suc(Q)-L, S),
	(	select(Q, X, X0)-> true
	;	X0 = X
	),
	rest_list(L, X0, Y, S).
%
rest_list([], X, X, _).
rest_list([_-Q|Qs], X, Y, S):-
	rest(Q, X, X0, S),
	rest_list(Qs, X0, Y, S).


% ?- zdd((dbg_qp_update(p(0,0)-p(1,0), [p(0,0)-p(0,0),p(1,0)-p(1,0)], L))).
% ?- spy(dbg_qp_update).
% ?- zdd((dbg_qp_update(p(2,2)-p(2,3), [p(0,0)-p(2,2),p(0,3)-p(0,3), p(1,3)-p(1,3), p(2,3)-p(2,3)], L))).

dbg_qp_update(E, X, Y):- @(dbg_qp_update(E, X, Y)).
%
dbg_qp_update(E, X, Y, S):- qp_list(Q, X, S),
	qp_update(E, Q, Q0, S),
	qp_list(Q0, Y, S).


% ?- zdd((rect_path_tree(rect(3,2), T), card(T, C),
%   zdd_insert_atoms([p(0,0)-p(2,3),p(0,3)-p(0,3),p(1,3)-p(1,3)], 1, W),
%	mirror(T, MT), card(MT, CMT),
%	rect_path_tree(rect(2,3), T0), card(T0, C0), mirror(T0, MT0),
%	card(MT0, CM0),
%	zdd_subtr(T, MT0, D), card(D, CD), psa(D)
%)).


% Test to show getmemo/2,3 works well.
% ?- time((numlist(1, 10000, Ns),
%	zdd((maplist( pred([A]:- getmemo(a(A)-A, _)), Ns),
%		 maplist( pred(([A]:- getmemo(a(A)-B, Old), B is 2*Old)),  Ns),
%		 foldl( pred(([I, B, C]:- getmemo(a(I)-J, J),
%							 C is B + J)), Ns, 0, Sum))))).
%@ % 938,719 inferences, 0.074 CPU in 0.075 seconds (99% CPU, 12701698 Lips)
%@ Ns = [1, 2, 3, 4, 5, 6, 7, 8, 9|...],
%@ Sum = 100010000.
