:- module(path_count6, [zdd0/1, co_compare/3]).

:- use_module(library(apply)).
:- use_module(library(apply_macros)).
:- use_module(library(clpfd)).
:- use_module(library(statistics)).
:- use_module(zdd('zdd-array')).
:- use_module(util(math)).
:- use_module(util('pac-meta')).
:- use_module(pac(basic)).
:- use_module(pac(meta)).
:- use_module(util(misc)).
:- use_module(pac('expand-pac')). % For the kind block.
:- use_module(zdd('zdd-misc')).
:- use_module(zdd(zdd)).
:- use_module(zdd('mqp-update2')).
:- use_module(pac(op)).


:- set_prolog_flag(stack_limit, 10_200_147_483_648).

 :- op(1060, xfy, ~).		% equivalence
 :- op(1060, xfy, #).		% exor
 :- op(1060, xfy, <->).		% equivalence
 :- op(1050, yfx, <-).
 :- op(1060, xfy, <=> ).	% equivalence
 :- op(1040, xfy, \/).		% OR
 :- op(1030, xfy, /\).		% AND
 :- op(1020, fy, \).		% NOT
 :- op(700, xfx, :=).		% Assignment
 :- op(1000, xfy, &).

% for pac query.
 :- pac:op(1000, xfy, &).
 :- pac:op(700, xfx, :=).

term_expansion --> pac:expand_pac.

% ?- zdd0((true)).
% ?- zdd0((X<<pow([a,b]),  psa(X))).

:- meta_predicate zdd0(:).
zdd0(X):- zdd((set_compare(co_compare),
   				   set_pred(admissibility, admissible_qp_link),
				   X)).

% ?- admissible_qp(u-v, a-b).  % false.
% ?- admissible_qp(a-b, l-m).

admissible_qp(X, Y):- X = (_-_),
	(inadmissible_qp(X, Y) -> false ; true).


% ?- admissible_qp_link(p(2,0)-p(3,0),  p(3,3)-p(3,4)).
% ?- admissible_qp_link(p(1,0)-p(2,0),  p(3,3)-p(3,4)).  % false

admissible_qp_link(p(I, _)-p(J, _), p(A, B)-p(C, B)):- !, U is min(A, C),
	I >= U,
	J >= U.
admissible_qp_link(p(I, _)-p(J, _), p(A, _)-p(A, _)):- U is A - 1,
	I >= U,
	J >= U.

% ?- inadmissible_qp(l-m, a-x).
% ?- inadmissible_qp(3-5, 3-3).	% false.
inadmissible_qp(A-B, P-Q):-min(A, B, M),
	(P @< M ; Q @< M), !.

% ?- test_mqp_update.
test_mqp_update:-
	forall(	mqp_query(Mqp, Links),
			zdd0((X<<Mqp,
				 nl,
				 format("Links = ~w, mqp = ~w\n", [Links, Mqp]),
				 mqp_update_links(Links, X, Y),
				 sets(Y, Y0),
				 maplist(writeln, Y0)))).
%
mqp_query({[]},  [a-b]).
mqp_query({[a-a, b-b]}, [a-b]).
mqp_query({[a-a, b-b]}, [b-a]).
mqp_query({[a-a, b-b], [a-a, c-c]}, [a-b]).
mqp_query({[a-a, b-b], [a-a, b-b, c-c]}, [a-b]).
mqp_query({[a-a, b-b], [a-a, b-b, c-c]}, [c-d]).
mqp_query({[a-a, b-c], [a-a, b-b, c-c]}, [c-b]).
mqp_query({[a-a, b-c], [a-a, b-b, c-c]}, [a-b, c-b]).
mqp_query({[a-a, b-c], [a-a, b-b, c-c]}, [a-b, c-a]).
mqp_query({[a-a, b-b]}, [a-b]).
mqp_query({[a-a, b-b]}, [b-a]).
mqp_query({[a-a, b-b, c-c]},[a-b, b-c]).
mqp_query({[a-a, b-b, c-c]}, [b-a, c-b]).

% ?- co_compare(C, c-d, c->d).
% ?- co_compare(C, c->d, c->d).
% ?- co_compare(C, c->d, c-d).
% ?- co_compare(C, c-d, c-d).
% ?- predsort(co_compare, [a->b, b->a], R).
% ?- predsort(co_compare, [a-b, b-a], R).
% ?- predsort(co_compare, [b->a, a->b, a-b, a->b, c-b, a-b, b-a], R).
% ?- zdd((set_compare(co_compare), zdd_compare(C, a-b, c-d))).
% ?- zdd((set_compare(co_compare), zdd_compare(C, c-d, c->d))).
% ?- zdd((set_compare(co_compare), zdd_compare(C, c->d, c-d))).
% ?- zdd((set_compare(co_compare), zdd_insert_atoms([a-b, c-d], 1, X))).
% ?- zdd((set_compare(co_compare), zdd_insert_atoms([a->b, c-d], 1, X), psa(X))).

co_compare(C, X, Y):- compare(C0, X, Y),
	(	C0 = (=)  -> C = (=)
	;	functor(X, Fx, _),
		functor(Y, Fy, _),
		(	Fx == Fy
		-> 	(	Fx = (-)
			->  ( C0 = (<) -> C = (>) ; C = (<) )
			;	C = C0
			)
		;	(	Fx = (-) -> C = (<) ; C = (>) )
		)
	).

% ?- predsort(co_compare, [a->b, b->a, a-b, a->b, c-b, b-a, a-b ], R).
% ?- predsort(co_compare, [a->b, b->a], R).

% some tiny.
% x(p(X, _), X).
% y(p(_, Y), Y).
link_symbol(_->_).
change_symbol(A-B, A->B).
%
touch(A-_, _-A):-!.
touch(_-A, A-_).
% ?- connect_pairs(a-b, [b-c, d-a], W).
% ?- connect_pairs(a-b, [a-c, a-b], W).  % false

connect_pairs(X-Y, [U-X, Y-V], U-V):-!.
connect_pairs(X-Y, [Y-V, U-X], U-V).

% ?- zdd0((X<<{[p(0,0)-p(2,1)], [p(1,1)-p(2,3), p(2,2)-p(2,2)]},
%	E = p(2,1)-p(3,1),
%	prune_frontier(X, E, p(0,0), Y), psa(Y))).

% ?- zdd0((X<<{[p(0,0)-p(2,1)], [p(1,1)-p(1,1), p(2,2)-p(2,2)]},
%	E = p(2,1)-p(3,1),
%	prune_frontier(X, E, p(0,0), Y), psa(Y))).

% ?- zdd0((X<<{[p(0,0)-p(2,1)], [p(1,1)-p(1,1), p(2,2)-p(2,3)]},
%	E = p(2,1)-p(3,1),
%	prune_frontier(X, 2, p(0,0), Y), psa(Y))).

prune_frontier(F, E, P, F0):- @(prune_frontier(F, E, P, F0)).
%
prune_frontier(F, _, _, F, _):- F<2, !.
prune_frontier(F, E, P, F0, S):- s_cofact(F, t(A, L, R), S),
	prune_frontier(L, E, P, L0, S),
	(	(	A = (X-X)
		;	A = (P-_)
		; 	admissible_qp(E, A)
		)
	->	prune_frontier(R, E, P, R0, S)
	;	R0 = 0
	),
	s_cofact(F0, t(A, L0, R0), S).

%
prune_frontier_by_level(F, K, P, F0):- @(prune_frontier_by_level(F, K, P, F0)).
%
prune_frontier_by_level(F, _, _, F, _):- F<2, !.
prune_frontier_by_level(F, K, P, F0, S):- s_cofact(F, t(A, L, R), S),
	prune_frontier_by_level(L, K, P, L0, S),
	(	(	A = (X-X)
		;	A = P - p(K,_)
		; 	A = p(K, _)-p(K, _)
		)
	->	prune_frontier_by_level(R, K, P, R0, S)
	;	R0 = 0
	),
	s_cofact(F0, t(A, L0, R0), S).




		/****************
		*     bridge    *
		****************/

% ?- zdd0((count_path(p(0,0), p(0, 1), C))).
% ?- zdd0((count_path(p(0,0), p(1, 0), C))).
% ?- zdd0((count_path(p(0,0), p(1,1), C))).
% ?- zdd0((count_path(p(0,0), p(2,2), C))).
% ?- zdd0((count_path(p(0,0), p(3,3), C))).
% ?- zdd0((count_path(p(1,1), p(2,1), C))).
% ?- zdd0((count_path(p(1,2), p(3,2), C))).
% ?- zdd0((count_path(p(1,2),p(10,2), C))).
% ?- zdd0((count_path(p(1,2),p(12,2), C))).
% ?- zdd0((count_path(p(1,2),p(13,2), C))).
% ?- zdd0((count_path(p(1,2),p(14,2), C))).
% ?- zdd0((count_path(p(1,2),p(100,2), C))).
% ?- time(zdd0((count_path_line(p(1,2),p(1000, 2), C)))).
%@ % 4,691,349 inferences, 0.982 CPU in 1.154 seconds
% ?- time(zdd0((count_path_line(p(1,2),p(10000,2), C)))).
%@ % 61,745,569 inferences, 59.803 CPU in 60.290 seconds

% ?- zdd0((count_path(p(0,0), p(1,1), Z, U),  psa(U))).

count_path(X, Y, Z, U):- @(count_path(X, Y, Z, U)).
%
count_path(p(I, J), p(_K, L), Bridge, Col, S):-
	initial_bridge(I, J, L, Bridge),
	initial_column(I, J, L, Col, S).


% ?- bridge_shift(h, [p(1,1)-p(1,1)], X).
% ?- bridge_shift(v, [p(1,1)-p(1,1)], X).
bridge_shift(F, X, Y):- maplist(shift_link(F), X, Y).

% ?- zdd((convert_to_col([[1-2], [3-4]], Col, 0),
%	map_h_shift(Col, Col0))).
map_h_shift(X, Y):- @(map_h_shift(X, Y)).
%
map_h_shift([], [], _).
map_h_shift([X|Y], [X0|Y0], S):- h_shift(X, X0, S),
	map_h_shift(Y, Y0, S).

% ?- zdd((convert_to_row([[1-2], [3-4]], Row, 0),
%	map_v_shift(Row, Row0))).
map_v_shift(X, Y):- @(map_v_shift(X, Y)).
%
map_v_shift([], [], _).
map_v_shift([X|Y], [X0|Y0], S):- v_shift(X, X0, S),
	map_v_shift(Y, Y0, S).

%
h_shift(X, Y):- @(h_shift(X, Y)).
%
h_shift(X, X, _):- X < 2, !.
h_shift(X, Y, S):- cofact(X, t(p(I, J)-p(I, K), L, R), S),
	I0 is I + 1,
	h_shift(L, L0, S),
	h_shift(R, R0, S),
	cofact(Y, t(p(I0, J)-p(I0, K), L0, R0), S).

% ?- zdd((qp_along_line(1, 2, X), convert_to_row(X, Y, 3),
%	maplist(v_shift, Y, Y0),
%	forall(member(M, Y0), (qp_list(M, L), writeln(L))))).

v_shift(X, Y):- @(v_shift(X, Y)).
%
v_shift(X, X, _):- X < 2, !.
v_shift(X, Y, S):- cofact(X, t(p(I, J)-p(K, J), L, R), S),
	v_shift(L, L0, S),
	v_shift(R, R0, S),
	J0 is J + 1,
	cofact(Y, t(p(I, J0)-p(K, J0), L0, R0), S).

% ?- spy(solve_rect).
% ?- zdd((solve_rect(p(0,0),p(1,0), Z))).
% ?- zdd((solve_rect(p(0,0),p(1,1), Z))).
% ?- zdd((solve_rect(p(0,0),p(2,2), Z))).
% ?- zdd((solve_rect(p(0,0),p(3,3), Z))).
% ?- call_with_time_limit(300, zdd((solve_rect(p(0,0),p(4,4), Z)))).

% ?- zdd0((solve_rect(p(0,0),p(1,1), Z, C))).
% ?- zdd0((solve_rect(p(0,0),p(1,0), Z, C))).
% ?- zdd0((solve_rect(p(0,0),p(2,0), Z, C))).
% ?- zdd0((solve_rect(p(0,0),p(32,0), Z, C))).
% ?- zdd0((solve_rect(p(0,0),p(1,1), Z, C), psa(Z))).
% ?- zdd0((solve_rect(p(0,0),p(3,3), Z, C))).
% ?- zdd0((solve_rect(p(0,0),p(3,4), Z, C))).
% ?- zdd0((solve_rect(p(0,0),p(3,1), Z, C))).
% ?- zdd0((solve_rect(p(0,0),p(20,1), Z, C))).
% ?- zdd0((solve_rect(p(0,0),p(4,4), Z, C))).
% ?- zdd0((solve_rect(p(0,0),p(5,5), Z, C))).
% ?- zdd0((solve_rect(p(0,0),p(6,6), Z, C))).
% ?- time(zdd0((solve_rect(p(0,0),p(7,7), Z, C)))).
% ?- zdd0((solve_rect(p(0,0),p(8,8), Z, C))).
% ?- zdd0((solve_rect(p(0,0),p(9,9), Z, C))).
% ?- zdd0((solve_rect(p(0,0),p(10,10), Z, C))).
% ?- zdd0((solve_rect(p(0,0),p(13,13), Z, C))).
% ?- zdd0((solve_rect(p(0,0),p(15,15), Z, C))).
% ?- zdd0((solve_rect(p(0,0),p(20,20), Z, C))).
% ?- zdd0((solve_rect(p(0,0),p(30,30), Z, C))).

% ?- zdd0((solve_rect(p(0,0),p(2,2), Z, C))).
% ?- zdd0((solve_rect(p(0,0),p(3,3), Z, C))).
% ?- zdd0((solve_rect(p(0,0),p(4,4), Z, C))).
% ?- zdd0((solve_rect(p(0,0),p(5,5), Z, C))).
% ?- zdd0((solve_rect(p(0,0),p(6,6), Z, C))).
% ?- zdd0((solve_rect(p(0,0),p(7,7), Z, C))).
% ?- zdd0((solve_rect(p(0,0),p(8,8), Z, C))).

solve_rect(X, Y, Z, C):- @(solve_rect(X, Y, Z, C)).
%
solve_rect(p(I, J), p(I, L), Z, 1, S):-!, initial_column(I, J, L, Z, S).
solve_rect(p(I, J), p(K, L), Z, C, S):-
	initial_col_bridge(I, J, L, Col, B, S),
	N is (K-I),
	repeat_bridge(N, Col, I, p(I,J), Col, B, Z, S),
	card(Z, C, S).

s_psa(X):- @(s_psa(X)).

s_psa(X, @(S,_,_)):-!, psa(X, S).
s_psa(X, S):- psa(X, S).
%
repeat_bridge(0, X, _, _, _, _, X, _):-!, writeln(0).
repeat_bridge(N, X, K, P, C, B, Y, S):- writeln(N),
	prune_frontier_by_level(X, K, P, X0, S),
	mqp_shift(h, C, C0, S),
	s_zdd_merge(X0, C0, X1, S), % s_psa(X0, S), s_psa(X1, S),
	s_mqp_update_links(B, X1, Y0, S), % nl, writeln(B), s_psa(Y0, S),
	maplist(shift_link(h), B, B0),
	K0 is K + 1,
	N0 is N - 1,
	repeat_bridge(N0, Y0, K0, P, C0, B0, Y, S).

% ?- zdd0((initial_col_bridge(0, 0, 3, C, B))).
% ?- time(zdd0((initial_col_bridge(0, 0, 15, C, B)))).
initial_col_bridge(I, L, H, C, B):-	@(initial_col_bridge(I, L, H, C, B)).
%
initial_col_bridge(I, L, H, C, B, S):- initial_column(I, L, H, C, S),
	initial_bridge(I, L, H, B).


% ?- zdd0((initial_column(1, 2, 4, Col), psa(Col))).
% ?- zdd0((initial_column(1, 0, 10, Col), card(Col, C))).
% ?- zdd0((initial_column(1, 0, 3, Col), card(Col, C))).  % seems good.
% ?- zdd0((initial_column(1, 0, 3, Col),
%	mqp_shift(h, Col, Col1),
%	mqp_shift(v, Col1, Col2),
%	card(Col, C),
%	card(Col1, C1),
%	card(Col2, C2))).

initial_column(I, Low, Hi, Col):- @(initial_column(I, Low, Hi, Col)).
%
initial_column(I, Low, Hi, Col, S):-
	mqp_linear_grid(Low, Hi, X, S),
	mqp_lift(x(I), X, Col, S).

% ?- initial_bridge(0, 1, 4, Bridge), maplist(writeln, Bridge).
initial_bridge(I, Low, Hi, Bridge):- J is I + 1,
	findall(A,
			(	member(A, [p(I, V)-p(J, V), p(J, V)-p(I, V)]),
				between(Low, Hi, V)
			),
			Bridge0),
	sort(Bridge0, Bridge).

		/****************
		*     prune     *
		****************/

% ?- zdd((qp_along_line(1, 13, X), length(X, L), convert_to_col(X, Y, 3),
%	prune_by_column(Y, 3, p(3,1)-p(3,1), Y0), length(Y, Ly))).

prune_by_column(F, X, P, F0):- @(prune_by_column(F, X, P, F0)).
%
prune_by_column([], _, _, [], _).
prune_by_column([Q|F], X, P, [Q|F0], S):- admissible_qp(Q, X, P, S), !,
	prune_by_column(F, X, P, F0, S).
prune_by_column([_|F], X, P, F0, S):-
	prune_by_column(F, X, P, F0, S).
%
admissible_qp(Q, X, P):- @(admissible_qp(Q, X, P)).
%
admissible_qp(Q, X, P, S):- qp_list(Q, List, S),
	admissible_qp_list(List, X, P).

% ?- admissible_qp_list([p(0,0)-p(2,1), p(3,1)-p(3,1)], 2, p(0,0)).
admissible_qp_list([], _, _):-!.
admissible_qp_list([P-p(X, _)|List], X, P):-!,
	admissible_qp_list(List, X, P).
admissible_qp_list([p(X, _)-p(X, _)|List], X, P):- !,
	admissible_qp_list(List, X, P).
admissible_qp_list([p(Z, _)-_|_], X, _):- Z>X, !.
admissible_qp_list([P-P|List], X, P):- admissible_qp_list(List, X, P).

% ?- rect((qp_along_line(1,1, Ls), maplist_qp(Ls, Q))).
% ?- rect((qp_along_line(1,10, Ls), maplist_qp(Ls, Q))).
% ?- time(rect((qp_along_line(1, 15, Ls), maplist_qp(Ls, Q)))).
%@ % 99,784,517 inferences, 8.136 CPU in 8.371 seconds (97% CPU, 12264944 Lips)


% ?- spy(mqp_linear_grid).
% ?- I = 0, zdd0((mqp_linear_grid(0,I,X), psa(X))).
% ?- I = 1, zdd0((mqp_linear_grid(0,I,X), psa(X))).
% ?- I = 2, zdd0((mqp_linear_grid(0,I,X), psa(X))).
% ?- I = 3, zdd0((mqp_linear_grid(0,I,X), psa(X))).
% Checked by hand.

% ?- forall(between(0, 20, I), zdd0((mqp_linear_grid(0,I,X), card(X, C), format("count(~w) = ~w\n", [I, C])))).

% simple line quasi path
mqp_linear_grid(X, Y, Z):- @(mqp_linear_grid(X, Y, Z)).

mqp_linear_grid(E, X, Y, S):-
	open_state(M, [hash_size(128)]),
	get_key(admissibility, Pred, S),
	s_mqp_linear_grid(E, X, Y, @(S, M, Pred)),
	close_state(M).
%
s_mqp_linear_grid(I, I, P, S):-!, s_zdd_singleton(I-I, P, S).
s_mqp_linear_grid(I, J, Q, S):-
	M is (I+J)//2,
	M0 is M+1,
	s_mqp_linear_grid(I, M, R, S),
	s_mqp_linear_grid(M0, J, R0, S),
	s_zdd_merge(R, R0, Q0, S),
	s_mqp_update_links([M-M0, M0-M], Q0, Q, S).


% ?- I = 3, zdd0((mqp_linear_grid(0,I,X), mqp_lift(y(0), X, Y), card(Y, C))).
%
mqp_lift(F, X, Y):- @(mqp_lift(F, X, Y)).
%
mqp_lift(F, X, Y, S):-open_state(M, [hash_size(128)]),
	mqp_lift(F, X, Y, S, M),
	close_state(M).
%
mqp_lift(_, X, X, _, _):- X<2, !.
mqp_lift(F, X, Y, S, M):- memo(mqp_lift(X)-Y, M),  % F dropped.
	(	nonvar(Y)-> true
	;	s_cofact(X, t(A, L, R), S),
		mqp_lift(F, L, L0, S, M),
		mqp_lift(F, R, R0, S, M),
		lift_link(F, A, B),
		s_cofact(Y, t(B, L0, R0), S)
	).

% ?- I = 3, zdd0((mqp_linear_grid(0,I,X), mqp_lift(y(0), X, Y), card(Y, C))).
% ?- I = 3, zdd0((mqp_linear_grid(0,I,X), mqp_lift(y(0), X, Y), card(Y, C),
%	mqp_shift(v, Y, Z), psa(Z), card(Z, D))).
% ?- I = 3, zdd0((mqp_linear_grid(0,I,X), mqp_lift(y(0), X, Y),
%	mqp_shift(v, Y, Z), mqp_shift(v, Z, U), psa(U), card(U, C))).

%
mqp_shift(F, X, Y):- @(mqp_shift(F, X, Y)).
%
mqp_shift(F, X, Y, S):-
	open_state(M, [hash_size(128)]),
	mqp_shift(F, X, Y, S, M),
	close_state(M).

%
mqp_shift(_, X, X, _, _):- X<2, !.
mqp_shift(F, X, Y, S, M):- memo(mqp_shift(X)-Y, M),  % F dropped.
	(	nonvar(Y)-> true
	;	s_cofact(X, t(A, L, R), S),
		mqp_shift(F, L, L0, S, M),
		mqp_shift(F, R, R0, S, M),
		shift_link(F, A, B),
		s_cofact(Y, t(B, L0, R0), S)
	).


% ?- lift_point(3, x(2), R).
% ?- lift_point(3, y(2), R).
%
lift_point(J, x(I), p(I, J)):-!.
lift_point(J, y(I), p(J, I)).

% ?- lift_link(x(2), 4-5, R).
lift_link(Ctr, A-B, P-Q):-!, lift_point(A, Ctr, P),
	lift_point(B, Ctr, Q).
% ?- lift_link(x(2), 4->5, R).
lift_link(Ctr, A->B, P->Q):- lift_point(A, Ctr, P),
	lift_point(B, Ctr, Q).
%
shift_link(h, A, B):-!, h_shift_link(A, B).
shift_link(v, A, B):- v_shift_link(A, B).

% ?- h_shift_point(p(1,1), R).
% ?- h_shift_point(-1, p(1,1), R).
h_shift_point(p(I, J), p(I0, J)):- I0 is I + 1.
h_shift_point(K, p(I, J), p(I0, J)):- I0 is I + K.
%
v_shift_point(p(I, J), p(I, J0)):- J0 is J + 1.
v_shift_point(K, p(I, J), p(I, J0)):- J0 is J + K.

%
h_shift_link(P-Q, P0-Q0):-!, h_shift_point(P, P0),
	h_shift_point(Q, Q0).
h_shift_link(P->Q, P0->Q0):- h_shift_point(P, P0),
	h_shift_point(Q, Q0).
%
v_shift_link(P-Q, P0-Q0):-!, v_shift_point(P, P0),
	v_shift_point(Q, Q0).
v_shift_link(P->Q, P0->Q0):- v_shift_point(P, P0),
	v_shift_point(Q, Q0).

% ?- simple_bridge(1-2, [1-1], [2-2], R).
% ?- simple_bridge(2-1, [1-1], [2-2], R).
% ?- simple_bridge(2-3, [1-1], [2-2], R).
simple_bridge(A-B, L, M, N):-	% assuming qp  X < qp Y
	(	select(B-V, M, M0),
		select(U-A, L, L0)
	;	select(B-V, L, L0),
		select(U-A, M, M0)
	),
	!,
	(	U @> V
	-> 	ord_union(L0, [U-V|M0], N)
	; 	ord_union([U-V], M0, M1),
		ord_union(L0, M1, N)
	).
simple_bridge(_, _, _, []).

%
simple_bridge_links([], _, _, Z, Z).
simple_bridge_links([E|Es], X, Y, Z, U):-
	simple_bridge_link(X, E, Y, Z, Z0),
	simple_bridge_links(Es, X, Y, Z0, U).
%
simple_bridge_link([], _, _, Z, Z).
simple_bridge_link([X|Xs], E, Y, Z, U):-
	simple_bridge_basic(Y, X, E, Z, Z0),
	simple_bridge_link(Xs, E, Y, Z0, U).
%
simple_bridge_basic([], _, _, Z, Z).
simple_bridge_basic([Y|Ys], X, E, Z, U):- simple_bridge(E, X, Y, V),
	(	V = []
	->	Z0 = Z
	;	Z  = [V|Z0]
	),
	simple_bridge_basic(Ys, X, E, Z0, U).

% ?- Q = [a-a], Q0=[b-b], Q1 = [c-c],
%	simple_prod_union([Q, Q0], [Q1], Z).

%
simple_prod_union(X, Y, Z):- simple_prod_union(X, Y, Z, []).
%
simple_prod_union([], _, U, U).
simple_prod_union([L|Y], Z, U, V):-
	simple_union(Z, L, U, U0),
	simple_prod_union(Y, Z, U0, V).
%
simple_union([], _, U, U).
simple_union([M|As], L, [N|U], V):-
	ord_union(L, M, N),
	simple_union(As, L, U, V).

%% test basics.
% ?- zdd((
%	qp_list(X, [a-a]), qp_list(Y, [b-b]),
%	qp_joint([a-b, b-a], [X], [Y], Z),
%	maplist(pred(([U]:-qp_list(U, List), writeln(List))), Z))).
% ?- zdd((
%	qp_list(X, [a-a]), qp_list(Y, [b-b]),
%	qp_joint([a-b, b-a], [X], [Y], Z),
%	zdd_join(X, Y, A),
%	memo(qp_suc(A)-L), writeln(qp_suc(A)-L),
%	maplist(pred(([U]:-qp_list(U, List), writeln(List))), Z))).
% ?- zdd((
%	qp_list(X, [a-a]), qp_list(Y, [b-b]),
%	qp_joint([a-b, b-a], [X], [Y], Z),
%	zdd_join(X, Y, A),
%	memo(qp_suc(A)-L), writeln(qp_suc(A)-L),
%	maplist(pred(([U]:-qp_list(U, List), writeln(List))), Z))).

qp_joint(X, Y, Z, U):-@(qp_joint(X, Y, Z, U)).
%
qp_joint(X, Y, Z, U, S):- map_prod_sum(Y, Z, V, [], S),
	qp_bridge_links(X, Y, Z, U, V, S).

% ?- zdd((qp_list(X, [a-a]), qp_list(Y, [b-b]),
%	qp_bridge(a-b, X, Y, Z), qp_list(Z, U))).
% ?- zdd((qp_list(X, [a-a]), qp_list(Y, [b-b]),
%	qp_bridge(b-a, X, Y, Z), qp_list(Z, U))).

qp_bridge(E, X, Y, Z):- @(qp_bridge(E, X, Y, Z)).
%
qp_bridge(A-B, X, Y, Z, S):- % assuming qp  X < qp Y
	qp_list(X, L, S),
	qp_list(Y, M, S),
	(	select(U-A, L, L0),
		select(B-V, M, M0)
	;	select(B-V, L, L0),
		select(U-A, M, M0)
	),
	!,
	ord_union(L0, M0, N0),
	ord_union([U-V], N0, N),
	qp_list(Z, N, S).
qp_bridge(_, _, _, 0, _).

%
qp_bridge_links(X, Y, Z, U, V):- @(qp_bridge_links(X, Y, Z, U, V)).
%
qp_bridge_links([], _, _, Z, Z, _).
qp_bridge_links([E|Es], X, Y, Z, U, S):-
	qp_bridge_link(X, E, Y, Z, Z0, S),
	qp_bridge_links(Es, X, Y, Z0, U, S).
%
qp_bridge_link([], _, _, Z, Z, _).
qp_bridge_link([X|Xs], E, Y, Z, U, S):-
	qp_bridge_basic(Y, X, E, Z, Z0, S),
	qp_bridge_link(Xs, E, Y, Z0, U, S).
%
qp_bridge_basic([], _, _, Z, Z, _).
qp_bridge_basic([Y|Ys], X, E, Z, U, S):- qp_bridge(E, X, Y, V, S),
	(	V = 0
	->	Z0 = Z
	;	Z  = [V|Z0],
		select_goto(E, X, Y, V, S)
	),
	qp_bridge_basic(Ys, X, E, Z0, U, S).
%
select_goto(E, X, Y, Z):- @(select_goto(E, X, Y, Z)).
%
select_goto(A-B, X, Y, Z, S):-
	G = (A-B)-Z,
	( A@<B -> W = X ; W = Y ),
	getmemo(qp_suc(W)-U, V, S),
	(	var(V) -> V = [] ; true ),
	add_new(G, V, U).

		/***********************
		*     end of bridge    *
		***********************/

% ?-  belong(3, 3, 4).
% ?-  belong(3, 4, 1).
belong(X, I, J):- I @=< J, !, I @=< X, X @=< J.
belong(X, I, J):- J @=< X, X @=< I.
%
sub_interval(I, J, X, Y):- belong(I, X, Y), belong(J, X, Y).
%
disjoint_interval(I, J, X, Y):- \+ belong(I, X, Y), \+ belong(J, X, Y).

% ?- compatible_interval(1, 2, 3, 4).
% ?- compatible_interval(1, 3, 2, 4). % false
% ?- compatible_interval(2, 4, 5, 7).
% ?- compatible_interval(7, 5, 2, 4).

compatible_interval(I, J, X, Y):- disjoint_interval(I, J, X, Y), !.
compatible_interval(I, J, X, Y):- sub_interval(I, J, X, Y), !.
compatible_interval(I, J, X, Y):- sub_interval(X, Y, I, J).


% go
% ?- spy(s_mqp_update_links).
% ?- rect_dg(rect(1,1), X, Y), findall(A-A, member(A,  X), Init), reverse(Y, Y0),
%	zdd0(( zdd_insert_atoms(Init, 1, Init0), psa(Init0), s_mqp_update_links(Y0, Init0, R), psa(R))).
% ?- rect_dg(rect(1,1), X, Y), findall(A-A, member(A,  X), Init), reverse(Y, Y0),
%	zdd0(( zdd_insert_atoms(Init, 1, Init0), psa(Init0), s_mqp_update_links(Y0, Init0, R), psa(R))).

% ?- rect_dg(rect(2,2), X, Y), findall(A-A, member(A,  X), Init),
%	zdd0(( zdd_insert_atoms(Init, 1, Init0), psa(Init0), s_mqp_update_links(Y, Init0, R))).
% ?- rect_dg(rect(3,3), X, Y), findall(A-A, member(A,  X), Init),
%	zdd0(( zdd_insert_atoms(Init, 1, Init0), s_mqp_update_links(Y, Init0, R))).


% ?- rect_dg(rect(0,1), X, Y).
% ?- rect_dg(rect(0,0), X, Y).

rect_dg(rect(W, H), Nodes, Links):-
	rect_nodes(W, H, Nodes),
	rect_links(W, H, Nodes, Links).
% ?- rect_nodes(4,5,Ns), length(Ns, L).

rect_nodes(W, H, Nodes):-
	findall(p(I, J),
			(	between(0, W, I),
				between(0, H, J)
			),
			Nodes0),
	sort(Nodes0, Nodes).

%
rect_links(W, H, Ns, Links):-
	findall(P-Q,
			(	member(P, Ns),
				P = p(I, J),
				(	Q = p(I, J1),
					(	J1 is J-1,
						J1 >= 0
					;	J1 is J+1,
						J1 =< H
					)
				;   Q = p(I1, J),
					(	I1 is I-1,
						I1 >= 0
					;	I1 is I+1,
						I1 =< W
					)
				)
			),
			Links0),
	sort(Links0, Links).

		/********************
		*   frontier based  *
		********************/

% ?- zdd(path_count(rect(0,0), C)).
% ?- zdd(path_count(rect(1,0), C)).
% ?- zdd(path_count(rect(0,1), C)).
% ?- zdd(path_count(rect(1,1), C)).
% ?- zdd(path_count(rect(2, 1), C)).
% ?- zdd(path_count(rect(3, 1), C)).
% ?- zdd(path_count(rect(1, 3), C)).
% ?- zdd(path_count(rect(4, 1), C)).
% ?- zdd(path_count(rect(2, 3), C)).
% ?- zdd(path_count(rect(4,3), C)).
%@ C = 976 .
% ?- zdd(path_count(rect(4,4), C)).
%@ C= 8512
% ?- call_with_time_limit(360, zdd(path_count(rect(5, 5), C))).

path_count(Rect, C):- @(path_count(Rect, C)).

%
path_count(rect(0,0), 1, _):-!.
path_count(Rect, C, S):-
	rect_dg(Rect, Ns, Links),
	findall(A-A, member(A, Ns), Ids),
	Rect = rect(W, H),
	set_key(final, p(W, H), S),
	qp_list(Q, Ids, S),
	memo(qp_suc(Q)-[], S),
	frontier(Links, [Q], FinalQs, S),
	fold_qp_memo_final(FinalQs, S),
	path_tree(Q, Tree, S),
	card(Tree, C, S).

%
frontier([], Qs, Qs, _).
frontier([E|Es], Qs, Qs0, S):-
	frontier_link(E, Qs, Qs1, Qs, S),
	prune_frontier(Qs1, E, Qs2, S),
	frontier(Es, Qs2, Qs0, S).

%
frontier_link(_, [], Qs, Qs, _):-!.
frontier_link(E, [X|Qs], Qs0, Qs1, S):- X<2, !,
 	frontier_link(E, Qs, Qs0, Qs1, S).
frontier_link(E, [Q|Qs], Next, Qs1, S):-
	qp_update(E, Q, Q0, S),
	(	Q0 < 2
	->	Next = Qs0
	;	qp_memo_final(Q0, S),
		Next = [Q0|Qs0]
	),
	qp_update_succs(E-Q0, Q, S),
	frontier_link(E, Qs, Qs0, Qs1, S).

% ?- zdd((qp_list(Q, [a]), qp_list(Q0, [b]), qp_list(Q1, [c]),
%	map_prod_sum([Q, Q0], [Q1], Z),
%	maplist(pred([A, M]:-qp_list(A, M)), Z, W))).
%
map_prod_sum(X, Y, Z):- map_prod_sum(X, Y, Z, []).
%
map_prod_sum(X, Y, Z, U):- @(map_prod_sum(X, Y, Z, U)).

map_prod_sum([], _, U, U, _).
map_prod_sum([Q|Y], Z, U, V, S):- qp_list(Q, L, S),
	map_qp_sum(Z, L, U, U0, S),
	map_prod_sum(Y, Z, U0, V, S).
%
map_qp_sum([], _, U, U, _).
map_qp_sum([Q|As], L, [Q0|U], V, S):-qp_list(Q, M, S),
	ord_union(L, M, N),
	qp_list(Q0, N, S),
	map_qp_sum(As, L, U, V, S).


% ?- zdd((path_count(rect(1,0), C))).
% ?- zdd((path_count(rect(0,1), C))).
% ?- zdd((path_count(rect(1,1), C))).
% ?- zdd((path_count(rect(3,1), C))).
% ?- time(zdd((path_count(rect(1,8), C)))).   % take time.
%@ % 86,018,490 inferences, 10.604 CPU in 11.323
% ?- time(zdd((path_count(rect(8,1), C)))).
%@ % 9,254,518 inferences, 3.089 CPU in 4.850
% ?- zdd((path_count(rect(2,3), C))).
% ?- zdd((path_count(rect(3,2), C))).
% ?- time(zdd((path_count(rect(3,3), C)))).
%@ % 50,584,360 inferences, 7.362 CPU in 9.480
%@ C = 184 .
% ?- time(zdd((path_count(rect(4,4), C)))).
%@ % 9,474,429,287 inferences, 1256.779 CPU in 1433.734 seconds (88% CPU, 7538660 Lips)
%@ C = 8512 .

qp_update(E, X, Y):- @(qp_update(E, X, Y)).
%
qp_update(_, X, 0, _):- X < 2.
qp_update(E, X, Y, S):- qp_list(X, U, S),
	memo(qp_update(E, X)-Y, S),
	(	nonvar(Y)  ->	true
	;	select_touch(E, [A, B], U, U0)
	->	connect_pairs(E, [A, B], W),
		ord_union([W], U0, V),
		qp_list(Y, V, S)
	;	Y = 0
	).
qp_update(_, _, 0,  _).

% ?- zdd((qp_list(X, [a-a, b-b]), memo(qp_suc(X)-[]))).
qp_update_succs(G, Q):- @(qp_update_succs(G, Q)).

% ?- zdd(qp_update_succs(a, 1)).
qp_update_succs(G, Q, S):- getmemo(qp_suc(Q)-U, V, S),
	G = E-Q0,
	(	var(V) -> U = [E-Q0]
	;	add_new(E-Q0, V, U)
	).

qp_admissible(L):-
	forall((select(A-B, L, L0), select(C-D, L0, _)),
		   compatible_interval(A, B, C, D)).

% ?- compatible_interval(1, 2, 3, 4).
% ?- compatible_interval(1, 3, 2, 4). % false
% ?- compatible_interval(2, 4, 5, 7).
% ?- compatible_interval(7, 5, 2, 4).


% compatible_interval(I, J, X, Y):- disjoint_interval(I, J, X, Y), !.
% compatible_interval(I, J, X, Y):- sub_interval(I, J, X, Y), !.
% compatible_interval(I, J, X, Y):- sub_interval(X, Y, I, J).
%@ true.
%@ path-count4.pl qcompiled.

% ?- select_touch(a-b, [A, B], [a-a, b-b], X).
% ?- select_touch(a-b, [A, B], [a-a, b-b], X),
%	connect_pairs(a-b, [A, B], Y).
%@ false.

% ?- select_touch(a-b, [A, B], [a-a, b-c], X),
%	connect_pairs(a-b, [A, B], Y).
% ?- select_touch(a-b, [A, B], [a-a, c-b], X).

select_touch(_, [], V, V).
select_touch(E, [A|As], [A|U], V):- touch(E, A), !,
	select_touch(E, As, U, V).
select_touch(E, As, [A|U], [A|V]):-select_touch(E, As, U, V).

% ?- add_new(a,[b,c,a], X).
add_new(X, Y, Y):- memberchk(X, Y), !.
add_new(X, Y, [X|Y]).

% ?- zdd((qp_list(Q, [p(0,0)-p(1,0)]), qp_list(Q, L))).
%@ L = [p(0, 0)-p(1, 0)].
%
qp_list(Q, List):- @(qp_list(Q, List)).
%
qp_list(0, 0, _):-!.
qp_list(1, [], _):-!.   %% ???
qp_list(Q, List, S):- var(Q), !,
	zdd_insert_atoms(List, 1, Q, S),
	memo(qp_suc(Q)-U, S),
	(	var(U) -> U = []
	;	true			% Already registered.
	).
qp_list(Q, [A|As], S):- cofact(Q, t(A, 0, R), S),
	qp_list(R, As, S).

%
qp_memo_final(X):- @(qp_memo_final(X)).
%
qp_memo_final(Q, S):-  memoq(qp_final(Q)-true, S), !.
qp_memo_final(Q, S):- qp_list(Q, List, S),
	get_key(final, F, S),
	(	qp_final(List, F)
	->	memo(qp_final(Q)-true, S)
	;	true
	).
%
fold_qp_memo_final([], _).
fold_qp_memo_final([Q|Qs], S):- qp_memo_final(Q, S),
	fold_qp_memo_final(Qs, S).

%
path_tree(Q, T):- @(path_tree(Q, T)).
%

path_tree(X, 0,  _):- X<1, !.
path_tree(Q, 1, S):- memoq(qp_final(Q)-true, S), !.
path_tree(Q, T, S):- memo(path_tree(Q)-T, S),
	(	nonvar(T) -> true
	;	memo(qp_suc(Q) - L, S),		% writeln(qp_suc(Q)-L),
		path_tree_list(L, 0, T, S)
	).
%
path_tree_list([], C, C, _).
path_tree_list([E-Q|Qs], C0, C, S):-
	path_tree(Q, A, S),
	zdd_insert(E, A, A0, S),
	zdd_join(C0, A0, C1, S),
	path_tree_list(Qs, C1, C, S).

% ?- qp_final([p(0,0)-p(2,3),p(0,3)-p(0,3),p(1,3)-p(1,3)], p(2,3)).

qp_final([X-Y|R], X, Y):-!, qp_all_id(R).
qp_final([A-A|R], X, Y):- qp_final(R, X, Y).

qp_final([p(0,0)-F|R], F):-!, qp_all_id(R).
qp_final([A-A|R], F):- qp_final(R, F).

%
qp_all_id([]).
qp_all_id([A-A|R]):- qp_all_id(R).

%
dag_card(Q, C):- @(dag_card(Q, C)).
%
dag_card(Q, C, S):- open_state(M),
	dag_card(Q, C, S, M),
	close_state(M).
%
dag_card(Q, 0,  _, _):- Q<2, !.
dag_card(Q, 1, S, _):- memoq(qp_final(Q)-true, S), !.
dag_card(Q, C, S, M):- memo(dag_card(Q)-C, M),
	(	nonvar(C) -> true
	;	memo(qp_suc(Q) - L, S),
		dag_card_list(L, 0, C, S, M)
	).
%
dag_card_list([], C, C, _, _).
dag_card_list([_-Q|Qs], C0, C, S, M):-
	dag_card(Q, A, S, M),
	C1 is C0 + A,
	dag_card_list(Qs, C1, C, S, M).


		/***************
		*     debug    *
		***************/
%
snap_qp_update(E, X, Y, S):- qp_update(E, X, Y, S),
	qp_list(X, X0, S),
	qp_list(Y, Y0, S),
	format("~w\n~w\n~w\n\n", [qp_update(E, X, Y), X0, Y0]).
%
dump_frontier(Qs, S):- maplist(pred(S,
	([Q]:- qp_list(Q, U, S), writeln(U))), Qs).
%
dump_qp(X, _):-  X<2, !, writeln(X).
dump_qp(X, S):-  qp_list(X, L, S),
	writeln(X=L).

%
delta_star([], Q, Q, _).
delta_star([E|Es], Q,  Q0, S):-  memo(qp_suc(Q)-L, S),
	writeln(E),
	memberchk(E-Q1, L),
	dump_qp(Q1, S),
	delta_star(Es, Q1, Q0, S).

%
mirror(X, Y):- @(mirror(X, Y)).
%
mirror(X, X, _):- X<2, !.
mirror(X, Y, S):- cofact(X, t(p(I,J)-p(K,L), Lx, Rx), S),
	mirror(Lx, MLx, S),
	mirror(Rx, MRx, S),
	zdd_insert(p(J, I)-p(L, K), MRx, MRx0, S),
	zdd_join(MLx, MRx0, Y, S).

%
rect_path_tree(Rect, T):-@(rect_path_tree(Rect, T)).

rect_path_tree(Rect, T, S):-
	sort_links_by_rank(Rect, Rs, Ids),
	Rect = rect(W, H),
	set_key(final, p(W, H), S),
	qp_list(Q, Ids, S),
	repeat_frontier(Rs, [Q], FinalQs, S),
	fold_qp_memo_final(FinalQs, S),
	path_tree(Q, T, S).

% ?- spy(remove_end_links).
% ?- remove_end_links([1-[c-a]], R0, a-d).
% ?- remove_end_links([1-[c-a, d-c, a-d]], R0, a-d).

remove_end_links(R, R0, ST):- maplist(pred(ST,
	(	[J-Es, J-Es0]:- foldl(
		pred([ST, Es, Es0],
			 ( [P-Q, U, V]:- ST = S-T,
					(	( Q==S; P==T)
					->  V = U
					;	U = [P-Q|V]
					))),  Es,  Es0, []))),  R, R0).

%
rest(Q, X, Y, S):- memo(qp_suc(Q)-L, S),
	(	select(Q, X, X0)-> true
	;	X0 = X
	),
	rest_list(L, X0, Y, S).
%
rest_list([], X, X, _).
rest_list([_-Q|Qs], X, Y, S):-
	rest(Q, X, X0, S),
	rest_list(Qs, X0, Y, S).


% ?- zdd((dbg_qp_update(p(0,0)-p(1,0), [p(0,0)-p(0,0),p(1,0)-p(1,0)], L))).
% ?- spy(dbg_qp_update).
% ?- zdd((dbg_qp_update(p(2,2)-p(2,3), [p(0,0)-p(2,2),p(0,3)-p(0,3), p(1,3)-p(1,3), p(2,3)-p(2,3)], L))).

dbg_qp_update(E, X, Y):- @(dbg_qp_update(E, X, Y)).
%
dbg_qp_update(E, X, Y, S):- qp_list(Q, X, S),
	qp_update(E, Q, Q0, S),
	qp_list(Q0, Y, S).


% ?- zdd((rect_path_tree(rect(3,2), T), card(T, C),
%   zdd_insert_atoms([p(0,0)-p(2,3),p(0,3)-p(0,3),p(1,3)-p(1,3)], 1, W),
%	mirror(T, MT), card(MT, CMT),
%	rect_path_tree(rect(2,3), T0), card(T0, C0), mirror(T0, MT0),
%	card(MT0, CM0),
%	zdd_subtr(T, MT0, D), card(D, CD), psa(D)
%)).


% Test to show getmemo/2,3 works well.
% ?- time((numlist(1, 10000, Ns),
%	zdd((maplist( pred([A]:- getmemo(a(A)-A, _)), Ns),
%		 maplist( pred(([A]:- getmemo(a(A)-B, Old), B is 2*Old)),  Ns),
%		 foldl( pred(([I, B, C]:- getmemo(a(I)-J, J),
%							 C is B + J)), Ns, 0, Sum))))).
%@ % 938,719 inferences, 0.074 CPU in 0.075 seconds (99% CPU, 12701698 Lips)
%@ Ns = [1, 2, 3, 4, 5, 6, 7, 8, 9|...],
%@ Sum = 100010000.
