:- module(zdd,
  [		(@)/1, (<<)/2, (<<)/3, (#)/1, get_sharp/1
	,	card/2, card/3
	,	ltr_cofact/2, ltr_cofact/3
	,	ltr_join/3,	ltr_join/4, ltr_zdd/1
  	,	ltr_merge/3, ltr_merge/4, ltr_card/3, ltr_compare/3
	,   card_filter_gte/4, card_filter_lte/4, card_filter_between/5
	,	max_length/3
  	,	zdd_sat/1, zdd_sat_count/1, zdd_sat_count/3
    ,	fos/1, fos_count/0, fos_count/1
	,	set_at/2, set_at/3
	,	dnf/2, dnf/3, cnf/2, cnf/3
	,	nnf_dnf/3, zdd_sat_count/3
	,	all_fun/3, all_fun/4, s_map/3, s_map/4
	,	all_mono/4, all_epi/4, merge_funs/3
	,	cofact0/3, cofact0/4
	,	bdd_list/3, bdd_append/4, bdd_interleave/4
	,	opp_compare/3
	,	zdd/1, zdd0/1, zdd_eval/3, zdd_eval/4
	, 	zdd_append/3,  zdd_append/4
	, 	zdd_insert/3, zdd_insert/4, zdd_insert/5, zdd_insert_atoms/3, zdd_cons/4
	,	zdd_insert_atoms/4
	,	zdd_ord_insert/4, zdd_reorder/4
  	,   zdd_insert_as_list/3, zdd_insert_as_list/4
	,	zdd_has_1/2
	,	zdd_insert_as_list/3, zdd_insert_as_list/4
  	,	zdd_memberchk/2, zdd_memberchk/3
    ,	zdd_family/2, zdd_family/3
    ,	zdd_subfamily/2, zdd_subfamily/3
	,	zdd_join/3, zdd_join/4, zdd_join_1/3, zdd_join_list/3, zdd_join_list/4
  	,	zdd_singleton/2, zdd_singleton/3
	,	zdd_merge/3, zdd_merge/4
	,	zdd_merge_list/2, zdd_merge_list/3, zdd_merge_list/4
	,	zdd_meet/3,  zdd_meet/4
	,	zdd_subtr/3, zdd_subtr/4, zdd_subtr_1/3
  	,	zdd_div/3,  zdd_div/4, zdd_mod/3, zdd_mod/4, zdd_divmod/5, zdd_div_div/5
	,	zdd_mod_by_list/4, zdd_div_by_list/4, zdd_divisible_by_list/4
	,	zdd_power/2, zdd_power/3
	,	zdd_rand_path/3
	,	ltr_meet_filter/4, ltr_join_filter/4
	,	get_extra/1, get_extra/2
	,	get_key/2, get_key/3, get_key/4, atom_index/3
	,	set_key/2, set_key/3, update_key/3, update_key/4, bump_up/3
	,   set_pred/2, set_pred/3, set_pred/2, set_pred/3
	,	zdd_compare/3, zdd_compare/4
	,	zdd_sort/2, zdd_sort/3
  	,	open_key/2, open_key/3, close_key/1, close_key/2
	,	set_compare/1, set_compare/2, get_compare/1, get_compare/2
	,	use_zdd/1, use_zdd/2, pred_zdd/2, pred_zdd/3
	,	zdd_sort_rev/2, zdd_sort_rev/3
	,   zdd_shift/1, zdd_shift/2, map_zdd/3, map_zdd/4
    ,	psa/1, psa/2, mp/2, mp/3
    ,	sets/2, sets/3, ppoly/1, ppoly/2, poly_term/2, poly_term/3
    ,	eqns/2, eqns/3
	,	zdd_list/2, zdd_list/3
	,	significant_length/3
	,	charlist/2, charlist/3, atomlist/2
	,	op(200, fy, #)
  ]).

% :- doc_server(7000).
% :- use_module(library(pldoc/doc_library)).
% :- doc_load_library.
% :- doc_browser.

:- use_module(library(apply)).
:- use_module(library(apply_macros)).
:- use_module(library(clpfd)).
:- use_module(library(statistics)).
:- use_module(zdd('zdd-array')).
:- use_module(util(math)).
:- use_module(util(meta2)).
:- use_module(pac(basic)).
:- use_module(pac(meta)).
:- use_module(util(misc)).
:- use_module(pac('expand-pac')). % For the kind block.
:- use_module(pac(op)).

:- set_prolog_flag(stack_limit, 10_200_147_483_648).

attr_unify_hook(V, Y):-
	(	get_attr(Y, zdd, A)
	->	zdd_unify(V, A)
	; 	zdd_unify(V, Y)
	).

 :- op(200, fy, #).
 :- op(1060, xfy, ~).		% equivalence
 :- op(1060, xfy, ~).		% equivalence
 :- op(1060, xfy, #).		% exor
 :- op(1060, xfy, <->).		% equivalence
 :- op(1050, yfx, <-).
 :- op(1060, xfy, <=> ).	% equivalence
 :- op(1040, xfy, \/).		% OR
 :- op(1030, xfy, /\).		% AND
 :- op(1020, fy, \).		% NOT
 :- op(700, xfx, :=).		% Assignment
 :- op(1000, xfy, &).

% for pac query.
 :- pac:op(1000, xfy, &).
 :- pac:op(700, xfx, :=).

term_expansion --> pac:expand_pac.

:- meta_predicate @(:).
@(X):- shift(X).

% ?- demo_cofact([[a,b], [c]], T).
% ?- zdd((X<<fos([[a,b], [a, c], [c, d], [c]]), psa(X))).
demo_cofact(F, t(A, L, R)):-
	zdd((X<<fos(F), cofact(X, t(A, L0, R0)),
	sets(L0, Ls),
	sets(R0, Rs))),
	maplist(maplist, Ls, L),
	maplist(maplist, Rs, R).

% ?- zdd((X<<{[a,b],[a,b,a,b]}, cofact(X, t(A, L, R)), cofact(Y, t(A, L, R))))  .
% ?- zdd((cofact(X, t(A,  R)), cofact(Y, t(A, L, R))))  .

% ?- zdd((X<<family([[a],[a,b]]), psa(X))).
% ?- zdd({X=1}).
% ?- zdd((append([], [], X))).
% ?- zdd((true, true)).
% ?- zdd((Y<<cnf((a+b)*(-(a)+c)), ltr_remove(-(a), Y, U),  sets(U, S))).
% ?- numlist(1, 16, L), zdd((X<< pow(L), card(X, C))).
% ?- numlist(1, 16, L), zdd((X<< pow(L))).
% ?- numlist(1, 10000, L), zdd((X<< pow(L), card(X, C))).
% ?- numlist(1, 16, L),
% zdd((X<< pow(L), Y<<pow(L), Z << (X+Y),  U<< (X*Y), card(X, C))).
% ?- time((numlist(1, 40, L),
% zdd((X << pow(L), Y<<pow(L), Z << merge(X, Y), card(Z, C))))).
% ?- zdd((set_memo(a-1), set_memo(a-2), memo(a-V))).
% ?- zdd(X<<{[1,1]}).
% ?- zdd((X<<pow([1,2,3]), prz(X))).
% ?- zdd((X<< *([1,2,3,1]), prz(X))).
% ?- zdd((X<< +([1,2,3,1]), psa(X))).
% ?- zdd((X<< family([[1],[1,2], [1]]), prz(X))).
% ?- zdd((X<< family([[a-b, b-c]]), prz(X))).
% ?- zdd( use_zdd((memo(a-1), memo(a-X)))).
% ?- zdd(use_zdd(zdd(use_zdd((memo(a-1), memo(a-X)))))).
% ?- zdd((X<<pow([1,2]), {zdd((Y<<pow([a,b]), sets(Y, Sy)))}, sets(X, Sx))).
% ?- zdd((X<<pow([1,2]), {zdd((Y<<pow([a,b]), sets(Y, Sy)))},
%	sets(X, Sx), sets(Y, Sy1))).
% ?- zdd((set_key(root, hello), get_key(root, R))).
% ?- zdd((all_fun([a],[b], F), psa(F))).
% ?- zdd((all_fun([],[b], F), psa(F))).
% ?- zdd((all_fun([a],[], F), psa(F))).
% ?- zdd((all_fun([a,b],[c,d], F), psa(F), card(F, C))).
% ?- N = 20, numlist(1, N, A),  raise_list(A, 2, A2), zdd((all_fun(A2, A, F), card(F, C))).
% ?- time((N=100, M=100,  numlist(1, N, Ns), numlist(1, M, Ms),
%	zdd((all_fun(Ns, Ms, F), card(F, C))))).

% Cardinality # of Y^X = { f | f: X-> Y }.
%   ({1,...,K}^L)^({1,...,I}^J) = (K^L)^(I^J).

% ?- I=1, J=1, K=1, L=1,
%	numlist(1, I, X), numlist(1, K, Y),
%	raise_list(X, J, Xj),
%	raise_list(Y, L, Yl),
%	time( ( call_with_time_limit(200, (
%		zdd((all_fun(Xj, Yl, F, default_pair), card(F, C))))),
%		C =:= (K^L)^(I^J),
%	    significant_length(C, Keta, 10))).

% ?- I=3, J=5, K=3, L=5,
%	numlist(1, I, X), numlist(1, K, Y),
%	raise_list(X, J, Xj),
%	raise_list(Y, L, Yl),
%	time( ( call_with_time_limit(200, (
%		zdd((all_fun(Xj, Yl, F), card(F, C))))),
%		C =:= (K^L)^(I^J),
%		significant_length(C, Keta, 10))).
%@ % 530,528,187 inferences, 58.036 CPU in 59.124 seconds (98% CPU, 9141365 Lips)


significant_length(0, 0, _):-!.
significant_length(X, 1, Radix):- X < Radix, !.
significant_length(X, L, Radix):- Y is X // Radix,
	significant_length(Y, L0, Radix),
	L is L0+1.

%
zdd_atom(X):- shift(zdd_atom(X)).
%
zdd_atom(X, S):- get_key(is_atom, Pred, S), !,	call(Pred, X).
zdd_atom('$VAR'(_), _):-!.
zdd_atom(X, _):- atomic(X).

%
hyphen(X, Y, X-Y).
comma(X, Y, (X,Y)).
equality(X, Y, (X=Y)).

		/*******************************************
		*     findall functions/mono/epi in ZDD    *
		*******************************************/

% ?- #(S), all_functions([a,b],[1,2], F, S), psa(F, S).
% ?- #(S), all_functions([a,b],[1,2], F, comma, S), psa(F, S).
% ?- #(S), all_functions([a,b],[1,2,3], F, equality, S), psa(F, S).
% ?- numlist(1, 10, Ns), #(S), all_functions(Ns, Ns, F, S), card(F, C, S).
% ?- numlist(1, 15, Ns), #(S), all_functions(Ns, Ns, F, S), card(F, C, S).
% ?- numlist(1, 20, Ns), #(S), all_functions(Ns, Ns, F, S), card(F, C, S).
:- meta_predicate  all_functions(?, ?, ?, 3, ?).
all_functions(X, Y, F, S):- zdd_sort(X, X0, S),
	zdd_sort(Y, Y0, S),
	all_functions(X0, Y0, F, hyphen, S).
%
all_functions([], _, 1, _, _).
all_functions([A|As], Bs, F, Pred, S):-
	all_functions(As, Bs, F0, Pred, S),
	extend_funs_by_one(A, Bs, F0, 0, F, Pred, S).
%
extend_funs_by_one(_, [], _, F, F, _, _):-!.
extend_funs_by_one(A, [B|Bs], H, F0, F, Pred, S):- call(Pred, A, B, Pair), !,
	zdd_insert(Pair, H, H0, S),
	zdd_join(H0, F0, F1, S),
	extend_funs_by_one(A, Bs, H, F1, F, Pred, S).


		/*****************************************
		*     all_fun/all_mono/all_epi in ZDD    *
		*****************************************/

% ?- #(S), all_fun([a,b],[1,2,3], F, S), card(F, C, S), psa(F, S).
% ?- #(S), all_fun([a, b, c],[1,2,3], F, S), card(F, C, S), psa(F, S).
% ?- #(S), all_fun([a, b, c, d, e],[1,2,3, 4], F, S), card(F, C, S).

% ?- N = 100, numlist(1, N, Ns), #(S), all_fun(Ns, Ns, F, S), card(F, C, S).
% ?-  numlist(1, 5, Ns),  numlist(6, 10, Ms),
% #(S), all_fun(Ns, Ns, F, S), all_fun(Ms, Ms, G, S),
%	zdd_merge(F, G, H, S), card(H, C, S), C =:= 5^5 * 5^5.

all_fun(D, R, F):- @(all_fun(D, R, F)).

%
all_fun(D, R, F, S):- zdd_sort(D, D0, S),
	zdd_sort(R, R0, S),
	length(D0, I),
	length(R0, J),
	(	I > 0, J = 0  -> F = 0
	;	zdd_append(D0, 1, D1, S),
		findall_fun(D1, R0, F, S)
	).
%
findall_fun(X, _, X, _):- X < 2, !.
findall_fun(X, Rng, Y, S):- memo(findall_fun(X)-Y, S),
	(	nonvar(Y) -> true		% , write(.)  % many hits.
	;	cofact(X, t(A, L, R), S),
		findall_fun(L, Rng, L0, S),
		findall_fun(R, Rng, R1, S),
		join_for_alt_val(A, Rng, R1, 0, R0, S),
		zdd_join(L0, R0, Y, S)
	).
%
join_for_alt_val(_, [], _, V, V, _).
join_for_alt_val(A, [B|Bs], F, F0, F1, S):-
	zdd_cons(F2, A-B, F, S),
	zdd_join(F0, F2, F3, S),
	join_for_alt_val(A, Bs, F, F3, F1, S).

% ?- #(S), all_mono([1],[a], Is, S), psa(Is, S).
% ?- #(S), all_mono([1,2],[a,b,c], Is, S), psa(Is, S).
% ?- #(S), all_mono([1,2],[a,b,c], Is, S), psa(Is, S).
% ?- #(S), all_mono([1,2,3],[a,b,c], Is, S), psa(Is, S), card(Is, C, S).
% ?- #(S), all_mono([1,2,3,4],[a,b,c,d], Is, S), card(Is, C, S).
% ?- time((numlist(1, 10, Ns), #(S), all_mono(Ns, Ns, F, S), card(F, C, S), factorial(10, C))).
% ?- time((numlist(1, 12, Ns), #(S), all_mono(Ns, Ns, F, S), card(F, C, S))).
% ?- N =10, time((numlist(1, N, Ns), #(S), time(all_mono(Ns, Ns, F, S)), card(F, C, S), factorial(N, C))).

all_mono(D, R, F, S):-  zdd_sort(D, D0, S),
	zdd_sort(R, R0, S),
	length(D0, I), length(R0, J),
	(	I @=< J ->
		zdd_append(D0, 1, D1, S),
		findall_mono(D1, R0, F, S)
	;	F = 0
	).


% ?- #S, zdd_append([], 1, X, S), findall_mono(X, [a], Y, S).
% ?- #S, zdd_append([1], 1, X, S), psa(X, S), findall_mono(X, [a], Y, S), psa(Y, S).
% ?- #S, zdd_append([1,2], 1, X, S), psa(X, S), findall_mono(X, [a,b], Y, S), psa(Y, S).
% ?- #S, zdd_append([1,2,3], 1, X, S), psa(X, S), findall_mono(X, [a,b,c], Y, S), psa(Y, S).
% ?- #S, N=14, numlist(1, N, Ns), zdd_append(Ns, 1, X, S),
%   time(findall_mono(X, Ns, Y, S)), card(Y, C, S),
%	factorial(14, D), ( D=:=C -> writeln("OK")).

findall_mono(X, _, X, _):- X < 2, !.
findall_mono(X, Rng, Y, S):- memo(findall_mono(X,Rng)-Y, S),
	(	nonvar(Y) -> true		% , write(.)  % many hits.
	;	cofact(X, t(A, L, R), S),
		findall_mono(L, Rng, L0, S),
		findall_mono(A, R, Rng, R0, S),
		zdd_join(L0, R0, Y, S)
	).
%
findall_mono(A, R, Rng, R0, S):- findall(B-U, select(B, Rng, U), Cases),
	findall_mono(Cases, A, R, 0, R0, S).
%
findall_mono([], _A, _R, V, V, _).
findall_mono([B-Rng|Cases], A, R, U, V, S):-
	findall_mono(R, Rng, U0, S),
	cofact(U1, t(A-B, 0, U0), S),
	zdd_join(U, U1, U2, S),
	findall_mono(Cases, A, R, U2, V, S).

% ?- #(S), all_epi([],[], F, S).
% ?- #(S), all_epi([a],[], F, S).
% ?- #(S), all_epi([a],[1], F, S), psa(F, S).
% ?- #(S), all_epi([a],[1,2], F, S), psa(F, S).
% ?- #(S), all_epi([a,b],[1], F, S), psa(F, S).
% ?- #(S), all_epi([a,b],[1,2], F, S), psa(F, S).
% ?- #(S), all_epi([a],[1,2], F, S), psa(F, S).
% ?- #(S), all_epi([a,b,c],[1,2], F, S), psa(F, S).
% ?- #(S), all_epi([a,b],[1,2,3], F, S), psa(F, S).
% ?- numlist(1, 10, Ns), numlist(1, 8, Ms), #(S), all_epi(Ns, Ms, F, S), psa(F, S).
% ?- numlist(1, 10, Ns), numlist(1, 9, Ms), #(S), all_epi(Ns, Ms, F, S).
% ?-time(( numlist(1, 10, Ns), numlist(1, 10, Ms), #(S), all_epi(Ns, Ms, F, S))).

all_epi(D, R, F, S):-
	zdd_sort(D, D0, S),
	zdd_sort(R, R0, S),
	length(D0, I), length(R0, J),
	(	I @>= J ->
		zdd_append(D0, 1, D1, S),
		findall_epi(D1, R0-R0, F, S)
	;	F = 0
	).

% ?- N=10, numlist(1, N, Ns),  #S, all_perm(Ns, X, S), card(X, C, S).
% ?- N=14, time((numlist(1, N, Ns),  #S, all_perm(Ns, X, S), card(X, C, S))).
%@ % 37,067,659 inferences, 12.701 CPU in 12.914 seconds (98% CPU, 2918525 Lips)
%@ N = 14,
%%@ X = 487425,
%@ C = 87,178,291,200.

% ?- N=15, time((numlist(1, N, Ns),  #S, all_perm(Ns, X, S), card(X, C, S))).
%@ % 84,073,484 inferences, 47.282 CPU in 47.647 seconds (99% CPU, 1778120 Lips)
%@ N = 15,
%@ X = 1105921,
%@ C = 1,307,674,368,000.
all_perm(D, P, S):-  zdd_sort(D, D0, S),
	findall_perm(D0, P, S).

%
findall_perm([], 1, _):-!.
findall_perm(D, X, S):- memo(perm(D)-X, S),
	( nonvar(X) -> true		%	, write(.) % many hits.
	;
	findall(I-D0, select(I, D, D0), U),
	join_perm_for_selection(U, 0, X, S)).
%
join_perm_for_selection([], X, X, _).
join_perm_for_selection([I-D|U], X, Y, S):-
	findall_perm(D, X0, S),
	zdd_cons(X1, I, X0, S),
	zdd_join(X1, X, X2, S),
	join_perm_for_selection(U, X2, Y, S).

% ?- #S, zdd_append([], 1, X, S), findall_epi(X, [a]-[a], Y, S).
% ?- #S, zdd_append([a], 1, X, S), findall_epi(X, [1]-[1], Y, S), card(Y, C, S).
% ?- #S, zdd_append([a,b], 1, X, S), findall_epi(X, [1]-[1], Y, S), card(Y, C, S).
% ?- #S, zdd_append([a,b], 1, X, S), findall_epi(X, [1,2]-[1,2], Y, S), card(Y, C, S).
% ?- #S, zdd_append([a,b, c], 1, X, S), findall_epi(X, [1,2]-[1,2], Y, S), card(Y, C, S), psa(Y, S).
% ?- N = 2, #S, numlist(1, N, Ns), zdd_append(Ns, 1, X, S), findall_epi(X, Ns-Ns, Y, S), card(Y, C, S).
% ?- N = 3, #S, numlist(1, N, Ns), zdd_append(Ns, 1, X, S), findall_epi(X, Ns-Ns, Y, S), card(Y, C, S).
% ?- N = 10, #S, numlist(1, N, Ns), zdd_append(Ns, 1, X, S), time(findall_epi(X, Ns-Ns, Y, S)), card(Y, C, S).
% ?- N = 11, #S, numlist(1, N, Ns), zdd_append(Ns, 1, X, S), time(findall_epi(X, Ns-Ns, Y, S)), card(Y, C, S).
% ?- N = 12, M=8, #S, numlist(1, N, Ns), numlist(1, M, Ms), zdd_append(Ns, 1, X, S), time(findall_epi(X, Ns-Ms, Y, S)), card(Y, C, S).
% ?- N = 12, M is N - 3, #S, numlist(1, N, Ns), numlist(1, M, Ms), zdd_append(Ns, 1, X, S), time(findall_epi(X, Ns-Ms, Y, S)), card(Y, C, S).
% ?- N = 13, M is N - 3, #S, numlist(1, N, Ns), numlist(1, M, Ms), zdd_append(Ns, 1, X, S), time(findall_epi(X, Ns-Ms, Y, S)), card(Y, C, S).
%
select_target(Range, U, B, V):- member(B, Range),
	(	select(B, U, V) -> true
	;	V = U
	).
%
findall_epi(0, _, 0, _):-!.
findall_epi(1, _-U, X, _):-!,
	(	U=[] -> X = 1
	;	X = 0
	).
findall_epi(X, RngU, Y, S):- memo(findall_epi(X, RngU)-Y, S),
	(	nonvar(Y) -> true		% , write(.)  % many hits.
	;	cofact(X, t(A, L, R), S),
		findall_epi(L, RngU, L0, S),
		findall_epi(A, R, RngU, R0, S),
		zdd_join(L0, R0, Y, S)
	).
%
findall_epi(A, R, Rng-U, R0, S):-
	findall(B-(Rng-V), select_target(Rng, U, B, V), Cases),
	findall_epi(Cases, A, R, 0, R0, S).
%
findall_epi([], _, _, V, V, _).
findall_epi([B-(Rng-Rng0)|Cases], A, R, U, V, S):-
	findall_epi(R, Rng-Rng0, U0, S),
	cofact(U1, t(A-B, 0, U0), S),
	zdd_join(U, U1, U2, S),
	findall_epi(Cases, A, R, U2, V, S).

% ?- #S, merge_funs([mono([1,2]-[a,b]), mono([3,4]-[c,d])], X, S), psa(X, S).
merge_funs(Fs, X, S):- fold_merge_funs(Fs, 1, X, S).
%
fold_merge_funs([], X, X, _).
fold_merge_funs([G|Fs], X, Y, S):-
	fun_block(G, X0, S),
	zdd_merge(X, X0, X1, S),
	fold_merge_funs(Fs, X1, Y, S).
%
fun_block(G, X, S):-
	(	G = mono(D-R) -> all_mono(D, R, X, S)
	;	G = epi(D-R) -> all_epi(D, R, X, S)
	;	G = fun(D-R),
		all_fun(D, R, X, S)
	).


% ?- N=2,  numlist(1, N, Ns),
%	#S, set_compare(opp_compare, S),
%	X<<pow(Ns), psa(X, S), set_compare(compare, S),
%	zdd_normalize(X, Y, S), psa(Y, S).

zdd_normalize(X, X, _):-X<2, !.
zdd_normalize(X, Y, S):- memo(normalize(X)-Y, S),
	(	nonvar(Y) -> true	%	, write(.) % Many hits.
	;	cofact(X, t(A, L, R), S),
		zdd_normalize(L, L0, S),
		zdd_normalize(R, R1, S),
		zdd_insert(A, R1, R0, S),
		zdd_join(L0, R0, Y, S)
	).

% ?- N = 1000, numlist(1, N, Ns),
%	time(zdd((X<<pow(Ns), rank_family_by_card(X, P),
%	memo(family_with_card(500)-L), card(L, C)))).
%@ % 90,174,772 inferences, 10.355 CPU in 10.540 seconds (98% CPU, 8708585 Lips)

% ?- N = 1000, numlist(1, N, Ns),
%	time(zdd(( X<<pow(Ns),
%			  get_family_of_rank(500, X, Y),
%			  card(Y, C)))).

%% get_family_of_rank(Order, X, F) is det.
%  Collect sets of a given cardinality.

get_family_of_rank(Order, X, F):- shift(get_family_of_rank(Order, X, F)).
%
get_family_of_rank(Order, X, F, S):- rank_family_by_card(X, _, S),
	memo(family_with_card(Order)-F, S).

% ?- zdd((X<<set([a]), rank_family_by_card(X, P),
%	memo(family_with_card(0)-L), sets(L, Sl),
%   memo(family_with_card(1)-M), sets(M, Sm))).

%% rank_family_by_card(X, Y[, S]) is det.
%
% Classifying sets in a family by size.
rank_family_by_card(X, Y):- shift(rank_family_by_card(X, Y)).
%
rank_family_by_card(0, 0, _):-!.
rank_family_by_card(1, 0, S):-!, memo(family_with_card(0)-1, S).
rank_family_by_card(I, P, S):- memo(rank_family_by_card_done(I)-B, S),
	(	nonvar(B) ->  true
	; 	cofact(I, t(A, L, R), S),
		rank_family_by_card(L, Pl, S),
		rank_family_by_card(R, Pr, S),
		max(Pl, Pr, P0),
		insert_one_to_family(A, P0, New, S),
		P is P0 + 1,
		memo(family_with_card(P)-New, S),
		B = true
	).
%
insert_one_to_family(A, 0, G, S):-!, memo(family_with_card(0)-F, S),
	zdd_insert(A, F, G, S).
insert_one_to_family(A, P, G, S):- P0 is P-1,
	insert_one_to_family(A, P0, G0, S),
	memo(family_with_card(P)-Fp, S),
	zdd_insert(A, Fp, G, S),
	zdd_join(Fp, G0, Gp, S),
	set_memo(family_with_card(P)-Gp, S).

		/**********************
		*     State access    *
		**********************/

get_extra(Extra):- shift(get_extra(Extra)).
get_extra(Extra, S):- arg(2, S, Extra).
%
set_extra(Extra):- shift(set_extra(Extra)).
set_extra(Extra, S):- setarg(2, S, Extra).

% ?- open_state(S), bump_up(a, N, S), bump_up(a, K, S).
bump_up(Key, N, S):- arg(2, S, Extra),
	(	select(Key-N0, Extra, Extra0) -> true
	;	N0 = 0,
		Extra0 = Extra
	),
	N is N0 + 1,
	setarg(2, S,  [Key-N|Extra0]).
%
:- meta_predicate set_compare(3).
:- meta_predicate set_compare(3, ?).
set_compare(Compare):- shift(set_compare(Compare)).
set_compare(Compare, S):- setarg(3, S, Compare).

% ?- zdd0((zdd_compare(C, a->b, a-b))).
% ?- zdd((zdd_compare(C, a, b))).
% ?- zdd0((zdd_compare(C, a, b))).
get_compare(Compare):- shift(get_compare(Compare)).
get_compare(Compare, S):- arg(3, S, Compare).
%
zdd_compare(C, X, Y):- shift(zdd_compare(C, X, Y)).
%
zdd_compare(C, X, Y, S):- arg(3, S, F), call(F, C, X, Y).

% ?- zdd((zdd_sort([1->1, 1->2, 1-1, 2-2], X))).
zdd_sort(X, Y):- shift(zdd_sort(X, Y)).
%
zdd_sort(X, Y, S):- get_compare(Comp, S), predsort(Comp, X, Y).

% ?- open_state(S), get_key(a, V, 1, S), get_key(a, W, S).
get_key(K, V):- shift(get_key(K, V)).
%
get_key(K, V, S):- arg(2, S, Assoc), memberchk(K-V, Assoc).
%
get_key(K, V, D, S):- arg(2, S, Assoc),
	(	memberchk(K-V, Assoc) -> true
	;	setarg(2, S, [K-D|Assoc]),
		V = D
	).
%
set_key(K, V):- shift(set_key(K, V)).
%
set_key(K, V, S) :- arg(2, S, Assoc),
	(	select(K-_, Assoc, Assoc0)
	->	setarg(2, S, [K-V|Assoc0])
	;	setarg(2, S, [K-V|Assoc])
	).

:- meta_predicate set_pred(?, :).
set_pred(K, V):- @(set_key(K, V)).

:- meta_predicate set_pred(?, :, ?).
set_pred(K, V, S) :- set_key(K, V, S).

% ?- zdd((open_key(a, []), update_key(a, X, Y), X=[1|Y],
%	get_key(a, Z), close_key(a))).
% ?- zdd((open_key(a, []), update_key(a, X, Y), X=[1|Y],
%	get_key(a, Z), close_key(a), get_key(a, U))).	% false
update_key(K, New, Old):- shift(update_key(K, New, Old)).
%
update_key(K, New, Old, S):- arg(2, S, Assoc),
	(	select(K-Old, Assoc, Assoc0) -> true
	;	Assoc0 = Assoc
	),
	setarg(2, S, [K-New|Assoc0]).
%
open_key(K, Val):- shift(open_key(K, Val)).
%
open_key(K, Val, S):- arg(2, S, Assoc),
	(	select(K-_, Assoc, Assoc0) 	->	true
	;	Assoc0 = Assoc
	),
	setarg(2, S, [K-Val|Assoc0]).

%
close_key(K):- shift(close_key(K)).
%
close_key(K, S):- arg(2, S, Extra),
	(	select(K-_, Extra, Extra0) -> setarg(2, S, Extra0)
	;	true
	).

% ?- zdd((varnum(D), varnum(E))).
%%	varnum(?D) is det.
% update a global variable varnum with D.
%
varnum(D):- shift(get_key(varnum, D)).

% ?- zdd((cofact0(X, a, 1), psa(X), cofact0(X, A, B))).
% ?- zdd((cofact0(X, 1, a), psa(X), cofact0(X, A, B))).
% ?- zdd((cofact0(X, 1, 0), psa(X), cofact0(X, A, B))).

%% cofact0(X, A, Y[, G]) is det.
%	 Restriction of cofact/[2,3] to a singleton family.

cofact0(X, A, Y):- shift(cofact0(X, A, Y)).
%
cofact0(X, A, Y, S):- var(X), !, cofact(X, t(A, 0, Y), S).
cofact0(X, A, Y, S):- X > 1, cofact(X, t(A, _, Y), S).

% ?- sort([1->1, 1->2, 1-1, 2-2], X).
% ?- predsort(compare_rev(compare), [1->1, 1->2, 1-1, 2-2], X).
% ?- zdd(zdd_sort_rev([1,2,3], X)).
compare_rev(Comp, C, A, B):- call(Comp, C, B, A).
%
zdd_sort_rev(X, Y):- shift(zdd_sort_rev(X, Y)).
zdd_sort_rev(X, Y, S):- get_compare(Comp, S),
	predsort(compare_rev(Comp), X, Y).

% ?- pred_zdd(path_count:opp_compare, zdd_sort([1-1, 2-2, 3-3], Y)).
:- meta_predicate pred_zdd(3, :).
pred_zdd(Comp, X):-  open_state(S), pred_zdd(Comp, X, S).
%
pred_zdd(Comp, X, S):- set_compare(Comp, S), zdd(X, S).

% ?- zdd((X<<pow([a,b,c]), zdd_memberchk([a,c], X))).
% ?- zdd((X<<pow([a,b,c]), zdd_memberchk([], X))).
% ?- zdd((X<<pow([a,b,c]), zdd_memberchk([a,c,d], X))). % false

%% zdd_memberchk(L, Z[, G]) is det.
%	A list L as a set is a member of a family Z of sets.
zdd_memberchk(L, Z):- shift(zdd_memberchk(L, Z)).
%
zdd_memberchk(L, Z, S):- zdd_sort(L, L0, S),
	zdd_ord_memberchk(L0, Z, S).
%
zdd_ord_memberchk(L, Z):- shift(zdd_ord_memberchk(L, Z)).
%
zdd_ord_memberchk([], Z, S):- !, zdd_has_1(Z, S).
zdd_ord_memberchk([A|As], Z, S):- Z>1,
	cofact(Z, t(B, L, R), S),
	(	A == B
	->	zdd_ord_memberchk(As, R, S)
	;	A @> B,
		zdd_ord_memberchk([A|As], L, S)
	).

% PASS
% ?- #S, X<<pow([a,b]), card(X, C, S).
% ?- zdd((X<<(pow([a,b])-pow([a])), card(X, C))).
% ?- zdd((X=1, Y =2, Z is X + Y)).
% ?- zdd((X<<set([a]), sets(X, S))).
% ?- zdd((X<<set([a]), S<<sets(X))).
% ?- zdd((X<<pow([a,b]), S<<sets(X))).
% ?- zdd((X<<pow([a,b,c,d,e,f]), Y<<pow([a,b,c,d,e,f]), U<<(X-Y), S<<sets(U))).
% ?- zdd(((X<< *(:append([a,b], [c,d]))), sets(X, Y))).
% ?- zdd(((X<< *(!append([a,b], [c,d]))), sets(X, Y))).
% ?- zdd((X<<([a,b]&[c,d]), S<<sets(X))).
% ?- zdd((X<<{[a],[b],[c]}, sets(X,S))).
% ?- zdd((X<<{[a],[b, c]}, sets(X,S))).
% ?- zdd((X<<{[a],[b, c]}, Y<<{[c, b, c]}, Z<<(X-Y), sets(Z,S))).
% ?- zdd((X<<{[a],[b, c]}, Y<<{[b, c], [a], [a]}, Z<< (X-Y), U<<sets(Z))).
% ?-  I =1000, J =2000,
%	time(zdd(( numlist(1, I, L), numlist(1, J, M),
%	X << pow(L), Y << pow(M),
%	Z << (Y - X), card(Z, C), C =:= 2^J-2^I))).

% PASS
% ?- zdd((X<<cnf(a), Y<<sets(X))).
% ?- zdd((X<<cnf(-a), Y<<sets(X))).
% ?- zdd((X<<cnf(a+b), Y<<sets(X))).
% ?- zdd((X<<cnf(-(a+b)), Y<<sets(X))).
% ?- zdd((X<<cnf(-(a+b+b+a)), Y<<sets(X))).
% ?- zdd((X<<cnf(-(-(a+b+b+a))), Y<<sets(X))).
% ?- zdd((X<<cnf(-(-a + -b)), Y<<sets(X))).
% ?- zdd((X<<dnf(a), Y<<sets(X), extern(Y, Y0))).
% ?- zdd((X<<dnf(a->b), Y<<sets(X), extern(Y, Y0))).
% ?- zdd((X<<dnf(a+b -> c*d), Y<<sets(X), extern(Y, Y0))).
% ?- zdd((X<<dnf(-(-a)), extern(X, X0), Y<<sets(X0))).
% ?- zdd((X<<dnf(a+b), Y<<sets(X))).
% ?- zdd((X<<dnf(-(a+b)), Y<<sets(X))).
% ?- zdd((X<<dnf(-(a+b+b+a)), Y<<sets(X))).
% ?- zdd((X<<dnf(-(-(a+b+b+a))), Y<<sets(X))).
% ?- zdd((X<<dnf(-(-a + -b)), Y<<sets(X))).
% ?- zdd((X<<dnf((-a + a)), Y<<sets(X))).
% ?- zdd((X<<dnf(-(-a + a)), Y<<sets(X))).
% ?- zdd(X << *(:append([a,b], [c,d]))).
% ?- zdd(X << *(!append([a,b], [c,d]))).
% ?- zdd((X << ((*[a,b]+ *[c,d])* *[c,d]), sets(X, S))).
% ?- zdd((X << *[a,b], sets(X, S))).
% ?- zdd((X << +[a,b], sets(X, S))).
% ?- zdd((X << dnf(a*b+c*d+c*d*d), sets(X, S))).
% ?- zdd((zdd_list_to_singletons([], X), prz(X))).
% ?- zdd((zdd_list_to_singletons([a], X), prz(X))).
% ?- zdd((zdd_list_to_singletons([a,b], X), prz(X))).
% ?- zdd((zdd_list_to_singletons([c, a, b], X), prz(X))).


% ERROR
% ?- I =1000, J is I + 1, numlist(1, I, L),
%	prepare_ltr_list(L, L0),
%	time(zdd(( X<<ltr_ord_power(L0), ltr_has_set([-(1), 3, 3,4, 6, I], X)))).

%! zdd_list_to_singletons(+As, -X, +S) is det.
%  X is unified with a ZDD which is the family of

%  singletons of an element of X.

%  Example.  If  X=[a,b,c] then X is [[a], [b], [c]]
%  as a family of sets in ZDD.

zdd_list_to_singletons(As, X):- shift(zdd_list_to_singletons(As, X)).
%
zdd_list_to_singletons([], 1, _).
zdd_list_to_singletons([A|As], X, S):-zdd_list_to_singletons(As, Y, S),
		zdd_singleton(A, Y0, S),
		zdd_join(Y0, Y, X, S).

% PASS.
% ?- zdd((zdd_partial_choices([], X), prz(X))).
% ?- zdd((zdd_partial_choices([[a], [b]], X), prz(X))).
% ?- zdd((zdd_partial_choices([[a, a1], [b, b1], [c, c1]], X), prz(X))).
% ?- zdd((zdd_partial_choices([[a, a], [a, b], [a, c]], X), prz(X))).
% ?- zdd((zdd_partial_choices([[b, a], [a, b], [a, a, b, b]], X), prz(X))).
% ?- zdd((zdd_partial_choices([[c, c], [b, b, b], [a, a]], X), prz(X))).
% ?- zdd((zdd_partial_choices([[a, a], [], [a, c]], X), prz(X)).)
% ?- zdd((zdd_partial_choices([[a], [b], [c], [a], [d]], X), prz(X), card(X, C))).
% ?- zdd((zdd_partial_choices([[a,a1], [b,b1]], X), prz(X), card(X, C))).
% ?- findall([I], between(1, 10000, I),  Ls),
%	zdd((zdd_partial_choices(Ls, X),
%	card(X, C))), !, C =:= 2^10000.

%!	zdd_partial_choices(+X, -Y, +S) is det.
%   Y is the ZDD of all possible partial choices from a list X of lists.

zdd_partial_choices(X, Y):- shift(zdd_partial_choices(X, Y)).
%
zdd_partial_choices([], 1, _):-!.
zdd_partial_choices([L|Ls], X, S):-  zdd_partial_choices(Ls, X0, S),
	sw_fold_insert(zdd, L, X0, 0, X1, S),
	zdd_join(X0, X1, X, S).

%!	s_map(+F:pred/3, ?Xs, ?Ys, +S) is det.
%   Almost equal to maplist(F(S), Xs, Ys).
%
:- meta_predicate s_map(:, ?, ?, ?).
%
s_map(_, [], [], _):-!.
s_map(F, [X|Xs], [Y|Ys], S):- call(F, X, Y, S), !, s_map(F, Xs, Ys, S).

%!	s_map(+F:pred/2, ?Xs, ?Ys, +S) is det.
%   Almost equal to maplist(F(S), Xs).

:- meta_predicate s_map(:, ?, ?).
%
s_map(_, [], _):-!.
s_map(F, [X|Xs], S):- call(F, X, S), !, s_map(F, Xs, S).

% PASS.
%
% ?- zdd((dnf0((A+B)*(C+D)*(E+F), X))).
% ?- zdd((dnf0((A+B)*(C+D)*(E+F), X), sets(X, G))), maplist(writeln, G).
% ?- zdd((dnf0((A+B)*(C+D)*(E+F), F1), sets(F1, G))), maplist(writeln, G).
% ?- zdd:zdd((P<<pow([1,2,3]), use_zdd((X=1;X=2)), card(P,C))).
% ?- zdd((P << (set([1,2,3,2,3])+set([4, 2,4])), prz(P))).
% ?- zdd({append([a,b], [c,d], X)}).
% ?- zdd(X << :append([a,b], [c,d])).
% ?- zdd(zdd(X << :append([a,b], [c,d]))).
% ?- zdd(zdd(zdd((X << :append([a,b], [c,d]))))).
% ?- zdd(zdd(zdd((X<<pow([1,2]), true, true)))).
% ?- zdd(zdd(zdd((X<<pow([1,2]), true, true, card(X, C))))).

%!	ltr_zdd(+E) is det.
%	Solve E as command on families of clauses.

:-meta_predicate ltr_zdd(:).
ltr_zdd(X):- zdd((set_compare(ltr_compare), X)).

%!	zdd(+E, +S) is det.
%   Solve E as commands on ZDD with compare/3 for ordering atoms.
:- meta_predicate zdd(:).
zdd(Exp):- use_zdd((set_key(atom_index, 2),
					set_compare(compare),
					Exp)).
%
:- meta_predicate zdd(0, ?).
zdd(M:A, S):- zdd_solve(A, S, M).

%!	zdd0(+E, +S) is det.
%   Solve E as commands on ZDD with opp_compare/3 for ordering atoms,
%	which is the reverse ordering of the standard compare predicate.

zdd0(X):- zdd((set_compare(opp_compare), X)).
:- meta_predicate zdd0(:).
zdd0(X):- zdd((set_compare(opp_compare), X)).

%!	opp_compare(-C, +X, +Y) is det.
%	Equivalent to opp_compare(C, Y, X).

% PASS.
% ?- zdd((zdd_sort([a-b,a->b, b-c, b->c], X))).
% ?- zdd0((zdd_sort([a-b,a->b, b-c, b->c], X))).
% ?- opp_compare(C, 1, 2).
% ?- opp_compare(C, a->b, b).
% ?- opp_compare(C, p(0,0), p(1,0)).
% ?- opp_compare(C, p(1,0)-p(0,0), p(1,0)-p(1,0)).

opp_compare(<, -(_,_), ->(_,_)):-!.
opp_compare(>, ->(_,_), -(_,_)):-!.
opp_compare(C, X, Y):- compare(C, Y, X). % reverse order

% PASS.
% ?- open_state(S), reset_continue((memo(a-1), memo(a-R)), S).
% ?- open_state(S), reset_continue(( zdd((X<<pow([1,2]),
%	memo(a-X), card(X, C)))), S).
% ?- zdd((#set_key(a, 1), #get_key(a, X))).
% ?- zdd((#set_key(a, 1), #get_key(a, X))).

%! reset_continue(+G, +S) is det.
%  Reset goal G and coninue the returned remaining goal.
:- meta_predicate reset_continue(:, ?).
reset_continue(Goal, S):- reset(Goal, Ball, Rest),
	(	var(Ball) -> true
	;	context_module(M),
		process_ball(Ball, S, M)
	),
	zdd(Rest, S).
%
process_ball(M:X, S, _):-!, process_ball(X, S, M).
process_ball((X, Y), S, M):-!, process_ball(X, S, M),
	process_ball(Y, S, M).
process_ball(X, S, M):- call(M:X, S).

%!	use_zdd(+X, +Options) is det.
%   Open a state S for given Options, then solve goal X.
%	On exit, the state S is closed.

:- meta_predicate  use_zdd(:).
use_zdd(X):- use_zdd(X, [hash_size(128)]).

:- meta_predicate  use_zdd(:, ?).
use_zdd(X, Options):-
	setup_call_cleanup(
		open_state(S, Options),
		zdd(X, S),
		close_state(S)).

%! zdd_shift(G) is det.
%  Perform shift-wrapped command. Short hand for used_zdd(shift(G))
%  for convenience.
:- meta_predicate zdd_shift(:).
zdd_shift(G):- use_zdd(shift(G), [hash_size(256)]).

:- meta_predicate zdd_shift(:, ?).
zdd_shift(G, Options):- use_zdd(shift(G), Options).

% PASS.
% ?- zdd((X<<pow([1,2]), memo(a-X), card(X, C))).
% ?- zdd(?(( =(U), =(V)))).
% ?- zdd(?((zdd_eval(pow([a,b]), X), psa(X)))).

%!  zdd_solve(E, S, M) is det.
%   E is command expression,  S a state, M a module name.
%	E is interpred as a command on S in module M.
%	(A, B)		(A, B).sequentiual (and)
%	X<<E		evaluate E, then unify X with it.
%	M:G			evaluate G under module M
%	true		true
%	0		    treated as true
%	{G}			plain prolog goal.
%	#(G)		call(G, State) via shif/reset
%	(A ; B)		(A; B) (or)
%	(A -> B)	A->B   (condional)
%	Other		Reset, and handle the ball and continuation.

% PASS.
% ?- X<< +[[a], [b]], #(S), psa(X, S).
% ?- X<< +[[a], [b]], #(S), psa(X, S).
% ?- <<(X,  +[[a], [b]], S), psa(X, S).

%!	#(?S) is det.
%	If S is uninstantiated then open new state S.
%	Store S to global variable `$zdd.`

#(S):-
	(	var(S) ->
		(	nb_current('$zdd', S), S\==[] -> true
		;	open_state(S),
			b_setval('$zdd', S)
		)
	;	b_setval('$zdd', S)
	).

get_sharp(S):- var(S), !, #S.
get_sharp(_).

%!	<<(-X, +E).
%	With state S stored the global variable '$zdd',
%	Evaluate expression E then unify obtained value with X.

:- meta_predicate <<(?, :).
<<(X, E) :- #(S), zdd_eval(E, X, S).

%!	<<(-X, +E, +S) is det.
%	Evaluate expression E in state S then unify obtained value with X.

:- meta_predicate <<(?, :, ?).
<<(X, E, S) :- #(S), zdd_eval(E, X, S).

%!	zdd_solve(+G, +S, +M) is det.
%	Solve goal G with state S and module M.

% ?- #S, X<< +[ *[1,2], *[3,4]], psa(X, S).
% ?- zdd((X<<pow(:numlist(1, 100)), #card(X, C))).

zdd_solve(0, _, _)		:-!.
zdd_solve(true, _, _)	:-!.
zdd_solve((A,B), S, M)	:-!, zdd_solve(A, S, M), zdd_solve(B, S, M).
zdd_solve(X<<E, S, M)	:-!, zdd_eval(E, X, S, M).
zdd_solve(M:G, S, _)	:-!, zdd_solve(G, S, M).
zdd_solve({G}, _, M)	:-!, call(M:G).
zdd_solve(#(G), S, M)	:-!, zdd_solve_sharp(G, S, M).  %  Was @
zdd_solve(?(G), S, M)	:-!, process_ball(G, S, M).
zdd_solve((A;B), S, M)	:-!, (zdd_solve(A, S, M); zdd_solve(B, S, M)).
zdd_solve((A->B), S, M)	:-!, (zdd_solve(A, S, M) -> zdd_solve(B, S, M)).
zdd_solve(Cont, S, M)	:- reset_continue(M:Cont, S).

%
zdd_solve_sharp((A, B), S, M):-!,
	zdd_solve_sharp(A, S, M),
	zdd_solve_sharp(B, S, M).
zdd_solve_sharp(A, S, M):-call(M:A, S).

% PASS.
% ?- zdd((X<<cofact(a, 1, 1), psa(X), Y<<cofact(X), card(X, C))).
% ?- zdd((X<<cofact(a, 1, 1), psa(X), Y<<fact(X), card(X, C))).

%!	zdd_eval(+E, -V, +S) is det.
%   Evaluate E and unify the obtained value with V in state S.
%
%	x			x  (nummber, string, or list)
%
%	set(L)		{L}		a singleton family.
%	*(L)		<==>	set(L)
%	atom(a)		{{a}}   singleton family.
%	{a,b,..}	{a, b, c, ...} a family of sets a, b, ...
%	fos(list of lists)	a family of sets a, b, ...
%   +		join (union)
%	-		subtract
%	\		subtract
%	&		intersection
%	merge   merge/multiply
%	*		merge/multiply
%	//		quotient
%	/		remainder
%	prod	product
%	**		product
%	pow		powerset
%	power	powerset
%	sets	convert to a form of list of lists
%	cnf		CNF
%	dnf		DNF
%   cofact/3	cofact  (for tutorial purpose, to be dropped)
%   cofact/1	cofact  (for tutorial purpose), to be dropped)
%   fact/1		cofact  (for tutorial purpose), to be dropped)
%	$E		E (quote).
%   :E		eval E as a usual prolog term, but  keeping the state.
%   !E		apply  E without state. Special case of :.
%	#E		eval E as a ZDD term referering to the state
%   @E		call E without evaluating args depending on modes
%   *E		eval vector E then fold list by merge
%   +E		eval vector E then fold list by join
%   E		evaluate args of E and then apply it.



% Pass.
% ?- zdd((X<<dnf(a*b), psa(X))).
% ?- zdd((X<<dnf0(a*b), psa(X))).
% ?- zdd((X<<dnf0((-a)* -b * a), psa(X))).
% ?- zdd((X<<dnf0((a->b)*(b->c) -> (a->c)), psa(X))).

% Pass.
% ?- ltr_zdd((X<<dnf0((a->b)*(b->c) -> (a->c)), @ltr_card(X, C),  psa(X))).
% ?- ltr_zdd((X<<dnf((a->a)), ltr_card(X, C),  psa(X))).
%
%% Pass.
% ?- zdd((X<<dnf((a->b)*(b->c)*a -> c), ltr_card(X, C))).
% ?- zdd((X<<dnf((a->b)*(b->c)->(b->c)), ltr_card(X, C))).
% ?- zdd((X<<dnf((a->b)*(b->c)->(b->c)), psa(X), ltr_card(X, C))).

% Pass.
% ?- zdd((X<<pow([a,b]), card(X, C))).
% ?- zdd((X<<(pow([a,b])-pow([a])), card(X, C))).
% ?- zdd((X<< a, psa(X))).
% ?- zdd((X<< atom(a), psa(X))).
% ?- zdd((X<< atom(a(1)), psa(X))).
% ?- zdd((X<< set(:(append([1,2],[3,4]))), psa(X))).
% ?- zdd((X<< +(:append([1,2],[3,4])), psa(X))).
% ?- zdd((X<< (a+b), psa(X))).
% ?- zdd((X<< ((a+b)*c), psa(X))).
% ?- zdd((X<< [a,b])).
% ?- zdd((X<< set(:(append([a,b],[c,d]))))).
% ?- zdd((X<< set([1]), Y<< (X+X), psa(X))).
% ?- zdd((X<< set([1]), psa(X))).
% ?- zdd((X<< set([]))).
% ?- zdd((X<< fos([]))).
% ?- zdd((X<< fos([[]]))).
% ?- zdd((X << {[a], [b], [c]}, psa(X))).
% ?- zdd((X << {[a], [b], [c]}, card(X, C))).
%

% PASS.
% ?- zdd((X<< pow(:append(numlist(1,3), numlist(4,5))), card(X, C))).
% ?- zdd((X<< set(:append(numlist(1,2), numlist(4,5))))).
% ?- zdd((X<< set(:append(:numlist(1,10), :numlist(11,20))))).
% ?- zdd((numlist(1,10,A), numlist(11,20, B), X<< pow( :append(A, B)),
%	card(X, C))).
% ?- zdd((X << pow(:(numlist(1,2))), card(X, C))).
% ?- zdd((X << pow(!(charlist(a,z))), card(X, C))).
% ?- zdd((X << pow(!(charlist(a-z))), card(X, C))).
% ?- zdd((X << pow([a,b]))).
% ?- zdd((X << *[a, b, c])).
% ?- zdd((X << (*[a, b, c] + *[a,b,c]), psa(X))).
% ?- zdd((X << (*[a, b, c] + *[atom(2),atom(3)]), psa(X))).
% ?- zdd((C << card(pow([a,b,c,1,2,3])))).
% ?- zdd((C << card(pow(:append([a,b,c], [1,2,3]))))).
% ?- zdd((C << pow(:numlist(1, 3)))).
% ?- zdd((C << atom(a), psa(C))).
% ?- zdd((C << atom(a), psa(C))).
% ?- zdd((C << atom([a]), psa(C))).

% Pass.
% ?- zdd((U << (pow([a])+pow([b,c]) + pow(:charlist($(a-z)))), card(U, C))).
% ?- zdd((U << (pow([a])+pow([b,c]) + pow(:charlist($a,$z))), card(U, C))).
% ?- zdd(((C<< pow(:charlist($a, $b)), card(C, N)))).
% ?- zdd(((C<< pow(:(atomlist($a(1,10)))), card(C, N)))).
% ?- zdd(((C<< pow(:charlist($a, $b)), card(C, N)))).
% ?- zdd((C<< set(:charlist($a,$z)))).
% ?- zdd((U << *(:append([a], [b])))).
% ?- zdd((U << :append([a], [b]), V<< *U, psa(V))).
% ?- zdd((U << :append([a], [b]), V<< +U, psa(V))).
% ?- zdd((U << *(:append(:append([a], [b]), :append([c], [d]))), card(U, C))).
%

% Pass.
% ?- zdd((X << atom($(a-z)), card(X, C))).
% ?- zdd(C << card(pow(:numlist(0, 100)))).
% ?- zdd(C << card(pow(:(numlist(0, 100))))).
% ?- zdd((X << pow([a, b, c]), zdd_super_power([a, b], X, Y), psa(Y))).
% ?- zdd((U << (pow(:append([a], [b])) + pow(:append([c], [d]))), card(U, C))).
% ?- zdd((U << (pow(:append([a], [b])) + pow([1, 2])), card(U, C))).
% ?- zdd((U << (pow([a])*b*c), psa(U))).
% ?- zdd((U << (pow([a])*b*c))).
% ?- zdd((U << (pow([a])), psa(U))).
% ?- zdd((U << (pow([a])* b), psa(U))).
% ?- zdd((U << (pow([a])* b), psa(U))).
% ?- zdd((U << (a*b*c), psa(U))).
% ?- zdd((U << a, psa(U))).
% ?- zdd((U << a, ?(sets(U, S)))).
% ?- zdd((U << a, shift(sets(U, S)))).
% ?- zdd((U << (a+b), psa(U))).
% ?- zdd((U << (a*b+c*d), psa(U))).
% ?- zdd((U << ((a+b)*(c+d)), psa(U))).
% ?- zdd((U << ((a+b)*1), psa(U))).
% ?- zdd((U << ((a+1+b)*1*1*1), psa(U))).
% ?- zdd((U << ((a+1+b) / (a+b)), psa(U))).
% ?- zdd((U << ((a+1+b) / (a*b)), psa(U))).
% ?- zdd((U << (pow(:charlist($a, $b)) mod (a+b)), psa(U))).
% ?- zdd((U << ((a+b+c)*(x+y+z)), psa(U))).
% ?- zdd((U << (atom(1)*(x+y+z)), psa(U))).
%

% Pass.
% ?- zdd((U << +(:(append([a,b,c], [d,e,f]))), psa(U))).
% ?- zdd((U << *(:(append([a,b,c], [d,e,f]))), psa(U))).
% ?- zdd((U << +(:(append(:append([a,b,c], [x,y,z]), [atom(1), atom(2)]))), psa(U))).
% ?- zdd((U << +[ *[a,b], *[c,d]],  psa(U))).
% ?- zdd((U << *[ +[a,b], +[c,d]],  psa(U))).
% ?- zdd((U << *[ *[a,b], *[c,d]],  psa(U))).
% ?- zdd((U << *[ *[a,b], +[c,d]],  psa(U))).
% ?- zdd((U << card(pow(:(append(:append([a,b,c], [x,y,z]), [1,2,3])))))).
%

% Pass.
% ?- zdd((U << :(=(3)))).
% ?- zdd((U << *([1,2,3]), psa(U))).
% ?- zdd((U << :plus(#(card(pow([a,b]))), #(card(pow([1,2])))))).
% ?- zdd((U << a)).  % singleton.
% ?- zdd((U << 2)).
% ?- zdd((U << [a,b])).
% ?- zdd((U << *[a,b], psa(U))).
% ?- zdd((U << +[a,b], psa(U))).
% ?- zdd((U << +[*[a,b], *[c,d]], psa(U))).
% ?- zdd((U << +[+[a,b], +[c,d]], psa(U))).
% ?- zdd((U << *[+[a,b], +[c,d]], psa(U))).
% ?- zdd((U << *[*[a,b], *[c,d]], psa(U))).
% ?- zdd((U << !(call(=, hello)))).
% ?- zdd((U << !(call(=, :append([a,b],[c,d]))))).
% ?- zdd((U << !call(=, :append([a,b],[c,d])))).
% ?- zdd((U << :call($(=), :append([a,b],[c,d])))).
% ?- #S, X<< *[a,b], Y<< *[c,d], Z<<  @zdd_join(X, Y), psa(Z, S).
% ?- #S, X<< *[a,b], Y<< *[c,d], Z<<  zdd_join(X, Y), psa(Z, S).
%
basic_type(X):-	number(X);
	string(X);
	is_list(X);
	var(X).


%!	zdd_eval(+X, -Y, +S) is det.
%	Evaluate X and unify the value with Y in state S.
zdd_eval(X, Y):- shift(zdd_eval(X, Y)).
%
zdd_eval(X, Y, S):- context_module(Mod), zdd_eval(X, Y, S, Mod).

% illegal use of pragma.
% ?- zdd((U << @(=(3)))).					% Error
% ?- zdd((U << #append([a], [b]))).			% Error
% ?- zdd((U << +(!append([a], [b])))), psa(U).	% Error
% ?- zdd((U << #(get_compare), psa(U))).		% Error
% ?- zdd( #(get_compare, X)).					% Error

% Normal use of pragmas, though meaningless.
% ?- zdd((U << !append([a], [b]))).
% ?- zdd((U << @(get_compare))).
% ?- zdd((U << +(!append([[a]], [[b]])), psa(U))).
% ?- zdd((U << get_compare, psa(U))).
% ?- zdd( #(get_compare(X))).

%!	zdd_eval(+E, -V, +S, +M) is det.
%	Evaluate E in state S in context module M, and unify the value with V.

%	Note that intgeger is used as a unique name of a ZDD.
%	use indicator `atom` as `atom(3)` for 3 as an atom.
%	Remark.  # is not allowed for zdd_eval.
zdd_eval(X, X, _, _)	:- basic_type(X), !.
zdd_eval({X}, Y, S, _)	:-!, associative_comma(X, U, []),
	zdd_family(U, Y, S).
zdd_eval($(X), X, _, _)	:-!.
zdd_eval(@(X), Y, S, M)	:-!, call(M:X, Y, S).
zdd_eval(:(X), Y, S, M)	:-!, prolog_eval(X, Y, S, M).
zdd_eval(!(X), Y, _, M)	:-!, call(M:X, Y).
zdd_eval(M:X, Y, S, _)	:-!, zdd_eval(X, Y, S, M).
zdd_eval(X, Y, S, M)	:-
	(	apply_type(X)	-> X0 = X
	;	zdd_eval_args(X, X0, S, M)
	),
	zdd_apply(X0, Y, S, M).
%
zdd_eval_flip(S, M, X, Y):- zdd_eval(X, Y, S, M).
%
zdd_eval_args(A, A, _, _):- atomic(A), !.
zdd_eval_args(A, B, S, M):-
	functor(A, Fa, Na),
	functor(B, Fa, Na),
	zdd_eval_args(1, A, B, S, M).
%
zdd_eval_args(I, A, B, S, M):- arg(I, A, U),
	zdd_eval(U, V, S, M),
	!,
	arg(I, B, V),
	J is I+1,
	zdd_eval_args(J, A, B, S, M).
zdd_eval_args(_, _, _, _, _).

%!	prolog_eval(E, V, S, M)	is det.
%	With context module M, evaluate E without referring to state S
%	as default, and unify the value with V.
%	Remark.
%	sharp(#) is for reserved zdd function symbols which is defined
%	in prolog_eval,  and is referred from prolog term.
%	#-term is interpreted in zdd_eval.
prolog_eval(X, X, _, _)		:- basic_type(X), !.
prolog_eval($(X), X, _, _)	:-!.
prolog_eval(@(X), Y, _, M)	:-!, call(M:X, Y).
prolog_eval(:(X), Y, S, M)	:-!, prolog_eval(X, Y, S, M).
prolog_eval(#(X), Y, S, M)	:-!, zdd_eval(X, Y, S, M).
prolog_eval(!(X), Y, _, M)	:-!, call(M:X, Y).
prolog_eval(M:X, Y, S, _)	:-!, prolog_eval(X, Y, S, M).
prolog_eval(X, Y,  S, M)	:- prolog_eval_args(X, X0, S, M),
							   call(M:X0, Y).

%
zdd_fold_join([], Z, Z, _).
zdd_fold_join([X|Xs], Z, Z0, S):-
	zdd_join(X, Z, Z1, S),
	zdd_fold_join(Xs, Z1, Z0, S).
%
zdd_fold_merge([], Z, Z, _).
zdd_fold_merge([X|Xs], Z, Z0, S):-
	zdd_merge(X, Z, Z1, S),
	zdd_fold_merge(Xs, Z1, Z0, S).
%
prolog_eval_args(M:A, M:B, S, _):-!, prolog_eval_args(A, B, S, M).
prolog_eval_args(A, A, _, _):- atom(A), !.
prolog_eval_args(A, B, S, M):-
	functor(A, Fa, Na),
	functor(B, Fa, Na),
	prolog_eval_args(1, A, B, S, M).
%
prolog_eval_args(I, A, B, S, M):- arg(I, A, U),
	prolog_eval(U, V, S, M),
	!,
	arg(I, B, V),
	J is I+1,
	prolog_eval_args(J, A, B, S, M).

prolog_eval_args(_, _, _, _, _).

%
apply_type(dnf(_)).
apply_type(cnf(_)).
apply_type(dnf0(_)).
apply_type(cnf0(_)).
apply_type(atom(_)).
apply_type(cofact(_,_,_)).
apply_type(fact(_,_,_)).
%
fold_singleton_join([], X, X, _).
fold_singleton_join([A|As], X, Y, S):-
	zdd_singleton(A, U, S),
	zdd_join(U, X, X0, S),
	fold_singleton_join(As, X0, Y, S).

		/*********************************
		*     zdd vecrtor expressions    *
		*********************************/

%!	zdd_vector_exp(+E, -V, +S) is det.
%	Evaluate vector notation for ZDD, and unify the value with S.

% ?- zdd((X<< *(: (=([[a,b],[c,d]]))), psa(X))).
% ?- zdd((X<< +(: (=([[a,b],[c,d]]))), psa(X))).
% ?- zdd((X<< +(: (=([*[a,b], *[c,d]]))), psa(X))).
% ?- zdd((X<< +[[1,2],[a,b]], psa(X), card(X, C))).
% ?- zdd((X<< *[[1,2],[a,b]], psa(X), card(X, C))).
% ?- zdd((X<< +[*[1,2],*[a,b]], psa(X), card(X, C))).
% ?- zdd((X<< +[*[x(1),x(2)],*[a(1),b(1)]], psa(X), card(X, C))).
% ?- zdd((X<< *[1,2], psa(X), card(X, C))).
% ?- zdd((X<< *[+[1,2],+[a,b]], psa(X), card(X, C))).
% ?- #S, zdd_vector_exp(+[[a],[b,c]], X, S), psa(X, S).
% ?- #S, zdd_vector_exp(+[[a],*[b,c]], X, S), psa(X, S).
% ?- #S, zdd_vector_exp(+[a,b,c], X, S), psa(X, S).
% ?- #S, zdd_vector_exp(*[a,b,c], X, S), psa(X, S).
% ?- #S, zdd_vector_exp(+[*[a,b], *[c,d]], X, S), psa(X, S).
% ?- #S, zdd_vector_exp(*[*[a,b], *[c,d]], X, S), psa(X, S).
% ?- #S, zdd_vector_exp(*[+[a,b], +[c,d]], X, S), psa(X, S).

zdd_vector_exp(*(X), Y, S):-!, zdd_vector_exp(*, X, Y, S).
zdd_vector_exp(+(X), Y, S):-!, zdd_vector_exp(+, X, Y, S).
%
zdd_vector_exp(*, [], 1, _):-!.
zdd_vector_exp(+, [], 0, _):-!.
zdd_vector_exp(F, [A|As], R, S):-!,
	zdd_vector_exp(F, A, U, S),
	zdd_vector_exp(F, As, V, S),
	apply_binary_basic(F, U, V, R, S).
zdd_vector_exp(_, *(A), V, S):-!, zdd_vector_exp(*, A, V, S).
zdd_vector_exp(_, +(A), V, S):-!, zdd_vector_exp(+, A, V, S).
zdd_vector_exp(_, A, R, S):- zdd_singleton(A, R, S).
%
apply_binary_basic(*, X, Y, Z, S):-!, zdd_merge(X, Y, Z, S).
apply_binary_basic(+, X, Y, Z, S):-!, zdd_join(X, Y, Z, S).

		/***********************************************
		*    Currently reserved names of operations.   *
		***********************************************/
%!	zdd_apply(+E, -V, +S, +M) is det.
%	With context module M and state S, apply E (without evaluate arguments)
%	and unify the value with V.

zdd_apply(atom(A), Y, S, _)	:-!, zdd_singleton(A, Y, S).
zdd_apply(fos(L), Y, S, _)	:-!, zdd_family(L, Y, S).
zdd_apply(set(L), Y, S, _)	:-!, zdd_append(L, 1, Y, S).
zdd_apply(ltr_set(L), Y, S, _)	:-!, ltr_append(L, 1, Y, S).
zdd_apply(dnf(A), X, S, _)	:-!, dnf(0, A, X, S).
zdd_apply(cnf(A), X, S, _)	:-!, cnf(0, A, X, S).
zdd_apply(dnf0(A), X, S, _)	:-!, dnf(1, A, X, S).
zdd_apply(cnf0(A), X, S, _)	:-!, cnf(1, A, X, S).
zdd_apply(gf2(A), X, S, _)	:-!, gf2(A, X, S).
zdd_apply(qp(A), X, S, _)	:-!, mqp(A, X, S).
zdd_apply(fos(A), X, S, _)	:-!, fos(A, X, S).
zdd_apply(X + Y, Z, S, _)	:-!, zdd_join(X, Y, Z, S).
zdd_apply(X * Y, Z, S, _)	:-!, zdd_merge(X, Y, Z, S).
zdd_apply(+(X), Y, S, _)	:-!, zdd_vector_exp(+(X), Y, S).
zdd_apply(*(X), Y, S, _)	:-!, zdd_vector_exp(*(X), Y, S).
zdd_apply(merge(X, Y), Z, S, _)	:-!, zdd_merge(X, Y, Z, S).
zdd_apply(X - Y, Z, S, _)	:-!, zdd_subtr(X, Y, Z, S).
zdd_apply(\(X,Y), Z, S, _)	:-!, zdd_apply(X-Y, Z, S).
zdd_apply((X // Y), Z, S, _):-!, zdd_div(X, Y, Z, S).
zdd_apply((X / Y), Z, S, _)	:-!, zdd_mod(X, Y, Z, S).
zdd_apply((X mod Y), Z, S, _)	:-!, zdd_mod(X, Y, Z, S).
zdd_apply(&(X, Y), Z, S, _)	:-!, zdd_meet(X, Y, Z, S).
zdd_apply(prod(X, Y), Z, S, _)	:-!, zdd_product(X, Y, Z, S).
zdd_apply(**(X, Y), Z, S, _):-!, zdd_product(X, Y, Z, S).
zdd_apply(pow(X), Z, S, _)	:-!, zdd_sort(X, Y, S),	zdd_power(Y, Z, S).
zdd_apply(power(X), Z, S, M):-!, zdd_apply(pow(X), Z, S, M).
zdd_apply(sets(X), Z, S, _)	:-!, sets(X, Z, S).
zdd_apply(cofact(A, L, R), X, S, _)	:-!, cofact(X, t(A, L, R), S).
zdd_apply(cofact(X), Y, S, _)	:-!, integer(X), X>1, cofact(X, Y, S).
zdd_apply(fact(X), Y, S, _)	:-!, integer(X), X>1, cofact(X, Y, S).
zdd_apply(X, Y, S, _)		:- zdd_singleton(X, Y, S).

% ?-zdd((X<<pow([1,2,3,4]), Y<<pow([1,2,3]),
%	card(X, C), card(Y, D), Z<<(X-Y), card(Z, E))).
%
zdd_super_power(X, Y, Z):- shift(zdd_super_power(X, Y, Z)).

zdd_super_power([], P, P, _).
zdd_super_power([A|As], P, Q, S):-
	zdd_super_power(As, P, Q0, S),
	zdd_insert(A, Q0, Q, S).

% ?- zdd((zdd_family([[a],[a,b]], X), card(X, C), psa(X))).
family(X, Y):- @(family(X, Y)).

family(X, Y, S):- zdd_family(X, Y, S).

fos(X, Y, S):- zdd_family(X, Y, S).

%!	zdd_family(+L, -X, +S) is det.
%	For a list L of lists of atoms, build a zdd in S
%	whose index is unified with X.

zdd_family(X, Y):- shift(zdd_family(X, Y)).
%
zdd_family(X, Y, S):- zdd_family(X, 0, Y, S).

% ?- zdd((zdd_family([[a],[a,b]], X), card(X, C), psa(X))).
% ?- zdd((shift(zdd_family([[], [a,a],[a,b]], X)), psa(X))).

zdd_family([], U, U, _).
zdd_family([X|Xs], U, V, S):-
	zdd_insert_atoms(X, 1, X1, S),
	zdd_join(X1, U, U0, S),
	zdd_family(Xs, U0, V, S).

% ?- zdd((ltr_family([[a, b], [c, d]], X), sets(X,S))).
% ?- zdd((ltr_family([[a, b], [c, -c]], X), sets(X,S))).
% ?- zdd((ltr_family([[a,-b, a],[b, a, -b]], X), sets(X,S))).
% ?- zdd((X << cnf((a+(-b)+a)*(b+a+(-b))), S<<sets(X))).
% ?- zdd((X << dnf((a+(-b)+a)), psa(X))).
% ?- zdd((X << dnf((a+(-b)+a)* (b + a+(-b))), psa(X))).
% ?- zdd((X << dnf((A+(-B)+A)* (B + A+(-B))), psa(X))).

% ?- zdd((X << cnf0((a+(-b)+a)*(b+a+(-b))), S<<sets(X))).
% ?- zdd((X << dnf0((a+(-b)+a)), psa(X))).
% ?- zdd((X << dnf0((a+(-b)+a)* (b + a+(-b))), psa(X))).
% ?- zdd((X << dnf0((A+(-B)+A)* (B + A+(-B))), psa(X))).
% ?- zdd((ltr_family([[a],[a,b]], X), @ltr_card(X, C), psa(X))).
% ?- zdd((ltr_family([[], [a],[a,b]], X), @ltr_card(X, C), psa(X))).
% ?- zdd((ltr_family([[a],[-b, -a,-c]], X), psa(X))).
%
ltr_family(X, Y):- shift(ltr_family(X, Y)).
%
ltr_family(X, Y, S):- ltr_family(X, 0, Y, S).

%
ltr_family([], U, U, _).
ltr_family([X|Xs], U, V, S):-
	ltr_append(X, 1, X0, S),
	ltr_join(X0, U, U0, S),
	ltr_family(Xs, U0, V, S).

		/************************
		*    generic choices    *
		************************/

%!	zdd_choices(+X, -Y, +S) is det.
%   Y = { W | U in W <==> all A in U some V in X such that A in X }.
%	where  <==> is for "if and only if".

% ?- zdd(( (X<< set([a,b,c]), zdd_choices(X, Y), psa(Y)))).
% ?- zdd(( (X<< {[a,b,c]}, zdd_choices(X, Y), psa(Y)))).
% ?- zdd(( (X<< {[a, b], [c, d]}), zdd_choices(X, Y), psa(Y))).
% ?- zdd(( (X<< {[], [a, b], [c, d]}), zdd_choices(X, Y), psa(Y))).
% ?- zdd(( (X<< {[a], [c, d]}), zdd_choices(X, Y), psa(Y))).
% ?- zdd(( X<< {[a,b], [c, d], [e,f]}, zdd_choices(X, Y), psa(Y))).
% ?- zdd(( X<< {[a,b]}, zdd_choices(X, Y), psa(Y))).
% ?- zdd(( X<< {[]}, zdd_choices(X, Y), psa(Y))).
% ?- zdd((X<<(set([a,b])+ set([1,2])), zdd_choices(X, Y), psa(Y), card(Y, C))).
% ?- charlist(a,z, AZ), zdd((A<< set(AZ), ltr_choices(A, Y), card(Y, C))).
% ?- zdd((A<< set(!(numlist(1, 2))), psa(A), card(A, C))).
% ?- zdd((A<< set(!charlist(a,z)), B<< set(!numlist(1,10)), X<<(A+B),
%	zdd_choices(X, Y), card(Y, C))).
% ?- zdd((A<< set(!(charlist(a, z))), psa(A), B<< set(:numlist(1,10)), X<<(A+B),
%	zdd_choices(X, Y), card(Y, C))).

zdd_choices(X, Y):- shift(sw_choices(zdd, X, Y)).

zdd_choices(X, Y, S):- sw_choices(zdd, X, Y, S).
%
sw_choices(SW, X, Y):- shift(sw_choices(SW, X, Y)).

%!	sw_choices(+X, -Y, +S).
%	Y is the set of all choices of X,
%	where X is a family of sets.
sw_choices(_, 1, 0, _):-!.
sw_choices(_, 0, 1, _):-!.
sw_choices(SW, X, Y, S):- memo(sw_choices(SW, X)-Y, S),
	(	nonvar(Y)	->	true
	;	sw_choices(SW, X, [], 1, Y, S)
	).
%
sw_choices(_, 0, _, Y, Y, _):-!.
sw_choices(SW, 1, As, Y0, Y, S):-!, sw_fold_insert(SW, As, Y0, 0, Y, S).
sw_choices(SW, X, As, Y0, Y, S):- cofact(X, t(A, L, R), S),
	(	L = 0
	->	sw_choices(SW, R, [A|As], Y0, Y, S)
	; 	L = 1
	->	(	As = []
		->  Y = 0
		;	sw_fold_insert(SW, As, Y0, 0, Y1, S),
			sw_choices(SW, R, [A|As], Y1, Y, S)
		)
	;	sw_choices(SW, L, As, Y0, Y1, S),
		sw_choices(SW, R, [A|As], Y1, Y, S)
	).
%
sw_fold_insert(_, [], _, Y, Y, _).
sw_fold_insert(SW, [A|As], Y0, U, Y, S):-
	(	SW == zdd ->
		zdd_insert(A, Y0, V, S),
		zdd_join(U, V, U0, S)
	;	ltr_insert(A, Y0, V, S),
		ltr_join(U, V, U0, S)
	),
	sw_fold_insert(SW, As, Y0, U0, Y, S).


		/*************************************************
		*   Operation on zdd join/merge/meet/subtr/cons  *
		*************************************************/

%!	zdd_join(+X, +Y, -Z, +S) is det.
%	Z is the union of X and Y.

zdd_join(X, Y, Z):- shift(zdd_join(X, Y, Z)).
%
zdd_join(0, X, X, _):-!.
zdd_join(X, 0, X, _):-!.
zdd_join(X, X, X, _):-!.
zdd_join(1, X, Y, S):-!, zdd_join_1(X, Y, S).
zdd_join(X, 1, Y, S):-!, zdd_join_1(X, Y, S).
zdd_join(X, Y, Z, S):-
	(	X@<Y -> memo(zdd_join(X,Y)-Z, S)
	;	memo(zdd_join(Y, X)-Z, S)
	),
	(	nonvar(Z) -> true  %, write(.)
	; 	cofact(X, t(A, L, R), S),
		cofact(Y, t(A0, L0, R0), S),
		zdd_compare(C, A, A0, S),
		(	C = (<)	->
 			zdd_join(L, Y, U, S),
			cofact(Z, t(A, U, R), S)
		;	C = (=)	->
			zdd_join(L, L0, U, S),
			zdd_join(R, R0, V, S),
			cofact(Z, t(A, U, V), S)
		;	zdd_join(L0, X, U, S),
			cofact(Z, t(A0, U, R0), S)
		)
	).

%!  zdd_cons(+X, +Y, -Z, +S) is det.
%	Short hand for cofact(Z, t(X, 0, Y), S).
%	Having analogy Z = [X|Y] in mind.

% ?- #(S), zdd_cons(X, a, 1, S), zdd_cons(X, A, B, S).
% ?- #(S), zdd_cons(X, a, 0, S), zdd_cons(X, A, B, S). % false
zdd_cons(X, Y, Z, S):- ( var(X);  X>1 ), !, cofact(X, t(Y, 0, Z), S).

% ?- zdd((X<<(*[a]+ *[b]+ []), psa(X))).  % false for [].
% ?- zdd((X<<(*[a]+ *[b]+ 1), psa(X))).
% ?- zdd((X<<(*[a]+ *[b]+ 0), psa(X))).
% ?- zdd((X<<(*[a]+ *[b]), psa(X))).
% ?- zdd((A<<{[a]},  psa(A), zdd_join_1(A, X), psa(X))).

%	zdd_join_1(+X, -Y, +G) is det.
%	Y is the union of X and 1 (the singleton {{}}).
zdd_join_1(X, Y):- shift(zdd_join_1(X, Y)).
%
zdd_join_1(X, Y, S):-
	(	X<2  -> Y = 1
	; 	cofact(X, t(A, L, R), S),
		zdd_join_1(L, L0, S),
		cofact(Y, t(A, L0, R), S)
	).

% ?- zdd((X<< *[a, b], Y<< *[a, b], Z << (X*Y), sets(Z, S))).
% ?- zdd((X<< *[a, a], Y<< *[a, b], Z << (X*Y), sets(Z, S))).
% ?- zdd((X<< *[a, b], Y<< *[a, b], Z << (X&Y), sets(Z, S))).
% ?- zdd((X<<(*[a, b]+ *[c,d]), Y<<( *[x,y]+ *[u,v]), Z << (X&Y), sets(Z, S))).
% ?- trace, zdd(X=[b, a]).
% ?- zdd((X<<(*[a, b]+ *[c,d]), Y<<(*[x,y]+ *[u,v]), Z << (X**Y), sets(Z, S), card(Z, C))).
% ?- zdd((X<< *[b, a], Y<< *[a, b], Z <<(X+Y), sets(Z, S))).

%!	zdd_meet(+X, +Y, -Z, +G) is det.
%	Z is the intersection of X and Y.

zdd_meet(X, Y, Z):- shift(zdd_meet(X, Y, Z)).

%
zdd_meet(0, _, 0, _):-!.
zdd_meet(_, 0, 0, _):-!.
zdd_meet(1, X, Y, S):-!, zdd_meet_1(X, Y, S).
zdd_meet(X, 1, Y, S):-!, zdd_meet_1(X, Y, S).
zdd_meet(X, X, X, _):-!.
zdd_meet(X, Y, Z, S):-
	(	X @< Y -> memo(zdd_meet(X,Y)-Z, S)
	;	memo(zdd_meet(Y,X)-Z, S)
	),
	(	nonvar(Z) -> true
	;	cofact(X, t(A, L, R), S),
		cofact(Y, t(A0, L0, R0), S),
		zdd_compare(C, A, A0, S),
		(	C = (<) -> zdd_meet(L, Y, Z, S)
		;	C = (=) ->
			zdd_meet(L, L0, U, S),
			zdd_meet(R, R0, V, S),
			cofact(Z, t(A, U, V), S)
		;	zdd_meet(L0, X, Z, S)
		)
	).

%!	zdd_meet_1(+X, -Y, +G) is semidet.
%	Y is the intersection of X and 1 {the singleton of the emptyset {{}}}.
%
zdd_meet_1(X, 1, S):- zdd_has_1(X, S), !.
zdd_meet_1(_, 0, _).

%!	zdd_subtr(+X, +Y, -Z, +G) is det.
%	Z is the subtraction of Y from X: Z = Y \ Z.

% ?- zdd((X<<{[a,b],[b,c]}, Y<<{[b,c]}, psa(Y), zdd_subtr(X, Y, Z), psa(Z))).
% ?- zdd0((X<<pow([a,b,c]), Y<<pow([a,b]), zdd_subtr(X, Y, Z), card(Z, C))).

zdd_subtr(X, Y, Z):-shift(zdd_subtr(X, Y, Z)).
%
zdd_subtr(X, X, 0, _):-!.
zdd_subtr(0, _, 0, _):-!.
zdd_subtr(X, 0, X, _):-!.
zdd_subtr(X, 1, Y, S):-!, zdd_subtr_1(X, Y, S).
zdd_subtr(1, X, Y, S):-!,
	(	zdd_has_1(X, S) ->	Y = 0
	;	Y = 1
	).
zdd_subtr(X, Y, Z, S):- memo(zdd_subtr(X,Y)-Z, S),
	(	nonvar(Z) -> true
	; 	cofact(X, t(A, L, R), S),
		cofact(Y, t(A0, L0, R0), S),
		zdd_compare(C, A, A0, S),
		(	C = (<)	->
			zdd_subtr(L, Y, U, S),
			cofact(Z, t(A, U, R), S)
		;	C = (=) ->
			zdd_subtr(L, L0, U, S),
			zdd_subtr(R, R0, V, S),
			cofact(Z, t(A, U, V), S)
		;	zdd_subtr(X, L0, Z, S)
		)
	).

%!	zdd_subtr_1(+X, -Y, +G) is det.
%	Y is the subtraction of 1 (the set {{}}) from X: Y = X \ {{}}.

% ?- zdd((X<<([]+[a]), zdd_subtr_1(X, Y), psa(Y))).
% ?- zdd((X<<{[a], [a,b,c]}, zdd_subtr_1(X, Y), psa(Y))).
% ?- zdd((X<<{[a], [a,b,c], []}, zdd_subtr_1(X, Y), psa(Y))).
zdd_subtr_1(X, Y):-shift(zdd_subtr_1(X, Y)).
%
zdd_subtr_1(X, 0, _):- X < 2, !. % empty family or singleton of the empty.
zdd_subtr_1(X, Y, S):- cofact(X, t(A, L, R), S),
	zdd_subtr_1(L, L0, S),
	cofact(Y, t(A, L0, R), S).

%!	zdd_merge(X, Y, Z, G) is det.
%
%	Z is the merge of X and Y, that is,
%   Z = { U | U is the union of A in X and B in Y}.
%
%	Remark. The merge is associative and commutative, but not idempotent.

% ?- zdd(( X<<{[a], [b]}, Y<<{[c]}, zdd_merge(X, Y, Z), psa(Z))).
% ?- time(zdd(( X<<pow(@charlist($(a-z))), Y<<pow(@numlist(1, 26)),
%	zdd_merge(X, Y, Z), U <<  zdd_meet(X, Z),  card(X, C)))), C is 2^26.
zdd_merge(X, Y, Z)	:- shift(zdd_merge(X, Y, Z)).
%
zdd_merge(0, _, 0, _):-!.
zdd_merge(_, 0, 0, _):-!.
zdd_merge(1, X, X, _):-!.
zdd_merge(X, 1, X, _):-!.
zdd_merge(X, Y, Z, S):-
	(	X > Y -> memo(zdd_merge(Y, X)-Z, S)
	;	memo(zdd_merge(X, Y)-Z, S)
	),
	(	nonvar(Z) -> true
	;	cofact(X, t(A, L, R), S),
		zdd_merge(L, Y, L0, S),
		zdd_merge(R, Y, R0, S),
		zdd_insert(A, R0, R1, S),
		zdd_join(L0, R1, Z, S)
	).
%
zdd_merge_list(X, Y):- shift(zdd_merge_list(X, 1, Y)).
%
zdd_merge_list(X, Y, Z):- shift(zdd_merge_list(X, Y, Z)).
%
zdd_merge_list([], X, X, _).
zdd_merge_list([A|As], X, Y, S):-
	(	integer(A) -> zdd_merge(A, X, X0, S)
	;	zdd_insert(A, X, X0, S)					% atom case
	),
	zdd_merge_list(As, X0, Y, S).

% ?- zdd((X<< {[a,b], [a,c]}, Y<<{[a,b]}, zdd_subfamily(Y, X))).
% ?- zdd((X<< {[a,b], [a,c]}, Y<<{[a,b]}, zdd_subfamily(X, Y))). % false
% ?- zdd((X<< {[a,b], [a,c], [a]},  Y<<{[a,b], [a]}, zdd_subfamily(Y, X))).
% ?- zdd((X<< {[a,b],  [a,c]}, Y<<{[a,b], [a]}, zdd_subfamily(Y, X))). % false
% ?- zdd((X<< pow([a,b]), Y<<pow([a,b,c]), zdd_subfamily(X, Y))).
% ?- zdd((X<< pow([a,b]), Y<<pow([a,b,c]), zdd_subfamily(Y, X))). % false

%!	zdd_subfamily(+X, +Y, +S) is det.
% True if a ZDD X is a subfamily of a ZDD Y.

zdd_subfamily(X, Y):- shift(zdd_subfamily(X, Y)).
%
zdd_subfamily(X, X, _):-!.
zdd_subfamily(0, _, _):-!.
zdd_subfamily(_, 0, _):-!, fail.
zdd_subfamily(1, X, S):-!, zdd_has_1(X, S).
zdd_subfamily(X, X0, S):-
		cofact(X, t(A, L, R), S),
		cofact(X0, t(A0, L0, R0), S),
		zdd_compare(C, A, A0, S),
		(	C = (=) ->
			zdd_subfamily(L, L0, S),
			zdd_subfamily(R, R0, S)
		;	C = (>),
			zdd_subfamily(X, L0, S)
		).

		/*********************************
		*     division/residue  in ZDD   *
		*********************************/

%!	zdd_divmod(X, Y, D, M, S) is det.
%	D = { A-B | A in X, B is Y, B is a subset of A }.
%	M = { A in X | forall B in Y B is not a subset of A }.
% ?- #S, X<< +[*[a,b], [c]], Y<< +[b], zdd_divmod(X, Y, D, M, S),
%  psa(X, S), psa(D, S), psa(M, S).

zdd_divmod(X, Y, D, M, S):-
	zdd_div_div(X, Y, D1, D, S),
	zdd_subtr(X, D1, M, S).

%!	zdd_divmod0(X, Y, D, D0, S) is det.
%	D = { A in X | A is a subset of B for some B in Y},
%   D0 = { A-B | A in X, B in Y, B is a subset of A}.

% ?- #S, X<< +[*[a,b], [c]], Y<< +[b], zdd_divmod0(X, Y, D, M, S),
%  psa(X, S), psa(D, S), psa(M, S).
zdd_divmod0(X, Y, D, M, S):-
	zdd_divisible(X, Y, D, S),
	zdd_subtr(X, D, M, S).

%!	zdd_div_div(X, Y, D, D0, S) is det.
%	D is  the set of elements in X  divisible by an element of Y,
%   D0 is the set A-B (=  A//B) s.t. A in X, B in Y and  B is subset of A.

% ?- #S, X<< +[b], zdd_div_div(X, 1, D, M, S),
%  psa(X, S), psa(D, S), psa(M, S).
% ?- #S, X<< +[b], Y<< +[b], zdd_div_div(X, Y, D, M, S),
%  psa(X, S), psa(Y, S), psa(D, S), psa(M, S).
% ?- #S, X<< +[b], Y<< +[b], zdd_div_div(X, Y, D, M, S),
%  psa(X, S), psa(Y, S), psa(D, S), psa(M, S).
% ?- #S, X<< +[*[a,b]], Y<< +[*[b]], psa(X, S), psa(Y, S), zdd_div_div(X, Y, D, M, S), psa(D, S), psa(M, S).
% ?- #S, X<< +[*[a,b], *[c,d]], Y<< +[*[b], *[d]],
%	zdd_div_div(X, Y, D, D1, S),
%	psa(D, S), psa(D1, S).

zdd_div_div(0, _, 0, 0, _):-!.
zdd_div_div(1, X, Y, Y, S):-!,
	(	zdd_has_1(X, S) ->	Y = 1
	;	Y = 0
	).
zdd_div_div(_, 0, 0, 0, _):-!.
zdd_div_div(X, 1, X, X, _):-!.
zdd_div_div(X, Y, Z, U, S):- memo(div_div(X, Y)- P, S),
	(	nonvar(P) -> P = (Z, U)
	;	cofact(X, t(A, L, R), S),
		zdd_div_div(L, Y, L0, U0, S),
		zdd_div_div(A, R, Y, R0, V0, S),
		zdd_join(L0, R0, Z, S),
		zdd_join(U0, V0, U, S),
		P = (Z, U)
	).
%
zdd_div_div(_, 0, _, 0, 0, _):-!.
zdd_div_div(_, 1, 0, 0, 0, _):-!.  % ???
zdd_div_div(A, 1, 1, Z, Z, S):-!, zdd_singleton(A, Z, S).
zdd_div_div(_, _, 0, 0, 0, _):-!.
zdd_div_div(A, X, 1, Z, Z, S):-!, cofact(Z, t(A, 0, X), S).
zdd_div_div(A, X, Y, Z, U, S):- cofact(Y, t(B, L, R), S),
		zdd_compare(C, A, B, S),
		(	C = (<) ->
			zdd_div_div(X, Y, Z0, U0, S),
			zdd_insert(A, Z0, Z, S),
			zdd_insert(A, U0, U, S)
		;	C = (=)	->
			zdd_div_div(X, L, Z0, U0, S),
			zdd_insert(A, Z0, Z1, S),
			zdd_insert(A, U0, U1, S),
			zdd_div_div(X, R, Z2, U2, S),
			zdd_insert(A, Z2, Z3, S),       % A is absorbed due to hit (A=B).
			zdd_join(Z1, Z3, Z, S),
			zdd_join(U1, U2, U, S)
		; 	zdd_div_div(A, X, L, Z, U, S)
		).

%!	zdd_div(+X, +Y, -Z, +S) is det.
%	Z is the quotient of X divided by Y;
%	Z = { A - B | A in X, B in Y, B is a subset of A }

% ?- zdd(( X<< {[a]}, zdd_div(X, X, Z), psa(Z))).
% ?- zdd(( X<< 0, Y<<{[a]}, zdd_div(X, Y, Z), psa(Z))).
% ?- zdd(( X<< 0, Y<< 0, zdd_div(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b]}, Y<<{[a]}, zdd_div(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a]}, Y<<{[a]}, zdd_div(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a,b]}, Y<<{[a]}, zdd_div(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b]}, Y<<1, zdd_div(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a]}, Y<<{[a]}, zdd_div(X, Y, Z), psa(Z))).
% ?- zdd(( X<<1, Y<<1, zdd_div(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b]}, Y<<{[a]}, zdd_div(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b, c], [b, c]}, Y<<{[a],[b]}, zdd_div(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b]}, Y<<{[a, b, c]}, zdd_div(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b], [a, c]}, Y<<{[a]}, zdd_div(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b, c], [a, c]}, Y<<{[a, b]}, zdd_div(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b, c], [a, c]}, Y<<{[a, b], [a, c]}, zdd_div(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b, c], [a, c]}, Y<<{[a, b]}, zdd_div(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[c]}, Y<<{[a, c]}, zdd_div(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a]}, Y<<{[a, b]}, zdd_div(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a,b]}, Y<<{[a, b, c]}, zdd_div(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a,b,c]}, Y<<{[a,b], [c]}, zdd_div(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b], [a, c]}, Y<<{[a], [b]}, zdd_div(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b], [a, c]}, Y<<{[b]}, zdd_div(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b], [a, c]}, Y<<{[b]}, zdd_div(X, Y, Z), psa(Z))).

zdd_div(X, Y, Z)	:- shift(zdd_div(X, Y, Z)).

%
zdd_div(0, _, 0, _):-!.
zdd_div(1, X, Y, S):-!,
	(	zdd_has_1(X, S) ->	Y = 1
	;	Y = 0
	).
zdd_div(X, Y, Z, S):- memo(zdd_div(X, Y)- Z, S),
	(	nonvar(Z) -> true
	;	cofact(X, t(A, L, R), S),
		zdd_div(L, Y, L0, S),
		zdd_div(A, R, Y, R0, S),
		zdd_join(L0, R0, Z, S)
	).
%
zdd_div(_, _, 0, 0, _):-!.
zdd_div(A, X, 1, Z, S):-!, cofact(Z, t(A, 0, X), S).
zdd_div(A, X, Y, Z, S):- cofact(Y, t(B, L, R), S),
		zdd_compare(C, A, B, S),
		(	C = (<) ->
			zdd_div(X, Y, Z0, S),
			cofact(Z, t(A, 0, Z0), S)
		;	C = (=)	->
			zdd_div(X, L, L0, S),
			zdd_div(X, R, R0, S),
			cofact(Z, t(A, R0, L0), S)
		;	zdd_div(A, X, L, Z, S)
		).


%!	zdd_divisible(+X, +Y, -Z, +S) is det.
%	Y = { A in X | for some B in Y, B is a subset of A }.

% ?- #S, N=10, numlist(1, N, Ns), numlist(1, 10, Js), X<<pow(Ns), Y<<{Js}, zdd_divisible(X, Y, Z, S).
% ?- zdd(( X<<{[a, c], [b]}, Y<<{[b]}, @zdd_divisible(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, b, c], [b]}, Y<<{[b]}, @zdd_divisible(X, Y, Z), psa(Z))).
% ?- zdd(( X<<{[a, c], [a]}, Y<<{[b]}, @zdd_divisible(X, Y, Z), psa(Z))).
% ?- zdd(( X<<pow([a, b,  c]), Y<<{[a], [b]}, @zdd_divisible(X, Y, Z), psa(Z))).

zdd_divisible(0, _, 0, _):-!.
zdd_divisible(1, X, Y, S):-!,
	(	zdd_has_1(X, S) ->	Y = 1
	;	Y = 0
	).
zdd_divisible(X, Y, Z, S):- memo(zdd_divisible(X, Y)- Z, S),
	(	nonvar(Z) -> true
	;	cofact(X, t(A, L, R), S),
		zdd_divisible(L, Y, L0, S),
		zdd_divisible(A, R, Y, R0, S),
		zdd_join(L0, R0, Z, S)
	).
%
zdd_divisible(_, _, 0, 0, _):-!.
zdd_divisible(A, X, 1, Z, S):-!, cofact(Z, t(A, 0, X), S).
zdd_divisible(A, X, Y, Z, S):- cofact(Y, t(B, L, R), S),
		zdd_compare(C, A, B, S),
		(	C = (<) ->
			zdd_divisible(X, Y, Z0, S),
			cofact(Z, t(A, 0, Z0), S)
		;	C = (=)	->
			zdd_divisible(X, L, L0, S),
			zdd_divisible(X, R, R0, S),
			zdd_join(L0, R0, Z0, S),
			cofact(Z, t(A, 0, Z0), S)
		;	zdd_divisible(A, X, L, Z, S)
		).


%! zdd_divisible_by_all(+X, +Y, -Z, +S) is det.
%	Z = { A in X | forall B in Y  B is a subset of A }.

% ?- #S, X<< +[*[a]], Y<< +[*[a]], zdd_divisible_by_all(X, Y, Z, S).
% ?- #S, X<< +[*[b]], Y<< +[*[a]], zdd_divisible_by_all(X, Y, Z, S).
% ?- #S, X<< +[*[a,b], *[c,d]], Y<< +[*[d]], zdd_divisible_by_all(X, Y, Z, S).
% ?- #S, X<< +[*[a,b]], Y<< +[*[c]], zdd_divisible_by_all(X, Y, Z, S).S).
% ?- N = 300, #S, numlist(1, N, Ns), X<<pow(Ns), Y << +[*Ns],
%	time(zdd_divisible_by_all(X, Y, Z, S)), psa(Z, S).

% ?- #S, X<< +[*[a,b], *[c,d], *[a, d]],  Y<< +[*[a], *[d]], zdd_divisible_by_all(X, Y, Z, S), psa(Z, S).
zdd_divisible_by_all(X, Y, X, _):-Y<2, !.    % It was hard to find this.
zdd_divisible_by_all(X, _, 0, _):-X<2, !.    % It was a little bit hard to find this.
zdd_divisible_by_all(X, Y, Z, S):- memo(div_by_all(X, Y)-Z, S),
	(	nonvar(Z) -> true
	; 	cofact(X, t(A, L, R), S),
		zdd_divisible_by_all(L, Y, L0, S),
		zdd_divisible_by_all(A, R, Y, R0, S),
		zdd_join(L0, R0, Z, S)
	).
%
zdd_divisible_by_all(A, X, Y, Z, S):- Y<2, !,
	cofact(Z, t(A, 0, X), S).
zdd_divisible_by_all(A, X, Y, Z, S):- cofact(Y, t(B, L, R), S),
	zdd_compare(C, A, B, S),
	(	C = (<) ->
		zdd_divisible_by_all(X, Y, Z0, S),
		zdd_insert(A, Z0, Z, S)
	;	C = (=) ->
		zdd_divisible_by_all(X, R, R1, S),
		zdd_insert(A, R1, R0, S),
		zdd_divisible_by_all(R0, L, Z, S)
	;	Z = 0				% It was hard to find this.
	).

%!	zdd_mod(+X, +Y, -Z, +S) is det.
%	Z = { A in X | forall B in Y B is not a subset of A }.

zdd_mod(X, Y, Z) :- shift(zdd_mod(X, Y, Z)).
%
zdd_mod(X, Y, Z, S)	:- zdd_divisible(Y, X, X0, S),
	zdd_subtr(X, X0, Z, S).

%!	zdd_multiple(+X, +Y, -Z, +S) is det.
%	Z = { B in Y | exists A in X  A is a subset of B}.

% ?- #S, zdd_list(X, [[a]], S), zdd_list(Y, [[a,c],[a,d]], S),
% zdd_multiple(X, Y, Z, S), psa(Z, S).
% ?- #S, zdd_list(X, [[a,b]], S), zdd_list(Y, [[a,c],[a,d]], S),
% zdd_multiple(X, Y, Z, S), psa(Z, S).
% ?- #S, zdd_list(X, [[a,b]], S), Y<<pow([a,b,c]),
% zdd_multiple(X, Y, Z, S), psa(Z, S).
% ?- #S, Y<<pow([a,b,c]), zdd_multiple(0, Y, Z, S), psa(Z, S).
% ?- #S, Y<<pow([a,b]), zdd_multiple(1, Y, Z, S), psa(Z, S).

zdd_multiple(0, _, 0, _):-!.
zdd_multiple(1, Y, Y, _):-!.
zdd_multiple(_, Y, 0, _):-Y<2, !.
zdd_multiple(X, Y, Z, S):-memo(multiple(X, Y)-Z, S),
	(	nonvar(Z) -> true
	;	cofact(X, t(A, L, R), S),
		zdd_multiple(L, Y, L0, S),
		zdd_multiple(A, R, Y, R0, S),
		zdd_join(L0, R0, Z, S)
	).
%
zdd_multiple(_, _, Y, 0, _):- Y<2, !.
zdd_multiple(A, X, Y, Z, S):- cofact(Y, t(B, L, R), S),
	zdd_multiple(A, X, L, L0, S),
	zdd_compare(C, A, B, S),
	(	C = (<) -> R0 = 0
	;	C = (=) ->
		zdd_multiple(X, R, R1, S),
		zdd_insert(A, R1, R0, S)
	;	zdd_multiple(A, X, R, R1, S),
		zdd_insert(B, R1, R0, S)
	),
	zdd_join(L0, R0, Z, S).

		/*********************************
		*     division/resudue in List   *
		*********************************/

%! zdd_mod_by_list(+X, +Y, -Z, +S) is det.
%	Equivalent to
%	zdd_list(Y, U, Z), zdd_mod_by_list(X, U, Z, S).

% ?- zdd((X<<pow([a,b]), shift(zdd_mod_by_list(X, [a], Y)), psa(Y))).
% ?- zdd((X<<pow([a,b]), shift(zdd_mod_by_list(X, [], Y)), psa(Y))).
% ?- zdd((X<<pow([a,b]), shift(zdd_mod_by_list(X, [a, b], Y)), psa(Y))).
% ?- zdd((X<<pow([a,b,c]), shift(zdd_mod_by_list(X, [b,c], Y)), psa(Y))).
% ?- zdd((X<<pow([a,b,c]), shift(zdd_mod_by_list(X, [a,c], Y)), psa(Y))).
% ?- zdd((X<<pow([a]), shift(zdd_mod_by_list(X, [c], Y)), psa(Y))).
% ?- N=100, numlist(1, N, Ns), numlist(1, N, Js), zdd((X<<pow(Ns),  shift(zdd_mod_by_list(X, Js, Z)), card(Z, C))).

zdd_mod_by_list(X, Y, Z, S):- zdd_divisible_by_list(X, Y, Z0, S),
	zdd_subtr(X, Z0, Z, S).

%!	zdd_div_by_list(+X, +Y, -Z, +S) is det.
%	Eqivalent to
%	zdd_sets(U, [X], Z, S), zdd_div(Y, U, Z, S).

% ?- zdd((X<<pow([a,b,c]), @(zdd_div_by_list(X, [b,c], Y)), psa(Y))).
% ?- zdd((X<<pow([a,b,c,d]), @(zdd_div_by_list(X, [b,c], Y)), psa(Y))).

zdd_div_by_list(X, [], X, _):-!.
zdd_div_by_list(X, F, Y, S):- F=[A|As],
 	(	X<2 -> Y = 0
	;	cofact(X, t(B, L, R), S),
		zdd_div_by_list(L, F, L0, S),
		zdd_compare(C, B, A, S),
		(	C = (>) -> R0 = 0
		;	C = (=) ->
			zdd_div_by_list(R, As, R0, S)
		;	zdd_div_by_list(R, F, R1, S),
			zdd_insert(B, R1, R0, S)
		),
		zdd_join(L0, R0, Y, S)
	).

%!	zdd_divisible_by_list(X, Y, Z, S) is det.
%	Equivalent to
%	zdd_sets(U, [X], S), zdd_divisible(Y, U, Z, S).

% ?- zdd((X<<pow([a,b,c]), @(zdd_divisible_by_list(X, [b,c], Y)), psa(Y))).
% ?- zdd((X<<pow([a,b,c,d]), @(zdd_divisible_by_list(X, [b,c], Y)), psa(Y))).
zdd_divisible_by_list(X, [], X, _):-!.
zdd_divisible_by_list(X, F, Y, S):- F=[A|As],
 	(	X<2 -> Y = 0
	;	cofact(X, t(B, L, R), S),
		zdd_divisible_by_list(L, F, L0, S),
		zdd_compare(C, B, A, S),
		(	C = (>) ->	R0 = 0
		;	C = (=) -> 	zdd_divisible_by_list(R, As, R0, S)
		;	zdd_divisible_by_list(R, F, R0, S)
		),
		cofact(Y, t(B, L0, R0), S)

	).

		/*********************
		*     meta on ZDD    *
		*********************/

% ?- zdd((X<< {[1],[2]}, map_zdd(pred([A, B]:-plus(A,B,0)), X, Y), sets(Y, S))).
% ?- zdd((X<<pow([1,2,4,5,6]), map_zdd(pred([A, B]:-plus(A,B,0)), X, Y), sets(Y, S))).
% ?- N=2, numlist(1, N, L), zdd((X<<pow(L), map_zdd(pred([A, B]:-plus(A,B,0)), X, Y), sets(Y, S))).
% ?- N=10, numlist(1, N, L), zdd((X<<pow(L),
%	map_zdd(pred([A, B]:-plus(A,B,0)), X, Y), sets(Y, S))).
% ?- N=10, numlist(1, N, L), zdd((X<<pow(L),
%	map_zdd(pred([A, B]:- B is A + 1), X, Y), card(Y, C))), C=:= 2^N.
% ?- N=1000, numlist(1, N, L), zdd((X<<pow(L),
%	map_zdd(pred([A, B]:- B is A + 1), X, Y), card(Y, C))), C=:= 2^N.

%! map_zdd(+F, +X, -Y, +G) is det.
%  Y is the set of all F_S with S running in X,
%  where f_S = { F(a) | a in S}.
%
%  map_zdd(F, X, Y)  works roughly like
%  maplist(maplist(F), X, Y)) with a list X of lists.

:- meta_predicate map_zdd(2,?,?).
map_zdd(F, X, Y):- shift(map_zdd(F, X, Y)).
:- meta_predicate map_zdd(2,?,?,?).

map_zdd(_, X, X, _):- X<2, !.
map_zdd(F, X, Y, S):- memo(map(X,F)-Y, S),
	(	nonvar(Y)-> true
	;	cofact(X, t(A, L, R), S),
		map_zdd(F, L, L0, S),
		map_zdd(F, R, R0, S),
		call(F, A, B),
		zdd_insert(B, R0, R1, S),
		zdd_join(L0, R1, Y, S)
	).

% ?- zdd((memo(a-1), set_memo(a-2), memo(a-V))).
% ?- zdd((X<<pow([a,b,c,d]), prz(X))).
% ?- zdd((X<<pow([a,b,c]), psa(X))).
% ?- zdd((prz(1))).
% ?- zdd((prz(0))).
% ?- zdd((mp(hello, 1))).
% ?- zdd((mp(hello, 0))).

%!	prz(Z) is det.
%	Print all sets of Z.
prz(X):- print_set_at(X).

%
psa(X):- @(psa(X)).

%!	psa(+X, +S) is det.
%	print all sets of X
psa(X, S):- print_set_at(X, S),
	writeln('-------------------').
%
mp(M, X):- shift(mp(M, X)).

%!	mp(+M, +X, +S) is det.
%	print all sets of X with message M.
mp(M, X, S):- format("~w\n", [M]), print_set_at(X, S).

%!	print_set_at(+X, _S) is det.
%	Pprint all sets in X.

print_set_at(X):- shift(print_set_at(X)).
%
print_set_at(0, _):-!, format("\szid = 0\n\t0\n").
print_set_at(1, _):-!, format("\szid = 1\n\t[]\n").
print_set_at(X, S):- format("\szid = ~w\n", [X]),
	forall(set_at(P, X, S), format("\t~w\n", [P])).

% ?- zdd((X<<(set([1-2, 2-3, 3-4]) + set([1-2,2-3])), set_at(U, X))).
% ?- zdd((X<<pow([1,2]), set_at(U, X))).
% ?- zdd((X<<pow([1,2]), mp(powerset, X))).

% set_at(P, X):- shift(set_at_(P, X)).
set_at(P, X):- shift(set_at(P, X)).
%
set_at([], 1, _):-!.
set_at(P, X, S):- X>1,
	cofact(X, t(A, L, R), S),
	(	set_at(P, L, S)
	;	set_at(As, R, S),
		P=[A|As]
	).

%!	zdd_list(?X, ?Y, +S) is det.
%	Two way converter between zdd and list.
%	Y is the list of zdd X.

% ?- zdd((X<<pow([a,b,c]), zdd_list(X, Y),  zdd_list(X0, Y))).
% ?- powerset([a,b,c], Y), zdd((zdd_list(X, Y),  zdd_list(X, Y0), zdd_list(X0, Y0))),
%  maplist(sort, Y, Y1),  sort(Y1, Y2).

zdd_list(X, Y):- shift(zdd_list(X, Y)).
%
zdd_list(X, Y, S):- nonvar(X), !, zdd_sets(X, Y, S).
zdd_list(X, Y, S):- zdd_family(Y, X, S).

% ?- zdd((X<<pow([a,b,c]), sets(X, P))).
% ?- zdd((set_compare(opp_compare), X<<pow([a,b,c]), sets(X, Y))).
% ?- zdd((X<<pow([]), sets(X, P))).
% ?- zdd((X<<pow([a]), sets(X, P))).
% ?- zdd((X<<pow(@numlist(1,3)), sets(X, Y))).
% ?- zdd((X<<pow([a,b,c]), sets(X, Y))).
% ?- zdd((X<<pow(@charlist($a, $c)), sets(X, Y))).
% ?- zdd((X<<(pow(@charlist($a, $c)) + pow(@numlist(1, 3))),  sets(X, Y))).

%! sets(+X, -Y, +S)
%  Y is the list in X.

sets(X, P):- shift(sets(X, P)).
%
sets(X, Y, S):- zdd_sets(X, Y0, S), zdd_sort(Y0, Y, S).
%
zdd_sets(1, [[]], _):-!.
zdd_sets(0, [], _):-!.
zdd_sets(X, P, S):-
	cofact(X, t(A,L,R), S),
	zdd_sets(L, Y, S),
	zdd_sets(R, Z, S),
	maplist(cons(A), Z, AZ),
	append(AZ, Y, P).

% ?- open_state(S), zeval(pow([a, b, c, d, e, f]), I, S), eqns(I, X, S),
%	maplist(writeln, X).
eqns(X, P):- shift(eqns(X, P)).
%
eqns(X, Y, S):- zdd_eqns(X, Y, S).

basic_zdd(0).
basic_zdd(1).

% ?- open_state(S), zeval(pow(@numlist(2, 100)), X, S), zdd_eqns(X, Es, S),
%	zdd_eqns(J, Es, S).
% ?- open_state(S), zeval(pow(@charlist($(a-z))), X, S), zdd_eqns(X, Es, S),
%	zdd_eqns(J, Es, S).
zdd_eqns(I, Es, S):- nonvar(I), var(Es), !,
	zdd_to_eqns(I, Es, [], S).
zdd_eqns(I, Es, S):- var(I), nonvar(Es), !,
	(	check_eqns_for_zdd(Es, S)
	->	Es =  [X=_|_],
		eqns_to_zdd(Es, X, I, S)
	;	throw('Violating-underlying-ordering-on-atoms.'(Es))
	).
%
zdd_to_eqns(I, X, S):- zdd_to_eqns(I, X, [], S),
	zdd_sort(X, X1, S),
	reverse(X1, X).
%
zdd_to_eqns(I, X, X, _):- I<2, !.
zdd_to_eqns(I, X, Y, S):- memo(visited(I)-F, S),
	(	nonvar(F) 	->	Y = X
	;	F = true,
		cofact(I, t(A, L, R), S),
		X = [I=t(A, L, R)|X0],
		zdd_to_eqns(L, X0, X1, S),
		zdd_to_eqns(R, X1, Y, S)
	).
%
check_eqns_for_zdd(Es, S):- are_eqns_closed(Es, Es, S),
	sort(Es, Fs),
	has_unique_left(Fs),
	sort(2, @=<, Es, Gs),
	has_unique_right(Gs).
%
are_eqns_closed([], _, _).
are_eqns_closed([_ = t(A, L, R)|Rest], Es, S):-
	(	basic_zdd(L) -> true
	; 	memberchk(L = t(B, _, _), Es),
		zdd_compare(<, A, B, S)
	),
	(	basic_zdd(R) -> true
	; 	memberchk(L = t(C, _, _), Es),
		zdd_compare(<, A, C, S)
	),
	are_eqns_closed(Rest, Es, S).
%
eqns_to_zdd(_, 0, 0, _):-!.
eqns_to_zdd(_, 1, 1, _):-!.
eqns_to_zdd(Es, X, I, S):- memo(solve(X)-I, S),
	(	nonvar(I) -> true
	;	memberchk(X = t(A, L, R), Es),
		eqns_to_zdd(Es, L, L0, S),
		eqns_to_zdd(Es, R, R0, S),
		cofact(I, t(A, L0, R0), S)
	).

%!	is_function(+X) is det.
%	True if X is a function.
% ?- is_function([a-1,b-2,c-3]).
% ?- is_function([a-1,b-2,a-3]).
is_function(X):- sort(X, Y), has_unique_left(Y).
%
has_unique_left([X,Y|Z]):-!,
	(	arg(1, X, A), arg(1, Y, A) -> false
	; 	has_unique_left([Y|Z])
	).
has_unique_left(_).

%!	check_one_to_one(X) is det.
%	True if X is one-to-one mapping.

% ?- check_one_to_one([a-1,b-2,c-3]).
% ?- check_one_to_one([a-1,b-2,c-1]).
check_one_to_one(X):- sort(2, @=<, X, Y), has_unique_right(Y).
%
has_unique_right([X,Y|Z]):-!,
	(	arg(2, X, A), arg(2, Y, A)  -> false
	; 	has_unique_right([Y|Z])
	).
has_unique_right(_).

%!	ppoly(+X) is det.
%	Print X in polynomical form.
ppoly(X):- @(ppoly(X)).
%
ppoly(X, S):- poly_term(X, Poly, S),
	writeln(Poly).

% ?- zdd((X<<pow([a,b]), poly_term(X, P))).
% ?- zdd((X<<pow([a,b,c]), poly_term(X, P))).
%
poly_term(X, Poly):- @(poly_term(X, Poly)).
%
poly_term(X, Poly, S):- zdd_sets(X, Sets, S),
	get_compare(Pred, S),
	maplist(predsort(Pred), Sets, Sets0),
	predsort(Pred, Sets0, Sets1),
	poly_term0(Sets1, Poly).

% ?- poly_term0([], Y).
% ?- poly_term0([[]], Y).
% ?- poly_term0([[a], [b]], Y).
% ?- poly_term0([[a*b], [c]], Y).

poly_term0(X, Y):- 	maplist(mul_term, X, X0),
	sum_term(X0, Y).
%
mul_term([], 1):-!.
mul_term([X], X):-!.
mul_term([X, Y|Xs], Z):-mul_term([X*Y|Xs], Z).
%
sum_term([], 0):-!.
sum_term([X], X):-!.
sum_term([X, Y|Xs], Z):-sum_term([X+Y|Xs], Z).

% ?- zdd((X << pow([a, b, c]), zdd_rand_path(X))).
% ?- zdd((X << {[a], [b], [c]}, zdd_rand_path(X))).
% ?- zdd((X << pow([a, b, c, d, e, f, g]), zdd_rand_path(X))).

%!	zdd_rand_path(+I, +S) is det.
%	Print a set at random in  In.
zdd_rand_path(I):- shift(zdd_rand_path(I)).
%
zdd_rand_path(I, S):- zdd_rand_path(I, P, [], S), writeln(P).
%
zdd_rand_path(I, P, S):- zdd_rand_path(I, P, [], S).
%
zdd_rand_path(1, X, X, _):-!.
zdd_rand_path(I, X, Y, S):- I>1,
	cofact(I, t(A, L, R), S),
	(	L == 0 ->
		X = [A|X0],
		zdd_rand_path(R, X0, Y, S)
	;	random(2) =:= 0 ->
	 	X = [A|X0],
		zdd_rand_path(R, X0, Y, S)
	; 	zdd_rand_path(L, X, Y, S)
	).

% ?- zdd((zdd_atoms(1, S))).
% ?- zdd((family([[a,b],[b,c]], X), zdd_atoms(X, S))).
% ?- N = 10000, time(zdd(({numlist(1, N, L)}, X<<power(L), zdd_atoms(X, S)))).

%!	zdd_atoms(+X, -Y, +S) is det.
%	Y is the set of  atoms appearing on ZDD X.
zdd_atoms(X, P):- shift(zdd_atoms(X, P)).

% ?- #S, X<<pow([a,b,c,d]), zdd_atoms(X, U, S).
zdd_atoms(X, [], _):- X<2, !.
zdd_atoms(X, P, S):- memo(atoms(X)-P, S),
	(	nonvar(P) -> true
	; 	cofact(X, t(A, L, R), S),
		zdd_atoms(L, Al, S),
		zdd_atoms(R, Ar, S),
		ord_union(Al, Ar, P0),
		P=[A|P0]
	).

% ?- delete(pred([a-_]), [a-b, c-d, a-e], L).
:- meta_predicate delete(1, ?, ?).
delete(X, Y, Z):- delete(X, Y, Z, []).

:- meta_predicate delete(1, ?, ?, ?).
delete(_, [], L, L).
delete(Pred, [X|Xs], L, L0):-
	(	call(Pred, X)
	->	delete(Pred, Xs, L, L0)
	;	L = [X|L1],
		delete(Pred, Xs, L1, L0)
	).

% ?- zdd((zdd_power([a,b], R), sets(R, U))).
% ?- zdd((:zdd_power(:charlist($(a-c)), R), sets(R, U))).

% ?- zdd((X<< :zdd_power(charlist($(a-c))), sets(X, U))).
% ?- zdd((R<< pow(:charlist($(a-d))), card(R, C))).

zdd_power(X, Y):- shift(zdd_power(X, Y)).
%
zdd_power(X, Y, S):- zdd_sort(X, X0, S),
	zdd_ord_power(X0, Y, S).

%!	zdd_ord_power(+L, -X, +S) is det.
%	X is the powerset of L.
% ?- #S, zdd_ord_power([a,b], X, S), psa(X, S).
zdd_ord_power([], 1, _).
zdd_ord_power([X|Xs], P, S):- zdd_ord_power(Xs, Q, S),
	cofact(P, t(X, Q, Q), S).

%!	atomlist(+T, -A) is det.
%	T is a term of the form a(e0,e1), where a is an atom,
%	e0 and e1 are integer expressions.
%	A is the list of atoms ai with suffix i (j=<i=<k),
%	where j and k is the value of e0 and e1.

% ?- atomlist(ax(1+1, 3+1), As).
%@ As = [ax2, ax3, ax4].
% ?- atomlist(ax(1, 3), As).
atomlist(X, As):- X=..[A, I, J],
	I0 is I,
	J0 is J,
	findall(Y, (	between(I0, J0, K),
					atom_concat(A, K, Y)
			   ),
			As).

%!	charlist(+:A-B, -I) is det.
%	I is the list of characters X such that A @=< X @=< B.

% ?- charlist(a-c, X).
%@ X = [a, b, c].

charlist(A-B, I):-
	findall(X, (	char_type(X, alnum),
				    X @>= A,
					X @=< B
			   ),
			I).

%!	charlist(+A, +B, X) is det.
%	Equivalent to charlist(A-B, X).
% ?- charlist(a, c, X).
%@ X = [a, b, c].
charlist(A, B, I):- charlist(A-B, I).

% ?- term_atoms(a+b=c, L).
% ?- term_atoms(a+b, L).
% ?- term_atoms(f(a+a), L).

%!	term_atoms(+X, -L, +S) is det.
%	Unify L with a sorted list of atomic boolean
%	subterms of X.

term_atoms(X, L):- term_atoms(X, L0, []),
	sort(L0, L).

%
term_atoms(X, L, L):- var(X), !.
term_atoms(X, [X|L], L):- atom(X), !.
term_atoms([X|Y], L, L0):-!,
	term_atoms(X, L, L1),
	term_atoms(Y, L1, L0).
term_atoms(X, L, L0):- compound(X), !,
	X =..[_|As],
	term_atoms(As, L, L0).
term_atoms(_, L, L).

% ?- subst_opp(f(a,b,c,f(a)), Y, [b-a]).
% ?- subst_opp(a, X, [b-a]).
% ?- subst_opp([a], X, [b-a]).

subst_opp(X, Y, L):-  memberchk(Y-X, L).
subst_opp([X|Y], [X0|Y0], L):-!,
	subst_opp(X, X0, L),
	subst_opp(Y, Y0, L).
subst_opp(X, Y, L):- compound(X), !,
	X =..[F|As],
	subst_opp(As, Bs, L),
	Y =..[F|Bs].
subst_opp(X, X, _).

% ?- subst_term(f(a), R, [a-b]).
% ?- subst_term(g(f(a),g(a)), R, [f(a)-b]).
% ?- subst_term(f(a), R, [a-X]).
% ?- subst_term(g(f(a),g(f(a))), R, [f(a)-X]).
% ?- subst_term(g(f(a),g(f(a))), R, [f(a)-X]).
% ?- subst_term(g(f(X),g(f(Y))), R, [f(X)-a, f(Y)-b]).

%!	subst_term(+X, -Y, +S) is det.
%	Apply an assoc list X as substitution S to X to unify Y with
%	the result.

subst_term(X, Y, S):-  member(X0-Y, S), X0==X, !.
subst_term(A, B, S):-  compound(A), !,
	(	A = [X|Y]	->
		B = [X0|Y0],
		subst_term(X, X0, S),
		subst_term(Y, Y0, S)
	; 	functor(A, F, N),
		functor(B, F, N),
		subst_term(N, A, B, S)
	).
subst_term(X, X, _).
%
subst_term(0, _, _, _):-!.
subst_term(I, X, Y, S):-
	arg(I, X, A),
	arg(I, Y, B),
	subst_term(A, B, S),
	J is I - 1,
	subst_term(J, X, Y, S).

% ?- all_instances([a,b], [0,1], a+b=b, R).
% ?- all_instances([x,y], [0,1], x+y=x, R).
%!	all_instances(+As, +Vs, +X, -Y) is det.
%	Unify Y with the list of possible variations P of X,
%	where P is a variation of X if for each A in As with some
%	V in Vs which depends on A,  P is obtained from X by replacing
%	all occurrences of A appearing in X with V for A in As.

all_instances(Ks, Ts, X, Y):- 	all_maps(Ks, Ts, Fs),
	apply_maps(Fs, X, Y, []).
%
apply_maps([], _, Y, Y).
apply_maps([F|Fs], X, [X0|Y], Z):-
	subst_term(X, X0, F),
	apply_maps(Fs, X, Y, Z).

% ?- mod_table(3, [a,b], a+b, Table), maplist(writeln, Table).
% ?- mod_table(3, [a,b,c], a+b+c, Table), maplist(writeln, Table).
% ?- mod_table(3, [a,b], a*b, Table), maplist(writeln, Table).
% ?- mod_table(2, [a], a*a, Table), maplist(writeln, Table).
%!	mod_table(+M, +X, +E, -T) is det.
%	Unify T with a set of form E' = V where
%	E' is a possible ground instance  of an integer expression E
%	and V is the value of E' with modulo M, where X is a set of
%	parameters appearing in E.

mod_table(M, Xs, Exp, Table):-	M0 is M-1,
	numlist(0, M0, V),
	all_instances(Xs, V, Exp, Exps),
	!,
	maplist(mod_arith(M), Exps, Table).
%
mod_arith(M, Exp, Exp=V):- V is Exp mod M.

%
unify_all([]).
unify_all([_]).
unify_all([X,Y|Z]):- zdd:(X=Y), unify_all([Y|Z]).

% ?- zdd((X<< {[a,b], [a,c]}, Y<<{[a,b]}, zdd_subfamily(Y, X))).
% ?- zdd((X<< {[a,b], [a,c]}, Y<<{[a,b]}, zdd_subfamily(X, Y))). % false
% ?- zdd((X<< {[a,b], [a,c], [a]},  Y<<{[a,b], [a]}, zdd_subfamily(Y, X))).
% ?- zdd((X<< {[a,b],  [a,c]}, Y<<{[a,b], [a]}, zdd_subfamily(Y, X))). % false
% ?- zdd((X<< pow([a,b]), Y<<pow([a,b,c]), zdd_subfamily(X, Y))).
% ?- zdd((X<< pow([a,b]), Y<<pow([a,b,c]), zdd_subfamily(Y, X))). % false

%!	zdd_appear(+A, +X, +S) is det.
%	True if A is a member of a member of X.
% ?- zdd((X <<([a,b]+[a,c]), zdd_appear(e, X))). % false
% ?- zdd((X <<(*[a,b]+ *[a,c]), zdd_appear(e, X))). % false
% ?- zdd((X <<(*[a,b]+ *[a,c]), psa(X), zdd_appear(c, X))).
% ?- numlist(1, 10000, Ns), zdd((X<<pow(Ns), zdd_appear(10000, X))).

zdd_appear(A, X):- shift(zdd_appear(A, X)).
zdd_appear(A, X, S):- zdd_shift(zdd_appear(A, X, S)).
%
zdd_appear(A, X, S, M):- X > 1,
	memo(visited(X)-B, M),
	var(B),
	cofact(X, t(U,L,R), S),
	zdd_compare(C, A, U, S),
	(	C = (=)
	->  true
	;	C = (>),
	 	(	zdd_appear(A, L, S, M) -> true
		;	L < 2
		->  zdd_appear(A, R, S, M)
		; 	memo(visited(L)-true, M),  %  for earlier fail.
			zdd_appear(A, R, S, M)
		)
	).

%!	zdd_singleton(+X, -P, +G) is det.
%	Unify P with a sigleton ZDD for X.

zdd_singleton(X, P):- shift(zdd_singleton(X, P)).
%
zdd_singleton(X, P, S):- cofact(P, t(X, 0, 1), S).

%!	zdd_has_1(+X, +G) is det.
%	true if X  has 1 (the empty set).
% ?- zdd0((X<<pow([a,b,c]), zdd_subtr(1, X, Y), card(Y, C))).
zdd_has_1(X):- shift(zdd_has_1(X)).
%
zdd_has_1(1, _):-!.
zdd_has_1(X, S):- X>1,
   cofact(X, t(_, L, _), S),
   zdd_has_1(L, S).


%!	zdd_subst(+X, +U, +P, -Q, +S) is det.
%	X : an atom.
%	U, P: zdd.
%	Unify Q with a substitute of P replacing all occurrences of atom X
%	with U.  The replacing operation is the merge of U and the child zdd of X.
%	This is analogy of the familiar substitution operation on factored expressions.
%   ex.  (a+b)/x  :  u*x*(y+z) --> a*u*y + a*u*z + b*u*y + b*u*z .

% ?- zdd((U<<{[a]}, P<<pow([x,y]), psa(P), zdd_subst(x, U, P, Q), psa(Q))).
zdd_subst(X, U, P, Q):- shift(zdd_subst(X, U, P, Q)).
%
zdd_subst(_, _, X, X, _):- X<2, !.
zdd_subst(X, U, P, Q, S):- memo(zdd_subst(X, U, P)-Q, S),
	(	nonvar(Q) -> true
	;	cofact(P, t(Y, L, R), S),
		zdd_subst(X, U, L, L0, S),
		zdd_compare(C, X, Y, S),
		(	C = (=) 	->	zdd_merge(U, R, R0, S)
		;	C = (<)		->  zdd_cons(R0, Y, R, S)
		;	zdd_subst(X, U, R, R1, S),
			zdd_insert(Y, R1, R0, S)
		),
		zdd_join(L0, R0, Q, S)
	).

		/***********************
		*    product on zdd    *
		***********************/

% ?- zdd((X<<(a*b), Y<<(c*d), Z<<(X**Y), card(Z, C))).
% ?- zdd((X<<(a*b), Y<<(c*d), Z<<(X**Y))).
% ?- zdd((X<< a, Y<< 0, Z<<(X**Y), card(Z, C))).
% ?- zdd((X<<(a*b), Y<<(c+e), Z<<(X**Y), card(Z, C))).
% ?- time(zdd((X<<pow(@numlist(1,16)), Z<<(X**X), card(Z, C)))).
% ?- call_with_time_limit(120, (numlist(1, 16, R),
%	time(zdd((X<<pow(R), Z<<(X**X), card(Z, C)))))).

zdd_product(X, Y, Z):- shift(zdd_product(X, Y, Z)).
%
zdd_product(X, Y, 0, _):- (X<2; Y<2), !.
zdd_product(X, Y, Z, S):- memo(prod(X, Y)-Z, S),
	(	nonvar(Z) -> true	 % , write(.)	% hits found.
	;	zdd_product(X, Y, [], Bs, 0, Z0, S),
		append_pairs_list(X, Bs, Z1, S),
		zdd_join(Z0, Z1, Z, S)
	).
%
zdd_product(_, X, Bs, Bs, Z, Z, _):- X<2, !.
zdd_product(X, Y, Bs, Q, U, V, S):- cofact(Y, t(B, L, R), S),
	zdd_product(X, L, L0, S),
	zdd_join(L0, U, U0, S),
	zdd_product(X, R, [B|Bs], Q, U0, V, S).
%
append_pairs_list(X, _, X, _):- X<2, !.
append_pairs_list(X, Bs, U, S):- cofact(X, t(A, L, R), S),
	append_pairs_list(R, Bs, R0, S),
	append_pairs_list(L, Bs, L0, S),
	insert_pairs(A, Bs, R0, R1, S),
	zdd_join(L0, R1, U, S).

%
insert_pairs(_, [], U, U, _):-!.
insert_pairs(A, [B|Bs], U, V, S):-
	zdd_insert((A-B), U, U0, S),
	insert_pairs(A, Bs, U0, V, S).

		/*************************
		*     Quotient on ZDD    *
		*************************/

%!	zdd_insert(+A, +Y, -Z, +S) is det.
%	Insert A to each set in a ZDD Y, and unify Z.
%	Z = { union of U and {A} | U in Y }.
%
% ?- zdd((zdd_insert(a, 1, X), zdd_insert(b, X, X1), prz(X1))).
% ?- zdd((X<<pow([a,b,c]), zdd_insert(x, X, X1), psa(X1))).
zdd_insert(X, Y, Z):- shift(zdd_insert(X, Y, Z)).
%
zdd_insert(_, 0, 0, _):-!.
zdd_insert(A, 1, J, S):-!, zdd_singleton(A, J, S).
zdd_insert(A, I, J, S):- memo(insert(A, I)-J, S),
		(	nonvar(J) -> true
		;	cofact(I, t(X, L, R), S),
			zdd_compare(C, A, X, S),
			(	C = (=)
			->	zdd_join(L, R, K, S),
				cofact(J, t(X, 0, K), S)
			;	C = (<)
			->	cofact(J, t(A, 0, I), S)
			;  	zdd_insert(A, L, L0, S),
				zdd_insert(A, R, R0, S),
				cofact(J, t(X, L0, R0), S)
			)
		).

% ?- #S, X<<pow([a,b]), Y<<pow([c,d]),
% psa(X, S), psa(Y, S),
% zdd_join(ahead_compare([d,c,b,a]), X, Y, Z, S), psa(Z, S).

% ?- #S, X<<pow([a,b]), zdd_insert(ahead_compare([a,c,b]), c, X, Y, S),
%	psa(Y, S).
% ?- listing(zdd_insert).

:- meta_predicate zdd_insert(3, ?, ?, ?, ?).
zdd_insert(_, _, 0, 0, _):-!.
zdd_insert(_, A, 1, Y, S):-!, zdd_singleton(A, Y, S).
zdd_insert(F, A, X, Y, S):- memo(zdd_insert(X, A, F)-Y, S),
	(	nonvar(Y) -> true  %	, write(.) % many hits
	;	cofact(X, t(B, L, R), S),
		zdd_insert(F, A, L, L0, S),
		call(F, C, A, B),
		(	C = (=) -> zdd_cons(R0, A, R, S)
		;	C = (<) -> bdd_append([A, B], R, R0, S)
		;	zdd_insert(F, A, R, R1, S),
			zdd_cons(R0, B, R1, S)
		),
		zdd_join(L0, R0, Y, S)
	).

% ?- zdd((zdd_insert_as_list(a, 1, X), zdd_insert_as_list(b, X, Y), cofact0(Y, A, Z))).
% ?- zdd0((zdd_insert_as_list(a, 1, X), zdd_insert_as_list(b, X, Y), cofact0(Y, A, Z))).
zdd_insert_as_list(X, Y, Z):- shift(zdd_insert_as_list(X, Y, Z)).

%
zdd_insert_as_list(A, 1, J, S):-!, zdd_singleton(A, J, S).
zdd_insert_as_list(_, 0, 0, _):-!.
zdd_insert_as_list(A, I, J, S):- cofact(I, t(X, _, R), S),
	zdd_compare(C, A, X, S),
	(	C = (=)	->  J = I
	;	C = (<) -> 	cofact(J, t(A, 0, I), S)
	;  	zdd_insert_as_list(A, R, R0, S),
		cofact(J, t(X, 0, R0), S)
	).

% ?- zdd((family([[a, b],[b]], X), sets(X, Sx),
%	zdd_insert_atoms([c, d], X, Y), sets(Y, Sy))).

%!	zdd_insert_atoms(+As, +X, -Y) is det.
%	Insert all atoms in As to X to get the result Y.

% ?- numlist(1, 10000, Ns), zdd((zdd_insert_atoms(Ns, 1, X), psa(X))).
zdd_insert_atoms(As, X, Y):- shift(zdd_insert_atoms(As, X, Y)).

% ?- zdd((X<<pow([1,2,3]), zdd_insert_atoms([4,5,6], X, Y), card(Y, C))).
% ?- zdd((X<<pow([1,2,3]), zdd_insert_atoms([1,2,3], X, Y), card(Y, C))).
zdd_insert_atoms(X, Y, Z):- shift(zdd_insert_atoms(X, Y, Z)).
%
zdd_insert_atoms(X, Y, Z, S):- zdd_sort(X, X0, S),
	zdd_ord_insert(X0, Y, Z, S).
%
zdd_ord_insert(_, 0, 0, _):-!.
zdd_ord_insert([], X, X, _):-!.
zdd_ord_insert(As, 1, Y, S):-!, zdd_ord_append(As, 1, Y, S).
zdd_ord_insert(As, X, Y, S):- memo(ord_insert(As, X)-Y, S),
	(	nonvar(Y) -> true % , write(.)   % many hits.
	;	As = [A|U],
		cofact(X, t(B, L, R), S),
		zdd_compare(C, A, B, S),
		(	C = (<) ->
			zdd_ord_insert(U, X, Y0, S),
			cofact(Y, t(A, 0, Y0), S)
		;	C = (=) ->
			zdd_ord_insert(U, L, L1, S),
			zdd_ord_insert(U, R, R1, S),
			zdd_join(R1, L1, Y0, S),
			cofact(Y, t(A, 0, Y0), S)
		;	zdd_ord_insert(As, L, L1, S),
			zdd_ord_insert(As, R, R1, S),
			cofact(Y, t(B, L1, R1), S)
		)
	).

%!	zdd_append(+X, +Y, -Z [, G]) is det.
%	Z is result of inserting all elements of X into Y.
%	Y is assumed to have no atom in X.
zdd_append(X, Y, Z):- shift(zdd_append(X, Y, Z)).

%
% ?- numlist(1, 100, Ns), zdd((X<<pow(Ns), card(X, C))).
% ?- numlist(1, 100, Ns), zdd((X<<pow(Ns))).

% ?- zdd((zdd_append([c, a, b,c], 1, Z), psa(Z))).
zdd_append(X, Y, Z, S):- zdd_sort(X, X0, S),
	zdd_ord_append(X0, Y, Z, S).
%
zdd_ord_append(X, Y, Z):- shift(zdd_ord_append(X, Y, Z)).
%
zdd_ord_append([], X, X, _).
zdd_ord_append([A|As], X, Y, S):-
	zdd_ord_append(As, X, Y0, S),
	cofact(Y, t(A, 0, Y0), S).


% ?- #S, X<<pow([a,b]), zdd_reorder(ahead_compare([b,a]), X, Y, S), psa(Y, S).
:- meta_predicate zdd_reorder(3, ?, ?, ?).
zdd_reorder(_, X, X, _):- X<2, !.
zdd_reorder(F, X, Y, S):- cofact(X, t(A, L, R), S),
	zdd_reorder(F, L, L0, S),
	zdd_reorder(F, R, R1, S),
	zdd_insert(F, A, R1, R0, S),
	zdd_join(L0, R0, Y, S).


		/*****************
		*     filters    *
		*****************/

% ?- #S, X<< +[*[a, b], *[b,c]], ltr_meet_filter([a, b], X, Y, S), psa(Y, S).
% ?- #S, X<< +[*[a, b], *[b]], ltr_meet_filter([-a, b], X, Y, S), psa(Y, S).
% ?- #S, numlist(1, 3, Ns), X<< pow(Ns), ltr_meet_filter([1, -(2), 3], X, Y, S), psa(Y, S).
% ?- #S, numlist(1, 3, Ns), X<< pow(Ns), ltr_meet_filter([1, 3], X, Y, S), psa(Y, S).
% ?- #S, numlist(1, 3, Ns), X<< pow(Ns), ltr_meet_filter([-(1), -(3)], X, Y, S), psa(Y, S).
% ?- #S, numlist(1, 1000, Ns), X<< pow(Ns), ltr_meet_filter([-(1000), -(1)], X, Y, S), card(Y, C, S),
%	card(X, D, S), compare(E, C, D), U is D-C.
%
ltr_meet_filter([], X, X, _):-!.
ltr_meet_filter([A|F], X, Y, S):-
	ltr_filter(A, X, X0, S),
	ltr_meet_filter(F, X0, Y, S).

% ?- #S, X<< +[*[a, b], *[b,c]], ltr_join_filter([a, b], X, Y, S), psa(Y, S).
% ?- #S, X<< +[*[a, b], *[b]], ltr_join_filter([-a, b], X, Y, S), psa(Y, S).
ltr_join_filter(F, X, Y, S):- ltr_join_filter(F, X, 0, Y, S).
%
ltr_join_filter([], _, Y, Y, _):-!.
ltr_join_filter([P|F], X, Y, Y0, S):-!,
	ltr_filter(P, X, Y1, S),
	zdd_join(Y, Y1, Y2, S),
	ltr_join_filter(F, X, Y2, Y0, S).
%
ltr_filter(-A, X, Y, S):-!, ltr_filter_negative(A, X, Y, S).
ltr_filter(A, X, Y, S):- ltr_filter_positive(A, X, Y, S).

%
ltr_filter_negative(_, X, X, _):- X<2, !.
ltr_filter_negative(A, X, Y, S):- memo(letral_neg(X, A)-Y, S),
	(	nonvar(Y) -> true
	;	cofact(X, t(B, L, R), S),
		ltr_filter_negative(A, L, L0, S),
		zdd_compare(C, A, B, S),
		(	C = (=) -> R0 = 0
		;	C = (<) ->  R0 = R
		;	ltr_filter_negative(A, R, R0, S)
		),
		cofact(Y, t(B, L0, R0), S)
	).
%
ltr_filter_positive(_, X, 0, _):- X<2, !.
ltr_filter_positive(A, X, Y, S):- memo(ltr_pos(X, A)-Y, S),
	(	nonvar(Y) -> true
	;	cofact(X, t(B, L, R), S),
		ltr_filter_positive(A, L, L0, S),
		zdd_compare(C, A, B, S),
		(	C = (=) -> R0 = R
		;	C = (<) ->  R0 = 0
		;	ltr_filter_positive(A, R, R0, S)
		),
		cofact(Y, t(B, L0, R0), S)
	).

% ?- zdd((X<<pow([a,b,c]), shift(dnf_filter([[-b,-c]], X, Y)), psa(Y))).
% ?- zdd((X<<pow([a,b]), shift(dnf_filter([[b]], X, Y)), psa(Y))).
% ?- zdd((X<<pow([a,b]), shift(dnf_filter([[a,b]], X, Y)), psa(Y))).
% ?- zdd((X<<pow([a,b]), shift(dnf_filter([], X, Y)), psa(Y))).
% ?- zdd((X<<pow([a,b,c]), shift(dnf_filter([[c]], X, Y)), psa(Y))).
% ?- zdd((X<<pow([a,b,c]), shift(dnf_filter([], X, Y)), psa(Y))).
% ?- zdd((X<<pow([a,b,c]), shift(dnf_filter([[a],[-c]], X, Y)), psa(Y))).
dnf_filter(DNF, X, Y, S):- dnf_filter(DNF, X, 0, Y, S).
%
dnf_filter([], _, Y, Y, _).
dnf_filter([F|Fs], X, Y, Z, S):-
	ltr_meet_filter(F, X, X0, S),
	zdd_join(X0, Y, Y0, S),
	dnf_filter(Fs, X, Y0, Z, S).

% ?- zdd((X<<pow([a,b,c]), shift(cnf_filter([[-b,-c]], X, Y)), psa(Y))).
% ?- zdd((X<<pow([a,b,c]), shift(cnf_filter([[-a, -b, -c]], X, Y)), psa(Y))).
cnf_filter([], X, X, _).
cnf_filter([F|Fs], X, Y, S):-
	ltr_join_filter(F, X, X0, S),
	cnf_filter(Fs, X0, Y, S).

%
zdd_meet_list([A], A, _):-!.
zdd_meet_list([A|As], B, S):- zdd_meet_list(As, A, B, S).
%
zdd_meet_list([], X, X, _).
zdd_meet_list([A|As], X, Y, S):- zdd_meet(A, X, X0, S),
	zdd_meet_list(As, X0, Y, S).
%
zdd_join_list([], 0, _):-!.
zdd_join_list([X|Xs], Y, S):- zdd_join_list(Xs, X, Y, S).
%
zdd_join_list([], X, X, _):-!.
zdd_join_list([A|As], X, Y, S):- zdd_join(A, X, X0, S),
	zdd_join_list(As, X0, Y, S).

%
ltr_join_list([], 0, _):-!.
ltr_join_list([X|Xs], Y, S):-ltr_join_list(Xs, X, Y, S).
%
ltr_join_list([], X, X, _):-!.
ltr_join_list([A|As], X, Y, S):- zdd_join(A, X, X0, S),
	ltr_join_list(As, X0, Y, S).

		/*********************
		*     remove atom    *
		*********************/

%!	zdd_remove(+A, +X, -Y, +S) is det.
%	Remove an atom A from each set in a ZDD X.
%	Y = { U\{A} | U in X }

% ?- zdd((family([[a,b,c],[a1,b,c1]], X), zdd_remove(b, X, Y),
%	sets(Y, Sy))).

zdd_remove(X, Y, Z):- shift(zdd_remove(X, Y, Z)).

zdd_remove(X, Y, Z, S):- zdd_shift(zdd_remove(X, Y, Z, S)).

%
zdd_remove(_, X, X,  _, _):- X<2, !.
zdd_remove(A, I, J, S, M):- memo(zdd_remove(I)-J, S),
	(	nonvar(J)	->  true
	; 	cofact(I, t(Ai, Li, Ri), S),
		zdd_compare(C, A, Ai, S),
		(	C = (<)  -> J = I
		;   C = (>)  -> zdd_remove(A, Li, Lia, S, M),
						zdd_remove(A, Ri, Ria, S, M),
						cofact(J, t(Ai, Lia, Ria), S)
		; 	zdd_join(Li, Ri, J, S)
		)
	).

% For debugggin.
dump_args(List) :- nl,
	writeln('** Completed Options:'),
	maplist(writeln, List).

% ?- zdd((X<<pow([a,b,c,d]), shift(zdd_dump(X)))).
% ?- zdd((X<<pow([a,b,c,d]), @(zdd_dump([X,[X,[X]]])))).

zdd_dump([], _):-!.
zdd_dump([X|Xs], S):-!, zdd_dump(X, S), zdd_dump(Xs, S).
zdd_dump(X, _):- X<2, !.
zdd_dump(X, S):- memo(dumped(X)-B, S),
	(	nonvar(B) -> true
	;	B = 0,
		cofact(X, t(A, L, R), S),
		writeln(X=t(A, L, R)),
		zdd_dump(L, S),
		zdd_dump(R, S)
	).

%
update_hit_count :-
	( nb_current(count_hit, C), integer(C)
	->	true
	;	C = 0
	),
	C0 is C + 1,
	b_setval(count_hit, C0).
%
print_hit_count	 :- (nb_current(count_hit, N) -> N0 = N; N0 = 0),
  format("~n~w", [count_hit = N0]).


%
newvars(Vars):- b_getval(boole_zip, Zip),
		length(Zip, N),
		I is N + 1,
		length(Vars, J),
		K is I + J - 1,
		(	K =<0 ->  Vars = []
		;	numlist(I, K, Vars)
		),
		zip(Vars, Vars, Zip0),
		append(Zip, Zip0, Zip1),
		b_setval(boole_zip, Zip1).
%
new_root_zdds(Zdds):- length(Zdds, N),
		length(Vars, N),
		newvars(Vars),
		maplist(var_to_zdd, Vars, Zdds).
%
var_to_zdd(I, X):- memo(truth(I)-X).

% ?- new_variables(1+2+A+B, [A], V).
new_variables(A, OldVs, NewVs):- term_variables(A, V0),
		subtract_by_memq(V0, OldVs, NewVs).

% ?- update_zip_by_term(A+B, [1-A], R), update_zip_by_term(A+B+C+D, R, S).
update_zip_by_term(A, OldZip, NewZip):- zip(_, Vs, OldZip),
	new_variables(A, Vs, NewVars),
	length(OldZip, Len),
	update_zip_by_term(NewVars, Len, OldZip, NewZip).
%
update_zip_by_term([], _, Z, Z).
update_zip_by_term([V|Vs], I, Z, NewZ):-
	I0 is I + 1,
	update_zip_by_term(Vs, I0, [I0-V|Z], NewZ).

% ?- subtract_by_memq([X, A, Y, B], [Y, X], U).
subtract_by_memq([], _, []).
subtract_by_memq([A|B], C, D):- memq(A, C), !,
	subtract_by_memq(B, C, D).
subtract_by_memq([A|B], C, [A|D]):-
	subtract_by_memq(B, C, D).
%
union_by_memq([], X, X).
union_by_memq([A|As], X, Y)	:- memq(A, X), !,
    union_by_memq(As, X, Y).
union_by_memq([A|As], X, [A|Y])	:-
	union_by_memq(As, X, Y).
%
union_by_memq([], []).
union_by_memq([A|B], C)	:- union_by_memq(B, C0),
				union_by_memq(A, C0, C).

% ?- solve([zdd_sat(a,X)]), zout(X).
zout_zip([]):-!.
zout_zip([I-Z|R]):-!, zdd_to_prop(Z, Z0),
		writeln(I-Z0),
		zout(R).
%
zout(X):- integer(X), !,
		zdd_to_prop(X, Y),
		writeln(node(X)=Y).
zout(X):- writeln(X).

%
zout(A, V):-!, zdd_to_prop(V, V0),
	   format("~w: ~w", [A, V0]).

%
zdd_to_prop(X, Y):- zdd_to_prop(X, Y, [slim_boole(true)]).
%
zdd_to_prop(X, Y, Options):- zdd_to_tree(X, T),
		tree_to_prop(T, Y0),
		( memberchk(slim_boole(true), Options)
		-> slim_boole(Y0, Y)
		;  Y=Y0
		).

% ?- tree_to_prop(if(1, truth(1), false), P, 1).
tree_to_prop(X, Y):- varnum(N), tree_to_prop(X, Y, N).
%
tree_to_prop(false, false, _):-!.
tree_to_prop(true, true, _):-!.
tree_to_prop(if(X, L, R), X*L0 + -(X)*R0, N):-
	tree_to_prop(L, TL, N),
	tree_to_prop(R, TR, N),
	depth(L, A, N),
	depth(R, B, N),
	X0 is X + 1,
	conj_of_negs(X0, A, TL, L0),
	conj_of_negs(X0, B, TR, R0).

% ?- conj_of_negs(1, 3, A).
conj_of_negs(I, J, A):- I=<J, conj_of_negs(I, J, -(J), A).

% ?- conj_of_negs(3, 1, a, X).
% ?- conj_of_negs(1, 3, a, X).
conj_of_negs(I, J, A, A):- I>=J, !.
conj_of_negs(I, J, A, B):- J0 is J - 1,
						conj_of_negs(I, J0, -(J0)*A, B).




		/*********************
		*     Cardinality    *
		*********************/

%!	card(+I, -C, +S) is det.
%	unify C with the cardinality of a family I of sets.

% ?- numlist(1, 10000, Ns), zdd((X<<pow(Ns), card(X, S))).
% ?- zdd((X<<pow([a,b]), card(X, C))).
card(X, Y):- shift(card(X, Y)).

%
card(X, Y, S):- setup_call_cleanup(
	open_state(M),
	card_with_memo(X, Y, S, M),
	close_state(M)).

%
card_with_memo(I, I, _, _):- I < 2, !.
card_with_memo(I, C, S, M):- memo(I-C, M),
	(	nonvar(C) -> true
	;	cofact(I, t(_, L, R), S),
		card_with_memo(R, Cr, S, M),
		card_with_memo(L, Cl, S, M),
		C is Cl + Cr
	).

%!	max_length(+X, -Max, +S) is det.
%	Unify Max with max of size of  members of X.

% ?- #S, numlist(1, 100, Ns), X<<pow(Ns), max_length(X, I, S).
max_length(X, Max, S):-
	setup_call_cleanup(
		open_memo(S),
		max_length_with_memo(X, Max, S),
		close_memo(S)).
%
max_length_with_memo(X, 0, _):- X<2, !.
max_length_with_memo(X, Max, S):-
	memo(max(X)-Max, S),
	cofact(X, t(_, L, R), S),
	(	nonvar(Max) -> true
	;	max_length_with_memo(L, ML, S),
		max_length_with_memo(R, MR, S),
		Max is max(1 + MR, ML)
	).

		/***********************
		*     Handy helpers    *
		***********************/

%								%
zip([], [], []).
zip([X|Xs], [Y|Ys], [X-Y|Zs]) :- zip(Xs, Ys, Zs).
%
unify_zip([]).
unify_zip([X-X|R]):- unify_zip(R).


		/******************************************
		*     Oprations on clauses of literals    *
		******************************************/

% ?- ltr_compare(C, a, -a).
% ?- ltr_compare(C, -(a), a).
% ?- ltr_compare(C, -(b), a).
% ?- ltr_compare(C, -(a), b).
% ?- ltr_compare(C, -(a), -(b)).
% ?- ltr_compare(C, -(a), -(a)).

%!  ltr_compare(-C, +X, +Y) is det.
%   Compare literals X and Y, and C is unified with
%	<, =, > as the standard compare/3.
%   ex. a<b, a < -a,  -a < b,   -a < -b.

ltr_compare(C, -(X), -(Y)):-!, compare(C, X, Y).
ltr_compare(C, X, -(Y)):-!, compare(C0, X, Y),
		(	C0 == (=) ->  C =(<)
		;	C0 = C
		).
ltr_compare(C, -(X), Y):-!, compare(C0, X, Y),
		(	C0 == (=) -> C = (>)
		;	C0 = C
		).
ltr_compare(C, X, Y):- compare(C, X, Y).

% ?- zdd((shift(ltr_compare(C, a, b)))).
% ?- zdd((shift(ltr_compare(C, -a, a)))).
% ?- zdd((shift(ltr_compare(C, a, -a)))).
% ?- zdd((shift(ltr_compare(C, -a, b)))).
%
ltr_compare(C, X, Y, S):- get_compare(F, S),
	compare_as_literal(C, X, Y, F).
%
compare_as_literal(F, C, -(X), -(Y)):-!, call(F, C, X, Y).
compare_as_literal(F, C, X, -(Y)):-!, call(F, C0, X, Y),
		(	C0 == (=) ->  C =(<)
		;	C0 = C
		).
compare_as_literal(F, C, -(X), Y):-!, call(F, C0, X, Y),
		(	C0 == (=) -> C = (>)
		;	C0 = C
		).
compare_as_literal(F, C, X, Y):- call(F, C, X, Y).

% ?- ltr_sort([b, b, a], X).
% ?- ltr_sort([-b, b,-b, -a], X).

%!	ltr_sort(+X, -Y) is det.
%   Y is unified with sorted X by ltr_compare.
% ?- zdd(shift(ltr_sort([-c, a, -b, -a, b, c], X))).
ltr_sort(X, Y):- predsort(ltr_compare, X, Y).

%!	ltr_sort(+X, -Y, +S) is det.
%   Y is the sorted as literals based on the order set in S.
ltr_sort(X, Y, S):- get_compare(F, S),
	predsort(compare_as_literal(F), X, Y).

% ?- zdd((X << ltr_pow([a, -b, c]), ltr_memberchk([a, -b], X))).
% ?- zdd((X << pow([a, -b, c]), ltr_has_set([a, b], X))).
% ?- zdd((X << ltr_power([a, -b, c]), ltr_has_set([a, b], X))).

ltr_memberchk(L, Z):- shift(ltr_memberchk(L, Z)).
%
ltr_memberchk(L, Z, S):- ltr_sort(L, L0),
	ltr_ord_memberchk(L0, Z, S).
%
ltr_ord_memberchk([], Z, S):- !,  zdd_has_1(Z, S).
ltr_ord_memberchk([A|As], Z, S):- Z > 1,
	cofact(Z, t(B, L, R), S),
	ltr_compare(C, A, B),
	(	C = (=) -> ltr_ord_memberchk(As, R, S)
	;	ltr_ord_memberchk([A|As], L, S)
	).

% ?- call_with_time_limit(100, time(zdd((a_big_cnf(30, X), resolve(X))))).
% ?- zdd((a_big_cnf(1, X), @ltr_card(X, C))).
% ?- zdd((a_big_cnf(2, X), @ltr_card(X, C))).
% ?- zdd((a_big_cnf(3, X), @ltr_card(X, C))).
% ?- zdd((a_big_cnf(10, X), @ltr_card(X, C))).
% ?- zdd((a_big_cnf(20, X), @ltr_card(X, C, N))).
% ?- time(zdd((a_big_cnf(100, X), @ltr_card(X, C, N)))).
% ?- N=10, time(zdd((a_big_cnf(N, X), @ltr_card(X, C)))), C=:=2^N.
% ?- N=300, time(zdd((a_big_cnf(N, X), @ltr_card(X, C, N)))), C=:=2^N.

a_big_cnf(X, Y):- shift(a_big_cnf(X, Y)).
%
a_big_cnf(0, 1, _).
a_big_cnf(N, X, S):- N>0,
	N0 is N-1,
	a_big_cnf(N0, X0, S),
	ltr_insert(N, X0, R, S),
	ltr_insert(-N, X0, L, S),
	zdd_join(L, R, X, S).

% ?- numlist(1, 3, Ns), zdd((X<<(pow(Ns)-1), ltr_choices(X, Y), psa(X), psa(Y), card(Y, C))).
% ?- time((numlist(1, 20, Ns), zdd((X<<(pow(Ns)-1), ltr_choices(X, Y), card(Y, C))))).
% ?- zdd((X<<((a*b)+c), ltr_choices(X, Y), psa(Y))).
% ?- zdd((X<<{[a,b],[c,d]},  ltr_choices(X, Y), ltr_choices(Y, Z),
%	ltr_choices(Z, U), psa(X),  psa(Y), psa(Z), psa(U), ltr_choices(U, V), psa(V))).
% ?- zdd(( (X<< {[a,b]}), ltr_choices(X, Y), psa(X), nl, psa(Y))).
% ?- zdd(( (X<< {[a,b,c]}), ltr_choices(X, Y),	psa(X), nl, psa(Y))).
% ?- zdd(( (X<< {[a, b], [c, d]}), ltr_choices(X, Y), psa(Y))).
% ?- zdd(( (X<< {[], [a, b], [c, d]}), ltr_choices(X, Y), psa(Y))).
% ?- zdd(( (X<< {[a], [c, d]}), ltr_choices(X, Y), psa(Y))).
% ?- zdd(( X<< {[a,b], [c, d], [e,f]}, ltr_choices(X, Y), psa(Y))).
% ?- zdd(( X<< {[a,b]}, ltr_choices(X, Y), psa(Y))).
% ?- zdd(( X<< {[]}, ltr_choices(X, Y), psa(Y))).
% ?- zdd(( cnf(a, X), ltr_choices(X, Y), psa(Y))).
% ?- zdd(( cnf(a*b, X), ltr_choices(X, Y), psa(Y))).
% ?- zdd(( cnf(a* (-a), X), ltr_choices(X, Y), psa(Y))).

%  Almost the same as zdd_chioce except ltr_insert for zdd_insert
ltr_choices(X, Y):- shift(sw_choices(ltr, X, Y)).

ltr_choices(X, Y, S):- sw_choices(ltr, X, Y, S).

% ?- spy(build_sample).
% ?- N=10, numlist(2, 10, Ns), zdd((shift(build_sample(Ns, X)), card(X, C))).
% ?- N=100, numlist(2, N, Ns), time(zdd((shift(sample_cnf(Ns, X)), card(X, C), resolve(X)))).
%@ % 2,532,002 inferences, 0.609 CPU in 1.005 seconds (61% CPU, 4154910 Lips)
%@ N = 100,
%@ Ns = [2, 3, 4, 5, 6, 7, 8, 9, 10|...],
%@ X = 298,
%@ C = 633825300114114700748351602688.

sample_cnf([], 1, _).
sample_cnf([J|Js], X, S):- sample_cnf(Js, Y, S),
	ltr_insert(J, Y, Z, S),
	ltr_insert(-J, Y, U, S),
	ltr_join(Z, U, X, S).

% ?- zdd_sat(-true).
% ?- zdd_sat(a+b), zdd_sat_count(C).
% ?- zdd_sat(a), zdd_sat(b).
% ?- zdd_sat(-(a + -a)).
% ?- zdd_sat(a), zdd_sat(b), zdd_sat_count(C).
% ?- zdd_sat(or(a, b)), zdd_sat_count(C).
% ?- zdd_sat(b), zdd_sat_count(C).
% ?- zdd_sat(xor(a, b)), zdd_sat_count(C).
% ?- zdd_sat(exor(a, b)), zdd_sat_count(C).
% ?- zdd_sat(e_x_o_r(a, b)), zdd_sat_count(C).
% ?- zdd_sat(a), zdd_sat(-a).	% false.
% ?- zdd_sat(a), zdd_sat(a+b), zdd_sat_count(C).
% ?- zdd_sat(a(1)+a(2)), zdd_sat_count(C).
% ?- zdd_sat(A+a(2)), zdd_sat_count(C).
% ?- zdd_sat(A + B), zdd_sat_count(C).
%@ A = C,
%@ B = D,
%@ C = 3.

% ?- zdd_sat(a+b+c+d), zdd_sat_count(C).
%@ C = 15.

% ?- zdd_sat(a+b), zdd_sat(c+d),  zdd_sat_count(C).
%@ C = 9.

% ?-zdd_sat(a->b), zdd_sat(b->c), zdd_sat(c->d), zdd_sat(d->a), zdd_sat_count(C).
%@ C = 2.

% ?- zdd_sat(X=(a=b)), zdd_sat(Y=(b+c)), zdd_sat(X=Y), zdd_sat_count(C).
%@ X = C,
%@ Y = F,
%@ C = 4.

% ?- zdd_sat(X=(a=b)), zdd_sat(Y=(b+c)), zdd_sat(X=Y), zdd_sat(X = (-Y)).
%@ false.

% ?- zdd_sat(*[p(1),p(2)]), zdd_sat_count(C).
%@ C = 1.
% ?- zdd_sat(+[p(1),p(2)]), zdd_sat_count(C).
%@ C = 3.
% ?- N=10, findall(p(I, J), (between(0, N, I), between(0, N, J)), L), zdd_sat(+L), zdd_sat_count(C).
%@ N = 10,
%@ L = [p(0, 0), p(0, 1), p(0, 2), p(0, 3), p(0, 4), p(0, 5), p(0, 6), p(0, 7), p(..., ...)|...],
%@ C = 2658455991569831745807614120560689151.
% ?- N=10, findall(p(I, J), (between(0, N, I), between(0, N, J)), L), zdd_sat(*L), zdd_sat_count(C).
%@ N = 10,
%@ L = [p(0, 0), p(0, 1), p(0, 2), p(0, 3), p(0, 4), p(0, 5), p(0, 6), p(0, 7), p(..., ...)|...],
%@ C = 1.

% ?-zdd_sat(a->b), zdd_sat(b->c), zdd_sat(c->d), zdd_sat(d->a), zdd_sat(-(d = b)).
%@ false.
% ?- E = (a=b)*(b=c)*(c=a), zdd_sat(E), zdd_sat_count(C).
%@ E =  (a=b)*(b=c)*(c=a),
%@ C = 2.
% ?- zdd_sat(A + B), zdd_sat_count(C).
%@ C = 3.
% ?- zdd_sat(A -> B), Vs = [A,B], zdd_sat(+([1|Vs])), zdd_sat_count(C).
%@ Vs = [A, B],
%@ C = 3 .
% ?- zdd_sat(A -> B), Vs = [A,B], zdd_sat_count(+([1|Vs]), C).
%@ Vs = [A, B],
%@ C = 4 .
% ?- zdd_sat(a=<(b=<a)), zdd_sat_count(Count).
% ?- zdd_sat(a), zdd_sat(b), zdd_sat_count(C).
% ?- zdd_sat(a+b+c), zdd_sat(-b), zdd_sat(-c), zdd_sat_count(C).
% ?- zdd_sat(X+A+B+C), zdd_sat(-B), zdd_sat(-C), zdd_sat_count(K).
% ?- Prop =
%	( (~(B) =< F) * ((B * F) =< ~(I)) * ((I + ~(B)) =< ~(F)) =< (I * F)),
%	term_variables(Prop, Vs),
%	sat(Prop),
%   labeling(Vs).
% ?- zdd_sat(a+b), zdd_sat(b+c), b_getval(sat_state, S),
%	get_key(root, R, S), ltr_card(R, C, S).
% ?- zdd_sat(a+b+c), zdd_sat(-a), zdd_sat(-b), zdd_sat(-c), print_root, zdd_sat_count(C).
% ?- zdd_sat(a=:=b), zdd_sat(b=:=c), zdd_sat_count(C), zdd_sat(-(a=:=c)), zdd_sat_count(D).
% ?- zdd_sat(a=:=b), zdd_sat(b=:=c), zdd_sat_count(C), zdd_sat(-(a=:=c)), b_getval(sat_state, U), get_key(root, A, U), psa(A, U).

zdd_sat(X):-
	(	nb_current(sat_state, S),  S \== [] -> true
	; 	open_state(S, [extra([root-1, atom_index-2])]),
		set_compare(ltr_compare, S),
		b_setval(sat_state, S)
	),
	b_getval(sat_state, S),
	get_key(root, Root, S),
	dnf(X, Y, S),
	ltr_merge(Y, Root, Root0, S),
	set_key(root, Root0, S),
	!,
	Root0 \== 0.

% ?- use_module(library(clpb)).
% ?- Prop =
%	( (~(B) =< F) * ((B * F) =< ~(I)) * ((I + ~(B)) =< ~(F)) =< (I * F)),
%	term_variables(Prop, Vs),
%	sat(Prop),
%   labeling(Vs).

% ?- Prop =
%	( (-(B) -> F) * ((B * F) -> -(I)) * ((I + -(B)) -> -(F)) -> (I * F)),
%	zdd_sat(Prop),
%	zdd_sat_count(Count).
%@ Prop =  ((-B->F)*(B*F-> -I)*(I+ -B-> -F)->I*F),
%@ Count = 5.
% ?- Prop =
%	( (-(B) -> F) * ((B * F) -> -(I)) * ((I + -(B)) -> -(F)) -> (I * F)),
%	zdd_sat(-Prop),
%	zdd_sat_count(Count).
%@ Prop =  ((-B->F)*(B*F-> -I)*(I+ -B-> -F)->I*F),

% ?- fos(A), fos_count(C).
% ?- fos(A), fos(B), fos(C), fos_count.
% ?- fos(a*b), fos_count(C).
% ?- fos(a), fos_count(C).
% ?- fos(pow([a, b])/a), fos_count.
% ?- fos(pow([a, b])//a), fos_count.
% ?- fos((a+b)*(x+Y)), fos_count.
% ?- fos((a+b)&(x+Y)), fos_count.
% ?- fos((a+b)&(a+c)), fos_count.

fos(X):-
	(	nb_current(fos_state, S0),  S0 \== [] -> true
	; 	open_state(S0, [extra([root-0, varnum-0])]),
		b_setval(fos_state, S0)
	),
	b_getval(fos_state, S),
	get_key(root, Root, S),
	get_key(varnum, N, S),
	numbervars(X, N, N0),
	set_key(varnum, N0, S),
	context_module(M),
	zdd_eval(X, Y, S, M),
	zdd_join(Y, Root, Root0, S),
	set_key(root, Root0, S).
%
fos_count :- fos_count(C), writeln(C).
%
fos_count(C):- b_getval(fos_state, S),
	get_key(root, R, S),
	card(R, C, S).

% ?- zdd_sat(a), zdd_sat_count(C).
% ?- zdd_sat(a), zdd_sat(b), zdd_sat_count(C).
% ?- zdd_sat(a+b+c), zdd_sat(-b), zdd_sat(-c), zdd_sat_count(C).
% ?- zdd_sat(X+A+B+C), zdd_sat(-B), zdd_sat(-C), zdd_sat_count(K).
zdd_sat_count(C):- b_getval(sat_state, S),
	get_key(root, X, S),
	get_key(atom_index, N, S),
	N0 is N - 2,
	ltr_card(N0, X, C, S).

% ?- zdd_sat(a), b_getval(sat_state, S), write(S),
% get_key(root, R, S),
% psa(R, S), get_key(atom_index, A, S), ltr_card(R, C, S).

zdd_sat_count(F, C):- b_getval(sat_state, S),
	dnf(F, X, S),
	get_key(atom_index, N, S),
	N0 is N-2,
    ltr_card(N0, X, C, S).
%
zdd_sat_count(X, C, S):-
	get_key(atom_index, N, S),
	N0 is N-2,
	ltr_card(N0, X, C, S).
%
print_root :- b_getval(sat_state, S),
	 get_key(root, R, S),
	 psa(R, S).
%
sat_clear:- nb_setval(sat_state, []).


% ?- prove(true).
% ?- prove(a + -a).
% ?- time(forall(valid_formula(A), prove(A))).

% ?- heavy_valid_formula(H), prove(H).
%@
heavy_valid_formula((((((((((((((((((((((((((((p13 <=> p12) <=> p11) <=> p10)
<=> p9) <=> p8) <=> p7) <=> p6) <=> p5) <=> p4) <=> p3)
<=> p2) <=> p1) <=> p0) <=> p13) <=> p12) <=> p11) <=> p10)
<=> p9) <=> p8) <=> p7) <=> p6) <=> p5) <=> p4) <=> p3)
<=> p2) <=> p1) <=> p0)).

% ?- prove(a).
% ?- prove(*[a, -a]).
% ?- run(1, 100).
% ?- run(2, 100).
% ?- run(11, 100).

run(N, T_limit) :-  forall( between(1, N, K),
	(	mk_left_assoc_term(<=>, K, F),
		format("Propositions: p(0) ... p(~w).~n", [K]),
		time(prove(F, [time_limit(T_limit), cnf(_)])))).

% ?- mk_left_assoc_term(<=>, 1, X), write_canonical(X).
% ?- mk_left_assoc_term(<=>, 13, X), write_canonical(X).
mk_left_assoc_term(Bop, N, Ex):- findall(p(I), between(0, N, I), L),
			  append(L, L, L2),
			  reverse(L2, R),
			  apply_left_assoc(Bop, R, Ex).
%
apply_left_assoc(Bop, [X|Y], Ex):- apply_left_assoc(Bop, X, Y, Ex).
%
apply_left_assoc(_, X, [], X):-!.
apply_left_assoc(Bop, U, [X|Y], W):- V=..[Bop, U, X],
	apply_left_assoc(Bop, V, Y, W).


% ?- forall(valid_formula(A), prove(A)).
% Valid formulae.
valid_formula( (a * b) + (a * -b) + (-a * b) + (-a * -b)).
valid_formula((-(-a) -> a) * (a -> -(-a))).		% Double Negation
valid_formula( ((a * b) -> -((-a + -b))) * (-((-a + -b)) -> (a * b))). % De Morgan’s Law
valid_formula( (-b -> f) * ((b *f) -> -i) * ((i + -b) -> -f) -> b).
valid_formula(a + -a).
valid_formula(((a -> b) -> a) -> a).  % Peirce’s Law
valid_formula(-(-a)->a).
valid_formula(a -> a).
valid_formula(a -> (b -> a)).
valid_formula((a->b)->((b->c)->(a->c))).
valid_formula(a->((a->b)->b)).
valid_formula((a*(a->b))->b).
valid_formula((a->b)->(-b -> -a)).
valid_formula((a->b)->((b->c)->(a->c))).
valid_formula((((((((p3<=> p2) <=> p1) <=> p0) <=> p3) <=> p2) <=> p1) <=> p0)).

% ?- forall(invalid_formula(A), prove(A)).
invalid_formula(a).
invalid_formula((a->b)->a).
invalid_formula((a->b)->(b->a)).
invalid_formula(a -> (b -> c)).
invalid_formula( (-b -> f) * ((b * f) -> -i) * ((i + -b) -> -f) -> (i * f)).
invalid_formula( -((-b -> f) * ((b * f) -> -i) * ((i + -b) -> -f) -> (i * f))).

%!	prove(X) is det.
%
%!	prove(+X, +Opts) is det.
%	X is a propositional formula, Opts is options list.
%	This predicate is auxiliary one for the case in which
%	processing X gets simple but exceptional form which is out of main
%	prover input form.  For the latter case, this predicate tries
%	to show X is a tautology by counting all solutions by
%	transforming X to a DNF form.

% ?- prove((-a) + a).
% ?- prove((-A) + A).
% ?- prove(A -> (p(B)->A)).
% ?- prove(-(a->a)).
% ?- prove(a->a).
% ?- prove((a->b)->(b->a)).
% ?- prove((a->b)->(a->b)).
% ?- prove((a->b)->(a)).
% ?- prove(a->(b->a)).
% ?- prove((a->a)).
% ?- prove(false).
% ?- prove((-(a) * a)).
% ?- prove(-(-(a) * a)).
% ?- prove((-a)+ a).
% ?- prove(a).
% ?- prove((-true)*true).
% ?- prove(-true + true).

prove(X):- prove(X, [time_limit(60)]).
%
prove(X, Opts):-
	(	memberchk(time_limit(Limit), Opts) -> true
	;	Limit = 60
	),
	open_state(S),
	set_compare(ltr_compare, S),
	(	call_with_time_limit(Limit, prove_boole(X, S))
	->	writeln('VALID')
	;	writeln('INVALID')
	),
	close_state(S).

% ?- #S, cnf(-( a + -a ), CNF, S), psa(CNF, S).
prove_boole(X, S):- cnf(-X, CNF, S), !,
	(	CNF > 0	->  resolve(CNF, S)
	;	dnf(X, DNF, S),
		ltr_card(DNF, Count, N, S),
		Count =:= 1 << N
	).

%!	resolve(+P, +S) is det.
%	P is a set of ground complementary-free clauses.
%	True if refutation by resolution prinficple for P is successful.

% ?- open_state(S), ltr_pow([a,b,c], P, S), resolve(P, S), psa(P, S).
% ?- open_state(S), numlist(1, 10, L), ltr_pow(L, P, S), card(P, C, S).
% ?- open_state(S), numlist(1, 100, L), ltr_pow(L, P, S), card(P, C, S).
% ?- time((open_state(S), numlist(1, 100, L), ltr_pow(L, P, S), card(P, C, S))).
% ?- time((open_state(S), numlist(1, 1000, L), ltr_pow(L, P, S), card(P, C, S))).
% ?- N = 100000, time((open_state(S), numlist(1, N, L), ltr_pow(L, P, S),
%	card(P, C, S))), C =:= 2^N, writeln('Refuted.'), fail; true.

resolve(X):- shift(resolve(X)).
%
resolve(X, S):- zdd_has_1(X, S), !.     % The empty clause found.
resolve(X, S):- X > 1,
	cofact(X, t(A, L, R), S),
	(	A = -(_) -> resolve(L, S)		% Negative pure literal rule.
	;	(	cofact(L, t(B, U, V), S),	% Pure literal rule.
			(	B = -(A)
			->  ltr_merge(V, R, M, S),  % Resolution + Tautology rule.
				ltr_join(U, M, W, S),
				resolve(W, S)			% Updated set of clauses.
			;	resolve(L, S)			% Posituve pure literal rule.
			)
		)
	).

% ?- zdd_sat(a), zdd_sat(b), zdd_sat(a=b), zdd_sat(a=c), zdd_sat(-(b=c)), zdd_sat_count(C).
% ?- N = 100, open_state(S), numlist(1, N, L), ltr_pow(L, P, S),
%	zeval(set(L), A, S), zdd_subtr(P, A, Q, S), card(Q, C, S),
%	writeln(C), resolve(Q, S).
%@ 1267650600228229401496703205375

% ?- N = 40, time((open_state(S), numlist(1, N, L), ltr_pow(L, P, S),
%	zeval(set(L), A, S), zdd_subtr(P, A, Q, S), card(Q, C, S),
%	writeln(C), resolve(Q, S))).
%@ 1099511627775
%@ % 616,170,135 inferences, 52.360 CPU in 59.584 seconds (88% CPU, 11767885 Lips)

% ?- N = 100000, time((open_state(S), numlist(1, N, L), ltr_pow(L, P, S),
%	writeln(C), resolve(Q, S))).
%@ 1099511627775
%@ % 616,170,135 inferences, 52.360 CPU in 59.584 seconds (88% CPU, 11767885 Lips)

atom_index(X, Y, S):- memo(atom_index(X)-Y, S),
	(	nonvar(Y) -> true
	;	get_key(atom_index, Y, S),
		memo(index_atom(Y)-X, S),
		Y0 is Y + 1,
		set_key(atom_index, Y0, S)
	).
%
numbervars_index(X):- shift(numbervars_index(X)).
%
numbervars_index(X, S):- term_variables(X, Vs),
	(	Vs = [] -> true
	;	get_key(atom_index, I, S),
		numbervars(Vs, I, J),
		index_dollar_var_list(Vs, I, S),
		set_key(atom_index, J, S)
	).
%
index_dollar_var_list([], _, _):-!.
index_dollar_var_list([V|Vs], K, S):-
	memo(atom_index(V)-K, S),
	memo(index_atom(K)-V, S),
	K0 is K + 1,
	index_dollar_var_list(Vs, K0, S).

% ?- zdd((dnf(-(a=:=b), X), psa(X))).
% ?- zdd((dnf((a=\=b), X), psa(X))).
boole_nnf(F, X, Y):- shift(boole_nnf(F, X, Y)).
%
boole_nnf(F, X, Y, S):- numbervars_index(X, S),
	basic_boole(F, X, Z, S),
	once(neg_normal(Z, Y)).

% convert if SW=1
basic_boole(0, K, K, _):- integer(K), !.
basic_boole(_, 0, 0, _):-!.
basic_boole(_, 1, 1, _):-!.
basic_boole(1, K, _, _):- integer(K), !,
	throw('integer > 2 found'(K)).
basic_boole(Flag, *(L), F, S):-!, basic_boole_vector(L, *, F0, S),
	basic_boole(Flag, F0, F, S).
basic_boole(Flag, +(L), F, S):-!, basic_boole_vector(L, +, F0, S),
	basic_boole(Flag, F0, F, S).
basic_boole(Flag, X, Y, S):- X=..[F|Xs],
	length(Xs, N),
	boole_op(N, As, Fs, Y),
	memberchk(F, Fs),
	!,
	basic_boole_list(Flag, Xs, As, S).
basic_boole(1, X, Y, S):-!, atom_index_chk(X, Y, S).
basic_boole(0, X, X, _S).
%
basic_boole_list(_, [], [], _):-!.
basic_boole_list(Flag, [X|Xs], [Y|Ys], S):- basic_boole(Flag, X, Y, S),
	basic_boole_list(Flag, Xs, Ys, S).
%
basic_boole_vector([], +, 0, _):-!.
basic_boole_vector([], *, 1, _):-!.
basic_boole_vector([X|Xs], F, Y, S):-
	basic_boole_vector(Xs, F, Z, S),
	Y=..[F, X, Z].
%
boole_op(0, [], [true], 1).
boole_op(0, [], [false], 0).
boole_op(1, [X], [-, ~, \+, not], -X).
boole_op(2, [X, Y], [and, &, /\, ',', *], X*Y).
boole_op(2, [X, Y], [or, '|', \/, ';', +], X+Y).
boole_op(2, [X, Y], [->, imply], -(X)+Y).
boole_op(2, [X, Y], [<-], X + -Y).
boole_op(2, [X, Y], [==, =,  =:=, equiv, ~, <->, <=>], -X* -Y + X*Y).
boole_op(2, [X, Y], [=\=, \=], -X*Y + X* -Y).
boole_op(2, [X, Y], [exor, #], X* -Y + -X *Y).
boole_op(2, [X, Y], [nand], -(X*Y)).

% ?- neg_normal(a->(b->a), X).
% ?- neg_normal(-(a + b + c), X).
neg_normal(-(0), 1).
neg_normal(-(1), 0).
neg_normal(-(-X), Z)	:- neg_normal(X, Z).
neg_normal(-(X+Y), Z)	:- neg_normal(-X* -Y, Z).
neg_normal(-(X*Y), Z)	:- neg_normal(-X+ -Y, Z).
neg_normal(X+Y, X0+Y0)	:- neg_normal(X, X0), neg_normal(Y, Y0).
neg_normal(X*Y, X0*Y0)	:- neg_normal(X, X0), neg_normal(Y, Y0).
neg_normal(X, X).

% ?- zdd((intern(-(a;b;c), X), boole_to_dnf(X, Z),  sets(Z, U), extern(U, Y))).
%%	extern(+X, -Y) is det.
% Convert an internal form X into an external one
% in order to unify Y with the result.

% ?- ltr_zdd((dnf0(A+B, X), sets(X, X0), extern(X0, Y))).
% ?- ltr_zdd((dnf0(X->Y, E), sets(E, E0), extern(E0, A))).
extern(X, Y):-shift(extern(X, Y)).
%
extern(X, Y, S):-integer(X), !,
	(	X < 2  -> Y = X
	;	memo(index_atom(X)-Y, S)
	).
extern(X, X, _):- atomic(X), !.
extern(X, Y, S):- X =.. [F|Xs],
	extern_list(Xs, Ys, S),
	Y =..[F|Ys].
%
extern_list([], [], _).
extern_list([X|Xs], [Y|Ys], S):- extern(X, Y, S),
	extern_list(Xs, Ys, S).

		/*********************************************
		*    Cofact/insert/join/merge on literals    *
		*********************************************/

%!	ltr_cofact(?X, ?Y:t(A, L, R), +S) is det.
%	Bidirectional.
%	If X is given then
%	Y is a triple t(A, L, R) such that
%	A is the minimum literal in X w.r.t. specified literal compare predicate,
%	L = { U in X | not ( A in U ) },
%	R = { V \ {A}  | V in X, A in V }.
%	If Y is given then
%	X = union of L and A*R = { unionf of U and {A} | U in R }.
%	Due to `ltr_cofact/3` and `itr_insert/4`, it is guaranteed that
%	every clause is complentary-free ( no complementary pair ).
%
ltr_cofact(X, Y):- shift(ltr_cofact(X, Y)).

%
ltr_cofact(Z, Y, S):- nonvar(Z), !, cofact(Z, Y, S).
ltr_cofact(Z, t(A, V, U), S):- U > 1, !,
	cofact(U, t(B, L, _), S),
	(	complementary(A, B)
	->  ltr_cofact(Z, t(A, V, L), S)
	;	cofact(Z, t(A, V, U), S)
	).
ltr_cofact(Z, T, S):- cofact(Z, T, S).

%!	ltr_insert(+X, +Y, -Z, +S) is det.
%	Insert a literal X into each set of literals
%	in Y, and the result is unified with Z.
%	Z = { union of U and {X} |
%			U in Y, and
%			the complemen of X is not in U }.

% PASS.
% ?- zdd((cnf(a*b, X), ltr_insert(c, X, Y), sets(Y, S))).
% ?- zdd((dnf((-a)*b, X), ltr_insert(a, X, Y), sets(Y, S))).
% ?- zdd((ltr_insert(a, 1, X), sets(X, S))).
% ?- zdd((ltr_insert(a, 1, X),	ltr_insert(b, X, Y), sets(Y, S))).
% ?- zdd((ltr_insert(a, 1, X), ltr_insert(b, X, Y), ltr_insert(-b, Y, Z),
%	sets(Z, S))).
% ?- zdd((X<<set([a]), ltr_insert(b, X, Y), psa(Y))).
% ?- zdd((X<<set([a]), ltr_insert(-a, X, Y), psa(Y))).
% ?- zdd((X<<set([b]), ltr_insert(a, X, Y), psa(Y))).
% ?- zdd((X<<set([a]), ltr_insert(-a, X, Y), psa(Y))).
% ?- zdd((X<<set([-a]), ltr_insert(a, X, Y), psa(Y))).
% ?- zdd((X<<{[a,b,c],[u,v]}, ltr_insert(-b, X, Y), psa(Y))).
% ?- zdd((X<<{[a,b,c],[u,v]}, ltr_insert(b, X, Y), psa(Y))).
% ?- zdd((X<<{[a,b,c],[u,v]}, ltr_insert(u, X, Y), psa(Y))).
% ?- zdd((X<<{[a]}, ltr_insert(a, X, Y), psa(Y))).
% ?- zdd((X<<{[a,b,c]}, ltr_insert(-b, X, Y), psa(Y))).
% ?- zdd((X<<{[a,b],[c]}, ltr_insert(-b, X, Y), psa(Y))).
% ?- zdd((X<<{[a,-b],[c]}, ltr_insert(b, X, Y), psa(Y))).
% ?- zdd((X<<{[a,b]}, ltr_insert(-b, X, Y), psa(Y))).
ltr_insert(X, Y, Z):- shift(ltr_insert(X, Y, Z)).
%
ltr_insert(_, 0, 0, _):-!.
ltr_insert(A, 1, J, S):-!, zdd_singleton(A, J, S).
ltr_insert(A, I, J, S):- memo(ltr_insert(A,I)-J, S),
	(	nonvar(J)	-> true
	;	cofact(I, t(B, L, R), S),
		ltr_insert(A, L, L0, S),
		ltr_insert(A, B, R, R0, S),
		ltr_join(L0, R0, J, S)
	).
%
ltr_insert(A, B, _, 0, _):-	complementary(A, B), !.
ltr_insert(A, B, X, Y, S):-	ltr_compare(C, A, B),
		(	C = (<)	->
			zdd_cons(Y0, B, X, S),
			zdd_cons(Y, A, Y0, S)
		;   C = (=)	-> zdd_cons(Y, A, X, S)
		;   ltr_insert(A, X, Y0, S),
			zdd_cons(Y, B, Y0, S)
		).

%!	ltr_join(+X, +Y, -Z, +S) is det.
%	Unify Z with a ZDD that represents the union of the ZDD X and Y
%	as a family of sets (complementary_free clauses) of literals.
ltr_join(X, Y, Z):- shift(ltr_join(X, Y, Z)).
%
ltr_join(X, X, X, _):-!.
ltr_join(0, X, X, _):-!.
ltr_join(X, 0, X, _):-!.
ltr_join(1, _, 1, _):-!.
ltr_join(_, 1, 1, _):-!.
ltr_join(X, Y, Z, S):-
	(	X<Y -> memo(ltr_join(X,Y)-Z, S)
	;	memo(ltr_join(Y,X)-Z, S)
	),
	(	nonvar(Z) -> true %, write(.)
	; 	cofact(X, t(A, L, R), S),
		cofact(Y, t(A0, L0, R0), S),
		ltr_compare(C, A, A0),
		(	C = (=)	->
			ltr_join(R, R0, U, S),
			ltr_join(L, L0, V, S),
			cofact(Z, t(A, V, U), S)
		;	C = (<) ->
			ltr_join(L, Y, U, S),
			cofact(Z, t(A, U, R), S)
		; 	ltr_join(L0, X, U, S),
			cofact(Z, t(A0, U, R0), S)
		)
	).

%!	ltr_merge(+X, +Y, -Z, +S) is det.
%	Z = { U |	A in X, B in Y,
%				U is the union of A and B,
%				U is complementary-free
%		}.
%

ltr_merge(X, Y, Z):- shift(ltr_merge(X, Y, Z)).
%
ltr_merge(0, _, 0, _):-!.
ltr_merge(_, 0, 0, _):-!.
ltr_merge(1, X, X, _):-!.
ltr_merge(X, 1, X, _):-!.
ltr_merge(X, Y, Z, S):-
	(	X < Y -> memo(ltr_merge(X,Y)-Z, S)
	;	memo(ltr_merge(Y,X)-Z, S)
	),
 	(	nonvar(Z) -> true
	;	cofact(Y, t(A, L, R), S),
		ltr_merge(X, R, U, S),
		ltr_merge(X, L, V, S),
		ltr_insert(A, U, U0, S),
		ltr_join(V, U0, Z, S)
	).

		/******************************************
		*     Auxiliary operations on literals    *
		******************************************/

%!	complementay(+A, -B) is det.
%   B is unfied with the complement of B.
ltr_invert(-A, A):-!.
ltr_invert(A, -A).

%!	complementay(+A, +B) is det.
%	True if A and B are form a complementary to each other.
% ?- complementary(a, -a).
% ?- complementary(-a, b).
complementary(-(A), A):-!.
complementary(A, -(A)).

%!	ltr_pow(+As, -P, +S) is det.
%	As is a list of atoms.
%	P is unified with the set of all complementary-free clauses C
%	such that every B in C is complementary with some A in As, and
%	for all A in As  A is complementary with some B in C.

% ?- open_state(S), ltr_pow([a, b], X, S), card(X, C, S).
% ?- open_state(S), numlist(1, 100, L), ltr_pow(L, X, S), card(X, C, S).
ltr_pow([], 1, _).
ltr_pow([A|As], P, S):- ltr_pow(As, Q, S),
	ltr_insert(A, Q, R, S),
	ltr_insert(-A, Q, L, S),
	ltr_join(L, R, P, S).

%!	ltr_append(+X, +Y, -Z, +S) is det.
%	X is a list of atoms and  zdd Y is a family of cluases.
%	Z is unified with zdd
%	{ U |	U is union of X and V in Y
%			such that U is complementary-free }.

% ?- zdd((ltr_append([-b, a, -d, c], 1, X), psa(X))).
% ?- zdd((ltr_append([-b, -d, c], 1, X), psa(X))).
% ?- zdd((ltr_append([-b, -d, c], 1, X), psa(X))).
ltr_append(X, Y, Z):- shift(ltr_append(X, Y, Z)).
%
ltr_append([], X, X, _).
ltr_append([A|As], X, Y, S):- ltr_append(As, X, X0, S),
	 	ltr_insert(A, X0, Y, S).

%!	ltr_set(+X, -Y, S) is det.
%	Euivalent to ltr_append(X, 1, Y, S).
% ?- zdd((X<<(ltr_set([a])+ltr_set([b])), psa(X))).
% ?- zdd((X<<(ltr_set([-a])+ltr_set([b])), psa(X))).
% ?- zdd((X<<(ltr_set([-a])+ltr_set([a])), psa(X))).

ltr_set(X, Y):- shift(ltr_set(X, Y)).
%
ltr_set(X, Y, S):- ltr_append(X, 1, Y, S).

%!	ltr_negate(+X, -Y, +S) is det.
%	X is a zdd of a family of clauses.
%	Y is unified with the zdd which is the choice of
%	the complement of X.

% ?- zdd((set_compare(ltr_compare), zdd_sort([-(3), 2, 3], L))).
% ?- zdd((set_compare(ltr_compare), X<<dnf(a*b), ltr_negate(X, Y), sets(Y, Y0))).
% ?- zdd((set_compare(ltr_compare), X<<dnf(a), ltr_negate(X, Y), sets(Y, Y0))).
% ?- zdd((set_compare(ltr_compare), X<<dnf(a*b), ltr_negate(X, Y), ltr_negate(Y, Z), sets(Y, Y0), sets(Z, Z0))).
% ?- zdd((set_compare(ltr_compare), X<<dnf(a+b), ltr_negate(X, Y), ltr_negate(Y, Z), sets(Y, Y0), sets(Z, Z0))).
ltr_negate(X, Y):- shift(ltr_negate(X, Y)).
%
ltr_negate(X, Y, S):- ltr_complement(X, X0, S),
		ltr_choices(X0, Y, S).

%!	ltr_complement(+X, -Y, +S) is det.
%	X is a zdd for a family of clauses.
%	Y is unified with a zdd of clauses C
%	such that for some D in X, if literal A in C then
%	the complement of A is in D, and if literal A in D then
%	the complement of A is in C.

% ?- zdd((X<<{[a, -b],[-b],[b, c]}, ltr_complement(X, Y), psa(Y))).
% ?- zdd((X<<{[-b],[-b],[b, c]}, psa(X), ltr_complement(X, Y),
%	nl, nl,  psa(Y))).
ltr_complement(X, Y):- shift(ltr_complement(X, Y)).
%
ltr_complement(X, X, _):- X<2, !.
ltr_complement(I, Y, S):- memo(ltr_complement(I)-Y, S),
	(	nonvar(Y) -> true
	; 	cofact(I, t(A, L, R), S),
		ltr_complement(L, L0, S),
		ltr_complement(R, R0, S),
		ltr_invert(A, NA),
		ltr_insert(NA, R0, R1, S),
		ltr_join(L0, R1, Y, S)
	).


		/************************************
		*  Convert Boolean Form to DNF/CNF  *
		************************************/

%!	dnf(+X, -Y, +S) is det.
%	Y is a disjuntive normal form of a boolean formula X.

%!	dnf0(+X, -Y, +S) is det.
%	Y is a disjuntive normal form of a boolean formula X.
%	All boolean atoms appearing in X are converted to integers > 1.

% ?- zdd((dnf(a, X), psa(X), card(X, C))).
% ?- zdd((dnf0(a, X), psa(X), card(X, C))).
% ?- zdd((dnf(-(a->b), X), extern(X, X0), psa(X))).
% ?- ltr_zdd((dnf0(-(a->b), X), extern(X, X0), writeln(X0), psa(X))).

dnf(X, Y):- shift(dnf(X, Y)).
%
dnf0(X, Y):- shift(dnf0(X, Y)). % atom => integer index

% convert if SW=1.
dnf(X, Y, S):- dnf(0, X, Y, S).
%
dnf0(X, Y, S):- dnf(1, X, Y, S). % atom => integer index
%
dnf(F, X, Y, S):- boole_nnf(F, X, Z, S), nnf_dnf(Z, Y, S).

%!	nnf_dnf(+E, -Z, +S) is det.
%   E is a nnf (Negation Normal Form). Z is a dnf for E.

nnf_dnf(X * Y, Z, S):-!, memo(dnf(X*Y)-Z, S),
	(	nonvar(Z) -> true
	;	nnf_dnf(X, U, S),
		nnf_dnf(Y, V, S),
		ltr_merge(U, V, Z, S)
	).
nnf_dnf(X + Y, Z, S):-!, memo(dnf(X+Y)-Z, S),
	(	nonvar(Z) -> true
	; 	nnf_dnf(X, U, S),
		nnf_dnf(Y, V, S),
		ltr_join(U, V, Z, S)
	).
nnf_dnf(X, X, _):- integer(X), X<2, !.
nnf_dnf(X, Y, S):- zdd_singleton(X, Y, S).

%!	atom_index_chk(+X, ?I, +S) is det.
%	Unify I with the index of an atom X. When `$VAR(_)` appears
%	inside X as a subterm, throws exception (as unexpected form of
%	atom).

atom_index_chk('$VAR'(K), I, S):-!, atom_index('$VAR'(K), I, S).
atom_index_chk(X, I, S):-
	(	dollar_var_found(X) -> throw(dollar_var_found(X))
	;	atom_index(X, I, S)
	).
%
dollar_var_found('$VAR'(_)):-!.
dollar_var_found(X):-compound(X),
	X =..[_|Xs],
	member(A, Xs),
	dollar_var_found(A).

%!	index_atoms(+X, -Y, +Pred, +S) is det.
%	Y is unified with a substitution of X whose all atoms appearing
%	in X are substituted with integer indecies.
%	Fails if checking by Pred fails for some atom.

:- meta_predicate  inddex_atoms(?, ?, 1, ?).
index_atoms(K, K, _, _):- K<2, !.  % (K < 2 is reserved).
index_atoms(K, _, _, _):-!, throw('integer>2 found'(K)).
index_atoms(A, I, Pred, S):- call(Pred, A), !, atom_index(A, I, S).
index_atoms(X, Y, Pred, S):- X=..[F|Xs],
	index_atoms_list(Xs, Ys, Pred, S),
	Y=..[F|Ys].
%
index_atoms_list([], [], _, _).
index_atoms_list([X|Xs], [Y|Ys], Pred, S):-
	index_atoms(X, Y, Pred, S),
	index_atoms_list(Xs, Ys, Pred, S).

% ?- zdd((cnf(a, X), S<<sets(X))).
% ?- ltr_zdd((cnf(a*b, X), S<<sets(X))).
% ?- zdd((X<<cnf(-a), Y<<sets(X))).
% ?- zdd((X<<cnf(a+b), Y<<sets(X))).
% ?- zdd((X<<cnf(+([a,b,c])), Y<<sets(X))).
% ?- zdd((X<<cnf(*([a,b,c])), Y<<sets(X))).
% ?- zdd((X<<dnf(+([a,b,c])), Y<<sets(X))).
% ?- zdd((X<<dnf(*([a,b,c])), Y<<sets(X))).
% ?- zdd((cnf(a, X), psa(X))).
% ?- zdd((cnf(-a, X), psa(X))).
% ?- zdd((cnf(a+b, X), psa(X))).
% ?- zdd((cnf(a+b+(-c), X), psa(X))).
% ?- zdd((cnf(-a * a,  X), psa(X))).
% ?- zdd((cnf(a->a,  X), psa(X))).
% ?- zdd((cnf(-a * a,  X), psa(X))).
% ?- zdd((cnf( a + a,  X), psa(X))).
% ?- zdd((cnf(-(a*b), X), psa(X))).
% ?- zdd((cnf(*([a,b,c]), X), psa(X))).
% ?- #S, N = 10, numlist(2, N, Ns), cnf(+(Ns), X, S), psa(X, S), ltr_card(X, C, K, S).
% ?- #S, N = 10, numlist(2, N, Ns), cnf(*(Ns), X, S), ltr_card(X, C, K, S).
% ?- zdd((cnf0(a+b, X), psa(X))).
% ?- ltr_zdd((dnf0(-(a->b), X), psa(X))).
% ?- ltr_zdd((dnf0(a, X), psa(X))).
% ?- ltr_zdd((dnf(a, X), psa(X))).

cnf(X, Y):- shift(cnf(X, Y)).

% ?- ltr_zdd((cnf0(-(a->b), X), psa(X))).
% ?- ltr_zdd((cnf0(a, X), psa(X))).
% ?- ltr_zdd((cnf(-(a->b), X), psa(X))).
% ?- ltr_zdd((cnf(a, X), psa(X))).

cnf(X, Y):- shift(cnf(X, Y)).
cnf0(X, Y):-shift(cnf0(X, Y)).   % atom => integer index

cnf(X, Y, S):- cnf(0, X, Y, S).
cnf0(X, Y, S):- cnf(1, X, Y, S). % atom => integer index
%
cnf(F, X, Y, S):- boole_nnf(F, X, Z, S), nnf_cnf(Z, Y, S).
%
nnf_cnf(X, Y):- shift(nnf_cnf(X,Y)).
%
nnf_cnf(X * Y, Z, S):-!, memo(cnf(X*Y)-Z, S),
	(	nonvar(Z)->true	%, write(.)   % many hits.
	;	nnf_cnf(X, U, S),
		nnf_cnf(Y, V, S),
		ltr_merge(U, V, Z, S)
	).
nnf_cnf(X + Y, Z, S):-!, memo(cnf(X+Y)-Z, S),
	(	nonvar(Z)->true	%, write(+)  % many hits.
	;	nnf_cnf(X, U, S),
		nnf_cnf(Y, V, S),
		ltr_join(U, V, Z, S)
	).
nnf_cnf(X, X, _):- integer(X), X<2, !.
nnf_cnf(X, Y, S):- zdd_singleton(X, Y, S).

		/*****************************************************
		*     ltr_card/[3,4] Counting solutions of a DNF.    *
		*****************************************************/

% ?- sat_count(a, C, N).
% ?- sat_count(a+b, C, N).
% ?- sat_count((a->b)*(b->c)->(b->c), C, N).
% ?- sat_count((a->b)*(b->c)->(b->c), C, N).

sat_count(X, C, N):- open_state(S),
	set_compare(ltr_compare, S),
	zdd_eval(dnf(X), Y, S),
	sat_count(Y, C, N, S),
	close_state(S).
%
sat_count(X, C, N, S):-
	zdd_boole_atoms(X, Us, S),
	zdd_sort(Us, Vs, S),
	expand_dnf(Vs, X, Y, S),
	card(Y, C, S),
	length(Vs, N).

% ?- ltr_zdd((X<<dnf(a+ -a*b), @zdd_boole_atoms(X, As))), sort(As, Bs).
zdd_boole_atoms(X, P):- shift(zdd_boole_atoms(X, P)).

zdd_boole_atoms(X, [], _):- X<2, !.
zdd_boole_atoms(X, P, S):- memo(bool_atoms(X)-P, S),
	(	nonvar(P) -> true
	; 	cofact(X, t(A, L, R), S),
		zdd_boole_atoms(L, Al, S),
		zdd_boole_atoms(R, Ar, S),
		ltr_var(A, A0),
		union([A0|Al], Ar, P)
	).

		/******************************
		*     find_counter_example    *
		******************************/
% ?- findall_counter_examples(a, X, S), psa(X, S).
% ?- findall_counter_examples((a->b)->a, X, S), psa(X, S).
% ?- findall_counter_examples((a->b)->(b->a), X, S), psa(X, S).
% ?- findall_counter_examples(a->b, X, S), psa(X, S).
% ?- findall_counter_examples((a->b)->(b->a), X, S), psa(X, S).

%!	ltr_open_state(-S) is det.
%	open_state(S) and set `ltr_compare/3` in S as the compare predicate.

ltr_open_state(S):- open_state(S), set_compare(ltr_compare, S).


%!	ltr_card(+X, -C, -N, S) is det.
%	C is the number of solutions of boolean formula X.
%	N is the number of boolean variable appearing in X.

% ?- ltr_zdd((X<< dnf(((a->b) * (b->c) * a -> c)), @ltr_card(X, C))).
% ?- ltr_zdd((X<< dnf(a), psa(X), @ltr_card(X, C))).
% ?- ltr_zdd((X<< dnf(a->a), @ltr_card(X, C))).
% ?- ltr_zdd((X<< dnf(((a->b) * (b->c) * a -> c)), @ltr_card(X, C))).
% ?- ltr_zdd((X<< dnf((a->b)*(b->c)->(a->c)), @ltr_card(X, C))).
% ?- findall(p(J), between(1, 100, J), Ps),
%	time(ltr_zdd((X<< dnf(+Ps), @ltr_card(X, C)))).
% ?- N = 1000, findall(p(J), between(1, N, J), Ps),
%	time(ltr_zdd((X<< dnf(+Ps), @ltr_card(X, C)))),
%	C =:= 2^1000 - 1.
ltr_card(In, Out):- @ltr_card(In, Out).

ltr_card(In, Out, S):- ltr_card(In, Out, _, S).

ltr_card(In, Out, N, S):-
	zdd_boole_atoms(In, Us, S),
	zdd_sort(Us, Vs, S),
	expand_dnf(Vs, In, Y, S),
	length(Vs, N),
	card(Y, Out, S).

% ?- ltr_var(-(5), X).
ltr_var(-(V), V):-!.
ltr_var(V, V).

%!	findall_counter_examples(+In, -Out, +S) is det.
%	Experimental.

% ?- findall_counter_examples(a->b, Out, S), psa(Out,S).
findall_counter_examples(In, Out, S):-
	ltr_open_state(S),
	zdd_eval(dnf(In), InZ, S),
	zdd_boole_atoms(InZ, Us, S),
	zdd_sort(Us, Vs, S),
	expand_prefix_dnf(Vs, 1, All, S),
	expand_dnf(Vs, InZ, Y, S),
	zdd_subtr(All, Y, Out, S).

%!	expand_dnf(+Vs, +X, -Y, +S) is det.
%	With U the set of all atoms appearing in X,
%	Y is the all maximal complemental-free clauses C over the union of Vs
%	and U such that C is an extension of some clause in X.

% ?- ltr_zdd((X<<dnf(a->a), psa(X), @expand_dnf([a], X, Y), psa(Y))).
% ?- ltr_zdd((X<<dnf(a->(b->a)), @expand_dnf([a,b], X, Y), card(Y, C))).
% ?- ltr_zdd((X<<dnf(a->((b->c)->a)), @expand_dnf([a,b,c], X, Y), card(Y, C))).
% ?- ltr_zdd((X<<dnf(a*b->b), psa(X), @expand_dnf([a,b], X, Y), psa(Y))).
% ?- ltr_zdd((X<<dnf((a->b)*(b->c)->(b->c)), psa(X), @expand_dnf([a,b,c], X, Y), psa(Y), card(Y, C))).
% ?- ltr_zdd((X<<dnf((a->b)*(b->c)->(a->c)), psa(X), @expand_dnf([a,b,c], X, Y), psa(Y), card(Y, C))).
% ?- ltr_zdd((X<<dnf(a + -a), psa(X), @expand_dnf([a,b,c], X, Y), psa(Y), card(Y, C))).

expand_dnf([], X, X, _):-!.
expand_dnf(_, 0, 0, _):-!.
expand_dnf(Vs, 1, Y, S):-!, expand_prefix_dnf(Vs, 1, Y, S).
expand_dnf(Vs, X, Y, S):- memo(expand_dnf(X,Vs)-Y, S),
	(	nonvar(Y) -> true	%, write(.) %  Many hits.
	;	cofact(X, t(A, L, R), S),
		ltr_var(A, K),
		divide_ord_list(Vs, K, [], Us, Ws),
		expand_dnf(Us, R, R0, S),
		ltr_insert(A, R0, R1, S),
		expand_left_dnf(K, Us, L, L0, S),
		ltr_join(L0, R1, X0, S),
		expand_prefix_dnf(Ws, X0, Y, S)
	).

%!	divide_ord_list(+L, +A, +U, -T, -W) is det.
%	Divide list L at A into two parts head H and tail T
%	so that reverse of H =  diffrence list W-U.
%	?- divide_ord_list([a,b,c,d,e], c, [u], T, W).
%@ T = [d, e],
%@ W = [b, a, u].
%	?- divide_ord_list([a,b,c, d], e, [], X, Y).   % false

divide_ord_list([V|Vs], V, Us, Vs, Us):-!.
divide_ord_list([V|Vs], U, Us, Ws, Ps):- ltr_compare(<, V, U),
	divide_ord_list(Vs, U, [V|Us], Ws, Ps).

%!	expand_prefix_dnf(+Vs, +X, -Y, +S) is det.
%	With U the set of all atoms appearing in X,
%	Y is the all maximal complemental-free clauses C over the union of Vs
%	and U such that C is consistent with all clauses in X.

% ?- ltr_zdd((@expand_prefix_dnf([c,d], 1, Y), @expand_prefix_dnf([a,b], Y, Z), psa(Z), card(Z, C))).
expand_prefix_dnf(Vs, X, Y, S):- zdd_sort(Vs, OrdVs, S),
	expand_ord_prefix_dnf(OrdVs, X, Y, S).

%
expand_ord_prefix_dnf([], X, X, _):-!.
expand_ord_prefix_dnf([V|Vs], X, Y, S):-
	expand_ord_prefix_dnf(Vs, X, X0, S),
	ltr_insert(V, X0, X1, S),
	ltr_insert(-V, X0, X2, S),
	ltr_join(X1, X2, Y, S).

%!	expand_left_dnf(+K, +Vs, +X, -Y, +S) is det.
%   X is assumed to have no occurrence of positive literal K
%   (negative on may appear).
%	With U the set of all atoms appearing in X,
%	Y is the all maximal complemental-free clauses C over the union of Vs
%	and U such that C is an extension of some clause in X and
%	has not the postive literal K.

expand_left_dnf(_, _, 0, 0, _):-!.
expand_left_dnf(K, Us, 1, Y, S):-!, expand_prefix_dnf([K|Us], 1, Y, S).
expand_left_dnf(K, Us, X, Y, S):- cofact(X, t(A, L, R), S),
	ltr_var(A, J),
	(	K = J ->
		expand_dnf(Us, R, R0, S),
		ltr_insert(A, R0, R1, S)
	;	divide_ord_list([K|Us], J, [], Vs, Ws),
		expand_dnf(Vs, R, R0, S),
		ltr_insert(A, R0, Q, S),
		expand_prefix_dnf(Ws, Q, R1, S)
	),
	expand_dnf([K|Us], L, L0, S),
	ltr_join(L0, R1, Y, S).

		/*******************************
		*     Filter on cardinality    *
		*******************************/

%!	card_filter_gte(+N, +X, -Y, +S) is det.
%	Y = { A in X | #A >= N }, where #A is the number of elements
%	of A.

% ?- zdd((X<<pow([a,b,c]), @(card_filter_gte(2, X, Y)), psa(Y))).
card_filter_gte(0, X, X, _):- !.	% gte: greater than or equal
card_filter_gte(1, X, Y, S):- !, zdd_subtr(X, 1, Y, S).
card_filter_gte(_, X, 0, _):- X<2, !.
card_filter_gte(N, X, Y, S):- memo(filter_gte(N,X)-Y, S),
	(	nonvar(Y) -> true		% many hits.
	;   cofact(X, t(A, L, R), S),
		N0 is N - 1,
		card_filter_gte(N, L, L0, S),
		card_filter_gte(N0, R, R0, S),
		cofact(Y, t(A, L0, R0), S)
	).

%!	card_filter_lte(+N, +X, -Y, +S) is det.
%	Y = { A in X | #A =< N }, where #A is the
%	cardinality of A.

% ?- zdd((X<<pow([a,b,c]), @(card_filter_lte(2, X, Y)), psa(Y))).
card_filter_lte(0, X, Y, S):-    % lte: less than or equal
	(	zdd_has_1(X, S) -> Y = 1
	;	Y = 0
	).
card_filter_lte(_, X, X, _):- X<2, !.
card_filter_lte(N, X, Y, S):- memo(filter_lte(N,X)-Y, S),
	(	nonvar(Y) -> true		% many hits.
	;    cofact(X, t(A, L, R), S),
		N0 is N - 1,
		card_filter_lte(N, L, L0, S),
		card_filter_lte(N0, R, R0, S),
		cofact(Y, t(A, L0, R0), S)
	).

%!	card_filter_between(+I, +J, +X, -Y, +S) is det.
%	Y = { A in X | I =< #A =< J }.

% ?- time(( N = 100, numlist(1, N, Ns), zdd((X<<pow(Ns), @(card_filter_between(50, 51, X, Y)), card(Y, C))))).
card_filter_between(I, J, X, Y, S):-
	card_filter_gte(I, X, Z, S),
	card_filter_lte(J, Z, Y, S).

% ?- time(( N = 100, numlist(1, N, Ns), zdd((X<<pow(Ns), @(card_filter_between_by_meet(50, 51, X, Y)), card(Y, C))))).
%@ % 924,870 inferences, 0.060 CPU in 0.064 seconds (94% CPU, 15335522 Lips)
%@ N = 100,
%@ Ns = [1, 2, 3, 4, 5, 6, 7, 8, 9|...],
%@ X = 101,
%@ Y = 7600,
%@ C = 199804427433372226016001220056 .
card_filter_between_by_meet(I, J, X, Y, S):-
	card_filter_gte(I, X, Z, S),
	card_filter_lte(J, X, U, S),
	zdd_meet(Z, U, Y, S).


		/*********************************************************
		*     Solve boolean equations and verify the solution    *
		*********************************************************/

%!	dnf_subst(+V, +T, +X, -Y, +S) is det.
%	Y is the ZDD obtained by substituting each
%	occurrence of atom V in X with T.

% ?-ltr_zdd((X<<dnf(-a), Z<<dnf(b), @(dnf_subst(a, Z, X, Y)), psa(X), psa(Y))).
% ?-ltr_zdd((X<<dnf(-a), Z<<dnf(-b), @(dnf_subst(a, Z, X, Y)), psa(X), psa(Y))).
% ?-ltr_zdd((X<<dnf(a+b), Z<<dnf(b), @(dnf_subst(a, Z, X, Y)), psa(X), psa(Y))).

dnf_subst(_, _, X, X, _):- X < 2, !.
dnf_subst(V, D, X, Y, S):-
	cofact(X, t(A, L, R), S),
	dnf_subst(V, D, L, L0, S),
	once( A = -(U); U = A ),
	ltr_compare(C, V, U),
	(	C = (<) -> Y = X
	;	(	C = (=) ->
			(	A = (-V) ->
				ltr_negate(D, D0, S),
				ltr_merge(D0, R, R0, S)
			;	A = V ->
				ltr_merge(D, R, R0, S)
			)
		;	dnf_subst(V, D, R, R1, S),
			ltr_insert(A, R1, R0, S)
		),
		ltr_join(L0, R0, Y, S)
	).

%!	solve_boole_with_check(+Exp, +Vs, +Ps, +S) is det.
%	Solve bolean equation equation expressed as Exp
%	with unknown variables Vs and free parameters Ps.
%	Check that the result is really solutions.
%	Ref. Hosoi.

%  exists x  s.t. C0*(L,-x) + C1*x = 1
% 	if and only if
%  C0 + C1 = 1  satisfiable
%  x = a*C1 + (-C0).

% ?- solve_boole_with_check(x*y=a, [x,y], [u,v], S).
% ?- solve_boole_with_check(x*y*z=a, [x,y,z], [u,v,w], S).
% ?- solve_boole_with_check(x*y*z*u=a, [x,y,z,u], [l, m, n, o], S).

solve_boole_with_check(Exp, Xs, Ps, S):- open_state(S),
	set_compare(ltr_compare, S),
	zdd_eval(dnf(Exp), E, S),
	solve_boolean_euations(E, Xs, Ps, Sols, S),
	eliminate_variables(E, Sols, Cond, S),
	!,
	dnf_valid_check_by_count(Cond, S).

% ?- ltr_zdd((X<<dnf0((-a)* -v + (-a)*v + (-a)* -u +a), @ltr_card(X, C))).
% ?- prove((-a)* -v + (-a)*v + (-a)* -u +a).
% ?- ltr_zdd((E<<dnf(x*y=a), @solve_boolean_euations(E, [x,y], [u,v], Sols), @solutions_to_sets(Sols, Sols0))).
% ?- ltr_zdd((E<<dnf(x*y=a), @solve_boolean_euations(E, [x,y,z], [u,v,w], Sols), @solutions_to_sets(Sols, Sols0))).

solve_boolean_euations(_, [], _, [], _):-!.
solve_boolean_euations(E, [X|Xs], [A|As], [X=Sol|Sols], S):-
	solve_boolean_euations_basic(E, X, A, Sol0, E0, S),
	solve_boolean_euations(E0, Xs, As, Sols, S),
	dnf_subst_list(Sols, Sol0, Sol, S).

% ?- ltr_zdd((E<<dnf(a*b+c*(-a)+b), @(solve_boolean_euations_basic(E, a, u, Sol, Cond)), psa(E), psa(Sol), psa(Cond))).
solve_boolean_euations_basic(E, X, A, Sol, Cond, S):-
	dnf_subst(X, 1,  E, C1, S),
	dnf_subst(X, 0,  E, C0, S),
	ltr_join(C0, C0, Cond, S),
	ltr_negate(C0, NC0, S),
	ltr_insert(A, C1, AC1, S),
	ltr_join(NC0, AC1, Sol, S).
%
dnf_subst_list([], E, E, _).
dnf_subst_list([X=A|Eqs], E, F, S):-
	dnf_subst(X, A, E, E0, S),
	dnf_subst_list(Eqs, E0, F, S).
%
solutions_to_sets([], [], _).
solutions_to_sets([X = E|Sols], [X = E0|Sols0], S):-
	sets(E, E0, S),
	solutions_to_sets(Sols, Sols0, S).

% ?- ltr_zdd((E<<dnf(x*y=a), @solve_boolean_euations(E, [x,y,z], [u,v,w], Sols), @solutions_to_sets(Sols, Sols0), @eliminate_variables(E, Sols, Cond), sets(Cond, Check))).

% ?- ltr_zdd((E<<dnf(x*y=a), @solve_boolean_euations(E, [x,y], [u,v], Sols), @solutions_to_sets(Sols, Sols0), @eliminate_variables(E, Sols, Cond), sets(Cond, Check), psa(Cond), @dnf_valid_check_by_count(Cond))).

% ?- ltr_zdd((E<<dnf(x*y=a), @solve_boolean_euations(E, [x,y], [u,v], Sols), @solutions_to_sets(Sols, Sols0), @eliminate_variables(E, Sols, Cond), sets(Cond, Check), psa(Cond))).

eliminate_variables(Exp, Eqns, Cond, S):-
	apply_subst_list(Eqns, Exp, Cond, S).
%
apply_subst_list([], E, E, _).
apply_subst_list([X=U|Eqns], E, E0, S):-
	dnf_subst(X, U, E, E1, S),
	dnf_subst_list(Eqns, E1, E0, S).

% Note that boolean variable index 0, 1 are reserved.
dnf_valid_check_by_count(X, S):-
	ltr_card(X, C, N, S),
	(	C is 1<< N -> writeln('Solved and Verified.')
	;	writeln('Solved but NOT verified.')
	).

		/***********************************
		*  operations on  reducible zdd    *
		***********************************/

% ?- #(S), bdd_append([x, a, y, b], 1, Y, S), psa(Y, S).
% ?- #(S), bdd_append([a, a, a, a], 1, Y, S), psa(Y, S).
% ?- #(S), bdd_append([b, b], 1, Y, S), bdd_append([b,b], Y, Z, S), psa(Z, S).
bdd_append([], X, X, _).
bdd_append([X|Y], Z, U, S):-
	bdd_append(Y, Z, Z0, S),
	cofact(U, t(X, 0, Z0), S).

% ?- #(S), bdd_list([b, b], Y, S), bdd_<append([b,b], Y, Z, S), psa(Z, S).
% ?- #(S), bdd_list([b, b], Y, S), bdd_list(X,  Y, S).
bdd_list(List, X, S):- var(X), !, bdd_append(List, 1, X, S).
bdd_list(List, X, S):- get_bdd_list(X, List, S).

% ?- #(S), bdd_append([x, a, y, b], 1, Y, S), psa(Y, S), get_bdd_list(Y, L, S).
% ?- #(S), bdd_append([a, a, a, a], 1, Y, S), psa(Y, S), get_bdd_list(Y, L, S).
get_bdd_list(1, [], _):-!.
get_bdd_list(X, [A|As], S):- X>1, zdd_cons(X, A, X0, S),
	get_bdd_list(X0, As, S).

% ?-  #(S), bdd_append([a,a,b,b], 1, X, S),
%	bdd_append([1,1,2,2], 1, Y, S),
%	bdd_zip(X, Y, Z, S), psa(Z, S).
bdd_zip(0, _, 0, _):-!.
bdd_zip(_, 0, 0, _):-!.
bdd_zip(1, _, 1, _):-!.
bdd_zip(_, 1, 1, _):-!.
bdd_zip(X, Y, Z, S):- zdd_cons(X, A, X0, S),
	zdd_cons(Y, B, Y0, S),
	bdd_zip(X0, Y0, Z0, S),
	zdd_cons(Z, A-B, Z0, S).

% ?- #(S), bdd_append([a,b,a,b], 1, X, S), bdd_normal(X, Y, S), psa(Y, S).
bdd_normal(X, X, _):- X<2, !.
bdd_normal(X, Y, S):- cofact(X, t(A, L, R), S),
	bdd_normal(L, L0, S),
	bdd_normal(R, R1, S),
	zdd_insert(A, R1, R0, S),
	zdd_join(L0, R0, Y, S).

% ?- #(S), bdd_append([a,b,a,b], 1, X, S),
%	bdd_append([a,b, c, a,b, c], 1, Y, S),
%	bdd_append([b, a, b, a, a,b ], 1, Z, S),
% bdd_list_normal([X, Y, Z], R, S), psa(R, S).

bdd_list_normal([], 0, _).
bdd_list_normal([A|As], R, S):- bdd_list_normal(As, R0, S),
	bdd_normal(A, A0, S),
	zdd_join(A0, R0, R, S).

%
bdd_insert(_, 0, 0, _):-!.
bdd_insert(A, 1, X, S):-!, zdd_singleton(A, X, S).
bdd_insert(A, X, Y, S):- cofact(X, t(B, L, R), S),
	bdd_insert(A, L, L0, S),
	bdd_append([A,B], R, R0, S),
	zdd_join(L0, R0, Y, S).

% Remark: zdd_join, zdd_meet, zdd_subtr, which does not use zdd_insert,
% also work for bddered zdd.

% ?- #(S), X<<pow([a]), Y<<pow([b]), bdd_merge(X, Y, Z, S), zdd_subtr(Z, Y, U, S),
%  psa(X, S), psa(Y, S), psa(Z, S), psa(U, S).
% ?- #(S), X<<pow([a, b]), Y<<pow([b,c]), bdd_merge(X, Y, Z, S),
%	psa(Z, S), card(Z, C, S).
% ?- #(S), numlist(1, 10, Ns), X<<pow(Ns), Y<<pow(Ns), bdd_merge(X, Y, Z, S),
%	card(Z, C, S).
% ?- #(S), numlist(1, 10, Ns), X<<pow(Ns), Y<<pow(Ns), zdd_merge(X, Y, Z, S),
%	card(Z, C, S), card(X, D, S).
bdd_merge(0, _, 0, _):-!.
bdd_merge(_, 0, 0, _):-!.
bdd_merge(1, X, X, _):-!.
bdd_merge(X, 1, X, _):-!.
bdd_merge(X, Y, Z, S):- memo(bdd_merge(X, Y)-Z, S),
	(	nonvar(Z) -> true		% , write(.)	% So so frequently hits.
	;	cofact(X, t(A, L, R), S),
		bdd_merge(L, Y, L0, S),
		bdd_merge(R, Y, R1, S),
		zdd_cons(R0, A, R1, S),
		zdd_join(L0, R0, Z, S)
	).

		/**************n***************************
		*    Interleave bddered atoms in BDD    *
		******************************************/

%!	bdd_interleave(+X, +Y, -Z, +S) is det.
%	Z = { an interleave of A and B | A in X, B in Y },
%	where a list M is an interleave of a list A and and a list B if
%	length(M) = length(A)+length(B), and both A and B are sublists of M.
%	A list U is a sublist of a list V if U is a subsequence of V,
%	provided that a list is viewed as a sequence.

% ?- #S, X<< +[*[a,b], *[x,y]],  Y<< +[*[1]],
%	bdd_interleave(X, Y, Z, S), psa(Z, S).
% ?- #S, X<< *[a], Y<< *[b], bdd_interleave(X, Y, Z, S), psa(Z, S).
% ?- #S, X<< *[a,b], Y<< *[1,2], bdd_interleave(X, Y, Z, S), psa(Z, S).

bdd_interleave(0, _, 0, _):-!.
bdd_interleave(_, 0, 0, _):-!.
bdd_interleave(1, X, X, _):-!.
bdd_interleave(X, 1, X, _):-!.
bdd_interleave(X, Y, Z, S):-	 % memo is useless here.
	cofact(Y, t(B, L, R), S),
	(	L < 2  ->  L0 = 0
	;	bdd_interleave(X, L, L0, S)
	),
	bdd_interleave(X, B, R, R0, S),
	zdd_join(L0, R0, Z, S).
%
bdd_interleave(0, _, _, 0, _):-!.
bdd_interleave(1, B, Y, Z, S):-!, cofact(Z, t(B, 0, Y), S).
bdd_interleave(X, B, Y, Z, S):- memo(merge(X, Y, B)-Z, S),
	(	nonvar(Z) -> true			% , write(+)			% many hits.
	;	cofact(X, t(A, L, R), S),
		(	L < 2 -> L0 = 0
		;	bdd_interleave(L, B, Y, L0, S)
		),
		bdd_interleave(A, R, B, Y, R0, S),
		zdd_join(L0, R0, Z, S)
	).
%
bdd_interleave(A, X, B, Y, Z, S):-  % memo is useless here.
	bdd_interleave(X, B, Y, U, S),
	cofact(Z0, t(A, 0, U), S),
	bdd_interleave(Y, A, X, V, S),
	cofact(Z1, t(B, 0, V), S),
	zdd_join(Z0, Z1, Z, S).

%!	bdd_funs(+PartExp, -F, +S) is det.
%	PartExp is an expressions built from mono(A-B), epi(A-B), fun(A-B),
%	and operator * (merge), & (interleave).
%	F is the function space in ZDD built from these parts.
%	mono(A-B) means function restricted to A is one-to-one mappings to B.
%	epi(A-B) one-to-onto, and fun(A-B) mapping.

% ?- #S, bdd_funs(mono([1,2]-[a,b])*mono([3,4]-[c,d]), X, S), psa(X, S).
% ?- #S, bdd_funs(mono([1,2]-[a,b])&mono([3,4]-[c,d]), X, S), psa(X, S).
% ?- #S, bdd_funs(mono([1,2]-[a,b])&mono([3,4]-[c,d]), X, S),
%	zdd_normalize(X, Y, S), psa(Y, S).

bdd_funs(A*B, F, S):-!, bdd_funs(A, A0, S),
	bdd_funs(B, B0, S),
	bdd_merge(A0, B0, F, S).
bdd_funs(A&B, F, S):-!, bdd_funs(A, A0, S),
	bdd_funs(B, B0, S),
	bdd_interleave(A0, B0, F, S).
bdd_funs(A, F, S):- fun_block(A, F, S).

		/**************
		*     misc    *
		**************/

%!	zdd_flatten(+X, -Y, +S) is det.
%	Y = { {A} | A in U, U in X }.

% ?- numlist(1, 100, Ns), zdd((X<<pow(Ns), shift(zdd_flatten(X, Y)), psa(Y))).
%  zdd_flatten(X, Y, S) :- Y = { A |  for some B in X, A in B }

zdd_flatten(X, 0, _):- X<2, !.
zdd_flatten(X, Y, S):- memo(flatten(X)-Y, S),
	(	nonvar(Y) -> true	% Many hits.
	;	cofact(X, t(A, L, R), S),
		zdd_flatten(L, L0, S),
		zdd_flatten(R, R0, S),
		zdd_join(L0, R0, Y0, S),
   		cofact(Y, t(A, Y0, 1), S)
	).
